export default function Landing() {
    return (
        <main className="flex-1 flex items-center justify-center">
            <h1 className="text-4xl md:text-6xl font-bold text-center">Welcome to XspecT</h1>
        </main>
    )
}