function F5(e,t){for(var n=0;n<t.length;n++){const r=t[n];if(typeof r!="string"&&!Array.isArray(r)){for(const i in r)if(i!=="default"&&!(i in e)){const a=Object.getOwnPropertyDescriptor(r,i);a&&Object.defineProperty(e,i,a.get?a:{enumerable:!0,get:()=>r[i]})}}}return Object.freeze(Object.defineProperty(e,Symbol.toStringTag,{value:"Module"}))}(function(){const t=document.createElement("link").relList;if(t&&t.supports&&t.supports("modulepreload"))return;for(const i of document.querySelectorAll('link[rel="modulepreload"]'))r(i);new MutationObserver(i=>{for(const a of i)if(a.type==="childList")for(const s of a.addedNodes)s.tagName==="LINK"&&s.rel==="modulepreload"&&r(s)}).observe(document,{childList:!0,subtree:!0});function n(i){const a={};return i.integrity&&(a.integrity=i.integrity),i.referrerPolicy&&(a.referrerPolicy=i.referrerPolicy),i.crossOrigin==="use-credentials"?a.credentials="include":i.crossOrigin==="anonymous"?a.credentials="omit":a.credentials="same-origin",a}function r(i){if(i.ep)return;i.ep=!0;const a=n(i);fetch(i.href,a)}})();var Up=typeof globalThis<"u"?globalThis:typeof window<"u"?window:typeof global<"u"?global:typeof self<"u"?self:{};function Rt(e){return e&&e.__esModule&&Object.prototype.hasOwnProperty.call(e,"default")?e.default:e}var Rb={exports:{}},bc={};/**
 * @license React
 * react-jsx-runtime.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var RM;function q5(){if(RM)return bc;RM=1;var e=Symbol.for("react.transitional.element"),t=Symbol.for("react.fragment");function n(r,i,a){var s=null;if(a!==void 0&&(s=""+a),i.key!==void 0&&(s=""+i.key),"key"in i){a={};for(var u in i)u!=="key"&&(a[u]=i[u])}else a=i;return i=a.ref,{$$typeof:e,type:r,key:s,ref:i!==void 0?i:null,props:a}}return bc.Fragment=t,bc.jsx=n,bc.jsxs=n,bc}var TM;function z5(){return TM||(TM=1,Rb.exports=q5()),Rb.exports}var E=z5(),Tb={exports:{}},Qe={};/**
 * @license React
 * react.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var MM;function U5(){if(MM)return Qe;MM=1;var e=Symbol.for("react.transitional.element"),t=Symbol.for("react.portal"),n=Symbol.for("react.fragment"),r=Symbol.for("react.strict_mode"),i=Symbol.for("react.profiler"),a=Symbol.for("react.consumer"),s=Symbol.for("react.context"),u=Symbol.for("react.forward_ref"),c=Symbol.for("react.suspense"),d=Symbol.for("react.memo"),p=Symbol.for("react.lazy"),m=Symbol.iterator;function v($){return $===null||typeof $!="object"?null:($=m&&$[m]||$["@@iterator"],typeof $=="function"?$:null)}var b={isMounted:function(){return!1},enqueueForceUpdate:function(){},enqueueReplaceState:function(){},enqueueSetState:function(){}},S=Object.assign,w={};function x($,J,q){this.props=$,this.context=J,this.refs=w,this.updater=q||b}x.prototype.isReactComponent={},x.prototype.setState=function($,J){if(typeof $!="object"&&typeof $!="function"&&$!=null)throw Error("takes an object of state variables to update or a function which returns an object of state variables.");this.updater.enqueueSetState(this,$,J,"setState")},x.prototype.forceUpdate=function($){this.updater.enqueueForceUpdate(this,$,"forceUpdate")};function C(){}C.prototype=x.prototype;function O($,J,q){this.props=$,this.context=J,this.refs=w,this.updater=q||b}var A=O.prototype=new C;A.constructor=O,S(A,x.prototype),A.isPureReactComponent=!0;var T=Array.isArray,R={H:null,A:null,T:null,S:null,V:null},M=Object.prototype.hasOwnProperty;function P($,J,q,X,oe,le){return q=le.ref,{$$typeof:e,type:$,key:J,ref:q!==void 0?q:null,props:le}}function N($,J){return P($.type,J,void 0,void 0,void 0,$.props)}function k($){return typeof $=="object"&&$!==null&&$.$$typeof===e}function H($){var J={"=":"=0",":":"=2"};return"$"+$.replace(/[=:]/g,function(q){return J[q]})}var L=/\/+/g;function F($,J){return typeof $=="object"&&$!==null&&$.key!=null?H(""+$.key):J.toString(36)}function z(){}function Z($){switch($.status){case"fulfilled":return $.value;case"rejected":throw $.reason;default:switch(typeof $.status=="string"?$.then(z,z):($.status="pending",$.then(function(J){$.status==="pending"&&($.status="fulfilled",$.value=J)},function(J){$.status==="pending"&&($.status="rejected",$.reason=J)})),$.status){case"fulfilled":return $.value;case"rejected":throw $.reason}}throw $}function Y($,J,q,X,oe){var le=typeof $;(le==="undefined"||le==="boolean")&&($=null);var ve=!1;if($===null)ve=!0;else switch(le){case"bigint":case"string":case"number":ve=!0;break;case"object":switch($.$$typeof){case e:case t:ve=!0;break;case p:return ve=$._init,Y(ve($._payload),J,q,X,oe)}}if(ve)return oe=oe($),ve=X===""?"."+F($,0):X,T(oe)?(q="",ve!=null&&(q=ve.replace(L,"$&/")+"/"),Y(oe,J,q,"",function(ce){return ce})):oe!=null&&(k(oe)&&(oe=N(oe,q+(oe.key==null||$&&$.key===oe.key?"":(""+oe.key).replace(L,"$&/")+"/")+ve)),J.push(oe)),1;ve=0;var de=X===""?".":X+":";if(T($))for(var Se=0;Se<$.length;Se++)X=$[Se],le=de+F(X,Se),ve+=Y(X,J,q,le,oe);else if(Se=v($),typeof Se=="function")for($=Se.call($),Se=0;!(X=$.next()).done;)X=X.value,le=de+F(X,Se++),ve+=Y(X,J,q,le,oe);else if(le==="object"){if(typeof $.then=="function")return Y(Z($),J,q,X,oe);throw J=String($),Error("Objects are not valid as a React child (found: "+(J==="[object Object]"?"object with keys {"+Object.keys($).join(", ")+"}":J)+"). If you meant to render a collection of children, use an array instead.")}return ve}function I($,J,q){if($==null)return $;var X=[],oe=0;return Y($,X,"","",function(le){return J.call(q,le,oe++)}),X}function W($){if($._status===-1){var J=$._result;J=J(),J.then(function(q){($._status===0||$._status===-1)&&($._status=1,$._result=q)},function(q){($._status===0||$._status===-1)&&($._status=2,$._result=q)}),$._status===-1&&($._status=0,$._result=J)}if($._status===1)return $._result.default;throw $._result}var B=typeof reportError=="function"?reportError:function($){if(typeof window=="object"&&typeof window.ErrorEvent=="function"){var J=new window.ErrorEvent("error",{bubbles:!0,cancelable:!0,message:typeof $=="object"&&$!==null&&typeof $.message=="string"?String($.message):String($),error:$});if(!window.dispatchEvent(J))return}else if(typeof process=="object"&&typeof process.emit=="function"){process.emit("uncaughtException",$);return}console.error($)};function Q(){}return Qe.Children={map:I,forEach:function($,J,q){I($,function(){J.apply(this,arguments)},q)},count:function($){var J=0;return I($,function(){J++}),J},toArray:function($){return I($,function(J){return J})||[]},only:function($){if(!k($))throw Error("React.Children.only expected to receive a single React element child.");return $}},Qe.Component=x,Qe.Fragment=n,Qe.Profiler=i,Qe.PureComponent=O,Qe.StrictMode=r,Qe.Suspense=c,Qe.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE=R,Qe.__COMPILER_RUNTIME={__proto__:null,c:function($){return R.H.useMemoCache($)}},Qe.cache=function($){return function(){return $.apply(null,arguments)}},Qe.cloneElement=function($,J,q){if($==null)throw Error("The argument must be a React element, but you passed "+$+".");var X=S({},$.props),oe=$.key,le=void 0;if(J!=null)for(ve in J.ref!==void 0&&(le=void 0),J.key!==void 0&&(oe=""+J.key),J)!M.call(J,ve)||ve==="key"||ve==="__self"||ve==="__source"||ve==="ref"&&J.ref===void 0||(X[ve]=J[ve]);var ve=arguments.length-2;if(ve===1)X.children=q;else if(1<ve){for(var de=Array(ve),Se=0;Se<ve;Se++)de[Se]=arguments[Se+2];X.children=de}return P($.type,oe,void 0,void 0,le,X)},Qe.createContext=function($){return $={$$typeof:s,_currentValue:$,_currentValue2:$,_threadCount:0,Provider:null,Consumer:null},$.Provider=$,$.Consumer={$$typeof:a,_context:$},$},Qe.createElement=function($,J,q){var X,oe={},le=null;if(J!=null)for(X in J.key!==void 0&&(le=""+J.key),J)M.call(J,X)&&X!=="key"&&X!=="__self"&&X!=="__source"&&(oe[X]=J[X]);var ve=arguments.length-2;if(ve===1)oe.children=q;else if(1<ve){for(var de=Array(ve),Se=0;Se<ve;Se++)de[Se]=arguments[Se+2];oe.children=de}if($&&$.defaultProps)for(X in ve=$.defaultProps,ve)oe[X]===void 0&&(oe[X]=ve[X]);return P($,le,void 0,void 0,null,oe)},Qe.createRef=function(){return{current:null}},Qe.forwardRef=function($){return{$$typeof:u,render:$}},Qe.isValidElement=k,Qe.lazy=function($){return{$$typeof:p,_payload:{_status:-1,_result:$},_init:W}},Qe.memo=function($,J){return{$$typeof:d,type:$,compare:J===void 0?null:J}},Qe.startTransition=function($){var J=R.T,q={};R.T=q;try{var X=$(),oe=R.S;oe!==null&&oe(q,X),typeof X=="object"&&X!==null&&typeof X.then=="function"&&X.then(Q,B)}catch(le){B(le)}finally{R.T=J}},Qe.unstable_useCacheRefresh=function(){return R.H.useCacheRefresh()},Qe.use=function($){return R.H.use($)},Qe.useActionState=function($,J,q){return R.H.useActionState($,J,q)},Qe.useCallback=function($,J){return R.H.useCallback($,J)},Qe.useContext=function($){return R.H.useContext($)},Qe.useDebugValue=function(){},Qe.useDeferredValue=function($,J){return R.H.useDeferredValue($,J)},Qe.useEffect=function($,J,q){var X=R.H;if(typeof q=="function")throw Error("useEffect CRUD overload is not enabled in this build of React.");return X.useEffect($,J)},Qe.useId=function(){return R.H.useId()},Qe.useImperativeHandle=function($,J,q){return R.H.useImperativeHandle($,J,q)},Qe.useInsertionEffect=function($,J){return R.H.useInsertionEffect($,J)},Qe.useLayoutEffect=function($,J){return R.H.useLayoutEffect($,J)},Qe.useMemo=function($,J){return R.H.useMemo($,J)},Qe.useOptimistic=function($,J){return R.H.useOptimistic($,J)},Qe.useReducer=function($,J,q){return R.H.useReducer($,J,q)},Qe.useRef=function($){return R.H.useRef($)},Qe.useState=function($){return R.H.useState($)},Qe.useSyncExternalStore=function($,J,q){return R.H.useSyncExternalStore($,J,q)},Qe.useTransition=function(){return R.H.useTransition()},Qe.version="19.1.0",Qe}var PM;function CE(){return PM||(PM=1,Tb.exports=U5()),Tb.exports}var g=CE();const V=Rt(g),ZI=F5({__proto__:null,default:V},[g]);var Mb={exports:{}},xc={},Pb={exports:{}},jb={};/**
 * @license React
 * scheduler.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var jM;function V5(){return jM||(jM=1,function(e){function t(I,W){var B=I.length;I.push(W);e:for(;0<B;){var Q=B-1>>>1,$=I[Q];if(0<i($,W))I[Q]=W,I[B]=$,B=Q;else break e}}function n(I){return I.length===0?null:I[0]}function r(I){if(I.length===0)return null;var W=I[0],B=I.pop();if(B!==W){I[0]=B;e:for(var Q=0,$=I.length,J=$>>>1;Q<J;){var q=2*(Q+1)-1,X=I[q],oe=q+1,le=I[oe];if(0>i(X,B))oe<$&&0>i(le,X)?(I[Q]=le,I[oe]=B,Q=oe):(I[Q]=X,I[q]=B,Q=q);else if(oe<$&&0>i(le,B))I[Q]=le,I[oe]=B,Q=oe;else break e}}return W}function i(I,W){var B=I.sortIndex-W.sortIndex;return B!==0?B:I.id-W.id}if(e.unstable_now=void 0,typeof performance=="object"&&typeof performance.now=="function"){var a=performance;e.unstable_now=function(){return a.now()}}else{var s=Date,u=s.now();e.unstable_now=function(){return s.now()-u}}var c=[],d=[],p=1,m=null,v=3,b=!1,S=!1,w=!1,x=!1,C=typeof setTimeout=="function"?setTimeout:null,O=typeof clearTimeout=="function"?clearTimeout:null,A=typeof setImmediate<"u"?setImmediate:null;function T(I){for(var W=n(d);W!==null;){if(W.callback===null)r(d);else if(W.startTime<=I)r(d),W.sortIndex=W.expirationTime,t(c,W);else break;W=n(d)}}function R(I){if(w=!1,T(I),!S)if(n(c)!==null)S=!0,M||(M=!0,F());else{var W=n(d);W!==null&&Y(R,W.startTime-I)}}var M=!1,P=-1,N=5,k=-1;function H(){return x?!0:!(e.unstable_now()-k<N)}function L(){if(x=!1,M){var I=e.unstable_now();k=I;var W=!0;try{e:{S=!1,w&&(w=!1,O(P),P=-1),b=!0;var B=v;try{t:{for(T(I),m=n(c);m!==null&&!(m.expirationTime>I&&H());){var Q=m.callback;if(typeof Q=="function"){m.callback=null,v=m.priorityLevel;var $=Q(m.expirationTime<=I);if(I=e.unstable_now(),typeof $=="function"){m.callback=$,T(I),W=!0;break t}m===n(c)&&r(c),T(I)}else r(c);m=n(c)}if(m!==null)W=!0;else{var J=n(d);J!==null&&Y(R,J.startTime-I),W=!1}}break e}finally{m=null,v=B,b=!1}W=void 0}}finally{W?F():M=!1}}}var F;if(typeof A=="function")F=function(){A(L)};else if(typeof MessageChannel<"u"){var z=new MessageChannel,Z=z.port2;z.port1.onmessage=L,F=function(){Z.postMessage(null)}}else F=function(){C(L,0)};function Y(I,W){P=C(function(){I(e.unstable_now())},W)}e.unstable_IdlePriority=5,e.unstable_ImmediatePriority=1,e.unstable_LowPriority=4,e.unstable_NormalPriority=3,e.unstable_Profiling=null,e.unstable_UserBlockingPriority=2,e.unstable_cancelCallback=function(I){I.callback=null},e.unstable_forceFrameRate=function(I){0>I||125<I?console.error("forceFrameRate takes a positive int between 0 and 125, forcing frame rates higher than 125 fps is not supported"):N=0<I?Math.floor(1e3/I):5},e.unstable_getCurrentPriorityLevel=function(){return v},e.unstable_next=function(I){switch(v){case 1:case 2:case 3:var W=3;break;default:W=v}var B=v;v=W;try{return I()}finally{v=B}},e.unstable_requestPaint=function(){x=!0},e.unstable_runWithPriority=function(I,W){switch(I){case 1:case 2:case 3:case 4:case 5:break;default:I=3}var B=v;v=I;try{return W()}finally{v=B}},e.unstable_scheduleCallback=function(I,W,B){var Q=e.unstable_now();switch(typeof B=="object"&&B!==null?(B=B.delay,B=typeof B=="number"&&0<B?Q+B:Q):B=Q,I){case 1:var $=-1;break;case 2:$=250;break;case 5:$=1073741823;break;case 4:$=1e4;break;default:$=5e3}return $=B+$,I={id:p++,callback:W,priorityLevel:I,startTime:B,expirationTime:$,sortIndex:-1},B>Q?(I.sortIndex=B,t(d,I),n(c)===null&&I===n(d)&&(w?(O(P),P=-1):w=!0,Y(R,B-Q))):(I.sortIndex=$,t(c,I),S||b||(S=!0,M||(M=!0,F()))),I},e.unstable_shouldYield=H,e.unstable_wrapCallback=function(I){var W=v;return function(){var B=v;v=W;try{return I.apply(this,arguments)}finally{v=B}}}}(jb)),jb}var NM;function H5(){return NM||(NM=1,Pb.exports=V5()),Pb.exports}var Nb={exports:{}},An={};/**
 * @license React
 * react-dom.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var DM;function G5(){if(DM)return An;DM=1;var e=CE();function t(c){var d="https://react.dev/errors/"+c;if(1<arguments.length){d+="?args[]="+encodeURIComponent(arguments[1]);for(var p=2;p<arguments.length;p++)d+="&args[]="+encodeURIComponent(arguments[p])}return"Minified React error #"+c+"; visit "+d+" for the full message or use the non-minified dev environment for full errors and additional helpful warnings."}function n(){}var r={d:{f:n,r:function(){throw Error(t(522))},D:n,C:n,L:n,m:n,X:n,S:n,M:n},p:0,findDOMNode:null},i=Symbol.for("react.portal");function a(c,d,p){var m=3<arguments.length&&arguments[3]!==void 0?arguments[3]:null;return{$$typeof:i,key:m==null?null:""+m,children:c,containerInfo:d,implementation:p}}var s=e.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE;function u(c,d){if(c==="font")return"";if(typeof d=="string")return d==="use-credentials"?d:""}return An.__DOM_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE=r,An.createPortal=function(c,d){var p=2<arguments.length&&arguments[2]!==void 0?arguments[2]:null;if(!d||d.nodeType!==1&&d.nodeType!==9&&d.nodeType!==11)throw Error(t(299));return a(c,d,null,p)},An.flushSync=function(c){var d=s.T,p=r.p;try{if(s.T=null,r.p=2,c)return c()}finally{s.T=d,r.p=p,r.d.f()}},An.preconnect=function(c,d){typeof c=="string"&&(d?(d=d.crossOrigin,d=typeof d=="string"?d==="use-credentials"?d:"":void 0):d=null,r.d.C(c,d))},An.prefetchDNS=function(c){typeof c=="string"&&r.d.D(c)},An.preinit=function(c,d){if(typeof c=="string"&&d&&typeof d.as=="string"){var p=d.as,m=u(p,d.crossOrigin),v=typeof d.integrity=="string"?d.integrity:void 0,b=typeof d.fetchPriority=="string"?d.fetchPriority:void 0;p==="style"?r.d.S(c,typeof d.precedence=="string"?d.precedence:void 0,{crossOrigin:m,integrity:v,fetchPriority:b}):p==="script"&&r.d.X(c,{crossOrigin:m,integrity:v,fetchPriority:b,nonce:typeof d.nonce=="string"?d.nonce:void 0})}},An.preinitModule=function(c,d){if(typeof c=="string")if(typeof d=="object"&&d!==null){if(d.as==null||d.as==="script"){var p=u(d.as,d.crossOrigin);r.d.M(c,{crossOrigin:p,integrity:typeof d.integrity=="string"?d.integrity:void 0,nonce:typeof d.nonce=="string"?d.nonce:void 0})}}else d==null&&r.d.M(c)},An.preload=function(c,d){if(typeof c=="string"&&typeof d=="object"&&d!==null&&typeof d.as=="string"){var p=d.as,m=u(p,d.crossOrigin);r.d.L(c,p,{crossOrigin:m,integrity:typeof d.integrity=="string"?d.integrity:void 0,nonce:typeof d.nonce=="string"?d.nonce:void 0,type:typeof d.type=="string"?d.type:void 0,fetchPriority:typeof d.fetchPriority=="string"?d.fetchPriority:void 0,referrerPolicy:typeof d.referrerPolicy=="string"?d.referrerPolicy:void 0,imageSrcSet:typeof d.imageSrcSet=="string"?d.imageSrcSet:void 0,imageSizes:typeof d.imageSizes=="string"?d.imageSizes:void 0,media:typeof d.media=="string"?d.media:void 0})}},An.preloadModule=function(c,d){if(typeof c=="string")if(d){var p=u(d.as,d.crossOrigin);r.d.m(c,{as:typeof d.as=="string"&&d.as!=="script"?d.as:void 0,crossOrigin:p,integrity:typeof d.integrity=="string"?d.integrity:void 0})}else r.d.m(c)},An.requestFormReset=function(c){r.d.r(c)},An.unstable_batchedUpdates=function(c,d){return c(d)},An.useFormState=function(c,d,p){return s.H.useFormState(c,d,p)},An.useFormStatus=function(){return s.H.useHostTransitionStatus()},An.version="19.1.0",An}var kM;function XI(){if(kM)return Nb.exports;kM=1;function e(){if(!(typeof __REACT_DEVTOOLS_GLOBAL_HOOK__>"u"||typeof __REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE!="function"))try{__REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE(e)}catch(t){console.error(t)}}return e(),Nb.exports=G5(),Nb.exports}/**
 * @license React
 * react-dom-client.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var $M;function W5(){if($M)return xc;$M=1;var e=H5(),t=CE(),n=XI();function r(o){var l="https://react.dev/errors/"+o;if(1<arguments.length){l+="?args[]="+encodeURIComponent(arguments[1]);for(var f=2;f<arguments.length;f++)l+="&args[]="+encodeURIComponent(arguments[f])}return"Minified React error #"+o+"; visit "+l+" for the full message or use the non-minified dev environment for full errors and additional helpful warnings."}function i(o){return!(!o||o.nodeType!==1&&o.nodeType!==9&&o.nodeType!==11)}function a(o){var l=o,f=o;if(o.alternate)for(;l.return;)l=l.return;else{o=l;do l=o,(l.flags&4098)!==0&&(f=l.return),o=l.return;while(o)}return l.tag===3?f:null}function s(o){if(o.tag===13){var l=o.memoizedState;if(l===null&&(o=o.alternate,o!==null&&(l=o.memoizedState)),l!==null)return l.dehydrated}return null}function u(o){if(a(o)!==o)throw Error(r(188))}function c(o){var l=o.alternate;if(!l){if(l=a(o),l===null)throw Error(r(188));return l!==o?null:o}for(var f=o,h=l;;){var y=f.return;if(y===null)break;var _=y.alternate;if(_===null){if(h=y.return,h!==null){f=h;continue}break}if(y.child===_.child){for(_=y.child;_;){if(_===f)return u(y),o;if(_===h)return u(y),l;_=_.sibling}throw Error(r(188))}if(f.return!==h.return)f=y,h=_;else{for(var j=!1,D=y.child;D;){if(D===f){j=!0,f=y,h=_;break}if(D===h){j=!0,h=y,f=_;break}D=D.sibling}if(!j){for(D=_.child;D;){if(D===f){j=!0,f=_,h=y;break}if(D===h){j=!0,h=_,f=y;break}D=D.sibling}if(!j)throw Error(r(189))}}if(f.alternate!==h)throw Error(r(190))}if(f.tag!==3)throw Error(r(188));return f.stateNode.current===f?o:l}function d(o){var l=o.tag;if(l===5||l===26||l===27||l===6)return o;for(o=o.child;o!==null;){if(l=d(o),l!==null)return l;o=o.sibling}return null}var p=Object.assign,m=Symbol.for("react.element"),v=Symbol.for("react.transitional.element"),b=Symbol.for("react.portal"),S=Symbol.for("react.fragment"),w=Symbol.for("react.strict_mode"),x=Symbol.for("react.profiler"),C=Symbol.for("react.provider"),O=Symbol.for("react.consumer"),A=Symbol.for("react.context"),T=Symbol.for("react.forward_ref"),R=Symbol.for("react.suspense"),M=Symbol.for("react.suspense_list"),P=Symbol.for("react.memo"),N=Symbol.for("react.lazy"),k=Symbol.for("react.activity"),H=Symbol.for("react.memo_cache_sentinel"),L=Symbol.iterator;function F(o){return o===null||typeof o!="object"?null:(o=L&&o[L]||o["@@iterator"],typeof o=="function"?o:null)}var z=Symbol.for("react.client.reference");function Z(o){if(o==null)return null;if(typeof o=="function")return o.$$typeof===z?null:o.displayName||o.name||null;if(typeof o=="string")return o;switch(o){case S:return"Fragment";case x:return"Profiler";case w:return"StrictMode";case R:return"Suspense";case M:return"SuspenseList";case k:return"Activity"}if(typeof o=="object")switch(o.$$typeof){case b:return"Portal";case A:return(o.displayName||"Context")+".Provider";case O:return(o._context.displayName||"Context")+".Consumer";case T:var l=o.render;return o=o.displayName,o||(o=l.displayName||l.name||"",o=o!==""?"ForwardRef("+o+")":"ForwardRef"),o;case P:return l=o.displayName||null,l!==null?l:Z(o.type)||"Memo";case N:l=o._payload,o=o._init;try{return Z(o(l))}catch{}}return null}var Y=Array.isArray,I=t.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE,W=n.__DOM_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE,B={pending:!1,data:null,method:null,action:null},Q=[],$=-1;function J(o){return{current:o}}function q(o){0>$||(o.current=Q[$],Q[$]=null,$--)}function X(o,l){$++,Q[$]=o.current,o.current=l}var oe=J(null),le=J(null),ve=J(null),de=J(null);function Se(o,l){switch(X(ve,l),X(le,o),X(oe,null),l.nodeType){case 9:case 11:o=(o=l.documentElement)&&(o=o.namespaceURI)?nM(o):0;break;default:if(o=l.tagName,l=l.namespaceURI)l=nM(l),o=rM(l,o);else switch(o){case"svg":o=1;break;case"math":o=2;break;default:o=0}}q(oe),X(oe,o)}function ce(){q(oe),q(le),q(ve)}function we(o){o.memoizedState!==null&&X(de,o);var l=oe.current,f=rM(l,o.type);l!==f&&(X(le,o),X(oe,f))}function _e(o){le.current===o&&(q(oe),q(le)),de.current===o&&(q(de),hc._currentValue=B)}var se=Object.prototype.hasOwnProperty,Ye=e.unstable_scheduleCallback,Be=e.unstable_cancelCallback,dt=e.unstable_shouldYield,ft=e.unstable_requestPaint,it=e.unstable_now,xt=e.unstable_getCurrentPriorityLevel,Tt=e.unstable_ImmediatePriority,bn=e.unstable_UserBlockingPriority,Bt=e.unstable_NormalPriority,G=e.unstable_LowPriority,re=e.unstable_IdlePriority,fe=e.log,Ae=e.unstable_setDisableYieldValue,Ee=null,ye=null;function Te(o){if(typeof fe=="function"&&Ae(o),ye&&typeof ye.setStrictMode=="function")try{ye.setStrictMode(Ee,o)}catch{}}var Ke=Math.clz32?Math.clz32:vg,Vt=Math.log,dn=Math.LN2;function vg(o){return o>>>=0,o===0?32:31-(Vt(o)/dn|0)|0}var _a=256,Ea=4194304;function ji(o){var l=o&42;if(l!==0)return l;switch(o&-o){case 1:return 1;case 2:return 2;case 4:return 4;case 8:return 8;case 16:return 16;case 32:return 32;case 64:return 64;case 128:return 128;case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:return o&4194048;case 4194304:case 8388608:case 16777216:case 33554432:return o&62914560;case 67108864:return 67108864;case 134217728:return 134217728;case 268435456:return 268435456;case 536870912:return 536870912;case 1073741824:return 0;default:return o}}function bo(o,l,f){var h=o.pendingLanes;if(h===0)return 0;var y=0,_=o.suspendedLanes,j=o.pingedLanes;o=o.warmLanes;var D=h&134217727;return D!==0?(h=D&~_,h!==0?y=ji(h):(j&=D,j!==0?y=ji(j):f||(f=D&~o,f!==0&&(y=ji(f))))):(D=h&~_,D!==0?y=ji(D):j!==0?y=ji(j):f||(f=h&~o,f!==0&&(y=ji(f)))),y===0?0:l!==0&&l!==y&&(l&_)===0&&(_=y&-y,f=l&-l,_>=f||_===32&&(f&4194048)!==0)?l:y}function xo(o,l){return(o.pendingLanes&~(o.suspendedLanes&~o.pingedLanes)&l)===0}function jd(o,l){switch(o){case 1:case 2:case 4:case 8:case 64:return l+250;case 16:case 32:case 128:case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:return l+5e3;case 4194304:case 8388608:case 16777216:case 33554432:return-1;case 67108864:case 134217728:case 268435456:case 536870912:case 1073741824:return-1;default:return-1}}function BO(){var o=_a;return _a<<=1,(_a&4194048)===0&&(_a=256),o}function FO(){var o=Ea;return Ea<<=1,(Ea&62914560)===0&&(Ea=4194304),o}function gg(o){for(var l=[],f=0;31>f;f++)l.push(o);return l}function Eu(o,l){o.pendingLanes|=l,l!==268435456&&(o.suspendedLanes=0,o.pingedLanes=0,o.warmLanes=0)}function R4(o,l,f,h,y,_){var j=o.pendingLanes;o.pendingLanes=f,o.suspendedLanes=0,o.pingedLanes=0,o.warmLanes=0,o.expiredLanes&=f,o.entangledLanes&=f,o.errorRecoveryDisabledLanes&=f,o.shellSuspendCounter=0;var D=o.entanglements,U=o.expirationTimes,ne=o.hiddenUpdates;for(f=j&~f;0<f;){var he=31-Ke(f),ge=1<<he;D[he]=0,U[he]=-1;var ie=ne[he];if(ie!==null)for(ne[he]=null,he=0;he<ie.length;he++){var ae=ie[he];ae!==null&&(ae.lane&=-536870913)}f&=~ge}h!==0&&qO(o,h,0),_!==0&&y===0&&o.tag!==0&&(o.suspendedLanes|=_&~(j&~l))}function qO(o,l,f){o.pendingLanes|=l,o.suspendedLanes&=~l;var h=31-Ke(l);o.entangledLanes|=l,o.entanglements[h]=o.entanglements[h]|1073741824|f&4194090}function zO(o,l){var f=o.entangledLanes|=l;for(o=o.entanglements;f;){var h=31-Ke(f),y=1<<h;y&l|o[h]&l&&(o[h]|=l),f&=~y}}function yg(o){switch(o){case 2:o=1;break;case 8:o=4;break;case 32:o=16;break;case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:case 4194304:case 8388608:case 16777216:case 33554432:o=128;break;case 268435456:o=134217728;break;default:o=0}return o}function bg(o){return o&=-o,2<o?8<o?(o&134217727)!==0?32:268435456:8:2}function UO(){var o=W.p;return o!==0?o:(o=window.event,o===void 0?32:SM(o.type))}function T4(o,l){var f=W.p;try{return W.p=o,l()}finally{W.p=f}}var Ca=Math.random().toString(36).slice(2),Cn="__reactFiber$"+Ca,Vn="__reactProps$"+Ca,Es="__reactContainer$"+Ca,xg="__reactEvents$"+Ca,M4="__reactListeners$"+Ca,P4="__reactHandles$"+Ca,VO="__reactResources$"+Ca,Cu="__reactMarker$"+Ca;function wg(o){delete o[Cn],delete o[Vn],delete o[xg],delete o[M4],delete o[P4]}function Cs(o){var l=o[Cn];if(l)return l;for(var f=o.parentNode;f;){if(l=f[Es]||f[Cn]){if(f=l.alternate,l.child!==null||f!==null&&f.child!==null)for(o=sM(o);o!==null;){if(f=o[Cn])return f;o=sM(o)}return l}o=f,f=o.parentNode}return null}function Os(o){if(o=o[Cn]||o[Es]){var l=o.tag;if(l===5||l===6||l===13||l===26||l===27||l===3)return o}return null}function Ou(o){var l=o.tag;if(l===5||l===26||l===27||l===6)return o.stateNode;throw Error(r(33))}function As(o){var l=o[VO];return l||(l=o[VO]={hoistableStyles:new Map,hoistableScripts:new Map}),l}function pn(o){o[Cu]=!0}var HO=new Set,GO={};function wo(o,l){Rs(o,l),Rs(o+"Capture",l)}function Rs(o,l){for(GO[o]=l,o=0;o<l.length;o++)HO.add(l[o])}var j4=RegExp("^[:A-Z_a-z\\u00C0-\\u00D6\\u00D8-\\u00F6\\u00F8-\\u02FF\\u0370-\\u037D\\u037F-\\u1FFF\\u200C-\\u200D\\u2070-\\u218F\\u2C00-\\u2FEF\\u3001-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFFD][:A-Z_a-z\\u00C0-\\u00D6\\u00D8-\\u00F6\\u00F8-\\u02FF\\u0370-\\u037D\\u037F-\\u1FFF\\u200C-\\u200D\\u2070-\\u218F\\u2C00-\\u2FEF\\u3001-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFFD\\-.0-9\\u00B7\\u0300-\\u036F\\u203F-\\u2040]*$"),WO={},KO={};function N4(o){return se.call(KO,o)?!0:se.call(WO,o)?!1:j4.test(o)?KO[o]=!0:(WO[o]=!0,!1)}function Nd(o,l,f){if(N4(l))if(f===null)o.removeAttribute(l);else{switch(typeof f){case"undefined":case"function":case"symbol":o.removeAttribute(l);return;case"boolean":var h=l.toLowerCase().slice(0,5);if(h!=="data-"&&h!=="aria-"){o.removeAttribute(l);return}}o.setAttribute(l,""+f)}}function Dd(o,l,f){if(f===null)o.removeAttribute(l);else{switch(typeof f){case"undefined":case"function":case"symbol":case"boolean":o.removeAttribute(l);return}o.setAttribute(l,""+f)}}function Ni(o,l,f,h){if(h===null)o.removeAttribute(f);else{switch(typeof h){case"undefined":case"function":case"symbol":case"boolean":o.removeAttribute(f);return}o.setAttributeNS(l,f,""+h)}}var Sg,YO;function Ts(o){if(Sg===void 0)try{throw Error()}catch(f){var l=f.stack.trim().match(/\n( *(at )?)/);Sg=l&&l[1]||"",YO=-1<f.stack.indexOf(`
    at`)?" (<anonymous>)":-1<f.stack.indexOf("@")?"@unknown:0:0":""}return`
`+Sg+o+YO}var _g=!1;function Eg(o,l){if(!o||_g)return"";_g=!0;var f=Error.prepareStackTrace;Error.prepareStackTrace=void 0;try{var h={DetermineComponentFrameRoot:function(){try{if(l){var ge=function(){throw Error()};if(Object.defineProperty(ge.prototype,"props",{set:function(){throw Error()}}),typeof Reflect=="object"&&Reflect.construct){try{Reflect.construct(ge,[])}catch(ae){var ie=ae}Reflect.construct(o,[],ge)}else{try{ge.call()}catch(ae){ie=ae}o.call(ge.prototype)}}else{try{throw Error()}catch(ae){ie=ae}(ge=o())&&typeof ge.catch=="function"&&ge.catch(function(){})}}catch(ae){if(ae&&ie&&typeof ae.stack=="string")return[ae.stack,ie.stack]}return[null,null]}};h.DetermineComponentFrameRoot.displayName="DetermineComponentFrameRoot";var y=Object.getOwnPropertyDescriptor(h.DetermineComponentFrameRoot,"name");y&&y.configurable&&Object.defineProperty(h.DetermineComponentFrameRoot,"name",{value:"DetermineComponentFrameRoot"});var _=h.DetermineComponentFrameRoot(),j=_[0],D=_[1];if(j&&D){var U=j.split(`
`),ne=D.split(`
`);for(y=h=0;h<U.length&&!U[h].includes("DetermineComponentFrameRoot");)h++;for(;y<ne.length&&!ne[y].includes("DetermineComponentFrameRoot");)y++;if(h===U.length||y===ne.length)for(h=U.length-1,y=ne.length-1;1<=h&&0<=y&&U[h]!==ne[y];)y--;for(;1<=h&&0<=y;h--,y--)if(U[h]!==ne[y]){if(h!==1||y!==1)do if(h--,y--,0>y||U[h]!==ne[y]){var he=`
`+U[h].replace(" at new "," at ");return o.displayName&&he.includes("<anonymous>")&&(he=he.replace("<anonymous>",o.displayName)),he}while(1<=h&&0<=y);break}}}finally{_g=!1,Error.prepareStackTrace=f}return(f=o?o.displayName||o.name:"")?Ts(f):""}function D4(o){switch(o.tag){case 26:case 27:case 5:return Ts(o.type);case 16:return Ts("Lazy");case 13:return Ts("Suspense");case 19:return Ts("SuspenseList");case 0:case 15:return Eg(o.type,!1);case 11:return Eg(o.type.render,!1);case 1:return Eg(o.type,!0);case 31:return Ts("Activity");default:return""}}function ZO(o){try{var l="";do l+=D4(o),o=o.return;while(o);return l}catch(f){return`
Error generating stack: `+f.message+`
`+f.stack}}function yr(o){switch(typeof o){case"bigint":case"boolean":case"number":case"string":case"undefined":return o;case"object":return o;default:return""}}function XO(o){var l=o.type;return(o=o.nodeName)&&o.toLowerCase()==="input"&&(l==="checkbox"||l==="radio")}function k4(o){var l=XO(o)?"checked":"value",f=Object.getOwnPropertyDescriptor(o.constructor.prototype,l),h=""+o[l];if(!o.hasOwnProperty(l)&&typeof f<"u"&&typeof f.get=="function"&&typeof f.set=="function"){var y=f.get,_=f.set;return Object.defineProperty(o,l,{configurable:!0,get:function(){return y.call(this)},set:function(j){h=""+j,_.call(this,j)}}),Object.defineProperty(o,l,{enumerable:f.enumerable}),{getValue:function(){return h},setValue:function(j){h=""+j},stopTracking:function(){o._valueTracker=null,delete o[l]}}}}function kd(o){o._valueTracker||(o._valueTracker=k4(o))}function QO(o){if(!o)return!1;var l=o._valueTracker;if(!l)return!0;var f=l.getValue(),h="";return o&&(h=XO(o)?o.checked?"true":"false":o.value),o=h,o!==f?(l.setValue(o),!0):!1}function $d(o){if(o=o||(typeof document<"u"?document:void 0),typeof o>"u")return null;try{return o.activeElement||o.body}catch{return o.body}}var $4=/[\n"\\]/g;function br(o){return o.replace($4,function(l){return"\\"+l.charCodeAt(0).toString(16)+" "})}function Cg(o,l,f,h,y,_,j,D){o.name="",j!=null&&typeof j!="function"&&typeof j!="symbol"&&typeof j!="boolean"?o.type=j:o.removeAttribute("type"),l!=null?j==="number"?(l===0&&o.value===""||o.value!=l)&&(o.value=""+yr(l)):o.value!==""+yr(l)&&(o.value=""+yr(l)):j!=="submit"&&j!=="reset"||o.removeAttribute("value"),l!=null?Og(o,j,yr(l)):f!=null?Og(o,j,yr(f)):h!=null&&o.removeAttribute("value"),y==null&&_!=null&&(o.defaultChecked=!!_),y!=null&&(o.checked=y&&typeof y!="function"&&typeof y!="symbol"),D!=null&&typeof D!="function"&&typeof D!="symbol"&&typeof D!="boolean"?o.name=""+yr(D):o.removeAttribute("name")}function JO(o,l,f,h,y,_,j,D){if(_!=null&&typeof _!="function"&&typeof _!="symbol"&&typeof _!="boolean"&&(o.type=_),l!=null||f!=null){if(!(_!=="submit"&&_!=="reset"||l!=null))return;f=f!=null?""+yr(f):"",l=l!=null?""+yr(l):f,D||l===o.value||(o.value=l),o.defaultValue=l}h=h??y,h=typeof h!="function"&&typeof h!="symbol"&&!!h,o.checked=D?o.checked:!!h,o.defaultChecked=!!h,j!=null&&typeof j!="function"&&typeof j!="symbol"&&typeof j!="boolean"&&(o.name=j)}function Og(o,l,f){l==="number"&&$d(o.ownerDocument)===o||o.defaultValue===""+f||(o.defaultValue=""+f)}function Ms(o,l,f,h){if(o=o.options,l){l={};for(var y=0;y<f.length;y++)l["$"+f[y]]=!0;for(f=0;f<o.length;f++)y=l.hasOwnProperty("$"+o[f].value),o[f].selected!==y&&(o[f].selected=y),y&&h&&(o[f].defaultSelected=!0)}else{for(f=""+yr(f),l=null,y=0;y<o.length;y++){if(o[y].value===f){o[y].selected=!0,h&&(o[y].defaultSelected=!0);return}l!==null||o[y].disabled||(l=o[y])}l!==null&&(l.selected=!0)}}function eA(o,l,f){if(l!=null&&(l=""+yr(l),l!==o.value&&(o.value=l),f==null)){o.defaultValue!==l&&(o.defaultValue=l);return}o.defaultValue=f!=null?""+yr(f):""}function tA(o,l,f,h){if(l==null){if(h!=null){if(f!=null)throw Error(r(92));if(Y(h)){if(1<h.length)throw Error(r(93));h=h[0]}f=h}f==null&&(f=""),l=f}f=yr(l),o.defaultValue=f,h=o.textContent,h===f&&h!==""&&h!==null&&(o.value=h)}function Ps(o,l){if(l){var f=o.firstChild;if(f&&f===o.lastChild&&f.nodeType===3){f.nodeValue=l;return}}o.textContent=l}var I4=new Set("animationIterationCount aspectRatio borderImageOutset borderImageSlice borderImageWidth boxFlex boxFlexGroup boxOrdinalGroup columnCount columns flex flexGrow flexPositive flexShrink flexNegative flexOrder gridArea gridRow gridRowEnd gridRowSpan gridRowStart gridColumn gridColumnEnd gridColumnSpan gridColumnStart fontWeight lineClamp lineHeight opacity order orphans scale tabSize widows zIndex zoom fillOpacity floodOpacity stopOpacity strokeDasharray strokeDashoffset strokeMiterlimit strokeOpacity strokeWidth MozAnimationIterationCount MozBoxFlex MozBoxFlexGroup MozLineClamp msAnimationIterationCount msFlex msZoom msFlexGrow msFlexNegative msFlexOrder msFlexPositive msFlexShrink msGridColumn msGridColumnSpan msGridRow msGridRowSpan WebkitAnimationIterationCount WebkitBoxFlex WebKitBoxFlexGroup WebkitBoxOrdinalGroup WebkitColumnCount WebkitColumns WebkitFlex WebkitFlexGrow WebkitFlexPositive WebkitFlexShrink WebkitLineClamp".split(" "));function nA(o,l,f){var h=l.indexOf("--")===0;f==null||typeof f=="boolean"||f===""?h?o.setProperty(l,""):l==="float"?o.cssFloat="":o[l]="":h?o.setProperty(l,f):typeof f!="number"||f===0||I4.has(l)?l==="float"?o.cssFloat=f:o[l]=(""+f).trim():o[l]=f+"px"}function rA(o,l,f){if(l!=null&&typeof l!="object")throw Error(r(62));if(o=o.style,f!=null){for(var h in f)!f.hasOwnProperty(h)||l!=null&&l.hasOwnProperty(h)||(h.indexOf("--")===0?o.setProperty(h,""):h==="float"?o.cssFloat="":o[h]="");for(var y in l)h=l[y],l.hasOwnProperty(y)&&f[y]!==h&&nA(o,y,h)}else for(var _ in l)l.hasOwnProperty(_)&&nA(o,_,l[_])}function Ag(o){if(o.indexOf("-")===-1)return!1;switch(o){case"annotation-xml":case"color-profile":case"font-face":case"font-face-src":case"font-face-uri":case"font-face-format":case"font-face-name":case"missing-glyph":return!1;default:return!0}}var L4=new Map([["acceptCharset","accept-charset"],["htmlFor","for"],["httpEquiv","http-equiv"],["crossOrigin","crossorigin"],["accentHeight","accent-height"],["alignmentBaseline","alignment-baseline"],["arabicForm","arabic-form"],["baselineShift","baseline-shift"],["capHeight","cap-height"],["clipPath","clip-path"],["clipRule","clip-rule"],["colorInterpolation","color-interpolation"],["colorInterpolationFilters","color-interpolation-filters"],["colorProfile","color-profile"],["colorRendering","color-rendering"],["dominantBaseline","dominant-baseline"],["enableBackground","enable-background"],["fillOpacity","fill-opacity"],["fillRule","fill-rule"],["floodColor","flood-color"],["floodOpacity","flood-opacity"],["fontFamily","font-family"],["fontSize","font-size"],["fontSizeAdjust","font-size-adjust"],["fontStretch","font-stretch"],["fontStyle","font-style"],["fontVariant","font-variant"],["fontWeight","font-weight"],["glyphName","glyph-name"],["glyphOrientationHorizontal","glyph-orientation-horizontal"],["glyphOrientationVertical","glyph-orientation-vertical"],["horizAdvX","horiz-adv-x"],["horizOriginX","horiz-origin-x"],["imageRendering","image-rendering"],["letterSpacing","letter-spacing"],["lightingColor","lighting-color"],["markerEnd","marker-end"],["markerMid","marker-mid"],["markerStart","marker-start"],["overlinePosition","overline-position"],["overlineThickness","overline-thickness"],["paintOrder","paint-order"],["panose-1","panose-1"],["pointerEvents","pointer-events"],["renderingIntent","rendering-intent"],["shapeRendering","shape-rendering"],["stopColor","stop-color"],["stopOpacity","stop-opacity"],["strikethroughPosition","strikethrough-position"],["strikethroughThickness","strikethrough-thickness"],["strokeDasharray","stroke-dasharray"],["strokeDashoffset","stroke-dashoffset"],["strokeLinecap","stroke-linecap"],["strokeLinejoin","stroke-linejoin"],["strokeMiterlimit","stroke-miterlimit"],["strokeOpacity","stroke-opacity"],["strokeWidth","stroke-width"],["textAnchor","text-anchor"],["textDecoration","text-decoration"],["textRendering","text-rendering"],["transformOrigin","transform-origin"],["underlinePosition","underline-position"],["underlineThickness","underline-thickness"],["unicodeBidi","unicode-bidi"],["unicodeRange","unicode-range"],["unitsPerEm","units-per-em"],["vAlphabetic","v-alphabetic"],["vHanging","v-hanging"],["vIdeographic","v-ideographic"],["vMathematical","v-mathematical"],["vectorEffect","vector-effect"],["vertAdvY","vert-adv-y"],["vertOriginX","vert-origin-x"],["vertOriginY","vert-origin-y"],["wordSpacing","word-spacing"],["writingMode","writing-mode"],["xmlnsXlink","xmlns:xlink"],["xHeight","x-height"]]),B4=/^[\u0000-\u001F ]*j[\r\n\t]*a[\r\n\t]*v[\r\n\t]*a[\r\n\t]*s[\r\n\t]*c[\r\n\t]*r[\r\n\t]*i[\r\n\t]*p[\r\n\t]*t[\r\n\t]*:/i;function Id(o){return B4.test(""+o)?"javascript:throw new Error('React has blocked a javascript: URL as a security precaution.')":o}var Rg=null;function Tg(o){return o=o.target||o.srcElement||window,o.correspondingUseElement&&(o=o.correspondingUseElement),o.nodeType===3?o.parentNode:o}var js=null,Ns=null;function iA(o){var l=Os(o);if(l&&(o=l.stateNode)){var f=o[Vn]||null;e:switch(o=l.stateNode,l.type){case"input":if(Cg(o,f.value,f.defaultValue,f.defaultValue,f.checked,f.defaultChecked,f.type,f.name),l=f.name,f.type==="radio"&&l!=null){for(f=o;f.parentNode;)f=f.parentNode;for(f=f.querySelectorAll('input[name="'+br(""+l)+'"][type="radio"]'),l=0;l<f.length;l++){var h=f[l];if(h!==o&&h.form===o.form){var y=h[Vn]||null;if(!y)throw Error(r(90));Cg(h,y.value,y.defaultValue,y.defaultValue,y.checked,y.defaultChecked,y.type,y.name)}}for(l=0;l<f.length;l++)h=f[l],h.form===o.form&&QO(h)}break e;case"textarea":eA(o,f.value,f.defaultValue);break e;case"select":l=f.value,l!=null&&Ms(o,!!f.multiple,l,!1)}}}var Mg=!1;function aA(o,l,f){if(Mg)return o(l,f);Mg=!0;try{var h=o(l);return h}finally{if(Mg=!1,(js!==null||Ns!==null)&&(Sp(),js&&(l=js,o=Ns,Ns=js=null,iA(l),o)))for(l=0;l<o.length;l++)iA(o[l])}}function Au(o,l){var f=o.stateNode;if(f===null)return null;var h=f[Vn]||null;if(h===null)return null;f=h[l];e:switch(l){case"onClick":case"onClickCapture":case"onDoubleClick":case"onDoubleClickCapture":case"onMouseDown":case"onMouseDownCapture":case"onMouseMove":case"onMouseMoveCapture":case"onMouseUp":case"onMouseUpCapture":case"onMouseEnter":(h=!h.disabled)||(o=o.type,h=!(o==="button"||o==="input"||o==="select"||o==="textarea")),o=!h;break e;default:o=!1}if(o)return null;if(f&&typeof f!="function")throw Error(r(231,l,typeof f));return f}var Di=!(typeof window>"u"||typeof window.document>"u"||typeof window.document.createElement>"u"),Pg=!1;if(Di)try{var Ru={};Object.defineProperty(Ru,"passive",{get:function(){Pg=!0}}),window.addEventListener("test",Ru,Ru),window.removeEventListener("test",Ru,Ru)}catch{Pg=!1}var Oa=null,jg=null,Ld=null;function oA(){if(Ld)return Ld;var o,l=jg,f=l.length,h,y="value"in Oa?Oa.value:Oa.textContent,_=y.length;for(o=0;o<f&&l[o]===y[o];o++);var j=f-o;for(h=1;h<=j&&l[f-h]===y[_-h];h++);return Ld=y.slice(o,1<h?1-h:void 0)}function Bd(o){var l=o.keyCode;return"charCode"in o?(o=o.charCode,o===0&&l===13&&(o=13)):o=l,o===10&&(o=13),32<=o||o===13?o:0}function Fd(){return!0}function sA(){return!1}function Hn(o){function l(f,h,y,_,j){this._reactName=f,this._targetInst=y,this.type=h,this.nativeEvent=_,this.target=j,this.currentTarget=null;for(var D in o)o.hasOwnProperty(D)&&(f=o[D],this[D]=f?f(_):_[D]);return this.isDefaultPrevented=(_.defaultPrevented!=null?_.defaultPrevented:_.returnValue===!1)?Fd:sA,this.isPropagationStopped=sA,this}return p(l.prototype,{preventDefault:function(){this.defaultPrevented=!0;var f=this.nativeEvent;f&&(f.preventDefault?f.preventDefault():typeof f.returnValue!="unknown"&&(f.returnValue=!1),this.isDefaultPrevented=Fd)},stopPropagation:function(){var f=this.nativeEvent;f&&(f.stopPropagation?f.stopPropagation():typeof f.cancelBubble!="unknown"&&(f.cancelBubble=!0),this.isPropagationStopped=Fd)},persist:function(){},isPersistent:Fd}),l}var So={eventPhase:0,bubbles:0,cancelable:0,timeStamp:function(o){return o.timeStamp||Date.now()},defaultPrevented:0,isTrusted:0},qd=Hn(So),Tu=p({},So,{view:0,detail:0}),F4=Hn(Tu),Ng,Dg,Mu,zd=p({},Tu,{screenX:0,screenY:0,clientX:0,clientY:0,pageX:0,pageY:0,ctrlKey:0,shiftKey:0,altKey:0,metaKey:0,getModifierState:$g,button:0,buttons:0,relatedTarget:function(o){return o.relatedTarget===void 0?o.fromElement===o.srcElement?o.toElement:o.fromElement:o.relatedTarget},movementX:function(o){return"movementX"in o?o.movementX:(o!==Mu&&(Mu&&o.type==="mousemove"?(Ng=o.screenX-Mu.screenX,Dg=o.screenY-Mu.screenY):Dg=Ng=0,Mu=o),Ng)},movementY:function(o){return"movementY"in o?o.movementY:Dg}}),lA=Hn(zd),q4=p({},zd,{dataTransfer:0}),z4=Hn(q4),U4=p({},Tu,{relatedTarget:0}),kg=Hn(U4),V4=p({},So,{animationName:0,elapsedTime:0,pseudoElement:0}),H4=Hn(V4),G4=p({},So,{clipboardData:function(o){return"clipboardData"in o?o.clipboardData:window.clipboardData}}),W4=Hn(G4),K4=p({},So,{data:0}),uA=Hn(K4),Y4={Esc:"Escape",Spacebar:" ",Left:"ArrowLeft",Up:"ArrowUp",Right:"ArrowRight",Down:"ArrowDown",Del:"Delete",Win:"OS",Menu:"ContextMenu",Apps:"ContextMenu",Scroll:"ScrollLock",MozPrintableKey:"Unidentified"},Z4={8:"Backspace",9:"Tab",12:"Clear",13:"Enter",16:"Shift",17:"Control",18:"Alt",19:"Pause",20:"CapsLock",27:"Escape",32:" ",33:"PageUp",34:"PageDown",35:"End",36:"Home",37:"ArrowLeft",38:"ArrowUp",39:"ArrowRight",40:"ArrowDown",45:"Insert",46:"Delete",112:"F1",113:"F2",114:"F3",115:"F4",116:"F5",117:"F6",118:"F7",119:"F8",120:"F9",121:"F10",122:"F11",123:"F12",144:"NumLock",145:"ScrollLock",224:"Meta"},X4={Alt:"altKey",Control:"ctrlKey",Meta:"metaKey",Shift:"shiftKey"};function Q4(o){var l=this.nativeEvent;return l.getModifierState?l.getModifierState(o):(o=X4[o])?!!l[o]:!1}function $g(){return Q4}var J4=p({},Tu,{key:function(o){if(o.key){var l=Y4[o.key]||o.key;if(l!=="Unidentified")return l}return o.type==="keypress"?(o=Bd(o),o===13?"Enter":String.fromCharCode(o)):o.type==="keydown"||o.type==="keyup"?Z4[o.keyCode]||"Unidentified":""},code:0,location:0,ctrlKey:0,shiftKey:0,altKey:0,metaKey:0,repeat:0,locale:0,getModifierState:$g,charCode:function(o){return o.type==="keypress"?Bd(o):0},keyCode:function(o){return o.type==="keydown"||o.type==="keyup"?o.keyCode:0},which:function(o){return o.type==="keypress"?Bd(o):o.type==="keydown"||o.type==="keyup"?o.keyCode:0}}),eH=Hn(J4),tH=p({},zd,{pointerId:0,width:0,height:0,pressure:0,tangentialPressure:0,tiltX:0,tiltY:0,twist:0,pointerType:0,isPrimary:0}),cA=Hn(tH),nH=p({},Tu,{touches:0,targetTouches:0,changedTouches:0,altKey:0,metaKey:0,ctrlKey:0,shiftKey:0,getModifierState:$g}),rH=Hn(nH),iH=p({},So,{propertyName:0,elapsedTime:0,pseudoElement:0}),aH=Hn(iH),oH=p({},zd,{deltaX:function(o){return"deltaX"in o?o.deltaX:"wheelDeltaX"in o?-o.wheelDeltaX:0},deltaY:function(o){return"deltaY"in o?o.deltaY:"wheelDeltaY"in o?-o.wheelDeltaY:"wheelDelta"in o?-o.wheelDelta:0},deltaZ:0,deltaMode:0}),sH=Hn(oH),lH=p({},So,{newState:0,oldState:0}),uH=Hn(lH),cH=[9,13,27,32],Ig=Di&&"CompositionEvent"in window,Pu=null;Di&&"documentMode"in document&&(Pu=document.documentMode);var fH=Di&&"TextEvent"in window&&!Pu,fA=Di&&(!Ig||Pu&&8<Pu&&11>=Pu),dA=" ",pA=!1;function hA(o,l){switch(o){case"keyup":return cH.indexOf(l.keyCode)!==-1;case"keydown":return l.keyCode!==229;case"keypress":case"mousedown":case"focusout":return!0;default:return!1}}function mA(o){return o=o.detail,typeof o=="object"&&"data"in o?o.data:null}var Ds=!1;function dH(o,l){switch(o){case"compositionend":return mA(l);case"keypress":return l.which!==32?null:(pA=!0,dA);case"textInput":return o=l.data,o===dA&&pA?null:o;default:return null}}function pH(o,l){if(Ds)return o==="compositionend"||!Ig&&hA(o,l)?(o=oA(),Ld=jg=Oa=null,Ds=!1,o):null;switch(o){case"paste":return null;case"keypress":if(!(l.ctrlKey||l.altKey||l.metaKey)||l.ctrlKey&&l.altKey){if(l.char&&1<l.char.length)return l.char;if(l.which)return String.fromCharCode(l.which)}return null;case"compositionend":return fA&&l.locale!=="ko"?null:l.data;default:return null}}var hH={color:!0,date:!0,datetime:!0,"datetime-local":!0,email:!0,month:!0,number:!0,password:!0,range:!0,search:!0,tel:!0,text:!0,time:!0,url:!0,week:!0};function vA(o){var l=o&&o.nodeName&&o.nodeName.toLowerCase();return l==="input"?!!hH[o.type]:l==="textarea"}function gA(o,l,f,h){js?Ns?Ns.push(h):Ns=[h]:js=h,l=Rp(l,"onChange"),0<l.length&&(f=new qd("onChange","change",null,f,h),o.push({event:f,listeners:l}))}var ju=null,Nu=null;function mH(o){XT(o,0)}function Ud(o){var l=Ou(o);if(QO(l))return o}function yA(o,l){if(o==="change")return l}var bA=!1;if(Di){var Lg;if(Di){var Bg="oninput"in document;if(!Bg){var xA=document.createElement("div");xA.setAttribute("oninput","return;"),Bg=typeof xA.oninput=="function"}Lg=Bg}else Lg=!1;bA=Lg&&(!document.documentMode||9<document.documentMode)}function wA(){ju&&(ju.detachEvent("onpropertychange",SA),Nu=ju=null)}function SA(o){if(o.propertyName==="value"&&Ud(Nu)){var l=[];gA(l,Nu,o,Tg(o)),aA(mH,l)}}function vH(o,l,f){o==="focusin"?(wA(),ju=l,Nu=f,ju.attachEvent("onpropertychange",SA)):o==="focusout"&&wA()}function gH(o){if(o==="selectionchange"||o==="keyup"||o==="keydown")return Ud(Nu)}function yH(o,l){if(o==="click")return Ud(l)}function bH(o,l){if(o==="input"||o==="change")return Ud(l)}function xH(o,l){return o===l&&(o!==0||1/o===1/l)||o!==o&&l!==l}var Jn=typeof Object.is=="function"?Object.is:xH;function Du(o,l){if(Jn(o,l))return!0;if(typeof o!="object"||o===null||typeof l!="object"||l===null)return!1;var f=Object.keys(o),h=Object.keys(l);if(f.length!==h.length)return!1;for(h=0;h<f.length;h++){var y=f[h];if(!se.call(l,y)||!Jn(o[y],l[y]))return!1}return!0}function _A(o){for(;o&&o.firstChild;)o=o.firstChild;return o}function EA(o,l){var f=_A(o);o=0;for(var h;f;){if(f.nodeType===3){if(h=o+f.textContent.length,o<=l&&h>=l)return{node:f,offset:l-o};o=h}e:{for(;f;){if(f.nextSibling){f=f.nextSibling;break e}f=f.parentNode}f=void 0}f=_A(f)}}function CA(o,l){return o&&l?o===l?!0:o&&o.nodeType===3?!1:l&&l.nodeType===3?CA(o,l.parentNode):"contains"in o?o.contains(l):o.compareDocumentPosition?!!(o.compareDocumentPosition(l)&16):!1:!1}function OA(o){o=o!=null&&o.ownerDocument!=null&&o.ownerDocument.defaultView!=null?o.ownerDocument.defaultView:window;for(var l=$d(o.document);l instanceof o.HTMLIFrameElement;){try{var f=typeof l.contentWindow.location.href=="string"}catch{f=!1}if(f)o=l.contentWindow;else break;l=$d(o.document)}return l}function Fg(o){var l=o&&o.nodeName&&o.nodeName.toLowerCase();return l&&(l==="input"&&(o.type==="text"||o.type==="search"||o.type==="tel"||o.type==="url"||o.type==="password")||l==="textarea"||o.contentEditable==="true")}var wH=Di&&"documentMode"in document&&11>=document.documentMode,ks=null,qg=null,ku=null,zg=!1;function AA(o,l,f){var h=f.window===f?f.document:f.nodeType===9?f:f.ownerDocument;zg||ks==null||ks!==$d(h)||(h=ks,"selectionStart"in h&&Fg(h)?h={start:h.selectionStart,end:h.selectionEnd}:(h=(h.ownerDocument&&h.ownerDocument.defaultView||window).getSelection(),h={anchorNode:h.anchorNode,anchorOffset:h.anchorOffset,focusNode:h.focusNode,focusOffset:h.focusOffset}),ku&&Du(ku,h)||(ku=h,h=Rp(qg,"onSelect"),0<h.length&&(l=new qd("onSelect","select",null,l,f),o.push({event:l,listeners:h}),l.target=ks)))}function _o(o,l){var f={};return f[o.toLowerCase()]=l.toLowerCase(),f["Webkit"+o]="webkit"+l,f["Moz"+o]="moz"+l,f}var $s={animationend:_o("Animation","AnimationEnd"),animationiteration:_o("Animation","AnimationIteration"),animationstart:_o("Animation","AnimationStart"),transitionrun:_o("Transition","TransitionRun"),transitionstart:_o("Transition","TransitionStart"),transitioncancel:_o("Transition","TransitionCancel"),transitionend:_o("Transition","TransitionEnd")},Ug={},RA={};Di&&(RA=document.createElement("div").style,"AnimationEvent"in window||(delete $s.animationend.animation,delete $s.animationiteration.animation,delete $s.animationstart.animation),"TransitionEvent"in window||delete $s.transitionend.transition);function Eo(o){if(Ug[o])return Ug[o];if(!$s[o])return o;var l=$s[o],f;for(f in l)if(l.hasOwnProperty(f)&&f in RA)return Ug[o]=l[f];return o}var TA=Eo("animationend"),MA=Eo("animationiteration"),PA=Eo("animationstart"),SH=Eo("transitionrun"),_H=Eo("transitionstart"),EH=Eo("transitioncancel"),jA=Eo("transitionend"),NA=new Map,Vg="abort auxClick beforeToggle cancel canPlay canPlayThrough click close contextMenu copy cut drag dragEnd dragEnter dragExit dragLeave dragOver dragStart drop durationChange emptied encrypted ended error gotPointerCapture input invalid keyDown keyPress keyUp load loadedData loadedMetadata loadStart lostPointerCapture mouseDown mouseMove mouseOut mouseOver mouseUp paste pause play playing pointerCancel pointerDown pointerMove pointerOut pointerOver pointerUp progress rateChange reset resize seeked seeking stalled submit suspend timeUpdate touchCancel touchEnd touchStart volumeChange scroll toggle touchMove waiting wheel".split(" ");Vg.push("scrollEnd");function Lr(o,l){NA.set(o,l),wo(l,[o])}var DA=new WeakMap;function xr(o,l){if(typeof o=="object"&&o!==null){var f=DA.get(o);return f!==void 0?f:(l={value:o,source:l,stack:ZO(l)},DA.set(o,l),l)}return{value:o,source:l,stack:ZO(l)}}var wr=[],Is=0,Hg=0;function Vd(){for(var o=Is,l=Hg=Is=0;l<o;){var f=wr[l];wr[l++]=null;var h=wr[l];wr[l++]=null;var y=wr[l];wr[l++]=null;var _=wr[l];if(wr[l++]=null,h!==null&&y!==null){var j=h.pending;j===null?y.next=y:(y.next=j.next,j.next=y),h.pending=y}_!==0&&kA(f,y,_)}}function Hd(o,l,f,h){wr[Is++]=o,wr[Is++]=l,wr[Is++]=f,wr[Is++]=h,Hg|=h,o.lanes|=h,o=o.alternate,o!==null&&(o.lanes|=h)}function Gg(o,l,f,h){return Hd(o,l,f,h),Gd(o)}function Ls(o,l){return Hd(o,null,null,l),Gd(o)}function kA(o,l,f){o.lanes|=f;var h=o.alternate;h!==null&&(h.lanes|=f);for(var y=!1,_=o.return;_!==null;)_.childLanes|=f,h=_.alternate,h!==null&&(h.childLanes|=f),_.tag===22&&(o=_.stateNode,o===null||o._visibility&1||(y=!0)),o=_,_=_.return;return o.tag===3?(_=o.stateNode,y&&l!==null&&(y=31-Ke(f),o=_.hiddenUpdates,h=o[y],h===null?o[y]=[l]:h.push(l),l.lane=f|536870912),_):null}function Gd(o){if(50<oc)throw oc=0,Qy=null,Error(r(185));for(var l=o.return;l!==null;)o=l,l=o.return;return o.tag===3?o.stateNode:null}var Bs={};function CH(o,l,f,h){this.tag=o,this.key=f,this.sibling=this.child=this.return=this.stateNode=this.type=this.elementType=null,this.index=0,this.refCleanup=this.ref=null,this.pendingProps=l,this.dependencies=this.memoizedState=this.updateQueue=this.memoizedProps=null,this.mode=h,this.subtreeFlags=this.flags=0,this.deletions=null,this.childLanes=this.lanes=0,this.alternate=null}function er(o,l,f,h){return new CH(o,l,f,h)}function Wg(o){return o=o.prototype,!(!o||!o.isReactComponent)}function ki(o,l){var f=o.alternate;return f===null?(f=er(o.tag,l,o.key,o.mode),f.elementType=o.elementType,f.type=o.type,f.stateNode=o.stateNode,f.alternate=o,o.alternate=f):(f.pendingProps=l,f.type=o.type,f.flags=0,f.subtreeFlags=0,f.deletions=null),f.flags=o.flags&65011712,f.childLanes=o.childLanes,f.lanes=o.lanes,f.child=o.child,f.memoizedProps=o.memoizedProps,f.memoizedState=o.memoizedState,f.updateQueue=o.updateQueue,l=o.dependencies,f.dependencies=l===null?null:{lanes:l.lanes,firstContext:l.firstContext},f.sibling=o.sibling,f.index=o.index,f.ref=o.ref,f.refCleanup=o.refCleanup,f}function $A(o,l){o.flags&=65011714;var f=o.alternate;return f===null?(o.childLanes=0,o.lanes=l,o.child=null,o.subtreeFlags=0,o.memoizedProps=null,o.memoizedState=null,o.updateQueue=null,o.dependencies=null,o.stateNode=null):(o.childLanes=f.childLanes,o.lanes=f.lanes,o.child=f.child,o.subtreeFlags=0,o.deletions=null,o.memoizedProps=f.memoizedProps,o.memoizedState=f.memoizedState,o.updateQueue=f.updateQueue,o.type=f.type,l=f.dependencies,o.dependencies=l===null?null:{lanes:l.lanes,firstContext:l.firstContext}),o}function Wd(o,l,f,h,y,_){var j=0;if(h=o,typeof o=="function")Wg(o)&&(j=1);else if(typeof o=="string")j=A5(o,f,oe.current)?26:o==="html"||o==="head"||o==="body"?27:5;else e:switch(o){case k:return o=er(31,f,l,y),o.elementType=k,o.lanes=_,o;case S:return Co(f.children,y,_,l);case w:j=8,y|=24;break;case x:return o=er(12,f,l,y|2),o.elementType=x,o.lanes=_,o;case R:return o=er(13,f,l,y),o.elementType=R,o.lanes=_,o;case M:return o=er(19,f,l,y),o.elementType=M,o.lanes=_,o;default:if(typeof o=="object"&&o!==null)switch(o.$$typeof){case C:case A:j=10;break e;case O:j=9;break e;case T:j=11;break e;case P:j=14;break e;case N:j=16,h=null;break e}j=29,f=Error(r(130,o===null?"null":typeof o,"")),h=null}return l=er(j,f,l,y),l.elementType=o,l.type=h,l.lanes=_,l}function Co(o,l,f,h){return o=er(7,o,h,l),o.lanes=f,o}function Kg(o,l,f){return o=er(6,o,null,l),o.lanes=f,o}function Yg(o,l,f){return l=er(4,o.children!==null?o.children:[],o.key,l),l.lanes=f,l.stateNode={containerInfo:o.containerInfo,pendingChildren:null,implementation:o.implementation},l}var Fs=[],qs=0,Kd=null,Yd=0,Sr=[],_r=0,Oo=null,$i=1,Ii="";function Ao(o,l){Fs[qs++]=Yd,Fs[qs++]=Kd,Kd=o,Yd=l}function IA(o,l,f){Sr[_r++]=$i,Sr[_r++]=Ii,Sr[_r++]=Oo,Oo=o;var h=$i;o=Ii;var y=32-Ke(h)-1;h&=~(1<<y),f+=1;var _=32-Ke(l)+y;if(30<_){var j=y-y%5;_=(h&(1<<j)-1).toString(32),h>>=j,y-=j,$i=1<<32-Ke(l)+y|f<<y|h,Ii=_+o}else $i=1<<_|f<<y|h,Ii=o}function Zg(o){o.return!==null&&(Ao(o,1),IA(o,1,0))}function Xg(o){for(;o===Kd;)Kd=Fs[--qs],Fs[qs]=null,Yd=Fs[--qs],Fs[qs]=null;for(;o===Oo;)Oo=Sr[--_r],Sr[_r]=null,Ii=Sr[--_r],Sr[_r]=null,$i=Sr[--_r],Sr[_r]=null}var kn=null,Ht=null,mt=!1,Ro=null,ii=!1,Qg=Error(r(519));function To(o){var l=Error(r(418,""));throw Lu(xr(l,o)),Qg}function LA(o){var l=o.stateNode,f=o.type,h=o.memoizedProps;switch(l[Cn]=o,l[Vn]=h,f){case"dialog":ot("cancel",l),ot("close",l);break;case"iframe":case"object":case"embed":ot("load",l);break;case"video":case"audio":for(f=0;f<lc.length;f++)ot(lc[f],l);break;case"source":ot("error",l);break;case"img":case"image":case"link":ot("error",l),ot("load",l);break;case"details":ot("toggle",l);break;case"input":ot("invalid",l),JO(l,h.value,h.defaultValue,h.checked,h.defaultChecked,h.type,h.name,!0),kd(l);break;case"select":ot("invalid",l);break;case"textarea":ot("invalid",l),tA(l,h.value,h.defaultValue,h.children),kd(l)}f=h.children,typeof f!="string"&&typeof f!="number"&&typeof f!="bigint"||l.textContent===""+f||h.suppressHydrationWarning===!0||tM(l.textContent,f)?(h.popover!=null&&(ot("beforetoggle",l),ot("toggle",l)),h.onScroll!=null&&ot("scroll",l),h.onScrollEnd!=null&&ot("scrollend",l),h.onClick!=null&&(l.onclick=Tp),l=!0):l=!1,l||To(o)}function BA(o){for(kn=o.return;kn;)switch(kn.tag){case 5:case 13:ii=!1;return;case 27:case 3:ii=!0;return;default:kn=kn.return}}function $u(o){if(o!==kn)return!1;if(!mt)return BA(o),mt=!0,!1;var l=o.tag,f;if((f=l!==3&&l!==27)&&((f=l===5)&&(f=o.type,f=!(f!=="form"&&f!=="button")||hb(o.type,o.memoizedProps)),f=!f),f&&Ht&&To(o),BA(o),l===13){if(o=o.memoizedState,o=o!==null?o.dehydrated:null,!o)throw Error(r(317));e:{for(o=o.nextSibling,l=0;o;){if(o.nodeType===8)if(f=o.data,f==="/$"){if(l===0){Ht=Fr(o.nextSibling);break e}l--}else f!=="$"&&f!=="$!"&&f!=="$?"||l++;o=o.nextSibling}Ht=null}}else l===27?(l=Ht,za(o.type)?(o=yb,yb=null,Ht=o):Ht=l):Ht=kn?Fr(o.stateNode.nextSibling):null;return!0}function Iu(){Ht=kn=null,mt=!1}function FA(){var o=Ro;return o!==null&&(Kn===null?Kn=o:Kn.push.apply(Kn,o),Ro=null),o}function Lu(o){Ro===null?Ro=[o]:Ro.push(o)}var Jg=J(null),Mo=null,Li=null;function Aa(o,l,f){X(Jg,l._currentValue),l._currentValue=f}function Bi(o){o._currentValue=Jg.current,q(Jg)}function ey(o,l,f){for(;o!==null;){var h=o.alternate;if((o.childLanes&l)!==l?(o.childLanes|=l,h!==null&&(h.childLanes|=l)):h!==null&&(h.childLanes&l)!==l&&(h.childLanes|=l),o===f)break;o=o.return}}function ty(o,l,f,h){var y=o.child;for(y!==null&&(y.return=o);y!==null;){var _=y.dependencies;if(_!==null){var j=y.child;_=_.firstContext;e:for(;_!==null;){var D=_;_=y;for(var U=0;U<l.length;U++)if(D.context===l[U]){_.lanes|=f,D=_.alternate,D!==null&&(D.lanes|=f),ey(_.return,f,o),h||(j=null);break e}_=D.next}}else if(y.tag===18){if(j=y.return,j===null)throw Error(r(341));j.lanes|=f,_=j.alternate,_!==null&&(_.lanes|=f),ey(j,f,o),j=null}else j=y.child;if(j!==null)j.return=y;else for(j=y;j!==null;){if(j===o){j=null;break}if(y=j.sibling,y!==null){y.return=j.return,j=y;break}j=j.return}y=j}}function Bu(o,l,f,h){o=null;for(var y=l,_=!1;y!==null;){if(!_){if((y.flags&524288)!==0)_=!0;else if((y.flags&262144)!==0)break}if(y.tag===10){var j=y.alternate;if(j===null)throw Error(r(387));if(j=j.memoizedProps,j!==null){var D=y.type;Jn(y.pendingProps.value,j.value)||(o!==null?o.push(D):o=[D])}}else if(y===de.current){if(j=y.alternate,j===null)throw Error(r(387));j.memoizedState.memoizedState!==y.memoizedState.memoizedState&&(o!==null?o.push(hc):o=[hc])}y=y.return}o!==null&&ty(l,o,f,h),l.flags|=262144}function Zd(o){for(o=o.firstContext;o!==null;){if(!Jn(o.context._currentValue,o.memoizedValue))return!0;o=o.next}return!1}function Po(o){Mo=o,Li=null,o=o.dependencies,o!==null&&(o.firstContext=null)}function On(o){return qA(Mo,o)}function Xd(o,l){return Mo===null&&Po(o),qA(o,l)}function qA(o,l){var f=l._currentValue;if(l={context:l,memoizedValue:f,next:null},Li===null){if(o===null)throw Error(r(308));Li=l,o.dependencies={lanes:0,firstContext:l},o.flags|=524288}else Li=Li.next=l;return f}var OH=typeof AbortController<"u"?AbortController:function(){var o=[],l=this.signal={aborted:!1,addEventListener:function(f,h){o.push(h)}};this.abort=function(){l.aborted=!0,o.forEach(function(f){return f()})}},AH=e.unstable_scheduleCallback,RH=e.unstable_NormalPriority,tn={$$typeof:A,Consumer:null,Provider:null,_currentValue:null,_currentValue2:null,_threadCount:0};function ny(){return{controller:new OH,data:new Map,refCount:0}}function Fu(o){o.refCount--,o.refCount===0&&AH(RH,function(){o.controller.abort()})}var qu=null,ry=0,zs=0,Us=null;function TH(o,l){if(qu===null){var f=qu=[];ry=0,zs=ab(),Us={status:"pending",value:void 0,then:function(h){f.push(h)}}}return ry++,l.then(zA,zA),l}function zA(){if(--ry===0&&qu!==null){Us!==null&&(Us.status="fulfilled");var o=qu;qu=null,zs=0,Us=null;for(var l=0;l<o.length;l++)(0,o[l])()}}function MH(o,l){var f=[],h={status:"pending",value:null,reason:null,then:function(y){f.push(y)}};return o.then(function(){h.status="fulfilled",h.value=l;for(var y=0;y<f.length;y++)(0,f[y])(l)},function(y){for(h.status="rejected",h.reason=y,y=0;y<f.length;y++)(0,f[y])(void 0)}),h}var UA=I.S;I.S=function(o,l){typeof l=="object"&&l!==null&&typeof l.then=="function"&&TH(o,l),UA!==null&&UA(o,l)};var jo=J(null);function iy(){var o=jo.current;return o!==null?o:Nt.pooledCache}function Qd(o,l){l===null?X(jo,jo.current):X(jo,l.pool)}function VA(){var o=iy();return o===null?null:{parent:tn._currentValue,pool:o}}var zu=Error(r(460)),HA=Error(r(474)),Jd=Error(r(542)),ay={then:function(){}};function GA(o){return o=o.status,o==="fulfilled"||o==="rejected"}function ep(){}function WA(o,l,f){switch(f=o[f],f===void 0?o.push(l):f!==l&&(l.then(ep,ep),l=f),l.status){case"fulfilled":return l.value;case"rejected":throw o=l.reason,YA(o),o;default:if(typeof l.status=="string")l.then(ep,ep);else{if(o=Nt,o!==null&&100<o.shellSuspendCounter)throw Error(r(482));o=l,o.status="pending",o.then(function(h){if(l.status==="pending"){var y=l;y.status="fulfilled",y.value=h}},function(h){if(l.status==="pending"){var y=l;y.status="rejected",y.reason=h}})}switch(l.status){case"fulfilled":return l.value;case"rejected":throw o=l.reason,YA(o),o}throw Uu=l,zu}}var Uu=null;function KA(){if(Uu===null)throw Error(r(459));var o=Uu;return Uu=null,o}function YA(o){if(o===zu||o===Jd)throw Error(r(483))}var Ra=!1;function oy(o){o.updateQueue={baseState:o.memoizedState,firstBaseUpdate:null,lastBaseUpdate:null,shared:{pending:null,lanes:0,hiddenCallbacks:null},callbacks:null}}function sy(o,l){o=o.updateQueue,l.updateQueue===o&&(l.updateQueue={baseState:o.baseState,firstBaseUpdate:o.firstBaseUpdate,lastBaseUpdate:o.lastBaseUpdate,shared:o.shared,callbacks:null})}function Ta(o){return{lane:o,tag:0,payload:null,callback:null,next:null}}function Ma(o,l,f){var h=o.updateQueue;if(h===null)return null;if(h=h.shared,(wt&2)!==0){var y=h.pending;return y===null?l.next=l:(l.next=y.next,y.next=l),h.pending=l,l=Gd(o),kA(o,null,f),l}return Hd(o,h,l,f),Gd(o)}function Vu(o,l,f){if(l=l.updateQueue,l!==null&&(l=l.shared,(f&4194048)!==0)){var h=l.lanes;h&=o.pendingLanes,f|=h,l.lanes=f,zO(o,f)}}function ly(o,l){var f=o.updateQueue,h=o.alternate;if(h!==null&&(h=h.updateQueue,f===h)){var y=null,_=null;if(f=f.firstBaseUpdate,f!==null){do{var j={lane:f.lane,tag:f.tag,payload:f.payload,callback:null,next:null};_===null?y=_=j:_=_.next=j,f=f.next}while(f!==null);_===null?y=_=l:_=_.next=l}else y=_=l;f={baseState:h.baseState,firstBaseUpdate:y,lastBaseUpdate:_,shared:h.shared,callbacks:h.callbacks},o.updateQueue=f;return}o=f.lastBaseUpdate,o===null?f.firstBaseUpdate=l:o.next=l,f.lastBaseUpdate=l}var uy=!1;function Hu(){if(uy){var o=Us;if(o!==null)throw o}}function Gu(o,l,f,h){uy=!1;var y=o.updateQueue;Ra=!1;var _=y.firstBaseUpdate,j=y.lastBaseUpdate,D=y.shared.pending;if(D!==null){y.shared.pending=null;var U=D,ne=U.next;U.next=null,j===null?_=ne:j.next=ne,j=U;var he=o.alternate;he!==null&&(he=he.updateQueue,D=he.lastBaseUpdate,D!==j&&(D===null?he.firstBaseUpdate=ne:D.next=ne,he.lastBaseUpdate=U))}if(_!==null){var ge=y.baseState;j=0,he=ne=U=null,D=_;do{var ie=D.lane&-536870913,ae=ie!==D.lane;if(ae?(lt&ie)===ie:(h&ie)===ie){ie!==0&&ie===zs&&(uy=!0),he!==null&&(he=he.next={lane:0,tag:D.tag,payload:D.payload,callback:null,next:null});e:{var Ve=o,Fe=D;ie=l;var Ct=f;switch(Fe.tag){case 1:if(Ve=Fe.payload,typeof Ve=="function"){ge=Ve.call(Ct,ge,ie);break e}ge=Ve;break e;case 3:Ve.flags=Ve.flags&-65537|128;case 0:if(Ve=Fe.payload,ie=typeof Ve=="function"?Ve.call(Ct,ge,ie):Ve,ie==null)break e;ge=p({},ge,ie);break e;case 2:Ra=!0}}ie=D.callback,ie!==null&&(o.flags|=64,ae&&(o.flags|=8192),ae=y.callbacks,ae===null?y.callbacks=[ie]:ae.push(ie))}else ae={lane:ie,tag:D.tag,payload:D.payload,callback:D.callback,next:null},he===null?(ne=he=ae,U=ge):he=he.next=ae,j|=ie;if(D=D.next,D===null){if(D=y.shared.pending,D===null)break;ae=D,D=ae.next,ae.next=null,y.lastBaseUpdate=ae,y.shared.pending=null}}while(!0);he===null&&(U=ge),y.baseState=U,y.firstBaseUpdate=ne,y.lastBaseUpdate=he,_===null&&(y.shared.lanes=0),La|=j,o.lanes=j,o.memoizedState=ge}}function ZA(o,l){if(typeof o!="function")throw Error(r(191,o));o.call(l)}function XA(o,l){var f=o.callbacks;if(f!==null)for(o.callbacks=null,o=0;o<f.length;o++)ZA(f[o],l)}var Vs=J(null),tp=J(0);function QA(o,l){o=Gi,X(tp,o),X(Vs,l),Gi=o|l.baseLanes}function cy(){X(tp,Gi),X(Vs,Vs.current)}function fy(){Gi=tp.current,q(Vs),q(tp)}var Pa=0,tt=null,_t=null,Qt=null,np=!1,Hs=!1,No=!1,rp=0,Wu=0,Gs=null,PH=0;function Kt(){throw Error(r(321))}function dy(o,l){if(l===null)return!1;for(var f=0;f<l.length&&f<o.length;f++)if(!Jn(o[f],l[f]))return!1;return!0}function py(o,l,f,h,y,_){return Pa=_,tt=l,l.memoizedState=null,l.updateQueue=null,l.lanes=0,I.H=o===null||o.memoizedState===null?kR:$R,No=!1,_=f(h,y),No=!1,Hs&&(_=eR(l,f,h,y)),JA(o),_}function JA(o){I.H=up;var l=_t!==null&&_t.next!==null;if(Pa=0,Qt=_t=tt=null,np=!1,Wu=0,Gs=null,l)throw Error(r(300));o===null||hn||(o=o.dependencies,o!==null&&Zd(o)&&(hn=!0))}function eR(o,l,f,h){tt=o;var y=0;do{if(Hs&&(Gs=null),Wu=0,Hs=!1,25<=y)throw Error(r(301));if(y+=1,Qt=_t=null,o.updateQueue!=null){var _=o.updateQueue;_.lastEffect=null,_.events=null,_.stores=null,_.memoCache!=null&&(_.memoCache.index=0)}I.H=LH,_=l(f,h)}while(Hs);return _}function jH(){var o=I.H,l=o.useState()[0];return l=typeof l.then=="function"?Ku(l):l,o=o.useState()[0],(_t!==null?_t.memoizedState:null)!==o&&(tt.flags|=1024),l}function hy(){var o=rp!==0;return rp=0,o}function my(o,l,f){l.updateQueue=o.updateQueue,l.flags&=-2053,o.lanes&=~f}function vy(o){if(np){for(o=o.memoizedState;o!==null;){var l=o.queue;l!==null&&(l.pending=null),o=o.next}np=!1}Pa=0,Qt=_t=tt=null,Hs=!1,Wu=rp=0,Gs=null}function Gn(){var o={memoizedState:null,baseState:null,baseQueue:null,queue:null,next:null};return Qt===null?tt.memoizedState=Qt=o:Qt=Qt.next=o,Qt}function Jt(){if(_t===null){var o=tt.alternate;o=o!==null?o.memoizedState:null}else o=_t.next;var l=Qt===null?tt.memoizedState:Qt.next;if(l!==null)Qt=l,_t=o;else{if(o===null)throw tt.alternate===null?Error(r(467)):Error(r(310));_t=o,o={memoizedState:_t.memoizedState,baseState:_t.baseState,baseQueue:_t.baseQueue,queue:_t.queue,next:null},Qt===null?tt.memoizedState=Qt=o:Qt=Qt.next=o}return Qt}function gy(){return{lastEffect:null,events:null,stores:null,memoCache:null}}function Ku(o){var l=Wu;return Wu+=1,Gs===null&&(Gs=[]),o=WA(Gs,o,l),l=tt,(Qt===null?l.memoizedState:Qt.next)===null&&(l=l.alternate,I.H=l===null||l.memoizedState===null?kR:$R),o}function ip(o){if(o!==null&&typeof o=="object"){if(typeof o.then=="function")return Ku(o);if(o.$$typeof===A)return On(o)}throw Error(r(438,String(o)))}function yy(o){var l=null,f=tt.updateQueue;if(f!==null&&(l=f.memoCache),l==null){var h=tt.alternate;h!==null&&(h=h.updateQueue,h!==null&&(h=h.memoCache,h!=null&&(l={data:h.data.map(function(y){return y.slice()}),index:0})))}if(l==null&&(l={data:[],index:0}),f===null&&(f=gy(),tt.updateQueue=f),f.memoCache=l,f=l.data[l.index],f===void 0)for(f=l.data[l.index]=Array(o),h=0;h<o;h++)f[h]=H;return l.index++,f}function Fi(o,l){return typeof l=="function"?l(o):l}function ap(o){var l=Jt();return by(l,_t,o)}function by(o,l,f){var h=o.queue;if(h===null)throw Error(r(311));h.lastRenderedReducer=f;var y=o.baseQueue,_=h.pending;if(_!==null){if(y!==null){var j=y.next;y.next=_.next,_.next=j}l.baseQueue=y=_,h.pending=null}if(_=o.baseState,y===null)o.memoizedState=_;else{l=y.next;var D=j=null,U=null,ne=l,he=!1;do{var ge=ne.lane&-536870913;if(ge!==ne.lane?(lt&ge)===ge:(Pa&ge)===ge){var ie=ne.revertLane;if(ie===0)U!==null&&(U=U.next={lane:0,revertLane:0,action:ne.action,hasEagerState:ne.hasEagerState,eagerState:ne.eagerState,next:null}),ge===zs&&(he=!0);else if((Pa&ie)===ie){ne=ne.next,ie===zs&&(he=!0);continue}else ge={lane:0,revertLane:ne.revertLane,action:ne.action,hasEagerState:ne.hasEagerState,eagerState:ne.eagerState,next:null},U===null?(D=U=ge,j=_):U=U.next=ge,tt.lanes|=ie,La|=ie;ge=ne.action,No&&f(_,ge),_=ne.hasEagerState?ne.eagerState:f(_,ge)}else ie={lane:ge,revertLane:ne.revertLane,action:ne.action,hasEagerState:ne.hasEagerState,eagerState:ne.eagerState,next:null},U===null?(D=U=ie,j=_):U=U.next=ie,tt.lanes|=ge,La|=ge;ne=ne.next}while(ne!==null&&ne!==l);if(U===null?j=_:U.next=D,!Jn(_,o.memoizedState)&&(hn=!0,he&&(f=Us,f!==null)))throw f;o.memoizedState=_,o.baseState=j,o.baseQueue=U,h.lastRenderedState=_}return y===null&&(h.lanes=0),[o.memoizedState,h.dispatch]}function xy(o){var l=Jt(),f=l.queue;if(f===null)throw Error(r(311));f.lastRenderedReducer=o;var h=f.dispatch,y=f.pending,_=l.memoizedState;if(y!==null){f.pending=null;var j=y=y.next;do _=o(_,j.action),j=j.next;while(j!==y);Jn(_,l.memoizedState)||(hn=!0),l.memoizedState=_,l.baseQueue===null&&(l.baseState=_),f.lastRenderedState=_}return[_,h]}function tR(o,l,f){var h=tt,y=Jt(),_=mt;if(_){if(f===void 0)throw Error(r(407));f=f()}else f=l();var j=!Jn((_t||y).memoizedState,f);j&&(y.memoizedState=f,hn=!0),y=y.queue;var D=iR.bind(null,h,y,o);if(Yu(2048,8,D,[o]),y.getSnapshot!==l||j||Qt!==null&&Qt.memoizedState.tag&1){if(h.flags|=2048,Ws(9,op(),rR.bind(null,h,y,f,l),null),Nt===null)throw Error(r(349));_||(Pa&124)!==0||nR(h,l,f)}return f}function nR(o,l,f){o.flags|=16384,o={getSnapshot:l,value:f},l=tt.updateQueue,l===null?(l=gy(),tt.updateQueue=l,l.stores=[o]):(f=l.stores,f===null?l.stores=[o]:f.push(o))}function rR(o,l,f,h){l.value=f,l.getSnapshot=h,aR(l)&&oR(o)}function iR(o,l,f){return f(function(){aR(l)&&oR(o)})}function aR(o){var l=o.getSnapshot;o=o.value;try{var f=l();return!Jn(o,f)}catch{return!0}}function oR(o){var l=Ls(o,2);l!==null&&ar(l,o,2)}function wy(o){var l=Gn();if(typeof o=="function"){var f=o;if(o=f(),No){Te(!0);try{f()}finally{Te(!1)}}}return l.memoizedState=l.baseState=o,l.queue={pending:null,lanes:0,dispatch:null,lastRenderedReducer:Fi,lastRenderedState:o},l}function sR(o,l,f,h){return o.baseState=f,by(o,_t,typeof h=="function"?h:Fi)}function NH(o,l,f,h,y){if(lp(o))throw Error(r(485));if(o=l.action,o!==null){var _={payload:y,action:o,next:null,isTransition:!0,status:"pending",value:null,reason:null,listeners:[],then:function(j){_.listeners.push(j)}};I.T!==null?f(!0):_.isTransition=!1,h(_),f=l.pending,f===null?(_.next=l.pending=_,lR(l,_)):(_.next=f.next,l.pending=f.next=_)}}function lR(o,l){var f=l.action,h=l.payload,y=o.state;if(l.isTransition){var _=I.T,j={};I.T=j;try{var D=f(y,h),U=I.S;U!==null&&U(j,D),uR(o,l,D)}catch(ne){Sy(o,l,ne)}finally{I.T=_}}else try{_=f(y,h),uR(o,l,_)}catch(ne){Sy(o,l,ne)}}function uR(o,l,f){f!==null&&typeof f=="object"&&typeof f.then=="function"?f.then(function(h){cR(o,l,h)},function(h){return Sy(o,l,h)}):cR(o,l,f)}function cR(o,l,f){l.status="fulfilled",l.value=f,fR(l),o.state=f,l=o.pending,l!==null&&(f=l.next,f===l?o.pending=null:(f=f.next,l.next=f,lR(o,f)))}function Sy(o,l,f){var h=o.pending;if(o.pending=null,h!==null){h=h.next;do l.status="rejected",l.reason=f,fR(l),l=l.next;while(l!==h)}o.action=null}function fR(o){o=o.listeners;for(var l=0;l<o.length;l++)(0,o[l])()}function dR(o,l){return l}function pR(o,l){if(mt){var f=Nt.formState;if(f!==null){e:{var h=tt;if(mt){if(Ht){t:{for(var y=Ht,_=ii;y.nodeType!==8;){if(!_){y=null;break t}if(y=Fr(y.nextSibling),y===null){y=null;break t}}_=y.data,y=_==="F!"||_==="F"?y:null}if(y){Ht=Fr(y.nextSibling),h=y.data==="F!";break e}}To(h)}h=!1}h&&(l=f[0])}}return f=Gn(),f.memoizedState=f.baseState=l,h={pending:null,lanes:0,dispatch:null,lastRenderedReducer:dR,lastRenderedState:l},f.queue=h,f=jR.bind(null,tt,h),h.dispatch=f,h=wy(!1),_=Ay.bind(null,tt,!1,h.queue),h=Gn(),y={state:l,dispatch:null,action:o,pending:null},h.queue=y,f=NH.bind(null,tt,y,_,f),y.dispatch=f,h.memoizedState=o,[l,f,!1]}function hR(o){var l=Jt();return mR(l,_t,o)}function mR(o,l,f){if(l=by(o,l,dR)[0],o=ap(Fi)[0],typeof l=="object"&&l!==null&&typeof l.then=="function")try{var h=Ku(l)}catch(j){throw j===zu?Jd:j}else h=l;l=Jt();var y=l.queue,_=y.dispatch;return f!==l.memoizedState&&(tt.flags|=2048,Ws(9,op(),DH.bind(null,y,f),null)),[h,_,o]}function DH(o,l){o.action=l}function vR(o){var l=Jt(),f=_t;if(f!==null)return mR(l,f,o);Jt(),l=l.memoizedState,f=Jt();var h=f.queue.dispatch;return f.memoizedState=o,[l,h,!1]}function Ws(o,l,f,h){return o={tag:o,create:f,deps:h,inst:l,next:null},l=tt.updateQueue,l===null&&(l=gy(),tt.updateQueue=l),f=l.lastEffect,f===null?l.lastEffect=o.next=o:(h=f.next,f.next=o,o.next=h,l.lastEffect=o),o}function op(){return{destroy:void 0,resource:void 0}}function gR(){return Jt().memoizedState}function sp(o,l,f,h){var y=Gn();h=h===void 0?null:h,tt.flags|=o,y.memoizedState=Ws(1|l,op(),f,h)}function Yu(o,l,f,h){var y=Jt();h=h===void 0?null:h;var _=y.memoizedState.inst;_t!==null&&h!==null&&dy(h,_t.memoizedState.deps)?y.memoizedState=Ws(l,_,f,h):(tt.flags|=o,y.memoizedState=Ws(1|l,_,f,h))}function yR(o,l){sp(8390656,8,o,l)}function bR(o,l){Yu(2048,8,o,l)}function xR(o,l){return Yu(4,2,o,l)}function wR(o,l){return Yu(4,4,o,l)}function SR(o,l){if(typeof l=="function"){o=o();var f=l(o);return function(){typeof f=="function"?f():l(null)}}if(l!=null)return o=o(),l.current=o,function(){l.current=null}}function _R(o,l,f){f=f!=null?f.concat([o]):null,Yu(4,4,SR.bind(null,l,o),f)}function _y(){}function ER(o,l){var f=Jt();l=l===void 0?null:l;var h=f.memoizedState;return l!==null&&dy(l,h[1])?h[0]:(f.memoizedState=[o,l],o)}function CR(o,l){var f=Jt();l=l===void 0?null:l;var h=f.memoizedState;if(l!==null&&dy(l,h[1]))return h[0];if(h=o(),No){Te(!0);try{o()}finally{Te(!1)}}return f.memoizedState=[h,l],h}function Ey(o,l,f){return f===void 0||(Pa&1073741824)!==0?o.memoizedState=l:(o.memoizedState=f,o=RT(),tt.lanes|=o,La|=o,f)}function OR(o,l,f,h){return Jn(f,l)?f:Vs.current!==null?(o=Ey(o,f,h),Jn(o,l)||(hn=!0),o):(Pa&42)===0?(hn=!0,o.memoizedState=f):(o=RT(),tt.lanes|=o,La|=o,l)}function AR(o,l,f,h,y){var _=W.p;W.p=_!==0&&8>_?_:8;var j=I.T,D={};I.T=D,Ay(o,!1,l,f);try{var U=y(),ne=I.S;if(ne!==null&&ne(D,U),U!==null&&typeof U=="object"&&typeof U.then=="function"){var he=MH(U,h);Zu(o,l,he,ir(o))}else Zu(o,l,h,ir(o))}catch(ge){Zu(o,l,{then:function(){},status:"rejected",reason:ge},ir())}finally{W.p=_,I.T=j}}function kH(){}function Cy(o,l,f,h){if(o.tag!==5)throw Error(r(476));var y=RR(o).queue;AR(o,y,l,B,f===null?kH:function(){return TR(o),f(h)})}function RR(o){var l=o.memoizedState;if(l!==null)return l;l={memoizedState:B,baseState:B,baseQueue:null,queue:{pending:null,lanes:0,dispatch:null,lastRenderedReducer:Fi,lastRenderedState:B},next:null};var f={};return l.next={memoizedState:f,baseState:f,baseQueue:null,queue:{pending:null,lanes:0,dispatch:null,lastRenderedReducer:Fi,lastRenderedState:f},next:null},o.memoizedState=l,o=o.alternate,o!==null&&(o.memoizedState=l),l}function TR(o){var l=RR(o).next.queue;Zu(o,l,{},ir())}function Oy(){return On(hc)}function MR(){return Jt().memoizedState}function PR(){return Jt().memoizedState}function $H(o){for(var l=o.return;l!==null;){switch(l.tag){case 24:case 3:var f=ir();o=Ta(f);var h=Ma(l,o,f);h!==null&&(ar(h,l,f),Vu(h,l,f)),l={cache:ny()},o.payload=l;return}l=l.return}}function IH(o,l,f){var h=ir();f={lane:h,revertLane:0,action:f,hasEagerState:!1,eagerState:null,next:null},lp(o)?NR(l,f):(f=Gg(o,l,f,h),f!==null&&(ar(f,o,h),DR(f,l,h)))}function jR(o,l,f){var h=ir();Zu(o,l,f,h)}function Zu(o,l,f,h){var y={lane:h,revertLane:0,action:f,hasEagerState:!1,eagerState:null,next:null};if(lp(o))NR(l,y);else{var _=o.alternate;if(o.lanes===0&&(_===null||_.lanes===0)&&(_=l.lastRenderedReducer,_!==null))try{var j=l.lastRenderedState,D=_(j,f);if(y.hasEagerState=!0,y.eagerState=D,Jn(D,j))return Hd(o,l,y,0),Nt===null&&Vd(),!1}catch{}finally{}if(f=Gg(o,l,y,h),f!==null)return ar(f,o,h),DR(f,l,h),!0}return!1}function Ay(o,l,f,h){if(h={lane:2,revertLane:ab(),action:h,hasEagerState:!1,eagerState:null,next:null},lp(o)){if(l)throw Error(r(479))}else l=Gg(o,f,h,2),l!==null&&ar(l,o,2)}function lp(o){var l=o.alternate;return o===tt||l!==null&&l===tt}function NR(o,l){Hs=np=!0;var f=o.pending;f===null?l.next=l:(l.next=f.next,f.next=l),o.pending=l}function DR(o,l,f){if((f&4194048)!==0){var h=l.lanes;h&=o.pendingLanes,f|=h,l.lanes=f,zO(o,f)}}var up={readContext:On,use:ip,useCallback:Kt,useContext:Kt,useEffect:Kt,useImperativeHandle:Kt,useLayoutEffect:Kt,useInsertionEffect:Kt,useMemo:Kt,useReducer:Kt,useRef:Kt,useState:Kt,useDebugValue:Kt,useDeferredValue:Kt,useTransition:Kt,useSyncExternalStore:Kt,useId:Kt,useHostTransitionStatus:Kt,useFormState:Kt,useActionState:Kt,useOptimistic:Kt,useMemoCache:Kt,useCacheRefresh:Kt},kR={readContext:On,use:ip,useCallback:function(o,l){return Gn().memoizedState=[o,l===void 0?null:l],o},useContext:On,useEffect:yR,useImperativeHandle:function(o,l,f){f=f!=null?f.concat([o]):null,sp(4194308,4,SR.bind(null,l,o),f)},useLayoutEffect:function(o,l){return sp(4194308,4,o,l)},useInsertionEffect:function(o,l){sp(4,2,o,l)},useMemo:function(o,l){var f=Gn();l=l===void 0?null:l;var h=o();if(No){Te(!0);try{o()}finally{Te(!1)}}return f.memoizedState=[h,l],h},useReducer:function(o,l,f){var h=Gn();if(f!==void 0){var y=f(l);if(No){Te(!0);try{f(l)}finally{Te(!1)}}}else y=l;return h.memoizedState=h.baseState=y,o={pending:null,lanes:0,dispatch:null,lastRenderedReducer:o,lastRenderedState:y},h.queue=o,o=o.dispatch=IH.bind(null,tt,o),[h.memoizedState,o]},useRef:function(o){var l=Gn();return o={current:o},l.memoizedState=o},useState:function(o){o=wy(o);var l=o.queue,f=jR.bind(null,tt,l);return l.dispatch=f,[o.memoizedState,f]},useDebugValue:_y,useDeferredValue:function(o,l){var f=Gn();return Ey(f,o,l)},useTransition:function(){var o=wy(!1);return o=AR.bind(null,tt,o.queue,!0,!1),Gn().memoizedState=o,[!1,o]},useSyncExternalStore:function(o,l,f){var h=tt,y=Gn();if(mt){if(f===void 0)throw Error(r(407));f=f()}else{if(f=l(),Nt===null)throw Error(r(349));(lt&124)!==0||nR(h,l,f)}y.memoizedState=f;var _={value:f,getSnapshot:l};return y.queue=_,yR(iR.bind(null,h,_,o),[o]),h.flags|=2048,Ws(9,op(),rR.bind(null,h,_,f,l),null),f},useId:function(){var o=Gn(),l=Nt.identifierPrefix;if(mt){var f=Ii,h=$i;f=(h&~(1<<32-Ke(h)-1)).toString(32)+f,l="«"+l+"R"+f,f=rp++,0<f&&(l+="H"+f.toString(32)),l+="»"}else f=PH++,l="«"+l+"r"+f.toString(32)+"»";return o.memoizedState=l},useHostTransitionStatus:Oy,useFormState:pR,useActionState:pR,useOptimistic:function(o){var l=Gn();l.memoizedState=l.baseState=o;var f={pending:null,lanes:0,dispatch:null,lastRenderedReducer:null,lastRenderedState:null};return l.queue=f,l=Ay.bind(null,tt,!0,f),f.dispatch=l,[o,l]},useMemoCache:yy,useCacheRefresh:function(){return Gn().memoizedState=$H.bind(null,tt)}},$R={readContext:On,use:ip,useCallback:ER,useContext:On,useEffect:bR,useImperativeHandle:_R,useInsertionEffect:xR,useLayoutEffect:wR,useMemo:CR,useReducer:ap,useRef:gR,useState:function(){return ap(Fi)},useDebugValue:_y,useDeferredValue:function(o,l){var f=Jt();return OR(f,_t.memoizedState,o,l)},useTransition:function(){var o=ap(Fi)[0],l=Jt().memoizedState;return[typeof o=="boolean"?o:Ku(o),l]},useSyncExternalStore:tR,useId:MR,useHostTransitionStatus:Oy,useFormState:hR,useActionState:hR,useOptimistic:function(o,l){var f=Jt();return sR(f,_t,o,l)},useMemoCache:yy,useCacheRefresh:PR},LH={readContext:On,use:ip,useCallback:ER,useContext:On,useEffect:bR,useImperativeHandle:_R,useInsertionEffect:xR,useLayoutEffect:wR,useMemo:CR,useReducer:xy,useRef:gR,useState:function(){return xy(Fi)},useDebugValue:_y,useDeferredValue:function(o,l){var f=Jt();return _t===null?Ey(f,o,l):OR(f,_t.memoizedState,o,l)},useTransition:function(){var o=xy(Fi)[0],l=Jt().memoizedState;return[typeof o=="boolean"?o:Ku(o),l]},useSyncExternalStore:tR,useId:MR,useHostTransitionStatus:Oy,useFormState:vR,useActionState:vR,useOptimistic:function(o,l){var f=Jt();return _t!==null?sR(f,_t,o,l):(f.baseState=o,[o,f.queue.dispatch])},useMemoCache:yy,useCacheRefresh:PR},Ks=null,Xu=0;function cp(o){var l=Xu;return Xu+=1,Ks===null&&(Ks=[]),WA(Ks,o,l)}function Qu(o,l){l=l.props.ref,o.ref=l!==void 0?l:null}function fp(o,l){throw l.$$typeof===m?Error(r(525)):(o=Object.prototype.toString.call(l),Error(r(31,o==="[object Object]"?"object with keys {"+Object.keys(l).join(", ")+"}":o)))}function IR(o){var l=o._init;return l(o._payload)}function LR(o){function l(ee,K){if(o){var te=ee.deletions;te===null?(ee.deletions=[K],ee.flags|=16):te.push(K)}}function f(ee,K){if(!o)return null;for(;K!==null;)l(ee,K),K=K.sibling;return null}function h(ee){for(var K=new Map;ee!==null;)ee.key!==null?K.set(ee.key,ee):K.set(ee.index,ee),ee=ee.sibling;return K}function y(ee,K){return ee=ki(ee,K),ee.index=0,ee.sibling=null,ee}function _(ee,K,te){return ee.index=te,o?(te=ee.alternate,te!==null?(te=te.index,te<K?(ee.flags|=67108866,K):te):(ee.flags|=67108866,K)):(ee.flags|=1048576,K)}function j(ee){return o&&ee.alternate===null&&(ee.flags|=67108866),ee}function D(ee,K,te,me){return K===null||K.tag!==6?(K=Kg(te,ee.mode,me),K.return=ee,K):(K=y(K,te),K.return=ee,K)}function U(ee,K,te,me){var je=te.type;return je===S?he(ee,K,te.props.children,me,te.key):K!==null&&(K.elementType===je||typeof je=="object"&&je!==null&&je.$$typeof===N&&IR(je)===K.type)?(K=y(K,te.props),Qu(K,te),K.return=ee,K):(K=Wd(te.type,te.key,te.props,null,ee.mode,me),Qu(K,te),K.return=ee,K)}function ne(ee,K,te,me){return K===null||K.tag!==4||K.stateNode.containerInfo!==te.containerInfo||K.stateNode.implementation!==te.implementation?(K=Yg(te,ee.mode,me),K.return=ee,K):(K=y(K,te.children||[]),K.return=ee,K)}function he(ee,K,te,me,je){return K===null||K.tag!==7?(K=Co(te,ee.mode,me,je),K.return=ee,K):(K=y(K,te),K.return=ee,K)}function ge(ee,K,te){if(typeof K=="string"&&K!==""||typeof K=="number"||typeof K=="bigint")return K=Kg(""+K,ee.mode,te),K.return=ee,K;if(typeof K=="object"&&K!==null){switch(K.$$typeof){case v:return te=Wd(K.type,K.key,K.props,null,ee.mode,te),Qu(te,K),te.return=ee,te;case b:return K=Yg(K,ee.mode,te),K.return=ee,K;case N:var me=K._init;return K=me(K._payload),ge(ee,K,te)}if(Y(K)||F(K))return K=Co(K,ee.mode,te,null),K.return=ee,K;if(typeof K.then=="function")return ge(ee,cp(K),te);if(K.$$typeof===A)return ge(ee,Xd(ee,K),te);fp(ee,K)}return null}function ie(ee,K,te,me){var je=K!==null?K.key:null;if(typeof te=="string"&&te!==""||typeof te=="number"||typeof te=="bigint")return je!==null?null:D(ee,K,""+te,me);if(typeof te=="object"&&te!==null){switch(te.$$typeof){case v:return te.key===je?U(ee,K,te,me):null;case b:return te.key===je?ne(ee,K,te,me):null;case N:return je=te._init,te=je(te._payload),ie(ee,K,te,me)}if(Y(te)||F(te))return je!==null?null:he(ee,K,te,me,null);if(typeof te.then=="function")return ie(ee,K,cp(te),me);if(te.$$typeof===A)return ie(ee,K,Xd(ee,te),me);fp(ee,te)}return null}function ae(ee,K,te,me,je){if(typeof me=="string"&&me!==""||typeof me=="number"||typeof me=="bigint")return ee=ee.get(te)||null,D(K,ee,""+me,je);if(typeof me=="object"&&me!==null){switch(me.$$typeof){case v:return ee=ee.get(me.key===null?te:me.key)||null,U(K,ee,me,je);case b:return ee=ee.get(me.key===null?te:me.key)||null,ne(K,ee,me,je);case N:var rt=me._init;return me=rt(me._payload),ae(ee,K,te,me,je)}if(Y(me)||F(me))return ee=ee.get(te)||null,he(K,ee,me,je,null);if(typeof me.then=="function")return ae(ee,K,te,cp(me),je);if(me.$$typeof===A)return ae(ee,K,te,Xd(K,me),je);fp(K,me)}return null}function Ve(ee,K,te,me){for(var je=null,rt=null,Le=K,qe=K=0,vn=null;Le!==null&&qe<te.length;qe++){Le.index>qe?(vn=Le,Le=null):vn=Le.sibling;var pt=ie(ee,Le,te[qe],me);if(pt===null){Le===null&&(Le=vn);break}o&&Le&&pt.alternate===null&&l(ee,Le),K=_(pt,K,qe),rt===null?je=pt:rt.sibling=pt,rt=pt,Le=vn}if(qe===te.length)return f(ee,Le),mt&&Ao(ee,qe),je;if(Le===null){for(;qe<te.length;qe++)Le=ge(ee,te[qe],me),Le!==null&&(K=_(Le,K,qe),rt===null?je=Le:rt.sibling=Le,rt=Le);return mt&&Ao(ee,qe),je}for(Le=h(Le);qe<te.length;qe++)vn=ae(Le,ee,qe,te[qe],me),vn!==null&&(o&&vn.alternate!==null&&Le.delete(vn.key===null?qe:vn.key),K=_(vn,K,qe),rt===null?je=vn:rt.sibling=vn,rt=vn);return o&&Le.forEach(function(Wa){return l(ee,Wa)}),mt&&Ao(ee,qe),je}function Fe(ee,K,te,me){if(te==null)throw Error(r(151));for(var je=null,rt=null,Le=K,qe=K=0,vn=null,pt=te.next();Le!==null&&!pt.done;qe++,pt=te.next()){Le.index>qe?(vn=Le,Le=null):vn=Le.sibling;var Wa=ie(ee,Le,pt.value,me);if(Wa===null){Le===null&&(Le=vn);break}o&&Le&&Wa.alternate===null&&l(ee,Le),K=_(Wa,K,qe),rt===null?je=Wa:rt.sibling=Wa,rt=Wa,Le=vn}if(pt.done)return f(ee,Le),mt&&Ao(ee,qe),je;if(Le===null){for(;!pt.done;qe++,pt=te.next())pt=ge(ee,pt.value,me),pt!==null&&(K=_(pt,K,qe),rt===null?je=pt:rt.sibling=pt,rt=pt);return mt&&Ao(ee,qe),je}for(Le=h(Le);!pt.done;qe++,pt=te.next())pt=ae(Le,ee,qe,pt.value,me),pt!==null&&(o&&pt.alternate!==null&&Le.delete(pt.key===null?qe:pt.key),K=_(pt,K,qe),rt===null?je=pt:rt.sibling=pt,rt=pt);return o&&Le.forEach(function(B5){return l(ee,B5)}),mt&&Ao(ee,qe),je}function Ct(ee,K,te,me){if(typeof te=="object"&&te!==null&&te.type===S&&te.key===null&&(te=te.props.children),typeof te=="object"&&te!==null){switch(te.$$typeof){case v:e:{for(var je=te.key;K!==null;){if(K.key===je){if(je=te.type,je===S){if(K.tag===7){f(ee,K.sibling),me=y(K,te.props.children),me.return=ee,ee=me;break e}}else if(K.elementType===je||typeof je=="object"&&je!==null&&je.$$typeof===N&&IR(je)===K.type){f(ee,K.sibling),me=y(K,te.props),Qu(me,te),me.return=ee,ee=me;break e}f(ee,K);break}else l(ee,K);K=K.sibling}te.type===S?(me=Co(te.props.children,ee.mode,me,te.key),me.return=ee,ee=me):(me=Wd(te.type,te.key,te.props,null,ee.mode,me),Qu(me,te),me.return=ee,ee=me)}return j(ee);case b:e:{for(je=te.key;K!==null;){if(K.key===je)if(K.tag===4&&K.stateNode.containerInfo===te.containerInfo&&K.stateNode.implementation===te.implementation){f(ee,K.sibling),me=y(K,te.children||[]),me.return=ee,ee=me;break e}else{f(ee,K);break}else l(ee,K);K=K.sibling}me=Yg(te,ee.mode,me),me.return=ee,ee=me}return j(ee);case N:return je=te._init,te=je(te._payload),Ct(ee,K,te,me)}if(Y(te))return Ve(ee,K,te,me);if(F(te)){if(je=F(te),typeof je!="function")throw Error(r(150));return te=je.call(te),Fe(ee,K,te,me)}if(typeof te.then=="function")return Ct(ee,K,cp(te),me);if(te.$$typeof===A)return Ct(ee,K,Xd(ee,te),me);fp(ee,te)}return typeof te=="string"&&te!==""||typeof te=="number"||typeof te=="bigint"?(te=""+te,K!==null&&K.tag===6?(f(ee,K.sibling),me=y(K,te),me.return=ee,ee=me):(f(ee,K),me=Kg(te,ee.mode,me),me.return=ee,ee=me),j(ee)):f(ee,K)}return function(ee,K,te,me){try{Xu=0;var je=Ct(ee,K,te,me);return Ks=null,je}catch(Le){if(Le===zu||Le===Jd)throw Le;var rt=er(29,Le,null,ee.mode);return rt.lanes=me,rt.return=ee,rt}finally{}}}var Ys=LR(!0),BR=LR(!1),Er=J(null),ai=null;function ja(o){var l=o.alternate;X(nn,nn.current&1),X(Er,o),ai===null&&(l===null||Vs.current!==null||l.memoizedState!==null)&&(ai=o)}function FR(o){if(o.tag===22){if(X(nn,nn.current),X(Er,o),ai===null){var l=o.alternate;l!==null&&l.memoizedState!==null&&(ai=o)}}else Na()}function Na(){X(nn,nn.current),X(Er,Er.current)}function qi(o){q(Er),ai===o&&(ai=null),q(nn)}var nn=J(0);function dp(o){for(var l=o;l!==null;){if(l.tag===13){var f=l.memoizedState;if(f!==null&&(f=f.dehydrated,f===null||f.data==="$?"||gb(f)))return l}else if(l.tag===19&&l.memoizedProps.revealOrder!==void 0){if((l.flags&128)!==0)return l}else if(l.child!==null){l.child.return=l,l=l.child;continue}if(l===o)break;for(;l.sibling===null;){if(l.return===null||l.return===o)return null;l=l.return}l.sibling.return=l.return,l=l.sibling}return null}function Ry(o,l,f,h){l=o.memoizedState,f=f(h,l),f=f==null?l:p({},l,f),o.memoizedState=f,o.lanes===0&&(o.updateQueue.baseState=f)}var Ty={enqueueSetState:function(o,l,f){o=o._reactInternals;var h=ir(),y=Ta(h);y.payload=l,f!=null&&(y.callback=f),l=Ma(o,y,h),l!==null&&(ar(l,o,h),Vu(l,o,h))},enqueueReplaceState:function(o,l,f){o=o._reactInternals;var h=ir(),y=Ta(h);y.tag=1,y.payload=l,f!=null&&(y.callback=f),l=Ma(o,y,h),l!==null&&(ar(l,o,h),Vu(l,o,h))},enqueueForceUpdate:function(o,l){o=o._reactInternals;var f=ir(),h=Ta(f);h.tag=2,l!=null&&(h.callback=l),l=Ma(o,h,f),l!==null&&(ar(l,o,f),Vu(l,o,f))}};function qR(o,l,f,h,y,_,j){return o=o.stateNode,typeof o.shouldComponentUpdate=="function"?o.shouldComponentUpdate(h,_,j):l.prototype&&l.prototype.isPureReactComponent?!Du(f,h)||!Du(y,_):!0}function zR(o,l,f,h){o=l.state,typeof l.componentWillReceiveProps=="function"&&l.componentWillReceiveProps(f,h),typeof l.UNSAFE_componentWillReceiveProps=="function"&&l.UNSAFE_componentWillReceiveProps(f,h),l.state!==o&&Ty.enqueueReplaceState(l,l.state,null)}function Do(o,l){var f=l;if("ref"in l){f={};for(var h in l)h!=="ref"&&(f[h]=l[h])}if(o=o.defaultProps){f===l&&(f=p({},f));for(var y in o)f[y]===void 0&&(f[y]=o[y])}return f}var pp=typeof reportError=="function"?reportError:function(o){if(typeof window=="object"&&typeof window.ErrorEvent=="function"){var l=new window.ErrorEvent("error",{bubbles:!0,cancelable:!0,message:typeof o=="object"&&o!==null&&typeof o.message=="string"?String(o.message):String(o),error:o});if(!window.dispatchEvent(l))return}else if(typeof process=="object"&&typeof process.emit=="function"){process.emit("uncaughtException",o);return}console.error(o)};function UR(o){pp(o)}function VR(o){console.error(o)}function HR(o){pp(o)}function hp(o,l){try{var f=o.onUncaughtError;f(l.value,{componentStack:l.stack})}catch(h){setTimeout(function(){throw h})}}function GR(o,l,f){try{var h=o.onCaughtError;h(f.value,{componentStack:f.stack,errorBoundary:l.tag===1?l.stateNode:null})}catch(y){setTimeout(function(){throw y})}}function My(o,l,f){return f=Ta(f),f.tag=3,f.payload={element:null},f.callback=function(){hp(o,l)},f}function WR(o){return o=Ta(o),o.tag=3,o}function KR(o,l,f,h){var y=f.type.getDerivedStateFromError;if(typeof y=="function"){var _=h.value;o.payload=function(){return y(_)},o.callback=function(){GR(l,f,h)}}var j=f.stateNode;j!==null&&typeof j.componentDidCatch=="function"&&(o.callback=function(){GR(l,f,h),typeof y!="function"&&(Ba===null?Ba=new Set([this]):Ba.add(this));var D=h.stack;this.componentDidCatch(h.value,{componentStack:D!==null?D:""})})}function BH(o,l,f,h,y){if(f.flags|=32768,h!==null&&typeof h=="object"&&typeof h.then=="function"){if(l=f.alternate,l!==null&&Bu(l,f,y,!0),f=Er.current,f!==null){switch(f.tag){case 13:return ai===null?eb():f.alternate===null&&Gt===0&&(Gt=3),f.flags&=-257,f.flags|=65536,f.lanes=y,h===ay?f.flags|=16384:(l=f.updateQueue,l===null?f.updateQueue=new Set([h]):l.add(h),nb(o,h,y)),!1;case 22:return f.flags|=65536,h===ay?f.flags|=16384:(l=f.updateQueue,l===null?(l={transitions:null,markerInstances:null,retryQueue:new Set([h])},f.updateQueue=l):(f=l.retryQueue,f===null?l.retryQueue=new Set([h]):f.add(h)),nb(o,h,y)),!1}throw Error(r(435,f.tag))}return nb(o,h,y),eb(),!1}if(mt)return l=Er.current,l!==null?((l.flags&65536)===0&&(l.flags|=256),l.flags|=65536,l.lanes=y,h!==Qg&&(o=Error(r(422),{cause:h}),Lu(xr(o,f)))):(h!==Qg&&(l=Error(r(423),{cause:h}),Lu(xr(l,f))),o=o.current.alternate,o.flags|=65536,y&=-y,o.lanes|=y,h=xr(h,f),y=My(o.stateNode,h,y),ly(o,y),Gt!==4&&(Gt=2)),!1;var _=Error(r(520),{cause:h});if(_=xr(_,f),ac===null?ac=[_]:ac.push(_),Gt!==4&&(Gt=2),l===null)return!0;h=xr(h,f),f=l;do{switch(f.tag){case 3:return f.flags|=65536,o=y&-y,f.lanes|=o,o=My(f.stateNode,h,o),ly(f,o),!1;case 1:if(l=f.type,_=f.stateNode,(f.flags&128)===0&&(typeof l.getDerivedStateFromError=="function"||_!==null&&typeof _.componentDidCatch=="function"&&(Ba===null||!Ba.has(_))))return f.flags|=65536,y&=-y,f.lanes|=y,y=WR(y),KR(y,o,f,h),ly(f,y),!1}f=f.return}while(f!==null);return!1}var YR=Error(r(461)),hn=!1;function xn(o,l,f,h){l.child=o===null?BR(l,null,f,h):Ys(l,o.child,f,h)}function ZR(o,l,f,h,y){f=f.render;var _=l.ref;if("ref"in h){var j={};for(var D in h)D!=="ref"&&(j[D]=h[D])}else j=h;return Po(l),h=py(o,l,f,j,_,y),D=hy(),o!==null&&!hn?(my(o,l,y),zi(o,l,y)):(mt&&D&&Zg(l),l.flags|=1,xn(o,l,h,y),l.child)}function XR(o,l,f,h,y){if(o===null){var _=f.type;return typeof _=="function"&&!Wg(_)&&_.defaultProps===void 0&&f.compare===null?(l.tag=15,l.type=_,QR(o,l,_,h,y)):(o=Wd(f.type,null,h,l,l.mode,y),o.ref=l.ref,o.return=l,l.child=o)}if(_=o.child,!Ly(o,y)){var j=_.memoizedProps;if(f=f.compare,f=f!==null?f:Du,f(j,h)&&o.ref===l.ref)return zi(o,l,y)}return l.flags|=1,o=ki(_,h),o.ref=l.ref,o.return=l,l.child=o}function QR(o,l,f,h,y){if(o!==null){var _=o.memoizedProps;if(Du(_,h)&&o.ref===l.ref)if(hn=!1,l.pendingProps=h=_,Ly(o,y))(o.flags&131072)!==0&&(hn=!0);else return l.lanes=o.lanes,zi(o,l,y)}return Py(o,l,f,h,y)}function JR(o,l,f){var h=l.pendingProps,y=h.children,_=o!==null?o.memoizedState:null;if(h.mode==="hidden"){if((l.flags&128)!==0){if(h=_!==null?_.baseLanes|f:f,o!==null){for(y=l.child=o.child,_=0;y!==null;)_=_|y.lanes|y.childLanes,y=y.sibling;l.childLanes=_&~h}else l.childLanes=0,l.child=null;return eT(o,l,h,f)}if((f&536870912)!==0)l.memoizedState={baseLanes:0,cachePool:null},o!==null&&Qd(l,_!==null?_.cachePool:null),_!==null?QA(l,_):cy(),FR(l);else return l.lanes=l.childLanes=536870912,eT(o,l,_!==null?_.baseLanes|f:f,f)}else _!==null?(Qd(l,_.cachePool),QA(l,_),Na(),l.memoizedState=null):(o!==null&&Qd(l,null),cy(),Na());return xn(o,l,y,f),l.child}function eT(o,l,f,h){var y=iy();return y=y===null?null:{parent:tn._currentValue,pool:y},l.memoizedState={baseLanes:f,cachePool:y},o!==null&&Qd(l,null),cy(),FR(l),o!==null&&Bu(o,l,h,!0),null}function mp(o,l){var f=l.ref;if(f===null)o!==null&&o.ref!==null&&(l.flags|=4194816);else{if(typeof f!="function"&&typeof f!="object")throw Error(r(284));(o===null||o.ref!==f)&&(l.flags|=4194816)}}function Py(o,l,f,h,y){return Po(l),f=py(o,l,f,h,void 0,y),h=hy(),o!==null&&!hn?(my(o,l,y),zi(o,l,y)):(mt&&h&&Zg(l),l.flags|=1,xn(o,l,f,y),l.child)}function tT(o,l,f,h,y,_){return Po(l),l.updateQueue=null,f=eR(l,h,f,y),JA(o),h=hy(),o!==null&&!hn?(my(o,l,_),zi(o,l,_)):(mt&&h&&Zg(l),l.flags|=1,xn(o,l,f,_),l.child)}function nT(o,l,f,h,y){if(Po(l),l.stateNode===null){var _=Bs,j=f.contextType;typeof j=="object"&&j!==null&&(_=On(j)),_=new f(h,_),l.memoizedState=_.state!==null&&_.state!==void 0?_.state:null,_.updater=Ty,l.stateNode=_,_._reactInternals=l,_=l.stateNode,_.props=h,_.state=l.memoizedState,_.refs={},oy(l),j=f.contextType,_.context=typeof j=="object"&&j!==null?On(j):Bs,_.state=l.memoizedState,j=f.getDerivedStateFromProps,typeof j=="function"&&(Ry(l,f,j,h),_.state=l.memoizedState),typeof f.getDerivedStateFromProps=="function"||typeof _.getSnapshotBeforeUpdate=="function"||typeof _.UNSAFE_componentWillMount!="function"&&typeof _.componentWillMount!="function"||(j=_.state,typeof _.componentWillMount=="function"&&_.componentWillMount(),typeof _.UNSAFE_componentWillMount=="function"&&_.UNSAFE_componentWillMount(),j!==_.state&&Ty.enqueueReplaceState(_,_.state,null),Gu(l,h,_,y),Hu(),_.state=l.memoizedState),typeof _.componentDidMount=="function"&&(l.flags|=4194308),h=!0}else if(o===null){_=l.stateNode;var D=l.memoizedProps,U=Do(f,D);_.props=U;var ne=_.context,he=f.contextType;j=Bs,typeof he=="object"&&he!==null&&(j=On(he));var ge=f.getDerivedStateFromProps;he=typeof ge=="function"||typeof _.getSnapshotBeforeUpdate=="function",D=l.pendingProps!==D,he||typeof _.UNSAFE_componentWillReceiveProps!="function"&&typeof _.componentWillReceiveProps!="function"||(D||ne!==j)&&zR(l,_,h,j),Ra=!1;var ie=l.memoizedState;_.state=ie,Gu(l,h,_,y),Hu(),ne=l.memoizedState,D||ie!==ne||Ra?(typeof ge=="function"&&(Ry(l,f,ge,h),ne=l.memoizedState),(U=Ra||qR(l,f,U,h,ie,ne,j))?(he||typeof _.UNSAFE_componentWillMount!="function"&&typeof _.componentWillMount!="function"||(typeof _.componentWillMount=="function"&&_.componentWillMount(),typeof _.UNSAFE_componentWillMount=="function"&&_.UNSAFE_componentWillMount()),typeof _.componentDidMount=="function"&&(l.flags|=4194308)):(typeof _.componentDidMount=="function"&&(l.flags|=4194308),l.memoizedProps=h,l.memoizedState=ne),_.props=h,_.state=ne,_.context=j,h=U):(typeof _.componentDidMount=="function"&&(l.flags|=4194308),h=!1)}else{_=l.stateNode,sy(o,l),j=l.memoizedProps,he=Do(f,j),_.props=he,ge=l.pendingProps,ie=_.context,ne=f.contextType,U=Bs,typeof ne=="object"&&ne!==null&&(U=On(ne)),D=f.getDerivedStateFromProps,(ne=typeof D=="function"||typeof _.getSnapshotBeforeUpdate=="function")||typeof _.UNSAFE_componentWillReceiveProps!="function"&&typeof _.componentWillReceiveProps!="function"||(j!==ge||ie!==U)&&zR(l,_,h,U),Ra=!1,ie=l.memoizedState,_.state=ie,Gu(l,h,_,y),Hu();var ae=l.memoizedState;j!==ge||ie!==ae||Ra||o!==null&&o.dependencies!==null&&Zd(o.dependencies)?(typeof D=="function"&&(Ry(l,f,D,h),ae=l.memoizedState),(he=Ra||qR(l,f,he,h,ie,ae,U)||o!==null&&o.dependencies!==null&&Zd(o.dependencies))?(ne||typeof _.UNSAFE_componentWillUpdate!="function"&&typeof _.componentWillUpdate!="function"||(typeof _.componentWillUpdate=="function"&&_.componentWillUpdate(h,ae,U),typeof _.UNSAFE_componentWillUpdate=="function"&&_.UNSAFE_componentWillUpdate(h,ae,U)),typeof _.componentDidUpdate=="function"&&(l.flags|=4),typeof _.getSnapshotBeforeUpdate=="function"&&(l.flags|=1024)):(typeof _.componentDidUpdate!="function"||j===o.memoizedProps&&ie===o.memoizedState||(l.flags|=4),typeof _.getSnapshotBeforeUpdate!="function"||j===o.memoizedProps&&ie===o.memoizedState||(l.flags|=1024),l.memoizedProps=h,l.memoizedState=ae),_.props=h,_.state=ae,_.context=U,h=he):(typeof _.componentDidUpdate!="function"||j===o.memoizedProps&&ie===o.memoizedState||(l.flags|=4),typeof _.getSnapshotBeforeUpdate!="function"||j===o.memoizedProps&&ie===o.memoizedState||(l.flags|=1024),h=!1)}return _=h,mp(o,l),h=(l.flags&128)!==0,_||h?(_=l.stateNode,f=h&&typeof f.getDerivedStateFromError!="function"?null:_.render(),l.flags|=1,o!==null&&h?(l.child=Ys(l,o.child,null,y),l.child=Ys(l,null,f,y)):xn(o,l,f,y),l.memoizedState=_.state,o=l.child):o=zi(o,l,y),o}function rT(o,l,f,h){return Iu(),l.flags|=256,xn(o,l,f,h),l.child}var jy={dehydrated:null,treeContext:null,retryLane:0,hydrationErrors:null};function Ny(o){return{baseLanes:o,cachePool:VA()}}function Dy(o,l,f){return o=o!==null?o.childLanes&~f:0,l&&(o|=Cr),o}function iT(o,l,f){var h=l.pendingProps,y=!1,_=(l.flags&128)!==0,j;if((j=_)||(j=o!==null&&o.memoizedState===null?!1:(nn.current&2)!==0),j&&(y=!0,l.flags&=-129),j=(l.flags&32)!==0,l.flags&=-33,o===null){if(mt){if(y?ja(l):Na(),mt){var D=Ht,U;if(U=D){e:{for(U=D,D=ii;U.nodeType!==8;){if(!D){D=null;break e}if(U=Fr(U.nextSibling),U===null){D=null;break e}}D=U}D!==null?(l.memoizedState={dehydrated:D,treeContext:Oo!==null?{id:$i,overflow:Ii}:null,retryLane:536870912,hydrationErrors:null},U=er(18,null,null,0),U.stateNode=D,U.return=l,l.child=U,kn=l,Ht=null,U=!0):U=!1}U||To(l)}if(D=l.memoizedState,D!==null&&(D=D.dehydrated,D!==null))return gb(D)?l.lanes=32:l.lanes=536870912,null;qi(l)}return D=h.children,h=h.fallback,y?(Na(),y=l.mode,D=vp({mode:"hidden",children:D},y),h=Co(h,y,f,null),D.return=l,h.return=l,D.sibling=h,l.child=D,y=l.child,y.memoizedState=Ny(f),y.childLanes=Dy(o,j,f),l.memoizedState=jy,h):(ja(l),ky(l,D))}if(U=o.memoizedState,U!==null&&(D=U.dehydrated,D!==null)){if(_)l.flags&256?(ja(l),l.flags&=-257,l=$y(o,l,f)):l.memoizedState!==null?(Na(),l.child=o.child,l.flags|=128,l=null):(Na(),y=h.fallback,D=l.mode,h=vp({mode:"visible",children:h.children},D),y=Co(y,D,f,null),y.flags|=2,h.return=l,y.return=l,h.sibling=y,l.child=h,Ys(l,o.child,null,f),h=l.child,h.memoizedState=Ny(f),h.childLanes=Dy(o,j,f),l.memoizedState=jy,l=y);else if(ja(l),gb(D)){if(j=D.nextSibling&&D.nextSibling.dataset,j)var ne=j.dgst;j=ne,h=Error(r(419)),h.stack="",h.digest=j,Lu({value:h,source:null,stack:null}),l=$y(o,l,f)}else if(hn||Bu(o,l,f,!1),j=(f&o.childLanes)!==0,hn||j){if(j=Nt,j!==null&&(h=f&-f,h=(h&42)!==0?1:yg(h),h=(h&(j.suspendedLanes|f))!==0?0:h,h!==0&&h!==U.retryLane))throw U.retryLane=h,Ls(o,h),ar(j,o,h),YR;D.data==="$?"||eb(),l=$y(o,l,f)}else D.data==="$?"?(l.flags|=192,l.child=o.child,l=null):(o=U.treeContext,Ht=Fr(D.nextSibling),kn=l,mt=!0,Ro=null,ii=!1,o!==null&&(Sr[_r++]=$i,Sr[_r++]=Ii,Sr[_r++]=Oo,$i=o.id,Ii=o.overflow,Oo=l),l=ky(l,h.children),l.flags|=4096);return l}return y?(Na(),y=h.fallback,D=l.mode,U=o.child,ne=U.sibling,h=ki(U,{mode:"hidden",children:h.children}),h.subtreeFlags=U.subtreeFlags&65011712,ne!==null?y=ki(ne,y):(y=Co(y,D,f,null),y.flags|=2),y.return=l,h.return=l,h.sibling=y,l.child=h,h=y,y=l.child,D=o.child.memoizedState,D===null?D=Ny(f):(U=D.cachePool,U!==null?(ne=tn._currentValue,U=U.parent!==ne?{parent:ne,pool:ne}:U):U=VA(),D={baseLanes:D.baseLanes|f,cachePool:U}),y.memoizedState=D,y.childLanes=Dy(o,j,f),l.memoizedState=jy,h):(ja(l),f=o.child,o=f.sibling,f=ki(f,{mode:"visible",children:h.children}),f.return=l,f.sibling=null,o!==null&&(j=l.deletions,j===null?(l.deletions=[o],l.flags|=16):j.push(o)),l.child=f,l.memoizedState=null,f)}function ky(o,l){return l=vp({mode:"visible",children:l},o.mode),l.return=o,o.child=l}function vp(o,l){return o=er(22,o,null,l),o.lanes=0,o.stateNode={_visibility:1,_pendingMarkers:null,_retryCache:null,_transitions:null},o}function $y(o,l,f){return Ys(l,o.child,null,f),o=ky(l,l.pendingProps.children),o.flags|=2,l.memoizedState=null,o}function aT(o,l,f){o.lanes|=l;var h=o.alternate;h!==null&&(h.lanes|=l),ey(o.return,l,f)}function Iy(o,l,f,h,y){var _=o.memoizedState;_===null?o.memoizedState={isBackwards:l,rendering:null,renderingStartTime:0,last:h,tail:f,tailMode:y}:(_.isBackwards=l,_.rendering=null,_.renderingStartTime=0,_.last=h,_.tail=f,_.tailMode=y)}function oT(o,l,f){var h=l.pendingProps,y=h.revealOrder,_=h.tail;if(xn(o,l,h.children,f),h=nn.current,(h&2)!==0)h=h&1|2,l.flags|=128;else{if(o!==null&&(o.flags&128)!==0)e:for(o=l.child;o!==null;){if(o.tag===13)o.memoizedState!==null&&aT(o,f,l);else if(o.tag===19)aT(o,f,l);else if(o.child!==null){o.child.return=o,o=o.child;continue}if(o===l)break e;for(;o.sibling===null;){if(o.return===null||o.return===l)break e;o=o.return}o.sibling.return=o.return,o=o.sibling}h&=1}switch(X(nn,h),y){case"forwards":for(f=l.child,y=null;f!==null;)o=f.alternate,o!==null&&dp(o)===null&&(y=f),f=f.sibling;f=y,f===null?(y=l.child,l.child=null):(y=f.sibling,f.sibling=null),Iy(l,!1,y,f,_);break;case"backwards":for(f=null,y=l.child,l.child=null;y!==null;){if(o=y.alternate,o!==null&&dp(o)===null){l.child=y;break}o=y.sibling,y.sibling=f,f=y,y=o}Iy(l,!0,f,null,_);break;case"together":Iy(l,!1,null,null,void 0);break;default:l.memoizedState=null}return l.child}function zi(o,l,f){if(o!==null&&(l.dependencies=o.dependencies),La|=l.lanes,(f&l.childLanes)===0)if(o!==null){if(Bu(o,l,f,!1),(f&l.childLanes)===0)return null}else return null;if(o!==null&&l.child!==o.child)throw Error(r(153));if(l.child!==null){for(o=l.child,f=ki(o,o.pendingProps),l.child=f,f.return=l;o.sibling!==null;)o=o.sibling,f=f.sibling=ki(o,o.pendingProps),f.return=l;f.sibling=null}return l.child}function Ly(o,l){return(o.lanes&l)!==0?!0:(o=o.dependencies,!!(o!==null&&Zd(o)))}function FH(o,l,f){switch(l.tag){case 3:Se(l,l.stateNode.containerInfo),Aa(l,tn,o.memoizedState.cache),Iu();break;case 27:case 5:we(l);break;case 4:Se(l,l.stateNode.containerInfo);break;case 10:Aa(l,l.type,l.memoizedProps.value);break;case 13:var h=l.memoizedState;if(h!==null)return h.dehydrated!==null?(ja(l),l.flags|=128,null):(f&l.child.childLanes)!==0?iT(o,l,f):(ja(l),o=zi(o,l,f),o!==null?o.sibling:null);ja(l);break;case 19:var y=(o.flags&128)!==0;if(h=(f&l.childLanes)!==0,h||(Bu(o,l,f,!1),h=(f&l.childLanes)!==0),y){if(h)return oT(o,l,f);l.flags|=128}if(y=l.memoizedState,y!==null&&(y.rendering=null,y.tail=null,y.lastEffect=null),X(nn,nn.current),h)break;return null;case 22:case 23:return l.lanes=0,JR(o,l,f);case 24:Aa(l,tn,o.memoizedState.cache)}return zi(o,l,f)}function sT(o,l,f){if(o!==null)if(o.memoizedProps!==l.pendingProps)hn=!0;else{if(!Ly(o,f)&&(l.flags&128)===0)return hn=!1,FH(o,l,f);hn=(o.flags&131072)!==0}else hn=!1,mt&&(l.flags&1048576)!==0&&IA(l,Yd,l.index);switch(l.lanes=0,l.tag){case 16:e:{o=l.pendingProps;var h=l.elementType,y=h._init;if(h=y(h._payload),l.type=h,typeof h=="function")Wg(h)?(o=Do(h,o),l.tag=1,l=nT(null,l,h,o,f)):(l.tag=0,l=Py(null,l,h,o,f));else{if(h!=null){if(y=h.$$typeof,y===T){l.tag=11,l=ZR(null,l,h,o,f);break e}else if(y===P){l.tag=14,l=XR(null,l,h,o,f);break e}}throw l=Z(h)||h,Error(r(306,l,""))}}return l;case 0:return Py(o,l,l.type,l.pendingProps,f);case 1:return h=l.type,y=Do(h,l.pendingProps),nT(o,l,h,y,f);case 3:e:{if(Se(l,l.stateNode.containerInfo),o===null)throw Error(r(387));h=l.pendingProps;var _=l.memoizedState;y=_.element,sy(o,l),Gu(l,h,null,f);var j=l.memoizedState;if(h=j.cache,Aa(l,tn,h),h!==_.cache&&ty(l,[tn],f,!0),Hu(),h=j.element,_.isDehydrated)if(_={element:h,isDehydrated:!1,cache:j.cache},l.updateQueue.baseState=_,l.memoizedState=_,l.flags&256){l=rT(o,l,h,f);break e}else if(h!==y){y=xr(Error(r(424)),l),Lu(y),l=rT(o,l,h,f);break e}else{switch(o=l.stateNode.containerInfo,o.nodeType){case 9:o=o.body;break;default:o=o.nodeName==="HTML"?o.ownerDocument.body:o}for(Ht=Fr(o.firstChild),kn=l,mt=!0,Ro=null,ii=!0,f=BR(l,null,h,f),l.child=f;f;)f.flags=f.flags&-3|4096,f=f.sibling}else{if(Iu(),h===y){l=zi(o,l,f);break e}xn(o,l,h,f)}l=l.child}return l;case 26:return mp(o,l),o===null?(f=fM(l.type,null,l.pendingProps,null))?l.memoizedState=f:mt||(f=l.type,o=l.pendingProps,h=Mp(ve.current).createElement(f),h[Cn]=l,h[Vn]=o,Sn(h,f,o),pn(h),l.stateNode=h):l.memoizedState=fM(l.type,o.memoizedProps,l.pendingProps,o.memoizedState),null;case 27:return we(l),o===null&&mt&&(h=l.stateNode=lM(l.type,l.pendingProps,ve.current),kn=l,ii=!0,y=Ht,za(l.type)?(yb=y,Ht=Fr(h.firstChild)):Ht=y),xn(o,l,l.pendingProps.children,f),mp(o,l),o===null&&(l.flags|=4194304),l.child;case 5:return o===null&&mt&&((y=h=Ht)&&(h=h5(h,l.type,l.pendingProps,ii),h!==null?(l.stateNode=h,kn=l,Ht=Fr(h.firstChild),ii=!1,y=!0):y=!1),y||To(l)),we(l),y=l.type,_=l.pendingProps,j=o!==null?o.memoizedProps:null,h=_.children,hb(y,_)?h=null:j!==null&&hb(y,j)&&(l.flags|=32),l.memoizedState!==null&&(y=py(o,l,jH,null,null,f),hc._currentValue=y),mp(o,l),xn(o,l,h,f),l.child;case 6:return o===null&&mt&&((o=f=Ht)&&(f=m5(f,l.pendingProps,ii),f!==null?(l.stateNode=f,kn=l,Ht=null,o=!0):o=!1),o||To(l)),null;case 13:return iT(o,l,f);case 4:return Se(l,l.stateNode.containerInfo),h=l.pendingProps,o===null?l.child=Ys(l,null,h,f):xn(o,l,h,f),l.child;case 11:return ZR(o,l,l.type,l.pendingProps,f);case 7:return xn(o,l,l.pendingProps,f),l.child;case 8:return xn(o,l,l.pendingProps.children,f),l.child;case 12:return xn(o,l,l.pendingProps.children,f),l.child;case 10:return h=l.pendingProps,Aa(l,l.type,h.value),xn(o,l,h.children,f),l.child;case 9:return y=l.type._context,h=l.pendingProps.children,Po(l),y=On(y),h=h(y),l.flags|=1,xn(o,l,h,f),l.child;case 14:return XR(o,l,l.type,l.pendingProps,f);case 15:return QR(o,l,l.type,l.pendingProps,f);case 19:return oT(o,l,f);case 31:return h=l.pendingProps,f=l.mode,h={mode:h.mode,children:h.children},o===null?(f=vp(h,f),f.ref=l.ref,l.child=f,f.return=l,l=f):(f=ki(o.child,h),f.ref=l.ref,l.child=f,f.return=l,l=f),l;case 22:return JR(o,l,f);case 24:return Po(l),h=On(tn),o===null?(y=iy(),y===null&&(y=Nt,_=ny(),y.pooledCache=_,_.refCount++,_!==null&&(y.pooledCacheLanes|=f),y=_),l.memoizedState={parent:h,cache:y},oy(l),Aa(l,tn,y)):((o.lanes&f)!==0&&(sy(o,l),Gu(l,null,null,f),Hu()),y=o.memoizedState,_=l.memoizedState,y.parent!==h?(y={parent:h,cache:h},l.memoizedState=y,l.lanes===0&&(l.memoizedState=l.updateQueue.baseState=y),Aa(l,tn,h)):(h=_.cache,Aa(l,tn,h),h!==y.cache&&ty(l,[tn],f,!0))),xn(o,l,l.pendingProps.children,f),l.child;case 29:throw l.pendingProps}throw Error(r(156,l.tag))}function Ui(o){o.flags|=4}function lT(o,l){if(l.type!=="stylesheet"||(l.state.loading&4)!==0)o.flags&=-16777217;else if(o.flags|=16777216,!vM(l)){if(l=Er.current,l!==null&&((lt&4194048)===lt?ai!==null:(lt&62914560)!==lt&&(lt&536870912)===0||l!==ai))throw Uu=ay,HA;o.flags|=8192}}function gp(o,l){l!==null&&(o.flags|=4),o.flags&16384&&(l=o.tag!==22?FO():536870912,o.lanes|=l,Js|=l)}function Ju(o,l){if(!mt)switch(o.tailMode){case"hidden":l=o.tail;for(var f=null;l!==null;)l.alternate!==null&&(f=l),l=l.sibling;f===null?o.tail=null:f.sibling=null;break;case"collapsed":f=o.tail;for(var h=null;f!==null;)f.alternate!==null&&(h=f),f=f.sibling;h===null?l||o.tail===null?o.tail=null:o.tail.sibling=null:h.sibling=null}}function Ft(o){var l=o.alternate!==null&&o.alternate.child===o.child,f=0,h=0;if(l)for(var y=o.child;y!==null;)f|=y.lanes|y.childLanes,h|=y.subtreeFlags&65011712,h|=y.flags&65011712,y.return=o,y=y.sibling;else for(y=o.child;y!==null;)f|=y.lanes|y.childLanes,h|=y.subtreeFlags,h|=y.flags,y.return=o,y=y.sibling;return o.subtreeFlags|=h,o.childLanes=f,l}function qH(o,l,f){var h=l.pendingProps;switch(Xg(l),l.tag){case 31:case 16:case 15:case 0:case 11:case 7:case 8:case 12:case 9:case 14:return Ft(l),null;case 1:return Ft(l),null;case 3:return f=l.stateNode,h=null,o!==null&&(h=o.memoizedState.cache),l.memoizedState.cache!==h&&(l.flags|=2048),Bi(tn),ce(),f.pendingContext&&(f.context=f.pendingContext,f.pendingContext=null),(o===null||o.child===null)&&($u(l)?Ui(l):o===null||o.memoizedState.isDehydrated&&(l.flags&256)===0||(l.flags|=1024,FA())),Ft(l),null;case 26:return f=l.memoizedState,o===null?(Ui(l),f!==null?(Ft(l),lT(l,f)):(Ft(l),l.flags&=-16777217)):f?f!==o.memoizedState?(Ui(l),Ft(l),lT(l,f)):(Ft(l),l.flags&=-16777217):(o.memoizedProps!==h&&Ui(l),Ft(l),l.flags&=-16777217),null;case 27:_e(l),f=ve.current;var y=l.type;if(o!==null&&l.stateNode!=null)o.memoizedProps!==h&&Ui(l);else{if(!h){if(l.stateNode===null)throw Error(r(166));return Ft(l),null}o=oe.current,$u(l)?LA(l):(o=lM(y,h,f),l.stateNode=o,Ui(l))}return Ft(l),null;case 5:if(_e(l),f=l.type,o!==null&&l.stateNode!=null)o.memoizedProps!==h&&Ui(l);else{if(!h){if(l.stateNode===null)throw Error(r(166));return Ft(l),null}if(o=oe.current,$u(l))LA(l);else{switch(y=Mp(ve.current),o){case 1:o=y.createElementNS("http://www.w3.org/2000/svg",f);break;case 2:o=y.createElementNS("http://www.w3.org/1998/Math/MathML",f);break;default:switch(f){case"svg":o=y.createElementNS("http://www.w3.org/2000/svg",f);break;case"math":o=y.createElementNS("http://www.w3.org/1998/Math/MathML",f);break;case"script":o=y.createElement("div"),o.innerHTML="<script><\/script>",o=o.removeChild(o.firstChild);break;case"select":o=typeof h.is=="string"?y.createElement("select",{is:h.is}):y.createElement("select"),h.multiple?o.multiple=!0:h.size&&(o.size=h.size);break;default:o=typeof h.is=="string"?y.createElement(f,{is:h.is}):y.createElement(f)}}o[Cn]=l,o[Vn]=h;e:for(y=l.child;y!==null;){if(y.tag===5||y.tag===6)o.appendChild(y.stateNode);else if(y.tag!==4&&y.tag!==27&&y.child!==null){y.child.return=y,y=y.child;continue}if(y===l)break e;for(;y.sibling===null;){if(y.return===null||y.return===l)break e;y=y.return}y.sibling.return=y.return,y=y.sibling}l.stateNode=o;e:switch(Sn(o,f,h),f){case"button":case"input":case"select":case"textarea":o=!!h.autoFocus;break e;case"img":o=!0;break e;default:o=!1}o&&Ui(l)}}return Ft(l),l.flags&=-16777217,null;case 6:if(o&&l.stateNode!=null)o.memoizedProps!==h&&Ui(l);else{if(typeof h!="string"&&l.stateNode===null)throw Error(r(166));if(o=ve.current,$u(l)){if(o=l.stateNode,f=l.memoizedProps,h=null,y=kn,y!==null)switch(y.tag){case 27:case 5:h=y.memoizedProps}o[Cn]=l,o=!!(o.nodeValue===f||h!==null&&h.suppressHydrationWarning===!0||tM(o.nodeValue,f)),o||To(l)}else o=Mp(o).createTextNode(h),o[Cn]=l,l.stateNode=o}return Ft(l),null;case 13:if(h=l.memoizedState,o===null||o.memoizedState!==null&&o.memoizedState.dehydrated!==null){if(y=$u(l),h!==null&&h.dehydrated!==null){if(o===null){if(!y)throw Error(r(318));if(y=l.memoizedState,y=y!==null?y.dehydrated:null,!y)throw Error(r(317));y[Cn]=l}else Iu(),(l.flags&128)===0&&(l.memoizedState=null),l.flags|=4;Ft(l),y=!1}else y=FA(),o!==null&&o.memoizedState!==null&&(o.memoizedState.hydrationErrors=y),y=!0;if(!y)return l.flags&256?(qi(l),l):(qi(l),null)}if(qi(l),(l.flags&128)!==0)return l.lanes=f,l;if(f=h!==null,o=o!==null&&o.memoizedState!==null,f){h=l.child,y=null,h.alternate!==null&&h.alternate.memoizedState!==null&&h.alternate.memoizedState.cachePool!==null&&(y=h.alternate.memoizedState.cachePool.pool);var _=null;h.memoizedState!==null&&h.memoizedState.cachePool!==null&&(_=h.memoizedState.cachePool.pool),_!==y&&(h.flags|=2048)}return f!==o&&f&&(l.child.flags|=8192),gp(l,l.updateQueue),Ft(l),null;case 4:return ce(),o===null&&ub(l.stateNode.containerInfo),Ft(l),null;case 10:return Bi(l.type),Ft(l),null;case 19:if(q(nn),y=l.memoizedState,y===null)return Ft(l),null;if(h=(l.flags&128)!==0,_=y.rendering,_===null)if(h)Ju(y,!1);else{if(Gt!==0||o!==null&&(o.flags&128)!==0)for(o=l.child;o!==null;){if(_=dp(o),_!==null){for(l.flags|=128,Ju(y,!1),o=_.updateQueue,l.updateQueue=o,gp(l,o),l.subtreeFlags=0,o=f,f=l.child;f!==null;)$A(f,o),f=f.sibling;return X(nn,nn.current&1|2),l.child}o=o.sibling}y.tail!==null&&it()>xp&&(l.flags|=128,h=!0,Ju(y,!1),l.lanes=4194304)}else{if(!h)if(o=dp(_),o!==null){if(l.flags|=128,h=!0,o=o.updateQueue,l.updateQueue=o,gp(l,o),Ju(y,!0),y.tail===null&&y.tailMode==="hidden"&&!_.alternate&&!mt)return Ft(l),null}else 2*it()-y.renderingStartTime>xp&&f!==536870912&&(l.flags|=128,h=!0,Ju(y,!1),l.lanes=4194304);y.isBackwards?(_.sibling=l.child,l.child=_):(o=y.last,o!==null?o.sibling=_:l.child=_,y.last=_)}return y.tail!==null?(l=y.tail,y.rendering=l,y.tail=l.sibling,y.renderingStartTime=it(),l.sibling=null,o=nn.current,X(nn,h?o&1|2:o&1),l):(Ft(l),null);case 22:case 23:return qi(l),fy(),h=l.memoizedState!==null,o!==null?o.memoizedState!==null!==h&&(l.flags|=8192):h&&(l.flags|=8192),h?(f&536870912)!==0&&(l.flags&128)===0&&(Ft(l),l.subtreeFlags&6&&(l.flags|=8192)):Ft(l),f=l.updateQueue,f!==null&&gp(l,f.retryQueue),f=null,o!==null&&o.memoizedState!==null&&o.memoizedState.cachePool!==null&&(f=o.memoizedState.cachePool.pool),h=null,l.memoizedState!==null&&l.memoizedState.cachePool!==null&&(h=l.memoizedState.cachePool.pool),h!==f&&(l.flags|=2048),o!==null&&q(jo),null;case 24:return f=null,o!==null&&(f=o.memoizedState.cache),l.memoizedState.cache!==f&&(l.flags|=2048),Bi(tn),Ft(l),null;case 25:return null;case 30:return null}throw Error(r(156,l.tag))}function zH(o,l){switch(Xg(l),l.tag){case 1:return o=l.flags,o&65536?(l.flags=o&-65537|128,l):null;case 3:return Bi(tn),ce(),o=l.flags,(o&65536)!==0&&(o&128)===0?(l.flags=o&-65537|128,l):null;case 26:case 27:case 5:return _e(l),null;case 13:if(qi(l),o=l.memoizedState,o!==null&&o.dehydrated!==null){if(l.alternate===null)throw Error(r(340));Iu()}return o=l.flags,o&65536?(l.flags=o&-65537|128,l):null;case 19:return q(nn),null;case 4:return ce(),null;case 10:return Bi(l.type),null;case 22:case 23:return qi(l),fy(),o!==null&&q(jo),o=l.flags,o&65536?(l.flags=o&-65537|128,l):null;case 24:return Bi(tn),null;case 25:return null;default:return null}}function uT(o,l){switch(Xg(l),l.tag){case 3:Bi(tn),ce();break;case 26:case 27:case 5:_e(l);break;case 4:ce();break;case 13:qi(l);break;case 19:q(nn);break;case 10:Bi(l.type);break;case 22:case 23:qi(l),fy(),o!==null&&q(jo);break;case 24:Bi(tn)}}function ec(o,l){try{var f=l.updateQueue,h=f!==null?f.lastEffect:null;if(h!==null){var y=h.next;f=y;do{if((f.tag&o)===o){h=void 0;var _=f.create,j=f.inst;h=_(),j.destroy=h}f=f.next}while(f!==y)}}catch(D){Mt(l,l.return,D)}}function Da(o,l,f){try{var h=l.updateQueue,y=h!==null?h.lastEffect:null;if(y!==null){var _=y.next;h=_;do{if((h.tag&o)===o){var j=h.inst,D=j.destroy;if(D!==void 0){j.destroy=void 0,y=l;var U=f,ne=D;try{ne()}catch(he){Mt(y,U,he)}}}h=h.next}while(h!==_)}}catch(he){Mt(l,l.return,he)}}function cT(o){var l=o.updateQueue;if(l!==null){var f=o.stateNode;try{XA(l,f)}catch(h){Mt(o,o.return,h)}}}function fT(o,l,f){f.props=Do(o.type,o.memoizedProps),f.state=o.memoizedState;try{f.componentWillUnmount()}catch(h){Mt(o,l,h)}}function tc(o,l){try{var f=o.ref;if(f!==null){switch(o.tag){case 26:case 27:case 5:var h=o.stateNode;break;case 30:h=o.stateNode;break;default:h=o.stateNode}typeof f=="function"?o.refCleanup=f(h):f.current=h}}catch(y){Mt(o,l,y)}}function oi(o,l){var f=o.ref,h=o.refCleanup;if(f!==null)if(typeof h=="function")try{h()}catch(y){Mt(o,l,y)}finally{o.refCleanup=null,o=o.alternate,o!=null&&(o.refCleanup=null)}else if(typeof f=="function")try{f(null)}catch(y){Mt(o,l,y)}else f.current=null}function dT(o){var l=o.type,f=o.memoizedProps,h=o.stateNode;try{e:switch(l){case"button":case"input":case"select":case"textarea":f.autoFocus&&h.focus();break e;case"img":f.src?h.src=f.src:f.srcSet&&(h.srcset=f.srcSet)}}catch(y){Mt(o,o.return,y)}}function By(o,l,f){try{var h=o.stateNode;u5(h,o.type,f,l),h[Vn]=l}catch(y){Mt(o,o.return,y)}}function pT(o){return o.tag===5||o.tag===3||o.tag===26||o.tag===27&&za(o.type)||o.tag===4}function Fy(o){e:for(;;){for(;o.sibling===null;){if(o.return===null||pT(o.return))return null;o=o.return}for(o.sibling.return=o.return,o=o.sibling;o.tag!==5&&o.tag!==6&&o.tag!==18;){if(o.tag===27&&za(o.type)||o.flags&2||o.child===null||o.tag===4)continue e;o.child.return=o,o=o.child}if(!(o.flags&2))return o.stateNode}}function qy(o,l,f){var h=o.tag;if(h===5||h===6)o=o.stateNode,l?(f.nodeType===9?f.body:f.nodeName==="HTML"?f.ownerDocument.body:f).insertBefore(o,l):(l=f.nodeType===9?f.body:f.nodeName==="HTML"?f.ownerDocument.body:f,l.appendChild(o),f=f._reactRootContainer,f!=null||l.onclick!==null||(l.onclick=Tp));else if(h!==4&&(h===27&&za(o.type)&&(f=o.stateNode,l=null),o=o.child,o!==null))for(qy(o,l,f),o=o.sibling;o!==null;)qy(o,l,f),o=o.sibling}function yp(o,l,f){var h=o.tag;if(h===5||h===6)o=o.stateNode,l?f.insertBefore(o,l):f.appendChild(o);else if(h!==4&&(h===27&&za(o.type)&&(f=o.stateNode),o=o.child,o!==null))for(yp(o,l,f),o=o.sibling;o!==null;)yp(o,l,f),o=o.sibling}function hT(o){var l=o.stateNode,f=o.memoizedProps;try{for(var h=o.type,y=l.attributes;y.length;)l.removeAttributeNode(y[0]);Sn(l,h,f),l[Cn]=o,l[Vn]=f}catch(_){Mt(o,o.return,_)}}var Vi=!1,Yt=!1,zy=!1,mT=typeof WeakSet=="function"?WeakSet:Set,mn=null;function UH(o,l){if(o=o.containerInfo,db=$p,o=OA(o),Fg(o)){if("selectionStart"in o)var f={start:o.selectionStart,end:o.selectionEnd};else e:{f=(f=o.ownerDocument)&&f.defaultView||window;var h=f.getSelection&&f.getSelection();if(h&&h.rangeCount!==0){f=h.anchorNode;var y=h.anchorOffset,_=h.focusNode;h=h.focusOffset;try{f.nodeType,_.nodeType}catch{f=null;break e}var j=0,D=-1,U=-1,ne=0,he=0,ge=o,ie=null;t:for(;;){for(var ae;ge!==f||y!==0&&ge.nodeType!==3||(D=j+y),ge!==_||h!==0&&ge.nodeType!==3||(U=j+h),ge.nodeType===3&&(j+=ge.nodeValue.length),(ae=ge.firstChild)!==null;)ie=ge,ge=ae;for(;;){if(ge===o)break t;if(ie===f&&++ne===y&&(D=j),ie===_&&++he===h&&(U=j),(ae=ge.nextSibling)!==null)break;ge=ie,ie=ge.parentNode}ge=ae}f=D===-1||U===-1?null:{start:D,end:U}}else f=null}f=f||{start:0,end:0}}else f=null;for(pb={focusedElem:o,selectionRange:f},$p=!1,mn=l;mn!==null;)if(l=mn,o=l.child,(l.subtreeFlags&1024)!==0&&o!==null)o.return=l,mn=o;else for(;mn!==null;){switch(l=mn,_=l.alternate,o=l.flags,l.tag){case 0:break;case 11:case 15:break;case 1:if((o&1024)!==0&&_!==null){o=void 0,f=l,y=_.memoizedProps,_=_.memoizedState,h=f.stateNode;try{var Ve=Do(f.type,y,f.elementType===f.type);o=h.getSnapshotBeforeUpdate(Ve,_),h.__reactInternalSnapshotBeforeUpdate=o}catch(Fe){Mt(f,f.return,Fe)}}break;case 3:if((o&1024)!==0){if(o=l.stateNode.containerInfo,f=o.nodeType,f===9)vb(o);else if(f===1)switch(o.nodeName){case"HEAD":case"HTML":case"BODY":vb(o);break;default:o.textContent=""}}break;case 5:case 26:case 27:case 6:case 4:case 17:break;default:if((o&1024)!==0)throw Error(r(163))}if(o=l.sibling,o!==null){o.return=l.return,mn=o;break}mn=l.return}}function vT(o,l,f){var h=f.flags;switch(f.tag){case 0:case 11:case 15:ka(o,f),h&4&&ec(5,f);break;case 1:if(ka(o,f),h&4)if(o=f.stateNode,l===null)try{o.componentDidMount()}catch(j){Mt(f,f.return,j)}else{var y=Do(f.type,l.memoizedProps);l=l.memoizedState;try{o.componentDidUpdate(y,l,o.__reactInternalSnapshotBeforeUpdate)}catch(j){Mt(f,f.return,j)}}h&64&&cT(f),h&512&&tc(f,f.return);break;case 3:if(ka(o,f),h&64&&(o=f.updateQueue,o!==null)){if(l=null,f.child!==null)switch(f.child.tag){case 27:case 5:l=f.child.stateNode;break;case 1:l=f.child.stateNode}try{XA(o,l)}catch(j){Mt(f,f.return,j)}}break;case 27:l===null&&h&4&&hT(f);case 26:case 5:ka(o,f),l===null&&h&4&&dT(f),h&512&&tc(f,f.return);break;case 12:ka(o,f);break;case 13:ka(o,f),h&4&&bT(o,f),h&64&&(o=f.memoizedState,o!==null&&(o=o.dehydrated,o!==null&&(f=QH.bind(null,f),v5(o,f))));break;case 22:if(h=f.memoizedState!==null||Vi,!h){l=l!==null&&l.memoizedState!==null||Yt,y=Vi;var _=Yt;Vi=h,(Yt=l)&&!_?$a(o,f,(f.subtreeFlags&8772)!==0):ka(o,f),Vi=y,Yt=_}break;case 30:break;default:ka(o,f)}}function gT(o){var l=o.alternate;l!==null&&(o.alternate=null,gT(l)),o.child=null,o.deletions=null,o.sibling=null,o.tag===5&&(l=o.stateNode,l!==null&&wg(l)),o.stateNode=null,o.return=null,o.dependencies=null,o.memoizedProps=null,o.memoizedState=null,o.pendingProps=null,o.stateNode=null,o.updateQueue=null}var It=null,Wn=!1;function Hi(o,l,f){for(f=f.child;f!==null;)yT(o,l,f),f=f.sibling}function yT(o,l,f){if(ye&&typeof ye.onCommitFiberUnmount=="function")try{ye.onCommitFiberUnmount(Ee,f)}catch{}switch(f.tag){case 26:Yt||oi(f,l),Hi(o,l,f),f.memoizedState?f.memoizedState.count--:f.stateNode&&(f=f.stateNode,f.parentNode.removeChild(f));break;case 27:Yt||oi(f,l);var h=It,y=Wn;za(f.type)&&(It=f.stateNode,Wn=!1),Hi(o,l,f),cc(f.stateNode),It=h,Wn=y;break;case 5:Yt||oi(f,l);case 6:if(h=It,y=Wn,It=null,Hi(o,l,f),It=h,Wn=y,It!==null)if(Wn)try{(It.nodeType===9?It.body:It.nodeName==="HTML"?It.ownerDocument.body:It).removeChild(f.stateNode)}catch(_){Mt(f,l,_)}else try{It.removeChild(f.stateNode)}catch(_){Mt(f,l,_)}break;case 18:It!==null&&(Wn?(o=It,oM(o.nodeType===9?o.body:o.nodeName==="HTML"?o.ownerDocument.body:o,f.stateNode),yc(o)):oM(It,f.stateNode));break;case 4:h=It,y=Wn,It=f.stateNode.containerInfo,Wn=!0,Hi(o,l,f),It=h,Wn=y;break;case 0:case 11:case 14:case 15:Yt||Da(2,f,l),Yt||Da(4,f,l),Hi(o,l,f);break;case 1:Yt||(oi(f,l),h=f.stateNode,typeof h.componentWillUnmount=="function"&&fT(f,l,h)),Hi(o,l,f);break;case 21:Hi(o,l,f);break;case 22:Yt=(h=Yt)||f.memoizedState!==null,Hi(o,l,f),Yt=h;break;default:Hi(o,l,f)}}function bT(o,l){if(l.memoizedState===null&&(o=l.alternate,o!==null&&(o=o.memoizedState,o!==null&&(o=o.dehydrated,o!==null))))try{yc(o)}catch(f){Mt(l,l.return,f)}}function VH(o){switch(o.tag){case 13:case 19:var l=o.stateNode;return l===null&&(l=o.stateNode=new mT),l;case 22:return o=o.stateNode,l=o._retryCache,l===null&&(l=o._retryCache=new mT),l;default:throw Error(r(435,o.tag))}}function Uy(o,l){var f=VH(o);l.forEach(function(h){var y=JH.bind(null,o,h);f.has(h)||(f.add(h),h.then(y,y))})}function tr(o,l){var f=l.deletions;if(f!==null)for(var h=0;h<f.length;h++){var y=f[h],_=o,j=l,D=j;e:for(;D!==null;){switch(D.tag){case 27:if(za(D.type)){It=D.stateNode,Wn=!1;break e}break;case 5:It=D.stateNode,Wn=!1;break e;case 3:case 4:It=D.stateNode.containerInfo,Wn=!0;break e}D=D.return}if(It===null)throw Error(r(160));yT(_,j,y),It=null,Wn=!1,_=y.alternate,_!==null&&(_.return=null),y.return=null}if(l.subtreeFlags&13878)for(l=l.child;l!==null;)xT(l,o),l=l.sibling}var Br=null;function xT(o,l){var f=o.alternate,h=o.flags;switch(o.tag){case 0:case 11:case 14:case 15:tr(l,o),nr(o),h&4&&(Da(3,o,o.return),ec(3,o),Da(5,o,o.return));break;case 1:tr(l,o),nr(o),h&512&&(Yt||f===null||oi(f,f.return)),h&64&&Vi&&(o=o.updateQueue,o!==null&&(h=o.callbacks,h!==null&&(f=o.shared.hiddenCallbacks,o.shared.hiddenCallbacks=f===null?h:f.concat(h))));break;case 26:var y=Br;if(tr(l,o),nr(o),h&512&&(Yt||f===null||oi(f,f.return)),h&4){var _=f!==null?f.memoizedState:null;if(h=o.memoizedState,f===null)if(h===null)if(o.stateNode===null){e:{h=o.type,f=o.memoizedProps,y=y.ownerDocument||y;t:switch(h){case"title":_=y.getElementsByTagName("title")[0],(!_||_[Cu]||_[Cn]||_.namespaceURI==="http://www.w3.org/2000/svg"||_.hasAttribute("itemprop"))&&(_=y.createElement(h),y.head.insertBefore(_,y.querySelector("head > title"))),Sn(_,h,f),_[Cn]=o,pn(_),h=_;break e;case"link":var j=hM("link","href",y).get(h+(f.href||""));if(j){for(var D=0;D<j.length;D++)if(_=j[D],_.getAttribute("href")===(f.href==null||f.href===""?null:f.href)&&_.getAttribute("rel")===(f.rel==null?null:f.rel)&&_.getAttribute("title")===(f.title==null?null:f.title)&&_.getAttribute("crossorigin")===(f.crossOrigin==null?null:f.crossOrigin)){j.splice(D,1);break t}}_=y.createElement(h),Sn(_,h,f),y.head.appendChild(_);break;case"meta":if(j=hM("meta","content",y).get(h+(f.content||""))){for(D=0;D<j.length;D++)if(_=j[D],_.getAttribute("content")===(f.content==null?null:""+f.content)&&_.getAttribute("name")===(f.name==null?null:f.name)&&_.getAttribute("property")===(f.property==null?null:f.property)&&_.getAttribute("http-equiv")===(f.httpEquiv==null?null:f.httpEquiv)&&_.getAttribute("charset")===(f.charSet==null?null:f.charSet)){j.splice(D,1);break t}}_=y.createElement(h),Sn(_,h,f),y.head.appendChild(_);break;default:throw Error(r(468,h))}_[Cn]=o,pn(_),h=_}o.stateNode=h}else mM(y,o.type,o.stateNode);else o.stateNode=pM(y,h,o.memoizedProps);else _!==h?(_===null?f.stateNode!==null&&(f=f.stateNode,f.parentNode.removeChild(f)):_.count--,h===null?mM(y,o.type,o.stateNode):pM(y,h,o.memoizedProps)):h===null&&o.stateNode!==null&&By(o,o.memoizedProps,f.memoizedProps)}break;case 27:tr(l,o),nr(o),h&512&&(Yt||f===null||oi(f,f.return)),f!==null&&h&4&&By(o,o.memoizedProps,f.memoizedProps);break;case 5:if(tr(l,o),nr(o),h&512&&(Yt||f===null||oi(f,f.return)),o.flags&32){y=o.stateNode;try{Ps(y,"")}catch(ae){Mt(o,o.return,ae)}}h&4&&o.stateNode!=null&&(y=o.memoizedProps,By(o,y,f!==null?f.memoizedProps:y)),h&1024&&(zy=!0);break;case 6:if(tr(l,o),nr(o),h&4){if(o.stateNode===null)throw Error(r(162));h=o.memoizedProps,f=o.stateNode;try{f.nodeValue=h}catch(ae){Mt(o,o.return,ae)}}break;case 3:if(Np=null,y=Br,Br=Pp(l.containerInfo),tr(l,o),Br=y,nr(o),h&4&&f!==null&&f.memoizedState.isDehydrated)try{yc(l.containerInfo)}catch(ae){Mt(o,o.return,ae)}zy&&(zy=!1,wT(o));break;case 4:h=Br,Br=Pp(o.stateNode.containerInfo),tr(l,o),nr(o),Br=h;break;case 12:tr(l,o),nr(o);break;case 13:tr(l,o),nr(o),o.child.flags&8192&&o.memoizedState!==null!=(f!==null&&f.memoizedState!==null)&&(Yy=it()),h&4&&(h=o.updateQueue,h!==null&&(o.updateQueue=null,Uy(o,h)));break;case 22:y=o.memoizedState!==null;var U=f!==null&&f.memoizedState!==null,ne=Vi,he=Yt;if(Vi=ne||y,Yt=he||U,tr(l,o),Yt=he,Vi=ne,nr(o),h&8192)e:for(l=o.stateNode,l._visibility=y?l._visibility&-2:l._visibility|1,y&&(f===null||U||Vi||Yt||ko(o)),f=null,l=o;;){if(l.tag===5||l.tag===26){if(f===null){U=f=l;try{if(_=U.stateNode,y)j=_.style,typeof j.setProperty=="function"?j.setProperty("display","none","important"):j.display="none";else{D=U.stateNode;var ge=U.memoizedProps.style,ie=ge!=null&&ge.hasOwnProperty("display")?ge.display:null;D.style.display=ie==null||typeof ie=="boolean"?"":(""+ie).trim()}}catch(ae){Mt(U,U.return,ae)}}}else if(l.tag===6){if(f===null){U=l;try{U.stateNode.nodeValue=y?"":U.memoizedProps}catch(ae){Mt(U,U.return,ae)}}}else if((l.tag!==22&&l.tag!==23||l.memoizedState===null||l===o)&&l.child!==null){l.child.return=l,l=l.child;continue}if(l===o)break e;for(;l.sibling===null;){if(l.return===null||l.return===o)break e;f===l&&(f=null),l=l.return}f===l&&(f=null),l.sibling.return=l.return,l=l.sibling}h&4&&(h=o.updateQueue,h!==null&&(f=h.retryQueue,f!==null&&(h.retryQueue=null,Uy(o,f))));break;case 19:tr(l,o),nr(o),h&4&&(h=o.updateQueue,h!==null&&(o.updateQueue=null,Uy(o,h)));break;case 30:break;case 21:break;default:tr(l,o),nr(o)}}function nr(o){var l=o.flags;if(l&2){try{for(var f,h=o.return;h!==null;){if(pT(h)){f=h;break}h=h.return}if(f==null)throw Error(r(160));switch(f.tag){case 27:var y=f.stateNode,_=Fy(o);yp(o,_,y);break;case 5:var j=f.stateNode;f.flags&32&&(Ps(j,""),f.flags&=-33);var D=Fy(o);yp(o,D,j);break;case 3:case 4:var U=f.stateNode.containerInfo,ne=Fy(o);qy(o,ne,U);break;default:throw Error(r(161))}}catch(he){Mt(o,o.return,he)}o.flags&=-3}l&4096&&(o.flags&=-4097)}function wT(o){if(o.subtreeFlags&1024)for(o=o.child;o!==null;){var l=o;wT(l),l.tag===5&&l.flags&1024&&l.stateNode.reset(),o=o.sibling}}function ka(o,l){if(l.subtreeFlags&8772)for(l=l.child;l!==null;)vT(o,l.alternate,l),l=l.sibling}function ko(o){for(o=o.child;o!==null;){var l=o;switch(l.tag){case 0:case 11:case 14:case 15:Da(4,l,l.return),ko(l);break;case 1:oi(l,l.return);var f=l.stateNode;typeof f.componentWillUnmount=="function"&&fT(l,l.return,f),ko(l);break;case 27:cc(l.stateNode);case 26:case 5:oi(l,l.return),ko(l);break;case 22:l.memoizedState===null&&ko(l);break;case 30:ko(l);break;default:ko(l)}o=o.sibling}}function $a(o,l,f){for(f=f&&(l.subtreeFlags&8772)!==0,l=l.child;l!==null;){var h=l.alternate,y=o,_=l,j=_.flags;switch(_.tag){case 0:case 11:case 15:$a(y,_,f),ec(4,_);break;case 1:if($a(y,_,f),h=_,y=h.stateNode,typeof y.componentDidMount=="function")try{y.componentDidMount()}catch(ne){Mt(h,h.return,ne)}if(h=_,y=h.updateQueue,y!==null){var D=h.stateNode;try{var U=y.shared.hiddenCallbacks;if(U!==null)for(y.shared.hiddenCallbacks=null,y=0;y<U.length;y++)ZA(U[y],D)}catch(ne){Mt(h,h.return,ne)}}f&&j&64&&cT(_),tc(_,_.return);break;case 27:hT(_);case 26:case 5:$a(y,_,f),f&&h===null&&j&4&&dT(_),tc(_,_.return);break;case 12:$a(y,_,f);break;case 13:$a(y,_,f),f&&j&4&&bT(y,_);break;case 22:_.memoizedState===null&&$a(y,_,f),tc(_,_.return);break;case 30:break;default:$a(y,_,f)}l=l.sibling}}function Vy(o,l){var f=null;o!==null&&o.memoizedState!==null&&o.memoizedState.cachePool!==null&&(f=o.memoizedState.cachePool.pool),o=null,l.memoizedState!==null&&l.memoizedState.cachePool!==null&&(o=l.memoizedState.cachePool.pool),o!==f&&(o!=null&&o.refCount++,f!=null&&Fu(f))}function Hy(o,l){o=null,l.alternate!==null&&(o=l.alternate.memoizedState.cache),l=l.memoizedState.cache,l!==o&&(l.refCount++,o!=null&&Fu(o))}function si(o,l,f,h){if(l.subtreeFlags&10256)for(l=l.child;l!==null;)ST(o,l,f,h),l=l.sibling}function ST(o,l,f,h){var y=l.flags;switch(l.tag){case 0:case 11:case 15:si(o,l,f,h),y&2048&&ec(9,l);break;case 1:si(o,l,f,h);break;case 3:si(o,l,f,h),y&2048&&(o=null,l.alternate!==null&&(o=l.alternate.memoizedState.cache),l=l.memoizedState.cache,l!==o&&(l.refCount++,o!=null&&Fu(o)));break;case 12:if(y&2048){si(o,l,f,h),o=l.stateNode;try{var _=l.memoizedProps,j=_.id,D=_.onPostCommit;typeof D=="function"&&D(j,l.alternate===null?"mount":"update",o.passiveEffectDuration,-0)}catch(U){Mt(l,l.return,U)}}else si(o,l,f,h);break;case 13:si(o,l,f,h);break;case 23:break;case 22:_=l.stateNode,j=l.alternate,l.memoizedState!==null?_._visibility&2?si(o,l,f,h):nc(o,l):_._visibility&2?si(o,l,f,h):(_._visibility|=2,Zs(o,l,f,h,(l.subtreeFlags&10256)!==0)),y&2048&&Vy(j,l);break;case 24:si(o,l,f,h),y&2048&&Hy(l.alternate,l);break;default:si(o,l,f,h)}}function Zs(o,l,f,h,y){for(y=y&&(l.subtreeFlags&10256)!==0,l=l.child;l!==null;){var _=o,j=l,D=f,U=h,ne=j.flags;switch(j.tag){case 0:case 11:case 15:Zs(_,j,D,U,y),ec(8,j);break;case 23:break;case 22:var he=j.stateNode;j.memoizedState!==null?he._visibility&2?Zs(_,j,D,U,y):nc(_,j):(he._visibility|=2,Zs(_,j,D,U,y)),y&&ne&2048&&Vy(j.alternate,j);break;case 24:Zs(_,j,D,U,y),y&&ne&2048&&Hy(j.alternate,j);break;default:Zs(_,j,D,U,y)}l=l.sibling}}function nc(o,l){if(l.subtreeFlags&10256)for(l=l.child;l!==null;){var f=o,h=l,y=h.flags;switch(h.tag){case 22:nc(f,h),y&2048&&Vy(h.alternate,h);break;case 24:nc(f,h),y&2048&&Hy(h.alternate,h);break;default:nc(f,h)}l=l.sibling}}var rc=8192;function Xs(o){if(o.subtreeFlags&rc)for(o=o.child;o!==null;)_T(o),o=o.sibling}function _T(o){switch(o.tag){case 26:Xs(o),o.flags&rc&&o.memoizedState!==null&&T5(Br,o.memoizedState,o.memoizedProps);break;case 5:Xs(o);break;case 3:case 4:var l=Br;Br=Pp(o.stateNode.containerInfo),Xs(o),Br=l;break;case 22:o.memoizedState===null&&(l=o.alternate,l!==null&&l.memoizedState!==null?(l=rc,rc=16777216,Xs(o),rc=l):Xs(o));break;default:Xs(o)}}function ET(o){var l=o.alternate;if(l!==null&&(o=l.child,o!==null)){l.child=null;do l=o.sibling,o.sibling=null,o=l;while(o!==null)}}function ic(o){var l=o.deletions;if((o.flags&16)!==0){if(l!==null)for(var f=0;f<l.length;f++){var h=l[f];mn=h,OT(h,o)}ET(o)}if(o.subtreeFlags&10256)for(o=o.child;o!==null;)CT(o),o=o.sibling}function CT(o){switch(o.tag){case 0:case 11:case 15:ic(o),o.flags&2048&&Da(9,o,o.return);break;case 3:ic(o);break;case 12:ic(o);break;case 22:var l=o.stateNode;o.memoizedState!==null&&l._visibility&2&&(o.return===null||o.return.tag!==13)?(l._visibility&=-3,bp(o)):ic(o);break;default:ic(o)}}function bp(o){var l=o.deletions;if((o.flags&16)!==0){if(l!==null)for(var f=0;f<l.length;f++){var h=l[f];mn=h,OT(h,o)}ET(o)}for(o=o.child;o!==null;){switch(l=o,l.tag){case 0:case 11:case 15:Da(8,l,l.return),bp(l);break;case 22:f=l.stateNode,f._visibility&2&&(f._visibility&=-3,bp(l));break;default:bp(l)}o=o.sibling}}function OT(o,l){for(;mn!==null;){var f=mn;switch(f.tag){case 0:case 11:case 15:Da(8,f,l);break;case 23:case 22:if(f.memoizedState!==null&&f.memoizedState.cachePool!==null){var h=f.memoizedState.cachePool.pool;h!=null&&h.refCount++}break;case 24:Fu(f.memoizedState.cache)}if(h=f.child,h!==null)h.return=f,mn=h;else e:for(f=o;mn!==null;){h=mn;var y=h.sibling,_=h.return;if(gT(h),h===f){mn=null;break e}if(y!==null){y.return=_,mn=y;break e}mn=_}}}var HH={getCacheForType:function(o){var l=On(tn),f=l.data.get(o);return f===void 0&&(f=o(),l.data.set(o,f)),f}},GH=typeof WeakMap=="function"?WeakMap:Map,wt=0,Nt=null,at=null,lt=0,St=0,rr=null,Ia=!1,Qs=!1,Gy=!1,Gi=0,Gt=0,La=0,$o=0,Wy=0,Cr=0,Js=0,ac=null,Kn=null,Ky=!1,Yy=0,xp=1/0,wp=null,Ba=null,wn=0,Fa=null,el=null,tl=0,Zy=0,Xy=null,AT=null,oc=0,Qy=null;function ir(){if((wt&2)!==0&&lt!==0)return lt&-lt;if(I.T!==null){var o=zs;return o!==0?o:ab()}return UO()}function RT(){Cr===0&&(Cr=(lt&536870912)===0||mt?BO():536870912);var o=Er.current;return o!==null&&(o.flags|=32),Cr}function ar(o,l,f){(o===Nt&&(St===2||St===9)||o.cancelPendingCommit!==null)&&(nl(o,0),qa(o,lt,Cr,!1)),Eu(o,f),((wt&2)===0||o!==Nt)&&(o===Nt&&((wt&2)===0&&($o|=f),Gt===4&&qa(o,lt,Cr,!1)),li(o))}function TT(o,l,f){if((wt&6)!==0)throw Error(r(327));var h=!f&&(l&124)===0&&(l&o.expiredLanes)===0||xo(o,l),y=h?YH(o,l):tb(o,l,!0),_=h;do{if(y===0){Qs&&!h&&qa(o,l,0,!1);break}else{if(f=o.current.alternate,_&&!WH(f)){y=tb(o,l,!1),_=!1;continue}if(y===2){if(_=l,o.errorRecoveryDisabledLanes&_)var j=0;else j=o.pendingLanes&-536870913,j=j!==0?j:j&536870912?536870912:0;if(j!==0){l=j;e:{var D=o;y=ac;var U=D.current.memoizedState.isDehydrated;if(U&&(nl(D,j).flags|=256),j=tb(D,j,!1),j!==2){if(Gy&&!U){D.errorRecoveryDisabledLanes|=_,$o|=_,y=4;break e}_=Kn,Kn=y,_!==null&&(Kn===null?Kn=_:Kn.push.apply(Kn,_))}y=j}if(_=!1,y!==2)continue}}if(y===1){nl(o,0),qa(o,l,0,!0);break}e:{switch(h=o,_=y,_){case 0:case 1:throw Error(r(345));case 4:if((l&4194048)!==l)break;case 6:qa(h,l,Cr,!Ia);break e;case 2:Kn=null;break;case 3:case 5:break;default:throw Error(r(329))}if((l&62914560)===l&&(y=Yy+300-it(),10<y)){if(qa(h,l,Cr,!Ia),bo(h,0,!0)!==0)break e;h.timeoutHandle=iM(MT.bind(null,h,f,Kn,wp,Ky,l,Cr,$o,Js,Ia,_,2,-0,0),y);break e}MT(h,f,Kn,wp,Ky,l,Cr,$o,Js,Ia,_,0,-0,0)}}break}while(!0);li(o)}function MT(o,l,f,h,y,_,j,D,U,ne,he,ge,ie,ae){if(o.timeoutHandle=-1,ge=l.subtreeFlags,(ge&8192||(ge&16785408)===16785408)&&(pc={stylesheets:null,count:0,unsuspend:R5},_T(l),ge=M5(),ge!==null)){o.cancelPendingCommit=ge(IT.bind(null,o,l,_,f,h,y,j,D,U,he,1,ie,ae)),qa(o,_,j,!ne);return}IT(o,l,_,f,h,y,j,D,U)}function WH(o){for(var l=o;;){var f=l.tag;if((f===0||f===11||f===15)&&l.flags&16384&&(f=l.updateQueue,f!==null&&(f=f.stores,f!==null)))for(var h=0;h<f.length;h++){var y=f[h],_=y.getSnapshot;y=y.value;try{if(!Jn(_(),y))return!1}catch{return!1}}if(f=l.child,l.subtreeFlags&16384&&f!==null)f.return=l,l=f;else{if(l===o)break;for(;l.sibling===null;){if(l.return===null||l.return===o)return!0;l=l.return}l.sibling.return=l.return,l=l.sibling}}return!0}function qa(o,l,f,h){l&=~Wy,l&=~$o,o.suspendedLanes|=l,o.pingedLanes&=~l,h&&(o.warmLanes|=l),h=o.expirationTimes;for(var y=l;0<y;){var _=31-Ke(y),j=1<<_;h[_]=-1,y&=~j}f!==0&&qO(o,f,l)}function Sp(){return(wt&6)===0?(sc(0),!1):!0}function Jy(){if(at!==null){if(St===0)var o=at.return;else o=at,Li=Mo=null,vy(o),Ks=null,Xu=0,o=at;for(;o!==null;)uT(o.alternate,o),o=o.return;at=null}}function nl(o,l){var f=o.timeoutHandle;f!==-1&&(o.timeoutHandle=-1,f5(f)),f=o.cancelPendingCommit,f!==null&&(o.cancelPendingCommit=null,f()),Jy(),Nt=o,at=f=ki(o.current,null),lt=l,St=0,rr=null,Ia=!1,Qs=xo(o,l),Gy=!1,Js=Cr=Wy=$o=La=Gt=0,Kn=ac=null,Ky=!1,(l&8)!==0&&(l|=l&32);var h=o.entangledLanes;if(h!==0)for(o=o.entanglements,h&=l;0<h;){var y=31-Ke(h),_=1<<y;l|=o[y],h&=~_}return Gi=l,Vd(),f}function PT(o,l){tt=null,I.H=up,l===zu||l===Jd?(l=KA(),St=3):l===HA?(l=KA(),St=4):St=l===YR?8:l!==null&&typeof l=="object"&&typeof l.then=="function"?6:1,rr=l,at===null&&(Gt=1,hp(o,xr(l,o.current)))}function jT(){var o=I.H;return I.H=up,o===null?up:o}function NT(){var o=I.A;return I.A=HH,o}function eb(){Gt=4,Ia||(lt&4194048)!==lt&&Er.current!==null||(Qs=!0),(La&134217727)===0&&($o&134217727)===0||Nt===null||qa(Nt,lt,Cr,!1)}function tb(o,l,f){var h=wt;wt|=2;var y=jT(),_=NT();(Nt!==o||lt!==l)&&(wp=null,nl(o,l)),l=!1;var j=Gt;e:do try{if(St!==0&&at!==null){var D=at,U=rr;switch(St){case 8:Jy(),j=6;break e;case 3:case 2:case 9:case 6:Er.current===null&&(l=!0);var ne=St;if(St=0,rr=null,rl(o,D,U,ne),f&&Qs){j=0;break e}break;default:ne=St,St=0,rr=null,rl(o,D,U,ne)}}KH(),j=Gt;break}catch(he){PT(o,he)}while(!0);return l&&o.shellSuspendCounter++,Li=Mo=null,wt=h,I.H=y,I.A=_,at===null&&(Nt=null,lt=0,Vd()),j}function KH(){for(;at!==null;)DT(at)}function YH(o,l){var f=wt;wt|=2;var h=jT(),y=NT();Nt!==o||lt!==l?(wp=null,xp=it()+500,nl(o,l)):Qs=xo(o,l);e:do try{if(St!==0&&at!==null){l=at;var _=rr;t:switch(St){case 1:St=0,rr=null,rl(o,l,_,1);break;case 2:case 9:if(GA(_)){St=0,rr=null,kT(l);break}l=function(){St!==2&&St!==9||Nt!==o||(St=7),li(o)},_.then(l,l);break e;case 3:St=7;break e;case 4:St=5;break e;case 7:GA(_)?(St=0,rr=null,kT(l)):(St=0,rr=null,rl(o,l,_,7));break;case 5:var j=null;switch(at.tag){case 26:j=at.memoizedState;case 5:case 27:var D=at;if(!j||vM(j)){St=0,rr=null;var U=D.sibling;if(U!==null)at=U;else{var ne=D.return;ne!==null?(at=ne,_p(ne)):at=null}break t}}St=0,rr=null,rl(o,l,_,5);break;case 6:St=0,rr=null,rl(o,l,_,6);break;case 8:Jy(),Gt=6;break e;default:throw Error(r(462))}}ZH();break}catch(he){PT(o,he)}while(!0);return Li=Mo=null,I.H=h,I.A=y,wt=f,at!==null?0:(Nt=null,lt=0,Vd(),Gt)}function ZH(){for(;at!==null&&!dt();)DT(at)}function DT(o){var l=sT(o.alternate,o,Gi);o.memoizedProps=o.pendingProps,l===null?_p(o):at=l}function kT(o){var l=o,f=l.alternate;switch(l.tag){case 15:case 0:l=tT(f,l,l.pendingProps,l.type,void 0,lt);break;case 11:l=tT(f,l,l.pendingProps,l.type.render,l.ref,lt);break;case 5:vy(l);default:uT(f,l),l=at=$A(l,Gi),l=sT(f,l,Gi)}o.memoizedProps=o.pendingProps,l===null?_p(o):at=l}function rl(o,l,f,h){Li=Mo=null,vy(l),Ks=null,Xu=0;var y=l.return;try{if(BH(o,y,l,f,lt)){Gt=1,hp(o,xr(f,o.current)),at=null;return}}catch(_){if(y!==null)throw at=y,_;Gt=1,hp(o,xr(f,o.current)),at=null;return}l.flags&32768?(mt||h===1?o=!0:Qs||(lt&536870912)!==0?o=!1:(Ia=o=!0,(h===2||h===9||h===3||h===6)&&(h=Er.current,h!==null&&h.tag===13&&(h.flags|=16384))),$T(l,o)):_p(l)}function _p(o){var l=o;do{if((l.flags&32768)!==0){$T(l,Ia);return}o=l.return;var f=qH(l.alternate,l,Gi);if(f!==null){at=f;return}if(l=l.sibling,l!==null){at=l;return}at=l=o}while(l!==null);Gt===0&&(Gt=5)}function $T(o,l){do{var f=zH(o.alternate,o);if(f!==null){f.flags&=32767,at=f;return}if(f=o.return,f!==null&&(f.flags|=32768,f.subtreeFlags=0,f.deletions=null),!l&&(o=o.sibling,o!==null)){at=o;return}at=o=f}while(o!==null);Gt=6,at=null}function IT(o,l,f,h,y,_,j,D,U){o.cancelPendingCommit=null;do Ep();while(wn!==0);if((wt&6)!==0)throw Error(r(327));if(l!==null){if(l===o.current)throw Error(r(177));if(_=l.lanes|l.childLanes,_|=Hg,R4(o,f,_,j,D,U),o===Nt&&(at=Nt=null,lt=0),el=l,Fa=o,tl=f,Zy=_,Xy=y,AT=h,(l.subtreeFlags&10256)!==0||(l.flags&10256)!==0?(o.callbackNode=null,o.callbackPriority=0,e5(Bt,function(){return zT(),null})):(o.callbackNode=null,o.callbackPriority=0),h=(l.flags&13878)!==0,(l.subtreeFlags&13878)!==0||h){h=I.T,I.T=null,y=W.p,W.p=2,j=wt,wt|=4;try{UH(o,l,f)}finally{wt=j,W.p=y,I.T=h}}wn=1,LT(),BT(),FT()}}function LT(){if(wn===1){wn=0;var o=Fa,l=el,f=(l.flags&13878)!==0;if((l.subtreeFlags&13878)!==0||f){f=I.T,I.T=null;var h=W.p;W.p=2;var y=wt;wt|=4;try{xT(l,o);var _=pb,j=OA(o.containerInfo),D=_.focusedElem,U=_.selectionRange;if(j!==D&&D&&D.ownerDocument&&CA(D.ownerDocument.documentElement,D)){if(U!==null&&Fg(D)){var ne=U.start,he=U.end;if(he===void 0&&(he=ne),"selectionStart"in D)D.selectionStart=ne,D.selectionEnd=Math.min(he,D.value.length);else{var ge=D.ownerDocument||document,ie=ge&&ge.defaultView||window;if(ie.getSelection){var ae=ie.getSelection(),Ve=D.textContent.length,Fe=Math.min(U.start,Ve),Ct=U.end===void 0?Fe:Math.min(U.end,Ve);!ae.extend&&Fe>Ct&&(j=Ct,Ct=Fe,Fe=j);var ee=EA(D,Fe),K=EA(D,Ct);if(ee&&K&&(ae.rangeCount!==1||ae.anchorNode!==ee.node||ae.anchorOffset!==ee.offset||ae.focusNode!==K.node||ae.focusOffset!==K.offset)){var te=ge.createRange();te.setStart(ee.node,ee.offset),ae.removeAllRanges(),Fe>Ct?(ae.addRange(te),ae.extend(K.node,K.offset)):(te.setEnd(K.node,K.offset),ae.addRange(te))}}}}for(ge=[],ae=D;ae=ae.parentNode;)ae.nodeType===1&&ge.push({element:ae,left:ae.scrollLeft,top:ae.scrollTop});for(typeof D.focus=="function"&&D.focus(),D=0;D<ge.length;D++){var me=ge[D];me.element.scrollLeft=me.left,me.element.scrollTop=me.top}}$p=!!db,pb=db=null}finally{wt=y,W.p=h,I.T=f}}o.current=l,wn=2}}function BT(){if(wn===2){wn=0;var o=Fa,l=el,f=(l.flags&8772)!==0;if((l.subtreeFlags&8772)!==0||f){f=I.T,I.T=null;var h=W.p;W.p=2;var y=wt;wt|=4;try{vT(o,l.alternate,l)}finally{wt=y,W.p=h,I.T=f}}wn=3}}function FT(){if(wn===4||wn===3){wn=0,ft();var o=Fa,l=el,f=tl,h=AT;(l.subtreeFlags&10256)!==0||(l.flags&10256)!==0?wn=5:(wn=0,el=Fa=null,qT(o,o.pendingLanes));var y=o.pendingLanes;if(y===0&&(Ba=null),bg(f),l=l.stateNode,ye&&typeof ye.onCommitFiberRoot=="function")try{ye.onCommitFiberRoot(Ee,l,void 0,(l.current.flags&128)===128)}catch{}if(h!==null){l=I.T,y=W.p,W.p=2,I.T=null;try{for(var _=o.onRecoverableError,j=0;j<h.length;j++){var D=h[j];_(D.value,{componentStack:D.stack})}}finally{I.T=l,W.p=y}}(tl&3)!==0&&Ep(),li(o),y=o.pendingLanes,(f&4194090)!==0&&(y&42)!==0?o===Qy?oc++:(oc=0,Qy=o):oc=0,sc(0)}}function qT(o,l){(o.pooledCacheLanes&=l)===0&&(l=o.pooledCache,l!=null&&(o.pooledCache=null,Fu(l)))}function Ep(o){return LT(),BT(),FT(),zT()}function zT(){if(wn!==5)return!1;var o=Fa,l=Zy;Zy=0;var f=bg(tl),h=I.T,y=W.p;try{W.p=32>f?32:f,I.T=null,f=Xy,Xy=null;var _=Fa,j=tl;if(wn=0,el=Fa=null,tl=0,(wt&6)!==0)throw Error(r(331));var D=wt;if(wt|=4,CT(_.current),ST(_,_.current,j,f),wt=D,sc(0,!1),ye&&typeof ye.onPostCommitFiberRoot=="function")try{ye.onPostCommitFiberRoot(Ee,_)}catch{}return!0}finally{W.p=y,I.T=h,qT(o,l)}}function UT(o,l,f){l=xr(f,l),l=My(o.stateNode,l,2),o=Ma(o,l,2),o!==null&&(Eu(o,2),li(o))}function Mt(o,l,f){if(o.tag===3)UT(o,o,f);else for(;l!==null;){if(l.tag===3){UT(l,o,f);break}else if(l.tag===1){var h=l.stateNode;if(typeof l.type.getDerivedStateFromError=="function"||typeof h.componentDidCatch=="function"&&(Ba===null||!Ba.has(h))){o=xr(f,o),f=WR(2),h=Ma(l,f,2),h!==null&&(KR(f,h,l,o),Eu(h,2),li(h));break}}l=l.return}}function nb(o,l,f){var h=o.pingCache;if(h===null){h=o.pingCache=new GH;var y=new Set;h.set(l,y)}else y=h.get(l),y===void 0&&(y=new Set,h.set(l,y));y.has(f)||(Gy=!0,y.add(f),o=XH.bind(null,o,l,f),l.then(o,o))}function XH(o,l,f){var h=o.pingCache;h!==null&&h.delete(l),o.pingedLanes|=o.suspendedLanes&f,o.warmLanes&=~f,Nt===o&&(lt&f)===f&&(Gt===4||Gt===3&&(lt&62914560)===lt&&300>it()-Yy?(wt&2)===0&&nl(o,0):Wy|=f,Js===lt&&(Js=0)),li(o)}function VT(o,l){l===0&&(l=FO()),o=Ls(o,l),o!==null&&(Eu(o,l),li(o))}function QH(o){var l=o.memoizedState,f=0;l!==null&&(f=l.retryLane),VT(o,f)}function JH(o,l){var f=0;switch(o.tag){case 13:var h=o.stateNode,y=o.memoizedState;y!==null&&(f=y.retryLane);break;case 19:h=o.stateNode;break;case 22:h=o.stateNode._retryCache;break;default:throw Error(r(314))}h!==null&&h.delete(l),VT(o,f)}function e5(o,l){return Ye(o,l)}var Cp=null,il=null,rb=!1,Op=!1,ib=!1,Io=0;function li(o){o!==il&&o.next===null&&(il===null?Cp=il=o:il=il.next=o),Op=!0,rb||(rb=!0,n5())}function sc(o,l){if(!ib&&Op){ib=!0;do for(var f=!1,h=Cp;h!==null;){if(o!==0){var y=h.pendingLanes;if(y===0)var _=0;else{var j=h.suspendedLanes,D=h.pingedLanes;_=(1<<31-Ke(42|o)+1)-1,_&=y&~(j&~D),_=_&201326741?_&201326741|1:_?_|2:0}_!==0&&(f=!0,KT(h,_))}else _=lt,_=bo(h,h===Nt?_:0,h.cancelPendingCommit!==null||h.timeoutHandle!==-1),(_&3)===0||xo(h,_)||(f=!0,KT(h,_));h=h.next}while(f);ib=!1}}function t5(){HT()}function HT(){Op=rb=!1;var o=0;Io!==0&&(c5()&&(o=Io),Io=0);for(var l=it(),f=null,h=Cp;h!==null;){var y=h.next,_=GT(h,l);_===0?(h.next=null,f===null?Cp=y:f.next=y,y===null&&(il=f)):(f=h,(o!==0||(_&3)!==0)&&(Op=!0)),h=y}sc(o)}function GT(o,l){for(var f=o.suspendedLanes,h=o.pingedLanes,y=o.expirationTimes,_=o.pendingLanes&-62914561;0<_;){var j=31-Ke(_),D=1<<j,U=y[j];U===-1?((D&f)===0||(D&h)!==0)&&(y[j]=jd(D,l)):U<=l&&(o.expiredLanes|=D),_&=~D}if(l=Nt,f=lt,f=bo(o,o===l?f:0,o.cancelPendingCommit!==null||o.timeoutHandle!==-1),h=o.callbackNode,f===0||o===l&&(St===2||St===9)||o.cancelPendingCommit!==null)return h!==null&&h!==null&&Be(h),o.callbackNode=null,o.callbackPriority=0;if((f&3)===0||xo(o,f)){if(l=f&-f,l===o.callbackPriority)return l;switch(h!==null&&Be(h),bg(f)){case 2:case 8:f=bn;break;case 32:f=Bt;break;case 268435456:f=re;break;default:f=Bt}return h=WT.bind(null,o),f=Ye(f,h),o.callbackPriority=l,o.callbackNode=f,l}return h!==null&&h!==null&&Be(h),o.callbackPriority=2,o.callbackNode=null,2}function WT(o,l){if(wn!==0&&wn!==5)return o.callbackNode=null,o.callbackPriority=0,null;var f=o.callbackNode;if(Ep()&&o.callbackNode!==f)return null;var h=lt;return h=bo(o,o===Nt?h:0,o.cancelPendingCommit!==null||o.timeoutHandle!==-1),h===0?null:(TT(o,h,l),GT(o,it()),o.callbackNode!=null&&o.callbackNode===f?WT.bind(null,o):null)}function KT(o,l){if(Ep())return null;TT(o,l,!0)}function n5(){d5(function(){(wt&6)!==0?Ye(Tt,t5):HT()})}function ab(){return Io===0&&(Io=BO()),Io}function YT(o){return o==null||typeof o=="symbol"||typeof o=="boolean"?null:typeof o=="function"?o:Id(""+o)}function ZT(o,l){var f=l.ownerDocument.createElement("input");return f.name=l.name,f.value=l.value,o.id&&f.setAttribute("form",o.id),l.parentNode.insertBefore(f,l),o=new FormData(o),f.parentNode.removeChild(f),o}function r5(o,l,f,h,y){if(l==="submit"&&f&&f.stateNode===y){var _=YT((y[Vn]||null).action),j=h.submitter;j&&(l=(l=j[Vn]||null)?YT(l.formAction):j.getAttribute("formAction"),l!==null&&(_=l,j=null));var D=new qd("action","action",null,h,y);o.push({event:D,listeners:[{instance:null,listener:function(){if(h.defaultPrevented){if(Io!==0){var U=j?ZT(y,j):new FormData(y);Cy(f,{pending:!0,data:U,method:y.method,action:_},null,U)}}else typeof _=="function"&&(D.preventDefault(),U=j?ZT(y,j):new FormData(y),Cy(f,{pending:!0,data:U,method:y.method,action:_},_,U))},currentTarget:y}]})}}for(var ob=0;ob<Vg.length;ob++){var sb=Vg[ob],i5=sb.toLowerCase(),a5=sb[0].toUpperCase()+sb.slice(1);Lr(i5,"on"+a5)}Lr(TA,"onAnimationEnd"),Lr(MA,"onAnimationIteration"),Lr(PA,"onAnimationStart"),Lr("dblclick","onDoubleClick"),Lr("focusin","onFocus"),Lr("focusout","onBlur"),Lr(SH,"onTransitionRun"),Lr(_H,"onTransitionStart"),Lr(EH,"onTransitionCancel"),Lr(jA,"onTransitionEnd"),Rs("onMouseEnter",["mouseout","mouseover"]),Rs("onMouseLeave",["mouseout","mouseover"]),Rs("onPointerEnter",["pointerout","pointerover"]),Rs("onPointerLeave",["pointerout","pointerover"]),wo("onChange","change click focusin focusout input keydown keyup selectionchange".split(" ")),wo("onSelect","focusout contextmenu dragend focusin keydown keyup mousedown mouseup selectionchange".split(" ")),wo("onBeforeInput",["compositionend","keypress","textInput","paste"]),wo("onCompositionEnd","compositionend focusout keydown keypress keyup mousedown".split(" ")),wo("onCompositionStart","compositionstart focusout keydown keypress keyup mousedown".split(" ")),wo("onCompositionUpdate","compositionupdate focusout keydown keypress keyup mousedown".split(" "));var lc="abort canplay canplaythrough durationchange emptied encrypted ended error loadeddata loadedmetadata loadstart pause play playing progress ratechange resize seeked seeking stalled suspend timeupdate volumechange waiting".split(" "),o5=new Set("beforetoggle cancel close invalid load scroll scrollend toggle".split(" ").concat(lc));function XT(o,l){l=(l&4)!==0;for(var f=0;f<o.length;f++){var h=o[f],y=h.event;h=h.listeners;e:{var _=void 0;if(l)for(var j=h.length-1;0<=j;j--){var D=h[j],U=D.instance,ne=D.currentTarget;if(D=D.listener,U!==_&&y.isPropagationStopped())break e;_=D,y.currentTarget=ne;try{_(y)}catch(he){pp(he)}y.currentTarget=null,_=U}else for(j=0;j<h.length;j++){if(D=h[j],U=D.instance,ne=D.currentTarget,D=D.listener,U!==_&&y.isPropagationStopped())break e;_=D,y.currentTarget=ne;try{_(y)}catch(he){pp(he)}y.currentTarget=null,_=U}}}}function ot(o,l){var f=l[xg];f===void 0&&(f=l[xg]=new Set);var h=o+"__bubble";f.has(h)||(QT(l,o,2,!1),f.add(h))}function lb(o,l,f){var h=0;l&&(h|=4),QT(f,o,h,l)}var Ap="_reactListening"+Math.random().toString(36).slice(2);function ub(o){if(!o[Ap]){o[Ap]=!0,HO.forEach(function(f){f!=="selectionchange"&&(o5.has(f)||lb(f,!1,o),lb(f,!0,o))});var l=o.nodeType===9?o:o.ownerDocument;l===null||l[Ap]||(l[Ap]=!0,lb("selectionchange",!1,l))}}function QT(o,l,f,h){switch(SM(l)){case 2:var y=N5;break;case 8:y=D5;break;default:y=_b}f=y.bind(null,l,f,o),y=void 0,!Pg||l!=="touchstart"&&l!=="touchmove"&&l!=="wheel"||(y=!0),h?y!==void 0?o.addEventListener(l,f,{capture:!0,passive:y}):o.addEventListener(l,f,!0):y!==void 0?o.addEventListener(l,f,{passive:y}):o.addEventListener(l,f,!1)}function cb(o,l,f,h,y){var _=h;if((l&1)===0&&(l&2)===0&&h!==null)e:for(;;){if(h===null)return;var j=h.tag;if(j===3||j===4){var D=h.stateNode.containerInfo;if(D===y)break;if(j===4)for(j=h.return;j!==null;){var U=j.tag;if((U===3||U===4)&&j.stateNode.containerInfo===y)return;j=j.return}for(;D!==null;){if(j=Cs(D),j===null)return;if(U=j.tag,U===5||U===6||U===26||U===27){h=_=j;continue e}D=D.parentNode}}h=h.return}aA(function(){var ne=_,he=Tg(f),ge=[];e:{var ie=NA.get(o);if(ie!==void 0){var ae=qd,Ve=o;switch(o){case"keypress":if(Bd(f)===0)break e;case"keydown":case"keyup":ae=eH;break;case"focusin":Ve="focus",ae=kg;break;case"focusout":Ve="blur",ae=kg;break;case"beforeblur":case"afterblur":ae=kg;break;case"click":if(f.button===2)break e;case"auxclick":case"dblclick":case"mousedown":case"mousemove":case"mouseup":case"mouseout":case"mouseover":case"contextmenu":ae=lA;break;case"drag":case"dragend":case"dragenter":case"dragexit":case"dragleave":case"dragover":case"dragstart":case"drop":ae=z4;break;case"touchcancel":case"touchend":case"touchmove":case"touchstart":ae=rH;break;case TA:case MA:case PA:ae=H4;break;case jA:ae=aH;break;case"scroll":case"scrollend":ae=F4;break;case"wheel":ae=sH;break;case"copy":case"cut":case"paste":ae=W4;break;case"gotpointercapture":case"lostpointercapture":case"pointercancel":case"pointerdown":case"pointermove":case"pointerout":case"pointerover":case"pointerup":ae=cA;break;case"toggle":case"beforetoggle":ae=uH}var Fe=(l&4)!==0,Ct=!Fe&&(o==="scroll"||o==="scrollend"),ee=Fe?ie!==null?ie+"Capture":null:ie;Fe=[];for(var K=ne,te;K!==null;){var me=K;if(te=me.stateNode,me=me.tag,me!==5&&me!==26&&me!==27||te===null||ee===null||(me=Au(K,ee),me!=null&&Fe.push(uc(K,me,te))),Ct)break;K=K.return}0<Fe.length&&(ie=new ae(ie,Ve,null,f,he),ge.push({event:ie,listeners:Fe}))}}if((l&7)===0){e:{if(ie=o==="mouseover"||o==="pointerover",ae=o==="mouseout"||o==="pointerout",ie&&f!==Rg&&(Ve=f.relatedTarget||f.fromElement)&&(Cs(Ve)||Ve[Es]))break e;if((ae||ie)&&(ie=he.window===he?he:(ie=he.ownerDocument)?ie.defaultView||ie.parentWindow:window,ae?(Ve=f.relatedTarget||f.toElement,ae=ne,Ve=Ve?Cs(Ve):null,Ve!==null&&(Ct=a(Ve),Fe=Ve.tag,Ve!==Ct||Fe!==5&&Fe!==27&&Fe!==6)&&(Ve=null)):(ae=null,Ve=ne),ae!==Ve)){if(Fe=lA,me="onMouseLeave",ee="onMouseEnter",K="mouse",(o==="pointerout"||o==="pointerover")&&(Fe=cA,me="onPointerLeave",ee="onPointerEnter",K="pointer"),Ct=ae==null?ie:Ou(ae),te=Ve==null?ie:Ou(Ve),ie=new Fe(me,K+"leave",ae,f,he),ie.target=Ct,ie.relatedTarget=te,me=null,Cs(he)===ne&&(Fe=new Fe(ee,K+"enter",Ve,f,he),Fe.target=te,Fe.relatedTarget=Ct,me=Fe),Ct=me,ae&&Ve)t:{for(Fe=ae,ee=Ve,K=0,te=Fe;te;te=al(te))K++;for(te=0,me=ee;me;me=al(me))te++;for(;0<K-te;)Fe=al(Fe),K--;for(;0<te-K;)ee=al(ee),te--;for(;K--;){if(Fe===ee||ee!==null&&Fe===ee.alternate)break t;Fe=al(Fe),ee=al(ee)}Fe=null}else Fe=null;ae!==null&&JT(ge,ie,ae,Fe,!1),Ve!==null&&Ct!==null&&JT(ge,Ct,Ve,Fe,!0)}}e:{if(ie=ne?Ou(ne):window,ae=ie.nodeName&&ie.nodeName.toLowerCase(),ae==="select"||ae==="input"&&ie.type==="file")var je=yA;else if(vA(ie))if(bA)je=bH;else{je=gH;var rt=vH}else ae=ie.nodeName,!ae||ae.toLowerCase()!=="input"||ie.type!=="checkbox"&&ie.type!=="radio"?ne&&Ag(ne.elementType)&&(je=yA):je=yH;if(je&&(je=je(o,ne))){gA(ge,je,f,he);break e}rt&&rt(o,ie,ne),o==="focusout"&&ne&&ie.type==="number"&&ne.memoizedProps.value!=null&&Og(ie,"number",ie.value)}switch(rt=ne?Ou(ne):window,o){case"focusin":(vA(rt)||rt.contentEditable==="true")&&(ks=rt,qg=ne,ku=null);break;case"focusout":ku=qg=ks=null;break;case"mousedown":zg=!0;break;case"contextmenu":case"mouseup":case"dragend":zg=!1,AA(ge,f,he);break;case"selectionchange":if(wH)break;case"keydown":case"keyup":AA(ge,f,he)}var Le;if(Ig)e:{switch(o){case"compositionstart":var qe="onCompositionStart";break e;case"compositionend":qe="onCompositionEnd";break e;case"compositionupdate":qe="onCompositionUpdate";break e}qe=void 0}else Ds?hA(o,f)&&(qe="onCompositionEnd"):o==="keydown"&&f.keyCode===229&&(qe="onCompositionStart");qe&&(fA&&f.locale!=="ko"&&(Ds||qe!=="onCompositionStart"?qe==="onCompositionEnd"&&Ds&&(Le=oA()):(Oa=he,jg="value"in Oa?Oa.value:Oa.textContent,Ds=!0)),rt=Rp(ne,qe),0<rt.length&&(qe=new uA(qe,o,null,f,he),ge.push({event:qe,listeners:rt}),Le?qe.data=Le:(Le=mA(f),Le!==null&&(qe.data=Le)))),(Le=fH?dH(o,f):pH(o,f))&&(qe=Rp(ne,"onBeforeInput"),0<qe.length&&(rt=new uA("onBeforeInput","beforeinput",null,f,he),ge.push({event:rt,listeners:qe}),rt.data=Le)),r5(ge,o,ne,f,he)}XT(ge,l)})}function uc(o,l,f){return{instance:o,listener:l,currentTarget:f}}function Rp(o,l){for(var f=l+"Capture",h=[];o!==null;){var y=o,_=y.stateNode;if(y=y.tag,y!==5&&y!==26&&y!==27||_===null||(y=Au(o,f),y!=null&&h.unshift(uc(o,y,_)),y=Au(o,l),y!=null&&h.push(uc(o,y,_))),o.tag===3)return h;o=o.return}return[]}function al(o){if(o===null)return null;do o=o.return;while(o&&o.tag!==5&&o.tag!==27);return o||null}function JT(o,l,f,h,y){for(var _=l._reactName,j=[];f!==null&&f!==h;){var D=f,U=D.alternate,ne=D.stateNode;if(D=D.tag,U!==null&&U===h)break;D!==5&&D!==26&&D!==27||ne===null||(U=ne,y?(ne=Au(f,_),ne!=null&&j.unshift(uc(f,ne,U))):y||(ne=Au(f,_),ne!=null&&j.push(uc(f,ne,U)))),f=f.return}j.length!==0&&o.push({event:l,listeners:j})}var s5=/\r\n?/g,l5=/\u0000|\uFFFD/g;function eM(o){return(typeof o=="string"?o:""+o).replace(s5,`
`).replace(l5,"")}function tM(o,l){return l=eM(l),eM(o)===l}function Tp(){}function Et(o,l,f,h,y,_){switch(f){case"children":typeof h=="string"?l==="body"||l==="textarea"&&h===""||Ps(o,h):(typeof h=="number"||typeof h=="bigint")&&l!=="body"&&Ps(o,""+h);break;case"className":Dd(o,"class",h);break;case"tabIndex":Dd(o,"tabindex",h);break;case"dir":case"role":case"viewBox":case"width":case"height":Dd(o,f,h);break;case"style":rA(o,h,_);break;case"data":if(l!=="object"){Dd(o,"data",h);break}case"src":case"href":if(h===""&&(l!=="a"||f!=="href")){o.removeAttribute(f);break}if(h==null||typeof h=="function"||typeof h=="symbol"||typeof h=="boolean"){o.removeAttribute(f);break}h=Id(""+h),o.setAttribute(f,h);break;case"action":case"formAction":if(typeof h=="function"){o.setAttribute(f,"javascript:throw new Error('A React form was unexpectedly submitted. If you called form.submit() manually, consider using form.requestSubmit() instead. If you\\'re trying to use event.stopPropagation() in a submit event handler, consider also calling event.preventDefault().')");break}else typeof _=="function"&&(f==="formAction"?(l!=="input"&&Et(o,l,"name",y.name,y,null),Et(o,l,"formEncType",y.formEncType,y,null),Et(o,l,"formMethod",y.formMethod,y,null),Et(o,l,"formTarget",y.formTarget,y,null)):(Et(o,l,"encType",y.encType,y,null),Et(o,l,"method",y.method,y,null),Et(o,l,"target",y.target,y,null)));if(h==null||typeof h=="symbol"||typeof h=="boolean"){o.removeAttribute(f);break}h=Id(""+h),o.setAttribute(f,h);break;case"onClick":h!=null&&(o.onclick=Tp);break;case"onScroll":h!=null&&ot("scroll",o);break;case"onScrollEnd":h!=null&&ot("scrollend",o);break;case"dangerouslySetInnerHTML":if(h!=null){if(typeof h!="object"||!("__html"in h))throw Error(r(61));if(f=h.__html,f!=null){if(y.children!=null)throw Error(r(60));o.innerHTML=f}}break;case"multiple":o.multiple=h&&typeof h!="function"&&typeof h!="symbol";break;case"muted":o.muted=h&&typeof h!="function"&&typeof h!="symbol";break;case"suppressContentEditableWarning":case"suppressHydrationWarning":case"defaultValue":case"defaultChecked":case"innerHTML":case"ref":break;case"autoFocus":break;case"xlinkHref":if(h==null||typeof h=="function"||typeof h=="boolean"||typeof h=="symbol"){o.removeAttribute("xlink:href");break}f=Id(""+h),o.setAttributeNS("http://www.w3.org/1999/xlink","xlink:href",f);break;case"contentEditable":case"spellCheck":case"draggable":case"value":case"autoReverse":case"externalResourcesRequired":case"focusable":case"preserveAlpha":h!=null&&typeof h!="function"&&typeof h!="symbol"?o.setAttribute(f,""+h):o.removeAttribute(f);break;case"inert":case"allowFullScreen":case"async":case"autoPlay":case"controls":case"default":case"defer":case"disabled":case"disablePictureInPicture":case"disableRemotePlayback":case"formNoValidate":case"hidden":case"loop":case"noModule":case"noValidate":case"open":case"playsInline":case"readOnly":case"required":case"reversed":case"scoped":case"seamless":case"itemScope":h&&typeof h!="function"&&typeof h!="symbol"?o.setAttribute(f,""):o.removeAttribute(f);break;case"capture":case"download":h===!0?o.setAttribute(f,""):h!==!1&&h!=null&&typeof h!="function"&&typeof h!="symbol"?o.setAttribute(f,h):o.removeAttribute(f);break;case"cols":case"rows":case"size":case"span":h!=null&&typeof h!="function"&&typeof h!="symbol"&&!isNaN(h)&&1<=h?o.setAttribute(f,h):o.removeAttribute(f);break;case"rowSpan":case"start":h==null||typeof h=="function"||typeof h=="symbol"||isNaN(h)?o.removeAttribute(f):o.setAttribute(f,h);break;case"popover":ot("beforetoggle",o),ot("toggle",o),Nd(o,"popover",h);break;case"xlinkActuate":Ni(o,"http://www.w3.org/1999/xlink","xlink:actuate",h);break;case"xlinkArcrole":Ni(o,"http://www.w3.org/1999/xlink","xlink:arcrole",h);break;case"xlinkRole":Ni(o,"http://www.w3.org/1999/xlink","xlink:role",h);break;case"xlinkShow":Ni(o,"http://www.w3.org/1999/xlink","xlink:show",h);break;case"xlinkTitle":Ni(o,"http://www.w3.org/1999/xlink","xlink:title",h);break;case"xlinkType":Ni(o,"http://www.w3.org/1999/xlink","xlink:type",h);break;case"xmlBase":Ni(o,"http://www.w3.org/XML/1998/namespace","xml:base",h);break;case"xmlLang":Ni(o,"http://www.w3.org/XML/1998/namespace","xml:lang",h);break;case"xmlSpace":Ni(o,"http://www.w3.org/XML/1998/namespace","xml:space",h);break;case"is":Nd(o,"is",h);break;case"innerText":case"textContent":break;default:(!(2<f.length)||f[0]!=="o"&&f[0]!=="O"||f[1]!=="n"&&f[1]!=="N")&&(f=L4.get(f)||f,Nd(o,f,h))}}function fb(o,l,f,h,y,_){switch(f){case"style":rA(o,h,_);break;case"dangerouslySetInnerHTML":if(h!=null){if(typeof h!="object"||!("__html"in h))throw Error(r(61));if(f=h.__html,f!=null){if(y.children!=null)throw Error(r(60));o.innerHTML=f}}break;case"children":typeof h=="string"?Ps(o,h):(typeof h=="number"||typeof h=="bigint")&&Ps(o,""+h);break;case"onScroll":h!=null&&ot("scroll",o);break;case"onScrollEnd":h!=null&&ot("scrollend",o);break;case"onClick":h!=null&&(o.onclick=Tp);break;case"suppressContentEditableWarning":case"suppressHydrationWarning":case"innerHTML":case"ref":break;case"innerText":case"textContent":break;default:if(!GO.hasOwnProperty(f))e:{if(f[0]==="o"&&f[1]==="n"&&(y=f.endsWith("Capture"),l=f.slice(2,y?f.length-7:void 0),_=o[Vn]||null,_=_!=null?_[f]:null,typeof _=="function"&&o.removeEventListener(l,_,y),typeof h=="function")){typeof _!="function"&&_!==null&&(f in o?o[f]=null:o.hasAttribute(f)&&o.removeAttribute(f)),o.addEventListener(l,h,y);break e}f in o?o[f]=h:h===!0?o.setAttribute(f,""):Nd(o,f,h)}}}function Sn(o,l,f){switch(l){case"div":case"span":case"svg":case"path":case"a":case"g":case"p":case"li":break;case"img":ot("error",o),ot("load",o);var h=!1,y=!1,_;for(_ in f)if(f.hasOwnProperty(_)){var j=f[_];if(j!=null)switch(_){case"src":h=!0;break;case"srcSet":y=!0;break;case"children":case"dangerouslySetInnerHTML":throw Error(r(137,l));default:Et(o,l,_,j,f,null)}}y&&Et(o,l,"srcSet",f.srcSet,f,null),h&&Et(o,l,"src",f.src,f,null);return;case"input":ot("invalid",o);var D=_=j=y=null,U=null,ne=null;for(h in f)if(f.hasOwnProperty(h)){var he=f[h];if(he!=null)switch(h){case"name":y=he;break;case"type":j=he;break;case"checked":U=he;break;case"defaultChecked":ne=he;break;case"value":_=he;break;case"defaultValue":D=he;break;case"children":case"dangerouslySetInnerHTML":if(he!=null)throw Error(r(137,l));break;default:Et(o,l,h,he,f,null)}}JO(o,_,D,U,ne,j,y,!1),kd(o);return;case"select":ot("invalid",o),h=j=_=null;for(y in f)if(f.hasOwnProperty(y)&&(D=f[y],D!=null))switch(y){case"value":_=D;break;case"defaultValue":j=D;break;case"multiple":h=D;default:Et(o,l,y,D,f,null)}l=_,f=j,o.multiple=!!h,l!=null?Ms(o,!!h,l,!1):f!=null&&Ms(o,!!h,f,!0);return;case"textarea":ot("invalid",o),_=y=h=null;for(j in f)if(f.hasOwnProperty(j)&&(D=f[j],D!=null))switch(j){case"value":h=D;break;case"defaultValue":y=D;break;case"children":_=D;break;case"dangerouslySetInnerHTML":if(D!=null)throw Error(r(91));break;default:Et(o,l,j,D,f,null)}tA(o,h,y,_),kd(o);return;case"option":for(U in f)if(f.hasOwnProperty(U)&&(h=f[U],h!=null))switch(U){case"selected":o.selected=h&&typeof h!="function"&&typeof h!="symbol";break;default:Et(o,l,U,h,f,null)}return;case"dialog":ot("beforetoggle",o),ot("toggle",o),ot("cancel",o),ot("close",o);break;case"iframe":case"object":ot("load",o);break;case"video":case"audio":for(h=0;h<lc.length;h++)ot(lc[h],o);break;case"image":ot("error",o),ot("load",o);break;case"details":ot("toggle",o);break;case"embed":case"source":case"link":ot("error",o),ot("load",o);case"area":case"base":case"br":case"col":case"hr":case"keygen":case"meta":case"param":case"track":case"wbr":case"menuitem":for(ne in f)if(f.hasOwnProperty(ne)&&(h=f[ne],h!=null))switch(ne){case"children":case"dangerouslySetInnerHTML":throw Error(r(137,l));default:Et(o,l,ne,h,f,null)}return;default:if(Ag(l)){for(he in f)f.hasOwnProperty(he)&&(h=f[he],h!==void 0&&fb(o,l,he,h,f,void 0));return}}for(D in f)f.hasOwnProperty(D)&&(h=f[D],h!=null&&Et(o,l,D,h,f,null))}function u5(o,l,f,h){switch(l){case"div":case"span":case"svg":case"path":case"a":case"g":case"p":case"li":break;case"input":var y=null,_=null,j=null,D=null,U=null,ne=null,he=null;for(ae in f){var ge=f[ae];if(f.hasOwnProperty(ae)&&ge!=null)switch(ae){case"checked":break;case"value":break;case"defaultValue":U=ge;default:h.hasOwnProperty(ae)||Et(o,l,ae,null,h,ge)}}for(var ie in h){var ae=h[ie];if(ge=f[ie],h.hasOwnProperty(ie)&&(ae!=null||ge!=null))switch(ie){case"type":_=ae;break;case"name":y=ae;break;case"checked":ne=ae;break;case"defaultChecked":he=ae;break;case"value":j=ae;break;case"defaultValue":D=ae;break;case"children":case"dangerouslySetInnerHTML":if(ae!=null)throw Error(r(137,l));break;default:ae!==ge&&Et(o,l,ie,ae,h,ge)}}Cg(o,j,D,U,ne,he,_,y);return;case"select":ae=j=D=ie=null;for(_ in f)if(U=f[_],f.hasOwnProperty(_)&&U!=null)switch(_){case"value":break;case"multiple":ae=U;default:h.hasOwnProperty(_)||Et(o,l,_,null,h,U)}for(y in h)if(_=h[y],U=f[y],h.hasOwnProperty(y)&&(_!=null||U!=null))switch(y){case"value":ie=_;break;case"defaultValue":D=_;break;case"multiple":j=_;default:_!==U&&Et(o,l,y,_,h,U)}l=D,f=j,h=ae,ie!=null?Ms(o,!!f,ie,!1):!!h!=!!f&&(l!=null?Ms(o,!!f,l,!0):Ms(o,!!f,f?[]:"",!1));return;case"textarea":ae=ie=null;for(D in f)if(y=f[D],f.hasOwnProperty(D)&&y!=null&&!h.hasOwnProperty(D))switch(D){case"value":break;case"children":break;default:Et(o,l,D,null,h,y)}for(j in h)if(y=h[j],_=f[j],h.hasOwnProperty(j)&&(y!=null||_!=null))switch(j){case"value":ie=y;break;case"defaultValue":ae=y;break;case"children":break;case"dangerouslySetInnerHTML":if(y!=null)throw Error(r(91));break;default:y!==_&&Et(o,l,j,y,h,_)}eA(o,ie,ae);return;case"option":for(var Ve in f)if(ie=f[Ve],f.hasOwnProperty(Ve)&&ie!=null&&!h.hasOwnProperty(Ve))switch(Ve){case"selected":o.selected=!1;break;default:Et(o,l,Ve,null,h,ie)}for(U in h)if(ie=h[U],ae=f[U],h.hasOwnProperty(U)&&ie!==ae&&(ie!=null||ae!=null))switch(U){case"selected":o.selected=ie&&typeof ie!="function"&&typeof ie!="symbol";break;default:Et(o,l,U,ie,h,ae)}return;case"img":case"link":case"area":case"base":case"br":case"col":case"embed":case"hr":case"keygen":case"meta":case"param":case"source":case"track":case"wbr":case"menuitem":for(var Fe in f)ie=f[Fe],f.hasOwnProperty(Fe)&&ie!=null&&!h.hasOwnProperty(Fe)&&Et(o,l,Fe,null,h,ie);for(ne in h)if(ie=h[ne],ae=f[ne],h.hasOwnProperty(ne)&&ie!==ae&&(ie!=null||ae!=null))switch(ne){case"children":case"dangerouslySetInnerHTML":if(ie!=null)throw Error(r(137,l));break;default:Et(o,l,ne,ie,h,ae)}return;default:if(Ag(l)){for(var Ct in f)ie=f[Ct],f.hasOwnProperty(Ct)&&ie!==void 0&&!h.hasOwnProperty(Ct)&&fb(o,l,Ct,void 0,h,ie);for(he in h)ie=h[he],ae=f[he],!h.hasOwnProperty(he)||ie===ae||ie===void 0&&ae===void 0||fb(o,l,he,ie,h,ae);return}}for(var ee in f)ie=f[ee],f.hasOwnProperty(ee)&&ie!=null&&!h.hasOwnProperty(ee)&&Et(o,l,ee,null,h,ie);for(ge in h)ie=h[ge],ae=f[ge],!h.hasOwnProperty(ge)||ie===ae||ie==null&&ae==null||Et(o,l,ge,ie,h,ae)}var db=null,pb=null;function Mp(o){return o.nodeType===9?o:o.ownerDocument}function nM(o){switch(o){case"http://www.w3.org/2000/svg":return 1;case"http://www.w3.org/1998/Math/MathML":return 2;default:return 0}}function rM(o,l){if(o===0)switch(l){case"svg":return 1;case"math":return 2;default:return 0}return o===1&&l==="foreignObject"?0:o}function hb(o,l){return o==="textarea"||o==="noscript"||typeof l.children=="string"||typeof l.children=="number"||typeof l.children=="bigint"||typeof l.dangerouslySetInnerHTML=="object"&&l.dangerouslySetInnerHTML!==null&&l.dangerouslySetInnerHTML.__html!=null}var mb=null;function c5(){var o=window.event;return o&&o.type==="popstate"?o===mb?!1:(mb=o,!0):(mb=null,!1)}var iM=typeof setTimeout=="function"?setTimeout:void 0,f5=typeof clearTimeout=="function"?clearTimeout:void 0,aM=typeof Promise=="function"?Promise:void 0,d5=typeof queueMicrotask=="function"?queueMicrotask:typeof aM<"u"?function(o){return aM.resolve(null).then(o).catch(p5)}:iM;function p5(o){setTimeout(function(){throw o})}function za(o){return o==="head"}function oM(o,l){var f=l,h=0,y=0;do{var _=f.nextSibling;if(o.removeChild(f),_&&_.nodeType===8)if(f=_.data,f==="/$"){if(0<h&&8>h){f=h;var j=o.ownerDocument;if(f&1&&cc(j.documentElement),f&2&&cc(j.body),f&4)for(f=j.head,cc(f),j=f.firstChild;j;){var D=j.nextSibling,U=j.nodeName;j[Cu]||U==="SCRIPT"||U==="STYLE"||U==="LINK"&&j.rel.toLowerCase()==="stylesheet"||f.removeChild(j),j=D}}if(y===0){o.removeChild(_),yc(l);return}y--}else f==="$"||f==="$?"||f==="$!"?y++:h=f.charCodeAt(0)-48;else h=0;f=_}while(f);yc(l)}function vb(o){var l=o.firstChild;for(l&&l.nodeType===10&&(l=l.nextSibling);l;){var f=l;switch(l=l.nextSibling,f.nodeName){case"HTML":case"HEAD":case"BODY":vb(f),wg(f);continue;case"SCRIPT":case"STYLE":continue;case"LINK":if(f.rel.toLowerCase()==="stylesheet")continue}o.removeChild(f)}}function h5(o,l,f,h){for(;o.nodeType===1;){var y=f;if(o.nodeName.toLowerCase()!==l.toLowerCase()){if(!h&&(o.nodeName!=="INPUT"||o.type!=="hidden"))break}else if(h){if(!o[Cu])switch(l){case"meta":if(!o.hasAttribute("itemprop"))break;return o;case"link":if(_=o.getAttribute("rel"),_==="stylesheet"&&o.hasAttribute("data-precedence"))break;if(_!==y.rel||o.getAttribute("href")!==(y.href==null||y.href===""?null:y.href)||o.getAttribute("crossorigin")!==(y.crossOrigin==null?null:y.crossOrigin)||o.getAttribute("title")!==(y.title==null?null:y.title))break;return o;case"style":if(o.hasAttribute("data-precedence"))break;return o;case"script":if(_=o.getAttribute("src"),(_!==(y.src==null?null:y.src)||o.getAttribute("type")!==(y.type==null?null:y.type)||o.getAttribute("crossorigin")!==(y.crossOrigin==null?null:y.crossOrigin))&&_&&o.hasAttribute("async")&&!o.hasAttribute("itemprop"))break;return o;default:return o}}else if(l==="input"&&o.type==="hidden"){var _=y.name==null?null:""+y.name;if(y.type==="hidden"&&o.getAttribute("name")===_)return o}else return o;if(o=Fr(o.nextSibling),o===null)break}return null}function m5(o,l,f){if(l==="")return null;for(;o.nodeType!==3;)if((o.nodeType!==1||o.nodeName!=="INPUT"||o.type!=="hidden")&&!f||(o=Fr(o.nextSibling),o===null))return null;return o}function gb(o){return o.data==="$!"||o.data==="$?"&&o.ownerDocument.readyState==="complete"}function v5(o,l){var f=o.ownerDocument;if(o.data!=="$?"||f.readyState==="complete")l();else{var h=function(){l(),f.removeEventListener("DOMContentLoaded",h)};f.addEventListener("DOMContentLoaded",h),o._reactRetry=h}}function Fr(o){for(;o!=null;o=o.nextSibling){var l=o.nodeType;if(l===1||l===3)break;if(l===8){if(l=o.data,l==="$"||l==="$!"||l==="$?"||l==="F!"||l==="F")break;if(l==="/$")return null}}return o}var yb=null;function sM(o){o=o.previousSibling;for(var l=0;o;){if(o.nodeType===8){var f=o.data;if(f==="$"||f==="$!"||f==="$?"){if(l===0)return o;l--}else f==="/$"&&l++}o=o.previousSibling}return null}function lM(o,l,f){switch(l=Mp(f),o){case"html":if(o=l.documentElement,!o)throw Error(r(452));return o;case"head":if(o=l.head,!o)throw Error(r(453));return o;case"body":if(o=l.body,!o)throw Error(r(454));return o;default:throw Error(r(451))}}function cc(o){for(var l=o.attributes;l.length;)o.removeAttributeNode(l[0]);wg(o)}var Or=new Map,uM=new Set;function Pp(o){return typeof o.getRootNode=="function"?o.getRootNode():o.nodeType===9?o:o.ownerDocument}var Wi=W.d;W.d={f:g5,r:y5,D:b5,C:x5,L:w5,m:S5,X:E5,S:_5,M:C5};function g5(){var o=Wi.f(),l=Sp();return o||l}function y5(o){var l=Os(o);l!==null&&l.tag===5&&l.type==="form"?TR(l):Wi.r(o)}var ol=typeof document>"u"?null:document;function cM(o,l,f){var h=ol;if(h&&typeof l=="string"&&l){var y=br(l);y='link[rel="'+o+'"][href="'+y+'"]',typeof f=="string"&&(y+='[crossorigin="'+f+'"]'),uM.has(y)||(uM.add(y),o={rel:o,crossOrigin:f,href:l},h.querySelector(y)===null&&(l=h.createElement("link"),Sn(l,"link",o),pn(l),h.head.appendChild(l)))}}function b5(o){Wi.D(o),cM("dns-prefetch",o,null)}function x5(o,l){Wi.C(o,l),cM("preconnect",o,l)}function w5(o,l,f){Wi.L(o,l,f);var h=ol;if(h&&o&&l){var y='link[rel="preload"][as="'+br(l)+'"]';l==="image"&&f&&f.imageSrcSet?(y+='[imagesrcset="'+br(f.imageSrcSet)+'"]',typeof f.imageSizes=="string"&&(y+='[imagesizes="'+br(f.imageSizes)+'"]')):y+='[href="'+br(o)+'"]';var _=y;switch(l){case"style":_=sl(o);break;case"script":_=ll(o)}Or.has(_)||(o=p({rel:"preload",href:l==="image"&&f&&f.imageSrcSet?void 0:o,as:l},f),Or.set(_,o),h.querySelector(y)!==null||l==="style"&&h.querySelector(fc(_))||l==="script"&&h.querySelector(dc(_))||(l=h.createElement("link"),Sn(l,"link",o),pn(l),h.head.appendChild(l)))}}function S5(o,l){Wi.m(o,l);var f=ol;if(f&&o){var h=l&&typeof l.as=="string"?l.as:"script",y='link[rel="modulepreload"][as="'+br(h)+'"][href="'+br(o)+'"]',_=y;switch(h){case"audioworklet":case"paintworklet":case"serviceworker":case"sharedworker":case"worker":case"script":_=ll(o)}if(!Or.has(_)&&(o=p({rel:"modulepreload",href:o},l),Or.set(_,o),f.querySelector(y)===null)){switch(h){case"audioworklet":case"paintworklet":case"serviceworker":case"sharedworker":case"worker":case"script":if(f.querySelector(dc(_)))return}h=f.createElement("link"),Sn(h,"link",o),pn(h),f.head.appendChild(h)}}}function _5(o,l,f){Wi.S(o,l,f);var h=ol;if(h&&o){var y=As(h).hoistableStyles,_=sl(o);l=l||"default";var j=y.get(_);if(!j){var D={loading:0,preload:null};if(j=h.querySelector(fc(_)))D.loading=5;else{o=p({rel:"stylesheet",href:o,"data-precedence":l},f),(f=Or.get(_))&&bb(o,f);var U=j=h.createElement("link");pn(U),Sn(U,"link",o),U._p=new Promise(function(ne,he){U.onload=ne,U.onerror=he}),U.addEventListener("load",function(){D.loading|=1}),U.addEventListener("error",function(){D.loading|=2}),D.loading|=4,jp(j,l,h)}j={type:"stylesheet",instance:j,count:1,state:D},y.set(_,j)}}}function E5(o,l){Wi.X(o,l);var f=ol;if(f&&o){var h=As(f).hoistableScripts,y=ll(o),_=h.get(y);_||(_=f.querySelector(dc(y)),_||(o=p({src:o,async:!0},l),(l=Or.get(y))&&xb(o,l),_=f.createElement("script"),pn(_),Sn(_,"link",o),f.head.appendChild(_)),_={type:"script",instance:_,count:1,state:null},h.set(y,_))}}function C5(o,l){Wi.M(o,l);var f=ol;if(f&&o){var h=As(f).hoistableScripts,y=ll(o),_=h.get(y);_||(_=f.querySelector(dc(y)),_||(o=p({src:o,async:!0,type:"module"},l),(l=Or.get(y))&&xb(o,l),_=f.createElement("script"),pn(_),Sn(_,"link",o),f.head.appendChild(_)),_={type:"script",instance:_,count:1,state:null},h.set(y,_))}}function fM(o,l,f,h){var y=(y=ve.current)?Pp(y):null;if(!y)throw Error(r(446));switch(o){case"meta":case"title":return null;case"style":return typeof f.precedence=="string"&&typeof f.href=="string"?(l=sl(f.href),f=As(y).hoistableStyles,h=f.get(l),h||(h={type:"style",instance:null,count:0,state:null},f.set(l,h)),h):{type:"void",instance:null,count:0,state:null};case"link":if(f.rel==="stylesheet"&&typeof f.href=="string"&&typeof f.precedence=="string"){o=sl(f.href);var _=As(y).hoistableStyles,j=_.get(o);if(j||(y=y.ownerDocument||y,j={type:"stylesheet",instance:null,count:0,state:{loading:0,preload:null}},_.set(o,j),(_=y.querySelector(fc(o)))&&!_._p&&(j.instance=_,j.state.loading=5),Or.has(o)||(f={rel:"preload",as:"style",href:f.href,crossOrigin:f.crossOrigin,integrity:f.integrity,media:f.media,hrefLang:f.hrefLang,referrerPolicy:f.referrerPolicy},Or.set(o,f),_||O5(y,o,f,j.state))),l&&h===null)throw Error(r(528,""));return j}if(l&&h!==null)throw Error(r(529,""));return null;case"script":return l=f.async,f=f.src,typeof f=="string"&&l&&typeof l!="function"&&typeof l!="symbol"?(l=ll(f),f=As(y).hoistableScripts,h=f.get(l),h||(h={type:"script",instance:null,count:0,state:null},f.set(l,h)),h):{type:"void",instance:null,count:0,state:null};default:throw Error(r(444,o))}}function sl(o){return'href="'+br(o)+'"'}function fc(o){return'link[rel="stylesheet"]['+o+"]"}function dM(o){return p({},o,{"data-precedence":o.precedence,precedence:null})}function O5(o,l,f,h){o.querySelector('link[rel="preload"][as="style"]['+l+"]")?h.loading=1:(l=o.createElement("link"),h.preload=l,l.addEventListener("load",function(){return h.loading|=1}),l.addEventListener("error",function(){return h.loading|=2}),Sn(l,"link",f),pn(l),o.head.appendChild(l))}function ll(o){return'[src="'+br(o)+'"]'}function dc(o){return"script[async]"+o}function pM(o,l,f){if(l.count++,l.instance===null)switch(l.type){case"style":var h=o.querySelector('style[data-href~="'+br(f.href)+'"]');if(h)return l.instance=h,pn(h),h;var y=p({},f,{"data-href":f.href,"data-precedence":f.precedence,href:null,precedence:null});return h=(o.ownerDocument||o).createElement("style"),pn(h),Sn(h,"style",y),jp(h,f.precedence,o),l.instance=h;case"stylesheet":y=sl(f.href);var _=o.querySelector(fc(y));if(_)return l.state.loading|=4,l.instance=_,pn(_),_;h=dM(f),(y=Or.get(y))&&bb(h,y),_=(o.ownerDocument||o).createElement("link"),pn(_);var j=_;return j._p=new Promise(function(D,U){j.onload=D,j.onerror=U}),Sn(_,"link",h),l.state.loading|=4,jp(_,f.precedence,o),l.instance=_;case"script":return _=ll(f.src),(y=o.querySelector(dc(_)))?(l.instance=y,pn(y),y):(h=f,(y=Or.get(_))&&(h=p({},f),xb(h,y)),o=o.ownerDocument||o,y=o.createElement("script"),pn(y),Sn(y,"link",h),o.head.appendChild(y),l.instance=y);case"void":return null;default:throw Error(r(443,l.type))}else l.type==="stylesheet"&&(l.state.loading&4)===0&&(h=l.instance,l.state.loading|=4,jp(h,f.precedence,o));return l.instance}function jp(o,l,f){for(var h=f.querySelectorAll('link[rel="stylesheet"][data-precedence],style[data-precedence]'),y=h.length?h[h.length-1]:null,_=y,j=0;j<h.length;j++){var D=h[j];if(D.dataset.precedence===l)_=D;else if(_!==y)break}_?_.parentNode.insertBefore(o,_.nextSibling):(l=f.nodeType===9?f.head:f,l.insertBefore(o,l.firstChild))}function bb(o,l){o.crossOrigin==null&&(o.crossOrigin=l.crossOrigin),o.referrerPolicy==null&&(o.referrerPolicy=l.referrerPolicy),o.title==null&&(o.title=l.title)}function xb(o,l){o.crossOrigin==null&&(o.crossOrigin=l.crossOrigin),o.referrerPolicy==null&&(o.referrerPolicy=l.referrerPolicy),o.integrity==null&&(o.integrity=l.integrity)}var Np=null;function hM(o,l,f){if(Np===null){var h=new Map,y=Np=new Map;y.set(f,h)}else y=Np,h=y.get(f),h||(h=new Map,y.set(f,h));if(h.has(o))return h;for(h.set(o,null),f=f.getElementsByTagName(o),y=0;y<f.length;y++){var _=f[y];if(!(_[Cu]||_[Cn]||o==="link"&&_.getAttribute("rel")==="stylesheet")&&_.namespaceURI!=="http://www.w3.org/2000/svg"){var j=_.getAttribute(l)||"";j=o+j;var D=h.get(j);D?D.push(_):h.set(j,[_])}}return h}function mM(o,l,f){o=o.ownerDocument||o,o.head.insertBefore(f,l==="title"?o.querySelector("head > title"):null)}function A5(o,l,f){if(f===1||l.itemProp!=null)return!1;switch(o){case"meta":case"title":return!0;case"style":if(typeof l.precedence!="string"||typeof l.href!="string"||l.href==="")break;return!0;case"link":if(typeof l.rel!="string"||typeof l.href!="string"||l.href===""||l.onLoad||l.onError)break;switch(l.rel){case"stylesheet":return o=l.disabled,typeof l.precedence=="string"&&o==null;default:return!0}case"script":if(l.async&&typeof l.async!="function"&&typeof l.async!="symbol"&&!l.onLoad&&!l.onError&&l.src&&typeof l.src=="string")return!0}return!1}function vM(o){return!(o.type==="stylesheet"&&(o.state.loading&3)===0)}var pc=null;function R5(){}function T5(o,l,f){if(pc===null)throw Error(r(475));var h=pc;if(l.type==="stylesheet"&&(typeof f.media!="string"||matchMedia(f.media).matches!==!1)&&(l.state.loading&4)===0){if(l.instance===null){var y=sl(f.href),_=o.querySelector(fc(y));if(_){o=_._p,o!==null&&typeof o=="object"&&typeof o.then=="function"&&(h.count++,h=Dp.bind(h),o.then(h,h)),l.state.loading|=4,l.instance=_,pn(_);return}_=o.ownerDocument||o,f=dM(f),(y=Or.get(y))&&bb(f,y),_=_.createElement("link"),pn(_);var j=_;j._p=new Promise(function(D,U){j.onload=D,j.onerror=U}),Sn(_,"link",f),l.instance=_}h.stylesheets===null&&(h.stylesheets=new Map),h.stylesheets.set(l,o),(o=l.state.preload)&&(l.state.loading&3)===0&&(h.count++,l=Dp.bind(h),o.addEventListener("load",l),o.addEventListener("error",l))}}function M5(){if(pc===null)throw Error(r(475));var o=pc;return o.stylesheets&&o.count===0&&wb(o,o.stylesheets),0<o.count?function(l){var f=setTimeout(function(){if(o.stylesheets&&wb(o,o.stylesheets),o.unsuspend){var h=o.unsuspend;o.unsuspend=null,h()}},6e4);return o.unsuspend=l,function(){o.unsuspend=null,clearTimeout(f)}}:null}function Dp(){if(this.count--,this.count===0){if(this.stylesheets)wb(this,this.stylesheets);else if(this.unsuspend){var o=this.unsuspend;this.unsuspend=null,o()}}}var kp=null;function wb(o,l){o.stylesheets=null,o.unsuspend!==null&&(o.count++,kp=new Map,l.forEach(P5,o),kp=null,Dp.call(o))}function P5(o,l){if(!(l.state.loading&4)){var f=kp.get(o);if(f)var h=f.get(null);else{f=new Map,kp.set(o,f);for(var y=o.querySelectorAll("link[data-precedence],style[data-precedence]"),_=0;_<y.length;_++){var j=y[_];(j.nodeName==="LINK"||j.getAttribute("media")!=="not all")&&(f.set(j.dataset.precedence,j),h=j)}h&&f.set(null,h)}y=l.instance,j=y.getAttribute("data-precedence"),_=f.get(j)||h,_===h&&f.set(null,y),f.set(j,y),this.count++,h=Dp.bind(this),y.addEventListener("load",h),y.addEventListener("error",h),_?_.parentNode.insertBefore(y,_.nextSibling):(o=o.nodeType===9?o.head:o,o.insertBefore(y,o.firstChild)),l.state.loading|=4}}var hc={$$typeof:A,Provider:null,Consumer:null,_currentValue:B,_currentValue2:B,_threadCount:0};function j5(o,l,f,h,y,_,j,D){this.tag=1,this.containerInfo=o,this.pingCache=this.current=this.pendingChildren=null,this.timeoutHandle=-1,this.callbackNode=this.next=this.pendingContext=this.context=this.cancelPendingCommit=null,this.callbackPriority=0,this.expirationTimes=gg(-1),this.entangledLanes=this.shellSuspendCounter=this.errorRecoveryDisabledLanes=this.expiredLanes=this.warmLanes=this.pingedLanes=this.suspendedLanes=this.pendingLanes=0,this.entanglements=gg(0),this.hiddenUpdates=gg(null),this.identifierPrefix=h,this.onUncaughtError=y,this.onCaughtError=_,this.onRecoverableError=j,this.pooledCache=null,this.pooledCacheLanes=0,this.formState=D,this.incompleteTransitions=new Map}function gM(o,l,f,h,y,_,j,D,U,ne,he,ge){return o=new j5(o,l,f,j,D,U,ne,ge),l=1,_===!0&&(l|=24),_=er(3,null,null,l),o.current=_,_.stateNode=o,l=ny(),l.refCount++,o.pooledCache=l,l.refCount++,_.memoizedState={element:h,isDehydrated:f,cache:l},oy(_),o}function yM(o){return o?(o=Bs,o):Bs}function bM(o,l,f,h,y,_){y=yM(y),h.context===null?h.context=y:h.pendingContext=y,h=Ta(l),h.payload={element:f},_=_===void 0?null:_,_!==null&&(h.callback=_),f=Ma(o,h,l),f!==null&&(ar(f,o,l),Vu(f,o,l))}function xM(o,l){if(o=o.memoizedState,o!==null&&o.dehydrated!==null){var f=o.retryLane;o.retryLane=f!==0&&f<l?f:l}}function Sb(o,l){xM(o,l),(o=o.alternate)&&xM(o,l)}function wM(o){if(o.tag===13){var l=Ls(o,67108864);l!==null&&ar(l,o,67108864),Sb(o,67108864)}}var $p=!0;function N5(o,l,f,h){var y=I.T;I.T=null;var _=W.p;try{W.p=2,_b(o,l,f,h)}finally{W.p=_,I.T=y}}function D5(o,l,f,h){var y=I.T;I.T=null;var _=W.p;try{W.p=8,_b(o,l,f,h)}finally{W.p=_,I.T=y}}function _b(o,l,f,h){if($p){var y=Eb(h);if(y===null)cb(o,l,h,Ip,f),_M(o,h);else if($5(y,o,l,f,h))h.stopPropagation();else if(_M(o,h),l&4&&-1<k5.indexOf(o)){for(;y!==null;){var _=Os(y);if(_!==null)switch(_.tag){case 3:if(_=_.stateNode,_.current.memoizedState.isDehydrated){var j=ji(_.pendingLanes);if(j!==0){var D=_;for(D.pendingLanes|=2,D.entangledLanes|=2;j;){var U=1<<31-Ke(j);D.entanglements[1]|=U,j&=~U}li(_),(wt&6)===0&&(xp=it()+500,sc(0))}}break;case 13:D=Ls(_,2),D!==null&&ar(D,_,2),Sp(),Sb(_,2)}if(_=Eb(h),_===null&&cb(o,l,h,Ip,f),_===y)break;y=_}y!==null&&h.stopPropagation()}else cb(o,l,h,null,f)}}function Eb(o){return o=Tg(o),Cb(o)}var Ip=null;function Cb(o){if(Ip=null,o=Cs(o),o!==null){var l=a(o);if(l===null)o=null;else{var f=l.tag;if(f===13){if(o=s(l),o!==null)return o;o=null}else if(f===3){if(l.stateNode.current.memoizedState.isDehydrated)return l.tag===3?l.stateNode.containerInfo:null;o=null}else l!==o&&(o=null)}}return Ip=o,null}function SM(o){switch(o){case"beforetoggle":case"cancel":case"click":case"close":case"contextmenu":case"copy":case"cut":case"auxclick":case"dblclick":case"dragend":case"dragstart":case"drop":case"focusin":case"focusout":case"input":case"invalid":case"keydown":case"keypress":case"keyup":case"mousedown":case"mouseup":case"paste":case"pause":case"play":case"pointercancel":case"pointerdown":case"pointerup":case"ratechange":case"reset":case"resize":case"seeked":case"submit":case"toggle":case"touchcancel":case"touchend":case"touchstart":case"volumechange":case"change":case"selectionchange":case"textInput":case"compositionstart":case"compositionend":case"compositionupdate":case"beforeblur":case"afterblur":case"beforeinput":case"blur":case"fullscreenchange":case"focus":case"hashchange":case"popstate":case"select":case"selectstart":return 2;case"drag":case"dragenter":case"dragexit":case"dragleave":case"dragover":case"mousemove":case"mouseout":case"mouseover":case"pointermove":case"pointerout":case"pointerover":case"scroll":case"touchmove":case"wheel":case"mouseenter":case"mouseleave":case"pointerenter":case"pointerleave":return 8;case"message":switch(xt()){case Tt:return 2;case bn:return 8;case Bt:case G:return 32;case re:return 268435456;default:return 32}default:return 32}}var Ob=!1,Ua=null,Va=null,Ha=null,mc=new Map,vc=new Map,Ga=[],k5="mousedown mouseup touchcancel touchend touchstart auxclick dblclick pointercancel pointerdown pointerup dragend dragstart drop compositionend compositionstart keydown keypress keyup input textInput copy cut paste click change contextmenu reset".split(" ");function _M(o,l){switch(o){case"focusin":case"focusout":Ua=null;break;case"dragenter":case"dragleave":Va=null;break;case"mouseover":case"mouseout":Ha=null;break;case"pointerover":case"pointerout":mc.delete(l.pointerId);break;case"gotpointercapture":case"lostpointercapture":vc.delete(l.pointerId)}}function gc(o,l,f,h,y,_){return o===null||o.nativeEvent!==_?(o={blockedOn:l,domEventName:f,eventSystemFlags:h,nativeEvent:_,targetContainers:[y]},l!==null&&(l=Os(l),l!==null&&wM(l)),o):(o.eventSystemFlags|=h,l=o.targetContainers,y!==null&&l.indexOf(y)===-1&&l.push(y),o)}function $5(o,l,f,h,y){switch(l){case"focusin":return Ua=gc(Ua,o,l,f,h,y),!0;case"dragenter":return Va=gc(Va,o,l,f,h,y),!0;case"mouseover":return Ha=gc(Ha,o,l,f,h,y),!0;case"pointerover":var _=y.pointerId;return mc.set(_,gc(mc.get(_)||null,o,l,f,h,y)),!0;case"gotpointercapture":return _=y.pointerId,vc.set(_,gc(vc.get(_)||null,o,l,f,h,y)),!0}return!1}function EM(o){var l=Cs(o.target);if(l!==null){var f=a(l);if(f!==null){if(l=f.tag,l===13){if(l=s(f),l!==null){o.blockedOn=l,T4(o.priority,function(){if(f.tag===13){var h=ir();h=yg(h);var y=Ls(f,h);y!==null&&ar(y,f,h),Sb(f,h)}});return}}else if(l===3&&f.stateNode.current.memoizedState.isDehydrated){o.blockedOn=f.tag===3?f.stateNode.containerInfo:null;return}}}o.blockedOn=null}function Lp(o){if(o.blockedOn!==null)return!1;for(var l=o.targetContainers;0<l.length;){var f=Eb(o.nativeEvent);if(f===null){f=o.nativeEvent;var h=new f.constructor(f.type,f);Rg=h,f.target.dispatchEvent(h),Rg=null}else return l=Os(f),l!==null&&wM(l),o.blockedOn=f,!1;l.shift()}return!0}function CM(o,l,f){Lp(o)&&f.delete(l)}function I5(){Ob=!1,Ua!==null&&Lp(Ua)&&(Ua=null),Va!==null&&Lp(Va)&&(Va=null),Ha!==null&&Lp(Ha)&&(Ha=null),mc.forEach(CM),vc.forEach(CM)}function Bp(o,l){o.blockedOn===l&&(o.blockedOn=null,Ob||(Ob=!0,e.unstable_scheduleCallback(e.unstable_NormalPriority,I5)))}var Fp=null;function OM(o){Fp!==o&&(Fp=o,e.unstable_scheduleCallback(e.unstable_NormalPriority,function(){Fp===o&&(Fp=null);for(var l=0;l<o.length;l+=3){var f=o[l],h=o[l+1],y=o[l+2];if(typeof h!="function"){if(Cb(h||f)===null)continue;break}var _=Os(f);_!==null&&(o.splice(l,3),l-=3,Cy(_,{pending:!0,data:y,method:f.method,action:h},h,y))}}))}function yc(o){function l(U){return Bp(U,o)}Ua!==null&&Bp(Ua,o),Va!==null&&Bp(Va,o),Ha!==null&&Bp(Ha,o),mc.forEach(l),vc.forEach(l);for(var f=0;f<Ga.length;f++){var h=Ga[f];h.blockedOn===o&&(h.blockedOn=null)}for(;0<Ga.length&&(f=Ga[0],f.blockedOn===null);)EM(f),f.blockedOn===null&&Ga.shift();if(f=(o.ownerDocument||o).$$reactFormReplay,f!=null)for(h=0;h<f.length;h+=3){var y=f[h],_=f[h+1],j=y[Vn]||null;if(typeof _=="function")j||OM(f);else if(j){var D=null;if(_&&_.hasAttribute("formAction")){if(y=_,j=_[Vn]||null)D=j.formAction;else if(Cb(y)!==null)continue}else D=j.action;typeof D=="function"?f[h+1]=D:(f.splice(h,3),h-=3),OM(f)}}}function Ab(o){this._internalRoot=o}qp.prototype.render=Ab.prototype.render=function(o){var l=this._internalRoot;if(l===null)throw Error(r(409));var f=l.current,h=ir();bM(f,h,o,l,null,null)},qp.prototype.unmount=Ab.prototype.unmount=function(){var o=this._internalRoot;if(o!==null){this._internalRoot=null;var l=o.containerInfo;bM(o.current,2,null,o,null,null),Sp(),l[Es]=null}};function qp(o){this._internalRoot=o}qp.prototype.unstable_scheduleHydration=function(o){if(o){var l=UO();o={blockedOn:null,target:o,priority:l};for(var f=0;f<Ga.length&&l!==0&&l<Ga[f].priority;f++);Ga.splice(f,0,o),f===0&&EM(o)}};var AM=t.version;if(AM!=="19.1.0")throw Error(r(527,AM,"19.1.0"));W.findDOMNode=function(o){var l=o._reactInternals;if(l===void 0)throw typeof o.render=="function"?Error(r(188)):(o=Object.keys(o).join(","),Error(r(268,o)));return o=c(l),o=o!==null?d(o):null,o=o===null?null:o.stateNode,o};var L5={bundleType:0,version:"19.1.0",rendererPackageName:"react-dom",currentDispatcherRef:I,reconcilerVersion:"19.1.0"};if(typeof __REACT_DEVTOOLS_GLOBAL_HOOK__<"u"){var zp=__REACT_DEVTOOLS_GLOBAL_HOOK__;if(!zp.isDisabled&&zp.supportsFiber)try{Ee=zp.inject(L5),ye=zp}catch{}}return xc.createRoot=function(o,l){if(!i(o))throw Error(r(299));var f=!1,h="",y=UR,_=VR,j=HR,D=null;return l!=null&&(l.unstable_strictMode===!0&&(f=!0),l.identifierPrefix!==void 0&&(h=l.identifierPrefix),l.onUncaughtError!==void 0&&(y=l.onUncaughtError),l.onCaughtError!==void 0&&(_=l.onCaughtError),l.onRecoverableError!==void 0&&(j=l.onRecoverableError),l.unstable_transitionCallbacks!==void 0&&(D=l.unstable_transitionCallbacks)),l=gM(o,1,!1,null,null,f,h,y,_,j,D,null),o[Es]=l.current,ub(o),new Ab(l)},xc.hydrateRoot=function(o,l,f){if(!i(o))throw Error(r(299));var h=!1,y="",_=UR,j=VR,D=HR,U=null,ne=null;return f!=null&&(f.unstable_strictMode===!0&&(h=!0),f.identifierPrefix!==void 0&&(y=f.identifierPrefix),f.onUncaughtError!==void 0&&(_=f.onUncaughtError),f.onCaughtError!==void 0&&(j=f.onCaughtError),f.onRecoverableError!==void 0&&(D=f.onRecoverableError),f.unstable_transitionCallbacks!==void 0&&(U=f.unstable_transitionCallbacks),f.formState!==void 0&&(ne=f.formState)),l=gM(o,1,!0,l,f??null,h,y,_,j,D,U,ne),l.context=yM(null),f=l.current,h=ir(),h=yg(h),y=Ta(h),y.callback=null,Ma(f,y,h),f=h,l.current.lanes=f,Eu(l,f),li(l),o[Es]=l.current,ub(o),new qp(l)},xc.version="19.1.0",xc}var IM;function K5(){if(IM)return Mb.exports;IM=1;function e(){if(!(typeof __REACT_DEVTOOLS_GLOBAL_HOOK__>"u"||typeof __REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE!="function"))try{__REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE(e)}catch(t){console.error(t)}}return e(),Mb.exports=W5(),Mb.exports}var Y5=K5(),ba=XI();const QI=Rt(ba);function Z5(e,t){const n=g.createContext(t),r=a=>{const{children:s,...u}=a,c=g.useMemo(()=>u,Object.values(u));return E.jsx(n.Provider,{value:c,children:s})};r.displayName=e+"Provider";function i(a){const s=g.useContext(n);if(s)return s;if(t!==void 0)return t;throw new Error(`\`${a}\` must be used within \`${e}\``)}return[r,i]}function jn(e,t=[]){let n=[];function r(a,s){const u=g.createContext(s),c=n.length;n=[...n,s];const d=m=>{var C;const{scope:v,children:b,...S}=m,w=((C=v==null?void 0:v[e])==null?void 0:C[c])||u,x=g.useMemo(()=>S,Object.values(S));return E.jsx(w.Provider,{value:x,children:b})};d.displayName=a+"Provider";function p(m,v){var w;const b=((w=v==null?void 0:v[e])==null?void 0:w[c])||u,S=g.useContext(b);if(S)return S;if(s!==void 0)return s;throw new Error(`\`${m}\` must be used within \`${a}\``)}return[d,p]}const i=()=>{const a=n.map(s=>g.createContext(s));return function(u){const c=(u==null?void 0:u[e])||a;return g.useMemo(()=>({[`__scope${e}`]:{...u,[e]:c}}),[u,c])}};return i.scopeName=e,[r,X5(i,...t)]}function X5(...e){const t=e[0];if(e.length===1)return t;const n=()=>{const r=e.map(i=>({useScope:i(),scopeName:i.scopeName}));return function(a){const s=r.reduce((u,{useScope:c,scopeName:d})=>{const m=c(a)[`__scope${d}`];return{...u,...m}},{});return g.useMemo(()=>({[`__scope${t.scopeName}`]:s}),[s])}};return n.scopeName=t.scopeName,n}function be(e,t,{checkForDefaultPrevented:n=!0}={}){return function(i){if(e==null||e(i),n===!1||!i.defaultPrevented)return t==null?void 0:t(i)}}function LM(e,t){if(typeof e=="function")return e(t);e!=null&&(e.current=t)}function wi(...e){return t=>{let n=!1;const r=e.map(i=>{const a=LM(i,t);return!n&&typeof a=="function"&&(n=!0),a});if(n)return()=>{for(let i=0;i<r.length;i++){const a=r[i];typeof a=="function"?a():LM(e[i],null)}}}}function Re(...e){return g.useCallback(wi(...e),e)}function Ah(e){const t=Q5(e),n=g.forwardRef((r,i)=>{const{children:a,...s}=r,u=g.Children.toArray(a),c=u.find(e6);if(c){const d=c.props.children,p=u.map(m=>m===c?g.Children.count(d)>1?g.Children.only(null):g.isValidElement(d)?d.props.children:null:m);return E.jsx(t,{...s,ref:i,children:g.isValidElement(d)?g.cloneElement(d,void 0,p):null})}return E.jsx(t,{...s,ref:i,children:a})});return n.displayName=`${e}.Slot`,n}var kr=Ah("Slot");function Q5(e){const t=g.forwardRef((n,r)=>{const{children:i,...a}=n;if(g.isValidElement(i)){const s=n6(i),u=t6(a,i.props);return i.type!==g.Fragment&&(u.ref=r?wi(r,s):s),g.cloneElement(i,u)}return g.Children.count(i)>1?g.Children.only(null):null});return t.displayName=`${e}.SlotClone`,t}var J5=Symbol("radix.slottable");function e6(e){return g.isValidElement(e)&&typeof e.type=="function"&&"__radixId"in e.type&&e.type.__radixId===J5}function t6(e,t){const n={...t};for(const r in t){const i=e[r],a=t[r];/^on[A-Z]/.test(r)?i&&a?n[r]=(...u)=>{a(...u),i(...u)}:i&&(n[r]=i):r==="style"?n[r]={...i,...a}:r==="className"&&(n[r]=[i,a].filter(Boolean).join(" "))}return{...e,...n}}function n6(e){var r,i;let t=(r=Object.getOwnPropertyDescriptor(e.props,"ref"))==null?void 0:r.get,n=t&&"isReactWarning"in t&&t.isReactWarning;return n?e.ref:(t=(i=Object.getOwnPropertyDescriptor(e,"ref"))==null?void 0:i.get,n=t&&"isReactWarning"in t&&t.isReactWarning,n?e.props.ref:e.props.ref||e.ref)}var r6=["a","button","div","form","h2","h3","img","input","label","li","nav","ol","p","select","span","svg","ul"],Ut=r6.reduce((e,t)=>{const n=Ah(`Primitive.${t}`),r=g.forwardRef((i,a)=>{const{asChild:s,...u}=i,c=s?n:t;return typeof window<"u"&&(window[Symbol.for("radix-ui")]=!0),E.jsx(c,{...u,ref:a})});return r.displayName=`Primitive.${t}`,{...e,[t]:r}},{});function d_(e,t){e&&ba.flushSync(()=>e.dispatchEvent(t))}var cn=globalThis!=null&&globalThis.document?g.useLayoutEffect:()=>{},i6=ZI[" useInsertionEffect ".trim().toString()]||cn;function Xr({prop:e,defaultProp:t,onChange:n=()=>{},caller:r}){const[i,a,s]=a6({defaultProp:t,onChange:n}),u=e!==void 0,c=u?e:i;{const p=g.useRef(e!==void 0);g.useEffect(()=>{const m=p.current;m!==u&&console.warn(`${r} is changing from ${m?"controlled":"uncontrolled"} to ${u?"controlled":"uncontrolled"}. Components should not switch from controlled to uncontrolled (or vice versa). Decide between using a controlled or uncontrolled value for the lifetime of the component.`),p.current=u},[u,r])}const d=g.useCallback(p=>{var m;if(u){const v=o6(p)?p(e):p;v!==e&&((m=s.current)==null||m.call(s,v))}else a(p)},[u,e,a,s]);return[c,d]}function a6({defaultProp:e,onChange:t}){const[n,r]=g.useState(e),i=g.useRef(n),a=g.useRef(t);return i6(()=>{a.current=t},[t]),g.useEffect(()=>{var s;i.current!==n&&((s=a.current)==null||s.call(a,n),i.current=n)},[n,i]),[n,r,a]}function o6(e){return typeof e=="function"}var s6=g.createContext(void 0);function iu(e){const t=g.useContext(s6);return e||t||"ltr"}function l6(e,t){return g.useReducer((n,r)=>t[n][r]??n,e)}var zn=e=>{const{present:t,children:n}=e,r=u6(t),i=typeof n=="function"?n({present:r.isPresent}):g.Children.only(n),a=Re(r.ref,c6(i));return typeof n=="function"||r.isPresent?g.cloneElement(i,{ref:a}):null};zn.displayName="Presence";function u6(e){const[t,n]=g.useState(),r=g.useRef(null),i=g.useRef(e),a=g.useRef("none"),s=e?"mounted":"unmounted",[u,c]=l6(s,{mounted:{UNMOUNT:"unmounted",ANIMATION_OUT:"unmountSuspended"},unmountSuspended:{MOUNT:"mounted",ANIMATION_END:"unmounted"},unmounted:{MOUNT:"mounted"}});return g.useEffect(()=>{const d=Vp(r.current);a.current=u==="mounted"?d:"none"},[u]),cn(()=>{const d=r.current,p=i.current;if(p!==e){const v=a.current,b=Vp(d);e?c("MOUNT"):b==="none"||(d==null?void 0:d.display)==="none"?c("UNMOUNT"):c(p&&v!==b?"ANIMATION_OUT":"UNMOUNT"),i.current=e}},[e,c]),cn(()=>{if(t){let d;const p=t.ownerDocument.defaultView??window,m=b=>{const w=Vp(r.current).includes(b.animationName);if(b.target===t&&w&&(c("ANIMATION_END"),!i.current)){const x=t.style.animationFillMode;t.style.animationFillMode="forwards",d=p.setTimeout(()=>{t.style.animationFillMode==="forwards"&&(t.style.animationFillMode=x)})}},v=b=>{b.target===t&&(a.current=Vp(r.current))};return t.addEventListener("animationstart",v),t.addEventListener("animationcancel",m),t.addEventListener("animationend",m),()=>{p.clearTimeout(d),t.removeEventListener("animationstart",v),t.removeEventListener("animationcancel",m),t.removeEventListener("animationend",m)}}else c("ANIMATION_END")},[t,c]),{isPresent:["mounted","unmountSuspended"].includes(u),ref:g.useCallback(d=>{r.current=d?getComputedStyle(d):null,n(d)},[])}}function Vp(e){return(e==null?void 0:e.animationName)||"none"}function c6(e){var r,i;let t=(r=Object.getOwnPropertyDescriptor(e.props,"ref"))==null?void 0:r.get,n=t&&"isReactWarning"in t&&t.isReactWarning;return n?e.ref:(t=(i=Object.getOwnPropertyDescriptor(e,"ref"))==null?void 0:i.get,n=t&&"isReactWarning"in t&&t.isReactWarning,n?e.props.ref:e.props.ref||e.ref)}var f6=ZI[" useId ".trim().toString()]||(()=>{}),d6=0;function ln(e){const[t,n]=g.useState(f6());return cn(()=>{n(r=>r??String(d6++))},[e]),t?`radix-${t}`:""}function JI(e){const t=e+"CollectionProvider",[n,r]=jn(t),[i,a]=n(t,{collectionRef:{current:null},itemMap:new Map}),s=w=>{const{scope:x,children:C}=w,O=V.useRef(null),A=V.useRef(new Map).current;return E.jsx(i,{scope:x,itemMap:A,collectionRef:O,children:C})};s.displayName=t;const u=e+"CollectionSlot",c=Ah(u),d=V.forwardRef((w,x)=>{const{scope:C,children:O}=w,A=a(u,C),T=Re(x,A.collectionRef);return E.jsx(c,{ref:T,children:O})});d.displayName=u;const p=e+"CollectionItemSlot",m="data-radix-collection-item",v=Ah(p),b=V.forwardRef((w,x)=>{const{scope:C,children:O,...A}=w,T=V.useRef(null),R=Re(x,T),M=a(p,C);return V.useEffect(()=>(M.itemMap.set(T,{ref:T,...A}),()=>void M.itemMap.delete(T))),E.jsx(v,{[m]:"",ref:R,children:O})});b.displayName=p;function S(w){const x=a(e+"CollectionConsumer",w);return V.useCallback(()=>{const O=x.collectionRef.current;if(!O)return[];const A=Array.from(O.querySelectorAll(`[${m}]`));return Array.from(x.itemMap.values()).sort((M,P)=>A.indexOf(M.ref.current)-A.indexOf(P.ref.current))},[x.collectionRef,x.itemMap])}return[{Provider:s,Slot:d,ItemSlot:b},S,r]}function $t(e){const t=g.useRef(e);return g.useEffect(()=>{t.current=e}),g.useMemo(()=>(...n)=>{var r;return(r=t.current)==null?void 0:r.call(t,...n)},[])}function rd(e,t=globalThis==null?void 0:globalThis.document){const n=$t(e);g.useEffect(()=>{const r=i=>{i.key==="Escape"&&n(i)};return t.addEventListener("keydown",r,{capture:!0}),()=>t.removeEventListener("keydown",r,{capture:!0})},[n,t])}var p6="DismissableLayer",p_="dismissableLayer.update",h6="dismissableLayer.pointerDownOutside",m6="dismissableLayer.focusOutside",BM,eL=g.createContext({layers:new Set,layersWithOutsidePointerEventsDisabled:new Set,branches:new Set}),tL=g.forwardRef((e,t)=>{const{disableOutsidePointerEvents:n=!1,onEscapeKeyDown:r,onPointerDownOutside:i,onFocusOutside:a,onInteractOutside:s,onDismiss:u,...c}=e,d=g.useContext(eL),[p,m]=g.useState(null),v=(p==null?void 0:p.ownerDocument)??(globalThis==null?void 0:globalThis.document),[,b]=g.useState({}),S=Re(t,P=>m(P)),w=Array.from(d.layers),[x]=[...d.layersWithOutsidePointerEventsDisabled].slice(-1),C=w.indexOf(x),O=p?w.indexOf(p):-1,A=d.layersWithOutsidePointerEventsDisabled.size>0,T=O>=C,R=y6(P=>{const N=P.target,k=[...d.branches].some(H=>H.contains(N));!T||k||(i==null||i(P),s==null||s(P),P.defaultPrevented||u==null||u())},v),M=b6(P=>{const N=P.target;[...d.branches].some(H=>H.contains(N))||(a==null||a(P),s==null||s(P),P.defaultPrevented||u==null||u())},v);return rd(P=>{O===d.layers.size-1&&(r==null||r(P),!P.defaultPrevented&&u&&(P.preventDefault(),u()))},v),g.useEffect(()=>{if(p)return n&&(d.layersWithOutsidePointerEventsDisabled.size===0&&(BM=v.body.style.pointerEvents,v.body.style.pointerEvents="none"),d.layersWithOutsidePointerEventsDisabled.add(p)),d.layers.add(p),FM(),()=>{n&&d.layersWithOutsidePointerEventsDisabled.size===1&&(v.body.style.pointerEvents=BM)}},[p,v,n,d]),g.useEffect(()=>()=>{p&&(d.layers.delete(p),d.layersWithOutsidePointerEventsDisabled.delete(p),FM())},[p,d]),g.useEffect(()=>{const P=()=>b({});return document.addEventListener(p_,P),()=>document.removeEventListener(p_,P)},[]),E.jsx(Ut.div,{...c,ref:S,style:{pointerEvents:A?T?"auto":"none":void 0,...e.style},onFocusCapture:be(e.onFocusCapture,M.onFocusCapture),onBlurCapture:be(e.onBlurCapture,M.onBlurCapture),onPointerDownCapture:be(e.onPointerDownCapture,R.onPointerDownCapture)})});tL.displayName=p6;var v6="DismissableLayerBranch",g6=g.forwardRef((e,t)=>{const n=g.useContext(eL),r=g.useRef(null),i=Re(t,r);return g.useEffect(()=>{const a=r.current;if(a)return n.branches.add(a),()=>{n.branches.delete(a)}},[n.branches]),E.jsx(Ut.div,{...e,ref:i})});g6.displayName=v6;function y6(e,t=globalThis==null?void 0:globalThis.document){const n=$t(e),r=g.useRef(!1),i=g.useRef(()=>{});return g.useEffect(()=>{const a=u=>{if(u.target&&!r.current){let c=function(){nL(h6,n,d,{discrete:!0})};const d={originalEvent:u};u.pointerType==="touch"?(t.removeEventListener("click",i.current),i.current=c,t.addEventListener("click",i.current,{once:!0})):c()}else t.removeEventListener("click",i.current);r.current=!1},s=window.setTimeout(()=>{t.addEventListener("pointerdown",a)},0);return()=>{window.clearTimeout(s),t.removeEventListener("pointerdown",a),t.removeEventListener("click",i.current)}},[t,n]),{onPointerDownCapture:()=>r.current=!0}}function b6(e,t=globalThis==null?void 0:globalThis.document){const n=$t(e),r=g.useRef(!1);return g.useEffect(()=>{const i=a=>{a.target&&!r.current&&nL(m6,n,{originalEvent:a},{discrete:!1})};return t.addEventListener("focusin",i),()=>t.removeEventListener("focusin",i)},[t,n]),{onFocusCapture:()=>r.current=!0,onBlurCapture:()=>r.current=!1}}function FM(){const e=new CustomEvent(p_);document.dispatchEvent(e)}function nL(e,t,n,{discrete:r}){const i=n.originalEvent.target,a=new CustomEvent(e,{bubbles:!1,cancelable:!0,detail:n});t&&i.addEventListener(e,t,{once:!0}),r?d_(i,a):i.dispatchEvent(a)}function Ym(e){const t=g.useRef({value:e,previous:e});return g.useMemo(()=>(t.current.value!==e&&(t.current.previous=t.current.value,t.current.value=e),t.current.previous),[e])}var x6=Object.freeze({position:"absolute",border:0,width:1,height:1,padding:0,margin:-1,overflow:"hidden",clip:"rect(0, 0, 0, 0)",whiteSpace:"nowrap",wordWrap:"normal"}),w6="VisuallyHidden",rL=g.forwardRef((e,t)=>E.jsx(Ut.span,{...e,ref:t,style:{...x6,...e.style}}));rL.displayName=w6;var S6=rL,hs="NavigationMenu",[OE,iL,_6]=JI(hs),[h_,E6,C6]=JI(hs),[AE,W0e]=jn(hs,[_6,C6]),[O6,$r]=AE(hs),[A6,R6]=AE(hs),aL=g.forwardRef((e,t)=>{const{__scopeNavigationMenu:n,value:r,onValueChange:i,defaultValue:a,delayDuration:s=200,skipDelayDuration:u=300,orientation:c="horizontal",dir:d,...p}=e,[m,v]=g.useState(null),b=Re(t,k=>v(k)),S=iu(d),w=g.useRef(0),x=g.useRef(0),C=g.useRef(0),[O,A]=g.useState(!0),[T,R]=Xr({prop:r,onChange:k=>{const H=k!=="",L=u>0;H?(window.clearTimeout(C.current),L&&A(!1)):(window.clearTimeout(C.current),C.current=window.setTimeout(()=>A(!0),u)),i==null||i(k)},defaultProp:a??"",caller:hs}),M=g.useCallback(()=>{window.clearTimeout(x.current),x.current=window.setTimeout(()=>R(""),150)},[R]),P=g.useCallback(k=>{window.clearTimeout(x.current),R(k)},[R]),N=g.useCallback(k=>{T===k?window.clearTimeout(x.current):w.current=window.setTimeout(()=>{window.clearTimeout(x.current),R(k)},s)},[T,R,s]);return g.useEffect(()=>()=>{window.clearTimeout(w.current),window.clearTimeout(x.current),window.clearTimeout(C.current)},[]),E.jsx(oL,{scope:n,isRootMenu:!0,value:T,dir:S,orientation:c,rootNavigationMenu:m,onTriggerEnter:k=>{window.clearTimeout(w.current),O?N(k):P(k)},onTriggerLeave:()=>{window.clearTimeout(w.current),M()},onContentEnter:()=>window.clearTimeout(x.current),onContentLeave:M,onItemSelect:k=>{R(H=>H===k?"":k)},onItemDismiss:()=>R(""),children:E.jsx(Ut.nav,{"aria-label":"Main","data-orientation":c,dir:S,...p,ref:b})})});aL.displayName=hs;var m_="NavigationMenuSub",T6=g.forwardRef((e,t)=>{const{__scopeNavigationMenu:n,value:r,onValueChange:i,defaultValue:a,orientation:s="horizontal",...u}=e,c=$r(m_,n),[d,p]=Xr({prop:r,onChange:i,defaultProp:a??"",caller:m_});return E.jsx(oL,{scope:n,isRootMenu:!1,value:d,dir:c.dir,orientation:s,rootNavigationMenu:c.rootNavigationMenu,onTriggerEnter:m=>p(m),onItemSelect:m=>p(m),onItemDismiss:()=>p(""),children:E.jsx(Ut.div,{"data-orientation":s,...u,ref:t})})});T6.displayName=m_;var oL=e=>{const{scope:t,isRootMenu:n,rootNavigationMenu:r,dir:i,orientation:a,children:s,value:u,onItemSelect:c,onItemDismiss:d,onTriggerEnter:p,onTriggerLeave:m,onContentEnter:v,onContentLeave:b}=e,[S,w]=g.useState(null),[x,C]=g.useState(new Map),[O,A]=g.useState(null);return E.jsx(O6,{scope:t,isRootMenu:n,rootNavigationMenu:r,value:u,previousValue:Ym(u),baseId:ln(),dir:i,orientation:a,viewport:S,onViewportChange:w,indicatorTrack:O,onIndicatorTrackChange:A,onTriggerEnter:$t(p),onTriggerLeave:$t(m),onContentEnter:$t(v),onContentLeave:$t(b),onItemSelect:$t(c),onItemDismiss:$t(d),onViewportContentChange:g.useCallback((T,R)=>{C(M=>(M.set(T,R),new Map(M)))},[]),onViewportContentRemove:g.useCallback(T=>{C(R=>R.has(T)?(R.delete(T),new Map(R)):R)},[]),children:E.jsx(OE.Provider,{scope:t,children:E.jsx(A6,{scope:t,items:x,children:s})})})},sL="NavigationMenuList",lL=g.forwardRef((e,t)=>{const{__scopeNavigationMenu:n,...r}=e,i=$r(sL,n),a=E.jsx(Ut.ul,{"data-orientation":i.orientation,...r,ref:t});return E.jsx(Ut.div,{style:{position:"relative"},ref:i.onIndicatorTrackChange,children:E.jsx(OE.Slot,{scope:n,children:i.isRootMenu?E.jsx(hL,{asChild:!0,children:a}):a})})});lL.displayName=sL;var uL="NavigationMenuItem",[M6,cL]=AE(uL),fL=g.forwardRef((e,t)=>{const{__scopeNavigationMenu:n,value:r,...i}=e,a=ln(),s=r||a||"LEGACY_REACT_AUTO_VALUE",u=g.useRef(null),c=g.useRef(null),d=g.useRef(null),p=g.useRef(()=>{}),m=g.useRef(!1),v=g.useCallback((S="start")=>{if(u.current){p.current();const w=g_(u.current);w.length&&ME(S==="start"?w:w.reverse())}},[]),b=g.useCallback(()=>{if(u.current){const S=g_(u.current);S.length&&(p.current=q6(S))}},[]);return E.jsx(M6,{scope:n,value:s,triggerRef:c,contentRef:u,focusProxyRef:d,wasEscapeCloseRef:m,onEntryKeyDown:v,onFocusProxyEnter:v,onRootContentClose:b,onContentFocusOutside:b,children:E.jsx(Ut.li,{...i,ref:t})})});fL.displayName=uL;var v_="NavigationMenuTrigger",P6=g.forwardRef((e,t)=>{const{__scopeNavigationMenu:n,disabled:r,...i}=e,a=$r(v_,e.__scopeNavigationMenu),s=cL(v_,e.__scopeNavigationMenu),u=g.useRef(null),c=Re(u,s.triggerRef,t),d=vL(a.baseId,s.value),p=gL(a.baseId,s.value),m=g.useRef(!1),v=g.useRef(!1),b=s.value===a.value;return E.jsxs(E.Fragment,{children:[E.jsx(OE.ItemSlot,{scope:n,value:s.value,children:E.jsx(mL,{asChild:!0,children:E.jsx(Ut.button,{id:d,disabled:r,"data-disabled":r?"":void 0,"data-state":PE(b),"aria-expanded":b,"aria-controls":p,...i,ref:c,onPointerEnter:be(e.onPointerEnter,()=>{v.current=!1,s.wasEscapeCloseRef.current=!1}),onPointerMove:be(e.onPointerMove,Rh(()=>{r||v.current||s.wasEscapeCloseRef.current||m.current||(a.onTriggerEnter(s.value),m.current=!0)})),onPointerLeave:be(e.onPointerLeave,Rh(()=>{r||(a.onTriggerLeave(),m.current=!1)})),onClick:be(e.onClick,()=>{a.onItemSelect(s.value),v.current=b}),onKeyDown:be(e.onKeyDown,S=>{const x={horizontal:"ArrowDown",vertical:a.dir==="rtl"?"ArrowLeft":"ArrowRight"}[a.orientation];b&&S.key===x&&(s.onEntryKeyDown(),S.preventDefault())})})})}),b&&E.jsxs(E.Fragment,{children:[E.jsx(S6,{"aria-hidden":!0,tabIndex:0,ref:s.focusProxyRef,onFocus:S=>{const w=s.contentRef.current,x=S.relatedTarget,C=x===u.current,O=w==null?void 0:w.contains(x);(C||!O)&&s.onFocusProxyEnter(C?"start":"end")}}),a.viewport&&E.jsx("span",{"aria-owns":p})]})]})});P6.displayName=v_;var j6="NavigationMenuLink",qM="navigationMenu.linkSelect",N6=g.forwardRef((e,t)=>{const{__scopeNavigationMenu:n,active:r,onSelect:i,...a}=e;return E.jsx(mL,{asChild:!0,children:E.jsx(Ut.a,{"data-active":r?"":void 0,"aria-current":r?"page":void 0,...a,ref:t,onClick:be(e.onClick,s=>{const u=s.target,c=new CustomEvent(qM,{bubbles:!0,cancelable:!0});if(u.addEventListener(qM,d=>i==null?void 0:i(d),{once:!0}),d_(u,c),!c.defaultPrevented&&!s.metaKey){const d=new CustomEvent(gh,{bubbles:!0,cancelable:!0});d_(u,d)}},{checkForDefaultPrevented:!1})})})});N6.displayName=j6;var RE="NavigationMenuIndicator",D6=g.forwardRef((e,t)=>{const{forceMount:n,...r}=e,i=$r(RE,e.__scopeNavigationMenu),a=!!i.value;return i.indicatorTrack?QI.createPortal(E.jsx(zn,{present:n||a,children:E.jsx(k6,{...r,ref:t})}),i.indicatorTrack):null});D6.displayName=RE;var k6=g.forwardRef((e,t)=>{const{__scopeNavigationMenu:n,...r}=e,i=$r(RE,n),a=iL(n),[s,u]=g.useState(null),[c,d]=g.useState(null),p=i.orientation==="horizontal",m=!!i.value;g.useEffect(()=>{var w;const S=(w=a().find(x=>x.value===i.value))==null?void 0:w.ref.current;S&&u(S)},[a,i.value]);const v=()=>{s&&d({size:p?s.offsetWidth:s.offsetHeight,offset:p?s.offsetLeft:s.offsetTop})};return y_(s,v),y_(i.indicatorTrack,v),c?E.jsx(Ut.div,{"aria-hidden":!0,"data-state":m?"visible":"hidden","data-orientation":i.orientation,...r,ref:t,style:{position:"absolute",...p?{left:0,width:c.size+"px",transform:`translateX(${c.offset}px)`}:{top:0,height:c.size+"px",transform:`translateY(${c.offset}px)`},...r.style}}):null}),Ml="NavigationMenuContent",$6=g.forwardRef((e,t)=>{const{forceMount:n,...r}=e,i=$r(Ml,e.__scopeNavigationMenu),a=cL(Ml,e.__scopeNavigationMenu),s=Re(a.contentRef,t),u=a.value===i.value,c={value:a.value,triggerRef:a.triggerRef,focusProxyRef:a.focusProxyRef,wasEscapeCloseRef:a.wasEscapeCloseRef,onContentFocusOutside:a.onContentFocusOutside,onRootContentClose:a.onRootContentClose,...r};return i.viewport?E.jsx(I6,{forceMount:n,...c,ref:s}):E.jsx(zn,{present:n||u,children:E.jsx(dL,{"data-state":PE(u),...c,ref:s,onPointerEnter:be(e.onPointerEnter,i.onContentEnter),onPointerLeave:be(e.onPointerLeave,Rh(i.onContentLeave)),style:{pointerEvents:!u&&i.isRootMenu?"none":void 0,...c.style}})})});$6.displayName=Ml;var I6=g.forwardRef((e,t)=>{const n=$r(Ml,e.__scopeNavigationMenu),{onViewportContentChange:r,onViewportContentRemove:i}=n;return cn(()=>{r(e.value,{ref:t,...e})},[e,t,r]),cn(()=>()=>i(e.value),[e.value,i]),null}),gh="navigationMenu.rootContentDismiss",dL=g.forwardRef((e,t)=>{const{__scopeNavigationMenu:n,value:r,triggerRef:i,focusProxyRef:a,wasEscapeCloseRef:s,onRootContentClose:u,onContentFocusOutside:c,...d}=e,p=$r(Ml,n),m=g.useRef(null),v=Re(m,t),b=vL(p.baseId,r),S=gL(p.baseId,r),w=iL(n),x=g.useRef(null),{onItemDismiss:C}=p;g.useEffect(()=>{const A=m.current;if(p.isRootMenu&&A){const T=()=>{var R;C(),u(),A.contains(document.activeElement)&&((R=i.current)==null||R.focus())};return A.addEventListener(gh,T),()=>A.removeEventListener(gh,T)}},[p.isRootMenu,e.value,i,C,u]);const O=g.useMemo(()=>{const T=w().map(H=>H.value);p.dir==="rtl"&&T.reverse();const R=T.indexOf(p.value),M=T.indexOf(p.previousValue),P=r===p.value,N=M===T.indexOf(r);if(!P&&!N)return x.current;const k=(()=>{if(R!==M){if(P&&M!==-1)return R>M?"from-end":"from-start";if(N&&R!==-1)return R>M?"to-start":"to-end"}return null})();return x.current=k,k},[p.previousValue,p.value,p.dir,w,r]);return E.jsx(hL,{asChild:!0,children:E.jsx(tL,{id:S,"aria-labelledby":b,"data-motion":O,"data-orientation":p.orientation,...d,ref:v,disableOutsidePointerEvents:!1,onDismiss:()=>{var T;const A=new Event(gh,{bubbles:!0,cancelable:!0});(T=m.current)==null||T.dispatchEvent(A)},onFocusOutside:be(e.onFocusOutside,A=>{var R;c();const T=A.target;(R=p.rootNavigationMenu)!=null&&R.contains(T)&&A.preventDefault()}),onPointerDownOutside:be(e.onPointerDownOutside,A=>{var P;const T=A.target,R=w().some(N=>{var k;return(k=N.ref.current)==null?void 0:k.contains(T)}),M=p.isRootMenu&&((P=p.viewport)==null?void 0:P.contains(T));(R||M||!p.isRootMenu)&&A.preventDefault()}),onKeyDown:be(e.onKeyDown,A=>{var M;const T=A.altKey||A.ctrlKey||A.metaKey;if(A.key==="Tab"&&!T){const P=g_(A.currentTarget),N=document.activeElement,k=P.findIndex(F=>F===N),L=A.shiftKey?P.slice(0,k).reverse():P.slice(k+1,P.length);ME(L)?A.preventDefault():(M=a.current)==null||M.focus()}}),onEscapeKeyDown:be(e.onEscapeKeyDown,A=>{s.current=!0})})})}),TE="NavigationMenuViewport",pL=g.forwardRef((e,t)=>{const{forceMount:n,...r}=e,a=!!$r(TE,e.__scopeNavigationMenu).value;return E.jsx(zn,{present:n||a,children:E.jsx(L6,{...r,ref:t})})});pL.displayName=TE;var L6=g.forwardRef((e,t)=>{const{__scopeNavigationMenu:n,children:r,...i}=e,a=$r(TE,n),s=Re(t,a.onViewportChange),u=R6(Ml,e.__scopeNavigationMenu),[c,d]=g.useState(null),[p,m]=g.useState(null),v=c?(c==null?void 0:c.width)+"px":void 0,b=c?(c==null?void 0:c.height)+"px":void 0,S=!!a.value,w=S?a.value:a.previousValue;return y_(p,()=>{p&&d({width:p.offsetWidth,height:p.offsetHeight})}),E.jsx(Ut.div,{"data-state":PE(S),"data-orientation":a.orientation,...i,ref:s,style:{pointerEvents:!S&&a.isRootMenu?"none":void 0,"--radix-navigation-menu-viewport-width":v,"--radix-navigation-menu-viewport-height":b,...i.style},onPointerEnter:be(e.onPointerEnter,a.onContentEnter),onPointerLeave:be(e.onPointerLeave,Rh(a.onContentLeave)),children:Array.from(u.items).map(([C,{ref:O,forceMount:A,...T}])=>{const R=w===C;return E.jsx(zn,{present:A||R,children:E.jsx(dL,{...T,ref:wi(O,M=>{R&&M&&m(M)})})},C)})})}),B6="FocusGroup",hL=g.forwardRef((e,t)=>{const{__scopeNavigationMenu:n,...r}=e,i=$r(B6,n);return E.jsx(h_.Provider,{scope:n,children:E.jsx(h_.Slot,{scope:n,children:E.jsx(Ut.div,{dir:i.dir,...r,ref:t})})})}),zM=["ArrowRight","ArrowLeft","ArrowUp","ArrowDown"],F6="FocusGroupItem",mL=g.forwardRef((e,t)=>{const{__scopeNavigationMenu:n,...r}=e,i=E6(n),a=$r(F6,n);return E.jsx(h_.ItemSlot,{scope:n,children:E.jsx(Ut.button,{...r,ref:t,onKeyDown:be(e.onKeyDown,s=>{if(["Home","End",...zM].includes(s.key)){let c=i().map(m=>m.ref.current);if([a.dir==="rtl"?"ArrowRight":"ArrowLeft","ArrowUp","End"].includes(s.key)&&c.reverse(),zM.includes(s.key)){const m=c.indexOf(s.currentTarget);c=c.slice(m+1)}setTimeout(()=>ME(c)),s.preventDefault()}})})})});function g_(e){const t=[],n=document.createTreeWalker(e,NodeFilter.SHOW_ELEMENT,{acceptNode:r=>{const i=r.tagName==="INPUT"&&r.type==="hidden";return r.disabled||r.hidden||i?NodeFilter.FILTER_SKIP:r.tabIndex>=0?NodeFilter.FILTER_ACCEPT:NodeFilter.FILTER_SKIP}});for(;n.nextNode();)t.push(n.currentNode);return t}function ME(e){const t=document.activeElement;return e.some(n=>n===t?!0:(n.focus(),document.activeElement!==t))}function q6(e){return e.forEach(t=>{t.dataset.tabindex=t.getAttribute("tabindex")||"",t.setAttribute("tabindex","-1")}),()=>{e.forEach(t=>{const n=t.dataset.tabindex;t.setAttribute("tabindex",n)})}}function y_(e,t){const n=$t(t);cn(()=>{let r=0;if(e){const i=new ResizeObserver(()=>{cancelAnimationFrame(r),r=window.requestAnimationFrame(n)});return i.observe(e),()=>{window.cancelAnimationFrame(r),i.unobserve(e)}}},[e,n])}function PE(e){return e?"open":"closed"}function vL(e,t){return`${e}-trigger-${t}`}function gL(e,t){return`${e}-content-${t}`}function Rh(e){return t=>t.pointerType==="mouse"?e(t):void 0}var z6=aL,U6=lL,V6=fL,H6=pL;function yL(e){var t,n,r="";if(typeof e=="string"||typeof e=="number")r+=e;else if(typeof e=="object")if(Array.isArray(e)){var i=e.length;for(t=0;t<i;t++)e[t]&&(n=yL(e[t]))&&(r&&(r+=" "),r+=n)}else for(n in e)e[n]&&(r&&(r+=" "),r+=n);return r}function ct(){for(var e,t,n=0,r="",i=arguments.length;n<i;n++)(e=arguments[n])&&(t=yL(e))&&(r&&(r+=" "),r+=t);return r}const UM=e=>typeof e=="boolean"?`${e}`:e===0?"0":e,VM=ct,G6=(e,t)=>n=>{var r;if((t==null?void 0:t.variants)==null)return VM(e,n==null?void 0:n.class,n==null?void 0:n.className);const{variants:i,defaultVariants:a}=t,s=Object.keys(i).map(d=>{const p=n==null?void 0:n[d],m=a==null?void 0:a[d];if(p===null)return null;const v=UM(p)||UM(m);return i[d][v]}),u=n&&Object.entries(n).reduce((d,p)=>{let[m,v]=p;return v===void 0||(d[m]=v),d},{}),c=t==null||(r=t.compoundVariants)===null||r===void 0?void 0:r.reduce((d,p)=>{let{class:m,className:v,...b}=p;return Object.entries(b).every(S=>{let[w,x]=S;return Array.isArray(x)?x.includes({...a,...u}[w]):{...a,...u}[w]===x})?[...d,m,v]:d},[]);return VM(e,s,c,n==null?void 0:n.class,n==null?void 0:n.className)};/**
 * @license lucide-react v0.503.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const W6=e=>e.replace(/([a-z0-9])([A-Z])/g,"$1-$2").toLowerCase(),K6=e=>e.replace(/^([A-Z])|[\s-_]+(\w)/g,(t,n,r)=>r?r.toUpperCase():n.toLowerCase()),HM=e=>{const t=K6(e);return t.charAt(0).toUpperCase()+t.slice(1)},bL=(...e)=>e.filter((t,n,r)=>!!t&&t.trim()!==""&&r.indexOf(t)===n).join(" ").trim(),Y6=e=>{for(const t in e)if(t.startsWith("aria-")||t==="role"||t==="title")return!0};/**
 * @license lucide-react v0.503.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */var Z6={xmlns:"http://www.w3.org/2000/svg",width:24,height:24,viewBox:"0 0 24 24",fill:"none",stroke:"currentColor",strokeWidth:2,strokeLinecap:"round",strokeLinejoin:"round"};/**
 * @license lucide-react v0.503.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const X6=g.forwardRef(({color:e="currentColor",size:t=24,strokeWidth:n=2,absoluteStrokeWidth:r,className:i="",children:a,iconNode:s,...u},c)=>g.createElement("svg",{ref:c,...Z6,width:t,height:t,stroke:e,strokeWidth:r?Number(n)*24/Number(t):n,className:bL("lucide",i),...!a&&!Y6(u)&&{"aria-hidden":"true"},...u},[...s.map(([d,p])=>g.createElement(d,p)),...Array.isArray(a)?a:[a]]));/**
 * @license lucide-react v0.503.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const Nn=(e,t)=>{const n=g.forwardRef(({className:r,...i},a)=>g.createElement(X6,{ref:a,iconNode:t,className:bL(`lucide-${W6(HM(e))}`,`lucide-${e}`,r),...i}));return n.displayName=HM(e),n};/**
 * @license lucide-react v0.503.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const Q6=[["path",{d:"M5 12h14",key:"1ays0h"}],["path",{d:"m12 5 7 7-7 7",key:"xquz4c"}]],J6=Nn("arrow-right",Q6);/**
 * @license lucide-react v0.503.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const e8=[["path",{d:"M20 6 9 17l-5-5",key:"1gmf2c"}]],jE=Nn("check",e8);/**
 * @license lucide-react v0.503.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const t8=[["path",{d:"m6 9 6 6 6-6",key:"qrunsl"}]],Zm=Nn("chevron-down",t8);/**
 * @license lucide-react v0.503.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const n8=[["path",{d:"m18 15-6-6-6 6",key:"153udz"}]],r8=Nn("chevron-up",n8);/**
 * @license lucide-react v0.503.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const i8=[["path",{d:"m7 15 5 5 5-5",key:"1hf1tw"}],["path",{d:"m7 9 5-5 5 5",key:"sgt6xg"}]],a8=Nn("chevrons-up-down",i8);/**
 * @license lucide-react v0.503.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const o8=[["path",{d:"M10 12v-1",key:"v7bkov"}],["path",{d:"M10 18v-2",key:"1cjy8d"}],["path",{d:"M10 7V6",key:"dljcrl"}],["path",{d:"M14 2v4a2 2 0 0 0 2 2h4",key:"tnqrlb"}],["path",{d:"M15.5 22H18a2 2 0 0 0 2-2V7l-5-5H6a2 2 0 0 0-2 2v16a2 2 0 0 0 .274 1.01",key:"gkbcor"}],["circle",{cx:"10",cy:"20",r:"2",key:"1xzdoj"}]],s8=Nn("file-archive",o8);/**
 * @license lucide-react v0.503.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const l8=[["path",{d:"M17.5 22h.5a2 2 0 0 0 2-2V7l-5-5H6a2 2 0 0 0-2 2v3",key:"rslqgf"}],["path",{d:"M14 2v4a2 2 0 0 0 2 2h4",key:"tnqrlb"}],["path",{d:"M2 19a2 2 0 1 1 4 0v1a2 2 0 1 1-4 0v-4a6 6 0 0 1 12 0v4a2 2 0 1 1-4 0v-1a2 2 0 1 1 4 0",key:"9f7x3i"}]],u8=Nn("file-audio",l8);/**
 * @license lucide-react v0.503.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const c8=[["path",{d:"M10 12.5 8 15l2 2.5",key:"1tg20x"}],["path",{d:"m14 12.5 2 2.5-2 2.5",key:"yinavb"}],["path",{d:"M14 2v4a2 2 0 0 0 2 2h4",key:"tnqrlb"}],["path",{d:"M15 2H6a2 2 0 0 0-2 2v16a2 2 0 0 0 2 2h12a2 2 0 0 0 2-2V7z",key:"1mlx9k"}]],f8=Nn("file-code",c8);/**
 * @license lucide-react v0.503.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const d8=[["path",{d:"M14 2v4a2 2 0 0 0 2 2h4",key:"tnqrlb"}],["path",{d:"m2.305 15.53.923-.382",key:"yfp9st"}],["path",{d:"m3.228 12.852-.924-.383",key:"bckynb"}],["path",{d:"M4.677 21.5a2 2 0 0 0 1.313.5H18a2 2 0 0 0 2-2V7l-5-5H6a2 2 0 0 0-2 2v2.5",key:"1yo3oz"}],["path",{d:"m4.852 11.228-.383-.923",key:"1j88i9"}],["path",{d:"m4.852 16.772-.383.924",key:"sag1dv"}],["path",{d:"m7.148 11.228.383-.923",key:"rj39hk"}],["path",{d:"m7.53 17.696-.382-.924",key:"1uu5cs"}],["path",{d:"m8.772 12.852.923-.383",key:"13811l"}],["path",{d:"m8.772 15.148.923.383",key:"z1a5l0"}],["circle",{cx:"6",cy:"14",r:"3",key:"a1xfv6"}]],p8=Nn("file-cog",d8);/**
 * @license lucide-react v0.503.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const h8=[["path",{d:"M15 2H6a2 2 0 0 0-2 2v16a2 2 0 0 0 2 2h12a2 2 0 0 0 2-2V7Z",key:"1rqfz7"}],["path",{d:"M14 2v4a2 2 0 0 0 2 2h4",key:"tnqrlb"}],["path",{d:"M10 9H8",key:"b1mrlr"}],["path",{d:"M16 13H8",key:"t4e002"}],["path",{d:"M16 17H8",key:"z1uh3a"}]],m8=Nn("file-text",h8);/**
 * @license lucide-react v0.503.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const v8=[["path",{d:"M15 2H6a2 2 0 0 0-2 2v16a2 2 0 0 0 2 2h12a2 2 0 0 0 2-2V7Z",key:"1rqfz7"}],["path",{d:"M14 2v4a2 2 0 0 0 2 2h4",key:"tnqrlb"}],["path",{d:"m10 11 5 3-5 3v-6Z",key:"7ntvm4"}]],g8=Nn("file-video",v8);/**
 * @license lucide-react v0.503.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const y8=[["path",{d:"M15 2H6a2 2 0 0 0-2 2v16a2 2 0 0 0 2 2h12a2 2 0 0 0 2-2V7Z",key:"1rqfz7"}],["path",{d:"M14 2v4a2 2 0 0 0 2 2h4",key:"tnqrlb"}]],b8=Nn("file",y8);/**
 * @license lucide-react v0.503.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const x8=[["circle",{cx:"11",cy:"11",r:"8",key:"4ej97u"}],["path",{d:"m21 21-4.3-4.3",key:"1qie3q"}]],w8=Nn("search",x8);/**
 * @license lucide-react v0.503.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const S8=[["path",{d:"m21.73 18-8-14a2 2 0 0 0-3.48 0l-8 14A2 2 0 0 0 4 21h16a2 2 0 0 0 1.73-3",key:"wmoenq"}],["path",{d:"M12 9v4",key:"juzpu7"}],["path",{d:"M12 17h.01",key:"p32p05"}]],_8=Nn("triangle-alert",S8);/**
 * @license lucide-react v0.503.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const E8=[["path",{d:"M21 15v4a2 2 0 0 1-2 2H5a2 2 0 0 1-2-2v-4",key:"ih7n3h"}],["polyline",{points:"17 8 12 3 7 8",key:"t8dd8p"}],["line",{x1:"12",x2:"12",y1:"3",y2:"15",key:"widbto"}]],xL=Nn("upload",E8);/**
 * @license lucide-react v0.503.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const C8=[["path",{d:"M18 6 6 18",key:"1bl5f8"}],["path",{d:"m6 6 12 12",key:"d8bk6v"}]],wL=Nn("x",C8),NE="-",O8=e=>{const t=R8(e),{conflictingClassGroups:n,conflictingClassGroupModifiers:r}=e;return{getClassGroupId:s=>{const u=s.split(NE);return u[0]===""&&u.length!==1&&u.shift(),SL(u,t)||A8(s)},getConflictingClassGroupIds:(s,u)=>{const c=n[s]||[];return u&&r[s]?[...c,...r[s]]:c}}},SL=(e,t)=>{var s;if(e.length===0)return t.classGroupId;const n=e[0],r=t.nextPart.get(n),i=r?SL(e.slice(1),r):void 0;if(i)return i;if(t.validators.length===0)return;const a=e.join(NE);return(s=t.validators.find(({validator:u})=>u(a)))==null?void 0:s.classGroupId},GM=/^\[(.+)\]$/,A8=e=>{if(GM.test(e)){const t=GM.exec(e)[1],n=t==null?void 0:t.substring(0,t.indexOf(":"));if(n)return"arbitrary.."+n}},R8=e=>{const{theme:t,classGroups:n}=e,r={nextPart:new Map,validators:[]};for(const i in n)b_(n[i],r,i,t);return r},b_=(e,t,n,r)=>{e.forEach(i=>{if(typeof i=="string"){const a=i===""?t:WM(t,i);a.classGroupId=n;return}if(typeof i=="function"){if(T8(i)){b_(i(r),t,n,r);return}t.validators.push({validator:i,classGroupId:n});return}Object.entries(i).forEach(([a,s])=>{b_(s,WM(t,a),n,r)})})},WM=(e,t)=>{let n=e;return t.split(NE).forEach(r=>{n.nextPart.has(r)||n.nextPart.set(r,{nextPart:new Map,validators:[]}),n=n.nextPart.get(r)}),n},T8=e=>e.isThemeGetter,M8=e=>{if(e<1)return{get:()=>{},set:()=>{}};let t=0,n=new Map,r=new Map;const i=(a,s)=>{n.set(a,s),t++,t>e&&(t=0,r=n,n=new Map)};return{get(a){let s=n.get(a);if(s!==void 0)return s;if((s=r.get(a))!==void 0)return i(a,s),s},set(a,s){n.has(a)?n.set(a,s):i(a,s)}}},x_="!",w_=":",P8=w_.length,j8=e=>{const{prefix:t,experimentalParseClassName:n}=e;let r=i=>{const a=[];let s=0,u=0,c=0,d;for(let S=0;S<i.length;S++){let w=i[S];if(s===0&&u===0){if(w===w_){a.push(i.slice(c,S)),c=S+P8;continue}if(w==="/"){d=S;continue}}w==="["?s++:w==="]"?s--:w==="("?u++:w===")"&&u--}const p=a.length===0?i:i.substring(c),m=N8(p),v=m!==p,b=d&&d>c?d-c:void 0;return{modifiers:a,hasImportantModifier:v,baseClassName:m,maybePostfixModifierPosition:b}};if(t){const i=t+w_,a=r;r=s=>s.startsWith(i)?a(s.substring(i.length)):{isExternal:!0,modifiers:[],hasImportantModifier:!1,baseClassName:s,maybePostfixModifierPosition:void 0}}if(n){const i=r;r=a=>n({className:a,parseClassName:i})}return r},N8=e=>e.endsWith(x_)?e.substring(0,e.length-1):e.startsWith(x_)?e.substring(1):e,D8=e=>{const t=Object.fromEntries(e.orderSensitiveModifiers.map(r=>[r,!0]));return r=>{if(r.length<=1)return r;const i=[];let a=[];return r.forEach(s=>{s[0]==="["||t[s]?(i.push(...a.sort(),s),a=[]):a.push(s)}),i.push(...a.sort()),i}},k8=e=>({cache:M8(e.cacheSize),parseClassName:j8(e),sortModifiers:D8(e),...O8(e)}),$8=/\s+/,I8=(e,t)=>{const{parseClassName:n,getClassGroupId:r,getConflictingClassGroupIds:i,sortModifiers:a}=t,s=[],u=e.trim().split($8);let c="";for(let d=u.length-1;d>=0;d-=1){const p=u[d],{isExternal:m,modifiers:v,hasImportantModifier:b,baseClassName:S,maybePostfixModifierPosition:w}=n(p);if(m){c=p+(c.length>0?" "+c:c);continue}let x=!!w,C=r(x?S.substring(0,w):S);if(!C){if(!x){c=p+(c.length>0?" "+c:c);continue}if(C=r(S),!C){c=p+(c.length>0?" "+c:c);continue}x=!1}const O=a(v).join(":"),A=b?O+x_:O,T=A+C;if(s.includes(T))continue;s.push(T);const R=i(C,x);for(let M=0;M<R.length;++M){const P=R[M];s.push(A+P)}c=p+(c.length>0?" "+c:c)}return c};function L8(){let e=0,t,n,r="";for(;e<arguments.length;)(t=arguments[e++])&&(n=_L(t))&&(r&&(r+=" "),r+=n);return r}const _L=e=>{if(typeof e=="string")return e;let t,n="";for(let r=0;r<e.length;r++)e[r]&&(t=_L(e[r]))&&(n&&(n+=" "),n+=t);return n};function B8(e,...t){let n,r,i,a=s;function s(c){const d=t.reduce((p,m)=>m(p),e());return n=k8(d),r=n.cache.get,i=n.cache.set,a=u,u(c)}function u(c){const d=r(c);if(d)return d;const p=I8(c,n);return i(c,p),p}return function(){return a(L8.apply(null,arguments))}}const rn=e=>{const t=n=>n[e]||[];return t.isThemeGetter=!0,t},EL=/^\[(?:(\w[\w-]*):)?(.+)\]$/i,CL=/^\((?:(\w[\w-]*):)?(.+)\)$/i,F8=/^\d+\/\d+$/,q8=/^(\d+(\.\d+)?)?(xs|sm|md|lg|xl)$/,z8=/\d+(%|px|r?em|[sdl]?v([hwib]|min|max)|pt|pc|in|cm|mm|cap|ch|ex|r?lh|cq(w|h|i|b|min|max))|\b(calc|min|max|clamp)\(.+\)|^0$/,U8=/^(rgba?|hsla?|hwb|(ok)?(lab|lch))\(.+\)$/,V8=/^(inset_)?-?((\d+)?\.?(\d+)[a-z]+|0)_-?((\d+)?\.?(\d+)[a-z]+|0)/,H8=/^(url|image|image-set|cross-fade|element|(repeating-)?(linear|radial|conic)-gradient)\(.+\)$/,ul=e=>F8.test(e),nt=e=>!!e&&!Number.isNaN(Number(e)),Ka=e=>!!e&&Number.isInteger(Number(e)),Db=e=>e.endsWith("%")&&nt(e.slice(0,-1)),Ki=e=>q8.test(e),G8=()=>!0,W8=e=>z8.test(e)&&!U8.test(e),OL=()=>!1,K8=e=>V8.test(e),Y8=e=>H8.test(e),Z8=e=>!Ne(e)&&!De(e),X8=e=>au(e,TL,OL),Ne=e=>EL.test(e),Lo=e=>au(e,ML,W8),kb=e=>au(e,n9,nt),KM=e=>au(e,AL,OL),Q8=e=>au(e,RL,Y8),Hp=e=>au(e,PL,K8),De=e=>CL.test(e),wc=e=>ou(e,ML),J8=e=>ou(e,r9),YM=e=>ou(e,AL),e9=e=>ou(e,TL),t9=e=>ou(e,RL),Gp=e=>ou(e,PL,!0),au=(e,t,n)=>{const r=EL.exec(e);return r?r[1]?t(r[1]):n(r[2]):!1},ou=(e,t,n=!1)=>{const r=CL.exec(e);return r?r[1]?t(r[1]):n:!1},AL=e=>e==="position"||e==="percentage",RL=e=>e==="image"||e==="url",TL=e=>e==="length"||e==="size"||e==="bg-size",ML=e=>e==="length",n9=e=>e==="number",r9=e=>e==="family-name",PL=e=>e==="shadow",i9=()=>{const e=rn("color"),t=rn("font"),n=rn("text"),r=rn("font-weight"),i=rn("tracking"),a=rn("leading"),s=rn("breakpoint"),u=rn("container"),c=rn("spacing"),d=rn("radius"),p=rn("shadow"),m=rn("inset-shadow"),v=rn("text-shadow"),b=rn("drop-shadow"),S=rn("blur"),w=rn("perspective"),x=rn("aspect"),C=rn("ease"),O=rn("animate"),A=()=>["auto","avoid","all","avoid-page","page","left","right","column"],T=()=>["center","top","bottom","left","right","top-left","left-top","top-right","right-top","bottom-right","right-bottom","bottom-left","left-bottom"],R=()=>[...T(),De,Ne],M=()=>["auto","hidden","clip","visible","scroll"],P=()=>["auto","contain","none"],N=()=>[De,Ne,c],k=()=>[ul,"full","auto",...N()],H=()=>[Ka,"none","subgrid",De,Ne],L=()=>["auto",{span:["full",Ka,De,Ne]},Ka,De,Ne],F=()=>[Ka,"auto",De,Ne],z=()=>["auto","min","max","fr",De,Ne],Z=()=>["start","end","center","between","around","evenly","stretch","baseline","center-safe","end-safe"],Y=()=>["start","end","center","stretch","center-safe","end-safe"],I=()=>["auto",...N()],W=()=>[ul,"auto","full","dvw","dvh","lvw","lvh","svw","svh","min","max","fit",...N()],B=()=>[e,De,Ne],Q=()=>[...T(),YM,KM,{position:[De,Ne]}],$=()=>["no-repeat",{repeat:["","x","y","space","round"]}],J=()=>["auto","cover","contain",e9,X8,{size:[De,Ne]}],q=()=>[Db,wc,Lo],X=()=>["","none","full",d,De,Ne],oe=()=>["",nt,wc,Lo],le=()=>["solid","dashed","dotted","double"],ve=()=>["normal","multiply","screen","overlay","darken","lighten","color-dodge","color-burn","hard-light","soft-light","difference","exclusion","hue","saturation","color","luminosity"],de=()=>[nt,Db,YM,KM],Se=()=>["","none",S,De,Ne],ce=()=>["none",nt,De,Ne],we=()=>["none",nt,De,Ne],_e=()=>[nt,De,Ne],se=()=>[ul,"full",...N()];return{cacheSize:500,theme:{animate:["spin","ping","pulse","bounce"],aspect:["video"],blur:[Ki],breakpoint:[Ki],color:[G8],container:[Ki],"drop-shadow":[Ki],ease:["in","out","in-out"],font:[Z8],"font-weight":["thin","extralight","light","normal","medium","semibold","bold","extrabold","black"],"inset-shadow":[Ki],leading:["none","tight","snug","normal","relaxed","loose"],perspective:["dramatic","near","normal","midrange","distant","none"],radius:[Ki],shadow:[Ki],spacing:["px",nt],text:[Ki],"text-shadow":[Ki],tracking:["tighter","tight","normal","wide","wider","widest"]},classGroups:{aspect:[{aspect:["auto","square",ul,Ne,De,x]}],container:["container"],columns:[{columns:[nt,Ne,De,u]}],"break-after":[{"break-after":A()}],"break-before":[{"break-before":A()}],"break-inside":[{"break-inside":["auto","avoid","avoid-page","avoid-column"]}],"box-decoration":[{"box-decoration":["slice","clone"]}],box:[{box:["border","content"]}],display:["block","inline-block","inline","flex","inline-flex","table","inline-table","table-caption","table-cell","table-column","table-column-group","table-footer-group","table-header-group","table-row-group","table-row","flow-root","grid","inline-grid","contents","list-item","hidden"],sr:["sr-only","not-sr-only"],float:[{float:["right","left","none","start","end"]}],clear:[{clear:["left","right","both","none","start","end"]}],isolation:["isolate","isolation-auto"],"object-fit":[{object:["contain","cover","fill","none","scale-down"]}],"object-position":[{object:R()}],overflow:[{overflow:M()}],"overflow-x":[{"overflow-x":M()}],"overflow-y":[{"overflow-y":M()}],overscroll:[{overscroll:P()}],"overscroll-x":[{"overscroll-x":P()}],"overscroll-y":[{"overscroll-y":P()}],position:["static","fixed","absolute","relative","sticky"],inset:[{inset:k()}],"inset-x":[{"inset-x":k()}],"inset-y":[{"inset-y":k()}],start:[{start:k()}],end:[{end:k()}],top:[{top:k()}],right:[{right:k()}],bottom:[{bottom:k()}],left:[{left:k()}],visibility:["visible","invisible","collapse"],z:[{z:[Ka,"auto",De,Ne]}],basis:[{basis:[ul,"full","auto",u,...N()]}],"flex-direction":[{flex:["row","row-reverse","col","col-reverse"]}],"flex-wrap":[{flex:["nowrap","wrap","wrap-reverse"]}],flex:[{flex:[nt,ul,"auto","initial","none",Ne]}],grow:[{grow:["",nt,De,Ne]}],shrink:[{shrink:["",nt,De,Ne]}],order:[{order:[Ka,"first","last","none",De,Ne]}],"grid-cols":[{"grid-cols":H()}],"col-start-end":[{col:L()}],"col-start":[{"col-start":F()}],"col-end":[{"col-end":F()}],"grid-rows":[{"grid-rows":H()}],"row-start-end":[{row:L()}],"row-start":[{"row-start":F()}],"row-end":[{"row-end":F()}],"grid-flow":[{"grid-flow":["row","col","dense","row-dense","col-dense"]}],"auto-cols":[{"auto-cols":z()}],"auto-rows":[{"auto-rows":z()}],gap:[{gap:N()}],"gap-x":[{"gap-x":N()}],"gap-y":[{"gap-y":N()}],"justify-content":[{justify:[...Z(),"normal"]}],"justify-items":[{"justify-items":[...Y(),"normal"]}],"justify-self":[{"justify-self":["auto",...Y()]}],"align-content":[{content:["normal",...Z()]}],"align-items":[{items:[...Y(),{baseline:["","last"]}]}],"align-self":[{self:["auto",...Y(),{baseline:["","last"]}]}],"place-content":[{"place-content":Z()}],"place-items":[{"place-items":[...Y(),"baseline"]}],"place-self":[{"place-self":["auto",...Y()]}],p:[{p:N()}],px:[{px:N()}],py:[{py:N()}],ps:[{ps:N()}],pe:[{pe:N()}],pt:[{pt:N()}],pr:[{pr:N()}],pb:[{pb:N()}],pl:[{pl:N()}],m:[{m:I()}],mx:[{mx:I()}],my:[{my:I()}],ms:[{ms:I()}],me:[{me:I()}],mt:[{mt:I()}],mr:[{mr:I()}],mb:[{mb:I()}],ml:[{ml:I()}],"space-x":[{"space-x":N()}],"space-x-reverse":["space-x-reverse"],"space-y":[{"space-y":N()}],"space-y-reverse":["space-y-reverse"],size:[{size:W()}],w:[{w:[u,"screen",...W()]}],"min-w":[{"min-w":[u,"screen","none",...W()]}],"max-w":[{"max-w":[u,"screen","none","prose",{screen:[s]},...W()]}],h:[{h:["screen",...W()]}],"min-h":[{"min-h":["screen","none",...W()]}],"max-h":[{"max-h":["screen",...W()]}],"font-size":[{text:["base",n,wc,Lo]}],"font-smoothing":["antialiased","subpixel-antialiased"],"font-style":["italic","not-italic"],"font-weight":[{font:[r,De,kb]}],"font-stretch":[{"font-stretch":["ultra-condensed","extra-condensed","condensed","semi-condensed","normal","semi-expanded","expanded","extra-expanded","ultra-expanded",Db,Ne]}],"font-family":[{font:[J8,Ne,t]}],"fvn-normal":["normal-nums"],"fvn-ordinal":["ordinal"],"fvn-slashed-zero":["slashed-zero"],"fvn-figure":["lining-nums","oldstyle-nums"],"fvn-spacing":["proportional-nums","tabular-nums"],"fvn-fraction":["diagonal-fractions","stacked-fractions"],tracking:[{tracking:[i,De,Ne]}],"line-clamp":[{"line-clamp":[nt,"none",De,kb]}],leading:[{leading:[a,...N()]}],"list-image":[{"list-image":["none",De,Ne]}],"list-style-position":[{list:["inside","outside"]}],"list-style-type":[{list:["disc","decimal","none",De,Ne]}],"text-alignment":[{text:["left","center","right","justify","start","end"]}],"placeholder-color":[{placeholder:B()}],"text-color":[{text:B()}],"text-decoration":["underline","overline","line-through","no-underline"],"text-decoration-style":[{decoration:[...le(),"wavy"]}],"text-decoration-thickness":[{decoration:[nt,"from-font","auto",De,Lo]}],"text-decoration-color":[{decoration:B()}],"underline-offset":[{"underline-offset":[nt,"auto",De,Ne]}],"text-transform":["uppercase","lowercase","capitalize","normal-case"],"text-overflow":["truncate","text-ellipsis","text-clip"],"text-wrap":[{text:["wrap","nowrap","balance","pretty"]}],indent:[{indent:N()}],"vertical-align":[{align:["baseline","top","middle","bottom","text-top","text-bottom","sub","super",De,Ne]}],whitespace:[{whitespace:["normal","nowrap","pre","pre-line","pre-wrap","break-spaces"]}],break:[{break:["normal","words","all","keep"]}],wrap:[{wrap:["break-word","anywhere","normal"]}],hyphens:[{hyphens:["none","manual","auto"]}],content:[{content:["none",De,Ne]}],"bg-attachment":[{bg:["fixed","local","scroll"]}],"bg-clip":[{"bg-clip":["border","padding","content","text"]}],"bg-origin":[{"bg-origin":["border","padding","content"]}],"bg-position":[{bg:Q()}],"bg-repeat":[{bg:$()}],"bg-size":[{bg:J()}],"bg-image":[{bg:["none",{linear:[{to:["t","tr","r","br","b","bl","l","tl"]},Ka,De,Ne],radial:["",De,Ne],conic:[Ka,De,Ne]},t9,Q8]}],"bg-color":[{bg:B()}],"gradient-from-pos":[{from:q()}],"gradient-via-pos":[{via:q()}],"gradient-to-pos":[{to:q()}],"gradient-from":[{from:B()}],"gradient-via":[{via:B()}],"gradient-to":[{to:B()}],rounded:[{rounded:X()}],"rounded-s":[{"rounded-s":X()}],"rounded-e":[{"rounded-e":X()}],"rounded-t":[{"rounded-t":X()}],"rounded-r":[{"rounded-r":X()}],"rounded-b":[{"rounded-b":X()}],"rounded-l":[{"rounded-l":X()}],"rounded-ss":[{"rounded-ss":X()}],"rounded-se":[{"rounded-se":X()}],"rounded-ee":[{"rounded-ee":X()}],"rounded-es":[{"rounded-es":X()}],"rounded-tl":[{"rounded-tl":X()}],"rounded-tr":[{"rounded-tr":X()}],"rounded-br":[{"rounded-br":X()}],"rounded-bl":[{"rounded-bl":X()}],"border-w":[{border:oe()}],"border-w-x":[{"border-x":oe()}],"border-w-y":[{"border-y":oe()}],"border-w-s":[{"border-s":oe()}],"border-w-e":[{"border-e":oe()}],"border-w-t":[{"border-t":oe()}],"border-w-r":[{"border-r":oe()}],"border-w-b":[{"border-b":oe()}],"border-w-l":[{"border-l":oe()}],"divide-x":[{"divide-x":oe()}],"divide-x-reverse":["divide-x-reverse"],"divide-y":[{"divide-y":oe()}],"divide-y-reverse":["divide-y-reverse"],"border-style":[{border:[...le(),"hidden","none"]}],"divide-style":[{divide:[...le(),"hidden","none"]}],"border-color":[{border:B()}],"border-color-x":[{"border-x":B()}],"border-color-y":[{"border-y":B()}],"border-color-s":[{"border-s":B()}],"border-color-e":[{"border-e":B()}],"border-color-t":[{"border-t":B()}],"border-color-r":[{"border-r":B()}],"border-color-b":[{"border-b":B()}],"border-color-l":[{"border-l":B()}],"divide-color":[{divide:B()}],"outline-style":[{outline:[...le(),"none","hidden"]}],"outline-offset":[{"outline-offset":[nt,De,Ne]}],"outline-w":[{outline:["",nt,wc,Lo]}],"outline-color":[{outline:B()}],shadow:[{shadow:["","none",p,Gp,Hp]}],"shadow-color":[{shadow:B()}],"inset-shadow":[{"inset-shadow":["none",m,Gp,Hp]}],"inset-shadow-color":[{"inset-shadow":B()}],"ring-w":[{ring:oe()}],"ring-w-inset":["ring-inset"],"ring-color":[{ring:B()}],"ring-offset-w":[{"ring-offset":[nt,Lo]}],"ring-offset-color":[{"ring-offset":B()}],"inset-ring-w":[{"inset-ring":oe()}],"inset-ring-color":[{"inset-ring":B()}],"text-shadow":[{"text-shadow":["none",v,Gp,Hp]}],"text-shadow-color":[{"text-shadow":B()}],opacity:[{opacity:[nt,De,Ne]}],"mix-blend":[{"mix-blend":[...ve(),"plus-darker","plus-lighter"]}],"bg-blend":[{"bg-blend":ve()}],"mask-clip":[{"mask-clip":["border","padding","content","fill","stroke","view"]},"mask-no-clip"],"mask-composite":[{mask:["add","subtract","intersect","exclude"]}],"mask-image-linear-pos":[{"mask-linear":[nt]}],"mask-image-linear-from-pos":[{"mask-linear-from":de()}],"mask-image-linear-to-pos":[{"mask-linear-to":de()}],"mask-image-linear-from-color":[{"mask-linear-from":B()}],"mask-image-linear-to-color":[{"mask-linear-to":B()}],"mask-image-t-from-pos":[{"mask-t-from":de()}],"mask-image-t-to-pos":[{"mask-t-to":de()}],"mask-image-t-from-color":[{"mask-t-from":B()}],"mask-image-t-to-color":[{"mask-t-to":B()}],"mask-image-r-from-pos":[{"mask-r-from":de()}],"mask-image-r-to-pos":[{"mask-r-to":de()}],"mask-image-r-from-color":[{"mask-r-from":B()}],"mask-image-r-to-color":[{"mask-r-to":B()}],"mask-image-b-from-pos":[{"mask-b-from":de()}],"mask-image-b-to-pos":[{"mask-b-to":de()}],"mask-image-b-from-color":[{"mask-b-from":B()}],"mask-image-b-to-color":[{"mask-b-to":B()}],"mask-image-l-from-pos":[{"mask-l-from":de()}],"mask-image-l-to-pos":[{"mask-l-to":de()}],"mask-image-l-from-color":[{"mask-l-from":B()}],"mask-image-l-to-color":[{"mask-l-to":B()}],"mask-image-x-from-pos":[{"mask-x-from":de()}],"mask-image-x-to-pos":[{"mask-x-to":de()}],"mask-image-x-from-color":[{"mask-x-from":B()}],"mask-image-x-to-color":[{"mask-x-to":B()}],"mask-image-y-from-pos":[{"mask-y-from":de()}],"mask-image-y-to-pos":[{"mask-y-to":de()}],"mask-image-y-from-color":[{"mask-y-from":B()}],"mask-image-y-to-color":[{"mask-y-to":B()}],"mask-image-radial":[{"mask-radial":[De,Ne]}],"mask-image-radial-from-pos":[{"mask-radial-from":de()}],"mask-image-radial-to-pos":[{"mask-radial-to":de()}],"mask-image-radial-from-color":[{"mask-radial-from":B()}],"mask-image-radial-to-color":[{"mask-radial-to":B()}],"mask-image-radial-shape":[{"mask-radial":["circle","ellipse"]}],"mask-image-radial-size":[{"mask-radial":[{closest:["side","corner"],farthest:["side","corner"]}]}],"mask-image-radial-pos":[{"mask-radial-at":T()}],"mask-image-conic-pos":[{"mask-conic":[nt]}],"mask-image-conic-from-pos":[{"mask-conic-from":de()}],"mask-image-conic-to-pos":[{"mask-conic-to":de()}],"mask-image-conic-from-color":[{"mask-conic-from":B()}],"mask-image-conic-to-color":[{"mask-conic-to":B()}],"mask-mode":[{mask:["alpha","luminance","match"]}],"mask-origin":[{"mask-origin":["border","padding","content","fill","stroke","view"]}],"mask-position":[{mask:Q()}],"mask-repeat":[{mask:$()}],"mask-size":[{mask:J()}],"mask-type":[{"mask-type":["alpha","luminance"]}],"mask-image":[{mask:["none",De,Ne]}],filter:[{filter:["","none",De,Ne]}],blur:[{blur:Se()}],brightness:[{brightness:[nt,De,Ne]}],contrast:[{contrast:[nt,De,Ne]}],"drop-shadow":[{"drop-shadow":["","none",b,Gp,Hp]}],"drop-shadow-color":[{"drop-shadow":B()}],grayscale:[{grayscale:["",nt,De,Ne]}],"hue-rotate":[{"hue-rotate":[nt,De,Ne]}],invert:[{invert:["",nt,De,Ne]}],saturate:[{saturate:[nt,De,Ne]}],sepia:[{sepia:["",nt,De,Ne]}],"backdrop-filter":[{"backdrop-filter":["","none",De,Ne]}],"backdrop-blur":[{"backdrop-blur":Se()}],"backdrop-brightness":[{"backdrop-brightness":[nt,De,Ne]}],"backdrop-contrast":[{"backdrop-contrast":[nt,De,Ne]}],"backdrop-grayscale":[{"backdrop-grayscale":["",nt,De,Ne]}],"backdrop-hue-rotate":[{"backdrop-hue-rotate":[nt,De,Ne]}],"backdrop-invert":[{"backdrop-invert":["",nt,De,Ne]}],"backdrop-opacity":[{"backdrop-opacity":[nt,De,Ne]}],"backdrop-saturate":[{"backdrop-saturate":[nt,De,Ne]}],"backdrop-sepia":[{"backdrop-sepia":["",nt,De,Ne]}],"border-collapse":[{border:["collapse","separate"]}],"border-spacing":[{"border-spacing":N()}],"border-spacing-x":[{"border-spacing-x":N()}],"border-spacing-y":[{"border-spacing-y":N()}],"table-layout":[{table:["auto","fixed"]}],caption:[{caption:["top","bottom"]}],transition:[{transition:["","all","colors","opacity","shadow","transform","none",De,Ne]}],"transition-behavior":[{transition:["normal","discrete"]}],duration:[{duration:[nt,"initial",De,Ne]}],ease:[{ease:["linear","initial",C,De,Ne]}],delay:[{delay:[nt,De,Ne]}],animate:[{animate:["none",O,De,Ne]}],backface:[{backface:["hidden","visible"]}],perspective:[{perspective:[w,De,Ne]}],"perspective-origin":[{"perspective-origin":R()}],rotate:[{rotate:ce()}],"rotate-x":[{"rotate-x":ce()}],"rotate-y":[{"rotate-y":ce()}],"rotate-z":[{"rotate-z":ce()}],scale:[{scale:we()}],"scale-x":[{"scale-x":we()}],"scale-y":[{"scale-y":we()}],"scale-z":[{"scale-z":we()}],"scale-3d":["scale-3d"],skew:[{skew:_e()}],"skew-x":[{"skew-x":_e()}],"skew-y":[{"skew-y":_e()}],transform:[{transform:[De,Ne,"","none","gpu","cpu"]}],"transform-origin":[{origin:R()}],"transform-style":[{transform:["3d","flat"]}],translate:[{translate:se()}],"translate-x":[{"translate-x":se()}],"translate-y":[{"translate-y":se()}],"translate-z":[{"translate-z":se()}],"translate-none":["translate-none"],accent:[{accent:B()}],appearance:[{appearance:["none","auto"]}],"caret-color":[{caret:B()}],"color-scheme":[{scheme:["normal","dark","light","light-dark","only-dark","only-light"]}],cursor:[{cursor:["auto","default","pointer","wait","text","move","help","not-allowed","none","context-menu","progress","cell","crosshair","vertical-text","alias","copy","no-drop","grab","grabbing","all-scroll","col-resize","row-resize","n-resize","e-resize","s-resize","w-resize","ne-resize","nw-resize","se-resize","sw-resize","ew-resize","ns-resize","nesw-resize","nwse-resize","zoom-in","zoom-out",De,Ne]}],"field-sizing":[{"field-sizing":["fixed","content"]}],"pointer-events":[{"pointer-events":["auto","none"]}],resize:[{resize:["none","","y","x"]}],"scroll-behavior":[{scroll:["auto","smooth"]}],"scroll-m":[{"scroll-m":N()}],"scroll-mx":[{"scroll-mx":N()}],"scroll-my":[{"scroll-my":N()}],"scroll-ms":[{"scroll-ms":N()}],"scroll-me":[{"scroll-me":N()}],"scroll-mt":[{"scroll-mt":N()}],"scroll-mr":[{"scroll-mr":N()}],"scroll-mb":[{"scroll-mb":N()}],"scroll-ml":[{"scroll-ml":N()}],"scroll-p":[{"scroll-p":N()}],"scroll-px":[{"scroll-px":N()}],"scroll-py":[{"scroll-py":N()}],"scroll-ps":[{"scroll-ps":N()}],"scroll-pe":[{"scroll-pe":N()}],"scroll-pt":[{"scroll-pt":N()}],"scroll-pr":[{"scroll-pr":N()}],"scroll-pb":[{"scroll-pb":N()}],"scroll-pl":[{"scroll-pl":N()}],"snap-align":[{snap:["start","end","center","align-none"]}],"snap-stop":[{snap:["normal","always"]}],"snap-type":[{snap:["none","x","y","both"]}],"snap-strictness":[{snap:["mandatory","proximity"]}],touch:[{touch:["auto","none","manipulation"]}],"touch-x":[{"touch-pan":["x","left","right"]}],"touch-y":[{"touch-pan":["y","up","down"]}],"touch-pz":["touch-pinch-zoom"],select:[{select:["none","text","all","auto"]}],"will-change":[{"will-change":["auto","scroll","contents","transform",De,Ne]}],fill:[{fill:["none",...B()]}],"stroke-w":[{stroke:[nt,wc,Lo,kb]}],stroke:[{stroke:["none",...B()]}],"forced-color-adjust":[{"forced-color-adjust":["auto","none"]}]},conflictingClassGroups:{overflow:["overflow-x","overflow-y"],overscroll:["overscroll-x","overscroll-y"],inset:["inset-x","inset-y","start","end","top","right","bottom","left"],"inset-x":["right","left"],"inset-y":["top","bottom"],flex:["basis","grow","shrink"],gap:["gap-x","gap-y"],p:["px","py","ps","pe","pt","pr","pb","pl"],px:["pr","pl"],py:["pt","pb"],m:["mx","my","ms","me","mt","mr","mb","ml"],mx:["mr","ml"],my:["mt","mb"],size:["w","h"],"font-size":["leading"],"fvn-normal":["fvn-ordinal","fvn-slashed-zero","fvn-figure","fvn-spacing","fvn-fraction"],"fvn-ordinal":["fvn-normal"],"fvn-slashed-zero":["fvn-normal"],"fvn-figure":["fvn-normal"],"fvn-spacing":["fvn-normal"],"fvn-fraction":["fvn-normal"],"line-clamp":["display","overflow"],rounded:["rounded-s","rounded-e","rounded-t","rounded-r","rounded-b","rounded-l","rounded-ss","rounded-se","rounded-ee","rounded-es","rounded-tl","rounded-tr","rounded-br","rounded-bl"],"rounded-s":["rounded-ss","rounded-es"],"rounded-e":["rounded-se","rounded-ee"],"rounded-t":["rounded-tl","rounded-tr"],"rounded-r":["rounded-tr","rounded-br"],"rounded-b":["rounded-br","rounded-bl"],"rounded-l":["rounded-tl","rounded-bl"],"border-spacing":["border-spacing-x","border-spacing-y"],"border-w":["border-w-x","border-w-y","border-w-s","border-w-e","border-w-t","border-w-r","border-w-b","border-w-l"],"border-w-x":["border-w-r","border-w-l"],"border-w-y":["border-w-t","border-w-b"],"border-color":["border-color-x","border-color-y","border-color-s","border-color-e","border-color-t","border-color-r","border-color-b","border-color-l"],"border-color-x":["border-color-r","border-color-l"],"border-color-y":["border-color-t","border-color-b"],translate:["translate-x","translate-y","translate-none"],"translate-none":["translate","translate-x","translate-y","translate-z"],"scroll-m":["scroll-mx","scroll-my","scroll-ms","scroll-me","scroll-mt","scroll-mr","scroll-mb","scroll-ml"],"scroll-mx":["scroll-mr","scroll-ml"],"scroll-my":["scroll-mt","scroll-mb"],"scroll-p":["scroll-px","scroll-py","scroll-ps","scroll-pe","scroll-pt","scroll-pr","scroll-pb","scroll-pl"],"scroll-px":["scroll-pr","scroll-pl"],"scroll-py":["scroll-pt","scroll-pb"],touch:["touch-x","touch-y","touch-pz"],"touch-x":["touch"],"touch-y":["touch"],"touch-pz":["touch"]},conflictingClassGroupModifiers:{"font-size":["leading"]},orderSensitiveModifiers:["*","**","after","backdrop","before","details-content","file","first-letter","first-line","marker","placeholder","selection"]}},a9=B8(i9);function ke(...e){return a9(ct(e))}function o9({className:e,children:t,viewport:n=!0,...r}){return E.jsxs(z6,{"data-slot":"navigation-menu","data-viewport":n,className:ke("group/navigation-menu relative flex max-w-max flex-1 items-center justify-center",e),...r,children:[t,n&&E.jsx(u9,{})]})}function s9({className:e,...t}){return E.jsx(U6,{"data-slot":"navigation-menu-list",className:ke("group flex flex-1 list-none items-center justify-center gap-1",e),...t})}function l9({className:e,...t}){return E.jsx(V6,{"data-slot":"navigation-menu-item",className:ke("relative",e),...t})}function u9({className:e,...t}){return E.jsx("div",{className:ke("absolute top-full left-0 isolate z-50 flex justify-center"),children:E.jsx(H6,{"data-slot":"navigation-menu-viewport",className:ke("origin-top-center bg-popover text-popover-foreground data-[state=open]:animate-in data-[state=closed]:animate-out data-[state=closed]:zoom-out-95 data-[state=open]:zoom-in-90 relative mt-1.5 h-[var(--radix-navigation-menu-viewport-height)] w-full overflow-hidden rounded-md border shadow md:w-[var(--radix-navigation-menu-viewport-width)]",e),...t})})}var Sc={},ZM;function c9(){if(ZM)return Sc;ZM=1,Object.defineProperty(Sc,"__esModule",{value:!0}),Sc.parse=s,Sc.serialize=d;const e=/^[\u0021-\u003A\u003C\u003E-\u007E]+$/,t=/^[\u0021-\u003A\u003C-\u007E]*$/,n=/^([.]?[a-z0-9]([a-z0-9-]{0,61}[a-z0-9])?)([.][a-z0-9]([a-z0-9-]{0,61}[a-z0-9])?)*$/i,r=/^[\u0020-\u003A\u003D-\u007E]*$/,i=Object.prototype.toString,a=(()=>{const v=function(){};return v.prototype=Object.create(null),v})();function s(v,b){const S=new a,w=v.length;if(w<2)return S;const x=(b==null?void 0:b.decode)||p;let C=0;do{const O=v.indexOf("=",C);if(O===-1)break;const A=v.indexOf(";",C),T=A===-1?w:A;if(O>T){C=v.lastIndexOf(";",O-1)+1;continue}const R=u(v,C,O),M=c(v,O,R),P=v.slice(R,M);if(S[P]===void 0){let N=u(v,O+1,T),k=c(v,T,N);const H=x(v.slice(N,k));S[P]=H}C=T+1}while(C<w);return S}function u(v,b,S){do{const w=v.charCodeAt(b);if(w!==32&&w!==9)return b}while(++b<S);return S}function c(v,b,S){for(;b>S;){const w=v.charCodeAt(--b);if(w!==32&&w!==9)return b+1}return S}function d(v,b,S){const w=(S==null?void 0:S.encode)||encodeURIComponent;if(!e.test(v))throw new TypeError(`argument name is invalid: ${v}`);const x=w(b);if(!t.test(x))throw new TypeError(`argument val is invalid: ${b}`);let C=v+"="+x;if(!S)return C;if(S.maxAge!==void 0){if(!Number.isInteger(S.maxAge))throw new TypeError(`option maxAge is invalid: ${S.maxAge}`);C+="; Max-Age="+S.maxAge}if(S.domain){if(!n.test(S.domain))throw new TypeError(`option domain is invalid: ${S.domain}`);C+="; Domain="+S.domain}if(S.path){if(!r.test(S.path))throw new TypeError(`option path is invalid: ${S.path}`);C+="; Path="+S.path}if(S.expires){if(!m(S.expires)||!Number.isFinite(S.expires.valueOf()))throw new TypeError(`option expires is invalid: ${S.expires}`);C+="; Expires="+S.expires.toUTCString()}if(S.httpOnly&&(C+="; HttpOnly"),S.secure&&(C+="; Secure"),S.partitioned&&(C+="; Partitioned"),S.priority)switch(typeof S.priority=="string"?S.priority.toLowerCase():void 0){case"low":C+="; Priority=Low";break;case"medium":C+="; Priority=Medium";break;case"high":C+="; Priority=High";break;default:throw new TypeError(`option priority is invalid: ${S.priority}`)}if(S.sameSite)switch(typeof S.sameSite=="string"?S.sameSite.toLowerCase():S.sameSite){case!0:case"strict":C+="; SameSite=Strict";break;case"lax":C+="; SameSite=Lax";break;case"none":C+="; SameSite=None";break;default:throw new TypeError(`option sameSite is invalid: ${S.sameSite}`)}return C}function p(v){if(v.indexOf("%")===-1)return v;try{return decodeURIComponent(v)}catch{return v}}function m(v){return i.call(v)==="[object Date]"}return Sc}c9();var XM="popstate";function f9(e={}){function t(r,i){let{pathname:a,search:s,hash:u}=r.location;return S_("",{pathname:a,search:s,hash:u},i.state&&i.state.usr||null,i.state&&i.state.key||"default")}function n(r,i){return typeof i=="string"?i:Jc(i)}return p9(t,n,null,e)}function zt(e,t){if(e===!1||e===null||typeof e>"u")throw new Error(t)}function Si(e,t){if(!e){typeof console<"u"&&console.warn(t);try{throw new Error(t)}catch{}}}function d9(){return Math.random().toString(36).substring(2,10)}function QM(e,t){return{usr:e.state,key:e.key,idx:t}}function S_(e,t,n=null,r){return{pathname:typeof e=="string"?e:e.pathname,search:"",hash:"",...typeof t=="string"?su(t):t,state:n,key:t&&t.key||r||d9()}}function Jc({pathname:e="/",search:t="",hash:n=""}){return t&&t!=="?"&&(e+=t.charAt(0)==="?"?t:"?"+t),n&&n!=="#"&&(e+=n.charAt(0)==="#"?n:"#"+n),e}function su(e){let t={};if(e){let n=e.indexOf("#");n>=0&&(t.hash=e.substring(n),e=e.substring(0,n));let r=e.indexOf("?");r>=0&&(t.search=e.substring(r),e=e.substring(0,r)),e&&(t.pathname=e)}return t}function p9(e,t,n,r={}){let{window:i=document.defaultView,v5Compat:a=!1}=r,s=i.history,u="POP",c=null,d=p();d==null&&(d=0,s.replaceState({...s.state,idx:d},""));function p(){return(s.state||{idx:null}).idx}function m(){u="POP";let x=p(),C=x==null?null:x-d;d=x,c&&c({action:u,location:w.location,delta:C})}function v(x,C){u="PUSH";let O=S_(w.location,x,C);d=p()+1;let A=QM(O,d),T=w.createHref(O);try{s.pushState(A,"",T)}catch(R){if(R instanceof DOMException&&R.name==="DataCloneError")throw R;i.location.assign(T)}a&&c&&c({action:u,location:w.location,delta:1})}function b(x,C){u="REPLACE";let O=S_(w.location,x,C);d=p();let A=QM(O,d),T=w.createHref(O);s.replaceState(A,"",T),a&&c&&c({action:u,location:w.location,delta:0})}function S(x){let C=i.location.origin!=="null"?i.location.origin:i.location.href,O=typeof x=="string"?x:Jc(x);return O=O.replace(/ $/,"%20"),zt(C,`No window.location.(origin|href) available to create URL for href: ${O}`),new URL(O,C)}let w={get action(){return u},get location(){return e(i,s)},listen(x){if(c)throw new Error("A history only accepts one active listener");return i.addEventListener(XM,m),c=x,()=>{i.removeEventListener(XM,m),c=null}},createHref(x){return t(i,x)},createURL:S,encodeLocation(x){let C=S(x);return{pathname:C.pathname,search:C.search,hash:C.hash}},push:v,replace:b,go(x){return s.go(x)}};return w}function jL(e,t,n="/"){return h9(e,t,n,!1)}function h9(e,t,n,r){let i=typeof t=="string"?su(t):t,a=ca(i.pathname||"/",n);if(a==null)return null;let s=NL(e);m9(s);let u=null;for(let c=0;u==null&&c<s.length;++c){let d=O9(a);u=E9(s[c],d,r)}return u}function NL(e,t=[],n=[],r=""){let i=(a,s,u)=>{let c={relativePath:u===void 0?a.path||"":u,caseSensitive:a.caseSensitive===!0,childrenIndex:s,route:a};c.relativePath.startsWith("/")&&(zt(c.relativePath.startsWith(r),`Absolute route path "${c.relativePath}" nested under path "${r}" is not valid. An absolute child route path must start with the combined path of all its parent routes.`),c.relativePath=c.relativePath.slice(r.length));let d=ia([r,c.relativePath]),p=n.concat(c);a.children&&a.children.length>0&&(zt(a.index!==!0,`Index routes must not have child routes. Please remove all child routes from route path "${d}".`),NL(a.children,t,p,d)),!(a.path==null&&!a.index)&&t.push({path:d,score:S9(d,a.index),routesMeta:p})};return e.forEach((a,s)=>{var u;if(a.path===""||!((u=a.path)!=null&&u.includes("?")))i(a,s);else for(let c of DL(a.path))i(a,s,c)}),t}function DL(e){let t=e.split("/");if(t.length===0)return[];let[n,...r]=t,i=n.endsWith("?"),a=n.replace(/\?$/,"");if(r.length===0)return i?[a,""]:[a];let s=DL(r.join("/")),u=[];return u.push(...s.map(c=>c===""?a:[a,c].join("/"))),i&&u.push(...s),u.map(c=>e.startsWith("/")&&c===""?"/":c)}function m9(e){e.sort((t,n)=>t.score!==n.score?n.score-t.score:_9(t.routesMeta.map(r=>r.childrenIndex),n.routesMeta.map(r=>r.childrenIndex)))}var v9=/^:[\w-]+$/,g9=3,y9=2,b9=1,x9=10,w9=-2,JM=e=>e==="*";function S9(e,t){let n=e.split("/"),r=n.length;return n.some(JM)&&(r+=w9),t&&(r+=y9),n.filter(i=>!JM(i)).reduce((i,a)=>i+(v9.test(a)?g9:a===""?b9:x9),r)}function _9(e,t){return e.length===t.length&&e.slice(0,-1).every((r,i)=>r===t[i])?e[e.length-1]-t[t.length-1]:0}function E9(e,t,n=!1){let{routesMeta:r}=e,i={},a="/",s=[];for(let u=0;u<r.length;++u){let c=r[u],d=u===r.length-1,p=a==="/"?t:t.slice(a.length)||"/",m=Th({path:c.relativePath,caseSensitive:c.caseSensitive,end:d},p),v=c.route;if(!m&&d&&n&&!r[r.length-1].route.index&&(m=Th({path:c.relativePath,caseSensitive:c.caseSensitive,end:!1},p)),!m)return null;Object.assign(i,m.params),s.push({params:i,pathname:ia([a,m.pathname]),pathnameBase:M9(ia([a,m.pathnameBase])),route:v}),m.pathnameBase!=="/"&&(a=ia([a,m.pathnameBase]))}return s}function Th(e,t){typeof e=="string"&&(e={path:e,caseSensitive:!1,end:!0});let[n,r]=C9(e.path,e.caseSensitive,e.end),i=t.match(n);if(!i)return null;let a=i[0],s=a.replace(/(.)\/+$/,"$1"),u=i.slice(1);return{params:r.reduce((d,{paramName:p,isOptional:m},v)=>{if(p==="*"){let S=u[v]||"";s=a.slice(0,a.length-S.length).replace(/(.)\/+$/,"$1")}const b=u[v];return m&&!b?d[p]=void 0:d[p]=(b||"").replace(/%2F/g,"/"),d},{}),pathname:a,pathnameBase:s,pattern:e}}function C9(e,t=!1,n=!0){Si(e==="*"||!e.endsWith("*")||e.endsWith("/*"),`Route path "${e}" will be treated as if it were "${e.replace(/\*$/,"/*")}" because the \`*\` character must always follow a \`/\` in the pattern. To get rid of this warning, please change the route path to "${e.replace(/\*$/,"/*")}".`);let r=[],i="^"+e.replace(/\/*\*?$/,"").replace(/^\/*/,"/").replace(/[\\.*+^${}|()[\]]/g,"\\$&").replace(/\/:([\w-]+)(\?)?/g,(s,u,c)=>(r.push({paramName:u,isOptional:c!=null}),c?"/?([^\\/]+)?":"/([^\\/]+)"));return e.endsWith("*")?(r.push({paramName:"*"}),i+=e==="*"||e==="/*"?"(.*)$":"(?:\\/(.+)|\\/*)$"):n?i+="\\/*$":e!==""&&e!=="/"&&(i+="(?:(?=\\/|$))"),[new RegExp(i,t?void 0:"i"),r]}function O9(e){try{return e.split("/").map(t=>decodeURIComponent(t).replace(/\//g,"%2F")).join("/")}catch(t){return Si(!1,`The URL path "${e}" could not be decoded because it is a malformed URL segment. This is probably due to a bad percent encoding (${t}).`),e}}function ca(e,t){if(t==="/")return e;if(!e.toLowerCase().startsWith(t.toLowerCase()))return null;let n=t.endsWith("/")?t.length-1:t.length,r=e.charAt(n);return r&&r!=="/"?null:e.slice(n)||"/"}function A9(e,t="/"){let{pathname:n,search:r="",hash:i=""}=typeof e=="string"?su(e):e;return{pathname:n?n.startsWith("/")?n:R9(n,t):t,search:P9(r),hash:j9(i)}}function R9(e,t){let n=t.replace(/\/+$/,"").split("/");return e.split("/").forEach(i=>{i===".."?n.length>1&&n.pop():i!=="."&&n.push(i)}),n.length>1?n.join("/"):"/"}function $b(e,t,n,r){return`Cannot include a '${e}' character in a manually specified \`to.${t}\` field [${JSON.stringify(r)}].  Please separate it out to the \`to.${n}\` field. Alternatively you may provide the full path as a string in <Link to="..."> and the router will parse it for you.`}function T9(e){return e.filter((t,n)=>n===0||t.route.path&&t.route.path.length>0)}function kL(e){let t=T9(e);return t.map((n,r)=>r===t.length-1?n.pathname:n.pathnameBase)}function $L(e,t,n,r=!1){let i;typeof e=="string"?i=su(e):(i={...e},zt(!i.pathname||!i.pathname.includes("?"),$b("?","pathname","search",i)),zt(!i.pathname||!i.pathname.includes("#"),$b("#","pathname","hash",i)),zt(!i.search||!i.search.includes("#"),$b("#","search","hash",i)));let a=e===""||i.pathname==="",s=a?"/":i.pathname,u;if(s==null)u=n;else{let m=t.length-1;if(!r&&s.startsWith("..")){let v=s.split("/");for(;v[0]==="..";)v.shift(),m-=1;i.pathname=v.join("/")}u=m>=0?t[m]:"/"}let c=A9(i,u),d=s&&s!=="/"&&s.endsWith("/"),p=(a||s===".")&&n.endsWith("/");return!c.pathname.endsWith("/")&&(d||p)&&(c.pathname+="/"),c}var ia=e=>e.join("/").replace(/\/\/+/g,"/"),M9=e=>e.replace(/\/+$/,"").replace(/^\/*/,"/"),P9=e=>!e||e==="?"?"":e.startsWith("?")?e:"?"+e,j9=e=>!e||e==="#"?"":e.startsWith("#")?e:"#"+e;function N9(e){return e!=null&&typeof e.status=="number"&&typeof e.statusText=="string"&&typeof e.internal=="boolean"&&"data"in e}var IL=["POST","PUT","PATCH","DELETE"];new Set(IL);var D9=["GET",...IL];new Set(D9);var lu=g.createContext(null);lu.displayName="DataRouter";var Xm=g.createContext(null);Xm.displayName="DataRouterState";var LL=g.createContext({isTransitioning:!1});LL.displayName="ViewTransition";var k9=g.createContext(new Map);k9.displayName="Fetchers";var $9=g.createContext(null);$9.displayName="Await";var Ai=g.createContext(null);Ai.displayName="Navigation";var id=g.createContext(null);id.displayName="Location";var Ri=g.createContext({outlet:null,matches:[],isDataRoute:!1});Ri.displayName="Route";var DE=g.createContext(null);DE.displayName="RouteError";function I9(e,{relative:t}={}){zt(ad(),"useHref() may be used only in the context of a <Router> component.");let{basename:n,navigator:r}=g.useContext(Ai),{hash:i,pathname:a,search:s}=od(e,{relative:t}),u=a;return n!=="/"&&(u=a==="/"?n:ia([n,a])),r.createHref({pathname:u,search:s,hash:i})}function ad(){return g.useContext(id)!=null}function ms(){return zt(ad(),"useLocation() may be used only in the context of a <Router> component."),g.useContext(id).location}var BL="You should call navigate() in a React.useEffect(), not when your component is first rendered.";function FL(e){g.useContext(Ai).static||g.useLayoutEffect(e)}function kE(){let{isDataRoute:e}=g.useContext(Ri);return e?Z9():L9()}function L9(){zt(ad(),"useNavigate() may be used only in the context of a <Router> component.");let e=g.useContext(lu),{basename:t,navigator:n}=g.useContext(Ai),{matches:r}=g.useContext(Ri),{pathname:i}=ms(),a=JSON.stringify(kL(r)),s=g.useRef(!1);return FL(()=>{s.current=!0}),g.useCallback((c,d={})=>{if(Si(s.current,BL),!s.current)return;if(typeof c=="number"){n.go(c);return}let p=$L(c,JSON.parse(a),i,d.relative==="path");e==null&&t!=="/"&&(p.pathname=p.pathname==="/"?t:ia([t,p.pathname])),(d.replace?n.replace:n.push)(p,d.state,d)},[t,n,a,i,e])}g.createContext(null);function $E(){let{matches:e}=g.useContext(Ri),t=e[e.length-1];return t?t.params:{}}function od(e,{relative:t}={}){let{matches:n}=g.useContext(Ri),{pathname:r}=ms(),i=JSON.stringify(kL(n));return g.useMemo(()=>$L(e,JSON.parse(i),r,t==="path"),[e,i,r,t])}function B9(e,t){return qL(e,t)}function qL(e,t,n,r){var O;zt(ad(),"useRoutes() may be used only in the context of a <Router> component.");let{navigator:i,static:a}=g.useContext(Ai),{matches:s}=g.useContext(Ri),u=s[s.length-1],c=u?u.params:{},d=u?u.pathname:"/",p=u?u.pathnameBase:"/",m=u&&u.route;{let A=m&&m.path||"";zL(d,!m||A.endsWith("*")||A.endsWith("*?"),`You rendered descendant <Routes> (or called \`useRoutes()\`) at "${d}" (under <Route path="${A}">) but the parent route path has no trailing "*". This means if you navigate deeper, the parent won't match anymore and therefore the child routes will never render.

Please change the parent <Route path="${A}"> to <Route path="${A==="/"?"*":`${A}/*`}">.`)}let v=ms(),b;if(t){let A=typeof t=="string"?su(t):t;zt(p==="/"||((O=A.pathname)==null?void 0:O.startsWith(p)),`When overriding the location using \`<Routes location>\` or \`useRoutes(routes, location)\`, the location pathname must begin with the portion of the URL pathname that was matched by all parent routes. The current pathname base is "${p}" but pathname "${A.pathname}" was given in the \`location\` prop.`),b=A}else b=v;let S=b.pathname||"/",w=S;if(p!=="/"){let A=p.replace(/^\//,"").split("/");w="/"+S.replace(/^\//,"").split("/").slice(A.length).join("/")}let x=!a&&n&&n.matches&&n.matches.length>0?n.matches:jL(e,{pathname:w});Si(m||x!=null,`No routes matched location "${b.pathname}${b.search}${b.hash}" `),Si(x==null||x[x.length-1].route.element!==void 0||x[x.length-1].route.Component!==void 0||x[x.length-1].route.lazy!==void 0,`Matched leaf route at location "${b.pathname}${b.search}${b.hash}" does not have an element or Component. This means it will render an <Outlet /> with a null value by default resulting in an "empty" page.`);let C=V9(x&&x.map(A=>Object.assign({},A,{params:Object.assign({},c,A.params),pathname:ia([p,i.encodeLocation?i.encodeLocation(A.pathname).pathname:A.pathname]),pathnameBase:A.pathnameBase==="/"?p:ia([p,i.encodeLocation?i.encodeLocation(A.pathnameBase).pathname:A.pathnameBase])})),s,n,r);return t&&C?g.createElement(id.Provider,{value:{location:{pathname:"/",search:"",hash:"",state:null,key:"default",...b},navigationType:"POP"}},C):C}function F9(){let e=Y9(),t=N9(e)?`${e.status} ${e.statusText}`:e instanceof Error?e.message:JSON.stringify(e),n=e instanceof Error?e.stack:null,r="rgba(200,200,200, 0.5)",i={padding:"0.5rem",backgroundColor:r},a={padding:"2px 4px",backgroundColor:r},s=null;return console.error("Error handled by React Router default ErrorBoundary:",e),s=g.createElement(g.Fragment,null,g.createElement("p",null,"💿 Hey developer 👋"),g.createElement("p",null,"You can provide a way better UX than this when your app throws errors by providing your own ",g.createElement("code",{style:a},"ErrorBoundary")," or"," ",g.createElement("code",{style:a},"errorElement")," prop on your route.")),g.createElement(g.Fragment,null,g.createElement("h2",null,"Unexpected Application Error!"),g.createElement("h3",{style:{fontStyle:"italic"}},t),n?g.createElement("pre",{style:i},n):null,s)}var q9=g.createElement(F9,null),z9=class extends g.Component{constructor(e){super(e),this.state={location:e.location,revalidation:e.revalidation,error:e.error}}static getDerivedStateFromError(e){return{error:e}}static getDerivedStateFromProps(e,t){return t.location!==e.location||t.revalidation!=="idle"&&e.revalidation==="idle"?{error:e.error,location:e.location,revalidation:e.revalidation}:{error:e.error!==void 0?e.error:t.error,location:t.location,revalidation:e.revalidation||t.revalidation}}componentDidCatch(e,t){console.error("React Router caught the following error during render",e,t)}render(){return this.state.error!==void 0?g.createElement(Ri.Provider,{value:this.props.routeContext},g.createElement(DE.Provider,{value:this.state.error,children:this.props.component})):this.props.children}};function U9({routeContext:e,match:t,children:n}){let r=g.useContext(lu);return r&&r.static&&r.staticContext&&(t.route.errorElement||t.route.ErrorBoundary)&&(r.staticContext._deepestRenderedBoundaryId=t.route.id),g.createElement(Ri.Provider,{value:e},n)}function V9(e,t=[],n=null,r=null){if(e==null){if(!n)return null;if(n.errors)e=n.matches;else if(t.length===0&&!n.initialized&&n.matches.length>0)e=n.matches;else return null}let i=e,a=n==null?void 0:n.errors;if(a!=null){let c=i.findIndex(d=>d.route.id&&(a==null?void 0:a[d.route.id])!==void 0);zt(c>=0,`Could not find a matching route for errors on route IDs: ${Object.keys(a).join(",")}`),i=i.slice(0,Math.min(i.length,c+1))}let s=!1,u=-1;if(n)for(let c=0;c<i.length;c++){let d=i[c];if((d.route.HydrateFallback||d.route.hydrateFallbackElement)&&(u=c),d.route.id){let{loaderData:p,errors:m}=n,v=d.route.loader&&!p.hasOwnProperty(d.route.id)&&(!m||m[d.route.id]===void 0);if(d.route.lazy||v){s=!0,u>=0?i=i.slice(0,u+1):i=[i[0]];break}}}return i.reduceRight((c,d,p)=>{let m,v=!1,b=null,S=null;n&&(m=a&&d.route.id?a[d.route.id]:void 0,b=d.route.errorElement||q9,s&&(u<0&&p===0?(zL("route-fallback",!1,"No `HydrateFallback` element provided to render during initial hydration"),v=!0,S=null):u===p&&(v=!0,S=d.route.hydrateFallbackElement||null)));let w=t.concat(i.slice(0,p+1)),x=()=>{let C;return m?C=b:v?C=S:d.route.Component?C=g.createElement(d.route.Component,null):d.route.element?C=d.route.element:C=c,g.createElement(U9,{match:d,routeContext:{outlet:c,matches:w,isDataRoute:n!=null},children:C})};return n&&(d.route.ErrorBoundary||d.route.errorElement||p===0)?g.createElement(z9,{location:n.location,revalidation:n.revalidation,component:b,error:m,children:x(),routeContext:{outlet:null,matches:w,isDataRoute:!0}}):x()},null)}function IE(e){return`${e} must be used within a data router.  See https://reactrouter.com/en/main/routers/picking-a-router.`}function H9(e){let t=g.useContext(lu);return zt(t,IE(e)),t}function G9(e){let t=g.useContext(Xm);return zt(t,IE(e)),t}function W9(e){let t=g.useContext(Ri);return zt(t,IE(e)),t}function LE(e){let t=W9(e),n=t.matches[t.matches.length-1];return zt(n.route.id,`${e} can only be used on routes that contain a unique "id"`),n.route.id}function K9(){return LE("useRouteId")}function Y9(){var r;let e=g.useContext(DE),t=G9("useRouteError"),n=LE("useRouteError");return e!==void 0?e:(r=t.errors)==null?void 0:r[n]}function Z9(){let{router:e}=H9("useNavigate"),t=LE("useNavigate"),n=g.useRef(!1);return FL(()=>{n.current=!0}),g.useCallback(async(i,a={})=>{Si(n.current,BL),n.current&&(typeof i=="number"?e.navigate(i):await e.navigate(i,{fromRouteId:t,...a}))},[e,t])}var eP={};function zL(e,t,n){!t&&!eP[e]&&(eP[e]=!0,Si(!1,n))}g.memo(X9);function X9({routes:e,future:t,state:n}){return qL(e,void 0,n,t)}function Ya(e){zt(!1,"A <Route> is only ever to be used as the child of <Routes> element, never rendered directly. Please wrap your <Route> in a <Routes>.")}function Q9({basename:e="/",children:t=null,location:n,navigationType:r="POP",navigator:i,static:a=!1}){zt(!ad(),"You cannot render a <Router> inside another <Router>. You should never have more than one in your app.");let s=e.replace(/^\/*/,"/"),u=g.useMemo(()=>({basename:s,navigator:i,static:a,future:{}}),[s,i,a]);typeof n=="string"&&(n=su(n));let{pathname:c="/",search:d="",hash:p="",state:m=null,key:v="default"}=n,b=g.useMemo(()=>{let S=ca(c,s);return S==null?null:{location:{pathname:S,search:d,hash:p,state:m,key:v},navigationType:r}},[s,c,d,p,m,v,r]);return Si(b!=null,`<Router basename="${s}"> is not able to match the URL "${c}${d}${p}" because it does not start with the basename, so the <Router> won't render anything.`),b==null?null:g.createElement(Ai.Provider,{value:u},g.createElement(id.Provider,{children:t,value:b}))}function J9({children:e,location:t}){return B9(__(e),t)}function __(e,t=[]){let n=[];return g.Children.forEach(e,(r,i)=>{if(!g.isValidElement(r))return;let a=[...t,i];if(r.type===g.Fragment){n.push.apply(n,__(r.props.children,a));return}zt(r.type===Ya,`[${typeof r.type=="string"?r.type:r.type.name}] is not a <Route> component. All component children of <Routes> must be a <Route> or <React.Fragment>`),zt(!r.props.index||!r.props.children,"An index route cannot have child routes.");let s={id:r.props.id||a.join("-"),caseSensitive:r.props.caseSensitive,element:r.props.element,Component:r.props.Component,index:r.props.index,path:r.props.path,loader:r.props.loader,action:r.props.action,hydrateFallbackElement:r.props.hydrateFallbackElement,HydrateFallback:r.props.HydrateFallback,errorElement:r.props.errorElement,ErrorBoundary:r.props.ErrorBoundary,hasErrorBoundary:r.props.hasErrorBoundary===!0||r.props.ErrorBoundary!=null||r.props.errorElement!=null,shouldRevalidate:r.props.shouldRevalidate,handle:r.props.handle,lazy:r.props.lazy};r.props.children&&(s.children=__(r.props.children,a)),n.push(s)}),n}var yh="get",bh="application/x-www-form-urlencoded";function Qm(e){return e!=null&&typeof e.tagName=="string"}function eG(e){return Qm(e)&&e.tagName.toLowerCase()==="button"}function tG(e){return Qm(e)&&e.tagName.toLowerCase()==="form"}function nG(e){return Qm(e)&&e.tagName.toLowerCase()==="input"}function rG(e){return!!(e.metaKey||e.altKey||e.ctrlKey||e.shiftKey)}function iG(e,t){return e.button===0&&(!t||t==="_self")&&!rG(e)}var Wp=null;function aG(){if(Wp===null)try{new FormData(document.createElement("form"),0),Wp=!1}catch{Wp=!0}return Wp}var oG=new Set(["application/x-www-form-urlencoded","multipart/form-data","text/plain"]);function Ib(e){return e!=null&&!oG.has(e)?(Si(!1,`"${e}" is not a valid \`encType\` for \`<Form>\`/\`<fetcher.Form>\` and will default to "${bh}"`),null):e}function sG(e,t){let n,r,i,a,s;if(tG(e)){let u=e.getAttribute("action");r=u?ca(u,t):null,n=e.getAttribute("method")||yh,i=Ib(e.getAttribute("enctype"))||bh,a=new FormData(e)}else if(eG(e)||nG(e)&&(e.type==="submit"||e.type==="image")){let u=e.form;if(u==null)throw new Error('Cannot submit a <button> or <input type="submit"> without a <form>');let c=e.getAttribute("formaction")||u.getAttribute("action");if(r=c?ca(c,t):null,n=e.getAttribute("formmethod")||u.getAttribute("method")||yh,i=Ib(e.getAttribute("formenctype"))||Ib(u.getAttribute("enctype"))||bh,a=new FormData(u,e),!aG()){let{name:d,type:p,value:m}=e;if(p==="image"){let v=d?`${d}.`:"";a.append(`${v}x`,"0"),a.append(`${v}y`,"0")}else d&&a.append(d,m)}}else{if(Qm(e))throw new Error('Cannot submit element that is not <form>, <button>, or <input type="submit|image">');n=yh,r=null,i=bh,s=e}return a&&i==="text/plain"&&(s=a,a=void 0),{action:r,method:n.toLowerCase(),encType:i,formData:a,body:s}}function BE(e,t){if(e===!1||e===null||typeof e>"u")throw new Error(t)}async function lG(e,t){if(e.id in t)return t[e.id];try{let n=await import(e.module);return t[e.id]=n,n}catch(n){return console.error(`Error loading route module \`${e.module}\`, reloading page...`),console.error(n),window.__reactRouterContext&&window.__reactRouterContext.isSpaMode,window.location.reload(),new Promise(()=>{})}}function uG(e){return e==null?!1:e.href==null?e.rel==="preload"&&typeof e.imageSrcSet=="string"&&typeof e.imageSizes=="string":typeof e.rel=="string"&&typeof e.href=="string"}async function cG(e,t,n){let r=await Promise.all(e.map(async i=>{let a=t.routes[i.route.id];if(a){let s=await lG(a,n);return s.links?s.links():[]}return[]}));return hG(r.flat(1).filter(uG).filter(i=>i.rel==="stylesheet"||i.rel==="preload").map(i=>i.rel==="stylesheet"?{...i,rel:"prefetch",as:"style"}:{...i,rel:"prefetch"}))}function tP(e,t,n,r,i,a){let s=(c,d)=>n[d]?c.route.id!==n[d].route.id:!0,u=(c,d)=>{var p;return n[d].pathname!==c.pathname||((p=n[d].route.path)==null?void 0:p.endsWith("*"))&&n[d].params["*"]!==c.params["*"]};return a==="assets"?t.filter((c,d)=>s(c,d)||u(c,d)):a==="data"?t.filter((c,d)=>{var m;let p=r.routes[c.route.id];if(!p||!p.hasLoader)return!1;if(s(c,d)||u(c,d))return!0;if(c.route.shouldRevalidate){let v=c.route.shouldRevalidate({currentUrl:new URL(i.pathname+i.search+i.hash,window.origin),currentParams:((m=n[0])==null?void 0:m.params)||{},nextUrl:new URL(e,window.origin),nextParams:c.params,defaultShouldRevalidate:!0});if(typeof v=="boolean")return v}return!0}):[]}function fG(e,t,{includeHydrateFallback:n}={}){return dG(e.map(r=>{let i=t.routes[r.route.id];if(!i)return[];let a=[i.module];return i.clientActionModule&&(a=a.concat(i.clientActionModule)),i.clientLoaderModule&&(a=a.concat(i.clientLoaderModule)),n&&i.hydrateFallbackModule&&(a=a.concat(i.hydrateFallbackModule)),i.imports&&(a=a.concat(i.imports)),a}).flat(1))}function dG(e){return[...new Set(e)]}function pG(e){let t={},n=Object.keys(e).sort();for(let r of n)t[r]=e[r];return t}function hG(e,t){let n=new Set;return new Set(t),e.reduce((r,i)=>{let a=JSON.stringify(pG(i));return n.has(a)||(n.add(a),r.push({key:a,link:i})),r},[])}var mG=new Set([100,101,204,205]);function vG(e,t){let n=typeof e=="string"?new URL(e,typeof window>"u"?"server://singlefetch/":window.location.origin):e;return n.pathname==="/"?n.pathname="_root.data":t&&ca(n.pathname,t)==="/"?n.pathname=`${t.replace(/\/$/,"")}/_root.data`:n.pathname=`${n.pathname.replace(/\/$/,"")}.data`,n}function UL(){let e=g.useContext(lu);return BE(e,"You must render this element inside a <DataRouterContext.Provider> element"),e}function gG(){let e=g.useContext(Xm);return BE(e,"You must render this element inside a <DataRouterStateContext.Provider> element"),e}var FE=g.createContext(void 0);FE.displayName="FrameworkContext";function VL(){let e=g.useContext(FE);return BE(e,"You must render this element inside a <HydratedRouter> element"),e}function yG(e,t){let n=g.useContext(FE),[r,i]=g.useState(!1),[a,s]=g.useState(!1),{onFocus:u,onBlur:c,onMouseEnter:d,onMouseLeave:p,onTouchStart:m}=t,v=g.useRef(null);g.useEffect(()=>{if(e==="render"&&s(!0),e==="viewport"){let w=C=>{C.forEach(O=>{s(O.isIntersecting)})},x=new IntersectionObserver(w,{threshold:.5});return v.current&&x.observe(v.current),()=>{x.disconnect()}}},[e]),g.useEffect(()=>{if(r){let w=setTimeout(()=>{s(!0)},100);return()=>{clearTimeout(w)}}},[r]);let b=()=>{i(!0)},S=()=>{i(!1),s(!1)};return n?e!=="intent"?[a,v,{}]:[a,v,{onFocus:_c(u,b),onBlur:_c(c,S),onMouseEnter:_c(d,b),onMouseLeave:_c(p,S),onTouchStart:_c(m,b)}]:[!1,v,{}]}function _c(e,t){return n=>{e&&e(n),n.defaultPrevented||t(n)}}function bG({page:e,...t}){let{router:n}=UL(),r=g.useMemo(()=>jL(n.routes,e,n.basename),[n.routes,e,n.basename]);return r?g.createElement(wG,{page:e,matches:r,...t}):null}function xG(e){let{manifest:t,routeModules:n}=VL(),[r,i]=g.useState([]);return g.useEffect(()=>{let a=!1;return cG(e,t,n).then(s=>{a||i(s)}),()=>{a=!0}},[e,t,n]),r}function wG({page:e,matches:t,...n}){let r=ms(),{manifest:i,routeModules:a}=VL(),{basename:s}=UL(),{loaderData:u,matches:c}=gG(),d=g.useMemo(()=>tP(e,t,c,i,r,"data"),[e,t,c,i,r]),p=g.useMemo(()=>tP(e,t,c,i,r,"assets"),[e,t,c,i,r]),m=g.useMemo(()=>{if(e===r.pathname+r.search+r.hash)return[];let S=new Set,w=!1;if(t.forEach(C=>{var A;let O=i.routes[C.route.id];!O||!O.hasLoader||(!d.some(T=>T.route.id===C.route.id)&&C.route.id in u&&((A=a[C.route.id])!=null&&A.shouldRevalidate)||O.hasClientLoader?w=!0:S.add(C.route.id))}),S.size===0)return[];let x=vG(e,s);return w&&S.size>0&&x.searchParams.set("_routes",t.filter(C=>S.has(C.route.id)).map(C=>C.route.id).join(",")),[x.pathname+x.search]},[s,u,r,i,d,t,e,a]),v=g.useMemo(()=>fG(p,i),[p,i]),b=xG(p);return g.createElement(g.Fragment,null,m.map(S=>g.createElement("link",{key:S,rel:"prefetch",as:"fetch",href:S,...n})),v.map(S=>g.createElement("link",{key:S,rel:"modulepreload",href:S,...n})),b.map(({key:S,link:w})=>g.createElement("link",{key:S,...w})))}function SG(...e){return t=>{e.forEach(n=>{typeof n=="function"?n(t):n!=null&&(n.current=t)})}}var HL=typeof window<"u"&&typeof window.document<"u"&&typeof window.document.createElement<"u";try{HL&&(window.__reactRouterVersion="7.5.3")}catch{}function _G({basename:e,children:t,window:n}){let r=g.useRef();r.current==null&&(r.current=f9({window:n,v5Compat:!0}));let i=r.current,[a,s]=g.useState({action:i.action,location:i.location}),u=g.useCallback(c=>{g.startTransition(()=>s(c))},[s]);return g.useLayoutEffect(()=>i.listen(u),[i,u]),g.createElement(Q9,{basename:e,children:t,location:a.location,navigationType:a.action,navigator:i})}var GL=/^(?:[a-z][a-z0-9+.-]*:|\/\/)/i,aa=g.forwardRef(function({onClick:t,discover:n="render",prefetch:r="none",relative:i,reloadDocument:a,replace:s,state:u,target:c,to:d,preventScrollReset:p,viewTransition:m,...v},b){let{basename:S}=g.useContext(Ai),w=typeof d=="string"&&GL.test(d),x,C=!1;if(typeof d=="string"&&w&&(x=d,HL))try{let k=new URL(window.location.href),H=d.startsWith("//")?new URL(k.protocol+d):new URL(d),L=ca(H.pathname,S);H.origin===k.origin&&L!=null?d=L+H.search+H.hash:C=!0}catch{Si(!1,`<Link to="${d}"> contains an invalid URL which will probably break when clicked - please update to a valid URL path.`)}let O=I9(d,{relative:i}),[A,T,R]=yG(r,v),M=OG(d,{replace:s,state:u,target:c,preventScrollReset:p,relative:i,viewTransition:m});function P(k){t&&t(k),k.defaultPrevented||M(k)}let N=g.createElement("a",{...v,...R,href:x||O,onClick:C||a?t:P,ref:SG(b,T),target:c,"data-discover":!w&&n==="render"?"true":void 0});return A&&!w?g.createElement(g.Fragment,null,N,g.createElement(bG,{page:O})):N});aa.displayName="Link";var WL=g.forwardRef(function({"aria-current":t="page",caseSensitive:n=!1,className:r="",end:i=!1,style:a,to:s,viewTransition:u,children:c,...d},p){let m=od(s,{relative:d.relative}),v=ms(),b=g.useContext(Xm),{navigator:S,basename:w}=g.useContext(Ai),x=b!=null&&PG(m)&&u===!0,C=S.encodeLocation?S.encodeLocation(m).pathname:m.pathname,O=v.pathname,A=b&&b.navigation&&b.navigation.location?b.navigation.location.pathname:null;n||(O=O.toLowerCase(),A=A?A.toLowerCase():null,C=C.toLowerCase()),A&&w&&(A=ca(A,w)||A);const T=C!=="/"&&C.endsWith("/")?C.length-1:C.length;let R=O===C||!i&&O.startsWith(C)&&O.charAt(T)==="/",M=A!=null&&(A===C||!i&&A.startsWith(C)&&A.charAt(C.length)==="/"),P={isActive:R,isPending:M,isTransitioning:x},N=R?t:void 0,k;typeof r=="function"?k=r(P):k=[r,R?"active":null,M?"pending":null,x?"transitioning":null].filter(Boolean).join(" ");let H=typeof a=="function"?a(P):a;return g.createElement(aa,{...d,"aria-current":N,className:k,ref:p,style:H,to:s,viewTransition:u},typeof c=="function"?c(P):c)});WL.displayName="NavLink";var EG=g.forwardRef(({discover:e="render",fetcherKey:t,navigate:n,reloadDocument:r,replace:i,state:a,method:s=yh,action:u,onSubmit:c,relative:d,preventScrollReset:p,viewTransition:m,...v},b)=>{let S=TG(),w=MG(u,{relative:d}),x=s.toLowerCase()==="get"?"get":"post",C=typeof u=="string"&&GL.test(u),O=A=>{if(c&&c(A),A.defaultPrevented)return;A.preventDefault();let T=A.nativeEvent.submitter,R=(T==null?void 0:T.getAttribute("formmethod"))||s;S(T||A.currentTarget,{fetcherKey:t,method:R,navigate:n,replace:i,state:a,relative:d,preventScrollReset:p,viewTransition:m})};return g.createElement("form",{ref:b,method:x,action:w,onSubmit:r?c:O,...v,"data-discover":!C&&e==="render"?"true":void 0})});EG.displayName="Form";function CG(e){return`${e} must be used within a data router.  See https://reactrouter.com/en/main/routers/picking-a-router.`}function KL(e){let t=g.useContext(lu);return zt(t,CG(e)),t}function OG(e,{target:t,replace:n,state:r,preventScrollReset:i,relative:a,viewTransition:s}={}){let u=kE(),c=ms(),d=od(e,{relative:a});return g.useCallback(p=>{if(iG(p,t)){p.preventDefault();let m=n!==void 0?n:Jc(c)===Jc(d);u(e,{replace:m,state:r,preventScrollReset:i,relative:a,viewTransition:s})}},[c,u,d,n,r,t,e,i,a,s])}var AG=0,RG=()=>`__${String(++AG)}__`;function TG(){let{router:e}=KL("useSubmit"),{basename:t}=g.useContext(Ai),n=K9();return g.useCallback(async(r,i={})=>{let{action:a,method:s,encType:u,formData:c,body:d}=sG(r,t);if(i.navigate===!1){let p=i.fetcherKey||RG();await e.fetch(p,n,i.action||a,{preventScrollReset:i.preventScrollReset,formData:c,body:d,formMethod:i.method||s,formEncType:i.encType||u,flushSync:i.flushSync})}else await e.navigate(i.action||a,{preventScrollReset:i.preventScrollReset,formData:c,body:d,formMethod:i.method||s,formEncType:i.encType||u,replace:i.replace,state:i.state,fromRouteId:n,flushSync:i.flushSync,viewTransition:i.viewTransition})},[e,t,n])}function MG(e,{relative:t}={}){let{basename:n}=g.useContext(Ai),r=g.useContext(Ri);zt(r,"useFormAction must be used inside a RouteContext");let[i]=r.matches.slice(-1),a={...od(e||".",{relative:t})},s=ms();if(e==null){a.search=s.search;let u=new URLSearchParams(a.search),c=u.getAll("index");if(c.some(p=>p==="")){u.delete("index"),c.filter(m=>m).forEach(m=>u.append("index",m));let p=u.toString();a.search=p?`?${p}`:""}}return(!e||e===".")&&i.route.index&&(a.search=a.search?a.search.replace(/^\?/,"?index&"):"?index"),n!=="/"&&(a.pathname=a.pathname==="/"?n:ia([n,a.pathname])),Jc(a)}function PG(e,t={}){let n=g.useContext(LL);zt(n!=null,"`useViewTransitionState` must be used within `react-router-dom`'s `RouterProvider`.  Did you accidentally import `RouterProvider` from `react-router`?");let{basename:r}=KL("useViewTransitionState"),i=od(e,{relative:t.relative});if(!n.isTransitioning)return!1;let a=ca(n.currentLocation.pathname,r)||n.currentLocation.pathname,s=ca(n.nextLocation.pathname,r)||n.nextLocation.pathname;return Th(i.pathname,s)!=null||Th(i.pathname,a)!=null}new TextEncoder;[...mG];function jG(){const e=[{name:"Classify",href:"classify",external:!1},{name:"Filter",href:"filter",external:!1},{name:"Models",href:"models",external:!1},{name:"Documentation",href:"https://bionf.github.io/XspecT/index.html",external:!0}];return E.jsx("header",{className:"bg-white shadow-sm",children:E.jsx("div",{className:"container mx-auto px-4 py-4",children:E.jsxs("div",{className:"flex items-center justify-between",children:[E.jsx("div",{className:"flex items-center text-xl font-bold",children:E.jsx(aa,{to:"/",children:"XspecT"})}),E.jsx(o9,{children:E.jsx(s9,{children:e.map(t=>E.jsx(l9,{className:"text-gray-700 hover:text-gray-900 font-medium",children:E.jsx(WL,{className:"block px-4 py-2 rounded-md transition-colors duration-200",to:t.href,...t.external?{onClick:n=>{n.preventDefault(),window.open(t.href,"_blank")}}:{},children:t.name})},t.name))})})]})})})}function NG(){return E.jsx("main",{className:"flex-1 flex items-center justify-center",children:E.jsx("h1",{className:"text-4xl md:text-6xl font-bold text-center",children:"Welcome to XspecT"})})}var sd=e=>e.type==="checkbox",Ho=e=>e instanceof Date,Bn=e=>e==null;const YL=e=>typeof e=="object";var Xt=e=>!Bn(e)&&!Array.isArray(e)&&YL(e)&&!Ho(e),ZL=e=>Xt(e)&&e.target?sd(e.target)?e.target.checked:e.target.value:e,DG=e=>e.substring(0,e.search(/\.\d+(\.|$)/))||e,XL=(e,t)=>e.has(DG(t)),kG=e=>{const t=e.constructor&&e.constructor.prototype;return Xt(t)&&t.hasOwnProperty("isPrototypeOf")},qE=typeof window<"u"&&typeof window.HTMLElement<"u"&&typeof document<"u";function _n(e){let t;const n=Array.isArray(e),r=typeof FileList<"u"?e instanceof FileList:!1;if(e instanceof Date)t=new Date(e);else if(e instanceof Set)t=new Set(e);else if(!(qE&&(e instanceof Blob||r))&&(n||Xt(e)))if(t=n?[]:{},!n&&!kG(e))t=e;else for(const i in e)e.hasOwnProperty(i)&&(t[i]=_n(e[i]));else return e;return t}var Jm=e=>Array.isArray(e)?e.filter(Boolean):[],Zt=e=>e===void 0,Ce=(e,t,n)=>{if(!t||!Xt(e))return n;const r=Jm(t.split(/[,[\].]+?/)).reduce((i,a)=>Bn(i)?i:i[a],e);return Zt(r)||r===e?Zt(e[t])?n:e[t]:r},dr=e=>typeof e=="boolean",zE=e=>/^\w*$/.test(e),QL=e=>Jm(e.replace(/["|']|\]/g,"").split(/\.|\[/)),bt=(e,t,n)=>{let r=-1;const i=zE(t)?[t]:QL(t),a=i.length,s=a-1;for(;++r<a;){const u=i[r];let c=n;if(r!==s){const d=e[u];c=Xt(d)||Array.isArray(d)?d:isNaN(+i[r+1])?{}:[]}if(u==="__proto__"||u==="constructor"||u==="prototype")return;e[u]=c,e=e[u]}};const Mh={BLUR:"blur",FOCUS_OUT:"focusout",CHANGE:"change"},Hr={onBlur:"onBlur",onChange:"onChange",onSubmit:"onSubmit",onTouched:"onTouched",all:"all"},Yi={max:"max",min:"min",maxLength:"maxLength",minLength:"minLength",pattern:"pattern",required:"required",validate:"validate"},JL=V.createContext(null),ev=()=>V.useContext(JL),$G=e=>{const{children:t,...n}=e;return V.createElement(JL.Provider,{value:n},t)};var eB=(e,t,n,r=!0)=>{const i={defaultValues:t._defaultValues};for(const a in e)Object.defineProperty(i,a,{get:()=>{const s=a;return t._proxyFormState[s]!==Hr.all&&(t._proxyFormState[s]=!r||Hr.all),n&&(n[s]=!0),e[s]}});return i};const UE=typeof window<"u"?g.useLayoutEffect:g.useEffect;function tB(e){const t=ev(),{control:n=t.control,disabled:r,name:i,exact:a}=e||{},[s,u]=V.useState(n._formState),c=V.useRef({isDirty:!1,isLoading:!1,dirtyFields:!1,touchedFields:!1,validatingFields:!1,isValidating:!1,isValid:!1,errors:!1});return UE(()=>n._subscribe({name:i,formState:c.current,exact:a,callback:d=>{!r&&u({...n._formState,...d})}}),[i,r,a]),V.useEffect(()=>{c.current.isValid&&n._setValid(!0)},[n]),V.useMemo(()=>eB(s,n,c.current,!1),[s,n])}var pi=e=>typeof e=="string",nB=(e,t,n,r,i)=>pi(e)?(r&&t.watch.add(e),Ce(n,e,i)):Array.isArray(e)?e.map(a=>(r&&t.watch.add(a),Ce(n,a))):(r&&(t.watchAll=!0),n);function IG(e){const t=ev(),{control:n=t.control,name:r,defaultValue:i,disabled:a,exact:s}=e||{},u=V.useRef(i),[c,d]=V.useState(n._getWatch(r,u.current));return UE(()=>n._subscribe({name:r,formState:{values:!0},exact:s,callback:p=>!a&&d(nB(r,n._names,p.values||n._formValues,!1,u.current))}),[r,n,a,s]),V.useEffect(()=>n._removeUnmounted()),c}function LG(e){const t=ev(),{name:n,disabled:r,control:i=t.control,shouldUnregister:a}=e,s=XL(i._names.array,n),u=IG({control:i,name:n,defaultValue:Ce(i._formValues,n,Ce(i._defaultValues,n,e.defaultValue)),exact:!0}),c=tB({control:i,name:n,exact:!0}),d=V.useRef(e),p=V.useRef(i.register(n,{...e.rules,value:u,...dr(e.disabled)?{disabled:e.disabled}:{}})),m=V.useMemo(()=>Object.defineProperties({},{invalid:{enumerable:!0,get:()=>!!Ce(c.errors,n)},isDirty:{enumerable:!0,get:()=>!!Ce(c.dirtyFields,n)},isTouched:{enumerable:!0,get:()=>!!Ce(c.touchedFields,n)},isValidating:{enumerable:!0,get:()=>!!Ce(c.validatingFields,n)},error:{enumerable:!0,get:()=>Ce(c.errors,n)}}),[c,n]),v=V.useCallback(x=>p.current.onChange({target:{value:ZL(x),name:n},type:Mh.CHANGE}),[n]),b=V.useCallback(()=>p.current.onBlur({target:{value:Ce(i._formValues,n),name:n},type:Mh.BLUR}),[n,i._formValues]),S=V.useCallback(x=>{const C=Ce(i._fields,n);C&&x&&(C._f.ref=x)},[i._fields,n]),w=V.useMemo(()=>({name:n,value:u,...dr(r)||c.disabled?{disabled:c.disabled||r}:{},onChange:v,onBlur:b,ref:S}),[n,r,c.disabled,v,b,S,u]);return V.useEffect(()=>{const x=i._options.shouldUnregister||a;i.register(n,{...d.current.rules,...dr(d.current.disabled)?{disabled:d.current.disabled}:{}});const C=(O,A)=>{const T=Ce(i._fields,O);T&&T._f&&(T._f.mount=A)};if(C(n,!0),x){const O=_n(Ce(i._options.defaultValues,n));bt(i._defaultValues,n,O),Zt(Ce(i._formValues,n))&&bt(i._formValues,n,O)}return!s&&i.register(n),()=>{(s?x&&!i._state.action:x)?i.unregister(n):C(n,!1)}},[n,i,s,a]),V.useEffect(()=>{i._setDisabledField({disabled:r,name:n})},[r,n,i]),V.useMemo(()=>({field:w,formState:c,fieldState:m}),[w,c,m])}const BG=e=>e.render(LG(e));var rB=(e,t,n,r,i)=>t?{...n[e],types:{...n[e]&&n[e].types?n[e].types:{},[r]:i||!0}}:{},Hc=e=>Array.isArray(e)?e:[e],nP=()=>{let e=[];return{get observers(){return e},next:i=>{for(const a of e)a.next&&a.next(i)},subscribe:i=>(e.push(i),{unsubscribe:()=>{e=e.filter(a=>a!==i)}}),unsubscribe:()=>{e=[]}}},E_=e=>Bn(e)||!YL(e);function Ja(e,t){if(E_(e)||E_(t))return e===t;if(Ho(e)&&Ho(t))return e.getTime()===t.getTime();const n=Object.keys(e),r=Object.keys(t);if(n.length!==r.length)return!1;for(const i of n){const a=e[i];if(!r.includes(i))return!1;if(i!=="ref"){const s=t[i];if(Ho(a)&&Ho(s)||Xt(a)&&Xt(s)||Array.isArray(a)&&Array.isArray(s)?!Ja(a,s):a!==s)return!1}}return!0}var In=e=>Xt(e)&&!Object.keys(e).length,VE=e=>e.type==="file",Gr=e=>typeof e=="function",Ph=e=>{if(!qE)return!1;const t=e?e.ownerDocument:0;return e instanceof(t&&t.defaultView?t.defaultView.HTMLElement:HTMLElement)},iB=e=>e.type==="select-multiple",HE=e=>e.type==="radio",FG=e=>HE(e)||sd(e),Lb=e=>Ph(e)&&e.isConnected;function qG(e,t){const n=t.slice(0,-1).length;let r=0;for(;r<n;)e=Zt(e)?r++:e[t[r++]];return e}function zG(e){for(const t in e)if(e.hasOwnProperty(t)&&!Zt(e[t]))return!1;return!0}function on(e,t){const n=Array.isArray(t)?t:zE(t)?[t]:QL(t),r=n.length===1?e:qG(e,n),i=n.length-1,a=n[i];return r&&delete r[a],i!==0&&(Xt(r)&&In(r)||Array.isArray(r)&&zG(r))&&on(e,n.slice(0,-1)),e}var aB=e=>{for(const t in e)if(Gr(e[t]))return!0;return!1};function jh(e,t={}){const n=Array.isArray(e);if(Xt(e)||n)for(const r in e)Array.isArray(e[r])||Xt(e[r])&&!aB(e[r])?(t[r]=Array.isArray(e[r])?[]:{},jh(e[r],t[r])):Bn(e[r])||(t[r]=!0);return t}function oB(e,t,n){const r=Array.isArray(e);if(Xt(e)||r)for(const i in e)Array.isArray(e[i])||Xt(e[i])&&!aB(e[i])?Zt(t)||E_(n[i])?n[i]=Array.isArray(e[i])?jh(e[i],[]):{...jh(e[i])}:oB(e[i],Bn(t)?{}:t[i],n[i]):n[i]=!Ja(e[i],t[i]);return n}var Ec=(e,t)=>oB(e,t,jh(t));const rP={value:!1,isValid:!1},iP={value:!0,isValid:!0};var sB=e=>{if(Array.isArray(e)){if(e.length>1){const t=e.filter(n=>n&&n.checked&&!n.disabled).map(n=>n.value);return{value:t,isValid:!!t.length}}return e[0].checked&&!e[0].disabled?e[0].attributes&&!Zt(e[0].attributes.value)?Zt(e[0].value)||e[0].value===""?iP:{value:e[0].value,isValid:!0}:iP:rP}return rP},lB=(e,{valueAsNumber:t,valueAsDate:n,setValueAs:r})=>Zt(e)?e:t?e===""?NaN:e&&+e:n&&pi(e)?new Date(e):r?r(e):e;const aP={isValid:!1,value:null};var uB=e=>Array.isArray(e)?e.reduce((t,n)=>n&&n.checked&&!n.disabled?{isValid:!0,value:n.value}:t,aP):aP;function oP(e){const t=e.ref;return VE(t)?t.files:HE(t)?uB(e.refs).value:iB(t)?[...t.selectedOptions].map(({value:n})=>n):sd(t)?sB(e.refs).value:lB(Zt(t.value)?e.ref.value:t.value,e)}var UG=(e,t,n,r)=>{const i={};for(const a of e){const s=Ce(t,a);s&&bt(i,a,s._f)}return{criteriaMode:n,names:[...e],fields:i,shouldUseNativeValidation:r}},Nh=e=>e instanceof RegExp,Cc=e=>Zt(e)?e:Nh(e)?e.source:Xt(e)?Nh(e.value)?e.value.source:e.value:e,sP=e=>({isOnSubmit:!e||e===Hr.onSubmit,isOnBlur:e===Hr.onBlur,isOnChange:e===Hr.onChange,isOnAll:e===Hr.all,isOnTouch:e===Hr.onTouched});const lP="AsyncFunction";var VG=e=>!!e&&!!e.validate&&!!(Gr(e.validate)&&e.validate.constructor.name===lP||Xt(e.validate)&&Object.values(e.validate).find(t=>t.constructor.name===lP)),HG=e=>e.mount&&(e.required||e.min||e.max||e.maxLength||e.minLength||e.pattern||e.validate),uP=(e,t,n)=>!n&&(t.watchAll||t.watch.has(e)||[...t.watch].some(r=>e.startsWith(r)&&/^\.\w+/.test(e.slice(r.length))));const Gc=(e,t,n,r)=>{for(const i of n||Object.keys(e)){const a=Ce(e,i);if(a){const{_f:s,...u}=a;if(s){if(s.refs&&s.refs[0]&&t(s.refs[0],i)&&!r)return!0;if(s.ref&&t(s.ref,s.name)&&!r)return!0;if(Gc(u,t))break}else if(Xt(u)&&Gc(u,t))break}}};function cP(e,t,n){const r=Ce(e,n);if(r||zE(n))return{error:r,name:n};const i=n.split(".");for(;i.length;){const a=i.join("."),s=Ce(t,a),u=Ce(e,a);if(s&&!Array.isArray(s)&&n!==a)return{name:n};if(u&&u.type)return{name:a,error:u};i.pop()}return{name:n}}var GG=(e,t,n,r)=>{n(e);const{name:i,...a}=e;return In(a)||Object.keys(a).length>=Object.keys(t).length||Object.keys(a).find(s=>t[s]===(!r||Hr.all))},WG=(e,t,n)=>!e||!t||e===t||Hc(e).some(r=>r&&(n?r===t:r.startsWith(t)||t.startsWith(r))),KG=(e,t,n,r,i)=>i.isOnAll?!1:!n&&i.isOnTouch?!(t||e):(n?r.isOnBlur:i.isOnBlur)?!e:(n?r.isOnChange:i.isOnChange)?e:!0,YG=(e,t)=>!Jm(Ce(e,t)).length&&on(e,t),ZG=(e,t,n)=>{const r=Hc(Ce(e,n));return bt(r,"root",t[n]),bt(e,n,r),e},xh=e=>pi(e);function fP(e,t,n="validate"){if(xh(e)||Array.isArray(e)&&e.every(xh)||dr(e)&&!e)return{type:n,message:xh(e)?e:"",ref:t}}var cl=e=>Xt(e)&&!Nh(e)?e:{value:e,message:""},dP=async(e,t,n,r,i,a)=>{const{ref:s,refs:u,required:c,maxLength:d,minLength:p,min:m,max:v,pattern:b,validate:S,name:w,valueAsNumber:x,mount:C}=e._f,O=Ce(n,w);if(!C||t.has(w))return{};const A=u?u[0]:s,T=F=>{i&&A.reportValidity&&(A.setCustomValidity(dr(F)?"":F||""),A.reportValidity())},R={},M=HE(s),P=sd(s),N=M||P,k=(x||VE(s))&&Zt(s.value)&&Zt(O)||Ph(s)&&s.value===""||O===""||Array.isArray(O)&&!O.length,H=rB.bind(null,w,r,R),L=(F,z,Z,Y=Yi.maxLength,I=Yi.minLength)=>{const W=F?z:Z;R[w]={type:F?Y:I,message:W,ref:s,...H(F?Y:I,W)}};if(a?!Array.isArray(O)||!O.length:c&&(!N&&(k||Bn(O))||dr(O)&&!O||P&&!sB(u).isValid||M&&!uB(u).isValid)){const{value:F,message:z}=xh(c)?{value:!!c,message:c}:cl(c);if(F&&(R[w]={type:Yi.required,message:z,ref:A,...H(Yi.required,z)},!r))return T(z),R}if(!k&&(!Bn(m)||!Bn(v))){let F,z;const Z=cl(v),Y=cl(m);if(!Bn(O)&&!isNaN(O)){const I=s.valueAsNumber||O&&+O;Bn(Z.value)||(F=I>Z.value),Bn(Y.value)||(z=I<Y.value)}else{const I=s.valueAsDate||new Date(O),W=$=>new Date(new Date().toDateString()+" "+$),B=s.type=="time",Q=s.type=="week";pi(Z.value)&&O&&(F=B?W(O)>W(Z.value):Q?O>Z.value:I>new Date(Z.value)),pi(Y.value)&&O&&(z=B?W(O)<W(Y.value):Q?O<Y.value:I<new Date(Y.value))}if((F||z)&&(L(!!F,Z.message,Y.message,Yi.max,Yi.min),!r))return T(R[w].message),R}if((d||p)&&!k&&(pi(O)||a&&Array.isArray(O))){const F=cl(d),z=cl(p),Z=!Bn(F.value)&&O.length>+F.value,Y=!Bn(z.value)&&O.length<+z.value;if((Z||Y)&&(L(Z,F.message,z.message),!r))return T(R[w].message),R}if(b&&!k&&pi(O)){const{value:F,message:z}=cl(b);if(Nh(F)&&!O.match(F)&&(R[w]={type:Yi.pattern,message:z,ref:s,...H(Yi.pattern,z)},!r))return T(z),R}if(S){if(Gr(S)){const F=await S(O,n),z=fP(F,A);if(z&&(R[w]={...z,...H(Yi.validate,z.message)},!r))return T(z.message),R}else if(Xt(S)){let F={};for(const z in S){if(!In(F)&&!r)break;const Z=fP(await S[z](O,n),A,z);Z&&(F={...Z,...H(z,Z.message)},T(Z.message),r&&(R[w]=F))}if(!In(F)&&(R[w]={ref:A,...F},!r))return R}}return T(!0),R};const XG={mode:Hr.onSubmit,reValidateMode:Hr.onChange,shouldFocusError:!0};function QG(e={}){let t={...XG,...e},n={submitCount:0,isDirty:!1,isReady:!1,isLoading:Gr(t.defaultValues),isValidating:!1,isSubmitted:!1,isSubmitting:!1,isSubmitSuccessful:!1,isValid:!1,touchedFields:{},dirtyFields:{},validatingFields:{},errors:t.errors||{},disabled:t.disabled||!1};const r={};let i=Xt(t.defaultValues)||Xt(t.values)?_n(t.defaultValues||t.values)||{}:{},a=t.shouldUnregister?{}:_n(i),s={action:!1,mount:!1,watch:!1},u={mount:new Set,disabled:new Set,unMount:new Set,array:new Set,watch:new Set},c,d=0;const p={isDirty:!1,dirtyFields:!1,validatingFields:!1,touchedFields:!1,isValidating:!1,isValid:!1,errors:!1};let m={...p};const v={array:nP(),state:nP()},b=sP(t.mode),S=sP(t.reValidateMode),w=t.criteriaMode===Hr.all,x=G=>re=>{clearTimeout(d),d=setTimeout(G,re)},C=async G=>{if(!t.disabled&&(p.isValid||m.isValid||G)){const re=t.resolver?In((await k()).errors):await L(r,!0);re!==n.isValid&&v.state.next({isValid:re})}},O=(G,re)=>{!t.disabled&&(p.isValidating||p.validatingFields||m.isValidating||m.validatingFields)&&((G||Array.from(u.mount)).forEach(fe=>{fe&&(re?bt(n.validatingFields,fe,re):on(n.validatingFields,fe))}),v.state.next({validatingFields:n.validatingFields,isValidating:!In(n.validatingFields)}))},A=(G,re=[],fe,Ae,Ee=!0,ye=!0)=>{if(Ae&&fe&&!t.disabled){if(s.action=!0,ye&&Array.isArray(Ce(r,G))){const Te=fe(Ce(r,G),Ae.argA,Ae.argB);Ee&&bt(r,G,Te)}if(ye&&Array.isArray(Ce(n.errors,G))){const Te=fe(Ce(n.errors,G),Ae.argA,Ae.argB);Ee&&bt(n.errors,G,Te),YG(n.errors,G)}if((p.touchedFields||m.touchedFields)&&ye&&Array.isArray(Ce(n.touchedFields,G))){const Te=fe(Ce(n.touchedFields,G),Ae.argA,Ae.argB);Ee&&bt(n.touchedFields,G,Te)}(p.dirtyFields||m.dirtyFields)&&(n.dirtyFields=Ec(i,a)),v.state.next({name:G,isDirty:z(G,re),dirtyFields:n.dirtyFields,errors:n.errors,isValid:n.isValid})}else bt(a,G,re)},T=(G,re)=>{bt(n.errors,G,re),v.state.next({errors:n.errors})},R=G=>{n.errors=G,v.state.next({errors:n.errors,isValid:!1})},M=(G,re,fe,Ae)=>{const Ee=Ce(r,G);if(Ee){const ye=Ce(a,G,Zt(fe)?Ce(i,G):fe);Zt(ye)||Ae&&Ae.defaultChecked||re?bt(a,G,re?ye:oP(Ee._f)):I(G,ye),s.mount&&C()}},P=(G,re,fe,Ae,Ee)=>{let ye=!1,Te=!1;const Ke={name:G};if(!t.disabled){if(!fe||Ae){(p.isDirty||m.isDirty)&&(Te=n.isDirty,n.isDirty=Ke.isDirty=z(),ye=Te!==Ke.isDirty);const Vt=Ja(Ce(i,G),re);Te=!!Ce(n.dirtyFields,G),Vt?on(n.dirtyFields,G):bt(n.dirtyFields,G,!0),Ke.dirtyFields=n.dirtyFields,ye=ye||(p.dirtyFields||m.dirtyFields)&&Te!==!Vt}if(fe){const Vt=Ce(n.touchedFields,G);Vt||(bt(n.touchedFields,G,fe),Ke.touchedFields=n.touchedFields,ye=ye||(p.touchedFields||m.touchedFields)&&Vt!==fe)}ye&&Ee&&v.state.next(Ke)}return ye?Ke:{}},N=(G,re,fe,Ae)=>{const Ee=Ce(n.errors,G),ye=(p.isValid||m.isValid)&&dr(re)&&n.isValid!==re;if(t.delayError&&fe?(c=x(()=>T(G,fe)),c(t.delayError)):(clearTimeout(d),c=null,fe?bt(n.errors,G,fe):on(n.errors,G)),(fe?!Ja(Ee,fe):Ee)||!In(Ae)||ye){const Te={...Ae,...ye&&dr(re)?{isValid:re}:{},errors:n.errors,name:G};n={...n,...Te},v.state.next(Te)}},k=async G=>{O(G,!0);const re=await t.resolver(a,t.context,UG(G||u.mount,r,t.criteriaMode,t.shouldUseNativeValidation));return O(G),re},H=async G=>{const{errors:re}=await k(G);if(G)for(const fe of G){const Ae=Ce(re,fe);Ae?bt(n.errors,fe,Ae):on(n.errors,fe)}else n.errors=re;return re},L=async(G,re,fe={valid:!0})=>{for(const Ae in G){const Ee=G[Ae];if(Ee){const{_f:ye,...Te}=Ee;if(ye){const Ke=u.array.has(ye.name),Vt=Ee._f&&VG(Ee._f);Vt&&p.validatingFields&&O([Ae],!0);const dn=await dP(Ee,u.disabled,a,w,t.shouldUseNativeValidation&&!re,Ke);if(Vt&&p.validatingFields&&O([Ae]),dn[ye.name]&&(fe.valid=!1,re))break;!re&&(Ce(dn,ye.name)?Ke?ZG(n.errors,dn,ye.name):bt(n.errors,ye.name,dn[ye.name]):on(n.errors,ye.name))}!In(Te)&&await L(Te,re,fe)}}return fe.valid},F=()=>{for(const G of u.unMount){const re=Ce(r,G);re&&(re._f.refs?re._f.refs.every(fe=>!Lb(fe)):!Lb(re._f.ref))&&ce(G)}u.unMount=new Set},z=(G,re)=>!t.disabled&&(G&&re&&bt(a,G,re),!Ja(q(),i)),Z=(G,re,fe)=>nB(G,u,{...s.mount?a:Zt(re)?i:pi(G)?{[G]:re}:re},fe,re),Y=G=>Jm(Ce(s.mount?a:i,G,t.shouldUnregister?Ce(i,G,[]):[])),I=(G,re,fe={})=>{const Ae=Ce(r,G);let Ee=re;if(Ae){const ye=Ae._f;ye&&(!ye.disabled&&bt(a,G,lB(re,ye)),Ee=Ph(ye.ref)&&Bn(re)?"":re,iB(ye.ref)?[...ye.ref.options].forEach(Te=>Te.selected=Ee.includes(Te.value)):ye.refs?sd(ye.ref)?ye.refs.length>1?ye.refs.forEach(Te=>(!Te.defaultChecked||!Te.disabled)&&(Te.checked=Array.isArray(Ee)?!!Ee.find(Ke=>Ke===Te.value):Ee===Te.value)):ye.refs[0]&&(ye.refs[0].checked=!!Ee):ye.refs.forEach(Te=>Te.checked=Te.value===Ee):VE(ye.ref)?ye.ref.value="":(ye.ref.value=Ee,ye.ref.type||v.state.next({name:G,values:_n(a)})))}(fe.shouldDirty||fe.shouldTouch)&&P(G,Ee,fe.shouldTouch,fe.shouldDirty,!0),fe.shouldValidate&&J(G)},W=(G,re,fe)=>{for(const Ae in re){const Ee=re[Ae],ye=`${G}.${Ae}`,Te=Ce(r,ye);(u.array.has(G)||Xt(Ee)||Te&&!Te._f)&&!Ho(Ee)?W(ye,Ee,fe):I(ye,Ee,fe)}},B=(G,re,fe={})=>{const Ae=Ce(r,G),Ee=u.array.has(G),ye=_n(re);bt(a,G,ye),Ee?(v.array.next({name:G,values:_n(a)}),(p.isDirty||p.dirtyFields||m.isDirty||m.dirtyFields)&&fe.shouldDirty&&v.state.next({name:G,dirtyFields:Ec(i,a),isDirty:z(G,ye)})):Ae&&!Ae._f&&!Bn(ye)?W(G,ye,fe):I(G,ye,fe),uP(G,u)&&v.state.next({...n}),v.state.next({name:s.mount?G:void 0,values:_n(a)})},Q=async G=>{s.mount=!0;const re=G.target;let fe=re.name,Ae=!0;const Ee=Ce(r,fe),ye=Te=>{Ae=Number.isNaN(Te)||Ho(Te)&&isNaN(Te.getTime())||Ja(Te,Ce(a,fe,Te))};if(Ee){let Te,Ke;const Vt=re.type?oP(Ee._f):ZL(G),dn=G.type===Mh.BLUR||G.type===Mh.FOCUS_OUT,vg=!HG(Ee._f)&&!t.resolver&&!Ce(n.errors,fe)&&!Ee._f.deps||KG(dn,Ce(n.touchedFields,fe),n.isSubmitted,S,b),_a=uP(fe,u,dn);bt(a,fe,Vt),dn?(Ee._f.onBlur&&Ee._f.onBlur(G),c&&c(0)):Ee._f.onChange&&Ee._f.onChange(G);const Ea=P(fe,Vt,dn),ji=!In(Ea)||_a;if(!dn&&v.state.next({name:fe,type:G.type,values:_n(a)}),vg)return(p.isValid||m.isValid)&&(t.mode==="onBlur"?dn&&C():dn||C()),ji&&v.state.next({name:fe,..._a?{}:Ea});if(!dn&&_a&&v.state.next({...n}),t.resolver){const{errors:bo}=await k([fe]);if(ye(Vt),Ae){const xo=cP(n.errors,r,fe),jd=cP(bo,r,xo.name||fe);Te=jd.error,fe=jd.name,Ke=In(bo)}}else O([fe],!0),Te=(await dP(Ee,u.disabled,a,w,t.shouldUseNativeValidation))[fe],O([fe]),ye(Vt),Ae&&(Te?Ke=!1:(p.isValid||m.isValid)&&(Ke=await L(r,!0)));Ae&&(Ee._f.deps&&J(Ee._f.deps),N(fe,Ke,Te,Ea))}},$=(G,re)=>{if(Ce(n.errors,re)&&G.focus)return G.focus(),1},J=async(G,re={})=>{let fe,Ae;const Ee=Hc(G);if(t.resolver){const ye=await H(Zt(G)?G:Ee);fe=In(ye),Ae=G?!Ee.some(Te=>Ce(ye,Te)):fe}else G?(Ae=(await Promise.all(Ee.map(async ye=>{const Te=Ce(r,ye);return await L(Te&&Te._f?{[ye]:Te}:Te)}))).every(Boolean),!(!Ae&&!n.isValid)&&C()):Ae=fe=await L(r);return v.state.next({...!pi(G)||(p.isValid||m.isValid)&&fe!==n.isValid?{}:{name:G},...t.resolver||!G?{isValid:fe}:{},errors:n.errors}),re.shouldFocus&&!Ae&&Gc(r,$,G?Ee:u.mount),Ae},q=G=>{const re={...s.mount?a:i};return Zt(G)?re:pi(G)?Ce(re,G):G.map(fe=>Ce(re,fe))},X=(G,re)=>({invalid:!!Ce((re||n).errors,G),isDirty:!!Ce((re||n).dirtyFields,G),error:Ce((re||n).errors,G),isValidating:!!Ce(n.validatingFields,G),isTouched:!!Ce((re||n).touchedFields,G)}),oe=G=>{G&&Hc(G).forEach(re=>on(n.errors,re)),v.state.next({errors:G?n.errors:{}})},le=(G,re,fe)=>{const Ae=(Ce(r,G,{_f:{}})._f||{}).ref,Ee=Ce(n.errors,G)||{},{ref:ye,message:Te,type:Ke,...Vt}=Ee;bt(n.errors,G,{...Vt,...re,ref:Ae}),v.state.next({name:G,errors:n.errors,isValid:!1}),fe&&fe.shouldFocus&&Ae&&Ae.focus&&Ae.focus()},ve=(G,re)=>Gr(G)?v.state.subscribe({next:fe=>G(Z(void 0,re),fe)}):Z(G,re,!0),de=G=>v.state.subscribe({next:re=>{WG(G.name,re.name,G.exact)&&GG(re,G.formState||p,Tt,G.reRenderRoot)&&G.callback({values:{...a},...n,...re})}}).unsubscribe,Se=G=>(s.mount=!0,m={...m,...G.formState},de({...G,formState:m})),ce=(G,re={})=>{for(const fe of G?Hc(G):u.mount)u.mount.delete(fe),u.array.delete(fe),re.keepValue||(on(r,fe),on(a,fe)),!re.keepError&&on(n.errors,fe),!re.keepDirty&&on(n.dirtyFields,fe),!re.keepTouched&&on(n.touchedFields,fe),!re.keepIsValidating&&on(n.validatingFields,fe),!t.shouldUnregister&&!re.keepDefaultValue&&on(i,fe);v.state.next({values:_n(a)}),v.state.next({...n,...re.keepDirty?{isDirty:z()}:{}}),!re.keepIsValid&&C()},we=({disabled:G,name:re})=>{(dr(G)&&s.mount||G||u.disabled.has(re))&&(G?u.disabled.add(re):u.disabled.delete(re))},_e=(G,re={})=>{let fe=Ce(r,G);const Ae=dr(re.disabled)||dr(t.disabled);return bt(r,G,{...fe||{},_f:{...fe&&fe._f?fe._f:{ref:{name:G}},name:G,mount:!0,...re}}),u.mount.add(G),fe?we({disabled:dr(re.disabled)?re.disabled:t.disabled,name:G}):M(G,!0,re.value),{...Ae?{disabled:re.disabled||t.disabled}:{},...t.progressive?{required:!!re.required,min:Cc(re.min),max:Cc(re.max),minLength:Cc(re.minLength),maxLength:Cc(re.maxLength),pattern:Cc(re.pattern)}:{},name:G,onChange:Q,onBlur:Q,ref:Ee=>{if(Ee){_e(G,re),fe=Ce(r,G);const ye=Zt(Ee.value)&&Ee.querySelectorAll&&Ee.querySelectorAll("input,select,textarea")[0]||Ee,Te=FG(ye),Ke=fe._f.refs||[];if(Te?Ke.find(Vt=>Vt===ye):ye===fe._f.ref)return;bt(r,G,{_f:{...fe._f,...Te?{refs:[...Ke.filter(Lb),ye,...Array.isArray(Ce(i,G))?[{}]:[]],ref:{type:ye.type,name:G}}:{ref:ye}}}),M(G,!1,void 0,ye)}else fe=Ce(r,G,{}),fe._f&&(fe._f.mount=!1),(t.shouldUnregister||re.shouldUnregister)&&!(XL(u.array,G)&&s.action)&&u.unMount.add(G)}}},se=()=>t.shouldFocusError&&Gc(r,$,u.mount),Ye=G=>{dr(G)&&(v.state.next({disabled:G}),Gc(r,(re,fe)=>{const Ae=Ce(r,fe);Ae&&(re.disabled=Ae._f.disabled||G,Array.isArray(Ae._f.refs)&&Ae._f.refs.forEach(Ee=>{Ee.disabled=Ae._f.disabled||G}))},0,!1))},Be=(G,re)=>async fe=>{let Ae;fe&&(fe.preventDefault&&fe.preventDefault(),fe.persist&&fe.persist());let Ee=_n(a);if(v.state.next({isSubmitting:!0}),t.resolver){const{errors:ye,values:Te}=await k();n.errors=ye,Ee=Te}else await L(r);if(u.disabled.size)for(const ye of u.disabled)bt(Ee,ye,void 0);if(on(n.errors,"root"),In(n.errors)){v.state.next({errors:{}});try{await G(Ee,fe)}catch(ye){Ae=ye}}else re&&await re({...n.errors},fe),se(),setTimeout(se);if(v.state.next({isSubmitted:!0,isSubmitting:!1,isSubmitSuccessful:In(n.errors)&&!Ae,submitCount:n.submitCount+1,errors:n.errors}),Ae)throw Ae},dt=(G,re={})=>{Ce(r,G)&&(Zt(re.defaultValue)?B(G,_n(Ce(i,G))):(B(G,re.defaultValue),bt(i,G,_n(re.defaultValue))),re.keepTouched||on(n.touchedFields,G),re.keepDirty||(on(n.dirtyFields,G),n.isDirty=re.defaultValue?z(G,_n(Ce(i,G))):z()),re.keepError||(on(n.errors,G),p.isValid&&C()),v.state.next({...n}))},ft=(G,re={})=>{const fe=G?_n(G):i,Ae=_n(fe),Ee=In(G),ye=Ee?i:Ae;if(re.keepDefaultValues||(i=fe),!re.keepValues){if(re.keepDirtyValues){const Te=new Set([...u.mount,...Object.keys(Ec(i,a))]);for(const Ke of Array.from(Te))Ce(n.dirtyFields,Ke)?bt(ye,Ke,Ce(a,Ke)):B(Ke,Ce(ye,Ke))}else{if(qE&&Zt(G))for(const Te of u.mount){const Ke=Ce(r,Te);if(Ke&&Ke._f){const Vt=Array.isArray(Ke._f.refs)?Ke._f.refs[0]:Ke._f.ref;if(Ph(Vt)){const dn=Vt.closest("form");if(dn){dn.reset();break}}}}for(const Te of u.mount)B(Te,Ce(ye,Te))}a=_n(ye),v.array.next({values:{...ye}}),v.state.next({values:{...ye}})}u={mount:re.keepDirtyValues?u.mount:new Set,unMount:new Set,array:new Set,disabled:new Set,watch:new Set,watchAll:!1,focus:""},s.mount=!p.isValid||!!re.keepIsValid||!!re.keepDirtyValues,s.watch=!!t.shouldUnregister,v.state.next({submitCount:re.keepSubmitCount?n.submitCount:0,isDirty:Ee?!1:re.keepDirty?n.isDirty:!!(re.keepDefaultValues&&!Ja(G,i)),isSubmitted:re.keepIsSubmitted?n.isSubmitted:!1,dirtyFields:Ee?{}:re.keepDirtyValues?re.keepDefaultValues&&a?Ec(i,a):n.dirtyFields:re.keepDefaultValues&&G?Ec(i,G):re.keepDirty?n.dirtyFields:{},touchedFields:re.keepTouched?n.touchedFields:{},errors:re.keepErrors?n.errors:{},isSubmitSuccessful:re.keepIsSubmitSuccessful?n.isSubmitSuccessful:!1,isSubmitting:!1})},it=(G,re)=>ft(Gr(G)?G(a):G,re),xt=(G,re={})=>{const fe=Ce(r,G),Ae=fe&&fe._f;if(Ae){const Ee=Ae.refs?Ae.refs[0]:Ae.ref;Ee.focus&&(Ee.focus(),re.shouldSelect&&Gr(Ee.select)&&Ee.select())}},Tt=G=>{n={...n,...G}},Bt={control:{register:_e,unregister:ce,getFieldState:X,handleSubmit:Be,setError:le,_subscribe:de,_runSchema:k,_getWatch:Z,_getDirty:z,_setValid:C,_setFieldArray:A,_setDisabledField:we,_setErrors:R,_getFieldArray:Y,_reset:ft,_resetDefaultValues:()=>Gr(t.defaultValues)&&t.defaultValues().then(G=>{it(G,t.resetOptions),v.state.next({isLoading:!1})}),_removeUnmounted:F,_disableForm:Ye,_subjects:v,_proxyFormState:p,get _fields(){return r},get _formValues(){return a},get _state(){return s},set _state(G){s=G},get _defaultValues(){return i},get _names(){return u},set _names(G){u=G},get _formState(){return n},get _options(){return t},set _options(G){t={...t,...G}}},subscribe:Se,trigger:J,register:_e,handleSubmit:Be,watch:ve,setValue:B,getValues:q,reset:it,resetField:dt,clearErrors:oe,unregister:ce,setError:le,setFocus:xt,getFieldState:X};return{...Bt,formControl:Bt}}function cB(e={}){const t=V.useRef(void 0),n=V.useRef(void 0),[r,i]=V.useState({isDirty:!1,isValidating:!1,isLoading:Gr(e.defaultValues),isSubmitted:!1,isSubmitting:!1,isSubmitSuccessful:!1,isValid:!1,submitCount:0,dirtyFields:{},touchedFields:{},validatingFields:{},errors:e.errors||{},disabled:e.disabled||!1,isReady:!1,defaultValues:Gr(e.defaultValues)?void 0:e.defaultValues});t.current||(t.current={...e.formControl?e.formControl:QG(e),formState:r},e.formControl&&e.defaultValues&&!Gr(e.defaultValues)&&e.formControl.reset(e.defaultValues,e.resetOptions));const a=t.current.control;return a._options=e,UE(()=>{const s=a._subscribe({formState:a._proxyFormState,callback:()=>i({...a._formState}),reRenderRoot:!0});return i(u=>({...u,isReady:!0})),a._formState.isReady=!0,s},[a]),V.useEffect(()=>a._disableForm(e.disabled),[a,e.disabled]),V.useEffect(()=>{e.mode&&(a._options.mode=e.mode),e.reValidateMode&&(a._options.reValidateMode=e.reValidateMode),e.errors&&!In(e.errors)&&a._setErrors(e.errors)},[a,e.errors,e.mode,e.reValidateMode]),V.useEffect(()=>{e.shouldUnregister&&a._subjects.state.next({values:a._getWatch()})},[a,e.shouldUnregister]),V.useEffect(()=>{if(a._proxyFormState.isDirty){const s=a._getDirty();s!==r.isDirty&&a._subjects.state.next({isDirty:s})}},[a,r.isDirty]),V.useEffect(()=>{e.values&&!Ja(e.values,n.current)?(a._reset(e.values,a._options.resetOptions),n.current=e.values,i(s=>({...s}))):a._resetDefaultValues()},[a,e.values]),V.useEffect(()=>{a._state.mount||(a._setValid(),a._state.mount=!0),a._state.watch&&(a._state.watch=!1,a._subjects.state.next({...a._formState})),a._removeUnmounted()}),t.current.formState=eB(r,a),t.current}const pP=(e,t,n)=>{if(e&&"reportValidity"in e){const r=Ce(n,t);e.setCustomValidity(r&&r.message||""),e.reportValidity()}},fB=(e,t)=>{for(const n in t.fields){const r=t.fields[n];r&&r.ref&&"reportValidity"in r.ref?pP(r.ref,n,e):r&&r.refs&&r.refs.forEach(i=>pP(i,n,e))}},JG=(e,t)=>{t.shouldUseNativeValidation&&fB(e,t);const n={};for(const r in e){const i=Ce(t.fields,r),a=Object.assign(e[r]||{},{ref:i&&i.ref});if(e7(t.names||Object.keys(e),r)){const s=Object.assign({},Ce(n,r));bt(s,"root",a),bt(n,r,s)}else bt(n,r,a)}return n},e7=(e,t)=>{const n=hP(t);return e.some(r=>hP(r).match(`^${n}\\.\\d+`))};function hP(e){return e.replace(/\]|\[/g,"")}function t7(e,t){for(var n={};e.length;){var r=e[0],i=r.code,a=r.message,s=r.path.join(".");if(!n[s])if("unionErrors"in r){var u=r.unionErrors[0].errors[0];n[s]={message:u.message,type:u.code}}else n[s]={message:a,type:i};if("unionErrors"in r&&r.unionErrors.forEach(function(p){return p.errors.forEach(function(m){return e.push(m)})}),t){var c=n[s].types,d=c&&c[r.code];n[s]=rB(s,t,n,i,d?[].concat(d,r.message):r.message)}e.shift()}return n}function dB(e,t,n){return n===void 0&&(n={}),function(r,i,a){try{return Promise.resolve(function(s,u){try{var c=Promise.resolve(e[n.mode==="sync"?"parse":"parseAsync"](r,t)).then(function(d){return a.shouldUseNativeValidation&&fB({},a),{errors:{},values:n.raw?Object.assign({},r):d}})}catch(d){return u(d)}return c&&c.then?c.then(void 0,u):c}(0,function(s){if(function(u){return Array.isArray(u==null?void 0:u.errors)}(s))return{values:{},errors:JG(t7(s.errors,!a.shouldUseNativeValidation&&a.criteriaMode==="all"),a)};throw s}))}catch(s){return Promise.reject(s)}}}var ut;(function(e){e.assertEqual=i=>i;function t(i){}e.assertIs=t;function n(i){throw new Error}e.assertNever=n,e.arrayToEnum=i=>{const a={};for(const s of i)a[s]=s;return a},e.getValidEnumValues=i=>{const a=e.objectKeys(i).filter(u=>typeof i[i[u]]!="number"),s={};for(const u of a)s[u]=i[u];return e.objectValues(s)},e.objectValues=i=>e.objectKeys(i).map(function(a){return i[a]}),e.objectKeys=typeof Object.keys=="function"?i=>Object.keys(i):i=>{const a=[];for(const s in i)Object.prototype.hasOwnProperty.call(i,s)&&a.push(s);return a},e.find=(i,a)=>{for(const s of i)if(a(s))return s},e.isInteger=typeof Number.isInteger=="function"?i=>Number.isInteger(i):i=>typeof i=="number"&&isFinite(i)&&Math.floor(i)===i;function r(i,a=" | "){return i.map(s=>typeof s=="string"?`'${s}'`:s).join(a)}e.joinValues=r,e.jsonStringifyReplacer=(i,a)=>typeof a=="bigint"?a.toString():a})(ut||(ut={}));var C_;(function(e){e.mergeShapes=(t,n)=>({...t,...n})})(C_||(C_={}));const Me=ut.arrayToEnum(["string","nan","number","integer","float","boolean","date","bigint","symbol","function","undefined","null","array","object","unknown","promise","void","never","map","set"]),ea=e=>{switch(typeof e){case"undefined":return Me.undefined;case"string":return Me.string;case"number":return isNaN(e)?Me.nan:Me.number;case"boolean":return Me.boolean;case"function":return Me.function;case"bigint":return Me.bigint;case"symbol":return Me.symbol;case"object":return Array.isArray(e)?Me.array:e===null?Me.null:e.then&&typeof e.then=="function"&&e.catch&&typeof e.catch=="function"?Me.promise:typeof Map<"u"&&e instanceof Map?Me.map:typeof Set<"u"&&e instanceof Set?Me.set:typeof Date<"u"&&e instanceof Date?Me.date:Me.object;default:return Me.unknown}},xe=ut.arrayToEnum(["invalid_type","invalid_literal","custom","invalid_union","invalid_union_discriminator","invalid_enum_value","unrecognized_keys","invalid_arguments","invalid_return_type","invalid_date","invalid_string","too_small","too_big","invalid_intersection_types","not_multiple_of","not_finite"]),n7=e=>JSON.stringify(e,null,2).replace(/"([^"]+)":/g,"$1:");class mr extends Error{get errors(){return this.issues}constructor(t){super(),this.issues=[],this.addIssue=r=>{this.issues=[...this.issues,r]},this.addIssues=(r=[])=>{this.issues=[...this.issues,...r]};const n=new.target.prototype;Object.setPrototypeOf?Object.setPrototypeOf(this,n):this.__proto__=n,this.name="ZodError",this.issues=t}format(t){const n=t||function(a){return a.message},r={_errors:[]},i=a=>{for(const s of a.issues)if(s.code==="invalid_union")s.unionErrors.map(i);else if(s.code==="invalid_return_type")i(s.returnTypeError);else if(s.code==="invalid_arguments")i(s.argumentsError);else if(s.path.length===0)r._errors.push(n(s));else{let u=r,c=0;for(;c<s.path.length;){const d=s.path[c];c===s.path.length-1?(u[d]=u[d]||{_errors:[]},u[d]._errors.push(n(s))):u[d]=u[d]||{_errors:[]},u=u[d],c++}}};return i(this),r}static assert(t){if(!(t instanceof mr))throw new Error(`Not a ZodError: ${t}`)}toString(){return this.message}get message(){return JSON.stringify(this.issues,ut.jsonStringifyReplacer,2)}get isEmpty(){return this.issues.length===0}flatten(t=n=>n.message){const n={},r=[];for(const i of this.issues)i.path.length>0?(n[i.path[0]]=n[i.path[0]]||[],n[i.path[0]].push(t(i))):r.push(t(i));return{formErrors:r,fieldErrors:n}}get formErrors(){return this.flatten()}}mr.create=e=>new mr(e);const Pl=(e,t)=>{let n;switch(e.code){case xe.invalid_type:e.received===Me.undefined?n="Required":n=`Expected ${e.expected}, received ${e.received}`;break;case xe.invalid_literal:n=`Invalid literal value, expected ${JSON.stringify(e.expected,ut.jsonStringifyReplacer)}`;break;case xe.unrecognized_keys:n=`Unrecognized key(s) in object: ${ut.joinValues(e.keys,", ")}`;break;case xe.invalid_union:n="Invalid input";break;case xe.invalid_union_discriminator:n=`Invalid discriminator value. Expected ${ut.joinValues(e.options)}`;break;case xe.invalid_enum_value:n=`Invalid enum value. Expected ${ut.joinValues(e.options)}, received '${e.received}'`;break;case xe.invalid_arguments:n="Invalid function arguments";break;case xe.invalid_return_type:n="Invalid function return type";break;case xe.invalid_date:n="Invalid date";break;case xe.invalid_string:typeof e.validation=="object"?"includes"in e.validation?(n=`Invalid input: must include "${e.validation.includes}"`,typeof e.validation.position=="number"&&(n=`${n} at one or more positions greater than or equal to ${e.validation.position}`)):"startsWith"in e.validation?n=`Invalid input: must start with "${e.validation.startsWith}"`:"endsWith"in e.validation?n=`Invalid input: must end with "${e.validation.endsWith}"`:ut.assertNever(e.validation):e.validation!=="regex"?n=`Invalid ${e.validation}`:n="Invalid";break;case xe.too_small:e.type==="array"?n=`Array must contain ${e.exact?"exactly":e.inclusive?"at least":"more than"} ${e.minimum} element(s)`:e.type==="string"?n=`String must contain ${e.exact?"exactly":e.inclusive?"at least":"over"} ${e.minimum} character(s)`:e.type==="number"?n=`Number must be ${e.exact?"exactly equal to ":e.inclusive?"greater than or equal to ":"greater than "}${e.minimum}`:e.type==="date"?n=`Date must be ${e.exact?"exactly equal to ":e.inclusive?"greater than or equal to ":"greater than "}${new Date(Number(e.minimum))}`:n="Invalid input";break;case xe.too_big:e.type==="array"?n=`Array must contain ${e.exact?"exactly":e.inclusive?"at most":"less than"} ${e.maximum} element(s)`:e.type==="string"?n=`String must contain ${e.exact?"exactly":e.inclusive?"at most":"under"} ${e.maximum} character(s)`:e.type==="number"?n=`Number must be ${e.exact?"exactly":e.inclusive?"less than or equal to":"less than"} ${e.maximum}`:e.type==="bigint"?n=`BigInt must be ${e.exact?"exactly":e.inclusive?"less than or equal to":"less than"} ${e.maximum}`:e.type==="date"?n=`Date must be ${e.exact?"exactly":e.inclusive?"smaller than or equal to":"smaller than"} ${new Date(Number(e.maximum))}`:n="Invalid input";break;case xe.custom:n="Invalid input";break;case xe.invalid_intersection_types:n="Intersection results could not be merged";break;case xe.not_multiple_of:n=`Number must be a multiple of ${e.multipleOf}`;break;case xe.not_finite:n="Number must be finite";break;default:n=t.defaultError,ut.assertNever(e)}return{message:n}};let pB=Pl;function r7(e){pB=e}function Dh(){return pB}const kh=e=>{const{data:t,path:n,errorMaps:r,issueData:i}=e,a=[...n,...i.path||[]],s={...i,path:a};if(i.message!==void 0)return{...i,path:a,message:i.message};let u="";const c=r.filter(d=>!!d).slice().reverse();for(const d of c)u=d(s,{data:t,defaultError:u}).message;return{...i,path:a,message:u}},i7=[];function Oe(e,t){const n=Dh(),r=kh({issueData:t,data:e.data,path:e.path,errorMaps:[e.common.contextualErrorMap,e.schemaErrorMap,n,n===Pl?void 0:Pl].filter(i=>!!i)});e.common.issues.push(r)}class Pn{constructor(){this.value="valid"}dirty(){this.value==="valid"&&(this.value="dirty")}abort(){this.value!=="aborted"&&(this.value="aborted")}static mergeArray(t,n){const r=[];for(const i of n){if(i.status==="aborted")return He;i.status==="dirty"&&t.dirty(),r.push(i.value)}return{status:t.value,value:r}}static async mergeObjectAsync(t,n){const r=[];for(const i of n){const a=await i.key,s=await i.value;r.push({key:a,value:s})}return Pn.mergeObjectSync(t,r)}static mergeObjectSync(t,n){const r={};for(const i of n){const{key:a,value:s}=i;if(a.status==="aborted"||s.status==="aborted")return He;a.status==="dirty"&&t.dirty(),s.status==="dirty"&&t.dirty(),a.value!=="__proto__"&&(typeof s.value<"u"||i.alwaysSet)&&(r[a.value]=s.value)}return{status:t.value,value:r}}}const He=Object.freeze({status:"aborted"}),xl=e=>({status:"dirty",value:e}),qn=e=>({status:"valid",value:e}),O_=e=>e.status==="aborted",A_=e=>e.status==="dirty",is=e=>e.status==="valid",ef=e=>typeof Promise<"u"&&e instanceof Promise;function $h(e,t,n,r){if(typeof t=="function"?e!==t||!0:!t.has(e))throw new TypeError("Cannot read private member from an object whose class did not declare it");return t.get(e)}function hB(e,t,n,r,i){if(typeof t=="function"?e!==t||!0:!t.has(e))throw new TypeError("Cannot write private member to an object whose class did not declare it");return t.set(e,n),n}var Ie;(function(e){e.errToObj=t=>typeof t=="string"?{message:t}:t||{},e.toString=t=>typeof t=="string"?t:t==null?void 0:t.message})(Ie||(Ie={}));var Ic,Lc;class _i{constructor(t,n,r,i){this._cachedPath=[],this.parent=t,this.data=n,this._path=r,this._key=i}get path(){return this._cachedPath.length||(this._key instanceof Array?this._cachedPath.push(...this._path,...this._key):this._cachedPath.push(...this._path,this._key)),this._cachedPath}}const mP=(e,t)=>{if(is(t))return{success:!0,data:t.value};if(!e.common.issues.length)throw new Error("Validation failed but no issues detected.");return{success:!1,get error(){if(this._error)return this._error;const n=new mr(e.common.issues);return this._error=n,this._error}}};function Xe(e){if(!e)return{};const{errorMap:t,invalid_type_error:n,required_error:r,description:i}=e;if(t&&(n||r))throw new Error(`Can't use "invalid_type_error" or "required_error" in conjunction with custom error map.`);return t?{errorMap:t,description:i}:{errorMap:(s,u)=>{var c,d;const{message:p}=e;return s.code==="invalid_enum_value"?{message:p??u.defaultError}:typeof u.data>"u"?{message:(c=p??r)!==null&&c!==void 0?c:u.defaultError}:s.code!=="invalid_type"?{message:u.defaultError}:{message:(d=p??n)!==null&&d!==void 0?d:u.defaultError}},description:i}}class et{get description(){return this._def.description}_getType(t){return ea(t.data)}_getOrReturnCtx(t,n){return n||{common:t.parent.common,data:t.data,parsedType:ea(t.data),schemaErrorMap:this._def.errorMap,path:t.path,parent:t.parent}}_processInputParams(t){return{status:new Pn,ctx:{common:t.parent.common,data:t.data,parsedType:ea(t.data),schemaErrorMap:this._def.errorMap,path:t.path,parent:t.parent}}}_parseSync(t){const n=this._parse(t);if(ef(n))throw new Error("Synchronous parse encountered promise.");return n}_parseAsync(t){const n=this._parse(t);return Promise.resolve(n)}parse(t,n){const r=this.safeParse(t,n);if(r.success)return r.data;throw r.error}safeParse(t,n){var r;const i={common:{issues:[],async:(r=n==null?void 0:n.async)!==null&&r!==void 0?r:!1,contextualErrorMap:n==null?void 0:n.errorMap},path:(n==null?void 0:n.path)||[],schemaErrorMap:this._def.errorMap,parent:null,data:t,parsedType:ea(t)},a=this._parseSync({data:t,path:i.path,parent:i});return mP(i,a)}"~validate"(t){var n,r;const i={common:{issues:[],async:!!this["~standard"].async},path:[],schemaErrorMap:this._def.errorMap,parent:null,data:t,parsedType:ea(t)};if(!this["~standard"].async)try{const a=this._parseSync({data:t,path:[],parent:i});return is(a)?{value:a.value}:{issues:i.common.issues}}catch(a){!((r=(n=a==null?void 0:a.message)===null||n===void 0?void 0:n.toLowerCase())===null||r===void 0)&&r.includes("encountered")&&(this["~standard"].async=!0),i.common={issues:[],async:!0}}return this._parseAsync({data:t,path:[],parent:i}).then(a=>is(a)?{value:a.value}:{issues:i.common.issues})}async parseAsync(t,n){const r=await this.safeParseAsync(t,n);if(r.success)return r.data;throw r.error}async safeParseAsync(t,n){const r={common:{issues:[],contextualErrorMap:n==null?void 0:n.errorMap,async:!0},path:(n==null?void 0:n.path)||[],schemaErrorMap:this._def.errorMap,parent:null,data:t,parsedType:ea(t)},i=this._parse({data:t,path:r.path,parent:r}),a=await(ef(i)?i:Promise.resolve(i));return mP(r,a)}refine(t,n){const r=i=>typeof n=="string"||typeof n>"u"?{message:n}:typeof n=="function"?n(i):n;return this._refinement((i,a)=>{const s=t(i),u=()=>a.addIssue({code:xe.custom,...r(i)});return typeof Promise<"u"&&s instanceof Promise?s.then(c=>c?!0:(u(),!1)):s?!0:(u(),!1)})}refinement(t,n){return this._refinement((r,i)=>t(r)?!0:(i.addIssue(typeof n=="function"?n(r,i):n),!1))}_refinement(t){return new Qr({schema:this,typeName:Ue.ZodEffects,effect:{type:"refinement",refinement:t}})}superRefine(t){return this._refinement(t)}constructor(t){this.spa=this.safeParseAsync,this._def=t,this.parse=this.parse.bind(this),this.safeParse=this.safeParse.bind(this),this.parseAsync=this.parseAsync.bind(this),this.safeParseAsync=this.safeParseAsync.bind(this),this.spa=this.spa.bind(this),this.refine=this.refine.bind(this),this.refinement=this.refinement.bind(this),this.superRefine=this.superRefine.bind(this),this.optional=this.optional.bind(this),this.nullable=this.nullable.bind(this),this.nullish=this.nullish.bind(this),this.array=this.array.bind(this),this.promise=this.promise.bind(this),this.or=this.or.bind(this),this.and=this.and.bind(this),this.transform=this.transform.bind(this),this.brand=this.brand.bind(this),this.default=this.default.bind(this),this.catch=this.catch.bind(this),this.describe=this.describe.bind(this),this.pipe=this.pipe.bind(this),this.readonly=this.readonly.bind(this),this.isNullable=this.isNullable.bind(this),this.isOptional=this.isOptional.bind(this),this["~standard"]={version:1,vendor:"zod",validate:n=>this["~validate"](n)}}optional(){return vi.create(this,this._def)}nullable(){return oo.create(this,this._def)}nullish(){return this.nullable().optional()}array(){return Yr.create(this)}promise(){return Nl.create(this,this._def)}or(t){return af.create([this,t],this._def)}and(t){return of.create(this,t,this._def)}transform(t){return new Qr({...Xe(this._def),schema:this,typeName:Ue.ZodEffects,effect:{type:"transform",transform:t}})}default(t){const n=typeof t=="function"?t:()=>t;return new ff({...Xe(this._def),innerType:this,defaultValue:n,typeName:Ue.ZodDefault})}brand(){return new GE({typeName:Ue.ZodBranded,type:this,...Xe(this._def)})}catch(t){const n=typeof t=="function"?t:()=>t;return new df({...Xe(this._def),innerType:this,catchValue:n,typeName:Ue.ZodCatch})}describe(t){const n=this.constructor;return new n({...this._def,description:t})}pipe(t){return ld.create(this,t)}readonly(){return pf.create(this)}isOptional(){return this.safeParse(void 0).success}isNullable(){return this.safeParse(null).success}}const a7=/^c[^\s-]{8,}$/i,o7=/^[0-9a-z]+$/,s7=/^[0-9A-HJKMNP-TV-Z]{26}$/i,l7=/^[0-9a-fA-F]{8}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{12}$/i,u7=/^[a-z0-9_-]{21}$/i,c7=/^[A-Za-z0-9-_]+\.[A-Za-z0-9-_]+\.[A-Za-z0-9-_]*$/,f7=/^[-+]?P(?!$)(?:(?:[-+]?\d+Y)|(?:[-+]?\d+[.,]\d+Y$))?(?:(?:[-+]?\d+M)|(?:[-+]?\d+[.,]\d+M$))?(?:(?:[-+]?\d+W)|(?:[-+]?\d+[.,]\d+W$))?(?:(?:[-+]?\d+D)|(?:[-+]?\d+[.,]\d+D$))?(?:T(?=[\d+-])(?:(?:[-+]?\d+H)|(?:[-+]?\d+[.,]\d+H$))?(?:(?:[-+]?\d+M)|(?:[-+]?\d+[.,]\d+M$))?(?:[-+]?\d+(?:[.,]\d+)?S)?)??$/,d7=/^(?!\.)(?!.*\.\.)([A-Z0-9_'+\-\.]*)[A-Z0-9_+-]@([A-Z0-9][A-Z0-9\-]*\.)+[A-Z]{2,}$/i,p7="^(\\p{Extended_Pictographic}|\\p{Emoji_Component})+$";let Bb;const h7=/^(?:(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9][0-9]|[0-9])\.){3}(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9][0-9]|[0-9])$/,m7=/^(?:(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9][0-9]|[0-9])\.){3}(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9][0-9]|[0-9])\/(3[0-2]|[12]?[0-9])$/,v7=/^(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}:|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|:((:[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|([0-9a-fA-F]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))$/,g7=/^(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}:|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|:((:[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|([0-9a-fA-F]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))\/(12[0-8]|1[01][0-9]|[1-9]?[0-9])$/,y7=/^([0-9a-zA-Z+/]{4})*(([0-9a-zA-Z+/]{2}==)|([0-9a-zA-Z+/]{3}=))?$/,b7=/^([0-9a-zA-Z-_]{4})*(([0-9a-zA-Z-_]{2}(==)?)|([0-9a-zA-Z-_]{3}(=)?))?$/,mB="((\\d\\d[2468][048]|\\d\\d[13579][26]|\\d\\d0[48]|[02468][048]00|[13579][26]00)-02-29|\\d{4}-((0[13578]|1[02])-(0[1-9]|[12]\\d|3[01])|(0[469]|11)-(0[1-9]|[12]\\d|30)|(02)-(0[1-9]|1\\d|2[0-8])))",x7=new RegExp(`^${mB}$`);function vB(e){let t="[0-5]\\d";e.precision?t=`${t}\\.\\d{${e.precision}}`:e.precision==null&&(t=`${t}(\\.\\d+)?`);const n=e.precision?"+":"?";return`([01]\\d|2[0-3]):[0-5]\\d(:${t})${n}`}function w7(e){return new RegExp(`^${vB(e)}$`)}function gB(e){let t=`${mB}T${vB(e)}`;const n=[];return n.push(e.local?"Z?":"Z"),e.offset&&n.push("([+-]\\d{2}:?\\d{2})"),t=`${t}(${n.join("|")})`,new RegExp(`^${t}$`)}function S7(e,t){return!!((t==="v4"||!t)&&h7.test(e)||(t==="v6"||!t)&&v7.test(e))}function _7(e,t){if(!c7.test(e))return!1;try{const[n]=e.split("."),r=n.replace(/-/g,"+").replace(/_/g,"/").padEnd(n.length+(4-n.length%4)%4,"="),i=JSON.parse(atob(r));return!(typeof i!="object"||i===null||!i.typ||!i.alg||t&&i.alg!==t)}catch{return!1}}function E7(e,t){return!!((t==="v4"||!t)&&m7.test(e)||(t==="v6"||!t)&&g7.test(e))}class Kr extends et{_parse(t){if(this._def.coerce&&(t.data=String(t.data)),this._getType(t)!==Me.string){const a=this._getOrReturnCtx(t);return Oe(a,{code:xe.invalid_type,expected:Me.string,received:a.parsedType}),He}const r=new Pn;let i;for(const a of this._def.checks)if(a.kind==="min")t.data.length<a.value&&(i=this._getOrReturnCtx(t,i),Oe(i,{code:xe.too_small,minimum:a.value,type:"string",inclusive:!0,exact:!1,message:a.message}),r.dirty());else if(a.kind==="max")t.data.length>a.value&&(i=this._getOrReturnCtx(t,i),Oe(i,{code:xe.too_big,maximum:a.value,type:"string",inclusive:!0,exact:!1,message:a.message}),r.dirty());else if(a.kind==="length"){const s=t.data.length>a.value,u=t.data.length<a.value;(s||u)&&(i=this._getOrReturnCtx(t,i),s?Oe(i,{code:xe.too_big,maximum:a.value,type:"string",inclusive:!0,exact:!0,message:a.message}):u&&Oe(i,{code:xe.too_small,minimum:a.value,type:"string",inclusive:!0,exact:!0,message:a.message}),r.dirty())}else if(a.kind==="email")d7.test(t.data)||(i=this._getOrReturnCtx(t,i),Oe(i,{validation:"email",code:xe.invalid_string,message:a.message}),r.dirty());else if(a.kind==="emoji")Bb||(Bb=new RegExp(p7,"u")),Bb.test(t.data)||(i=this._getOrReturnCtx(t,i),Oe(i,{validation:"emoji",code:xe.invalid_string,message:a.message}),r.dirty());else if(a.kind==="uuid")l7.test(t.data)||(i=this._getOrReturnCtx(t,i),Oe(i,{validation:"uuid",code:xe.invalid_string,message:a.message}),r.dirty());else if(a.kind==="nanoid")u7.test(t.data)||(i=this._getOrReturnCtx(t,i),Oe(i,{validation:"nanoid",code:xe.invalid_string,message:a.message}),r.dirty());else if(a.kind==="cuid")a7.test(t.data)||(i=this._getOrReturnCtx(t,i),Oe(i,{validation:"cuid",code:xe.invalid_string,message:a.message}),r.dirty());else if(a.kind==="cuid2")o7.test(t.data)||(i=this._getOrReturnCtx(t,i),Oe(i,{validation:"cuid2",code:xe.invalid_string,message:a.message}),r.dirty());else if(a.kind==="ulid")s7.test(t.data)||(i=this._getOrReturnCtx(t,i),Oe(i,{validation:"ulid",code:xe.invalid_string,message:a.message}),r.dirty());else if(a.kind==="url")try{new URL(t.data)}catch{i=this._getOrReturnCtx(t,i),Oe(i,{validation:"url",code:xe.invalid_string,message:a.message}),r.dirty()}else a.kind==="regex"?(a.regex.lastIndex=0,a.regex.test(t.data)||(i=this._getOrReturnCtx(t,i),Oe(i,{validation:"regex",code:xe.invalid_string,message:a.message}),r.dirty())):a.kind==="trim"?t.data=t.data.trim():a.kind==="includes"?t.data.includes(a.value,a.position)||(i=this._getOrReturnCtx(t,i),Oe(i,{code:xe.invalid_string,validation:{includes:a.value,position:a.position},message:a.message}),r.dirty()):a.kind==="toLowerCase"?t.data=t.data.toLowerCase():a.kind==="toUpperCase"?t.data=t.data.toUpperCase():a.kind==="startsWith"?t.data.startsWith(a.value)||(i=this._getOrReturnCtx(t,i),Oe(i,{code:xe.invalid_string,validation:{startsWith:a.value},message:a.message}),r.dirty()):a.kind==="endsWith"?t.data.endsWith(a.value)||(i=this._getOrReturnCtx(t,i),Oe(i,{code:xe.invalid_string,validation:{endsWith:a.value},message:a.message}),r.dirty()):a.kind==="datetime"?gB(a).test(t.data)||(i=this._getOrReturnCtx(t,i),Oe(i,{code:xe.invalid_string,validation:"datetime",message:a.message}),r.dirty()):a.kind==="date"?x7.test(t.data)||(i=this._getOrReturnCtx(t,i),Oe(i,{code:xe.invalid_string,validation:"date",message:a.message}),r.dirty()):a.kind==="time"?w7(a).test(t.data)||(i=this._getOrReturnCtx(t,i),Oe(i,{code:xe.invalid_string,validation:"time",message:a.message}),r.dirty()):a.kind==="duration"?f7.test(t.data)||(i=this._getOrReturnCtx(t,i),Oe(i,{validation:"duration",code:xe.invalid_string,message:a.message}),r.dirty()):a.kind==="ip"?S7(t.data,a.version)||(i=this._getOrReturnCtx(t,i),Oe(i,{validation:"ip",code:xe.invalid_string,message:a.message}),r.dirty()):a.kind==="jwt"?_7(t.data,a.alg)||(i=this._getOrReturnCtx(t,i),Oe(i,{validation:"jwt",code:xe.invalid_string,message:a.message}),r.dirty()):a.kind==="cidr"?E7(t.data,a.version)||(i=this._getOrReturnCtx(t,i),Oe(i,{validation:"cidr",code:xe.invalid_string,message:a.message}),r.dirty()):a.kind==="base64"?y7.test(t.data)||(i=this._getOrReturnCtx(t,i),Oe(i,{validation:"base64",code:xe.invalid_string,message:a.message}),r.dirty()):a.kind==="base64url"?b7.test(t.data)||(i=this._getOrReturnCtx(t,i),Oe(i,{validation:"base64url",code:xe.invalid_string,message:a.message}),r.dirty()):ut.assertNever(a);return{status:r.value,value:t.data}}_regex(t,n,r){return this.refinement(i=>t.test(i),{validation:n,code:xe.invalid_string,...Ie.errToObj(r)})}_addCheck(t){return new Kr({...this._def,checks:[...this._def.checks,t]})}email(t){return this._addCheck({kind:"email",...Ie.errToObj(t)})}url(t){return this._addCheck({kind:"url",...Ie.errToObj(t)})}emoji(t){return this._addCheck({kind:"emoji",...Ie.errToObj(t)})}uuid(t){return this._addCheck({kind:"uuid",...Ie.errToObj(t)})}nanoid(t){return this._addCheck({kind:"nanoid",...Ie.errToObj(t)})}cuid(t){return this._addCheck({kind:"cuid",...Ie.errToObj(t)})}cuid2(t){return this._addCheck({kind:"cuid2",...Ie.errToObj(t)})}ulid(t){return this._addCheck({kind:"ulid",...Ie.errToObj(t)})}base64(t){return this._addCheck({kind:"base64",...Ie.errToObj(t)})}base64url(t){return this._addCheck({kind:"base64url",...Ie.errToObj(t)})}jwt(t){return this._addCheck({kind:"jwt",...Ie.errToObj(t)})}ip(t){return this._addCheck({kind:"ip",...Ie.errToObj(t)})}cidr(t){return this._addCheck({kind:"cidr",...Ie.errToObj(t)})}datetime(t){var n,r;return typeof t=="string"?this._addCheck({kind:"datetime",precision:null,offset:!1,local:!1,message:t}):this._addCheck({kind:"datetime",precision:typeof(t==null?void 0:t.precision)>"u"?null:t==null?void 0:t.precision,offset:(n=t==null?void 0:t.offset)!==null&&n!==void 0?n:!1,local:(r=t==null?void 0:t.local)!==null&&r!==void 0?r:!1,...Ie.errToObj(t==null?void 0:t.message)})}date(t){return this._addCheck({kind:"date",message:t})}time(t){return typeof t=="string"?this._addCheck({kind:"time",precision:null,message:t}):this._addCheck({kind:"time",precision:typeof(t==null?void 0:t.precision)>"u"?null:t==null?void 0:t.precision,...Ie.errToObj(t==null?void 0:t.message)})}duration(t){return this._addCheck({kind:"duration",...Ie.errToObj(t)})}regex(t,n){return this._addCheck({kind:"regex",regex:t,...Ie.errToObj(n)})}includes(t,n){return this._addCheck({kind:"includes",value:t,position:n==null?void 0:n.position,...Ie.errToObj(n==null?void 0:n.message)})}startsWith(t,n){return this._addCheck({kind:"startsWith",value:t,...Ie.errToObj(n)})}endsWith(t,n){return this._addCheck({kind:"endsWith",value:t,...Ie.errToObj(n)})}min(t,n){return this._addCheck({kind:"min",value:t,...Ie.errToObj(n)})}max(t,n){return this._addCheck({kind:"max",value:t,...Ie.errToObj(n)})}length(t,n){return this._addCheck({kind:"length",value:t,...Ie.errToObj(n)})}nonempty(t){return this.min(1,Ie.errToObj(t))}trim(){return new Kr({...this._def,checks:[...this._def.checks,{kind:"trim"}]})}toLowerCase(){return new Kr({...this._def,checks:[...this._def.checks,{kind:"toLowerCase"}]})}toUpperCase(){return new Kr({...this._def,checks:[...this._def.checks,{kind:"toUpperCase"}]})}get isDatetime(){return!!this._def.checks.find(t=>t.kind==="datetime")}get isDate(){return!!this._def.checks.find(t=>t.kind==="date")}get isTime(){return!!this._def.checks.find(t=>t.kind==="time")}get isDuration(){return!!this._def.checks.find(t=>t.kind==="duration")}get isEmail(){return!!this._def.checks.find(t=>t.kind==="email")}get isURL(){return!!this._def.checks.find(t=>t.kind==="url")}get isEmoji(){return!!this._def.checks.find(t=>t.kind==="emoji")}get isUUID(){return!!this._def.checks.find(t=>t.kind==="uuid")}get isNANOID(){return!!this._def.checks.find(t=>t.kind==="nanoid")}get isCUID(){return!!this._def.checks.find(t=>t.kind==="cuid")}get isCUID2(){return!!this._def.checks.find(t=>t.kind==="cuid2")}get isULID(){return!!this._def.checks.find(t=>t.kind==="ulid")}get isIP(){return!!this._def.checks.find(t=>t.kind==="ip")}get isCIDR(){return!!this._def.checks.find(t=>t.kind==="cidr")}get isBase64(){return!!this._def.checks.find(t=>t.kind==="base64")}get isBase64url(){return!!this._def.checks.find(t=>t.kind==="base64url")}get minLength(){let t=null;for(const n of this._def.checks)n.kind==="min"&&(t===null||n.value>t)&&(t=n.value);return t}get maxLength(){let t=null;for(const n of this._def.checks)n.kind==="max"&&(t===null||n.value<t)&&(t=n.value);return t}}Kr.create=e=>{var t;return new Kr({checks:[],typeName:Ue.ZodString,coerce:(t=e==null?void 0:e.coerce)!==null&&t!==void 0?t:!1,...Xe(e)})};function C7(e,t){const n=(e.toString().split(".")[1]||"").length,r=(t.toString().split(".")[1]||"").length,i=n>r?n:r,a=parseInt(e.toFixed(i).replace(".","")),s=parseInt(t.toFixed(i).replace(".",""));return a%s/Math.pow(10,i)}class ro extends et{constructor(){super(...arguments),this.min=this.gte,this.max=this.lte,this.step=this.multipleOf}_parse(t){if(this._def.coerce&&(t.data=Number(t.data)),this._getType(t)!==Me.number){const a=this._getOrReturnCtx(t);return Oe(a,{code:xe.invalid_type,expected:Me.number,received:a.parsedType}),He}let r;const i=new Pn;for(const a of this._def.checks)a.kind==="int"?ut.isInteger(t.data)||(r=this._getOrReturnCtx(t,r),Oe(r,{code:xe.invalid_type,expected:"integer",received:"float",message:a.message}),i.dirty()):a.kind==="min"?(a.inclusive?t.data<a.value:t.data<=a.value)&&(r=this._getOrReturnCtx(t,r),Oe(r,{code:xe.too_small,minimum:a.value,type:"number",inclusive:a.inclusive,exact:!1,message:a.message}),i.dirty()):a.kind==="max"?(a.inclusive?t.data>a.value:t.data>=a.value)&&(r=this._getOrReturnCtx(t,r),Oe(r,{code:xe.too_big,maximum:a.value,type:"number",inclusive:a.inclusive,exact:!1,message:a.message}),i.dirty()):a.kind==="multipleOf"?C7(t.data,a.value)!==0&&(r=this._getOrReturnCtx(t,r),Oe(r,{code:xe.not_multiple_of,multipleOf:a.value,message:a.message}),i.dirty()):a.kind==="finite"?Number.isFinite(t.data)||(r=this._getOrReturnCtx(t,r),Oe(r,{code:xe.not_finite,message:a.message}),i.dirty()):ut.assertNever(a);return{status:i.value,value:t.data}}gte(t,n){return this.setLimit("min",t,!0,Ie.toString(n))}gt(t,n){return this.setLimit("min",t,!1,Ie.toString(n))}lte(t,n){return this.setLimit("max",t,!0,Ie.toString(n))}lt(t,n){return this.setLimit("max",t,!1,Ie.toString(n))}setLimit(t,n,r,i){return new ro({...this._def,checks:[...this._def.checks,{kind:t,value:n,inclusive:r,message:Ie.toString(i)}]})}_addCheck(t){return new ro({...this._def,checks:[...this._def.checks,t]})}int(t){return this._addCheck({kind:"int",message:Ie.toString(t)})}positive(t){return this._addCheck({kind:"min",value:0,inclusive:!1,message:Ie.toString(t)})}negative(t){return this._addCheck({kind:"max",value:0,inclusive:!1,message:Ie.toString(t)})}nonpositive(t){return this._addCheck({kind:"max",value:0,inclusive:!0,message:Ie.toString(t)})}nonnegative(t){return this._addCheck({kind:"min",value:0,inclusive:!0,message:Ie.toString(t)})}multipleOf(t,n){return this._addCheck({kind:"multipleOf",value:t,message:Ie.toString(n)})}finite(t){return this._addCheck({kind:"finite",message:Ie.toString(t)})}safe(t){return this._addCheck({kind:"min",inclusive:!0,value:Number.MIN_SAFE_INTEGER,message:Ie.toString(t)})._addCheck({kind:"max",inclusive:!0,value:Number.MAX_SAFE_INTEGER,message:Ie.toString(t)})}get minValue(){let t=null;for(const n of this._def.checks)n.kind==="min"&&(t===null||n.value>t)&&(t=n.value);return t}get maxValue(){let t=null;for(const n of this._def.checks)n.kind==="max"&&(t===null||n.value<t)&&(t=n.value);return t}get isInt(){return!!this._def.checks.find(t=>t.kind==="int"||t.kind==="multipleOf"&&ut.isInteger(t.value))}get isFinite(){let t=null,n=null;for(const r of this._def.checks){if(r.kind==="finite"||r.kind==="int"||r.kind==="multipleOf")return!0;r.kind==="min"?(n===null||r.value>n)&&(n=r.value):r.kind==="max"&&(t===null||r.value<t)&&(t=r.value)}return Number.isFinite(n)&&Number.isFinite(t)}}ro.create=e=>new ro({checks:[],typeName:Ue.ZodNumber,coerce:(e==null?void 0:e.coerce)||!1,...Xe(e)});class io extends et{constructor(){super(...arguments),this.min=this.gte,this.max=this.lte}_parse(t){if(this._def.coerce)try{t.data=BigInt(t.data)}catch{return this._getInvalidInput(t)}if(this._getType(t)!==Me.bigint)return this._getInvalidInput(t);let r;const i=new Pn;for(const a of this._def.checks)a.kind==="min"?(a.inclusive?t.data<a.value:t.data<=a.value)&&(r=this._getOrReturnCtx(t,r),Oe(r,{code:xe.too_small,type:"bigint",minimum:a.value,inclusive:a.inclusive,message:a.message}),i.dirty()):a.kind==="max"?(a.inclusive?t.data>a.value:t.data>=a.value)&&(r=this._getOrReturnCtx(t,r),Oe(r,{code:xe.too_big,type:"bigint",maximum:a.value,inclusive:a.inclusive,message:a.message}),i.dirty()):a.kind==="multipleOf"?t.data%a.value!==BigInt(0)&&(r=this._getOrReturnCtx(t,r),Oe(r,{code:xe.not_multiple_of,multipleOf:a.value,message:a.message}),i.dirty()):ut.assertNever(a);return{status:i.value,value:t.data}}_getInvalidInput(t){const n=this._getOrReturnCtx(t);return Oe(n,{code:xe.invalid_type,expected:Me.bigint,received:n.parsedType}),He}gte(t,n){return this.setLimit("min",t,!0,Ie.toString(n))}gt(t,n){return this.setLimit("min",t,!1,Ie.toString(n))}lte(t,n){return this.setLimit("max",t,!0,Ie.toString(n))}lt(t,n){return this.setLimit("max",t,!1,Ie.toString(n))}setLimit(t,n,r,i){return new io({...this._def,checks:[...this._def.checks,{kind:t,value:n,inclusive:r,message:Ie.toString(i)}]})}_addCheck(t){return new io({...this._def,checks:[...this._def.checks,t]})}positive(t){return this._addCheck({kind:"min",value:BigInt(0),inclusive:!1,message:Ie.toString(t)})}negative(t){return this._addCheck({kind:"max",value:BigInt(0),inclusive:!1,message:Ie.toString(t)})}nonpositive(t){return this._addCheck({kind:"max",value:BigInt(0),inclusive:!0,message:Ie.toString(t)})}nonnegative(t){return this._addCheck({kind:"min",value:BigInt(0),inclusive:!0,message:Ie.toString(t)})}multipleOf(t,n){return this._addCheck({kind:"multipleOf",value:t,message:Ie.toString(n)})}get minValue(){let t=null;for(const n of this._def.checks)n.kind==="min"&&(t===null||n.value>t)&&(t=n.value);return t}get maxValue(){let t=null;for(const n of this._def.checks)n.kind==="max"&&(t===null||n.value<t)&&(t=n.value);return t}}io.create=e=>{var t;return new io({checks:[],typeName:Ue.ZodBigInt,coerce:(t=e==null?void 0:e.coerce)!==null&&t!==void 0?t:!1,...Xe(e)})};class tf extends et{_parse(t){if(this._def.coerce&&(t.data=!!t.data),this._getType(t)!==Me.boolean){const r=this._getOrReturnCtx(t);return Oe(r,{code:xe.invalid_type,expected:Me.boolean,received:r.parsedType}),He}return qn(t.data)}}tf.create=e=>new tf({typeName:Ue.ZodBoolean,coerce:(e==null?void 0:e.coerce)||!1,...Xe(e)});class as extends et{_parse(t){if(this._def.coerce&&(t.data=new Date(t.data)),this._getType(t)!==Me.date){const a=this._getOrReturnCtx(t);return Oe(a,{code:xe.invalid_type,expected:Me.date,received:a.parsedType}),He}if(isNaN(t.data.getTime())){const a=this._getOrReturnCtx(t);return Oe(a,{code:xe.invalid_date}),He}const r=new Pn;let i;for(const a of this._def.checks)a.kind==="min"?t.data.getTime()<a.value&&(i=this._getOrReturnCtx(t,i),Oe(i,{code:xe.too_small,message:a.message,inclusive:!0,exact:!1,minimum:a.value,type:"date"}),r.dirty()):a.kind==="max"?t.data.getTime()>a.value&&(i=this._getOrReturnCtx(t,i),Oe(i,{code:xe.too_big,message:a.message,inclusive:!0,exact:!1,maximum:a.value,type:"date"}),r.dirty()):ut.assertNever(a);return{status:r.value,value:new Date(t.data.getTime())}}_addCheck(t){return new as({...this._def,checks:[...this._def.checks,t]})}min(t,n){return this._addCheck({kind:"min",value:t.getTime(),message:Ie.toString(n)})}max(t,n){return this._addCheck({kind:"max",value:t.getTime(),message:Ie.toString(n)})}get minDate(){let t=null;for(const n of this._def.checks)n.kind==="min"&&(t===null||n.value>t)&&(t=n.value);return t!=null?new Date(t):null}get maxDate(){let t=null;for(const n of this._def.checks)n.kind==="max"&&(t===null||n.value<t)&&(t=n.value);return t!=null?new Date(t):null}}as.create=e=>new as({checks:[],coerce:(e==null?void 0:e.coerce)||!1,typeName:Ue.ZodDate,...Xe(e)});class Ih extends et{_parse(t){if(this._getType(t)!==Me.symbol){const r=this._getOrReturnCtx(t);return Oe(r,{code:xe.invalid_type,expected:Me.symbol,received:r.parsedType}),He}return qn(t.data)}}Ih.create=e=>new Ih({typeName:Ue.ZodSymbol,...Xe(e)});class nf extends et{_parse(t){if(this._getType(t)!==Me.undefined){const r=this._getOrReturnCtx(t);return Oe(r,{code:xe.invalid_type,expected:Me.undefined,received:r.parsedType}),He}return qn(t.data)}}nf.create=e=>new nf({typeName:Ue.ZodUndefined,...Xe(e)});class rf extends et{_parse(t){if(this._getType(t)!==Me.null){const r=this._getOrReturnCtx(t);return Oe(r,{code:xe.invalid_type,expected:Me.null,received:r.parsedType}),He}return qn(t.data)}}rf.create=e=>new rf({typeName:Ue.ZodNull,...Xe(e)});class jl extends et{constructor(){super(...arguments),this._any=!0}_parse(t){return qn(t.data)}}jl.create=e=>new jl({typeName:Ue.ZodAny,...Xe(e)});class Xo extends et{constructor(){super(...arguments),this._unknown=!0}_parse(t){return qn(t.data)}}Xo.create=e=>new Xo({typeName:Ue.ZodUnknown,...Xe(e)});class fa extends et{_parse(t){const n=this._getOrReturnCtx(t);return Oe(n,{code:xe.invalid_type,expected:Me.never,received:n.parsedType}),He}}fa.create=e=>new fa({typeName:Ue.ZodNever,...Xe(e)});class Lh extends et{_parse(t){if(this._getType(t)!==Me.undefined){const r=this._getOrReturnCtx(t);return Oe(r,{code:xe.invalid_type,expected:Me.void,received:r.parsedType}),He}return qn(t.data)}}Lh.create=e=>new Lh({typeName:Ue.ZodVoid,...Xe(e)});class Yr extends et{_parse(t){const{ctx:n,status:r}=this._processInputParams(t),i=this._def;if(n.parsedType!==Me.array)return Oe(n,{code:xe.invalid_type,expected:Me.array,received:n.parsedType}),He;if(i.exactLength!==null){const s=n.data.length>i.exactLength.value,u=n.data.length<i.exactLength.value;(s||u)&&(Oe(n,{code:s?xe.too_big:xe.too_small,minimum:u?i.exactLength.value:void 0,maximum:s?i.exactLength.value:void 0,type:"array",inclusive:!0,exact:!0,message:i.exactLength.message}),r.dirty())}if(i.minLength!==null&&n.data.length<i.minLength.value&&(Oe(n,{code:xe.too_small,minimum:i.minLength.value,type:"array",inclusive:!0,exact:!1,message:i.minLength.message}),r.dirty()),i.maxLength!==null&&n.data.length>i.maxLength.value&&(Oe(n,{code:xe.too_big,maximum:i.maxLength.value,type:"array",inclusive:!0,exact:!1,message:i.maxLength.message}),r.dirty()),n.common.async)return Promise.all([...n.data].map((s,u)=>i.type._parseAsync(new _i(n,s,n.path,u)))).then(s=>Pn.mergeArray(r,s));const a=[...n.data].map((s,u)=>i.type._parseSync(new _i(n,s,n.path,u)));return Pn.mergeArray(r,a)}get element(){return this._def.type}min(t,n){return new Yr({...this._def,minLength:{value:t,message:Ie.toString(n)}})}max(t,n){return new Yr({...this._def,maxLength:{value:t,message:Ie.toString(n)}})}length(t,n){return new Yr({...this._def,exactLength:{value:t,message:Ie.toString(n)}})}nonempty(t){return this.min(1,t)}}Yr.create=(e,t)=>new Yr({type:e,minLength:null,maxLength:null,exactLength:null,typeName:Ue.ZodArray,...Xe(t)});function vl(e){if(e instanceof qt){const t={};for(const n in e.shape){const r=e.shape[n];t[n]=vi.create(vl(r))}return new qt({...e._def,shape:()=>t})}else return e instanceof Yr?new Yr({...e._def,type:vl(e.element)}):e instanceof vi?vi.create(vl(e.unwrap())):e instanceof oo?oo.create(vl(e.unwrap())):e instanceof Ei?Ei.create(e.items.map(t=>vl(t))):e}class qt extends et{constructor(){super(...arguments),this._cached=null,this.nonstrict=this.passthrough,this.augment=this.extend}_getCached(){if(this._cached!==null)return this._cached;const t=this._def.shape(),n=ut.objectKeys(t);return this._cached={shape:t,keys:n}}_parse(t){if(this._getType(t)!==Me.object){const d=this._getOrReturnCtx(t);return Oe(d,{code:xe.invalid_type,expected:Me.object,received:d.parsedType}),He}const{status:r,ctx:i}=this._processInputParams(t),{shape:a,keys:s}=this._getCached(),u=[];if(!(this._def.catchall instanceof fa&&this._def.unknownKeys==="strip"))for(const d in i.data)s.includes(d)||u.push(d);const c=[];for(const d of s){const p=a[d],m=i.data[d];c.push({key:{status:"valid",value:d},value:p._parse(new _i(i,m,i.path,d)),alwaysSet:d in i.data})}if(this._def.catchall instanceof fa){const d=this._def.unknownKeys;if(d==="passthrough")for(const p of u)c.push({key:{status:"valid",value:p},value:{status:"valid",value:i.data[p]}});else if(d==="strict")u.length>0&&(Oe(i,{code:xe.unrecognized_keys,keys:u}),r.dirty());else if(d!=="strip")throw new Error("Internal ZodObject error: invalid unknownKeys value.")}else{const d=this._def.catchall;for(const p of u){const m=i.data[p];c.push({key:{status:"valid",value:p},value:d._parse(new _i(i,m,i.path,p)),alwaysSet:p in i.data})}}return i.common.async?Promise.resolve().then(async()=>{const d=[];for(const p of c){const m=await p.key,v=await p.value;d.push({key:m,value:v,alwaysSet:p.alwaysSet})}return d}).then(d=>Pn.mergeObjectSync(r,d)):Pn.mergeObjectSync(r,c)}get shape(){return this._def.shape()}strict(t){return Ie.errToObj,new qt({...this._def,unknownKeys:"strict",...t!==void 0?{errorMap:(n,r)=>{var i,a,s,u;const c=(s=(a=(i=this._def).errorMap)===null||a===void 0?void 0:a.call(i,n,r).message)!==null&&s!==void 0?s:r.defaultError;return n.code==="unrecognized_keys"?{message:(u=Ie.errToObj(t).message)!==null&&u!==void 0?u:c}:{message:c}}}:{}})}strip(){return new qt({...this._def,unknownKeys:"strip"})}passthrough(){return new qt({...this._def,unknownKeys:"passthrough"})}extend(t){return new qt({...this._def,shape:()=>({...this._def.shape(),...t})})}merge(t){return new qt({unknownKeys:t._def.unknownKeys,catchall:t._def.catchall,shape:()=>({...this._def.shape(),...t._def.shape()}),typeName:Ue.ZodObject})}setKey(t,n){return this.augment({[t]:n})}catchall(t){return new qt({...this._def,catchall:t})}pick(t){const n={};return ut.objectKeys(t).forEach(r=>{t[r]&&this.shape[r]&&(n[r]=this.shape[r])}),new qt({...this._def,shape:()=>n})}omit(t){const n={};return ut.objectKeys(this.shape).forEach(r=>{t[r]||(n[r]=this.shape[r])}),new qt({...this._def,shape:()=>n})}deepPartial(){return vl(this)}partial(t){const n={};return ut.objectKeys(this.shape).forEach(r=>{const i=this.shape[r];t&&!t[r]?n[r]=i:n[r]=i.optional()}),new qt({...this._def,shape:()=>n})}required(t){const n={};return ut.objectKeys(this.shape).forEach(r=>{if(t&&!t[r])n[r]=this.shape[r];else{let a=this.shape[r];for(;a instanceof vi;)a=a._def.innerType;n[r]=a}}),new qt({...this._def,shape:()=>n})}keyof(){return yB(ut.objectKeys(this.shape))}}qt.create=(e,t)=>new qt({shape:()=>e,unknownKeys:"strip",catchall:fa.create(),typeName:Ue.ZodObject,...Xe(t)});qt.strictCreate=(e,t)=>new qt({shape:()=>e,unknownKeys:"strict",catchall:fa.create(),typeName:Ue.ZodObject,...Xe(t)});qt.lazycreate=(e,t)=>new qt({shape:e,unknownKeys:"strip",catchall:fa.create(),typeName:Ue.ZodObject,...Xe(t)});class af extends et{_parse(t){const{ctx:n}=this._processInputParams(t),r=this._def.options;function i(a){for(const u of a)if(u.result.status==="valid")return u.result;for(const u of a)if(u.result.status==="dirty")return n.common.issues.push(...u.ctx.common.issues),u.result;const s=a.map(u=>new mr(u.ctx.common.issues));return Oe(n,{code:xe.invalid_union,unionErrors:s}),He}if(n.common.async)return Promise.all(r.map(async a=>{const s={...n,common:{...n.common,issues:[]},parent:null};return{result:await a._parseAsync({data:n.data,path:n.path,parent:s}),ctx:s}})).then(i);{let a;const s=[];for(const c of r){const d={...n,common:{...n.common,issues:[]},parent:null},p=c._parseSync({data:n.data,path:n.path,parent:d});if(p.status==="valid")return p;p.status==="dirty"&&!a&&(a={result:p,ctx:d}),d.common.issues.length&&s.push(d.common.issues)}if(a)return n.common.issues.push(...a.ctx.common.issues),a.result;const u=s.map(c=>new mr(c));return Oe(n,{code:xe.invalid_union,unionErrors:u}),He}}get options(){return this._def.options}}af.create=(e,t)=>new af({options:e,typeName:Ue.ZodUnion,...Xe(t)});const Xi=e=>e instanceof lf?Xi(e.schema):e instanceof Qr?Xi(e.innerType()):e instanceof uf?[e.value]:e instanceof ao?e.options:e instanceof cf?ut.objectValues(e.enum):e instanceof ff?Xi(e._def.innerType):e instanceof nf?[void 0]:e instanceof rf?[null]:e instanceof vi?[void 0,...Xi(e.unwrap())]:e instanceof oo?[null,...Xi(e.unwrap())]:e instanceof GE||e instanceof pf?Xi(e.unwrap()):e instanceof df?Xi(e._def.innerType):[];class tv extends et{_parse(t){const{ctx:n}=this._processInputParams(t);if(n.parsedType!==Me.object)return Oe(n,{code:xe.invalid_type,expected:Me.object,received:n.parsedType}),He;const r=this.discriminator,i=n.data[r],a=this.optionsMap.get(i);return a?n.common.async?a._parseAsync({data:n.data,path:n.path,parent:n}):a._parseSync({data:n.data,path:n.path,parent:n}):(Oe(n,{code:xe.invalid_union_discriminator,options:Array.from(this.optionsMap.keys()),path:[r]}),He)}get discriminator(){return this._def.discriminator}get options(){return this._def.options}get optionsMap(){return this._def.optionsMap}static create(t,n,r){const i=new Map;for(const a of n){const s=Xi(a.shape[t]);if(!s.length)throw new Error(`A discriminator value for key \`${t}\` could not be extracted from all schema options`);for(const u of s){if(i.has(u))throw new Error(`Discriminator property ${String(t)} has duplicate value ${String(u)}`);i.set(u,a)}}return new tv({typeName:Ue.ZodDiscriminatedUnion,discriminator:t,options:n,optionsMap:i,...Xe(r)})}}function R_(e,t){const n=ea(e),r=ea(t);if(e===t)return{valid:!0,data:e};if(n===Me.object&&r===Me.object){const i=ut.objectKeys(t),a=ut.objectKeys(e).filter(u=>i.indexOf(u)!==-1),s={...e,...t};for(const u of a){const c=R_(e[u],t[u]);if(!c.valid)return{valid:!1};s[u]=c.data}return{valid:!0,data:s}}else if(n===Me.array&&r===Me.array){if(e.length!==t.length)return{valid:!1};const i=[];for(let a=0;a<e.length;a++){const s=e[a],u=t[a],c=R_(s,u);if(!c.valid)return{valid:!1};i.push(c.data)}return{valid:!0,data:i}}else return n===Me.date&&r===Me.date&&+e==+t?{valid:!0,data:e}:{valid:!1}}class of extends et{_parse(t){const{status:n,ctx:r}=this._processInputParams(t),i=(a,s)=>{if(O_(a)||O_(s))return He;const u=R_(a.value,s.value);return u.valid?((A_(a)||A_(s))&&n.dirty(),{status:n.value,value:u.data}):(Oe(r,{code:xe.invalid_intersection_types}),He)};return r.common.async?Promise.all([this._def.left._parseAsync({data:r.data,path:r.path,parent:r}),this._def.right._parseAsync({data:r.data,path:r.path,parent:r})]).then(([a,s])=>i(a,s)):i(this._def.left._parseSync({data:r.data,path:r.path,parent:r}),this._def.right._parseSync({data:r.data,path:r.path,parent:r}))}}of.create=(e,t,n)=>new of({left:e,right:t,typeName:Ue.ZodIntersection,...Xe(n)});class Ei extends et{_parse(t){const{status:n,ctx:r}=this._processInputParams(t);if(r.parsedType!==Me.array)return Oe(r,{code:xe.invalid_type,expected:Me.array,received:r.parsedType}),He;if(r.data.length<this._def.items.length)return Oe(r,{code:xe.too_small,minimum:this._def.items.length,inclusive:!0,exact:!1,type:"array"}),He;!this._def.rest&&r.data.length>this._def.items.length&&(Oe(r,{code:xe.too_big,maximum:this._def.items.length,inclusive:!0,exact:!1,type:"array"}),n.dirty());const a=[...r.data].map((s,u)=>{const c=this._def.items[u]||this._def.rest;return c?c._parse(new _i(r,s,r.path,u)):null}).filter(s=>!!s);return r.common.async?Promise.all(a).then(s=>Pn.mergeArray(n,s)):Pn.mergeArray(n,a)}get items(){return this._def.items}rest(t){return new Ei({...this._def,rest:t})}}Ei.create=(e,t)=>{if(!Array.isArray(e))throw new Error("You must pass an array of schemas to z.tuple([ ... ])");return new Ei({items:e,typeName:Ue.ZodTuple,rest:null,...Xe(t)})};class sf extends et{get keySchema(){return this._def.keyType}get valueSchema(){return this._def.valueType}_parse(t){const{status:n,ctx:r}=this._processInputParams(t);if(r.parsedType!==Me.object)return Oe(r,{code:xe.invalid_type,expected:Me.object,received:r.parsedType}),He;const i=[],a=this._def.keyType,s=this._def.valueType;for(const u in r.data)i.push({key:a._parse(new _i(r,u,r.path,u)),value:s._parse(new _i(r,r.data[u],r.path,u)),alwaysSet:u in r.data});return r.common.async?Pn.mergeObjectAsync(n,i):Pn.mergeObjectSync(n,i)}get element(){return this._def.valueType}static create(t,n,r){return n instanceof et?new sf({keyType:t,valueType:n,typeName:Ue.ZodRecord,...Xe(r)}):new sf({keyType:Kr.create(),valueType:t,typeName:Ue.ZodRecord,...Xe(n)})}}class Bh extends et{get keySchema(){return this._def.keyType}get valueSchema(){return this._def.valueType}_parse(t){const{status:n,ctx:r}=this._processInputParams(t);if(r.parsedType!==Me.map)return Oe(r,{code:xe.invalid_type,expected:Me.map,received:r.parsedType}),He;const i=this._def.keyType,a=this._def.valueType,s=[...r.data.entries()].map(([u,c],d)=>({key:i._parse(new _i(r,u,r.path,[d,"key"])),value:a._parse(new _i(r,c,r.path,[d,"value"]))}));if(r.common.async){const u=new Map;return Promise.resolve().then(async()=>{for(const c of s){const d=await c.key,p=await c.value;if(d.status==="aborted"||p.status==="aborted")return He;(d.status==="dirty"||p.status==="dirty")&&n.dirty(),u.set(d.value,p.value)}return{status:n.value,value:u}})}else{const u=new Map;for(const c of s){const d=c.key,p=c.value;if(d.status==="aborted"||p.status==="aborted")return He;(d.status==="dirty"||p.status==="dirty")&&n.dirty(),u.set(d.value,p.value)}return{status:n.value,value:u}}}}Bh.create=(e,t,n)=>new Bh({valueType:t,keyType:e,typeName:Ue.ZodMap,...Xe(n)});class os extends et{_parse(t){const{status:n,ctx:r}=this._processInputParams(t);if(r.parsedType!==Me.set)return Oe(r,{code:xe.invalid_type,expected:Me.set,received:r.parsedType}),He;const i=this._def;i.minSize!==null&&r.data.size<i.minSize.value&&(Oe(r,{code:xe.too_small,minimum:i.minSize.value,type:"set",inclusive:!0,exact:!1,message:i.minSize.message}),n.dirty()),i.maxSize!==null&&r.data.size>i.maxSize.value&&(Oe(r,{code:xe.too_big,maximum:i.maxSize.value,type:"set",inclusive:!0,exact:!1,message:i.maxSize.message}),n.dirty());const a=this._def.valueType;function s(c){const d=new Set;for(const p of c){if(p.status==="aborted")return He;p.status==="dirty"&&n.dirty(),d.add(p.value)}return{status:n.value,value:d}}const u=[...r.data.values()].map((c,d)=>a._parse(new _i(r,c,r.path,d)));return r.common.async?Promise.all(u).then(c=>s(c)):s(u)}min(t,n){return new os({...this._def,minSize:{value:t,message:Ie.toString(n)}})}max(t,n){return new os({...this._def,maxSize:{value:t,message:Ie.toString(n)}})}size(t,n){return this.min(t,n).max(t,n)}nonempty(t){return this.min(1,t)}}os.create=(e,t)=>new os({valueType:e,minSize:null,maxSize:null,typeName:Ue.ZodSet,...Xe(t)});class El extends et{constructor(){super(...arguments),this.validate=this.implement}_parse(t){const{ctx:n}=this._processInputParams(t);if(n.parsedType!==Me.function)return Oe(n,{code:xe.invalid_type,expected:Me.function,received:n.parsedType}),He;function r(u,c){return kh({data:u,path:n.path,errorMaps:[n.common.contextualErrorMap,n.schemaErrorMap,Dh(),Pl].filter(d=>!!d),issueData:{code:xe.invalid_arguments,argumentsError:c}})}function i(u,c){return kh({data:u,path:n.path,errorMaps:[n.common.contextualErrorMap,n.schemaErrorMap,Dh(),Pl].filter(d=>!!d),issueData:{code:xe.invalid_return_type,returnTypeError:c}})}const a={errorMap:n.common.contextualErrorMap},s=n.data;if(this._def.returns instanceof Nl){const u=this;return qn(async function(...c){const d=new mr([]),p=await u._def.args.parseAsync(c,a).catch(b=>{throw d.addIssue(r(c,b)),d}),m=await Reflect.apply(s,this,p);return await u._def.returns._def.type.parseAsync(m,a).catch(b=>{throw d.addIssue(i(m,b)),d})})}else{const u=this;return qn(function(...c){const d=u._def.args.safeParse(c,a);if(!d.success)throw new mr([r(c,d.error)]);const p=Reflect.apply(s,this,d.data),m=u._def.returns.safeParse(p,a);if(!m.success)throw new mr([i(p,m.error)]);return m.data})}}parameters(){return this._def.args}returnType(){return this._def.returns}args(...t){return new El({...this._def,args:Ei.create(t).rest(Xo.create())})}returns(t){return new El({...this._def,returns:t})}implement(t){return this.parse(t)}strictImplement(t){return this.parse(t)}static create(t,n,r){return new El({args:t||Ei.create([]).rest(Xo.create()),returns:n||Xo.create(),typeName:Ue.ZodFunction,...Xe(r)})}}class lf extends et{get schema(){return this._def.getter()}_parse(t){const{ctx:n}=this._processInputParams(t);return this._def.getter()._parse({data:n.data,path:n.path,parent:n})}}lf.create=(e,t)=>new lf({getter:e,typeName:Ue.ZodLazy,...Xe(t)});class uf extends et{_parse(t){if(t.data!==this._def.value){const n=this._getOrReturnCtx(t);return Oe(n,{received:n.data,code:xe.invalid_literal,expected:this._def.value}),He}return{status:"valid",value:t.data}}get value(){return this._def.value}}uf.create=(e,t)=>new uf({value:e,typeName:Ue.ZodLiteral,...Xe(t)});function yB(e,t){return new ao({values:e,typeName:Ue.ZodEnum,...Xe(t)})}class ao extends et{constructor(){super(...arguments),Ic.set(this,void 0)}_parse(t){if(typeof t.data!="string"){const n=this._getOrReturnCtx(t),r=this._def.values;return Oe(n,{expected:ut.joinValues(r),received:n.parsedType,code:xe.invalid_type}),He}if($h(this,Ic)||hB(this,Ic,new Set(this._def.values)),!$h(this,Ic).has(t.data)){const n=this._getOrReturnCtx(t),r=this._def.values;return Oe(n,{received:n.data,code:xe.invalid_enum_value,options:r}),He}return qn(t.data)}get options(){return this._def.values}get enum(){const t={};for(const n of this._def.values)t[n]=n;return t}get Values(){const t={};for(const n of this._def.values)t[n]=n;return t}get Enum(){const t={};for(const n of this._def.values)t[n]=n;return t}extract(t,n=this._def){return ao.create(t,{...this._def,...n})}exclude(t,n=this._def){return ao.create(this.options.filter(r=>!t.includes(r)),{...this._def,...n})}}Ic=new WeakMap;ao.create=yB;class cf extends et{constructor(){super(...arguments),Lc.set(this,void 0)}_parse(t){const n=ut.getValidEnumValues(this._def.values),r=this._getOrReturnCtx(t);if(r.parsedType!==Me.string&&r.parsedType!==Me.number){const i=ut.objectValues(n);return Oe(r,{expected:ut.joinValues(i),received:r.parsedType,code:xe.invalid_type}),He}if($h(this,Lc)||hB(this,Lc,new Set(ut.getValidEnumValues(this._def.values))),!$h(this,Lc).has(t.data)){const i=ut.objectValues(n);return Oe(r,{received:r.data,code:xe.invalid_enum_value,options:i}),He}return qn(t.data)}get enum(){return this._def.values}}Lc=new WeakMap;cf.create=(e,t)=>new cf({values:e,typeName:Ue.ZodNativeEnum,...Xe(t)});class Nl extends et{unwrap(){return this._def.type}_parse(t){const{ctx:n}=this._processInputParams(t);if(n.parsedType!==Me.promise&&n.common.async===!1)return Oe(n,{code:xe.invalid_type,expected:Me.promise,received:n.parsedType}),He;const r=n.parsedType===Me.promise?n.data:Promise.resolve(n.data);return qn(r.then(i=>this._def.type.parseAsync(i,{path:n.path,errorMap:n.common.contextualErrorMap})))}}Nl.create=(e,t)=>new Nl({type:e,typeName:Ue.ZodPromise,...Xe(t)});class Qr extends et{innerType(){return this._def.schema}sourceType(){return this._def.schema._def.typeName===Ue.ZodEffects?this._def.schema.sourceType():this._def.schema}_parse(t){const{status:n,ctx:r}=this._processInputParams(t),i=this._def.effect||null,a={addIssue:s=>{Oe(r,s),s.fatal?n.abort():n.dirty()},get path(){return r.path}};if(a.addIssue=a.addIssue.bind(a),i.type==="preprocess"){const s=i.transform(r.data,a);if(r.common.async)return Promise.resolve(s).then(async u=>{if(n.value==="aborted")return He;const c=await this._def.schema._parseAsync({data:u,path:r.path,parent:r});return c.status==="aborted"?He:c.status==="dirty"||n.value==="dirty"?xl(c.value):c});{if(n.value==="aborted")return He;const u=this._def.schema._parseSync({data:s,path:r.path,parent:r});return u.status==="aborted"?He:u.status==="dirty"||n.value==="dirty"?xl(u.value):u}}if(i.type==="refinement"){const s=u=>{const c=i.refinement(u,a);if(r.common.async)return Promise.resolve(c);if(c instanceof Promise)throw new Error("Async refinement encountered during synchronous parse operation. Use .parseAsync instead.");return u};if(r.common.async===!1){const u=this._def.schema._parseSync({data:r.data,path:r.path,parent:r});return u.status==="aborted"?He:(u.status==="dirty"&&n.dirty(),s(u.value),{status:n.value,value:u.value})}else return this._def.schema._parseAsync({data:r.data,path:r.path,parent:r}).then(u=>u.status==="aborted"?He:(u.status==="dirty"&&n.dirty(),s(u.value).then(()=>({status:n.value,value:u.value}))))}if(i.type==="transform")if(r.common.async===!1){const s=this._def.schema._parseSync({data:r.data,path:r.path,parent:r});if(!is(s))return s;const u=i.transform(s.value,a);if(u instanceof Promise)throw new Error("Asynchronous transform encountered during synchronous parse operation. Use .parseAsync instead.");return{status:n.value,value:u}}else return this._def.schema._parseAsync({data:r.data,path:r.path,parent:r}).then(s=>is(s)?Promise.resolve(i.transform(s.value,a)).then(u=>({status:n.value,value:u})):s);ut.assertNever(i)}}Qr.create=(e,t,n)=>new Qr({schema:e,typeName:Ue.ZodEffects,effect:t,...Xe(n)});Qr.createWithPreprocess=(e,t,n)=>new Qr({schema:t,effect:{type:"preprocess",transform:e},typeName:Ue.ZodEffects,...Xe(n)});class vi extends et{_parse(t){return this._getType(t)===Me.undefined?qn(void 0):this._def.innerType._parse(t)}unwrap(){return this._def.innerType}}vi.create=(e,t)=>new vi({innerType:e,typeName:Ue.ZodOptional,...Xe(t)});class oo extends et{_parse(t){return this._getType(t)===Me.null?qn(null):this._def.innerType._parse(t)}unwrap(){return this._def.innerType}}oo.create=(e,t)=>new oo({innerType:e,typeName:Ue.ZodNullable,...Xe(t)});class ff extends et{_parse(t){const{ctx:n}=this._processInputParams(t);let r=n.data;return n.parsedType===Me.undefined&&(r=this._def.defaultValue()),this._def.innerType._parse({data:r,path:n.path,parent:n})}removeDefault(){return this._def.innerType}}ff.create=(e,t)=>new ff({innerType:e,typeName:Ue.ZodDefault,defaultValue:typeof t.default=="function"?t.default:()=>t.default,...Xe(t)});class df extends et{_parse(t){const{ctx:n}=this._processInputParams(t),r={...n,common:{...n.common,issues:[]}},i=this._def.innerType._parse({data:r.data,path:r.path,parent:{...r}});return ef(i)?i.then(a=>({status:"valid",value:a.status==="valid"?a.value:this._def.catchValue({get error(){return new mr(r.common.issues)},input:r.data})})):{status:"valid",value:i.status==="valid"?i.value:this._def.catchValue({get error(){return new mr(r.common.issues)},input:r.data})}}removeCatch(){return this._def.innerType}}df.create=(e,t)=>new df({innerType:e,typeName:Ue.ZodCatch,catchValue:typeof t.catch=="function"?t.catch:()=>t.catch,...Xe(t)});class Fh extends et{_parse(t){if(this._getType(t)!==Me.nan){const r=this._getOrReturnCtx(t);return Oe(r,{code:xe.invalid_type,expected:Me.nan,received:r.parsedType}),He}return{status:"valid",value:t.data}}}Fh.create=e=>new Fh({typeName:Ue.ZodNaN,...Xe(e)});const O7=Symbol("zod_brand");class GE extends et{_parse(t){const{ctx:n}=this._processInputParams(t),r=n.data;return this._def.type._parse({data:r,path:n.path,parent:n})}unwrap(){return this._def.type}}class ld extends et{_parse(t){const{status:n,ctx:r}=this._processInputParams(t);if(r.common.async)return(async()=>{const a=await this._def.in._parseAsync({data:r.data,path:r.path,parent:r});return a.status==="aborted"?He:a.status==="dirty"?(n.dirty(),xl(a.value)):this._def.out._parseAsync({data:a.value,path:r.path,parent:r})})();{const i=this._def.in._parseSync({data:r.data,path:r.path,parent:r});return i.status==="aborted"?He:i.status==="dirty"?(n.dirty(),{status:"dirty",value:i.value}):this._def.out._parseSync({data:i.value,path:r.path,parent:r})}}static create(t,n){return new ld({in:t,out:n,typeName:Ue.ZodPipeline})}}class pf extends et{_parse(t){const n=this._def.innerType._parse(t),r=i=>(is(i)&&(i.value=Object.freeze(i.value)),i);return ef(n)?n.then(i=>r(i)):r(n)}unwrap(){return this._def.innerType}}pf.create=(e,t)=>new pf({innerType:e,typeName:Ue.ZodReadonly,...Xe(t)});function vP(e,t){const n=typeof e=="function"?e(t):typeof e=="string"?{message:e}:e;return typeof n=="string"?{message:n}:n}function bB(e,t={},n){return e?jl.create().superRefine((r,i)=>{var a,s;const u=e(r);if(u instanceof Promise)return u.then(c=>{var d,p;if(!c){const m=vP(t,r),v=(p=(d=m.fatal)!==null&&d!==void 0?d:n)!==null&&p!==void 0?p:!0;i.addIssue({code:"custom",...m,fatal:v})}});if(!u){const c=vP(t,r),d=(s=(a=c.fatal)!==null&&a!==void 0?a:n)!==null&&s!==void 0?s:!0;i.addIssue({code:"custom",...c,fatal:d})}}):jl.create()}const A7={object:qt.lazycreate};var Ue;(function(e){e.ZodString="ZodString",e.ZodNumber="ZodNumber",e.ZodNaN="ZodNaN",e.ZodBigInt="ZodBigInt",e.ZodBoolean="ZodBoolean",e.ZodDate="ZodDate",e.ZodSymbol="ZodSymbol",e.ZodUndefined="ZodUndefined",e.ZodNull="ZodNull",e.ZodAny="ZodAny",e.ZodUnknown="ZodUnknown",e.ZodNever="ZodNever",e.ZodVoid="ZodVoid",e.ZodArray="ZodArray",e.ZodObject="ZodObject",e.ZodUnion="ZodUnion",e.ZodDiscriminatedUnion="ZodDiscriminatedUnion",e.ZodIntersection="ZodIntersection",e.ZodTuple="ZodTuple",e.ZodRecord="ZodRecord",e.ZodMap="ZodMap",e.ZodSet="ZodSet",e.ZodFunction="ZodFunction",e.ZodLazy="ZodLazy",e.ZodLiteral="ZodLiteral",e.ZodEnum="ZodEnum",e.ZodEffects="ZodEffects",e.ZodNativeEnum="ZodNativeEnum",e.ZodOptional="ZodOptional",e.ZodNullable="ZodNullable",e.ZodDefault="ZodDefault",e.ZodCatch="ZodCatch",e.ZodPromise="ZodPromise",e.ZodBranded="ZodBranded",e.ZodPipeline="ZodPipeline",e.ZodReadonly="ZodReadonly"})(Ue||(Ue={}));const R7=(e,t={message:`Input not instance of ${e.name}`})=>bB(n=>n instanceof e,t),xB=Kr.create,wB=ro.create,T7=Fh.create,M7=io.create,SB=tf.create,P7=as.create,j7=Ih.create,N7=nf.create,D7=rf.create,k7=jl.create,$7=Xo.create,I7=fa.create,L7=Lh.create,B7=Yr.create,F7=qt.create,q7=qt.strictCreate,z7=af.create,U7=tv.create,V7=of.create,H7=Ei.create,G7=sf.create,W7=Bh.create,K7=os.create,Y7=El.create,Z7=lf.create,X7=uf.create,Q7=ao.create,J7=cf.create,eW=Nl.create,gP=Qr.create,tW=vi.create,nW=oo.create,rW=Qr.createWithPreprocess,iW=ld.create,aW=()=>xB().optional(),oW=()=>wB().optional(),sW=()=>SB().optional(),lW={string:e=>Kr.create({...e,coerce:!0}),number:e=>ro.create({...e,coerce:!0}),boolean:e=>tf.create({...e,coerce:!0}),bigint:e=>io.create({...e,coerce:!0}),date:e=>as.create({...e,coerce:!0})},uW=He;var Ln=Object.freeze({__proto__:null,defaultErrorMap:Pl,setErrorMap:r7,getErrorMap:Dh,makeIssue:kh,EMPTY_PATH:i7,addIssueToContext:Oe,ParseStatus:Pn,INVALID:He,DIRTY:xl,OK:qn,isAborted:O_,isDirty:A_,isValid:is,isAsync:ef,get util(){return ut},get objectUtil(){return C_},ZodParsedType:Me,getParsedType:ea,ZodType:et,datetimeRegex:gB,ZodString:Kr,ZodNumber:ro,ZodBigInt:io,ZodBoolean:tf,ZodDate:as,ZodSymbol:Ih,ZodUndefined:nf,ZodNull:rf,ZodAny:jl,ZodUnknown:Xo,ZodNever:fa,ZodVoid:Lh,ZodArray:Yr,ZodObject:qt,ZodUnion:af,ZodDiscriminatedUnion:tv,ZodIntersection:of,ZodTuple:Ei,ZodRecord:sf,ZodMap:Bh,ZodSet:os,ZodFunction:El,ZodLazy:lf,ZodLiteral:uf,ZodEnum:ao,ZodNativeEnum:cf,ZodPromise:Nl,ZodEffects:Qr,ZodTransformer:Qr,ZodOptional:vi,ZodNullable:oo,ZodDefault:ff,ZodCatch:df,ZodNaN:Fh,BRAND:O7,ZodBranded:GE,ZodPipeline:ld,ZodReadonly:pf,custom:bB,Schema:et,ZodSchema:et,late:A7,get ZodFirstPartyTypeKind(){return Ue},coerce:lW,any:k7,array:B7,bigint:M7,boolean:SB,date:P7,discriminatedUnion:U7,effect:gP,enum:Q7,function:Y7,instanceof:R7,intersection:V7,lazy:Z7,literal:X7,map:W7,nan:T7,nativeEnum:J7,never:I7,null:D7,nullable:nW,number:wB,object:F7,oboolean:sW,onumber:oW,optional:tW,ostring:aW,pipeline:iW,preprocess:rW,promise:eW,record:G7,set:K7,strictObject:q7,string:xB,symbol:j7,transformer:gP,tuple:H7,undefined:N7,union:z7,unknown:$7,void:L7,NEVER:uW,ZodIssueCode:xe,quotelessJson:n7,ZodError:mr});const cW=G6("inline-flex items-center justify-center gap-2 whitespace-nowrap rounded-md text-sm font-medium transition-all disabled:pointer-events-none disabled:opacity-50 [&_svg]:pointer-events-none [&_svg:not([class*='size-'])]:size-4 shrink-0 [&_svg]:shrink-0 outline-none focus-visible:border-ring focus-visible:ring-ring/50 focus-visible:ring-[3px] aria-invalid:ring-destructive/20 dark:aria-invalid:ring-destructive/40 aria-invalid:border-destructive",{variants:{variant:{default:"bg-primary text-primary-foreground shadow-xs hover:bg-primary/90",destructive:"bg-destructive text-white shadow-xs hover:bg-destructive/90 focus-visible:ring-destructive/20 dark:focus-visible:ring-destructive/40 dark:bg-destructive/60",outline:"border bg-background shadow-xs hover:bg-accent hover:text-accent-foreground dark:bg-input/30 dark:border-input dark:hover:bg-input/50",secondary:"bg-secondary text-secondary-foreground shadow-xs hover:bg-secondary/80",ghost:"hover:bg-accent hover:text-accent-foreground dark:hover:bg-accent/50",link:"text-primary underline-offset-4 hover:underline"},size:{default:"h-9 px-4 py-2 has-[>svg]:px-3",sm:"h-8 rounded-md gap-1.5 px-3 has-[>svg]:px-2.5",lg:"h-10 rounded-md px-6 has-[>svg]:px-4",icon:"size-9"}},defaultVariants:{variant:"default",size:"default"}});function Pr({className:e,variant:t,size:n,asChild:r=!1,...i}){const a=r?kr:"button";return E.jsx(a,{"data-slot":"button",className:ke(cW({variant:t,size:n,className:e})),...i})}function WE(e){const[t,n]=g.useState(void 0);return cn(()=>{if(e){n({width:e.offsetWidth,height:e.offsetHeight});const r=new ResizeObserver(i=>{if(!Array.isArray(i)||!i.length)return;const a=i[0];let s,u;if("borderBoxSize"in a){const c=a.borderBoxSize,d=Array.isArray(c)?c[0]:c;s=d.inlineSize,u=d.blockSize}else s=e.offsetWidth,u=e.offsetHeight;n({width:s,height:u})});return r.observe(e,{box:"border-box"}),()=>r.unobserve(e)}else n(void 0)},[e]),t}function fW(e){const t=dW(e),n=g.forwardRef((r,i)=>{const{children:a,...s}=r,u=g.Children.toArray(a),c=u.find(hW);if(c){const d=c.props.children,p=u.map(m=>m===c?g.Children.count(d)>1?g.Children.only(null):g.isValidElement(d)?d.props.children:null:m);return E.jsx(t,{...s,ref:i,children:g.isValidElement(d)?g.cloneElement(d,void 0,p):null})}return E.jsx(t,{...s,ref:i,children:a})});return n.displayName=`${e}.Slot`,n}function dW(e){const t=g.forwardRef((n,r)=>{const{children:i,...a}=n,s=g.isValidElement(i)?vW(i):void 0,u=Re(s,r);if(g.isValidElement(i)){const c=mW(a,i.props);return i.type!==g.Fragment&&(c.ref=u),g.cloneElement(i,c)}return g.Children.count(i)>1?g.Children.only(null):null});return t.displayName=`${e}.SlotClone`,t}var pW=Symbol("radix.slottable");function hW(e){return g.isValidElement(e)&&typeof e.type=="function"&&"__radixId"in e.type&&e.type.__radixId===pW}function mW(e,t){const n={...t};for(const r in t){const i=e[r],a=t[r];/^on[A-Z]/.test(r)?i&&a?n[r]=(...u)=>{const c=a(...u);return i(...u),c}:i&&(n[r]=i):r==="style"?n[r]={...i,...a}:r==="className"&&(n[r]=[i,a].filter(Boolean).join(" "))}return{...e,...n}}function vW(e){var r,i;let t=(r=Object.getOwnPropertyDescriptor(e.props,"ref"))==null?void 0:r.get,n=t&&"isReactWarning"in t&&t.isReactWarning;return n?e.ref:(t=(i=Object.getOwnPropertyDescriptor(e,"ref"))==null?void 0:i.get,n=t&&"isReactWarning"in t&&t.isReactWarning,n?e.props.ref:e.props.ref||e.ref)}var gW=["a","button","div","form","h2","h3","img","input","label","li","nav","ol","p","select","span","svg","ul"],_B=gW.reduce((e,t)=>{const n=fW(`Primitive.${t}`),r=g.forwardRef((i,a)=>{const{asChild:s,...u}=i,c=s?n:t;return typeof window<"u"&&(window[Symbol.for("radix-ui")]=!0),E.jsx(c,{...u,ref:a})});return r.displayName=`Primitive.${t}`,{...e,[t]:r}},{}),nv="Switch",[yW,K0e]=jn(nv),[bW,xW]=yW(nv),EB=g.forwardRef((e,t)=>{const{__scopeSwitch:n,name:r,checked:i,defaultChecked:a,required:s,disabled:u,value:c="on",onCheckedChange:d,form:p,...m}=e,[v,b]=g.useState(null),S=Re(t,A=>b(A)),w=g.useRef(!1),x=v?p||!!v.closest("form"):!0,[C,O]=Xr({prop:i,defaultProp:a??!1,onChange:d,caller:nv});return E.jsxs(bW,{scope:n,checked:C,disabled:u,children:[E.jsx(_B.button,{type:"button",role:"switch","aria-checked":C,"aria-required":s,"data-state":RB(C),"data-disabled":u?"":void 0,disabled:u,value:c,...m,ref:S,onClick:be(e.onClick,A=>{O(T=>!T),x&&(w.current=A.isPropagationStopped(),w.current||A.stopPropagation())})}),x&&E.jsx(AB,{control:v,bubbles:!w.current,name:r,value:c,checked:C,required:s,disabled:u,form:p,style:{transform:"translateX(-100%)"}})]})});EB.displayName=nv;var CB="SwitchThumb",OB=g.forwardRef((e,t)=>{const{__scopeSwitch:n,...r}=e,i=xW(CB,n);return E.jsx(_B.span,{"data-state":RB(i.checked),"data-disabled":i.disabled?"":void 0,...r,ref:t})});OB.displayName=CB;var wW="SwitchBubbleInput",AB=g.forwardRef(({__scopeSwitch:e,control:t,checked:n,bubbles:r=!0,...i},a)=>{const s=g.useRef(null),u=Re(s,a),c=Ym(n),d=WE(t);return g.useEffect(()=>{const p=s.current;if(!p)return;const m=window.HTMLInputElement.prototype,b=Object.getOwnPropertyDescriptor(m,"checked").set;if(c!==n&&b){const S=new Event("click",{bubbles:r});b.call(p,n),p.dispatchEvent(S)}},[c,n,r]),E.jsx("input",{type:"checkbox","aria-hidden":!0,defaultChecked:n,...i,tabIndex:-1,ref:u,style:{...i.style,...d,position:"absolute",pointerEvents:"none",opacity:0,margin:0}})});AB.displayName=wW;function RB(e){return e?"checked":"unchecked"}var SW=EB,_W=OB;function TB({className:e,...t}){return E.jsx(SW,{"data-slot":"switch",className:ke("peer data-[state=checked]:bg-primary data-[state=unchecked]:bg-input focus-visible:border-ring focus-visible:ring-ring/50 dark:data-[state=unchecked]:bg-input/80 inline-flex h-[1.15rem] w-8 shrink-0 items-center rounded-full border border-transparent shadow-xs transition-all outline-none focus-visible:ring-[3px] disabled:cursor-not-allowed disabled:opacity-50",e),...t,children:E.jsx(_W,{"data-slot":"switch-thumb",className:ke("bg-background dark:data-[state=unchecked]:bg-foreground dark:data-[state=checked]:bg-primary-foreground pointer-events-none block size-4 rounded-full ring-0 transition-transform data-[state=checked]:translate-x-[calc(100%-2px)] data-[state=unchecked]:translate-x-0")})})}function T_({className:e,type:t,...n}){return E.jsx("input",{type:t,"data-slot":"input",className:ke("file:text-foreground placeholder:text-muted-foreground selection:bg-primary selection:text-primary-foreground dark:bg-input/30 border-input flex h-9 w-full min-w-0 rounded-md border bg-transparent px-3 py-1 text-base shadow-xs transition-[color,box-shadow] outline-none file:inline-flex file:h-7 file:border-0 file:bg-transparent file:text-sm file:font-medium disabled:pointer-events-none disabled:cursor-not-allowed disabled:opacity-50 md:text-sm","focus-visible:border-ring focus-visible:ring-ring/50 focus-visible:ring-[3px]","aria-invalid:ring-destructive/20 dark:aria-invalid:ring-destructive/40 aria-invalid:border-destructive",e),...n})}var EW="Label",MB=g.forwardRef((e,t)=>E.jsx(Ut.label,{...e,ref:t,onMouseDown:n=>{var i;n.target.closest("button, input, select, textarea")||((i=e.onMouseDown)==null||i.call(e,n),!n.defaultPrevented&&n.detail>1&&n.preventDefault())}}));MB.displayName=EW;var CW=MB;function OW({className:e,...t}){return E.jsx(CW,{"data-slot":"label",className:ke("flex items-center gap-2 text-sm leading-none font-medium select-none group-data-[disabled=true]:pointer-events-none group-data-[disabled=true]:opacity-50 peer-disabled:cursor-not-allowed peer-disabled:opacity-50",e),...t})}const PB=$G,jB=g.createContext({}),sr=({...e})=>E.jsx(jB.Provider,{value:{name:e.name},children:E.jsx(BG,{...e})}),rv=()=>{const e=g.useContext(jB),t=g.useContext(NB),{getFieldState:n}=ev(),r=tB({name:e.name}),i=n(e.name,r);if(!e)throw new Error("useFormField should be used within <FormField>");const{id:a}=t;return{id:a,name:e.name,formItemId:`${a}-form-item`,formDescriptionId:`${a}-form-item-description`,formMessageId:`${a}-form-item-message`,...i}},NB=g.createContext({});function lr({className:e,...t}){const n=g.useId();return E.jsx(NB.Provider,{value:{id:n},children:E.jsx("div",{"data-slot":"form-item",className:ke("grid gap-2",e),...t})})}function ur({className:e,...t}){const{error:n,formItemId:r}=rv();return E.jsx(OW,{"data-slot":"form-label","data-error":!!n,className:ke("data-[error=true]:text-destructive",e),htmlFor:r,...t})}function cr({...e}){const{error:t,formItemId:n,formDescriptionId:r,formMessageId:i}=rv();return E.jsx(kr,{"data-slot":"form-control",id:n,"aria-describedby":t?`${r} ${i}`:`${r}`,"aria-invalid":!!t,...e})}function fr({className:e,...t}){const{formDescriptionId:n}=rv();return E.jsx("p",{"data-slot":"form-description",id:n,className:ke("text-muted-foreground text-sm",e),...t})}function Vr({className:e,...t}){const{error:n,formMessageId:r}=rv(),i=n?String((n==null?void 0:n.message)??""):t.children;return i?E.jsx("p",{"data-slot":"form-message",id:r,className:ke("text-destructive text-sm",e),...t,children:i}):null}function qh(e,[t,n]){return Math.min(n,Math.max(t,e))}function zh(e){const t=AW(e),n=g.forwardRef((r,i)=>{const{children:a,...s}=r,u=g.Children.toArray(a),c=u.find(TW);if(c){const d=c.props.children,p=u.map(m=>m===c?g.Children.count(d)>1?g.Children.only(null):g.isValidElement(d)?d.props.children:null:m);return E.jsx(t,{...s,ref:i,children:g.isValidElement(d)?g.cloneElement(d,void 0,p):null})}return E.jsx(t,{...s,ref:i,children:a})});return n.displayName=`${e}.Slot`,n}function AW(e){const t=g.forwardRef((n,r)=>{const{children:i,...a}=n,s=g.isValidElement(i)?PW(i):void 0,u=Re(s,r);if(g.isValidElement(i)){const c=MW(a,i.props);return i.type!==g.Fragment&&(c.ref=u),g.cloneElement(i,c)}return g.Children.count(i)>1?g.Children.only(null):null});return t.displayName=`${e}.SlotClone`,t}var RW=Symbol("radix.slottable");function TW(e){return g.isValidElement(e)&&typeof e.type=="function"&&"__radixId"in e.type&&e.type.__radixId===RW}function MW(e,t){const n={...t};for(const r in t){const i=e[r],a=t[r];/^on[A-Z]/.test(r)?i&&a?n[r]=(...u)=>{const c=a(...u);return i(...u),c}:i&&(n[r]=i):r==="style"?n[r]={...i,...a}:r==="className"&&(n[r]=[i,a].filter(Boolean).join(" "))}return{...e,...n}}function PW(e){var r,i;let t=(r=Object.getOwnPropertyDescriptor(e.props,"ref"))==null?void 0:r.get,n=t&&"isReactWarning"in t&&t.isReactWarning;return n?e.ref:(t=(i=Object.getOwnPropertyDescriptor(e,"ref"))==null?void 0:i.get,n=t&&"isReactWarning"in t&&t.isReactWarning,n?e.props.ref:e.props.ref||e.ref)}function jW(e){const t=e+"CollectionProvider",[n,r]=jn(t),[i,a]=n(t,{collectionRef:{current:null},itemMap:new Map}),s=w=>{const{scope:x,children:C}=w,O=V.useRef(null),A=V.useRef(new Map).current;return E.jsx(i,{scope:x,itemMap:A,collectionRef:O,children:C})};s.displayName=t;const u=e+"CollectionSlot",c=zh(u),d=V.forwardRef((w,x)=>{const{scope:C,children:O}=w,A=a(u,C),T=Re(x,A.collectionRef);return E.jsx(c,{ref:T,children:O})});d.displayName=u;const p=e+"CollectionItemSlot",m="data-radix-collection-item",v=zh(p),b=V.forwardRef((w,x)=>{const{scope:C,children:O,...A}=w,T=V.useRef(null),R=Re(x,T),M=a(p,C);return V.useEffect(()=>(M.itemMap.set(T,{ref:T,...A}),()=>void M.itemMap.delete(T))),E.jsx(v,{[m]:"",ref:R,children:O})});b.displayName=p;function S(w){const x=a(e+"CollectionConsumer",w);return V.useCallback(()=>{const O=x.collectionRef.current;if(!O)return[];const A=Array.from(O.querySelectorAll(`[${m}]`));return Array.from(x.itemMap.values()).sort((M,P)=>A.indexOf(M.ref.current)-A.indexOf(P.ref.current))},[x.collectionRef,x.itemMap])}return[{Provider:s,Slot:d,ItemSlot:b},S,r]}var NW=["a","button","div","form","h2","h3","img","input","label","li","nav","ol","p","select","span","svg","ul"],Dn=NW.reduce((e,t)=>{const n=zh(`Primitive.${t}`),r=g.forwardRef((i,a)=>{const{asChild:s,...u}=i,c=s?n:t;return typeof window<"u"&&(window[Symbol.for("radix-ui")]=!0),E.jsx(c,{...u,ref:a})});return r.displayName=`Primitive.${t}`,{...e,[t]:r}},{});function DW(e,t){e&&ba.flushSync(()=>e.dispatchEvent(t))}var kW="DismissableLayer",M_="dismissableLayer.update",$W="dismissableLayer.pointerDownOutside",IW="dismissableLayer.focusOutside",yP,DB=g.createContext({layers:new Set,layersWithOutsidePointerEventsDisabled:new Set,branches:new Set}),kB=g.forwardRef((e,t)=>{const{disableOutsidePointerEvents:n=!1,onEscapeKeyDown:r,onPointerDownOutside:i,onFocusOutside:a,onInteractOutside:s,onDismiss:u,...c}=e,d=g.useContext(DB),[p,m]=g.useState(null),v=(p==null?void 0:p.ownerDocument)??(globalThis==null?void 0:globalThis.document),[,b]=g.useState({}),S=Re(t,P=>m(P)),w=Array.from(d.layers),[x]=[...d.layersWithOutsidePointerEventsDisabled].slice(-1),C=w.indexOf(x),O=p?w.indexOf(p):-1,A=d.layersWithOutsidePointerEventsDisabled.size>0,T=O>=C,R=FW(P=>{const N=P.target,k=[...d.branches].some(H=>H.contains(N));!T||k||(i==null||i(P),s==null||s(P),P.defaultPrevented||u==null||u())},v),M=qW(P=>{const N=P.target;[...d.branches].some(H=>H.contains(N))||(a==null||a(P),s==null||s(P),P.defaultPrevented||u==null||u())},v);return rd(P=>{O===d.layers.size-1&&(r==null||r(P),!P.defaultPrevented&&u&&(P.preventDefault(),u()))},v),g.useEffect(()=>{if(p)return n&&(d.layersWithOutsidePointerEventsDisabled.size===0&&(yP=v.body.style.pointerEvents,v.body.style.pointerEvents="none"),d.layersWithOutsidePointerEventsDisabled.add(p)),d.layers.add(p),bP(),()=>{n&&d.layersWithOutsidePointerEventsDisabled.size===1&&(v.body.style.pointerEvents=yP)}},[p,v,n,d]),g.useEffect(()=>()=>{p&&(d.layers.delete(p),d.layersWithOutsidePointerEventsDisabled.delete(p),bP())},[p,d]),g.useEffect(()=>{const P=()=>b({});return document.addEventListener(M_,P),()=>document.removeEventListener(M_,P)},[]),E.jsx(Dn.div,{...c,ref:S,style:{pointerEvents:A?T?"auto":"none":void 0,...e.style},onFocusCapture:be(e.onFocusCapture,M.onFocusCapture),onBlurCapture:be(e.onBlurCapture,M.onBlurCapture),onPointerDownCapture:be(e.onPointerDownCapture,R.onPointerDownCapture)})});kB.displayName=kW;var LW="DismissableLayerBranch",BW=g.forwardRef((e,t)=>{const n=g.useContext(DB),r=g.useRef(null),i=Re(t,r);return g.useEffect(()=>{const a=r.current;if(a)return n.branches.add(a),()=>{n.branches.delete(a)}},[n.branches]),E.jsx(Dn.div,{...e,ref:i})});BW.displayName=LW;function FW(e,t=globalThis==null?void 0:globalThis.document){const n=$t(e),r=g.useRef(!1),i=g.useRef(()=>{});return g.useEffect(()=>{const a=u=>{if(u.target&&!r.current){let c=function(){$B($W,n,d,{discrete:!0})};const d={originalEvent:u};u.pointerType==="touch"?(t.removeEventListener("click",i.current),i.current=c,t.addEventListener("click",i.current,{once:!0})):c()}else t.removeEventListener("click",i.current);r.current=!1},s=window.setTimeout(()=>{t.addEventListener("pointerdown",a)},0);return()=>{window.clearTimeout(s),t.removeEventListener("pointerdown",a),t.removeEventListener("click",i.current)}},[t,n]),{onPointerDownCapture:()=>r.current=!0}}function qW(e,t=globalThis==null?void 0:globalThis.document){const n=$t(e),r=g.useRef(!1);return g.useEffect(()=>{const i=a=>{a.target&&!r.current&&$B(IW,n,{originalEvent:a},{discrete:!1})};return t.addEventListener("focusin",i),()=>t.removeEventListener("focusin",i)},[t,n]),{onFocusCapture:()=>r.current=!0,onBlurCapture:()=>r.current=!1}}function bP(){const e=new CustomEvent(M_);document.dispatchEvent(e)}function $B(e,t,n,{discrete:r}){const i=n.originalEvent.target,a=new CustomEvent(e,{bubbles:!1,cancelable:!0,detail:n});t&&i.addEventListener(e,t,{once:!0}),r?DW(i,a):i.dispatchEvent(a)}var Fb=0;function iv(){g.useEffect(()=>{const e=document.querySelectorAll("[data-radix-focus-guard]");return document.body.insertAdjacentElement("afterbegin",e[0]??xP()),document.body.insertAdjacentElement("beforeend",e[1]??xP()),Fb++,()=>{Fb===1&&document.querySelectorAll("[data-radix-focus-guard]").forEach(t=>t.remove()),Fb--}},[])}function xP(){const e=document.createElement("span");return e.setAttribute("data-radix-focus-guard",""),e.tabIndex=0,e.style.outline="none",e.style.opacity="0",e.style.position="fixed",e.style.pointerEvents="none",e}function zW(e){const t=UW(e),n=g.forwardRef((r,i)=>{const{children:a,...s}=r,u=g.Children.toArray(a),c=u.find(HW);if(c){const d=c.props.children,p=u.map(m=>m===c?g.Children.count(d)>1?g.Children.only(null):g.isValidElement(d)?d.props.children:null:m);return E.jsx(t,{...s,ref:i,children:g.isValidElement(d)?g.cloneElement(d,void 0,p):null})}return E.jsx(t,{...s,ref:i,children:a})});return n.displayName=`${e}.Slot`,n}function UW(e){const t=g.forwardRef((n,r)=>{const{children:i,...a}=n,s=g.isValidElement(i)?WW(i):void 0,u=Re(s,r);if(g.isValidElement(i)){const c=GW(a,i.props);return i.type!==g.Fragment&&(c.ref=u),g.cloneElement(i,c)}return g.Children.count(i)>1?g.Children.only(null):null});return t.displayName=`${e}.SlotClone`,t}var VW=Symbol("radix.slottable");function HW(e){return g.isValidElement(e)&&typeof e.type=="function"&&"__radixId"in e.type&&e.type.__radixId===VW}function GW(e,t){const n={...t};for(const r in t){const i=e[r],a=t[r];/^on[A-Z]/.test(r)?i&&a?n[r]=(...u)=>{const c=a(...u);return i(...u),c}:i&&(n[r]=i):r==="style"?n[r]={...i,...a}:r==="className"&&(n[r]=[i,a].filter(Boolean).join(" "))}return{...e,...n}}function WW(e){var r,i;let t=(r=Object.getOwnPropertyDescriptor(e.props,"ref"))==null?void 0:r.get,n=t&&"isReactWarning"in t&&t.isReactWarning;return n?e.ref:(t=(i=Object.getOwnPropertyDescriptor(e,"ref"))==null?void 0:i.get,n=t&&"isReactWarning"in t&&t.isReactWarning,n?e.props.ref:e.props.ref||e.ref)}var KW=["a","button","div","form","h2","h3","img","input","label","li","nav","ol","p","select","span","svg","ul"],YW=KW.reduce((e,t)=>{const n=zW(`Primitive.${t}`),r=g.forwardRef((i,a)=>{const{asChild:s,...u}=i,c=s?n:t;return typeof window<"u"&&(window[Symbol.for("radix-ui")]=!0),E.jsx(c,{...u,ref:a})});return r.displayName=`Primitive.${t}`,{...e,[t]:r}},{}),qb="focusScope.autoFocusOnMount",zb="focusScope.autoFocusOnUnmount",wP={bubbles:!1,cancelable:!0},ZW="FocusScope",ud=g.forwardRef((e,t)=>{const{loop:n=!1,trapped:r=!1,onMountAutoFocus:i,onUnmountAutoFocus:a,...s}=e,[u,c]=g.useState(null),d=$t(i),p=$t(a),m=g.useRef(null),v=Re(t,w=>c(w)),b=g.useRef({paused:!1,pause(){this.paused=!0},resume(){this.paused=!1}}).current;g.useEffect(()=>{if(r){let w=function(A){if(b.paused||!u)return;const T=A.target;u.contains(T)?m.current=T:Za(m.current,{select:!0})},x=function(A){if(b.paused||!u)return;const T=A.relatedTarget;T!==null&&(u.contains(T)||Za(m.current,{select:!0}))},C=function(A){if(document.activeElement===document.body)for(const R of A)R.removedNodes.length>0&&Za(u)};document.addEventListener("focusin",w),document.addEventListener("focusout",x);const O=new MutationObserver(C);return u&&O.observe(u,{childList:!0,subtree:!0}),()=>{document.removeEventListener("focusin",w),document.removeEventListener("focusout",x),O.disconnect()}}},[r,u,b.paused]),g.useEffect(()=>{if(u){_P.add(b);const w=document.activeElement;if(!u.contains(w)){const C=new CustomEvent(qb,wP);u.addEventListener(qb,d),u.dispatchEvent(C),C.defaultPrevented||(XW(nK(IB(u)),{select:!0}),document.activeElement===w&&Za(u))}return()=>{u.removeEventListener(qb,d),setTimeout(()=>{const C=new CustomEvent(zb,wP);u.addEventListener(zb,p),u.dispatchEvent(C),C.defaultPrevented||Za(w??document.body,{select:!0}),u.removeEventListener(zb,p),_P.remove(b)},0)}}},[u,d,p,b]);const S=g.useCallback(w=>{if(!n&&!r||b.paused)return;const x=w.key==="Tab"&&!w.altKey&&!w.ctrlKey&&!w.metaKey,C=document.activeElement;if(x&&C){const O=w.currentTarget,[A,T]=QW(O);A&&T?!w.shiftKey&&C===T?(w.preventDefault(),n&&Za(A,{select:!0})):w.shiftKey&&C===A&&(w.preventDefault(),n&&Za(T,{select:!0})):C===O&&w.preventDefault()}},[n,r,b.paused]);return E.jsx(YW.div,{tabIndex:-1,...s,ref:v,onKeyDown:S})});ud.displayName=ZW;function XW(e,{select:t=!1}={}){const n=document.activeElement;for(const r of e)if(Za(r,{select:t}),document.activeElement!==n)return}function QW(e){const t=IB(e),n=SP(t,e),r=SP(t.reverse(),e);return[n,r]}function IB(e){const t=[],n=document.createTreeWalker(e,NodeFilter.SHOW_ELEMENT,{acceptNode:r=>{const i=r.tagName==="INPUT"&&r.type==="hidden";return r.disabled||r.hidden||i?NodeFilter.FILTER_SKIP:r.tabIndex>=0?NodeFilter.FILTER_ACCEPT:NodeFilter.FILTER_SKIP}});for(;n.nextNode();)t.push(n.currentNode);return t}function SP(e,t){for(const n of e)if(!JW(n,{upTo:t}))return n}function JW(e,{upTo:t}){if(getComputedStyle(e).visibility==="hidden")return!0;for(;e;){if(t!==void 0&&e===t)return!1;if(getComputedStyle(e).display==="none")return!0;e=e.parentElement}return!1}function eK(e){return e instanceof HTMLInputElement&&"select"in e}function Za(e,{select:t=!1}={}){if(e&&e.focus){const n=document.activeElement;e.focus({preventScroll:!0}),e!==n&&eK(e)&&t&&e.select()}}var _P=tK();function tK(){let e=[];return{add(t){const n=e[0];t!==n&&(n==null||n.pause()),e=EP(e,t),e.unshift(t)},remove(t){var n;e=EP(e,t),(n=e[0])==null||n.resume()}}}function EP(e,t){const n=[...e],r=n.indexOf(t);return r!==-1&&n.splice(r,1),n}function nK(e){return e.filter(t=>t.tagName!=="A")}const rK=["top","right","bottom","left"],so=Math.min,pr=Math.max,Uh=Math.round,Kp=Math.floor,gi=e=>({x:e,y:e}),iK={left:"right",right:"left",bottom:"top",top:"bottom"},aK={start:"end",end:"start"};function P_(e,t,n){return pr(e,so(t,n))}function da(e,t){return typeof e=="function"?e(t):e}function pa(e){return e.split("-")[0]}function uu(e){return e.split("-")[1]}function KE(e){return e==="x"?"y":"x"}function YE(e){return e==="y"?"height":"width"}function oa(e){return["top","bottom"].includes(pa(e))?"y":"x"}function ZE(e){return KE(oa(e))}function oK(e,t,n){n===void 0&&(n=!1);const r=uu(e),i=ZE(e),a=YE(i);let s=i==="x"?r===(n?"end":"start")?"right":"left":r==="start"?"bottom":"top";return t.reference[a]>t.floating[a]&&(s=Vh(s)),[s,Vh(s)]}function sK(e){const t=Vh(e);return[j_(e),t,j_(t)]}function j_(e){return e.replace(/start|end/g,t=>aK[t])}function lK(e,t,n){const r=["left","right"],i=["right","left"],a=["top","bottom"],s=["bottom","top"];switch(e){case"top":case"bottom":return n?t?i:r:t?r:i;case"left":case"right":return t?a:s;default:return[]}}function uK(e,t,n,r){const i=uu(e);let a=lK(pa(e),n==="start",r);return i&&(a=a.map(s=>s+"-"+i),t&&(a=a.concat(a.map(j_)))),a}function Vh(e){return e.replace(/left|right|bottom|top/g,t=>iK[t])}function cK(e){return{top:0,right:0,bottom:0,left:0,...e}}function LB(e){return typeof e!="number"?cK(e):{top:e,right:e,bottom:e,left:e}}function Hh(e){const{x:t,y:n,width:r,height:i}=e;return{width:r,height:i,top:n,left:t,right:t+r,bottom:n+i,x:t,y:n}}function CP(e,t,n){let{reference:r,floating:i}=e;const a=oa(t),s=ZE(t),u=YE(s),c=pa(t),d=a==="y",p=r.x+r.width/2-i.width/2,m=r.y+r.height/2-i.height/2,v=r[u]/2-i[u]/2;let b;switch(c){case"top":b={x:p,y:r.y-i.height};break;case"bottom":b={x:p,y:r.y+r.height};break;case"right":b={x:r.x+r.width,y:m};break;case"left":b={x:r.x-i.width,y:m};break;default:b={x:r.x,y:r.y}}switch(uu(t)){case"start":b[s]-=v*(n&&d?-1:1);break;case"end":b[s]+=v*(n&&d?-1:1);break}return b}const fK=async(e,t,n)=>{const{placement:r="bottom",strategy:i="absolute",middleware:a=[],platform:s}=n,u=a.filter(Boolean),c=await(s.isRTL==null?void 0:s.isRTL(t));let d=await s.getElementRects({reference:e,floating:t,strategy:i}),{x:p,y:m}=CP(d,r,c),v=r,b={},S=0;for(let w=0;w<u.length;w++){const{name:x,fn:C}=u[w],{x:O,y:A,data:T,reset:R}=await C({x:p,y:m,initialPlacement:r,placement:v,strategy:i,middlewareData:b,rects:d,platform:s,elements:{reference:e,floating:t}});p=O??p,m=A??m,b={...b,[x]:{...b[x],...T}},R&&S<=50&&(S++,typeof R=="object"&&(R.placement&&(v=R.placement),R.rects&&(d=R.rects===!0?await s.getElementRects({reference:e,floating:t,strategy:i}):R.rects),{x:p,y:m}=CP(d,v,c)),w=-1)}return{x:p,y:m,placement:v,strategy:i,middlewareData:b}};async function hf(e,t){var n;t===void 0&&(t={});const{x:r,y:i,platform:a,rects:s,elements:u,strategy:c}=e,{boundary:d="clippingAncestors",rootBoundary:p="viewport",elementContext:m="floating",altBoundary:v=!1,padding:b=0}=da(t,e),S=LB(b),x=u[v?m==="floating"?"reference":"floating":m],C=Hh(await a.getClippingRect({element:(n=await(a.isElement==null?void 0:a.isElement(x)))==null||n?x:x.contextElement||await(a.getDocumentElement==null?void 0:a.getDocumentElement(u.floating)),boundary:d,rootBoundary:p,strategy:c})),O=m==="floating"?{x:r,y:i,width:s.floating.width,height:s.floating.height}:s.reference,A=await(a.getOffsetParent==null?void 0:a.getOffsetParent(u.floating)),T=await(a.isElement==null?void 0:a.isElement(A))?await(a.getScale==null?void 0:a.getScale(A))||{x:1,y:1}:{x:1,y:1},R=Hh(a.convertOffsetParentRelativeRectToViewportRelativeRect?await a.convertOffsetParentRelativeRectToViewportRelativeRect({elements:u,rect:O,offsetParent:A,strategy:c}):O);return{top:(C.top-R.top+S.top)/T.y,bottom:(R.bottom-C.bottom+S.bottom)/T.y,left:(C.left-R.left+S.left)/T.x,right:(R.right-C.right+S.right)/T.x}}const dK=e=>({name:"arrow",options:e,async fn(t){const{x:n,y:r,placement:i,rects:a,platform:s,elements:u,middlewareData:c}=t,{element:d,padding:p=0}=da(e,t)||{};if(d==null)return{};const m=LB(p),v={x:n,y:r},b=ZE(i),S=YE(b),w=await s.getDimensions(d),x=b==="y",C=x?"top":"left",O=x?"bottom":"right",A=x?"clientHeight":"clientWidth",T=a.reference[S]+a.reference[b]-v[b]-a.floating[S],R=v[b]-a.reference[b],M=await(s.getOffsetParent==null?void 0:s.getOffsetParent(d));let P=M?M[A]:0;(!P||!await(s.isElement==null?void 0:s.isElement(M)))&&(P=u.floating[A]||a.floating[S]);const N=T/2-R/2,k=P/2-w[S]/2-1,H=so(m[C],k),L=so(m[O],k),F=H,z=P-w[S]-L,Z=P/2-w[S]/2+N,Y=P_(F,Z,z),I=!c.arrow&&uu(i)!=null&&Z!==Y&&a.reference[S]/2-(Z<F?H:L)-w[S]/2<0,W=I?Z<F?Z-F:Z-z:0;return{[b]:v[b]+W,data:{[b]:Y,centerOffset:Z-Y-W,...I&&{alignmentOffset:W}},reset:I}}}),pK=function(e){return e===void 0&&(e={}),{name:"flip",options:e,async fn(t){var n,r;const{placement:i,middlewareData:a,rects:s,initialPlacement:u,platform:c,elements:d}=t,{mainAxis:p=!0,crossAxis:m=!0,fallbackPlacements:v,fallbackStrategy:b="bestFit",fallbackAxisSideDirection:S="none",flipAlignment:w=!0,...x}=da(e,t);if((n=a.arrow)!=null&&n.alignmentOffset)return{};const C=pa(i),O=oa(u),A=pa(u)===u,T=await(c.isRTL==null?void 0:c.isRTL(d.floating)),R=v||(A||!w?[Vh(u)]:sK(u)),M=S!=="none";!v&&M&&R.push(...uK(u,w,S,T));const P=[u,...R],N=await hf(t,x),k=[];let H=((r=a.flip)==null?void 0:r.overflows)||[];if(p&&k.push(N[C]),m){const Y=oK(i,s,T);k.push(N[Y[0]],N[Y[1]])}if(H=[...H,{placement:i,overflows:k}],!k.every(Y=>Y<=0)){var L,F;const Y=(((L=a.flip)==null?void 0:L.index)||0)+1,I=P[Y];if(I){var z;const B=m==="alignment"?O!==oa(I):!1,Q=((z=H[0])==null?void 0:z.overflows[0])>0;if(!B||Q)return{data:{index:Y,overflows:H},reset:{placement:I}}}let W=(F=H.filter(B=>B.overflows[0]<=0).sort((B,Q)=>B.overflows[1]-Q.overflows[1])[0])==null?void 0:F.placement;if(!W)switch(b){case"bestFit":{var Z;const B=(Z=H.filter(Q=>{if(M){const $=oa(Q.placement);return $===O||$==="y"}return!0}).map(Q=>[Q.placement,Q.overflows.filter($=>$>0).reduce(($,J)=>$+J,0)]).sort((Q,$)=>Q[1]-$[1])[0])==null?void 0:Z[0];B&&(W=B);break}case"initialPlacement":W=u;break}if(i!==W)return{reset:{placement:W}}}return{}}}};function OP(e,t){return{top:e.top-t.height,right:e.right-t.width,bottom:e.bottom-t.height,left:e.left-t.width}}function AP(e){return rK.some(t=>e[t]>=0)}const hK=function(e){return e===void 0&&(e={}),{name:"hide",options:e,async fn(t){const{rects:n}=t,{strategy:r="referenceHidden",...i}=da(e,t);switch(r){case"referenceHidden":{const a=await hf(t,{...i,elementContext:"reference"}),s=OP(a,n.reference);return{data:{referenceHiddenOffsets:s,referenceHidden:AP(s)}}}case"escaped":{const a=await hf(t,{...i,altBoundary:!0}),s=OP(a,n.floating);return{data:{escapedOffsets:s,escaped:AP(s)}}}default:return{}}}}};async function mK(e,t){const{placement:n,platform:r,elements:i}=e,a=await(r.isRTL==null?void 0:r.isRTL(i.floating)),s=pa(n),u=uu(n),c=oa(n)==="y",d=["left","top"].includes(s)?-1:1,p=a&&c?-1:1,m=da(t,e);let{mainAxis:v,crossAxis:b,alignmentAxis:S}=typeof m=="number"?{mainAxis:m,crossAxis:0,alignmentAxis:null}:{mainAxis:m.mainAxis||0,crossAxis:m.crossAxis||0,alignmentAxis:m.alignmentAxis};return u&&typeof S=="number"&&(b=u==="end"?S*-1:S),c?{x:b*p,y:v*d}:{x:v*d,y:b*p}}const vK=function(e){return e===void 0&&(e=0),{name:"offset",options:e,async fn(t){var n,r;const{x:i,y:a,placement:s,middlewareData:u}=t,c=await mK(t,e);return s===((n=u.offset)==null?void 0:n.placement)&&(r=u.arrow)!=null&&r.alignmentOffset?{}:{x:i+c.x,y:a+c.y,data:{...c,placement:s}}}}},gK=function(e){return e===void 0&&(e={}),{name:"shift",options:e,async fn(t){const{x:n,y:r,placement:i}=t,{mainAxis:a=!0,crossAxis:s=!1,limiter:u={fn:x=>{let{x:C,y:O}=x;return{x:C,y:O}}},...c}=da(e,t),d={x:n,y:r},p=await hf(t,c),m=oa(pa(i)),v=KE(m);let b=d[v],S=d[m];if(a){const x=v==="y"?"top":"left",C=v==="y"?"bottom":"right",O=b+p[x],A=b-p[C];b=P_(O,b,A)}if(s){const x=m==="y"?"top":"left",C=m==="y"?"bottom":"right",O=S+p[x],A=S-p[C];S=P_(O,S,A)}const w=u.fn({...t,[v]:b,[m]:S});return{...w,data:{x:w.x-n,y:w.y-r,enabled:{[v]:a,[m]:s}}}}}},yK=function(e){return e===void 0&&(e={}),{options:e,fn(t){const{x:n,y:r,placement:i,rects:a,middlewareData:s}=t,{offset:u=0,mainAxis:c=!0,crossAxis:d=!0}=da(e,t),p={x:n,y:r},m=oa(i),v=KE(m);let b=p[v],S=p[m];const w=da(u,t),x=typeof w=="number"?{mainAxis:w,crossAxis:0}:{mainAxis:0,crossAxis:0,...w};if(c){const A=v==="y"?"height":"width",T=a.reference[v]-a.floating[A]+x.mainAxis,R=a.reference[v]+a.reference[A]-x.mainAxis;b<T?b=T:b>R&&(b=R)}if(d){var C,O;const A=v==="y"?"width":"height",T=["top","left"].includes(pa(i)),R=a.reference[m]-a.floating[A]+(T&&((C=s.offset)==null?void 0:C[m])||0)+(T?0:x.crossAxis),M=a.reference[m]+a.reference[A]+(T?0:((O=s.offset)==null?void 0:O[m])||0)-(T?x.crossAxis:0);S<R?S=R:S>M&&(S=M)}return{[v]:b,[m]:S}}}},bK=function(e){return e===void 0&&(e={}),{name:"size",options:e,async fn(t){var n,r;const{placement:i,rects:a,platform:s,elements:u}=t,{apply:c=()=>{},...d}=da(e,t),p=await hf(t,d),m=pa(i),v=uu(i),b=oa(i)==="y",{width:S,height:w}=a.floating;let x,C;m==="top"||m==="bottom"?(x=m,C=v===(await(s.isRTL==null?void 0:s.isRTL(u.floating))?"start":"end")?"left":"right"):(C=m,x=v==="end"?"top":"bottom");const O=w-p.top-p.bottom,A=S-p.left-p.right,T=so(w-p[x],O),R=so(S-p[C],A),M=!t.middlewareData.shift;let P=T,N=R;if((n=t.middlewareData.shift)!=null&&n.enabled.x&&(N=A),(r=t.middlewareData.shift)!=null&&r.enabled.y&&(P=O),M&&!v){const H=pr(p.left,0),L=pr(p.right,0),F=pr(p.top,0),z=pr(p.bottom,0);b?N=S-2*(H!==0||L!==0?H+L:pr(p.left,p.right)):P=w-2*(F!==0||z!==0?F+z:pr(p.top,p.bottom))}await c({...t,availableWidth:N,availableHeight:P});const k=await s.getDimensions(u.floating);return S!==k.width||w!==k.height?{reset:{rects:!0}}:{}}}};function av(){return typeof window<"u"}function cu(e){return BB(e)?(e.nodeName||"").toLowerCase():"#document"}function vr(e){var t;return(e==null||(t=e.ownerDocument)==null?void 0:t.defaultView)||window}function Ti(e){var t;return(t=(BB(e)?e.ownerDocument:e.document)||window.document)==null?void 0:t.documentElement}function BB(e){return av()?e instanceof Node||e instanceof vr(e).Node:!1}function Jr(e){return av()?e instanceof Element||e instanceof vr(e).Element:!1}function Ci(e){return av()?e instanceof HTMLElement||e instanceof vr(e).HTMLElement:!1}function RP(e){return!av()||typeof ShadowRoot>"u"?!1:e instanceof ShadowRoot||e instanceof vr(e).ShadowRoot}function cd(e){const{overflow:t,overflowX:n,overflowY:r,display:i}=ei(e);return/auto|scroll|overlay|hidden|clip/.test(t+r+n)&&!["inline","contents"].includes(i)}function xK(e){return["table","td","th"].includes(cu(e))}function ov(e){return[":popover-open",":modal"].some(t=>{try{return e.matches(t)}catch{return!1}})}function XE(e){const t=QE(),n=Jr(e)?ei(e):e;return["transform","translate","scale","rotate","perspective"].some(r=>n[r]?n[r]!=="none":!1)||(n.containerType?n.containerType!=="normal":!1)||!t&&(n.backdropFilter?n.backdropFilter!=="none":!1)||!t&&(n.filter?n.filter!=="none":!1)||["transform","translate","scale","rotate","perspective","filter"].some(r=>(n.willChange||"").includes(r))||["paint","layout","strict","content"].some(r=>(n.contain||"").includes(r))}function wK(e){let t=lo(e);for(;Ci(t)&&!Dl(t);){if(XE(t))return t;if(ov(t))return null;t=lo(t)}return null}function QE(){return typeof CSS>"u"||!CSS.supports?!1:CSS.supports("-webkit-backdrop-filter","none")}function Dl(e){return["html","body","#document"].includes(cu(e))}function ei(e){return vr(e).getComputedStyle(e)}function sv(e){return Jr(e)?{scrollLeft:e.scrollLeft,scrollTop:e.scrollTop}:{scrollLeft:e.scrollX,scrollTop:e.scrollY}}function lo(e){if(cu(e)==="html")return e;const t=e.assignedSlot||e.parentNode||RP(e)&&e.host||Ti(e);return RP(t)?t.host:t}function FB(e){const t=lo(e);return Dl(t)?e.ownerDocument?e.ownerDocument.body:e.body:Ci(t)&&cd(t)?t:FB(t)}function mf(e,t,n){var r;t===void 0&&(t=[]),n===void 0&&(n=!0);const i=FB(e),a=i===((r=e.ownerDocument)==null?void 0:r.body),s=vr(i);if(a){const u=N_(s);return t.concat(s,s.visualViewport||[],cd(i)?i:[],u&&n?mf(u):[])}return t.concat(i,mf(i,[],n))}function N_(e){return e.parent&&Object.getPrototypeOf(e.parent)?e.frameElement:null}function qB(e){const t=ei(e);let n=parseFloat(t.width)||0,r=parseFloat(t.height)||0;const i=Ci(e),a=i?e.offsetWidth:n,s=i?e.offsetHeight:r,u=Uh(n)!==a||Uh(r)!==s;return u&&(n=a,r=s),{width:n,height:r,$:u}}function JE(e){return Jr(e)?e:e.contextElement}function Cl(e){const t=JE(e);if(!Ci(t))return gi(1);const n=t.getBoundingClientRect(),{width:r,height:i,$:a}=qB(t);let s=(a?Uh(n.width):n.width)/r,u=(a?Uh(n.height):n.height)/i;return(!s||!Number.isFinite(s))&&(s=1),(!u||!Number.isFinite(u))&&(u=1),{x:s,y:u}}const SK=gi(0);function zB(e){const t=vr(e);return!QE()||!t.visualViewport?SK:{x:t.visualViewport.offsetLeft,y:t.visualViewport.offsetTop}}function _K(e,t,n){return t===void 0&&(t=!1),!n||t&&n!==vr(e)?!1:t}function ss(e,t,n,r){t===void 0&&(t=!1),n===void 0&&(n=!1);const i=e.getBoundingClientRect(),a=JE(e);let s=gi(1);t&&(r?Jr(r)&&(s=Cl(r)):s=Cl(e));const u=_K(a,n,r)?zB(a):gi(0);let c=(i.left+u.x)/s.x,d=(i.top+u.y)/s.y,p=i.width/s.x,m=i.height/s.y;if(a){const v=vr(a),b=r&&Jr(r)?vr(r):r;let S=v,w=N_(S);for(;w&&r&&b!==S;){const x=Cl(w),C=w.getBoundingClientRect(),O=ei(w),A=C.left+(w.clientLeft+parseFloat(O.paddingLeft))*x.x,T=C.top+(w.clientTop+parseFloat(O.paddingTop))*x.y;c*=x.x,d*=x.y,p*=x.x,m*=x.y,c+=A,d+=T,S=vr(w),w=N_(S)}}return Hh({width:p,height:m,x:c,y:d})}function eC(e,t){const n=sv(e).scrollLeft;return t?t.left+n:ss(Ti(e)).left+n}function UB(e,t,n){n===void 0&&(n=!1);const r=e.getBoundingClientRect(),i=r.left+t.scrollLeft-(n?0:eC(e,r)),a=r.top+t.scrollTop;return{x:i,y:a}}function EK(e){let{elements:t,rect:n,offsetParent:r,strategy:i}=e;const a=i==="fixed",s=Ti(r),u=t?ov(t.floating):!1;if(r===s||u&&a)return n;let c={scrollLeft:0,scrollTop:0},d=gi(1);const p=gi(0),m=Ci(r);if((m||!m&&!a)&&((cu(r)!=="body"||cd(s))&&(c=sv(r)),Ci(r))){const b=ss(r);d=Cl(r),p.x=b.x+r.clientLeft,p.y=b.y+r.clientTop}const v=s&&!m&&!a?UB(s,c,!0):gi(0);return{width:n.width*d.x,height:n.height*d.y,x:n.x*d.x-c.scrollLeft*d.x+p.x+v.x,y:n.y*d.y-c.scrollTop*d.y+p.y+v.y}}function CK(e){return Array.from(e.getClientRects())}function OK(e){const t=Ti(e),n=sv(e),r=e.ownerDocument.body,i=pr(t.scrollWidth,t.clientWidth,r.scrollWidth,r.clientWidth),a=pr(t.scrollHeight,t.clientHeight,r.scrollHeight,r.clientHeight);let s=-n.scrollLeft+eC(e);const u=-n.scrollTop;return ei(r).direction==="rtl"&&(s+=pr(t.clientWidth,r.clientWidth)-i),{width:i,height:a,x:s,y:u}}function AK(e,t){const n=vr(e),r=Ti(e),i=n.visualViewport;let a=r.clientWidth,s=r.clientHeight,u=0,c=0;if(i){a=i.width,s=i.height;const d=QE();(!d||d&&t==="fixed")&&(u=i.offsetLeft,c=i.offsetTop)}return{width:a,height:s,x:u,y:c}}function RK(e,t){const n=ss(e,!0,t==="fixed"),r=n.top+e.clientTop,i=n.left+e.clientLeft,a=Ci(e)?Cl(e):gi(1),s=e.clientWidth*a.x,u=e.clientHeight*a.y,c=i*a.x,d=r*a.y;return{width:s,height:u,x:c,y:d}}function TP(e,t,n){let r;if(t==="viewport")r=AK(e,n);else if(t==="document")r=OK(Ti(e));else if(Jr(t))r=RK(t,n);else{const i=zB(e);r={x:t.x-i.x,y:t.y-i.y,width:t.width,height:t.height}}return Hh(r)}function VB(e,t){const n=lo(e);return n===t||!Jr(n)||Dl(n)?!1:ei(n).position==="fixed"||VB(n,t)}function TK(e,t){const n=t.get(e);if(n)return n;let r=mf(e,[],!1).filter(u=>Jr(u)&&cu(u)!=="body"),i=null;const a=ei(e).position==="fixed";let s=a?lo(e):e;for(;Jr(s)&&!Dl(s);){const u=ei(s),c=XE(s);!c&&u.position==="fixed"&&(i=null),(a?!c&&!i:!c&&u.position==="static"&&!!i&&["absolute","fixed"].includes(i.position)||cd(s)&&!c&&VB(e,s))?r=r.filter(p=>p!==s):i=u,s=lo(s)}return t.set(e,r),r}function MK(e){let{element:t,boundary:n,rootBoundary:r,strategy:i}=e;const s=[...n==="clippingAncestors"?ov(t)?[]:TK(t,this._c):[].concat(n),r],u=s[0],c=s.reduce((d,p)=>{const m=TP(t,p,i);return d.top=pr(m.top,d.top),d.right=so(m.right,d.right),d.bottom=so(m.bottom,d.bottom),d.left=pr(m.left,d.left),d},TP(t,u,i));return{width:c.right-c.left,height:c.bottom-c.top,x:c.left,y:c.top}}function PK(e){const{width:t,height:n}=qB(e);return{width:t,height:n}}function jK(e,t,n){const r=Ci(t),i=Ti(t),a=n==="fixed",s=ss(e,!0,a,t);let u={scrollLeft:0,scrollTop:0};const c=gi(0);function d(){c.x=eC(i)}if(r||!r&&!a)if((cu(t)!=="body"||cd(i))&&(u=sv(t)),r){const b=ss(t,!0,a,t);c.x=b.x+t.clientLeft,c.y=b.y+t.clientTop}else i&&d();a&&!r&&i&&d();const p=i&&!r&&!a?UB(i,u):gi(0),m=s.left+u.scrollLeft-c.x-p.x,v=s.top+u.scrollTop-c.y-p.y;return{x:m,y:v,width:s.width,height:s.height}}function Ub(e){return ei(e).position==="static"}function MP(e,t){if(!Ci(e)||ei(e).position==="fixed")return null;if(t)return t(e);let n=e.offsetParent;return Ti(e)===n&&(n=n.ownerDocument.body),n}function HB(e,t){const n=vr(e);if(ov(e))return n;if(!Ci(e)){let i=lo(e);for(;i&&!Dl(i);){if(Jr(i)&&!Ub(i))return i;i=lo(i)}return n}let r=MP(e,t);for(;r&&xK(r)&&Ub(r);)r=MP(r,t);return r&&Dl(r)&&Ub(r)&&!XE(r)?n:r||wK(e)||n}const NK=async function(e){const t=this.getOffsetParent||HB,n=this.getDimensions,r=await n(e.floating);return{reference:jK(e.reference,await t(e.floating),e.strategy),floating:{x:0,y:0,width:r.width,height:r.height}}};function DK(e){return ei(e).direction==="rtl"}const kK={convertOffsetParentRelativeRectToViewportRelativeRect:EK,getDocumentElement:Ti,getClippingRect:MK,getOffsetParent:HB,getElementRects:NK,getClientRects:CK,getDimensions:PK,getScale:Cl,isElement:Jr,isRTL:DK};function GB(e,t){return e.x===t.x&&e.y===t.y&&e.width===t.width&&e.height===t.height}function $K(e,t){let n=null,r;const i=Ti(e);function a(){var u;clearTimeout(r),(u=n)==null||u.disconnect(),n=null}function s(u,c){u===void 0&&(u=!1),c===void 0&&(c=1),a();const d=e.getBoundingClientRect(),{left:p,top:m,width:v,height:b}=d;if(u||t(),!v||!b)return;const S=Kp(m),w=Kp(i.clientWidth-(p+v)),x=Kp(i.clientHeight-(m+b)),C=Kp(p),A={rootMargin:-S+"px "+-w+"px "+-x+"px "+-C+"px",threshold:pr(0,so(1,c))||1};let T=!0;function R(M){const P=M[0].intersectionRatio;if(P!==c){if(!T)return s();P?s(!1,P):r=setTimeout(()=>{s(!1,1e-7)},1e3)}P===1&&!GB(d,e.getBoundingClientRect())&&s(),T=!1}try{n=new IntersectionObserver(R,{...A,root:i.ownerDocument})}catch{n=new IntersectionObserver(R,A)}n.observe(e)}return s(!0),a}function IK(e,t,n,r){r===void 0&&(r={});const{ancestorScroll:i=!0,ancestorResize:a=!0,elementResize:s=typeof ResizeObserver=="function",layoutShift:u=typeof IntersectionObserver=="function",animationFrame:c=!1}=r,d=JE(e),p=i||a?[...d?mf(d):[],...mf(t)]:[];p.forEach(C=>{i&&C.addEventListener("scroll",n,{passive:!0}),a&&C.addEventListener("resize",n)});const m=d&&u?$K(d,n):null;let v=-1,b=null;s&&(b=new ResizeObserver(C=>{let[O]=C;O&&O.target===d&&b&&(b.unobserve(t),cancelAnimationFrame(v),v=requestAnimationFrame(()=>{var A;(A=b)==null||A.observe(t)})),n()}),d&&!c&&b.observe(d),b.observe(t));let S,w=c?ss(e):null;c&&x();function x(){const C=ss(e);w&&!GB(w,C)&&n(),w=C,S=requestAnimationFrame(x)}return n(),()=>{var C;p.forEach(O=>{i&&O.removeEventListener("scroll",n),a&&O.removeEventListener("resize",n)}),m==null||m(),(C=b)==null||C.disconnect(),b=null,c&&cancelAnimationFrame(S)}}const LK=vK,BK=gK,FK=pK,qK=bK,zK=hK,PP=dK,UK=yK,VK=(e,t,n)=>{const r=new Map,i={platform:kK,...n},a={...i.platform,_c:r};return fK(e,t,{...i,platform:a})};var wh=typeof document<"u"?g.useLayoutEffect:g.useEffect;function Gh(e,t){if(e===t)return!0;if(typeof e!=typeof t)return!1;if(typeof e=="function"&&e.toString()===t.toString())return!0;let n,r,i;if(e&&t&&typeof e=="object"){if(Array.isArray(e)){if(n=e.length,n!==t.length)return!1;for(r=n;r--!==0;)if(!Gh(e[r],t[r]))return!1;return!0}if(i=Object.keys(e),n=i.length,n!==Object.keys(t).length)return!1;for(r=n;r--!==0;)if(!{}.hasOwnProperty.call(t,i[r]))return!1;for(r=n;r--!==0;){const a=i[r];if(!(a==="_owner"&&e.$$typeof)&&!Gh(e[a],t[a]))return!1}return!0}return e!==e&&t!==t}function WB(e){return typeof window>"u"?1:(e.ownerDocument.defaultView||window).devicePixelRatio||1}function jP(e,t){const n=WB(e);return Math.round(t*n)/n}function Vb(e){const t=g.useRef(e);return wh(()=>{t.current=e}),t}function HK(e){e===void 0&&(e={});const{placement:t="bottom",strategy:n="absolute",middleware:r=[],platform:i,elements:{reference:a,floating:s}={},transform:u=!0,whileElementsMounted:c,open:d}=e,[p,m]=g.useState({x:0,y:0,strategy:n,placement:t,middlewareData:{},isPositioned:!1}),[v,b]=g.useState(r);Gh(v,r)||b(r);const[S,w]=g.useState(null),[x,C]=g.useState(null),O=g.useCallback(B=>{B!==M.current&&(M.current=B,w(B))},[]),A=g.useCallback(B=>{B!==P.current&&(P.current=B,C(B))},[]),T=a||S,R=s||x,M=g.useRef(null),P=g.useRef(null),N=g.useRef(p),k=c!=null,H=Vb(c),L=Vb(i),F=Vb(d),z=g.useCallback(()=>{if(!M.current||!P.current)return;const B={placement:t,strategy:n,middleware:v};L.current&&(B.platform=L.current),VK(M.current,P.current,B).then(Q=>{const $={...Q,isPositioned:F.current!==!1};Z.current&&!Gh(N.current,$)&&(N.current=$,ba.flushSync(()=>{m($)}))})},[v,t,n,L,F]);wh(()=>{d===!1&&N.current.isPositioned&&(N.current.isPositioned=!1,m(B=>({...B,isPositioned:!1})))},[d]);const Z=g.useRef(!1);wh(()=>(Z.current=!0,()=>{Z.current=!1}),[]),wh(()=>{if(T&&(M.current=T),R&&(P.current=R),T&&R){if(H.current)return H.current(T,R,z);z()}},[T,R,z,H,k]);const Y=g.useMemo(()=>({reference:M,floating:P,setReference:O,setFloating:A}),[O,A]),I=g.useMemo(()=>({reference:T,floating:R}),[T,R]),W=g.useMemo(()=>{const B={position:n,left:0,top:0};if(!I.floating)return B;const Q=jP(I.floating,p.x),$=jP(I.floating,p.y);return u?{...B,transform:"translate("+Q+"px, "+$+"px)",...WB(I.floating)>=1.5&&{willChange:"transform"}}:{position:n,left:Q,top:$}},[n,u,I.floating,p.x,p.y]);return g.useMemo(()=>({...p,update:z,refs:Y,elements:I,floatingStyles:W}),[p,z,Y,I,W])}const GK=e=>{function t(n){return{}.hasOwnProperty.call(n,"current")}return{name:"arrow",options:e,fn(n){const{element:r,padding:i}=typeof e=="function"?e(n):e;return r&&t(r)?r.current!=null?PP({element:r.current,padding:i}).fn(n):{}:r?PP({element:r,padding:i}).fn(n):{}}}},WK=(e,t)=>({...LK(e),options:[e,t]}),KK=(e,t)=>({...BK(e),options:[e,t]}),YK=(e,t)=>({...UK(e),options:[e,t]}),ZK=(e,t)=>({...FK(e),options:[e,t]}),XK=(e,t)=>({...qK(e),options:[e,t]}),QK=(e,t)=>({...zK(e),options:[e,t]}),JK=(e,t)=>({...GK(e),options:[e,t]});function eY(e){const t=tY(e),n=g.forwardRef((r,i)=>{const{children:a,...s}=r,u=g.Children.toArray(a),c=u.find(rY);if(c){const d=c.props.children,p=u.map(m=>m===c?g.Children.count(d)>1?g.Children.only(null):g.isValidElement(d)?d.props.children:null:m);return E.jsx(t,{...s,ref:i,children:g.isValidElement(d)?g.cloneElement(d,void 0,p):null})}return E.jsx(t,{...s,ref:i,children:a})});return n.displayName=`${e}.Slot`,n}function tY(e){const t=g.forwardRef((n,r)=>{const{children:i,...a}=n,s=g.isValidElement(i)?aY(i):void 0,u=Re(s,r);if(g.isValidElement(i)){const c=iY(a,i.props);return i.type!==g.Fragment&&(c.ref=u),g.cloneElement(i,c)}return g.Children.count(i)>1?g.Children.only(null):null});return t.displayName=`${e}.SlotClone`,t}var nY=Symbol("radix.slottable");function rY(e){return g.isValidElement(e)&&typeof e.type=="function"&&"__radixId"in e.type&&e.type.__radixId===nY}function iY(e,t){const n={...t};for(const r in t){const i=e[r],a=t[r];/^on[A-Z]/.test(r)?i&&a?n[r]=(...u)=>{const c=a(...u);return i(...u),c}:i&&(n[r]=i):r==="style"?n[r]={...i,...a}:r==="className"&&(n[r]=[i,a].filter(Boolean).join(" "))}return{...e,...n}}function aY(e){var r,i;let t=(r=Object.getOwnPropertyDescriptor(e.props,"ref"))==null?void 0:r.get,n=t&&"isReactWarning"in t&&t.isReactWarning;return n?e.ref:(t=(i=Object.getOwnPropertyDescriptor(e,"ref"))==null?void 0:i.get,n=t&&"isReactWarning"in t&&t.isReactWarning,n?e.props.ref:e.props.ref||e.ref)}var oY=["a","button","div","form","h2","h3","img","input","label","li","nav","ol","p","select","span","svg","ul"],sY=oY.reduce((e,t)=>{const n=eY(`Primitive.${t}`),r=g.forwardRef((i,a)=>{const{asChild:s,...u}=i,c=s?n:t;return typeof window<"u"&&(window[Symbol.for("radix-ui")]=!0),E.jsx(c,{...u,ref:a})});return r.displayName=`Primitive.${t}`,{...e,[t]:r}},{}),lY="Arrow",KB=g.forwardRef((e,t)=>{const{children:n,width:r=10,height:i=5,...a}=e;return E.jsx(sY.svg,{...a,ref:t,width:r,height:i,viewBox:"0 0 30 10",preserveAspectRatio:"none",children:e.asChild?n:E.jsx("polygon",{points:"0,0 30,0 15,10"})})});KB.displayName=lY;var uY=KB;function cY(e){const t=fY(e),n=g.forwardRef((r,i)=>{const{children:a,...s}=r,u=g.Children.toArray(a),c=u.find(pY);if(c){const d=c.props.children,p=u.map(m=>m===c?g.Children.count(d)>1?g.Children.only(null):g.isValidElement(d)?d.props.children:null:m);return E.jsx(t,{...s,ref:i,children:g.isValidElement(d)?g.cloneElement(d,void 0,p):null})}return E.jsx(t,{...s,ref:i,children:a})});return n.displayName=`${e}.Slot`,n}function fY(e){const t=g.forwardRef((n,r)=>{const{children:i,...a}=n,s=g.isValidElement(i)?mY(i):void 0,u=Re(s,r);if(g.isValidElement(i)){const c=hY(a,i.props);return i.type!==g.Fragment&&(c.ref=u),g.cloneElement(i,c)}return g.Children.count(i)>1?g.Children.only(null):null});return t.displayName=`${e}.SlotClone`,t}var dY=Symbol("radix.slottable");function pY(e){return g.isValidElement(e)&&typeof e.type=="function"&&"__radixId"in e.type&&e.type.__radixId===dY}function hY(e,t){const n={...t};for(const r in t){const i=e[r],a=t[r];/^on[A-Z]/.test(r)?i&&a?n[r]=(...u)=>{const c=a(...u);return i(...u),c}:i&&(n[r]=i):r==="style"?n[r]={...i,...a}:r==="className"&&(n[r]=[i,a].filter(Boolean).join(" "))}return{...e,...n}}function mY(e){var r,i;let t=(r=Object.getOwnPropertyDescriptor(e.props,"ref"))==null?void 0:r.get,n=t&&"isReactWarning"in t&&t.isReactWarning;return n?e.ref:(t=(i=Object.getOwnPropertyDescriptor(e,"ref"))==null?void 0:i.get,n=t&&"isReactWarning"in t&&t.isReactWarning,n?e.props.ref:e.props.ref||e.ref)}var vY=["a","button","div","form","h2","h3","img","input","label","li","nav","ol","p","select","span","svg","ul"],YB=vY.reduce((e,t)=>{const n=cY(`Primitive.${t}`),r=g.forwardRef((i,a)=>{const{asChild:s,...u}=i,c=s?n:t;return typeof window<"u"&&(window[Symbol.for("radix-ui")]=!0),E.jsx(c,{...u,ref:a})});return r.displayName=`Primitive.${t}`,{...e,[t]:r}},{}),tC="Popper",[ZB,fu]=jn(tC),[gY,XB]=ZB(tC),QB=e=>{const{__scopePopper:t,children:n}=e,[r,i]=g.useState(null);return E.jsx(gY,{scope:t,anchor:r,onAnchorChange:i,children:n})};QB.displayName=tC;var JB="PopperAnchor",eF=g.forwardRef((e,t)=>{const{__scopePopper:n,virtualRef:r,...i}=e,a=XB(JB,n),s=g.useRef(null),u=Re(t,s);return g.useEffect(()=>{a.onAnchorChange((r==null?void 0:r.current)||s.current)}),r?null:E.jsx(YB.div,{...i,ref:u})});eF.displayName=JB;var nC="PopperContent",[yY,bY]=ZB(nC),tF=g.forwardRef((e,t)=>{var de,Se,ce,we,_e,se;const{__scopePopper:n,side:r="bottom",sideOffset:i=0,align:a="center",alignOffset:s=0,arrowPadding:u=0,avoidCollisions:c=!0,collisionBoundary:d=[],collisionPadding:p=0,sticky:m="partial",hideWhenDetached:v=!1,updatePositionStrategy:b="optimized",onPlaced:S,...w}=e,x=XB(nC,n),[C,O]=g.useState(null),A=Re(t,Ye=>O(Ye)),[T,R]=g.useState(null),M=WE(T),P=(M==null?void 0:M.width)??0,N=(M==null?void 0:M.height)??0,k=r+(a!=="center"?"-"+a:""),H=typeof p=="number"?p:{top:0,right:0,bottom:0,left:0,...p},L=Array.isArray(d)?d:[d],F=L.length>0,z={padding:H,boundary:L.filter(wY),altBoundary:F},{refs:Z,floatingStyles:Y,placement:I,isPositioned:W,middlewareData:B}=HK({strategy:"fixed",placement:k,whileElementsMounted:(...Ye)=>IK(...Ye,{animationFrame:b==="always"}),elements:{reference:x.anchor},middleware:[WK({mainAxis:i+N,alignmentAxis:s}),c&&KK({mainAxis:!0,crossAxis:!1,limiter:m==="partial"?YK():void 0,...z}),c&&ZK({...z}),XK({...z,apply:({elements:Ye,rects:Be,availableWidth:dt,availableHeight:ft})=>{const{width:it,height:xt}=Be.reference,Tt=Ye.floating.style;Tt.setProperty("--radix-popper-available-width",`${dt}px`),Tt.setProperty("--radix-popper-available-height",`${ft}px`),Tt.setProperty("--radix-popper-anchor-width",`${it}px`),Tt.setProperty("--radix-popper-anchor-height",`${xt}px`)}}),T&&JK({element:T,padding:u}),SY({arrowWidth:P,arrowHeight:N}),v&&QK({strategy:"referenceHidden",...z})]}),[Q,$]=iF(I),J=$t(S);cn(()=>{W&&(J==null||J())},[W,J]);const q=(de=B.arrow)==null?void 0:de.x,X=(Se=B.arrow)==null?void 0:Se.y,oe=((ce=B.arrow)==null?void 0:ce.centerOffset)!==0,[le,ve]=g.useState();return cn(()=>{C&&ve(window.getComputedStyle(C).zIndex)},[C]),E.jsx("div",{ref:Z.setFloating,"data-radix-popper-content-wrapper":"",style:{...Y,transform:W?Y.transform:"translate(0, -200%)",minWidth:"max-content",zIndex:le,"--radix-popper-transform-origin":[(we=B.transformOrigin)==null?void 0:we.x,(_e=B.transformOrigin)==null?void 0:_e.y].join(" "),...((se=B.hide)==null?void 0:se.referenceHidden)&&{visibility:"hidden",pointerEvents:"none"}},dir:e.dir,children:E.jsx(yY,{scope:n,placedSide:Q,onArrowChange:R,arrowX:q,arrowY:X,shouldHideArrow:oe,children:E.jsx(YB.div,{"data-side":Q,"data-align":$,...w,ref:A,style:{...w.style,animation:W?void 0:"none"}})})})});tF.displayName=nC;var nF="PopperArrow",xY={top:"bottom",right:"left",bottom:"top",left:"right"},rF=g.forwardRef(function(t,n){const{__scopePopper:r,...i}=t,a=bY(nF,r),s=xY[a.placedSide];return E.jsx("span",{ref:a.onArrowChange,style:{position:"absolute",left:a.arrowX,top:a.arrowY,[s]:0,transformOrigin:{top:"",right:"0 0",bottom:"center 0",left:"100% 0"}[a.placedSide],transform:{top:"translateY(100%)",right:"translateY(50%) rotate(90deg) translateX(-50%)",bottom:"rotate(180deg)",left:"translateY(50%) rotate(-90deg) translateX(50%)"}[a.placedSide],visibility:a.shouldHideArrow?"hidden":void 0},children:E.jsx(uY,{...i,ref:n,style:{...i.style,display:"block"}})})});rF.displayName=nF;function wY(e){return e!==null}var SY=e=>({name:"transformOrigin",options:e,fn(t){var x,C,O;const{placement:n,rects:r,middlewareData:i}=t,s=((x=i.arrow)==null?void 0:x.centerOffset)!==0,u=s?0:e.arrowWidth,c=s?0:e.arrowHeight,[d,p]=iF(n),m={start:"0%",center:"50%",end:"100%"}[p],v=(((C=i.arrow)==null?void 0:C.x)??0)+u/2,b=(((O=i.arrow)==null?void 0:O.y)??0)+c/2;let S="",w="";return d==="bottom"?(S=s?m:`${v}px`,w=`${-c}px`):d==="top"?(S=s?m:`${v}px`,w=`${r.floating.height+c}px`):d==="right"?(S=`${-c}px`,w=s?m:`${b}px`):d==="left"&&(S=`${r.floating.width+c}px`,w=s?m:`${b}px`),{data:{x:S,y:w}}}});function iF(e){const[t,n="center"]=e.split("-");return[t,n]}var rC=QB,lv=eF,iC=tF,aC=rF;function _Y(e){const t=EY(e),n=g.forwardRef((r,i)=>{const{children:a,...s}=r,u=g.Children.toArray(a),c=u.find(OY);if(c){const d=c.props.children,p=u.map(m=>m===c?g.Children.count(d)>1?g.Children.only(null):g.isValidElement(d)?d.props.children:null:m);return E.jsx(t,{...s,ref:i,children:g.isValidElement(d)?g.cloneElement(d,void 0,p):null})}return E.jsx(t,{...s,ref:i,children:a})});return n.displayName=`${e}.Slot`,n}function EY(e){const t=g.forwardRef((n,r)=>{const{children:i,...a}=n,s=g.isValidElement(i)?RY(i):void 0,u=Re(s,r);if(g.isValidElement(i)){const c=AY(a,i.props);return i.type!==g.Fragment&&(c.ref=u),g.cloneElement(i,c)}return g.Children.count(i)>1?g.Children.only(null):null});return t.displayName=`${e}.SlotClone`,t}var CY=Symbol("radix.slottable");function OY(e){return g.isValidElement(e)&&typeof e.type=="function"&&"__radixId"in e.type&&e.type.__radixId===CY}function AY(e,t){const n={...t};for(const r in t){const i=e[r],a=t[r];/^on[A-Z]/.test(r)?i&&a?n[r]=(...u)=>{const c=a(...u);return i(...u),c}:i&&(n[r]=i):r==="style"?n[r]={...i,...a}:r==="className"&&(n[r]=[i,a].filter(Boolean).join(" "))}return{...e,...n}}function RY(e){var r,i;let t=(r=Object.getOwnPropertyDescriptor(e.props,"ref"))==null?void 0:r.get,n=t&&"isReactWarning"in t&&t.isReactWarning;return n?e.ref:(t=(i=Object.getOwnPropertyDescriptor(e,"ref"))==null?void 0:i.get,n=t&&"isReactWarning"in t&&t.isReactWarning,n?e.props.ref:e.props.ref||e.ref)}var TY=["a","button","div","form","h2","h3","img","input","label","li","nav","ol","p","select","span","svg","ul"],MY=TY.reduce((e,t)=>{const n=_Y(`Primitive.${t}`),r=g.forwardRef((i,a)=>{const{asChild:s,...u}=i,c=s?n:t;return typeof window<"u"&&(window[Symbol.for("radix-ui")]=!0),E.jsx(c,{...u,ref:a})});return r.displayName=`Primitive.${t}`,{...e,[t]:r}},{}),PY="Portal",fd=g.forwardRef((e,t)=>{var u;const{container:n,...r}=e,[i,a]=g.useState(!1);cn(()=>a(!0),[]);const s=n||i&&((u=globalThis==null?void 0:globalThis.document)==null?void 0:u.body);return s?QI.createPortal(E.jsx(MY.div,{...r,ref:t}),s):null});fd.displayName=PY;var aF=Object.freeze({position:"absolute",border:0,width:1,height:1,padding:0,margin:-1,overflow:"hidden",clip:"rect(0, 0, 0, 0)",whiteSpace:"nowrap",wordWrap:"normal"}),jY="VisuallyHidden",NY=g.forwardRef((e,t)=>E.jsx(Dn.span,{...e,ref:t,style:{...aF,...e.style}}));NY.displayName=jY;var DY=function(e){if(typeof document>"u")return null;var t=Array.isArray(e)?e[0]:e;return t.ownerDocument.body},fl=new WeakMap,Yp=new WeakMap,Zp={},Hb=0,oF=function(e){return e&&(e.host||oF(e.parentNode))},kY=function(e,t){return t.map(function(n){if(e.contains(n))return n;var r=oF(n);return r&&e.contains(r)?r:(console.error("aria-hidden",n,"in not contained inside",e,". Doing nothing"),null)}).filter(function(n){return!!n})},$Y=function(e,t,n,r){var i=kY(t,Array.isArray(e)?e:[e]);Zp[n]||(Zp[n]=new WeakMap);var a=Zp[n],s=[],u=new Set,c=new Set(i),d=function(m){!m||u.has(m)||(u.add(m),d(m.parentNode))};i.forEach(d);var p=function(m){!m||c.has(m)||Array.prototype.forEach.call(m.children,function(v){if(u.has(v))p(v);else try{var b=v.getAttribute(r),S=b!==null&&b!=="false",w=(fl.get(v)||0)+1,x=(a.get(v)||0)+1;fl.set(v,w),a.set(v,x),s.push(v),w===1&&S&&Yp.set(v,!0),x===1&&v.setAttribute(n,"true"),S||v.setAttribute(r,"true")}catch(C){console.error("aria-hidden: cannot operate on ",v,C)}})};return p(t),u.clear(),Hb++,function(){s.forEach(function(m){var v=fl.get(m)-1,b=a.get(m)-1;fl.set(m,v),a.set(m,b),v||(Yp.has(m)||m.removeAttribute(r),Yp.delete(m)),b||m.removeAttribute(n)}),Hb--,Hb||(fl=new WeakMap,fl=new WeakMap,Yp=new WeakMap,Zp={})}},uv=function(e,t,n){n===void 0&&(n="data-aria-hidden");var r=Array.from(Array.isArray(e)?e:[e]),i=DY(e);return i?(r.push.apply(r,Array.from(i.querySelectorAll("[aria-live]"))),$Y(r,i,n,"aria-hidden")):function(){return null}},di=function(){return di=Object.assign||function(t){for(var n,r=1,i=arguments.length;r<i;r++){n=arguments[r];for(var a in n)Object.prototype.hasOwnProperty.call(n,a)&&(t[a]=n[a])}return t},di.apply(this,arguments)};function sF(e,t){var n={};for(var r in e)Object.prototype.hasOwnProperty.call(e,r)&&t.indexOf(r)<0&&(n[r]=e[r]);if(e!=null&&typeof Object.getOwnPropertySymbols=="function")for(var i=0,r=Object.getOwnPropertySymbols(e);i<r.length;i++)t.indexOf(r[i])<0&&Object.prototype.propertyIsEnumerable.call(e,r[i])&&(n[r[i]]=e[r[i]]);return n}function IY(e,t,n){if(n||arguments.length===2)for(var r=0,i=t.length,a;r<i;r++)(a||!(r in t))&&(a||(a=Array.prototype.slice.call(t,0,r)),a[r]=t[r]);return e.concat(a||Array.prototype.slice.call(t))}var Sh="right-scroll-bar-position",_h="width-before-scroll-bar",LY="with-scroll-bars-hidden",BY="--removed-body-scroll-bar-size";function Gb(e,t){return typeof e=="function"?e(t):e&&(e.current=t),e}function FY(e,t){var n=g.useState(function(){return{value:e,callback:t,facade:{get current(){return n.value},set current(r){var i=n.value;i!==r&&(n.value=r,n.callback(r,i))}}}})[0];return n.callback=t,n.facade}var qY=typeof window<"u"?g.useLayoutEffect:g.useEffect,NP=new WeakMap;function zY(e,t){var n=FY(null,function(r){return e.forEach(function(i){return Gb(i,r)})});return qY(function(){var r=NP.get(n);if(r){var i=new Set(r),a=new Set(e),s=n.current;i.forEach(function(u){a.has(u)||Gb(u,null)}),a.forEach(function(u){i.has(u)||Gb(u,s)})}NP.set(n,e)},[e]),n}function UY(e){return e}function VY(e,t){t===void 0&&(t=UY);var n=[],r=!1,i={read:function(){if(r)throw new Error("Sidecar: could not `read` from an `assigned` medium. `read` could be used only with `useMedium`.");return n.length?n[n.length-1]:e},useMedium:function(a){var s=t(a,r);return n.push(s),function(){n=n.filter(function(u){return u!==s})}},assignSyncMedium:function(a){for(r=!0;n.length;){var s=n;n=[],s.forEach(a)}n={push:function(u){return a(u)},filter:function(){return n}}},assignMedium:function(a){r=!0;var s=[];if(n.length){var u=n;n=[],u.forEach(a),s=n}var c=function(){var p=s;s=[],p.forEach(a)},d=function(){return Promise.resolve().then(c)};d(),n={push:function(p){s.push(p),d()},filter:function(p){return s=s.filter(p),n}}}};return i}function HY(e){e===void 0&&(e={});var t=VY(null);return t.options=di({async:!0,ssr:!1},e),t}var lF=function(e){var t=e.sideCar,n=sF(e,["sideCar"]);if(!t)throw new Error("Sidecar: please provide `sideCar` property to import the right car");var r=t.read();if(!r)throw new Error("Sidecar medium not found");return g.createElement(r,di({},n))};lF.isSideCarExport=!0;function GY(e,t){return e.useMedium(t),lF}var uF=HY(),Wb=function(){},cv=g.forwardRef(function(e,t){var n=g.useRef(null),r=g.useState({onScrollCapture:Wb,onWheelCapture:Wb,onTouchMoveCapture:Wb}),i=r[0],a=r[1],s=e.forwardProps,u=e.children,c=e.className,d=e.removeScrollBar,p=e.enabled,m=e.shards,v=e.sideCar,b=e.noIsolation,S=e.inert,w=e.allowPinchZoom,x=e.as,C=x===void 0?"div":x,O=e.gapMode,A=sF(e,["forwardProps","children","className","removeScrollBar","enabled","shards","sideCar","noIsolation","inert","allowPinchZoom","as","gapMode"]),T=v,R=zY([n,t]),M=di(di({},A),i);return g.createElement(g.Fragment,null,p&&g.createElement(T,{sideCar:uF,removeScrollBar:d,shards:m,noIsolation:b,inert:S,setCallbacks:a,allowPinchZoom:!!w,lockRef:n,gapMode:O}),s?g.cloneElement(g.Children.only(u),di(di({},M),{ref:R})):g.createElement(C,di({},M,{className:c,ref:R}),u))});cv.defaultProps={enabled:!0,removeScrollBar:!0,inert:!1};cv.classNames={fullWidth:_h,zeroRight:Sh};var WY=function(){if(typeof __webpack_nonce__<"u")return __webpack_nonce__};function KY(){if(!document)return null;var e=document.createElement("style");e.type="text/css";var t=WY();return t&&e.setAttribute("nonce",t),e}function YY(e,t){e.styleSheet?e.styleSheet.cssText=t:e.appendChild(document.createTextNode(t))}function ZY(e){var t=document.head||document.getElementsByTagName("head")[0];t.appendChild(e)}var XY=function(){var e=0,t=null;return{add:function(n){e==0&&(t=KY())&&(YY(t,n),ZY(t)),e++},remove:function(){e--,!e&&t&&(t.parentNode&&t.parentNode.removeChild(t),t=null)}}},QY=function(){var e=XY();return function(t,n){g.useEffect(function(){return e.add(t),function(){e.remove()}},[t&&n])}},cF=function(){var e=QY(),t=function(n){var r=n.styles,i=n.dynamic;return e(r,i),null};return t},JY={left:0,top:0,right:0,gap:0},Kb=function(e){return parseInt(e||"",10)||0},eZ=function(e){var t=window.getComputedStyle(document.body),n=t[e==="padding"?"paddingLeft":"marginLeft"],r=t[e==="padding"?"paddingTop":"marginTop"],i=t[e==="padding"?"paddingRight":"marginRight"];return[Kb(n),Kb(r),Kb(i)]},tZ=function(e){if(e===void 0&&(e="margin"),typeof window>"u")return JY;var t=eZ(e),n=document.documentElement.clientWidth,r=window.innerWidth;return{left:t[0],top:t[1],right:t[2],gap:Math.max(0,r-n+t[2]-t[0])}},nZ=cF(),Ol="data-scroll-locked",rZ=function(e,t,n,r){var i=e.left,a=e.top,s=e.right,u=e.gap;return n===void 0&&(n="margin"),`
  .`.concat(LY,` {
   overflow: hidden `).concat(r,`;
   padding-right: `).concat(u,"px ").concat(r,`;
  }
  body[`).concat(Ol,`] {
    overflow: hidden `).concat(r,`;
    overscroll-behavior: contain;
    `).concat([t&&"position: relative ".concat(r,";"),n==="margin"&&`
    padding-left: `.concat(i,`px;
    padding-top: `).concat(a,`px;
    padding-right: `).concat(s,`px;
    margin-left:0;
    margin-top:0;
    margin-right: `).concat(u,"px ").concat(r,`;
    `),n==="padding"&&"padding-right: ".concat(u,"px ").concat(r,";")].filter(Boolean).join(""),`
  }
  
  .`).concat(Sh,` {
    right: `).concat(u,"px ").concat(r,`;
  }
  
  .`).concat(_h,` {
    margin-right: `).concat(u,"px ").concat(r,`;
  }
  
  .`).concat(Sh," .").concat(Sh,` {
    right: 0 `).concat(r,`;
  }
  
  .`).concat(_h," .").concat(_h,` {
    margin-right: 0 `).concat(r,`;
  }
  
  body[`).concat(Ol,`] {
    `).concat(BY,": ").concat(u,`px;
  }
`)},DP=function(){var e=parseInt(document.body.getAttribute(Ol)||"0",10);return isFinite(e)?e:0},iZ=function(){g.useEffect(function(){return document.body.setAttribute(Ol,(DP()+1).toString()),function(){var e=DP()-1;e<=0?document.body.removeAttribute(Ol):document.body.setAttribute(Ol,e.toString())}},[])},aZ=function(e){var t=e.noRelative,n=e.noImportant,r=e.gapMode,i=r===void 0?"margin":r;iZ();var a=g.useMemo(function(){return tZ(i)},[i]);return g.createElement(nZ,{styles:rZ(a,!t,i,n?"":"!important")})},D_=!1;if(typeof window<"u")try{var Xp=Object.defineProperty({},"passive",{get:function(){return D_=!0,!0}});window.addEventListener("test",Xp,Xp),window.removeEventListener("test",Xp,Xp)}catch{D_=!1}var dl=D_?{passive:!1}:!1,oZ=function(e){return e.tagName==="TEXTAREA"},fF=function(e,t){if(!(e instanceof Element))return!1;var n=window.getComputedStyle(e);return n[t]!=="hidden"&&!(n.overflowY===n.overflowX&&!oZ(e)&&n[t]==="visible")},sZ=function(e){return fF(e,"overflowY")},lZ=function(e){return fF(e,"overflowX")},kP=function(e,t){var n=t.ownerDocument,r=t;do{typeof ShadowRoot<"u"&&r instanceof ShadowRoot&&(r=r.host);var i=dF(e,r);if(i){var a=pF(e,r),s=a[1],u=a[2];if(s>u)return!0}r=r.parentNode}while(r&&r!==n.body);return!1},uZ=function(e){var t=e.scrollTop,n=e.scrollHeight,r=e.clientHeight;return[t,n,r]},cZ=function(e){var t=e.scrollLeft,n=e.scrollWidth,r=e.clientWidth;return[t,n,r]},dF=function(e,t){return e==="v"?sZ(t):lZ(t)},pF=function(e,t){return e==="v"?uZ(t):cZ(t)},fZ=function(e,t){return e==="h"&&t==="rtl"?-1:1},dZ=function(e,t,n,r,i){var a=fZ(e,window.getComputedStyle(t).direction),s=a*r,u=n.target,c=t.contains(u),d=!1,p=s>0,m=0,v=0;do{var b=pF(e,u),S=b[0],w=b[1],x=b[2],C=w-x-a*S;(S||C)&&dF(e,u)&&(m+=C,v+=S),u instanceof ShadowRoot?u=u.host:u=u.parentNode}while(!c&&u!==document.body||c&&(t.contains(u)||t===u));return(p&&Math.abs(m)<1||!p&&Math.abs(v)<1)&&(d=!0),d},Qp=function(e){return"changedTouches"in e?[e.changedTouches[0].clientX,e.changedTouches[0].clientY]:[0,0]},$P=function(e){return[e.deltaX,e.deltaY]},IP=function(e){return e&&"current"in e?e.current:e},pZ=function(e,t){return e[0]===t[0]&&e[1]===t[1]},hZ=function(e){return`
  .block-interactivity-`.concat(e,` {pointer-events: none;}
  .allow-interactivity-`).concat(e,` {pointer-events: all;}
`)},mZ=0,pl=[];function vZ(e){var t=g.useRef([]),n=g.useRef([0,0]),r=g.useRef(),i=g.useState(mZ++)[0],a=g.useState(cF)[0],s=g.useRef(e);g.useEffect(function(){s.current=e},[e]),g.useEffect(function(){if(e.inert){document.body.classList.add("block-interactivity-".concat(i));var w=IY([e.lockRef.current],(e.shards||[]).map(IP),!0).filter(Boolean);return w.forEach(function(x){return x.classList.add("allow-interactivity-".concat(i))}),function(){document.body.classList.remove("block-interactivity-".concat(i)),w.forEach(function(x){return x.classList.remove("allow-interactivity-".concat(i))})}}},[e.inert,e.lockRef.current,e.shards]);var u=g.useCallback(function(w,x){if("touches"in w&&w.touches.length===2||w.type==="wheel"&&w.ctrlKey)return!s.current.allowPinchZoom;var C=Qp(w),O=n.current,A="deltaX"in w?w.deltaX:O[0]-C[0],T="deltaY"in w?w.deltaY:O[1]-C[1],R,M=w.target,P=Math.abs(A)>Math.abs(T)?"h":"v";if("touches"in w&&P==="h"&&M.type==="range")return!1;var N=kP(P,M);if(!N)return!0;if(N?R=P:(R=P==="v"?"h":"v",N=kP(P,M)),!N)return!1;if(!r.current&&"changedTouches"in w&&(A||T)&&(r.current=R),!R)return!0;var k=r.current||R;return dZ(k,x,w,k==="h"?A:T)},[]),c=g.useCallback(function(w){var x=w;if(!(!pl.length||pl[pl.length-1]!==a)){var C="deltaY"in x?$P(x):Qp(x),O=t.current.filter(function(R){return R.name===x.type&&(R.target===x.target||x.target===R.shadowParent)&&pZ(R.delta,C)})[0];if(O&&O.should){x.cancelable&&x.preventDefault();return}if(!O){var A=(s.current.shards||[]).map(IP).filter(Boolean).filter(function(R){return R.contains(x.target)}),T=A.length>0?u(x,A[0]):!s.current.noIsolation;T&&x.cancelable&&x.preventDefault()}}},[]),d=g.useCallback(function(w,x,C,O){var A={name:w,delta:x,target:C,should:O,shadowParent:gZ(C)};t.current.push(A),setTimeout(function(){t.current=t.current.filter(function(T){return T!==A})},1)},[]),p=g.useCallback(function(w){n.current=Qp(w),r.current=void 0},[]),m=g.useCallback(function(w){d(w.type,$P(w),w.target,u(w,e.lockRef.current))},[]),v=g.useCallback(function(w){d(w.type,Qp(w),w.target,u(w,e.lockRef.current))},[]);g.useEffect(function(){return pl.push(a),e.setCallbacks({onScrollCapture:m,onWheelCapture:m,onTouchMoveCapture:v}),document.addEventListener("wheel",c,dl),document.addEventListener("touchmove",c,dl),document.addEventListener("touchstart",p,dl),function(){pl=pl.filter(function(w){return w!==a}),document.removeEventListener("wheel",c,dl),document.removeEventListener("touchmove",c,dl),document.removeEventListener("touchstart",p,dl)}},[]);var b=e.removeScrollBar,S=e.inert;return g.createElement(g.Fragment,null,S?g.createElement(a,{styles:hZ(i)}):null,b?g.createElement(aZ,{gapMode:e.gapMode}):null)}function gZ(e){for(var t=null;e!==null;)e instanceof ShadowRoot&&(t=e.host,e=e.host),e=e.parentNode;return t}const yZ=GY(uF,vZ);var dd=g.forwardRef(function(e,t){return g.createElement(cv,di({},e,{ref:t,sideCar:yZ}))});dd.classNames=cv.classNames;var bZ=[" ","Enter","ArrowUp","ArrowDown"],xZ=[" ","Enter"],ls="Select",[fv,dv,wZ]=jW(ls),[du,Y0e]=jn(ls,[wZ,fu]),pv=fu(),[SZ,fo]=du(ls),[_Z,EZ]=du(ls),hF=e=>{const{__scopeSelect:t,children:n,open:r,defaultOpen:i,onOpenChange:a,value:s,defaultValue:u,onValueChange:c,dir:d,name:p,autoComplete:m,disabled:v,required:b,form:S}=e,w=pv(t),[x,C]=g.useState(null),[O,A]=g.useState(null),[T,R]=g.useState(!1),M=iu(d),[P,N]=Xr({prop:r,defaultProp:i??!1,onChange:a,caller:ls}),[k,H]=Xr({prop:s,defaultProp:u,onChange:c,caller:ls}),L=g.useRef(null),F=x?S||!!x.closest("form"):!0,[z,Z]=g.useState(new Set),Y=Array.from(z).map(I=>I.props.value).join(";");return E.jsx(rC,{...w,children:E.jsxs(SZ,{required:b,scope:t,trigger:x,onTriggerChange:C,valueNode:O,onValueNodeChange:A,valueNodeHasChildren:T,onValueNodeHasChildrenChange:R,contentId:ln(),value:k,onValueChange:H,open:P,onOpenChange:N,dir:M,triggerPointerDownPosRef:L,disabled:v,children:[E.jsx(fv.Provider,{scope:t,children:E.jsx(_Z,{scope:e.__scopeSelect,onNativeOptionAdd:g.useCallback(I=>{Z(W=>new Set(W).add(I))},[]),onNativeOptionRemove:g.useCallback(I=>{Z(W=>{const B=new Set(W);return B.delete(I),B})},[]),children:n})}),F?E.jsxs($F,{"aria-hidden":!0,required:b,tabIndex:-1,name:p,autoComplete:m,value:k,onChange:I=>H(I.target.value),disabled:v,form:S,children:[k===void 0?E.jsx("option",{value:""}):null,Array.from(z)]},Y):null]})})};hF.displayName=ls;var mF="SelectTrigger",vF=g.forwardRef((e,t)=>{const{__scopeSelect:n,disabled:r=!1,...i}=e,a=pv(n),s=fo(mF,n),u=s.disabled||r,c=Re(t,s.onTriggerChange),d=dv(n),p=g.useRef("touch"),[m,v,b]=LF(w=>{const x=d().filter(A=>!A.disabled),C=x.find(A=>A.value===s.value),O=BF(x,w,C);O!==void 0&&s.onValueChange(O.value)}),S=w=>{u||(s.onOpenChange(!0),b()),w&&(s.triggerPointerDownPosRef.current={x:Math.round(w.pageX),y:Math.round(w.pageY)})};return E.jsx(lv,{asChild:!0,...a,children:E.jsx(Dn.button,{type:"button",role:"combobox","aria-controls":s.contentId,"aria-expanded":s.open,"aria-required":s.required,"aria-autocomplete":"none",dir:s.dir,"data-state":s.open?"open":"closed",disabled:u,"data-disabled":u?"":void 0,"data-placeholder":IF(s.value)?"":void 0,...i,ref:c,onClick:be(i.onClick,w=>{w.currentTarget.focus(),p.current!=="mouse"&&S(w)}),onPointerDown:be(i.onPointerDown,w=>{p.current=w.pointerType;const x=w.target;x.hasPointerCapture(w.pointerId)&&x.releasePointerCapture(w.pointerId),w.button===0&&w.ctrlKey===!1&&w.pointerType==="mouse"&&(S(w),w.preventDefault())}),onKeyDown:be(i.onKeyDown,w=>{const x=m.current!=="";!(w.ctrlKey||w.altKey||w.metaKey)&&w.key.length===1&&v(w.key),!(x&&w.key===" ")&&bZ.includes(w.key)&&(S(),w.preventDefault())})})})});vF.displayName=mF;var gF="SelectValue",yF=g.forwardRef((e,t)=>{const{__scopeSelect:n,className:r,style:i,children:a,placeholder:s="",...u}=e,c=fo(gF,n),{onValueNodeHasChildrenChange:d}=c,p=a!==void 0,m=Re(t,c.onValueNodeChange);return cn(()=>{d(p)},[d,p]),E.jsx(Dn.span,{...u,ref:m,style:{pointerEvents:"none"},children:IF(c.value)?E.jsx(E.Fragment,{children:s}):a})});yF.displayName=gF;var CZ="SelectIcon",bF=g.forwardRef((e,t)=>{const{__scopeSelect:n,children:r,...i}=e;return E.jsx(Dn.span,{"aria-hidden":!0,...i,ref:t,children:r||"▼"})});bF.displayName=CZ;var OZ="SelectPortal",xF=e=>E.jsx(fd,{asChild:!0,...e});xF.displayName=OZ;var us="SelectContent",wF=g.forwardRef((e,t)=>{const n=fo(us,e.__scopeSelect),[r,i]=g.useState();if(cn(()=>{i(new DocumentFragment)},[]),!n.open){const a=r;return a?ba.createPortal(E.jsx(SF,{scope:e.__scopeSelect,children:E.jsx(fv.Slot,{scope:e.__scopeSelect,children:E.jsx("div",{children:e.children})})}),a):null}return E.jsx(_F,{...e,ref:t})});wF.displayName=us;var Ur=10,[SF,po]=du(us),AZ="SelectContentImpl",RZ=zh("SelectContent.RemoveScroll"),_F=g.forwardRef((e,t)=>{const{__scopeSelect:n,position:r="item-aligned",onCloseAutoFocus:i,onEscapeKeyDown:a,onPointerDownOutside:s,side:u,sideOffset:c,align:d,alignOffset:p,arrowPadding:m,collisionBoundary:v,collisionPadding:b,sticky:S,hideWhenDetached:w,avoidCollisions:x,...C}=e,O=fo(us,n),[A,T]=g.useState(null),[R,M]=g.useState(null),P=Re(t,de=>T(de)),[N,k]=g.useState(null),[H,L]=g.useState(null),F=dv(n),[z,Z]=g.useState(!1),Y=g.useRef(!1);g.useEffect(()=>{if(A)return uv(A)},[A]),iv();const I=g.useCallback(de=>{const[Se,...ce]=F().map(se=>se.ref.current),[we]=ce.slice(-1),_e=document.activeElement;for(const se of de)if(se===_e||(se==null||se.scrollIntoView({block:"nearest"}),se===Se&&R&&(R.scrollTop=0),se===we&&R&&(R.scrollTop=R.scrollHeight),se==null||se.focus(),document.activeElement!==_e))return},[F,R]),W=g.useCallback(()=>I([N,A]),[I,N,A]);g.useEffect(()=>{z&&W()},[z,W]);const{onOpenChange:B,triggerPointerDownPosRef:Q}=O;g.useEffect(()=>{if(A){let de={x:0,y:0};const Se=we=>{var _e,se;de={x:Math.abs(Math.round(we.pageX)-(((_e=Q.current)==null?void 0:_e.x)??0)),y:Math.abs(Math.round(we.pageY)-(((se=Q.current)==null?void 0:se.y)??0))}},ce=we=>{de.x<=10&&de.y<=10?we.preventDefault():A.contains(we.target)||B(!1),document.removeEventListener("pointermove",Se),Q.current=null};return Q.current!==null&&(document.addEventListener("pointermove",Se),document.addEventListener("pointerup",ce,{capture:!0,once:!0})),()=>{document.removeEventListener("pointermove",Se),document.removeEventListener("pointerup",ce,{capture:!0})}}},[A,B,Q]),g.useEffect(()=>{const de=()=>B(!1);return window.addEventListener("blur",de),window.addEventListener("resize",de),()=>{window.removeEventListener("blur",de),window.removeEventListener("resize",de)}},[B]);const[$,J]=LF(de=>{const Se=F().filter(_e=>!_e.disabled),ce=Se.find(_e=>_e.ref.current===document.activeElement),we=BF(Se,de,ce);we&&setTimeout(()=>we.ref.current.focus())}),q=g.useCallback((de,Se,ce)=>{const we=!Y.current&&!ce;(O.value!==void 0&&O.value===Se||we)&&(k(de),we&&(Y.current=!0))},[O.value]),X=g.useCallback(()=>A==null?void 0:A.focus(),[A]),oe=g.useCallback((de,Se,ce)=>{const we=!Y.current&&!ce;(O.value!==void 0&&O.value===Se||we)&&L(de)},[O.value]),le=r==="popper"?k_:EF,ve=le===k_?{side:u,sideOffset:c,align:d,alignOffset:p,arrowPadding:m,collisionBoundary:v,collisionPadding:b,sticky:S,hideWhenDetached:w,avoidCollisions:x}:{};return E.jsx(SF,{scope:n,content:A,viewport:R,onViewportChange:M,itemRefCallback:q,selectedItem:N,onItemLeave:X,itemTextRefCallback:oe,focusSelectedItem:W,selectedItemText:H,position:r,isPositioned:z,searchRef:$,children:E.jsx(dd,{as:RZ,allowPinchZoom:!0,children:E.jsx(ud,{asChild:!0,trapped:O.open,onMountAutoFocus:de=>{de.preventDefault()},onUnmountAutoFocus:be(i,de=>{var Se;(Se=O.trigger)==null||Se.focus({preventScroll:!0}),de.preventDefault()}),children:E.jsx(kB,{asChild:!0,disableOutsidePointerEvents:!0,onEscapeKeyDown:a,onPointerDownOutside:s,onFocusOutside:de=>de.preventDefault(),onDismiss:()=>O.onOpenChange(!1),children:E.jsx(le,{role:"listbox",id:O.contentId,"data-state":O.open?"open":"closed",dir:O.dir,onContextMenu:de=>de.preventDefault(),...C,...ve,onPlaced:()=>Z(!0),ref:P,style:{display:"flex",flexDirection:"column",outline:"none",...C.style},onKeyDown:be(C.onKeyDown,de=>{const Se=de.ctrlKey||de.altKey||de.metaKey;if(de.key==="Tab"&&de.preventDefault(),!Se&&de.key.length===1&&J(de.key),["ArrowUp","ArrowDown","Home","End"].includes(de.key)){let we=F().filter(_e=>!_e.disabled).map(_e=>_e.ref.current);if(["ArrowUp","End"].includes(de.key)&&(we=we.slice().reverse()),["ArrowUp","ArrowDown"].includes(de.key)){const _e=de.target,se=we.indexOf(_e);we=we.slice(se+1)}setTimeout(()=>I(we)),de.preventDefault()}})})})})})})});_F.displayName=AZ;var TZ="SelectItemAlignedPosition",EF=g.forwardRef((e,t)=>{const{__scopeSelect:n,onPlaced:r,...i}=e,a=fo(us,n),s=po(us,n),[u,c]=g.useState(null),[d,p]=g.useState(null),m=Re(t,P=>p(P)),v=dv(n),b=g.useRef(!1),S=g.useRef(!0),{viewport:w,selectedItem:x,selectedItemText:C,focusSelectedItem:O}=s,A=g.useCallback(()=>{if(a.trigger&&a.valueNode&&u&&d&&w&&x&&C){const P=a.trigger.getBoundingClientRect(),N=d.getBoundingClientRect(),k=a.valueNode.getBoundingClientRect(),H=C.getBoundingClientRect();if(a.dir!=="rtl"){const _e=H.left-N.left,se=k.left-_e,Ye=P.left-se,Be=P.width+Ye,dt=Math.max(Be,N.width),ft=window.innerWidth-Ur,it=qh(se,[Ur,Math.max(Ur,ft-dt)]);u.style.minWidth=Be+"px",u.style.left=it+"px"}else{const _e=N.right-H.right,se=window.innerWidth-k.right-_e,Ye=window.innerWidth-P.right-se,Be=P.width+Ye,dt=Math.max(Be,N.width),ft=window.innerWidth-Ur,it=qh(se,[Ur,Math.max(Ur,ft-dt)]);u.style.minWidth=Be+"px",u.style.right=it+"px"}const L=v(),F=window.innerHeight-Ur*2,z=w.scrollHeight,Z=window.getComputedStyle(d),Y=parseInt(Z.borderTopWidth,10),I=parseInt(Z.paddingTop,10),W=parseInt(Z.borderBottomWidth,10),B=parseInt(Z.paddingBottom,10),Q=Y+I+z+B+W,$=Math.min(x.offsetHeight*5,Q),J=window.getComputedStyle(w),q=parseInt(J.paddingTop,10),X=parseInt(J.paddingBottom,10),oe=P.top+P.height/2-Ur,le=F-oe,ve=x.offsetHeight/2,de=x.offsetTop+ve,Se=Y+I+de,ce=Q-Se;if(Se<=oe){const _e=L.length>0&&x===L[L.length-1].ref.current;u.style.bottom="0px";const se=d.clientHeight-w.offsetTop-w.offsetHeight,Ye=Math.max(le,ve+(_e?X:0)+se+W),Be=Se+Ye;u.style.height=Be+"px"}else{const _e=L.length>0&&x===L[0].ref.current;u.style.top="0px";const Ye=Math.max(oe,Y+w.offsetTop+(_e?q:0)+ve)+ce;u.style.height=Ye+"px",w.scrollTop=Se-oe+w.offsetTop}u.style.margin=`${Ur}px 0`,u.style.minHeight=$+"px",u.style.maxHeight=F+"px",r==null||r(),requestAnimationFrame(()=>b.current=!0)}},[v,a.trigger,a.valueNode,u,d,w,x,C,a.dir,r]);cn(()=>A(),[A]);const[T,R]=g.useState();cn(()=>{d&&R(window.getComputedStyle(d).zIndex)},[d]);const M=g.useCallback(P=>{P&&S.current===!0&&(A(),O==null||O(),S.current=!1)},[A,O]);return E.jsx(PZ,{scope:n,contentWrapper:u,shouldExpandOnScrollRef:b,onScrollButtonChange:M,children:E.jsx("div",{ref:c,style:{display:"flex",flexDirection:"column",position:"fixed",zIndex:T},children:E.jsx(Dn.div,{...i,ref:m,style:{boxSizing:"border-box",maxHeight:"100%",...i.style}})})})});EF.displayName=TZ;var MZ="SelectPopperPosition",k_=g.forwardRef((e,t)=>{const{__scopeSelect:n,align:r="start",collisionPadding:i=Ur,...a}=e,s=pv(n);return E.jsx(iC,{...s,...a,ref:t,align:r,collisionPadding:i,style:{boxSizing:"border-box",...a.style,"--radix-select-content-transform-origin":"var(--radix-popper-transform-origin)","--radix-select-content-available-width":"var(--radix-popper-available-width)","--radix-select-content-available-height":"var(--radix-popper-available-height)","--radix-select-trigger-width":"var(--radix-popper-anchor-width)","--radix-select-trigger-height":"var(--radix-popper-anchor-height)"}})});k_.displayName=MZ;var[PZ,oC]=du(us,{}),$_="SelectViewport",CF=g.forwardRef((e,t)=>{const{__scopeSelect:n,nonce:r,...i}=e,a=po($_,n),s=oC($_,n),u=Re(t,a.onViewportChange),c=g.useRef(0);return E.jsxs(E.Fragment,{children:[E.jsx("style",{dangerouslySetInnerHTML:{__html:"[data-radix-select-viewport]{scrollbar-width:none;-ms-overflow-style:none;-webkit-overflow-scrolling:touch;}[data-radix-select-viewport]::-webkit-scrollbar{display:none}"},nonce:r}),E.jsx(fv.Slot,{scope:n,children:E.jsx(Dn.div,{"data-radix-select-viewport":"",role:"presentation",...i,ref:u,style:{position:"relative",flex:1,overflow:"hidden auto",...i.style},onScroll:be(i.onScroll,d=>{const p=d.currentTarget,{contentWrapper:m,shouldExpandOnScrollRef:v}=s;if(v!=null&&v.current&&m){const b=Math.abs(c.current-p.scrollTop);if(b>0){const S=window.innerHeight-Ur*2,w=parseFloat(m.style.minHeight),x=parseFloat(m.style.height),C=Math.max(w,x);if(C<S){const O=C+b,A=Math.min(S,O),T=O-A;m.style.height=A+"px",m.style.bottom==="0px"&&(p.scrollTop=T>0?T:0,m.style.justifyContent="flex-end")}}}c.current=p.scrollTop})})})]})});CF.displayName=$_;var OF="SelectGroup",[jZ,NZ]=du(OF),DZ=g.forwardRef((e,t)=>{const{__scopeSelect:n,...r}=e,i=ln();return E.jsx(jZ,{scope:n,id:i,children:E.jsx(Dn.div,{role:"group","aria-labelledby":i,...r,ref:t})})});DZ.displayName=OF;var AF="SelectLabel",kZ=g.forwardRef((e,t)=>{const{__scopeSelect:n,...r}=e,i=NZ(AF,n);return E.jsx(Dn.div,{id:i.id,...r,ref:t})});kZ.displayName=AF;var Wh="SelectItem",[$Z,RF]=du(Wh),TF=g.forwardRef((e,t)=>{const{__scopeSelect:n,value:r,disabled:i=!1,textValue:a,...s}=e,u=fo(Wh,n),c=po(Wh,n),d=u.value===r,[p,m]=g.useState(a??""),[v,b]=g.useState(!1),S=Re(t,O=>{var A;return(A=c.itemRefCallback)==null?void 0:A.call(c,O,r,i)}),w=ln(),x=g.useRef("touch"),C=()=>{i||(u.onValueChange(r),u.onOpenChange(!1))};if(r==="")throw new Error("A <Select.Item /> must have a value prop that is not an empty string. This is because the Select value can be set to an empty string to clear the selection and show the placeholder.");return E.jsx($Z,{scope:n,value:r,disabled:i,textId:w,isSelected:d,onItemTextChange:g.useCallback(O=>{m(A=>A||((O==null?void 0:O.textContent)??"").trim())},[]),children:E.jsx(fv.ItemSlot,{scope:n,value:r,disabled:i,textValue:p,children:E.jsx(Dn.div,{role:"option","aria-labelledby":w,"data-highlighted":v?"":void 0,"aria-selected":d&&v,"data-state":d?"checked":"unchecked","aria-disabled":i||void 0,"data-disabled":i?"":void 0,tabIndex:i?void 0:-1,...s,ref:S,onFocus:be(s.onFocus,()=>b(!0)),onBlur:be(s.onBlur,()=>b(!1)),onClick:be(s.onClick,()=>{x.current!=="mouse"&&C()}),onPointerUp:be(s.onPointerUp,()=>{x.current==="mouse"&&C()}),onPointerDown:be(s.onPointerDown,O=>{x.current=O.pointerType}),onPointerMove:be(s.onPointerMove,O=>{var A;x.current=O.pointerType,i?(A=c.onItemLeave)==null||A.call(c):x.current==="mouse"&&O.currentTarget.focus({preventScroll:!0})}),onPointerLeave:be(s.onPointerLeave,O=>{var A;O.currentTarget===document.activeElement&&((A=c.onItemLeave)==null||A.call(c))}),onKeyDown:be(s.onKeyDown,O=>{var T;((T=c.searchRef)==null?void 0:T.current)!==""&&O.key===" "||(xZ.includes(O.key)&&C(),O.key===" "&&O.preventDefault())})})})})});TF.displayName=Wh;var Bc="SelectItemText",MF=g.forwardRef((e,t)=>{const{__scopeSelect:n,className:r,style:i,...a}=e,s=fo(Bc,n),u=po(Bc,n),c=RF(Bc,n),d=EZ(Bc,n),[p,m]=g.useState(null),v=Re(t,C=>m(C),c.onItemTextChange,C=>{var O;return(O=u.itemTextRefCallback)==null?void 0:O.call(u,C,c.value,c.disabled)}),b=p==null?void 0:p.textContent,S=g.useMemo(()=>E.jsx("option",{value:c.value,disabled:c.disabled,children:b},c.value),[c.disabled,c.value,b]),{onNativeOptionAdd:w,onNativeOptionRemove:x}=d;return cn(()=>(w(S),()=>x(S)),[w,x,S]),E.jsxs(E.Fragment,{children:[E.jsx(Dn.span,{id:c.textId,...a,ref:v}),c.isSelected&&s.valueNode&&!s.valueNodeHasChildren?ba.createPortal(a.children,s.valueNode):null]})});MF.displayName=Bc;var PF="SelectItemIndicator",jF=g.forwardRef((e,t)=>{const{__scopeSelect:n,...r}=e;return RF(PF,n).isSelected?E.jsx(Dn.span,{"aria-hidden":!0,...r,ref:t}):null});jF.displayName=PF;var I_="SelectScrollUpButton",NF=g.forwardRef((e,t)=>{const n=po(I_,e.__scopeSelect),r=oC(I_,e.__scopeSelect),[i,a]=g.useState(!1),s=Re(t,r.onScrollButtonChange);return cn(()=>{if(n.viewport&&n.isPositioned){let u=function(){const d=c.scrollTop>0;a(d)};const c=n.viewport;return u(),c.addEventListener("scroll",u),()=>c.removeEventListener("scroll",u)}},[n.viewport,n.isPositioned]),i?E.jsx(kF,{...e,ref:s,onAutoScroll:()=>{const{viewport:u,selectedItem:c}=n;u&&c&&(u.scrollTop=u.scrollTop-c.offsetHeight)}}):null});NF.displayName=I_;var L_="SelectScrollDownButton",DF=g.forwardRef((e,t)=>{const n=po(L_,e.__scopeSelect),r=oC(L_,e.__scopeSelect),[i,a]=g.useState(!1),s=Re(t,r.onScrollButtonChange);return cn(()=>{if(n.viewport&&n.isPositioned){let u=function(){const d=c.scrollHeight-c.clientHeight,p=Math.ceil(c.scrollTop)<d;a(p)};const c=n.viewport;return u(),c.addEventListener("scroll",u),()=>c.removeEventListener("scroll",u)}},[n.viewport,n.isPositioned]),i?E.jsx(kF,{...e,ref:s,onAutoScroll:()=>{const{viewport:u,selectedItem:c}=n;u&&c&&(u.scrollTop=u.scrollTop+c.offsetHeight)}}):null});DF.displayName=L_;var kF=g.forwardRef((e,t)=>{const{__scopeSelect:n,onAutoScroll:r,...i}=e,a=po("SelectScrollButton",n),s=g.useRef(null),u=dv(n),c=g.useCallback(()=>{s.current!==null&&(window.clearInterval(s.current),s.current=null)},[]);return g.useEffect(()=>()=>c(),[c]),cn(()=>{var p;const d=u().find(m=>m.ref.current===document.activeElement);(p=d==null?void 0:d.ref.current)==null||p.scrollIntoView({block:"nearest"})},[u]),E.jsx(Dn.div,{"aria-hidden":!0,...i,ref:t,style:{flexShrink:0,...i.style},onPointerDown:be(i.onPointerDown,()=>{s.current===null&&(s.current=window.setInterval(r,50))}),onPointerMove:be(i.onPointerMove,()=>{var d;(d=a.onItemLeave)==null||d.call(a),s.current===null&&(s.current=window.setInterval(r,50))}),onPointerLeave:be(i.onPointerLeave,()=>{c()})})}),IZ="SelectSeparator",LZ=g.forwardRef((e,t)=>{const{__scopeSelect:n,...r}=e;return E.jsx(Dn.div,{"aria-hidden":!0,...r,ref:t})});LZ.displayName=IZ;var B_="SelectArrow",BZ=g.forwardRef((e,t)=>{const{__scopeSelect:n,...r}=e,i=pv(n),a=fo(B_,n),s=po(B_,n);return a.open&&s.position==="popper"?E.jsx(aC,{...i,...r,ref:t}):null});BZ.displayName=B_;var FZ="SelectBubbleInput",$F=g.forwardRef(({__scopeSelect:e,value:t,...n},r)=>{const i=g.useRef(null),a=Re(r,i),s=Ym(t);return g.useEffect(()=>{const u=i.current;if(!u)return;const c=window.HTMLSelectElement.prototype,p=Object.getOwnPropertyDescriptor(c,"value").set;if(s!==t&&p){const m=new Event("change",{bubbles:!0});p.call(u,t),u.dispatchEvent(m)}},[s,t]),E.jsx(Dn.select,{...n,style:{...aF,...n.style},ref:a,defaultValue:t})});$F.displayName=FZ;function IF(e){return e===""||e===void 0}function LF(e){const t=$t(e),n=g.useRef(""),r=g.useRef(0),i=g.useCallback(s=>{const u=n.current+s;t(u),function c(d){n.current=d,window.clearTimeout(r.current),d!==""&&(r.current=window.setTimeout(()=>c(""),1e3))}(u)},[t]),a=g.useCallback(()=>{n.current="",window.clearTimeout(r.current)},[]);return g.useEffect(()=>()=>window.clearTimeout(r.current),[]),[n,i,a]}function BF(e,t,n){const i=t.length>1&&Array.from(t).every(d=>d===t[0])?t[0]:t,a=n?e.indexOf(n):-1;let s=qZ(e,Math.max(a,0));i.length===1&&(s=s.filter(d=>d!==n));const c=s.find(d=>d.textValue.toLowerCase().startsWith(i.toLowerCase()));return c!==n?c:void 0}function qZ(e,t){return e.map((n,r)=>e[(t+r)%e.length])}var zZ=hF,UZ=vF,VZ=yF,HZ=bF,GZ=xF,WZ=wF,KZ=CF,YZ=TF,ZZ=MF,XZ=jF,QZ=NF,JZ=DF;function Kh({...e}){return E.jsx(zZ,{"data-slot":"select",...e})}function Yh({...e}){return E.jsx(VZ,{"data-slot":"select-value",...e})}function Zh({className:e,size:t="default",children:n,...r}){return E.jsxs(UZ,{"data-slot":"select-trigger","data-size":t,className:ke("border-input data-[placeholder]:text-muted-foreground [&_svg:not([class*='text-'])]:text-muted-foreground focus-visible:border-ring focus-visible:ring-ring/50 aria-invalid:ring-destructive/20 dark:aria-invalid:ring-destructive/40 aria-invalid:border-destructive dark:bg-input/30 dark:hover:bg-input/50 flex w-fit items-center justify-between gap-2 rounded-md border bg-transparent px-3 py-2 text-sm whitespace-nowrap shadow-xs transition-[color,box-shadow] outline-none focus-visible:ring-[3px] disabled:cursor-not-allowed disabled:opacity-50 data-[size=default]:h-9 data-[size=sm]:h-8 *:data-[slot=select-value]:line-clamp-1 *:data-[slot=select-value]:flex *:data-[slot=select-value]:items-center *:data-[slot=select-value]:gap-2 [&_svg]:pointer-events-none [&_svg]:shrink-0 [&_svg:not([class*='size-'])]:size-4",e),...r,children:[n,E.jsx(HZ,{asChild:!0,children:E.jsx(Zm,{className:"size-4 opacity-50"})})]})}function Xh({className:e,children:t,position:n="popper",...r}){return E.jsx(GZ,{children:E.jsxs(WZ,{"data-slot":"select-content",className:ke("bg-popover text-popover-foreground data-[state=open]:animate-in data-[state=closed]:animate-out data-[state=closed]:fade-out-0 data-[state=open]:fade-in-0 data-[state=closed]:zoom-out-95 data-[state=open]:zoom-in-95 data-[side=bottom]:slide-in-from-top-2 data-[side=left]:slide-in-from-right-2 data-[side=right]:slide-in-from-left-2 data-[side=top]:slide-in-from-bottom-2 relative z-50 max-h-(--radix-select-content-available-height) min-w-[8rem] origin-(--radix-select-content-transform-origin) overflow-x-hidden overflow-y-auto rounded-md border shadow-md",n==="popper"&&"data-[side=bottom]:translate-y-1 data-[side=left]:-translate-x-1 data-[side=right]:translate-x-1 data-[side=top]:-translate-y-1",e),position:n,...r,children:[E.jsx(eX,{}),E.jsx(KZ,{className:ke("p-1",n==="popper"&&"h-[var(--radix-select-trigger-height)] w-full min-w-[var(--radix-select-trigger-width)] scroll-my-1"),children:t}),E.jsx(tX,{})]})})}function Go({className:e,children:t,...n}){return E.jsxs(YZ,{"data-slot":"select-item",className:ke("focus:bg-accent focus:text-accent-foreground [&_svg:not([class*='text-'])]:text-muted-foreground relative flex w-full cursor-default items-center gap-2 rounded-sm py-1.5 pr-8 pl-2 text-sm outline-hidden select-none data-[disabled]:pointer-events-none data-[disabled]:opacity-50 [&_svg]:pointer-events-none [&_svg]:shrink-0 [&_svg:not([class*='size-'])]:size-4 *:[span]:last:flex *:[span]:last:items-center *:[span]:last:gap-2",e),...n,children:[E.jsx("span",{className:"absolute right-2 flex size-3.5 items-center justify-center",children:E.jsx(XZ,{children:E.jsx(jE,{className:"size-4"})})}),E.jsx(ZZ,{children:t})]})}function eX({className:e,...t}){return E.jsx(QZ,{"data-slot":"select-scroll-up-button",className:ke("flex cursor-default items-center justify-center py-1",e),...t,children:E.jsx(r8,{className:"size-4"})})}function tX({className:e,...t}){return E.jsx(JZ,{"data-slot":"select-scroll-down-button",className:ke("flex cursor-default items-center justify-center py-1",e),...t,children:E.jsx(Zm,{className:"size-4"})})}function FF(e,t){return function(){return e.apply(t,arguments)}}const{toString:nX}=Object.prototype,{getPrototypeOf:sC}=Object,{iterator:hv,toStringTag:qF}=Symbol,mv=(e=>t=>{const n=nX.call(t);return e[n]||(e[n]=n.slice(8,-1).toLowerCase())})(Object.create(null)),ti=e=>(e=e.toLowerCase(),t=>mv(t)===e),vv=e=>t=>typeof t===e,{isArray:pu}=Array,vf=vv("undefined");function rX(e){return e!==null&&!vf(e)&&e.constructor!==null&&!vf(e.constructor)&&Zn(e.constructor.isBuffer)&&e.constructor.isBuffer(e)}const zF=ti("ArrayBuffer");function iX(e){let t;return typeof ArrayBuffer<"u"&&ArrayBuffer.isView?t=ArrayBuffer.isView(e):t=e&&e.buffer&&zF(e.buffer),t}const aX=vv("string"),Zn=vv("function"),UF=vv("number"),gv=e=>e!==null&&typeof e=="object",oX=e=>e===!0||e===!1,Eh=e=>{if(mv(e)!=="object")return!1;const t=sC(e);return(t===null||t===Object.prototype||Object.getPrototypeOf(t)===null)&&!(qF in e)&&!(hv in e)},sX=ti("Date"),lX=ti("File"),uX=ti("Blob"),cX=ti("FileList"),fX=e=>gv(e)&&Zn(e.pipe),dX=e=>{let t;return e&&(typeof FormData=="function"&&e instanceof FormData||Zn(e.append)&&((t=mv(e))==="formdata"||t==="object"&&Zn(e.toString)&&e.toString()==="[object FormData]"))},pX=ti("URLSearchParams"),[hX,mX,vX,gX]=["ReadableStream","Request","Response","Headers"].map(ti),yX=e=>e.trim?e.trim():e.replace(/^[\s\uFEFF\xA0]+|[\s\uFEFF\xA0]+$/g,"");function pd(e,t,{allOwnKeys:n=!1}={}){if(e===null||typeof e>"u")return;let r,i;if(typeof e!="object"&&(e=[e]),pu(e))for(r=0,i=e.length;r<i;r++)t.call(null,e[r],r,e);else{const a=n?Object.getOwnPropertyNames(e):Object.keys(e),s=a.length;let u;for(r=0;r<s;r++)u=a[r],t.call(null,e[u],u,e)}}function VF(e,t){t=t.toLowerCase();const n=Object.keys(e);let r=n.length,i;for(;r-- >0;)if(i=n[r],t===i.toLowerCase())return i;return null}const Wo=typeof globalThis<"u"?globalThis:typeof self<"u"?self:typeof window<"u"?window:global,HF=e=>!vf(e)&&e!==Wo;function F_(){const{caseless:e}=HF(this)&&this||{},t={},n=(r,i)=>{const a=e&&VF(t,i)||i;Eh(t[a])&&Eh(r)?t[a]=F_(t[a],r):Eh(r)?t[a]=F_({},r):pu(r)?t[a]=r.slice():t[a]=r};for(let r=0,i=arguments.length;r<i;r++)arguments[r]&&pd(arguments[r],n);return t}const bX=(e,t,n,{allOwnKeys:r}={})=>(pd(t,(i,a)=>{n&&Zn(i)?e[a]=FF(i,n):e[a]=i},{allOwnKeys:r}),e),xX=e=>(e.charCodeAt(0)===65279&&(e=e.slice(1)),e),wX=(e,t,n,r)=>{e.prototype=Object.create(t.prototype,r),e.prototype.constructor=e,Object.defineProperty(e,"super",{value:t.prototype}),n&&Object.assign(e.prototype,n)},SX=(e,t,n,r)=>{let i,a,s;const u={};if(t=t||{},e==null)return t;do{for(i=Object.getOwnPropertyNames(e),a=i.length;a-- >0;)s=i[a],(!r||r(s,e,t))&&!u[s]&&(t[s]=e[s],u[s]=!0);e=n!==!1&&sC(e)}while(e&&(!n||n(e,t))&&e!==Object.prototype);return t},_X=(e,t,n)=>{e=String(e),(n===void 0||n>e.length)&&(n=e.length),n-=t.length;const r=e.indexOf(t,n);return r!==-1&&r===n},EX=e=>{if(!e)return null;if(pu(e))return e;let t=e.length;if(!UF(t))return null;const n=new Array(t);for(;t-- >0;)n[t]=e[t];return n},CX=(e=>t=>e&&t instanceof e)(typeof Uint8Array<"u"&&sC(Uint8Array)),OX=(e,t)=>{const r=(e&&e[hv]).call(e);let i;for(;(i=r.next())&&!i.done;){const a=i.value;t.call(e,a[0],a[1])}},AX=(e,t)=>{let n;const r=[];for(;(n=e.exec(t))!==null;)r.push(n);return r},RX=ti("HTMLFormElement"),TX=e=>e.toLowerCase().replace(/[-_\s]([a-z\d])(\w*)/g,function(n,r,i){return r.toUpperCase()+i}),LP=(({hasOwnProperty:e})=>(t,n)=>e.call(t,n))(Object.prototype),MX=ti("RegExp"),GF=(e,t)=>{const n=Object.getOwnPropertyDescriptors(e),r={};pd(n,(i,a)=>{let s;(s=t(i,a,e))!==!1&&(r[a]=s||i)}),Object.defineProperties(e,r)},PX=e=>{GF(e,(t,n)=>{if(Zn(e)&&["arguments","caller","callee"].indexOf(n)!==-1)return!1;const r=e[n];if(Zn(r)){if(t.enumerable=!1,"writable"in t){t.writable=!1;return}t.set||(t.set=()=>{throw Error("Can not rewrite read-only method '"+n+"'")})}})},jX=(e,t)=>{const n={},r=i=>{i.forEach(a=>{n[a]=!0})};return pu(e)?r(e):r(String(e).split(t)),n},NX=()=>{},DX=(e,t)=>e!=null&&Number.isFinite(e=+e)?e:t;function kX(e){return!!(e&&Zn(e.append)&&e[qF]==="FormData"&&e[hv])}const $X=e=>{const t=new Array(10),n=(r,i)=>{if(gv(r)){if(t.indexOf(r)>=0)return;if(!("toJSON"in r)){t[i]=r;const a=pu(r)?[]:{};return pd(r,(s,u)=>{const c=n(s,i+1);!vf(c)&&(a[u]=c)}),t[i]=void 0,a}}return r};return n(e,0)},IX=ti("AsyncFunction"),LX=e=>e&&(gv(e)||Zn(e))&&Zn(e.then)&&Zn(e.catch),WF=((e,t)=>e?setImmediate:t?((n,r)=>(Wo.addEventListener("message",({source:i,data:a})=>{i===Wo&&a===n&&r.length&&r.shift()()},!1),i=>{r.push(i),Wo.postMessage(n,"*")}))(`axios@${Math.random()}`,[]):n=>setTimeout(n))(typeof setImmediate=="function",Zn(Wo.postMessage)),BX=typeof queueMicrotask<"u"?queueMicrotask.bind(Wo):typeof process<"u"&&process.nextTick||WF,FX=e=>e!=null&&Zn(e[hv]),ue={isArray:pu,isArrayBuffer:zF,isBuffer:rX,isFormData:dX,isArrayBufferView:iX,isString:aX,isNumber:UF,isBoolean:oX,isObject:gv,isPlainObject:Eh,isReadableStream:hX,isRequest:mX,isResponse:vX,isHeaders:gX,isUndefined:vf,isDate:sX,isFile:lX,isBlob:uX,isRegExp:MX,isFunction:Zn,isStream:fX,isURLSearchParams:pX,isTypedArray:CX,isFileList:cX,forEach:pd,merge:F_,extend:bX,trim:yX,stripBOM:xX,inherits:wX,toFlatObject:SX,kindOf:mv,kindOfTest:ti,endsWith:_X,toArray:EX,forEachEntry:OX,matchAll:AX,isHTMLForm:RX,hasOwnProperty:LP,hasOwnProp:LP,reduceDescriptors:GF,freezeMethods:PX,toObjectSet:jX,toCamelCase:TX,noop:NX,toFiniteNumber:DX,findKey:VF,global:Wo,isContextDefined:HF,isSpecCompliantForm:kX,toJSONObject:$X,isAsyncFn:IX,isThenable:LX,setImmediate:WF,asap:BX,isIterable:FX};function Ze(e,t,n,r,i){Error.call(this),Error.captureStackTrace?Error.captureStackTrace(this,this.constructor):this.stack=new Error().stack,this.message=e,this.name="AxiosError",t&&(this.code=t),n&&(this.config=n),r&&(this.request=r),i&&(this.response=i,this.status=i.status?i.status:null)}ue.inherits(Ze,Error,{toJSON:function(){return{message:this.message,name:this.name,description:this.description,number:this.number,fileName:this.fileName,lineNumber:this.lineNumber,columnNumber:this.columnNumber,stack:this.stack,config:ue.toJSONObject(this.config),code:this.code,status:this.status}}});const KF=Ze.prototype,YF={};["ERR_BAD_OPTION_VALUE","ERR_BAD_OPTION","ECONNABORTED","ETIMEDOUT","ERR_NETWORK","ERR_FR_TOO_MANY_REDIRECTS","ERR_DEPRECATED","ERR_BAD_RESPONSE","ERR_BAD_REQUEST","ERR_CANCELED","ERR_NOT_SUPPORT","ERR_INVALID_URL"].forEach(e=>{YF[e]={value:e}});Object.defineProperties(Ze,YF);Object.defineProperty(KF,"isAxiosError",{value:!0});Ze.from=(e,t,n,r,i,a)=>{const s=Object.create(KF);return ue.toFlatObject(e,s,function(c){return c!==Error.prototype},u=>u!=="isAxiosError"),Ze.call(s,e.message,t,n,r,i),s.cause=e,s.name=e.name,a&&Object.assign(s,a),s};const qX=null;function q_(e){return ue.isPlainObject(e)||ue.isArray(e)}function ZF(e){return ue.endsWith(e,"[]")?e.slice(0,-2):e}function BP(e,t,n){return e?e.concat(t).map(function(i,a){return i=ZF(i),!n&&a?"["+i+"]":i}).join(n?".":""):t}function zX(e){return ue.isArray(e)&&!e.some(q_)}const UX=ue.toFlatObject(ue,{},null,function(t){return/^is[A-Z]/.test(t)});function yv(e,t,n){if(!ue.isObject(e))throw new TypeError("target must be an object");t=t||new FormData,n=ue.toFlatObject(n,{metaTokens:!0,dots:!1,indexes:!1},!1,function(w,x){return!ue.isUndefined(x[w])});const r=n.metaTokens,i=n.visitor||p,a=n.dots,s=n.indexes,c=(n.Blob||typeof Blob<"u"&&Blob)&&ue.isSpecCompliantForm(t);if(!ue.isFunction(i))throw new TypeError("visitor must be a function");function d(S){if(S===null)return"";if(ue.isDate(S))return S.toISOString();if(!c&&ue.isBlob(S))throw new Ze("Blob is not supported. Use a Buffer instead.");return ue.isArrayBuffer(S)||ue.isTypedArray(S)?c&&typeof Blob=="function"?new Blob([S]):Buffer.from(S):S}function p(S,w,x){let C=S;if(S&&!x&&typeof S=="object"){if(ue.endsWith(w,"{}"))w=r?w:w.slice(0,-2),S=JSON.stringify(S);else if(ue.isArray(S)&&zX(S)||(ue.isFileList(S)||ue.endsWith(w,"[]"))&&(C=ue.toArray(S)))return w=ZF(w),C.forEach(function(A,T){!(ue.isUndefined(A)||A===null)&&t.append(s===!0?BP([w],T,a):s===null?w:w+"[]",d(A))}),!1}return q_(S)?!0:(t.append(BP(x,w,a),d(S)),!1)}const m=[],v=Object.assign(UX,{defaultVisitor:p,convertValue:d,isVisitable:q_});function b(S,w){if(!ue.isUndefined(S)){if(m.indexOf(S)!==-1)throw Error("Circular reference detected in "+w.join("."));m.push(S),ue.forEach(S,function(C,O){(!(ue.isUndefined(C)||C===null)&&i.call(t,C,ue.isString(O)?O.trim():O,w,v))===!0&&b(C,w?w.concat(O):[O])}),m.pop()}}if(!ue.isObject(e))throw new TypeError("data must be an object");return b(e),t}function FP(e){const t={"!":"%21","'":"%27","(":"%28",")":"%29","~":"%7E","%20":"+","%00":"\0"};return encodeURIComponent(e).replace(/[!'()~]|%20|%00/g,function(r){return t[r]})}function lC(e,t){this._pairs=[],e&&yv(e,this,t)}const XF=lC.prototype;XF.append=function(t,n){this._pairs.push([t,n])};XF.toString=function(t){const n=t?function(r){return t.call(this,r,FP)}:FP;return this._pairs.map(function(i){return n(i[0])+"="+n(i[1])},"").join("&")};function VX(e){return encodeURIComponent(e).replace(/%3A/gi,":").replace(/%24/g,"$").replace(/%2C/gi,",").replace(/%20/g,"+").replace(/%5B/gi,"[").replace(/%5D/gi,"]")}function QF(e,t,n){if(!t)return e;const r=n&&n.encode||VX;ue.isFunction(n)&&(n={serialize:n});const i=n&&n.serialize;let a;if(i?a=i(t,n):a=ue.isURLSearchParams(t)?t.toString():new lC(t,n).toString(r),a){const s=e.indexOf("#");s!==-1&&(e=e.slice(0,s)),e+=(e.indexOf("?")===-1?"?":"&")+a}return e}class qP{constructor(){this.handlers=[]}use(t,n,r){return this.handlers.push({fulfilled:t,rejected:n,synchronous:r?r.synchronous:!1,runWhen:r?r.runWhen:null}),this.handlers.length-1}eject(t){this.handlers[t]&&(this.handlers[t]=null)}clear(){this.handlers&&(this.handlers=[])}forEach(t){ue.forEach(this.handlers,function(r){r!==null&&t(r)})}}const JF={silentJSONParsing:!0,forcedJSONParsing:!0,clarifyTimeoutError:!1},HX=typeof URLSearchParams<"u"?URLSearchParams:lC,GX=typeof FormData<"u"?FormData:null,WX=typeof Blob<"u"?Blob:null,KX={isBrowser:!0,classes:{URLSearchParams:HX,FormData:GX,Blob:WX},protocols:["http","https","file","blob","url","data"]},uC=typeof window<"u"&&typeof document<"u",z_=typeof navigator=="object"&&navigator||void 0,YX=uC&&(!z_||["ReactNative","NativeScript","NS"].indexOf(z_.product)<0),ZX=typeof WorkerGlobalScope<"u"&&self instanceof WorkerGlobalScope&&typeof self.importScripts=="function",XX=uC&&window.location.href||"http://localhost",QX=Object.freeze(Object.defineProperty({__proto__:null,hasBrowserEnv:uC,hasStandardBrowserEnv:YX,hasStandardBrowserWebWorkerEnv:ZX,navigator:z_,origin:XX},Symbol.toStringTag,{value:"Module"})),Mn={...QX,...KX};function JX(e,t){return yv(e,new Mn.classes.URLSearchParams,Object.assign({visitor:function(n,r,i,a){return Mn.isNode&&ue.isBuffer(n)?(this.append(r,n.toString("base64")),!1):a.defaultVisitor.apply(this,arguments)}},t))}function eQ(e){return ue.matchAll(/\w+|\[(\w*)]/g,e).map(t=>t[0]==="[]"?"":t[1]||t[0])}function tQ(e){const t={},n=Object.keys(e);let r;const i=n.length;let a;for(r=0;r<i;r++)a=n[r],t[a]=e[a];return t}function eq(e){function t(n,r,i,a){let s=n[a++];if(s==="__proto__")return!0;const u=Number.isFinite(+s),c=a>=n.length;return s=!s&&ue.isArray(i)?i.length:s,c?(ue.hasOwnProp(i,s)?i[s]=[i[s],r]:i[s]=r,!u):((!i[s]||!ue.isObject(i[s]))&&(i[s]=[]),t(n,r,i[s],a)&&ue.isArray(i[s])&&(i[s]=tQ(i[s])),!u)}if(ue.isFormData(e)&&ue.isFunction(e.entries)){const n={};return ue.forEachEntry(e,(r,i)=>{t(eQ(r),i,n,0)}),n}return null}function nQ(e,t,n){if(ue.isString(e))try{return(t||JSON.parse)(e),ue.trim(e)}catch(r){if(r.name!=="SyntaxError")throw r}return(n||JSON.stringify)(e)}const hd={transitional:JF,adapter:["xhr","http","fetch"],transformRequest:[function(t,n){const r=n.getContentType()||"",i=r.indexOf("application/json")>-1,a=ue.isObject(t);if(a&&ue.isHTMLForm(t)&&(t=new FormData(t)),ue.isFormData(t))return i?JSON.stringify(eq(t)):t;if(ue.isArrayBuffer(t)||ue.isBuffer(t)||ue.isStream(t)||ue.isFile(t)||ue.isBlob(t)||ue.isReadableStream(t))return t;if(ue.isArrayBufferView(t))return t.buffer;if(ue.isURLSearchParams(t))return n.setContentType("application/x-www-form-urlencoded;charset=utf-8",!1),t.toString();let u;if(a){if(r.indexOf("application/x-www-form-urlencoded")>-1)return JX(t,this.formSerializer).toString();if((u=ue.isFileList(t))||r.indexOf("multipart/form-data")>-1){const c=this.env&&this.env.FormData;return yv(u?{"files[]":t}:t,c&&new c,this.formSerializer)}}return a||i?(n.setContentType("application/json",!1),nQ(t)):t}],transformResponse:[function(t){const n=this.transitional||hd.transitional,r=n&&n.forcedJSONParsing,i=this.responseType==="json";if(ue.isResponse(t)||ue.isReadableStream(t))return t;if(t&&ue.isString(t)&&(r&&!this.responseType||i)){const s=!(n&&n.silentJSONParsing)&&i;try{return JSON.parse(t)}catch(u){if(s)throw u.name==="SyntaxError"?Ze.from(u,Ze.ERR_BAD_RESPONSE,this,null,this.response):u}}return t}],timeout:0,xsrfCookieName:"XSRF-TOKEN",xsrfHeaderName:"X-XSRF-TOKEN",maxContentLength:-1,maxBodyLength:-1,env:{FormData:Mn.classes.FormData,Blob:Mn.classes.Blob},validateStatus:function(t){return t>=200&&t<300},headers:{common:{Accept:"application/json, text/plain, */*","Content-Type":void 0}}};ue.forEach(["delete","get","head","post","put","patch"],e=>{hd.headers[e]={}});const rQ=ue.toObjectSet(["age","authorization","content-length","content-type","etag","expires","from","host","if-modified-since","if-unmodified-since","last-modified","location","max-forwards","proxy-authorization","referer","retry-after","user-agent"]),iQ=e=>{const t={};let n,r,i;return e&&e.split(`
`).forEach(function(s){i=s.indexOf(":"),n=s.substring(0,i).trim().toLowerCase(),r=s.substring(i+1).trim(),!(!n||t[n]&&rQ[n])&&(n==="set-cookie"?t[n]?t[n].push(r):t[n]=[r]:t[n]=t[n]?t[n]+", "+r:r)}),t},zP=Symbol("internals");function Oc(e){return e&&String(e).trim().toLowerCase()}function Ch(e){return e===!1||e==null?e:ue.isArray(e)?e.map(Ch):String(e)}function aQ(e){const t=Object.create(null),n=/([^\s,;=]+)\s*(?:=\s*([^,;]+))?/g;let r;for(;r=n.exec(e);)t[r[1]]=r[2];return t}const oQ=e=>/^[-_a-zA-Z0-9^`|~,!#$%&'*+.]+$/.test(e.trim());function Yb(e,t,n,r,i){if(ue.isFunction(r))return r.call(this,t,n);if(i&&(t=n),!!ue.isString(t)){if(ue.isString(r))return t.indexOf(r)!==-1;if(ue.isRegExp(r))return r.test(t)}}function sQ(e){return e.trim().toLowerCase().replace(/([a-z\d])(\w*)/g,(t,n,r)=>n.toUpperCase()+r)}function lQ(e,t){const n=ue.toCamelCase(" "+t);["get","set","has"].forEach(r=>{Object.defineProperty(e,r+n,{value:function(i,a,s){return this[r].call(this,t,i,a,s)},configurable:!0})})}let Xn=class{constructor(t){t&&this.set(t)}set(t,n,r){const i=this;function a(u,c,d){const p=Oc(c);if(!p)throw new Error("header name must be a non-empty string");const m=ue.findKey(i,p);(!m||i[m]===void 0||d===!0||d===void 0&&i[m]!==!1)&&(i[m||c]=Ch(u))}const s=(u,c)=>ue.forEach(u,(d,p)=>a(d,p,c));if(ue.isPlainObject(t)||t instanceof this.constructor)s(t,n);else if(ue.isString(t)&&(t=t.trim())&&!oQ(t))s(iQ(t),n);else if(ue.isObject(t)&&ue.isIterable(t)){let u={},c,d;for(const p of t){if(!ue.isArray(p))throw TypeError("Object iterator must return a key-value pair");u[d=p[0]]=(c=u[d])?ue.isArray(c)?[...c,p[1]]:[c,p[1]]:p[1]}s(u,n)}else t!=null&&a(n,t,r);return this}get(t,n){if(t=Oc(t),t){const r=ue.findKey(this,t);if(r){const i=this[r];if(!n)return i;if(n===!0)return aQ(i);if(ue.isFunction(n))return n.call(this,i,r);if(ue.isRegExp(n))return n.exec(i);throw new TypeError("parser must be boolean|regexp|function")}}}has(t,n){if(t=Oc(t),t){const r=ue.findKey(this,t);return!!(r&&this[r]!==void 0&&(!n||Yb(this,this[r],r,n)))}return!1}delete(t,n){const r=this;let i=!1;function a(s){if(s=Oc(s),s){const u=ue.findKey(r,s);u&&(!n||Yb(r,r[u],u,n))&&(delete r[u],i=!0)}}return ue.isArray(t)?t.forEach(a):a(t),i}clear(t){const n=Object.keys(this);let r=n.length,i=!1;for(;r--;){const a=n[r];(!t||Yb(this,this[a],a,t,!0))&&(delete this[a],i=!0)}return i}normalize(t){const n=this,r={};return ue.forEach(this,(i,a)=>{const s=ue.findKey(r,a);if(s){n[s]=Ch(i),delete n[a];return}const u=t?sQ(a):String(a).trim();u!==a&&delete n[a],n[u]=Ch(i),r[u]=!0}),this}concat(...t){return this.constructor.concat(this,...t)}toJSON(t){const n=Object.create(null);return ue.forEach(this,(r,i)=>{r!=null&&r!==!1&&(n[i]=t&&ue.isArray(r)?r.join(", "):r)}),n}[Symbol.iterator](){return Object.entries(this.toJSON())[Symbol.iterator]()}toString(){return Object.entries(this.toJSON()).map(([t,n])=>t+": "+n).join(`
`)}getSetCookie(){return this.get("set-cookie")||[]}get[Symbol.toStringTag](){return"AxiosHeaders"}static from(t){return t instanceof this?t:new this(t)}static concat(t,...n){const r=new this(t);return n.forEach(i=>r.set(i)),r}static accessor(t){const r=(this[zP]=this[zP]={accessors:{}}).accessors,i=this.prototype;function a(s){const u=Oc(s);r[u]||(lQ(i,s),r[u]=!0)}return ue.isArray(t)?t.forEach(a):a(t),this}};Xn.accessor(["Content-Type","Content-Length","Accept","Accept-Encoding","User-Agent","Authorization"]);ue.reduceDescriptors(Xn.prototype,({value:e},t)=>{let n=t[0].toUpperCase()+t.slice(1);return{get:()=>e,set(r){this[n]=r}}});ue.freezeMethods(Xn);function Zb(e,t){const n=this||hd,r=t||n,i=Xn.from(r.headers);let a=r.data;return ue.forEach(e,function(u){a=u.call(n,a,i.normalize(),t?t.status:void 0)}),i.normalize(),a}function tq(e){return!!(e&&e.__CANCEL__)}function hu(e,t,n){Ze.call(this,e??"canceled",Ze.ERR_CANCELED,t,n),this.name="CanceledError"}ue.inherits(hu,Ze,{__CANCEL__:!0});function nq(e,t,n){const r=n.config.validateStatus;!n.status||!r||r(n.status)?e(n):t(new Ze("Request failed with status code "+n.status,[Ze.ERR_BAD_REQUEST,Ze.ERR_BAD_RESPONSE][Math.floor(n.status/100)-4],n.config,n.request,n))}function uQ(e){const t=/^([-+\w]{1,25})(:?\/\/|:)/.exec(e);return t&&t[1]||""}function cQ(e,t){e=e||10;const n=new Array(e),r=new Array(e);let i=0,a=0,s;return t=t!==void 0?t:1e3,function(c){const d=Date.now(),p=r[a];s||(s=d),n[i]=c,r[i]=d;let m=a,v=0;for(;m!==i;)v+=n[m++],m=m%e;if(i=(i+1)%e,i===a&&(a=(a+1)%e),d-s<t)return;const b=p&&d-p;return b?Math.round(v*1e3/b):void 0}}function fQ(e,t){let n=0,r=1e3/t,i,a;const s=(d,p=Date.now())=>{n=p,i=null,a&&(clearTimeout(a),a=null),e.apply(null,d)};return[(...d)=>{const p=Date.now(),m=p-n;m>=r?s(d,p):(i=d,a||(a=setTimeout(()=>{a=null,s(i)},r-m)))},()=>i&&s(i)]}const Qh=(e,t,n=3)=>{let r=0;const i=cQ(50,250);return fQ(a=>{const s=a.loaded,u=a.lengthComputable?a.total:void 0,c=s-r,d=i(c),p=s<=u;r=s;const m={loaded:s,total:u,progress:u?s/u:void 0,bytes:c,rate:d||void 0,estimated:d&&u&&p?(u-s)/d:void 0,event:a,lengthComputable:u!=null,[t?"download":"upload"]:!0};e(m)},n)},UP=(e,t)=>{const n=e!=null;return[r=>t[0]({lengthComputable:n,total:e,loaded:r}),t[1]]},VP=e=>(...t)=>ue.asap(()=>e(...t)),dQ=Mn.hasStandardBrowserEnv?((e,t)=>n=>(n=new URL(n,Mn.origin),e.protocol===n.protocol&&e.host===n.host&&(t||e.port===n.port)))(new URL(Mn.origin),Mn.navigator&&/(msie|trident)/i.test(Mn.navigator.userAgent)):()=>!0,pQ=Mn.hasStandardBrowserEnv?{write(e,t,n,r,i,a){const s=[e+"="+encodeURIComponent(t)];ue.isNumber(n)&&s.push("expires="+new Date(n).toGMTString()),ue.isString(r)&&s.push("path="+r),ue.isString(i)&&s.push("domain="+i),a===!0&&s.push("secure"),document.cookie=s.join("; ")},read(e){const t=document.cookie.match(new RegExp("(^|;\\s*)("+e+")=([^;]*)"));return t?decodeURIComponent(t[3]):null},remove(e){this.write(e,"",Date.now()-864e5)}}:{write(){},read(){return null},remove(){}};function hQ(e){return/^([a-z][a-z\d+\-.]*:)?\/\//i.test(e)}function mQ(e,t){return t?e.replace(/\/?\/$/,"")+"/"+t.replace(/^\/+/,""):e}function rq(e,t,n){let r=!hQ(t);return e&&(r||n==!1)?mQ(e,t):t}const HP=e=>e instanceof Xn?{...e}:e;function cs(e,t){t=t||{};const n={};function r(d,p,m,v){return ue.isPlainObject(d)&&ue.isPlainObject(p)?ue.merge.call({caseless:v},d,p):ue.isPlainObject(p)?ue.merge({},p):ue.isArray(p)?p.slice():p}function i(d,p,m,v){if(ue.isUndefined(p)){if(!ue.isUndefined(d))return r(void 0,d,m,v)}else return r(d,p,m,v)}function a(d,p){if(!ue.isUndefined(p))return r(void 0,p)}function s(d,p){if(ue.isUndefined(p)){if(!ue.isUndefined(d))return r(void 0,d)}else return r(void 0,p)}function u(d,p,m){if(m in t)return r(d,p);if(m in e)return r(void 0,d)}const c={url:a,method:a,data:a,baseURL:s,transformRequest:s,transformResponse:s,paramsSerializer:s,timeout:s,timeoutMessage:s,withCredentials:s,withXSRFToken:s,adapter:s,responseType:s,xsrfCookieName:s,xsrfHeaderName:s,onUploadProgress:s,onDownloadProgress:s,decompress:s,maxContentLength:s,maxBodyLength:s,beforeRedirect:s,transport:s,httpAgent:s,httpsAgent:s,cancelToken:s,socketPath:s,responseEncoding:s,validateStatus:u,headers:(d,p,m)=>i(HP(d),HP(p),m,!0)};return ue.forEach(Object.keys(Object.assign({},e,t)),function(p){const m=c[p]||i,v=m(e[p],t[p],p);ue.isUndefined(v)&&m!==u||(n[p]=v)}),n}const iq=e=>{const t=cs({},e);let{data:n,withXSRFToken:r,xsrfHeaderName:i,xsrfCookieName:a,headers:s,auth:u}=t;t.headers=s=Xn.from(s),t.url=QF(rq(t.baseURL,t.url,t.allowAbsoluteUrls),e.params,e.paramsSerializer),u&&s.set("Authorization","Basic "+btoa((u.username||"")+":"+(u.password?unescape(encodeURIComponent(u.password)):"")));let c;if(ue.isFormData(n)){if(Mn.hasStandardBrowserEnv||Mn.hasStandardBrowserWebWorkerEnv)s.setContentType(void 0);else if((c=s.getContentType())!==!1){const[d,...p]=c?c.split(";").map(m=>m.trim()).filter(Boolean):[];s.setContentType([d||"multipart/form-data",...p].join("; "))}}if(Mn.hasStandardBrowserEnv&&(r&&ue.isFunction(r)&&(r=r(t)),r||r!==!1&&dQ(t.url))){const d=i&&a&&pQ.read(a);d&&s.set(i,d)}return t},vQ=typeof XMLHttpRequest<"u",gQ=vQ&&function(e){return new Promise(function(n,r){const i=iq(e);let a=i.data;const s=Xn.from(i.headers).normalize();let{responseType:u,onUploadProgress:c,onDownloadProgress:d}=i,p,m,v,b,S;function w(){b&&b(),S&&S(),i.cancelToken&&i.cancelToken.unsubscribe(p),i.signal&&i.signal.removeEventListener("abort",p)}let x=new XMLHttpRequest;x.open(i.method.toUpperCase(),i.url,!0),x.timeout=i.timeout;function C(){if(!x)return;const A=Xn.from("getAllResponseHeaders"in x&&x.getAllResponseHeaders()),R={data:!u||u==="text"||u==="json"?x.responseText:x.response,status:x.status,statusText:x.statusText,headers:A,config:e,request:x};nq(function(P){n(P),w()},function(P){r(P),w()},R),x=null}"onloadend"in x?x.onloadend=C:x.onreadystatechange=function(){!x||x.readyState!==4||x.status===0&&!(x.responseURL&&x.responseURL.indexOf("file:")===0)||setTimeout(C)},x.onabort=function(){x&&(r(new Ze("Request aborted",Ze.ECONNABORTED,e,x)),x=null)},x.onerror=function(){r(new Ze("Network Error",Ze.ERR_NETWORK,e,x)),x=null},x.ontimeout=function(){let T=i.timeout?"timeout of "+i.timeout+"ms exceeded":"timeout exceeded";const R=i.transitional||JF;i.timeoutErrorMessage&&(T=i.timeoutErrorMessage),r(new Ze(T,R.clarifyTimeoutError?Ze.ETIMEDOUT:Ze.ECONNABORTED,e,x)),x=null},a===void 0&&s.setContentType(null),"setRequestHeader"in x&&ue.forEach(s.toJSON(),function(T,R){x.setRequestHeader(R,T)}),ue.isUndefined(i.withCredentials)||(x.withCredentials=!!i.withCredentials),u&&u!=="json"&&(x.responseType=i.responseType),d&&([v,S]=Qh(d,!0),x.addEventListener("progress",v)),c&&x.upload&&([m,b]=Qh(c),x.upload.addEventListener("progress",m),x.upload.addEventListener("loadend",b)),(i.cancelToken||i.signal)&&(p=A=>{x&&(r(!A||A.type?new hu(null,e,x):A),x.abort(),x=null)},i.cancelToken&&i.cancelToken.subscribe(p),i.signal&&(i.signal.aborted?p():i.signal.addEventListener("abort",p)));const O=uQ(i.url);if(O&&Mn.protocols.indexOf(O)===-1){r(new Ze("Unsupported protocol "+O+":",Ze.ERR_BAD_REQUEST,e));return}x.send(a||null)})},yQ=(e,t)=>{const{length:n}=e=e?e.filter(Boolean):[];if(t||n){let r=new AbortController,i;const a=function(d){if(!i){i=!0,u();const p=d instanceof Error?d:this.reason;r.abort(p instanceof Ze?p:new hu(p instanceof Error?p.message:p))}};let s=t&&setTimeout(()=>{s=null,a(new Ze(`timeout ${t} of ms exceeded`,Ze.ETIMEDOUT))},t);const u=()=>{e&&(s&&clearTimeout(s),s=null,e.forEach(d=>{d.unsubscribe?d.unsubscribe(a):d.removeEventListener("abort",a)}),e=null)};e.forEach(d=>d.addEventListener("abort",a));const{signal:c}=r;return c.unsubscribe=()=>ue.asap(u),c}},bQ=function*(e,t){let n=e.byteLength;if(n<t){yield e;return}let r=0,i;for(;r<n;)i=r+t,yield e.slice(r,i),r=i},xQ=async function*(e,t){for await(const n of wQ(e))yield*bQ(n,t)},wQ=async function*(e){if(e[Symbol.asyncIterator]){yield*e;return}const t=e.getReader();try{for(;;){const{done:n,value:r}=await t.read();if(n)break;yield r}}finally{await t.cancel()}},GP=(e,t,n,r)=>{const i=xQ(e,t);let a=0,s,u=c=>{s||(s=!0,r&&r(c))};return new ReadableStream({async pull(c){try{const{done:d,value:p}=await i.next();if(d){u(),c.close();return}let m=p.byteLength;if(n){let v=a+=m;n(v)}c.enqueue(new Uint8Array(p))}catch(d){throw u(d),d}},cancel(c){return u(c),i.return()}},{highWaterMark:2})},bv=typeof fetch=="function"&&typeof Request=="function"&&typeof Response=="function",aq=bv&&typeof ReadableStream=="function",SQ=bv&&(typeof TextEncoder=="function"?(e=>t=>e.encode(t))(new TextEncoder):async e=>new Uint8Array(await new Response(e).arrayBuffer())),oq=(e,...t)=>{try{return!!e(...t)}catch{return!1}},_Q=aq&&oq(()=>{let e=!1;const t=new Request(Mn.origin,{body:new ReadableStream,method:"POST",get duplex(){return e=!0,"half"}}).headers.has("Content-Type");return e&&!t}),WP=64*1024,U_=aq&&oq(()=>ue.isReadableStream(new Response("").body)),Jh={stream:U_&&(e=>e.body)};bv&&(e=>{["text","arrayBuffer","blob","formData","stream"].forEach(t=>{!Jh[t]&&(Jh[t]=ue.isFunction(e[t])?n=>n[t]():(n,r)=>{throw new Ze(`Response type '${t}' is not supported`,Ze.ERR_NOT_SUPPORT,r)})})})(new Response);const EQ=async e=>{if(e==null)return 0;if(ue.isBlob(e))return e.size;if(ue.isSpecCompliantForm(e))return(await new Request(Mn.origin,{method:"POST",body:e}).arrayBuffer()).byteLength;if(ue.isArrayBufferView(e)||ue.isArrayBuffer(e))return e.byteLength;if(ue.isURLSearchParams(e)&&(e=e+""),ue.isString(e))return(await SQ(e)).byteLength},CQ=async(e,t)=>{const n=ue.toFiniteNumber(e.getContentLength());return n??EQ(t)},OQ=bv&&(async e=>{let{url:t,method:n,data:r,signal:i,cancelToken:a,timeout:s,onDownloadProgress:u,onUploadProgress:c,responseType:d,headers:p,withCredentials:m="same-origin",fetchOptions:v}=iq(e);d=d?(d+"").toLowerCase():"text";let b=yQ([i,a&&a.toAbortSignal()],s),S;const w=b&&b.unsubscribe&&(()=>{b.unsubscribe()});let x;try{if(c&&_Q&&n!=="get"&&n!=="head"&&(x=await CQ(p,r))!==0){let R=new Request(t,{method:"POST",body:r,duplex:"half"}),M;if(ue.isFormData(r)&&(M=R.headers.get("content-type"))&&p.setContentType(M),R.body){const[P,N]=UP(x,Qh(VP(c)));r=GP(R.body,WP,P,N)}}ue.isString(m)||(m=m?"include":"omit");const C="credentials"in Request.prototype;S=new Request(t,{...v,signal:b,method:n.toUpperCase(),headers:p.normalize().toJSON(),body:r,duplex:"half",credentials:C?m:void 0});let O=await fetch(S);const A=U_&&(d==="stream"||d==="response");if(U_&&(u||A&&w)){const R={};["status","statusText","headers"].forEach(k=>{R[k]=O[k]});const M=ue.toFiniteNumber(O.headers.get("content-length")),[P,N]=u&&UP(M,Qh(VP(u),!0))||[];O=new Response(GP(O.body,WP,P,()=>{N&&N(),w&&w()}),R)}d=d||"text";let T=await Jh[ue.findKey(Jh,d)||"text"](O,e);return!A&&w&&w(),await new Promise((R,M)=>{nq(R,M,{data:T,headers:Xn.from(O.headers),status:O.status,statusText:O.statusText,config:e,request:S})})}catch(C){throw w&&w(),C&&C.name==="TypeError"&&/Load failed|fetch/i.test(C.message)?Object.assign(new Ze("Network Error",Ze.ERR_NETWORK,e,S),{cause:C.cause||C}):Ze.from(C,C&&C.code,e,S)}}),V_={http:qX,xhr:gQ,fetch:OQ};ue.forEach(V_,(e,t)=>{if(e){try{Object.defineProperty(e,"name",{value:t})}catch{}Object.defineProperty(e,"adapterName",{value:t})}});const KP=e=>`- ${e}`,AQ=e=>ue.isFunction(e)||e===null||e===!1,sq={getAdapter:e=>{e=ue.isArray(e)?e:[e];const{length:t}=e;let n,r;const i={};for(let a=0;a<t;a++){n=e[a];let s;if(r=n,!AQ(n)&&(r=V_[(s=String(n)).toLowerCase()],r===void 0))throw new Ze(`Unknown adapter '${s}'`);if(r)break;i[s||"#"+a]=r}if(!r){const a=Object.entries(i).map(([u,c])=>`adapter ${u} `+(c===!1?"is not supported by the environment":"is not available in the build"));let s=t?a.length>1?`since :
`+a.map(KP).join(`
`):" "+KP(a[0]):"as no adapter specified";throw new Ze("There is no suitable adapter to dispatch the request "+s,"ERR_NOT_SUPPORT")}return r},adapters:V_};function Xb(e){if(e.cancelToken&&e.cancelToken.throwIfRequested(),e.signal&&e.signal.aborted)throw new hu(null,e)}function YP(e){return Xb(e),e.headers=Xn.from(e.headers),e.data=Zb.call(e,e.transformRequest),["post","put","patch"].indexOf(e.method)!==-1&&e.headers.setContentType("application/x-www-form-urlencoded",!1),sq.getAdapter(e.adapter||hd.adapter)(e).then(function(r){return Xb(e),r.data=Zb.call(e,e.transformResponse,r),r.headers=Xn.from(r.headers),r},function(r){return tq(r)||(Xb(e),r&&r.response&&(r.response.data=Zb.call(e,e.transformResponse,r.response),r.response.headers=Xn.from(r.response.headers))),Promise.reject(r)})}const lq="1.9.0",xv={};["object","boolean","number","function","string","symbol"].forEach((e,t)=>{xv[e]=function(r){return typeof r===e||"a"+(t<1?"n ":" ")+e}});const ZP={};xv.transitional=function(t,n,r){function i(a,s){return"[Axios v"+lq+"] Transitional option '"+a+"'"+s+(r?". "+r:"")}return(a,s,u)=>{if(t===!1)throw new Ze(i(s," has been removed"+(n?" in "+n:"")),Ze.ERR_DEPRECATED);return n&&!ZP[s]&&(ZP[s]=!0,console.warn(i(s," has been deprecated since v"+n+" and will be removed in the near future"))),t?t(a,s,u):!0}};xv.spelling=function(t){return(n,r)=>(console.warn(`${r} is likely a misspelling of ${t}`),!0)};function RQ(e,t,n){if(typeof e!="object")throw new Ze("options must be an object",Ze.ERR_BAD_OPTION_VALUE);const r=Object.keys(e);let i=r.length;for(;i-- >0;){const a=r[i],s=t[a];if(s){const u=e[a],c=u===void 0||s(u,a,e);if(c!==!0)throw new Ze("option "+a+" must be "+c,Ze.ERR_BAD_OPTION_VALUE);continue}if(n!==!0)throw new Ze("Unknown option "+a,Ze.ERR_BAD_OPTION)}}const Oh={assertOptions:RQ,validators:xv},ui=Oh.validators;let Qo=class{constructor(t){this.defaults=t||{},this.interceptors={request:new qP,response:new qP}}async request(t,n){try{return await this._request(t,n)}catch(r){if(r instanceof Error){let i={};Error.captureStackTrace?Error.captureStackTrace(i):i=new Error;const a=i.stack?i.stack.replace(/^.+\n/,""):"";try{r.stack?a&&!String(r.stack).endsWith(a.replace(/^.+\n.+\n/,""))&&(r.stack+=`
`+a):r.stack=a}catch{}}throw r}}_request(t,n){typeof t=="string"?(n=n||{},n.url=t):n=t||{},n=cs(this.defaults,n);const{transitional:r,paramsSerializer:i,headers:a}=n;r!==void 0&&Oh.assertOptions(r,{silentJSONParsing:ui.transitional(ui.boolean),forcedJSONParsing:ui.transitional(ui.boolean),clarifyTimeoutError:ui.transitional(ui.boolean)},!1),i!=null&&(ue.isFunction(i)?n.paramsSerializer={serialize:i}:Oh.assertOptions(i,{encode:ui.function,serialize:ui.function},!0)),n.allowAbsoluteUrls!==void 0||(this.defaults.allowAbsoluteUrls!==void 0?n.allowAbsoluteUrls=this.defaults.allowAbsoluteUrls:n.allowAbsoluteUrls=!0),Oh.assertOptions(n,{baseUrl:ui.spelling("baseURL"),withXsrfToken:ui.spelling("withXSRFToken")},!0),n.method=(n.method||this.defaults.method||"get").toLowerCase();let s=a&&ue.merge(a.common,a[n.method]);a&&ue.forEach(["delete","get","head","post","put","patch","common"],S=>{delete a[S]}),n.headers=Xn.concat(s,a);const u=[];let c=!0;this.interceptors.request.forEach(function(w){typeof w.runWhen=="function"&&w.runWhen(n)===!1||(c=c&&w.synchronous,u.unshift(w.fulfilled,w.rejected))});const d=[];this.interceptors.response.forEach(function(w){d.push(w.fulfilled,w.rejected)});let p,m=0,v;if(!c){const S=[YP.bind(this),void 0];for(S.unshift.apply(S,u),S.push.apply(S,d),v=S.length,p=Promise.resolve(n);m<v;)p=p.then(S[m++],S[m++]);return p}v=u.length;let b=n;for(m=0;m<v;){const S=u[m++],w=u[m++];try{b=S(b)}catch(x){w.call(this,x);break}}try{p=YP.call(this,b)}catch(S){return Promise.reject(S)}for(m=0,v=d.length;m<v;)p=p.then(d[m++],d[m++]);return p}getUri(t){t=cs(this.defaults,t);const n=rq(t.baseURL,t.url,t.allowAbsoluteUrls);return QF(n,t.params,t.paramsSerializer)}};ue.forEach(["delete","get","head","options"],function(t){Qo.prototype[t]=function(n,r){return this.request(cs(r||{},{method:t,url:n,data:(r||{}).data}))}});ue.forEach(["post","put","patch"],function(t){function n(r){return function(a,s,u){return this.request(cs(u||{},{method:t,headers:r?{"Content-Type":"multipart/form-data"}:{},url:a,data:s}))}}Qo.prototype[t]=n(),Qo.prototype[t+"Form"]=n(!0)});let TQ=class uq{constructor(t){if(typeof t!="function")throw new TypeError("executor must be a function.");let n;this.promise=new Promise(function(a){n=a});const r=this;this.promise.then(i=>{if(!r._listeners)return;let a=r._listeners.length;for(;a-- >0;)r._listeners[a](i);r._listeners=null}),this.promise.then=i=>{let a;const s=new Promise(u=>{r.subscribe(u),a=u}).then(i);return s.cancel=function(){r.unsubscribe(a)},s},t(function(a,s,u){r.reason||(r.reason=new hu(a,s,u),n(r.reason))})}throwIfRequested(){if(this.reason)throw this.reason}subscribe(t){if(this.reason){t(this.reason);return}this._listeners?this._listeners.push(t):this._listeners=[t]}unsubscribe(t){if(!this._listeners)return;const n=this._listeners.indexOf(t);n!==-1&&this._listeners.splice(n,1)}toAbortSignal(){const t=new AbortController,n=r=>{t.abort(r)};return this.subscribe(n),t.signal.unsubscribe=()=>this.unsubscribe(n),t.signal}static source(){let t;return{token:new uq(function(i){t=i}),cancel:t}}};function MQ(e){return function(n){return e.apply(null,n)}}function PQ(e){return ue.isObject(e)&&e.isAxiosError===!0}const H_={Continue:100,SwitchingProtocols:101,Processing:102,EarlyHints:103,Ok:200,Created:201,Accepted:202,NonAuthoritativeInformation:203,NoContent:204,ResetContent:205,PartialContent:206,MultiStatus:207,AlreadyReported:208,ImUsed:226,MultipleChoices:300,MovedPermanently:301,Found:302,SeeOther:303,NotModified:304,UseProxy:305,Unused:306,TemporaryRedirect:307,PermanentRedirect:308,BadRequest:400,Unauthorized:401,PaymentRequired:402,Forbidden:403,NotFound:404,MethodNotAllowed:405,NotAcceptable:406,ProxyAuthenticationRequired:407,RequestTimeout:408,Conflict:409,Gone:410,LengthRequired:411,PreconditionFailed:412,PayloadTooLarge:413,UriTooLong:414,UnsupportedMediaType:415,RangeNotSatisfiable:416,ExpectationFailed:417,ImATeapot:418,MisdirectedRequest:421,UnprocessableEntity:422,Locked:423,FailedDependency:424,TooEarly:425,UpgradeRequired:426,PreconditionRequired:428,TooManyRequests:429,RequestHeaderFieldsTooLarge:431,UnavailableForLegalReasons:451,InternalServerError:500,NotImplemented:501,BadGateway:502,ServiceUnavailable:503,GatewayTimeout:504,HttpVersionNotSupported:505,VariantAlsoNegotiates:506,InsufficientStorage:507,LoopDetected:508,NotExtended:510,NetworkAuthenticationRequired:511};Object.entries(H_).forEach(([e,t])=>{H_[t]=e});function cq(e){const t=new Qo(e),n=FF(Qo.prototype.request,t);return ue.extend(n,Qo.prototype,t,{allOwnKeys:!0}),ue.extend(n,t,null,{allOwnKeys:!0}),n.create=function(i){return cq(cs(e,i))},n}const jt=cq(hd);jt.Axios=Qo;jt.CanceledError=hu;jt.CancelToken=TQ;jt.isCancel=tq;jt.VERSION=lq;jt.toFormData=yv;jt.AxiosError=Ze;jt.Cancel=jt.CanceledError;jt.all=function(t){return Promise.all(t)};jt.spread=MQ;jt.isAxiosError=PQ;jt.mergeConfig=cs;jt.AxiosHeaders=Xn;jt.formToJSON=e=>eq(ue.isHTMLForm(e)?new FormData(e):e);jt.getAdapter=sq.getAdapter;jt.HttpStatusCode=H_;jt.default=jt;const{Axios:Q0e,AxiosError:J0e,CanceledError:exe,isCancel:txe,CancelToken:nxe,VERSION:rxe,all:ixe,Cancel:axe,isAxiosError:oxe,spread:sxe,toFormData:lxe,AxiosHeaders:uxe,HttpStatusCode:cxe,formToJSON:fxe,getAdapter:dxe,mergeConfig:pxe}=jt;async function cC(){try{return(await jt.get("/api/list-models")).data}catch(e){throw console.error("Error fetching models:",e),new Error("Failed to fetch models")}}async function fq(e){try{const t=new FormData;return t.append("file",e),(await jt.post("/api/upload-file",t,{headers:{"Content-Type":"multipart/form-data"}})).data}catch(t){throw console.error("Error uploading file:",t),new Error("Failed to upload file")}}async function jQ(e,t,n,r){try{return(await jt.post(`/api/classify?classification_type=${t}&model=${n}&file=${e}&step=${r}`)).data}catch(i){throw console.error("Error classifying file:",i),new Error("Failed to classify file")}}async function NQ(e){try{return(await jt.get("/api/classification-result",{params:{uuid:e}})).data}catch(t){throw console.error("Error fetching classification result:",t),new Error("Failed to fetch classification result")}}async function fC(e){try{return(await jt.get("/api/model-metadata",{params:{model_slug:e}})).data}catch(t){throw console.error("Error fetching model metadata:",t),new Error("Failed to fetch model metadata")}}async function DQ(e,t,n,r,i="",a=1){try{return(await jt.post(`/api/filter?filter_type=${e}&genus=${t}&input_file=${n}&threshold=${r}&filter_species=${i}&step=${a}`)).data}catch(s){throw console.error("Error filtering sequences:",s),new Error("Failed to filter sequences")}}async function kQ(e){try{return(await jt.get("/api/filtering-result",{params:{uuid:e}})).data}catch(t){throw console.error("Error fetching filtering result:",t),new Error("Failed to fetch filtering result")}}const ci="FileUpload",gf="FileUploadDropzone",em="FileUploadTrigger",tm="FileUploadList",Xa="FileUploadItem",nm="FileUploadItemPreview",yf="FileUploadItemMetadata",rm="FileUploadItemProgress",bf="FileUploadItemDelete",xf="FileUploadClear",dC={[ci]:`\`${ci}\` must be used as root component`,[gf]:`\`${gf}\` must be within \`${ci}\``,[em]:`\`${em}\` must be within \`${ci}\``,[tm]:`\`${tm}\` must be within \`${ci}\``,[Xa]:`\`${Xa}\` must be within \`${ci}\``,[nm]:`\`${nm}\` must be within \`${Xa}\``,[yf]:`\`${yf}\` must be within \`${Xa}\``,[rm]:`\`${rm}\` must be within \`${Xa}\``,[bf]:`\`${bf}\` must be within \`${Xa}\``,[xf]:`\`${xf}\` must be within \`${ci}\``},$Q=typeof window<"u"?g.useLayoutEffect:g.useEffect;function md(e){const t=g.useRef(e);return $Q(()=>{t.current=e}),t}function G_(e){const t=g.useRef(null);return t.current===null&&(t.current=e()),t}const dq=g.createContext(void 0);function IQ(e){const t=g.useContext(dq);return e??t??"ltr"}function LQ(e,t,n,r){let a={files:t,dragOver:!1,invalid:r??!1};function s(p,m){switch(m.variant){case"ADD_FILES":{for(const v of m.files)t.set(v,{file:v,progress:0,status:"idle"});if(n){const v=Array.from(t.values()).map(b=>b.file);n(v)}return{...p,files:t}}case"SET_FILES":{const v=new Set(m.files);for(const b of t.keys())v.has(b)||t.delete(b);for(const b of m.files)t.get(b)||t.set(b,{file:b,progress:0,status:"idle"});return{...p,files:t}}case"SET_PROGRESS":{const v=t.get(m.file);return v&&t.set(m.file,{...v,progress:m.progress,status:"uploading"}),{...p,files:t}}case"SET_SUCCESS":{const v=t.get(m.file);return v&&t.set(m.file,{...v,progress:100,status:"success"}),{...p,files:t}}case"SET_ERROR":{const v=t.get(m.file);return v&&t.set(m.file,{...v,error:m.error,status:"error"}),{...p,files:t}}case"REMOVE_FILE":{if(t.delete(m.file),n){const v=Array.from(t.values()).map(b=>b.file);n(v)}return{...p,files:t}}case"SET_DRAG_OVER":return{...p,dragOver:m.dragOver};case"SET_INVALID":return{...p,invalid:m.invalid};case"CLEAR":return t.clear(),n&&n([]),{...p,files:t,invalid:!1};default:return p}}function u(){return a}function c(p){a=s(a,p);for(const m of e)m()}function d(p){return e.add(p),()=>e.delete(p)}return{getState:u,dispatch:c,subscribe:d}}const pC=g.createContext(null);pC.displayName=ci;function wv(e){const t=g.useContext(pC);if(!t)throw new Error(dC[e]);return t}function Jo(e){const t=wv(ci),n=G_(()=>null),r=g.useCallback(()=>{const i=t.getState(),a=n.current;if(a&&a.state===i)return a.value;const s=e(i);return n.current={value:s,state:i},s},[t,e,n]);return g.useSyncExternalStore(t.subscribe,r,r)}const pq=g.createContext(null);function mu(e){const t=g.useContext(pq);if(!t)throw new Error(dC[e]);return t}const hq=g.forwardRef((e,t)=>{const{value:n,defaultValue:r,onValueChange:i,onAccept:a,onFileAccept:s,onFileReject:u,onFileValidate:c,onUpload:d,accept:p,maxFiles:m,maxSize:v,dir:b,label:S,name:w,asChild:x,disabled:C=!1,invalid:O=!1,multiple:A=!1,required:T=!1,children:R,className:M,...P}=e,N=g.useId(),k=g.useId(),H=g.useId(),L=g.useId(),F=IQ(b),z=md(e),Z=G_(()=>new Set).current,Y=G_(()=>new Map).current,I=g.useRef(null),W=n!==void 0,B=g.useMemo(()=>LQ(Z,Y,i,O),[Z,Y,i,O]),Q=g.useMemo(()=>({dropzoneId:k,inputId:N,listId:H,labelId:L,dir:F,disabled:C,inputRef:I}),[k,N,H,L,F,C]);g.useEffect(()=>{W?B.dispatch({variant:"SET_FILES",files:n}):r&&r.length>0&&!B.getState().files.size&&B.dispatch({variant:"SET_FILES",files:r})},[n,r,W,B]);const $=g.useCallback(oe=>{var Se,ce,we,_e,se,Ye,Be,dt,ft,it;if(z.current.disabled)return;let le=[...oe],ve=!1;if(z.current.maxFiles){const xt=B.getState().files.size,Tt=Math.max(0,z.current.maxFiles-xt);if(Tt<le.length){const bn=le.slice(Tt);ve=!0,le=le.slice(0,Tt);for(const Bt of bn){let G=`Maximum ${z.current.maxFiles} files allowed`;if(z.current.onFileValidate){const re=z.current.onFileValidate(Bt);re&&(G=re)}(ce=(Se=z.current).onFileReject)==null||ce.call(Se,Bt,G)}}}const de=[];for(const xt of le){let Tt=!1,bn="";if(z.current.onFileValidate){const Bt=z.current.onFileValidate(xt);if(Bt){bn=Bt,(_e=(we=z.current).onFileReject)==null||_e.call(we,xt,bn),Tt=!0,ve=!0;continue}}if(z.current.accept){const Bt=z.current.accept.split(",").map(fe=>fe.trim()),G=xt.type,re=`.${xt.name.split(".").pop()}`;Bt.some(fe=>fe===G||fe===re||fe.includes("/*")&&G.startsWith(fe.replace("/*","/")))||(bn="File type not accepted",(Ye=(se=z.current).onFileReject)==null||Ye.call(se,xt,bn),Tt=!0,ve=!0)}z.current.maxSize&&xt.size>z.current.maxSize&&(bn="File too large",(dt=(Be=z.current).onFileReject)==null||dt.call(Be,xt,bn),Tt=!0,ve=!0),Tt||de.push(xt)}if(ve&&(B.dispatch({variant:"SET_INVALID",invalid:ve}),setTimeout(()=>{B.dispatch({variant:"SET_INVALID",invalid:!1})},2e3)),de.length>0){if(B.dispatch({variant:"ADD_FILES",files:de}),W&&z.current.onValueChange){const xt=Array.from(B.getState().files.values()).map(Tt=>Tt.file);z.current.onValueChange([...xt])}z.current.onAccept&&z.current.onAccept(de);for(const xt of de)(it=(ft=z.current).onFileAccept)==null||it.call(ft,xt);z.current.onUpload&&requestAnimationFrame(()=>{J(de)})}},[B,W,z]),J=g.useCallback(async oe=>{try{for(const le of oe)B.dispatch({variant:"SET_PROGRESS",file:le,progress:0});if(z.current.onUpload)await z.current.onUpload(oe,{onProgress:(le,ve)=>{B.dispatch({variant:"SET_PROGRESS",file:le,progress:Math.min(Math.max(0,ve),100)})},onSuccess:le=>{B.dispatch({variant:"SET_SUCCESS",file:le})},onError:(le,ve)=>{B.dispatch({variant:"SET_ERROR",file:le,error:ve.message??"Upload failed"})}});else for(const le of oe)B.dispatch({variant:"SET_SUCCESS",file:le})}catch(le){const ve=le instanceof Error?le.message:"Upload failed";for(const de of oe)B.dispatch({variant:"SET_ERROR",file:de,error:ve})}},[B,z.current.onUpload]),q=g.useCallback(oe=>{const le=Array.from(oe.target.files??[]);$(le),oe.target.value=""},[$]),X=x?kr:"div";return E.jsx(dq.Provider,{value:F,children:E.jsx(pC.Provider,{value:B,children:E.jsx(pq.Provider,{value:Q,children:E.jsxs(X,{"data-disabled":C?"":void 0,"data-slot":"file-upload",dir:F,...P,ref:t,className:ke("relative flex flex-col gap-2",M),children:[R,E.jsx("input",{type:"file",id:N,"aria-labelledby":L,"aria-describedby":k,ref:I,tabIndex:-1,accept:p,name:w,disabled:C,multiple:A,required:T,className:"sr-only",onChange:q}),E.jsx("span",{id:L,className:"sr-only",children:S??"File upload"})]})})})})});hq.displayName=ci;const hC=g.forwardRef((e,t)=>{const{asChild:n,className:r,...i}=e,a=mu(gf),s=wv(gf),u=Jo(O=>O.dragOver),c=Jo(O=>O.invalid),d=md(i),p=g.useCallback(O=>{var R,M,P;if((M=(R=d.current)==null?void 0:R.onClick)==null||M.call(R,O),O.defaultPrevented)return;const A=O.target;A instanceof HTMLElement&&A.closest('[data-slot="file-upload-trigger"]')||(P=a.inputRef.current)==null||P.click()},[a.inputRef,d]),m=g.useCallback(O=>{var A,T;(T=(A=d.current)==null?void 0:A.onDragOver)==null||T.call(A,O),!O.defaultPrevented&&(O.preventDefault(),s.dispatch({variant:"SET_DRAG_OVER",dragOver:!0}))},[s,d.current.onDragOver]),v=g.useCallback(O=>{var A,T;(T=(A=d.current)==null?void 0:A.onDragEnter)==null||T.call(A,O),!O.defaultPrevented&&(O.preventDefault(),s.dispatch({variant:"SET_DRAG_OVER",dragOver:!0}))},[s,d.current.onDragEnter]),b=g.useCallback(O=>{var T,R;if((R=(T=d.current)==null?void 0:T.onDragLeave)==null||R.call(T,O),O.defaultPrevented)return;const A=O.relatedTarget;A&&A instanceof Node&&O.currentTarget.contains(A)||(O.preventDefault(),s.dispatch({variant:"SET_DRAG_OVER",dragOver:!1}))},[s,d.current.onDragLeave]),S=g.useCallback(O=>{var M,P;if((P=(M=d.current)==null?void 0:M.onDrop)==null||P.call(M,O),O.defaultPrevented)return;O.preventDefault(),s.dispatch({variant:"SET_DRAG_OVER",dragOver:!1});const A=Array.from(O.dataTransfer.files),T=a.inputRef.current;if(!T)return;const R=new DataTransfer;for(const N of A)R.items.add(N);T.files=R.files,T.dispatchEvent(new Event("change",{bubbles:!0}))},[s,a.inputRef,d.current.onDrop]),w=g.useCallback(O=>{var P,N,k;if((N=(P=d.current)==null?void 0:P.onPaste)==null||N.call(P,O),O.defaultPrevented)return;O.preventDefault(),s.dispatch({variant:"SET_DRAG_OVER",dragOver:!1});const A=(k=O.clipboardData)==null?void 0:k.items;if(!A)return;const T=[];for(let H=0;H<A.length;H++){const L=A[H];if((L==null?void 0:L.kind)==="file"){const F=L.getAsFile();F&&T.push(F)}}if(T.length===0)return;const R=a.inputRef.current;if(!R)return;const M=new DataTransfer;for(const H of T)M.items.add(H);R.files=M.files,R.dispatchEvent(new Event("change",{bubbles:!0}))},[s,a.inputRef,d]),x=g.useCallback(O=>{var A,T,R;(T=(A=d.current)==null?void 0:A.onKeyDown)==null||T.call(A,O),!O.defaultPrevented&&(O.key==="Enter"||O.key===" ")&&(O.preventDefault(),(R=a.inputRef.current)==null||R.click())},[a.inputRef,d.current.onKeyDown]),C=n?kr:"div";return E.jsx(C,{role:"region",id:a.dropzoneId,"aria-controls":`${a.inputId} ${a.listId}`,"aria-disabled":a.disabled,"aria-invalid":c,"data-disabled":a.disabled?"":void 0,"data-dragging":u?"":void 0,"data-invalid":c?"":void 0,"data-slot":"file-upload-dropzone",dir:a.dir,tabIndex:a.disabled?void 0:0,...i,ref:t,className:ke("relative flex select-none flex-col items-center justify-center gap-2 rounded-lg border-2 border-dashed p-6 outline-none transition-colors hover:bg-accent/30 focus-visible:border-ring/50 data-[disabled]:pointer-events-none data-[dragging]:border-primary data-[invalid]:border-destructive data-[invalid]:ring-destructive/20",r),onClick:p,onDragEnter:v,onDragLeave:b,onDragOver:m,onDrop:S,onKeyDown:x,onPaste:w})});hC.displayName=gf;const mC=g.forwardRef((e,t)=>{const{asChild:n,...r}=e,i=mu(em),a=md(r),s=g.useCallback(c=>{var d,p,m;(p=(d=a.current)==null?void 0:d.onClick)==null||p.call(d,c),!c.defaultPrevented&&((m=i.inputRef.current)==null||m.click())},[i.inputRef,a.current]),u=n?kr:"button";return E.jsx(u,{type:"button","aria-controls":i.inputId,"data-disabled":i.disabled?"":void 0,"data-slot":"file-upload-trigger",...r,ref:t,disabled:i.disabled,onClick:s})});mC.displayName=em;const vC=g.forwardRef((e,t)=>{const{className:n,orientation:r="vertical",asChild:i,forceMount:a,...s}=e,u=mu(tm),c=a||Jo(p=>p.files.size>0);if(!c)return null;const d=i?kr:"div";return E.jsx(d,{role:"list",id:u.listId,"aria-orientation":r,"data-orientation":r,"data-slot":"file-upload-list","data-state":c?"active":"inactive",dir:u.dir,...s,ref:t,className:ke("data-[state=inactive]:fade-out-0 data-[state=active]:fade-in-0 data-[state=inactive]:slide-out-to-top-2 data-[state=active]:slide-in-from-top-2 flex flex-col gap-2 data-[state=active]:animate-in data-[state=inactive]:animate-out",r==="horizontal"&&"flex-row overflow-x-auto p-1.5",n)})});vC.displayName=tm;const mq=g.createContext(null);function Sv(e){const t=g.useContext(mq);if(!t)throw new Error(dC[e]);return t}const gC=g.forwardRef((e,t)=>{const{value:n,asChild:r,className:i,...a}=e,s=g.useId(),u=`${s}-status`,c=`${s}-name`,d=`${s}-size`,p=`${s}-message`,m=mu(Xa),v=Jo(O=>O.files.get(n)),b=Jo(O=>O.files.size),S=Jo(O=>Array.from(O.files.keys()).indexOf(n)+1),w=g.useMemo(()=>({id:s,fileState:v,nameId:c,sizeId:d,statusId:u,messageId:p}),[s,v,u,c,d,p]);if(!v)return null;const x=v.error?`Error: ${v.error}`:v.status==="uploading"?`Uploading: ${v.progress}% complete`:v.status==="success"?"Upload complete":"Ready to upload",C=r?kr:"div";return E.jsx(mq.Provider,{value:w,children:E.jsxs(C,{role:"listitem",id:s,"aria-setsize":b,"aria-posinset":S,"aria-describedby":`${c} ${d} ${u} ${v.error?p:""}`,"aria-labelledby":c,"data-slot":"file-upload-item",dir:m.dir,...a,ref:t,className:ke("relative flex items-center gap-2.5 rounded-md border p-3",i),children:[e.children,E.jsx("span",{id:u,className:"sr-only",children:x})]})})});gC.displayName=Xa;function BQ(e){if(e===0)return"0 B";const t=["B","KB","MB","GB","TB"],n=Math.floor(Math.log(e)/Math.log(1024));return`${(e/1024**n).toFixed(n?1:0)} ${t[n]}`}function FQ(e){var r;const t=e.type,n=((r=e.name.split(".").pop())==null?void 0:r.toLowerCase())??"";return t.startsWith("video/")?E.jsx(g8,{}):t.startsWith("audio/")?E.jsx(u8,{}):t.startsWith("text/")||["txt","md","rtf","pdf"].includes(n)?E.jsx(m8,{}):["html","css","js","jsx","ts","tsx","json","xml","php","py","rb","java","c","cpp","cs"].includes(n)?E.jsx(f8,{}):["zip","rar","7z","tar","gz","bz2"].includes(n)?E.jsx(s8,{}):["exe","msi","app","apk","deb","rpm"].includes(n)||t.startsWith("application/")?E.jsx(p8,{}):E.jsx(b8,{})}const yC=g.forwardRef((e,t)=>{var p;const{render:n,asChild:r,children:i,className:a,...s}=e,u=Sv(nm),c=g.useCallback(m=>{var v;return n?n(m):(v=u.fileState)!=null&&v.file.type.startsWith("image/")?E.jsx("img",{src:URL.createObjectURL(m),alt:m.name,className:"size-full object-cover",onLoad:b=>{b.target instanceof HTMLImageElement&&URL.revokeObjectURL(b.target.src)}}):FQ(m)},[n,(p=u.fileState)==null?void 0:p.file.type]);if(!u.fileState)return null;const d=r?kr:"div";return E.jsxs(d,{"aria-labelledby":u.nameId,"data-slot":"file-upload-preview",...s,ref:t,className:ke("relative flex size-10 shrink-0 items-center justify-center overflow-hidden rounded border bg-accent/50 [&>svg]:size-10",a),children:[c(u.fileState.file),i]})});yC.displayName=nm;const bC=g.forwardRef((e,t)=>{const{asChild:n,size:r="default",children:i,className:a,...s}=e,u=mu(yf),c=Sv(yf);if(!c.fileState)return null;const d=n?kr:"div";return E.jsx(d,{"data-slot":"file-upload-metadata",dir:u.dir,...s,ref:t,className:ke("flex min-w-0 flex-1 flex-col",a),children:i??E.jsxs(E.Fragment,{children:[E.jsx("span",{id:c.nameId,className:ke("truncate font-medium text-sm",r==="sm"&&"font-normal text-[13px] leading-snug"),children:c.fileState.file.name}),E.jsx("span",{id:c.sizeId,className:ke("truncate text-muted-foreground text-xs",r==="sm"&&"text-[11px]"),children:BQ(c.fileState.file.size)}),c.fileState.error&&E.jsx("span",{id:c.messageId,className:"text-destructive text-xs",children:c.fileState.error})]})})});bC.displayName=yf;const xC=g.forwardRef((e,t)=>{const{variant:n="linear",size:r=40,asChild:i,forceMount:a,className:s,...u}=e,c=Sv(rm);if(!c.fileState||!(a||c.fileState.progress!==100))return null;const p=i?kr:"div";switch(n){case"circular":{const m=2*Math.PI*((r-4)/2),v=m-c.fileState.progress/100*m;return E.jsx(p,{role:"progressbar","aria-valuemin":0,"aria-valuemax":100,"aria-valuenow":c.fileState.progress,"aria-valuetext":`${c.fileState.progress}%`,"aria-labelledby":c.nameId,"data-slot":"file-upload-progress",...u,ref:t,className:ke("-translate-x-1/2 -translate-y-1/2 absolute top-1/2 left-1/2",s),children:E.jsxs("svg",{className:"rotate-[-90deg] transform",width:r,height:r,viewBox:`0 0 ${r} ${r}`,fill:"none",stroke:"currentColor",children:[E.jsx("circle",{className:"text-primary/20",strokeWidth:"2",cx:r/2,cy:r/2,r:(r-4)/2}),E.jsx("circle",{className:"text-primary transition-[stroke-dashoffset] duration-300 ease-linear",strokeWidth:"2",strokeLinecap:"round",strokeDasharray:m,strokeDashoffset:v,cx:r/2,cy:r/2,r:(r-4)/2})]})})}case"fill":{const m=c.fileState.progress,v=100-m;return E.jsx(p,{role:"progressbar","aria-valuemin":0,"aria-valuemax":100,"aria-valuenow":m,"aria-valuetext":`${m}%`,"aria-labelledby":c.nameId,"data-slot":"file-upload-progress",...u,ref:t,className:ke("absolute inset-0 bg-primary/50 transition-[clip-path] duration-300 ease-linear",s),style:{clipPath:`inset(${v}% 0% 0% 0%)`}})}default:return E.jsx(p,{role:"progressbar","aria-valuemin":0,"aria-valuemax":100,"aria-valuenow":c.fileState.progress,"aria-valuetext":`${c.fileState.progress}%`,"aria-labelledby":c.nameId,"data-slot":"file-upload-progress",...u,ref:t,className:ke("relative h-1.5 w-full overflow-hidden rounded-full bg-primary/20",s),children:E.jsx("div",{className:"h-full w-full flex-1 bg-primary transition-transform duration-300 ease-linear",style:{transform:`translateX(-${100-c.fileState.progress}%)`}})})}});xC.displayName=rm;const wC=g.forwardRef((e,t)=>{var d;const{asChild:n,...r}=e,i=wv(bf),a=Sv(bf),s=md(r),u=g.useCallback(p=>{var m,v;(v=(m=s.current)==null?void 0:m.onClick)==null||v.call(m,p),!(!a.fileState||p.defaultPrevented)&&i.dispatch({variant:"REMOVE_FILE",file:a.fileState.file})},[i,a.fileState,(d=s.current)==null?void 0:d.onClick]);if(!a.fileState)return null;const c=n?kr:"button";return E.jsx(c,{type:"button","aria-controls":a.id,"aria-describedby":a.nameId,"data-slot":"file-upload-item-delete",...r,ref:t,onClick:u})});wC.displayName=bf;const qQ=g.forwardRef((e,t)=>{const{asChild:n,forceMount:r,disabled:i,...a}=e,s=mu(xf),u=wv(xf),c=md(a),d=i||s.disabled,p=g.useCallback(b=>{var S,w;(w=(S=c.current)==null?void 0:S.onClick)==null||w.call(S,b),!b.defaultPrevented&&u.dispatch({variant:"CLEAR"})},[u,c]);if(!(r||Jo(b=>b.files.size>0)))return null;const v=n?kr:"button";return E.jsx(v,{type:"button","aria-controls":s.listId,"data-slot":"file-upload-clear","data-disabled":d?"":void 0,...a,ref:t,disabled:d,onClick:p})});qQ.displayName=xf;const vq=hq,zQ=Ln.object({input_file:Ln.string().min(1,"Please upload a file"),classification_type:Ln.enum(["Genus","Species","MLST"]),model:Ln.string(),sparse_sampling:Ln.boolean(),sparse_sampling_step:Ln.number().min(1).max(500)});function UQ(){const[e,t]=g.useState({});g.useEffect(()=>{cC().then(d=>{t(d)}).catch(d=>{console.error("Error fetching models:",d)})},[]);const n=kE(),r=cB({resolver:dB(zQ),defaultValues:{classification_type:"Species",model:void 0,sparse_sampling:!1,sparse_sampling_step:1}}),[i,a]=g.useState([]),s=g.useCallback(async(d,{onSuccess:p,onError:m})=>{const v=d[0];try{const{filename:b}=await fq(v);p(v),console.log("File uploaded successfully:",b),r.setValue("input_file",b)}catch(b){m(v,b),console.error("Error uploading file:",b)}},[r]),u=g.useCallback((d,p)=>{console.log(p,{description:`"${d.name.length>20?`${d.name.slice(0,20)}...`:d.name}" has been rejected`})},[]);function c(d){jQ(d.input_file,d.classification_type,d.model,d.sparse_sampling_step).then(p=>{console.log("Classification response:",p),n(`/result/${p.uuid}`)}).catch(p=>{console.error("Error during classification:",p)})}return E.jsx(PB,{...r,children:E.jsxs("form",{onSubmit:r.handleSubmit(c),className:"space-y-6",children:[E.jsx(sr,{control:r.control,name:"input_file",render:({field:d})=>E.jsxs(lr,{children:[E.jsx(ur,{children:"Input File"}),E.jsx(cr,{children:E.jsxs(vq,{value:i,onValueChange:a,onUpload:s,onFileReject:u,maxFiles:1,multiple:!0,children:[i.length===0&&E.jsxs(hC,{children:[E.jsxs("div",{className:"flex flex-col items-center gap-1 text-center",children:[E.jsx("div",{className:"flex items-center justify-center rounded-full border p-2.5",children:E.jsx(xL,{className:"size-6 text-muted-foreground"})}),E.jsx("p",{className:"font-medium text-sm",children:"Drag & drop file here"}),E.jsx("p",{className:"text-muted-foreground text-xs",children:"Or click to browse"})]}),E.jsx(mC,{asChild:!0,children:E.jsx(Pr,{variant:"outline",size:"sm",className:"mt-2 w-fit",children:"Browse files"})})]}),E.jsx(vC,{children:i.map((p,m)=>E.jsxs(gC,{value:p,className:"flex-col",children:[E.jsxs("div",{className:"flex w-full items-center gap-2",children:[E.jsx(yC,{}),E.jsx(bC,{}),E.jsx(wC,{asChild:!0,children:E.jsx(Pr,{variant:"ghost",size:"icon",className:"size-7",onClick:()=>{a(v=>v.filter((b,S)=>S!==m)),r.setValue("input_file","")},children:E.jsx(wL,{})})})]}),E.jsx(xC,{})]},m))})]})}),E.jsx(fr,{children:"Upload the file you would like to classify."}),E.jsx(Vr,{})]})}),E.jsx(sr,{control:r.control,name:"classification_type",render:({field:d})=>E.jsxs(lr,{children:[E.jsx(ur,{children:"Classification Type"}),E.jsx(cr,{children:E.jsxs(Kh,{defaultValue:"Species",onValueChange:p=>d.onChange(p),children:[E.jsx(Zh,{className:"w-full",children:E.jsx(Yh,{})}),E.jsxs(Xh,{children:[E.jsx(Go,{value:"Genus",children:"Genus"}),E.jsx(Go,{value:"Species",children:"Species"}),E.jsx(Go,{value:"MLST",children:"MLST"})]})]})}),E.jsx(fr,{children:"Select the type of classification you would like to perform."}),E.jsx(Vr,{})]})}),E.jsx(sr,{control:r.control,name:"model",render:({field:d})=>{var p;return E.jsxs(lr,{children:[E.jsx(ur,{children:"Model"}),E.jsx(cr,{children:E.jsxs(Kh,{onValueChange:d.onChange,children:[E.jsx(Zh,{className:"w-full",children:E.jsx(Yh,{})}),E.jsx(Xh,{children:(p=e[r.watch("classification_type")])==null?void 0:p.map(m=>E.jsx(Go,{value:m,children:m},m))})]})}),E.jsx(fr,{children:"Select the model you would like to use for classification."}),E.jsx(Vr,{})]})}}),E.jsx(sr,{control:r.control,name:"sparse_sampling",render:({field:d})=>E.jsxs(lr,{className:"flex flex-row items-center justify-between pr-2",children:[E.jsxs("div",{className:"space-y-0.5",children:[E.jsx(ur,{children:"Sparse Sampling"}),E.jsx(fr,{children:"Enable sparse sampling for classification."})]}),E.jsx(cr,{children:E.jsx(TB,{checked:d.value,onCheckedChange:d.onChange})})]})}),r.watch("sparse_sampling")&&E.jsx(sr,{control:r.control,name:"sparse_sampling_step",render:({field:d})=>E.jsxs(lr,{children:[E.jsx(ur,{children:"Sparse Sampling Step"}),E.jsx(cr,{children:E.jsx(T_,{type:"number",value:d.value,onChange:p=>d.onChange(Number(p.target.value))})}),E.jsx(fr,{children:"Set the step size for sparse sampling."}),E.jsx(Vr,{})]})}),E.jsx(Pr,{type:"submit",children:"Classify"})]})})}function VQ(e){const t=HQ(e),n=g.forwardRef((r,i)=>{const{children:a,...s}=r,u=g.Children.toArray(a),c=u.find(WQ);if(c){const d=c.props.children,p=u.map(m=>m===c?g.Children.count(d)>1?g.Children.only(null):g.isValidElement(d)?d.props.children:null:m);return E.jsx(t,{...s,ref:i,children:g.isValidElement(d)?g.cloneElement(d,void 0,p):null})}return E.jsx(t,{...s,ref:i,children:a})});return n.displayName=`${e}.Slot`,n}function HQ(e){const t=g.forwardRef((n,r)=>{const{children:i,...a}=n,s=g.isValidElement(i)?YQ(i):void 0,u=Re(s,r);if(g.isValidElement(i)){const c=KQ(a,i.props);return i.type!==g.Fragment&&(c.ref=u),g.cloneElement(i,c)}return g.Children.count(i)>1?g.Children.only(null):null});return t.displayName=`${e}.SlotClone`,t}var GQ=Symbol("radix.slottable");function WQ(e){return g.isValidElement(e)&&typeof e.type=="function"&&"__radixId"in e.type&&e.type.__radixId===GQ}function KQ(e,t){const n={...t};for(const r in t){const i=e[r],a=t[r];/^on[A-Z]/.test(r)?i&&a?n[r]=(...u)=>{const c=a(...u);return i(...u),c}:i&&(n[r]=i):r==="style"?n[r]={...i,...a}:r==="className"&&(n[r]=[i,a].filter(Boolean).join(" "))}return{...e,...n}}function YQ(e){var r,i;let t=(r=Object.getOwnPropertyDescriptor(e.props,"ref"))==null?void 0:r.get,n=t&&"isReactWarning"in t&&t.isReactWarning;return n?e.ref:(t=(i=Object.getOwnPropertyDescriptor(e,"ref"))==null?void 0:i.get,n=t&&"isReactWarning"in t&&t.isReactWarning,n?e.props.ref:e.props.ref||e.ref)}var ZQ=["a","button","div","form","h2","h3","img","input","label","li","nav","ol","p","select","span","svg","ul"],XQ=ZQ.reduce((e,t)=>{const n=VQ(`Primitive.${t}`),r=g.forwardRef((i,a)=>{const{asChild:s,...u}=i,c=s?n:t;return typeof window<"u"&&(window[Symbol.for("radix-ui")]=!0),E.jsx(c,{...u,ref:a})});return r.displayName=`Primitive.${t}`,{...e,[t]:r}},{}),QQ="Separator",XP="horizontal",JQ=["horizontal","vertical"],gq=g.forwardRef((e,t)=>{const{decorative:n,orientation:r=XP,...i}=e,a=eJ(r)?r:XP,u=n?{role:"none"}:{"aria-orientation":a==="vertical"?a:void 0,role:"separator"};return E.jsx(XQ.div,{"data-orientation":a,...u,...i,ref:t})});gq.displayName=QQ;function eJ(e){return JQ.includes(e)}var tJ=gq;function Mr({className:e,orientation:t="horizontal",decorative:n=!0,...r}){return E.jsx(tJ,{"data-slot":"separator-root",decorative:n,orientation:t,className:ke("bg-border shrink-0 data-[orientation=horizontal]:h-px data-[orientation=horizontal]:w-full data-[orientation=vertical]:h-full data-[orientation=vertical]:w-px",e),...r})}function nJ(){return E.jsx("main",{className:"flex-1 flex flex-col items-center justify-center p-4",children:E.jsxs("div",{className:"w-1/2",children:[E.jsx("h1",{className:"text-xl font-bold",children:"Classify your data"}),E.jsx("p",{children:"Upload your data and select a model to classify it."}),E.jsx(Mr,{className:"my-4"}),E.jsx(UQ,{})]})})}function W_(e){const t=rJ(e),n=g.forwardRef((r,i)=>{const{children:a,...s}=r,u=g.Children.toArray(a),c=u.find(aJ);if(c){const d=c.props.children,p=u.map(m=>m===c?g.Children.count(d)>1?g.Children.only(null):g.isValidElement(d)?d.props.children:null:m);return E.jsx(t,{...s,ref:i,children:g.isValidElement(d)?g.cloneElement(d,void 0,p):null})}return E.jsx(t,{...s,ref:i,children:a})});return n.displayName=`${e}.Slot`,n}function rJ(e){const t=g.forwardRef((n,r)=>{const{children:i,...a}=n,s=g.isValidElement(i)?sJ(i):void 0,u=Re(s,r);if(g.isValidElement(i)){const c=oJ(a,i.props);return i.type!==g.Fragment&&(c.ref=u),g.cloneElement(i,c)}return g.Children.count(i)>1?g.Children.only(null):null});return t.displayName=`${e}.SlotClone`,t}var iJ=Symbol("radix.slottable");function aJ(e){return g.isValidElement(e)&&typeof e.type=="function"&&"__radixId"in e.type&&e.type.__radixId===iJ}function oJ(e,t){const n={...t};for(const r in t){const i=e[r],a=t[r];/^on[A-Z]/.test(r)?i&&a?n[r]=(...u)=>{const c=a(...u);return i(...u),c}:i&&(n[r]=i):r==="style"?n[r]={...i,...a}:r==="className"&&(n[r]=[i,a].filter(Boolean).join(" "))}return{...e,...n}}function sJ(e){var r,i;let t=(r=Object.getOwnPropertyDescriptor(e.props,"ref"))==null?void 0:r.get,n=t&&"isReactWarning"in t&&t.isReactWarning;return n?e.ref:(t=(i=Object.getOwnPropertyDescriptor(e,"ref"))==null?void 0:i.get,n=t&&"isReactWarning"in t&&t.isReactWarning,n?e.props.ref:e.props.ref||e.ref)}function lJ(e){const t=e+"CollectionProvider",[n,r]=jn(t),[i,a]=n(t,{collectionRef:{current:null},itemMap:new Map}),s=w=>{const{scope:x,children:C}=w,O=V.useRef(null),A=V.useRef(new Map).current;return E.jsx(i,{scope:x,itemMap:A,collectionRef:O,children:C})};s.displayName=t;const u=e+"CollectionSlot",c=W_(u),d=V.forwardRef((w,x)=>{const{scope:C,children:O}=w,A=a(u,C),T=Re(x,A.collectionRef);return E.jsx(c,{ref:T,children:O})});d.displayName=u;const p=e+"CollectionItemSlot",m="data-radix-collection-item",v=W_(p),b=V.forwardRef((w,x)=>{const{scope:C,children:O,...A}=w,T=V.useRef(null),R=Re(x,T),M=a(p,C);return V.useEffect(()=>(M.itemMap.set(T,{ref:T,...A}),()=>void M.itemMap.delete(T))),E.jsx(v,{[m]:"",ref:R,children:O})});b.displayName=p;function S(w){const x=a(e+"CollectionConsumer",w);return V.useCallback(()=>{const O=x.collectionRef.current;if(!O)return[];const A=Array.from(O.querySelectorAll(`[${m}]`));return Array.from(x.itemMap.values()).sort((M,P)=>A.indexOf(M.ref.current)-A.indexOf(P.ref.current))},[x.collectionRef,x.itemMap])}return[{Provider:s,Slot:d,ItemSlot:b},S,r]}var uJ=["a","button","div","form","h2","h3","img","input","label","li","nav","ol","p","select","span","svg","ul"],yq=uJ.reduce((e,t)=>{const n=W_(`Primitive.${t}`),r=g.forwardRef((i,a)=>{const{asChild:s,...u}=i,c=s?n:t;return typeof window<"u"&&(window[Symbol.for("radix-ui")]=!0),E.jsx(c,{...u,ref:a})});return r.displayName=`Primitive.${t}`,{...e,[t]:r}},{}),Qb="rovingFocusGroup.onEntryFocus",cJ={bubbles:!1,cancelable:!0},vd="RovingFocusGroup",[K_,bq,fJ]=lJ(vd),[dJ,_v]=jn(vd,[fJ]),[pJ,hJ]=dJ(vd),xq=g.forwardRef((e,t)=>E.jsx(K_.Provider,{scope:e.__scopeRovingFocusGroup,children:E.jsx(K_.Slot,{scope:e.__scopeRovingFocusGroup,children:E.jsx(mJ,{...e,ref:t})})}));xq.displayName=vd;var mJ=g.forwardRef((e,t)=>{const{__scopeRovingFocusGroup:n,orientation:r,loop:i=!1,dir:a,currentTabStopId:s,defaultCurrentTabStopId:u,onCurrentTabStopIdChange:c,onEntryFocus:d,preventScrollOnEntryFocus:p=!1,...m}=e,v=g.useRef(null),b=Re(t,v),S=iu(a),[w,x]=Xr({prop:s,defaultProp:u??null,onChange:c,caller:vd}),[C,O]=g.useState(!1),A=$t(d),T=bq(n),R=g.useRef(!1),[M,P]=g.useState(0);return g.useEffect(()=>{const N=v.current;if(N)return N.addEventListener(Qb,A),()=>N.removeEventListener(Qb,A)},[A]),E.jsx(pJ,{scope:n,orientation:r,dir:S,loop:i,currentTabStopId:w,onItemFocus:g.useCallback(N=>x(N),[x]),onItemShiftTab:g.useCallback(()=>O(!0),[]),onFocusableItemAdd:g.useCallback(()=>P(N=>N+1),[]),onFocusableItemRemove:g.useCallback(()=>P(N=>N-1),[]),children:E.jsx(yq.div,{tabIndex:C||M===0?-1:0,"data-orientation":r,...m,ref:b,style:{outline:"none",...e.style},onMouseDown:be(e.onMouseDown,()=>{R.current=!0}),onFocus:be(e.onFocus,N=>{const k=!R.current;if(N.target===N.currentTarget&&k&&!C){const H=new CustomEvent(Qb,cJ);if(N.currentTarget.dispatchEvent(H),!H.defaultPrevented){const L=T().filter(I=>I.focusable),F=L.find(I=>I.active),z=L.find(I=>I.id===w),Y=[F,z,...L].filter(Boolean).map(I=>I.ref.current);_q(Y,p)}}R.current=!1}),onBlur:be(e.onBlur,()=>O(!1))})})}),wq="RovingFocusGroupItem",Sq=g.forwardRef((e,t)=>{const{__scopeRovingFocusGroup:n,focusable:r=!0,active:i=!1,tabStopId:a,children:s,...u}=e,c=ln(),d=a||c,p=hJ(wq,n),m=p.currentTabStopId===d,v=bq(n),{onFocusableItemAdd:b,onFocusableItemRemove:S,currentTabStopId:w}=p;return g.useEffect(()=>{if(r)return b(),()=>S()},[r,b,S]),E.jsx(K_.ItemSlot,{scope:n,id:d,focusable:r,active:i,children:E.jsx(yq.span,{tabIndex:m?0:-1,"data-orientation":p.orientation,...u,ref:t,onMouseDown:be(e.onMouseDown,x=>{r?p.onItemFocus(d):x.preventDefault()}),onFocus:be(e.onFocus,()=>p.onItemFocus(d)),onKeyDown:be(e.onKeyDown,x=>{if(x.key==="Tab"&&x.shiftKey){p.onItemShiftTab();return}if(x.target!==x.currentTarget)return;const C=yJ(x,p.orientation,p.dir);if(C!==void 0){if(x.metaKey||x.ctrlKey||x.altKey||x.shiftKey)return;x.preventDefault();let A=v().filter(T=>T.focusable).map(T=>T.ref.current);if(C==="last")A.reverse();else if(C==="prev"||C==="next"){C==="prev"&&A.reverse();const T=A.indexOf(x.currentTarget);A=p.loop?bJ(A,T+1):A.slice(T+1)}setTimeout(()=>_q(A))}}),children:typeof s=="function"?s({isCurrentTabStop:m,hasTabStop:w!=null}):s})})});Sq.displayName=wq;var vJ={ArrowLeft:"prev",ArrowUp:"prev",ArrowRight:"next",ArrowDown:"next",PageUp:"first",Home:"first",PageDown:"last",End:"last"};function gJ(e,t){return t!=="rtl"?e:e==="ArrowLeft"?"ArrowRight":e==="ArrowRight"?"ArrowLeft":e}function yJ(e,t,n){const r=gJ(e.key,n);if(!(t==="vertical"&&["ArrowLeft","ArrowRight"].includes(r))&&!(t==="horizontal"&&["ArrowUp","ArrowDown"].includes(r)))return vJ[r]}function _q(e,t=!1){const n=document.activeElement;for(const r of e)if(r===n||(r.focus({preventScroll:t}),document.activeElement!==n))return}function bJ(e,t){return e.map((n,r)=>e[(t+r)%e.length])}var Eq=xq,Cq=Sq;function xJ(e){const t=wJ(e),n=g.forwardRef((r,i)=>{const{children:a,...s}=r,u=g.Children.toArray(a),c=u.find(_J);if(c){const d=c.props.children,p=u.map(m=>m===c?g.Children.count(d)>1?g.Children.only(null):g.isValidElement(d)?d.props.children:null:m);return E.jsx(t,{...s,ref:i,children:g.isValidElement(d)?g.cloneElement(d,void 0,p):null})}return E.jsx(t,{...s,ref:i,children:a})});return n.displayName=`${e}.Slot`,n}function wJ(e){const t=g.forwardRef((n,r)=>{const{children:i,...a}=n,s=g.isValidElement(i)?CJ(i):void 0,u=Re(s,r);if(g.isValidElement(i)){const c=EJ(a,i.props);return i.type!==g.Fragment&&(c.ref=u),g.cloneElement(i,c)}return g.Children.count(i)>1?g.Children.only(null):null});return t.displayName=`${e}.SlotClone`,t}var SJ=Symbol("radix.slottable");function _J(e){return g.isValidElement(e)&&typeof e.type=="function"&&"__radixId"in e.type&&e.type.__radixId===SJ}function EJ(e,t){const n={...t};for(const r in t){const i=e[r],a=t[r];/^on[A-Z]/.test(r)?i&&a?n[r]=(...u)=>{const c=a(...u);return i(...u),c}:i&&(n[r]=i):r==="style"?n[r]={...i,...a}:r==="className"&&(n[r]=[i,a].filter(Boolean).join(" "))}return{...e,...n}}function CJ(e){var r,i;let t=(r=Object.getOwnPropertyDescriptor(e.props,"ref"))==null?void 0:r.get,n=t&&"isReactWarning"in t&&t.isReactWarning;return n?e.ref:(t=(i=Object.getOwnPropertyDescriptor(e,"ref"))==null?void 0:i.get,n=t&&"isReactWarning"in t&&t.isReactWarning,n?e.props.ref:e.props.ref||e.ref)}var OJ=["a","button","div","form","h2","h3","img","input","label","li","nav","ol","p","select","span","svg","ul"],Ev=OJ.reduce((e,t)=>{const n=xJ(`Primitive.${t}`),r=g.forwardRef((i,a)=>{const{asChild:s,...u}=i,c=s?n:t;return typeof window<"u"&&(window[Symbol.for("radix-ui")]=!0),E.jsx(c,{...u,ref:a})});return r.displayName=`Primitive.${t}`,{...e,[t]:r}},{}),Cv="Tabs",[AJ,hxe]=jn(Cv,[_v]),Oq=_v(),[RJ,SC]=AJ(Cv),Aq=g.forwardRef((e,t)=>{const{__scopeTabs:n,value:r,onValueChange:i,defaultValue:a,orientation:s="horizontal",dir:u,activationMode:c="automatic",...d}=e,p=iu(u),[m,v]=Xr({prop:r,onChange:i,defaultProp:a??"",caller:Cv});return E.jsx(RJ,{scope:n,baseId:ln(),value:m,onValueChange:v,orientation:s,dir:p,activationMode:c,children:E.jsx(Ev.div,{dir:p,"data-orientation":s,...d,ref:t})})});Aq.displayName=Cv;var Rq="TabsList",Tq=g.forwardRef((e,t)=>{const{__scopeTabs:n,loop:r=!0,...i}=e,a=SC(Rq,n),s=Oq(n);return E.jsx(Eq,{asChild:!0,...s,orientation:a.orientation,dir:a.dir,loop:r,children:E.jsx(Ev.div,{role:"tablist","aria-orientation":a.orientation,...i,ref:t})})});Tq.displayName=Rq;var Mq="TabsTrigger",Pq=g.forwardRef((e,t)=>{const{__scopeTabs:n,value:r,disabled:i=!1,...a}=e,s=SC(Mq,n),u=Oq(n),c=Nq(s.baseId,r),d=Dq(s.baseId,r),p=r===s.value;return E.jsx(Cq,{asChild:!0,...u,focusable:!i,active:p,children:E.jsx(Ev.button,{type:"button",role:"tab","aria-selected":p,"aria-controls":d,"data-state":p?"active":"inactive","data-disabled":i?"":void 0,disabled:i,id:c,...a,ref:t,onMouseDown:be(e.onMouseDown,m=>{!i&&m.button===0&&m.ctrlKey===!1?s.onValueChange(r):m.preventDefault()}),onKeyDown:be(e.onKeyDown,m=>{[" ","Enter"].includes(m.key)&&s.onValueChange(r)}),onFocus:be(e.onFocus,()=>{const m=s.activationMode!=="manual";!p&&!i&&m&&s.onValueChange(r)})})})});Pq.displayName=Mq;var jq="TabsContent",TJ=g.forwardRef((e,t)=>{const{__scopeTabs:n,value:r,forceMount:i,children:a,...s}=e,u=SC(jq,n),c=Nq(u.baseId,r),d=Dq(u.baseId,r),p=r===u.value,m=g.useRef(p);return g.useEffect(()=>{const v=requestAnimationFrame(()=>m.current=!1);return()=>cancelAnimationFrame(v)},[]),E.jsx(zn,{present:i||p,children:({present:v})=>E.jsx(Ev.div,{"data-state":p?"active":"inactive","data-orientation":u.orientation,role:"tabpanel","aria-labelledby":c,hidden:!v,id:d,tabIndex:0,...s,ref:t,style:{...e.style,animationDuration:m.current?"0s":void 0},children:v&&a})})});TJ.displayName=jq;function Nq(e,t){return`${e}-trigger-${t}`}function Dq(e,t){return`${e}-content-${t}`}var MJ=Aq,PJ=Tq,jJ=Pq;function NJ({className:e,...t}){return E.jsx(MJ,{"data-slot":"tabs",className:ke("flex flex-col gap-2",e),...t})}function DJ({className:e,...t}){return E.jsx(PJ,{"data-slot":"tabs-list",className:ke("bg-muted text-muted-foreground inline-flex h-9 w-fit items-center justify-center rounded-lg p-[3px]",e),...t})}function QP({className:e,...t}){return E.jsx(jJ,{"data-slot":"tabs-trigger",className:ke("data-[state=active]:bg-background dark:data-[state=active]:text-foreground focus-visible:border-ring focus-visible:ring-ring/50 focus-visible:outline-ring dark:data-[state=active]:border-input dark:data-[state=active]:bg-input/30 text-foreground dark:text-muted-foreground inline-flex h-[calc(100%-1px)] flex-1 items-center justify-center gap-1.5 rounded-md border border-transparent px-2 py-1 text-sm font-medium whitespace-nowrap transition-[color,box-shadow] focus-visible:ring-[3px] focus-visible:outline-1 disabled:pointer-events-none disabled:opacity-50 data-[state=active]:shadow-sm [&_svg]:pointer-events-none [&_svg]:shrink-0 [&_svg:not([class*='size-'])]:size-4",e),...t})}function kq(e){const t=kJ(e),n=g.forwardRef((r,i)=>{const{children:a,...s}=r,u=g.Children.toArray(a),c=u.find(IJ);if(c){const d=c.props.children,p=u.map(m=>m===c?g.Children.count(d)>1?g.Children.only(null):g.isValidElement(d)?d.props.children:null:m);return E.jsx(t,{...s,ref:i,children:g.isValidElement(d)?g.cloneElement(d,void 0,p):null})}return E.jsx(t,{...s,ref:i,children:a})});return n.displayName=`${e}.Slot`,n}function kJ(e){const t=g.forwardRef((n,r)=>{const{children:i,...a}=n,s=g.isValidElement(i)?BJ(i):void 0,u=Re(s,r);if(g.isValidElement(i)){const c=LJ(a,i.props);return i.type!==g.Fragment&&(c.ref=u),g.cloneElement(i,c)}return g.Children.count(i)>1?g.Children.only(null):null});return t.displayName=`${e}.SlotClone`,t}var $J=Symbol("radix.slottable");function IJ(e){return g.isValidElement(e)&&typeof e.type=="function"&&"__radixId"in e.type&&e.type.__radixId===$J}function LJ(e,t){const n={...t};for(const r in t){const i=e[r],a=t[r];/^on[A-Z]/.test(r)?i&&a?n[r]=(...u)=>{const c=a(...u);return i(...u),c}:i&&(n[r]=i):r==="style"?n[r]={...i,...a}:r==="className"&&(n[r]=[i,a].filter(Boolean).join(" "))}return{...e,...n}}function BJ(e){var r,i;let t=(r=Object.getOwnPropertyDescriptor(e.props,"ref"))==null?void 0:r.get,n=t&&"isReactWarning"in t&&t.isReactWarning;return n?e.ref:(t=(i=Object.getOwnPropertyDescriptor(e,"ref"))==null?void 0:i.get,n=t&&"isReactWarning"in t&&t.isReactWarning,n?e.props.ref:e.props.ref||e.ref)}var FJ=["a","button","div","form","h2","h3","img","input","label","li","nav","ol","p","select","span","svg","ul"],Ov=FJ.reduce((e,t)=>{const n=kq(`Primitive.${t}`),r=g.forwardRef((i,a)=>{const{asChild:s,...u}=i,c=s?n:t;return typeof window<"u"&&(window[Symbol.for("radix-ui")]=!0),E.jsx(c,{...u,ref:a})});return r.displayName=`Primitive.${t}`,{...e,[t]:r}},{});function qJ(e,t){e&&ba.flushSync(()=>e.dispatchEvent(t))}var zJ="DismissableLayer",Y_="dismissableLayer.update",UJ="dismissableLayer.pointerDownOutside",VJ="dismissableLayer.focusOutside",JP,$q=g.createContext({layers:new Set,layersWithOutsidePointerEventsDisabled:new Set,branches:new Set}),Iq=g.forwardRef((e,t)=>{const{disableOutsidePointerEvents:n=!1,onEscapeKeyDown:r,onPointerDownOutside:i,onFocusOutside:a,onInteractOutside:s,onDismiss:u,...c}=e,d=g.useContext($q),[p,m]=g.useState(null),v=(p==null?void 0:p.ownerDocument)??(globalThis==null?void 0:globalThis.document),[,b]=g.useState({}),S=Re(t,P=>m(P)),w=Array.from(d.layers),[x]=[...d.layersWithOutsidePointerEventsDisabled].slice(-1),C=w.indexOf(x),O=p?w.indexOf(p):-1,A=d.layersWithOutsidePointerEventsDisabled.size>0,T=O>=C,R=WJ(P=>{const N=P.target,k=[...d.branches].some(H=>H.contains(N));!T||k||(i==null||i(P),s==null||s(P),P.defaultPrevented||u==null||u())},v),M=KJ(P=>{const N=P.target;[...d.branches].some(H=>H.contains(N))||(a==null||a(P),s==null||s(P),P.defaultPrevented||u==null||u())},v);return rd(P=>{O===d.layers.size-1&&(r==null||r(P),!P.defaultPrevented&&u&&(P.preventDefault(),u()))},v),g.useEffect(()=>{if(p)return n&&(d.layersWithOutsidePointerEventsDisabled.size===0&&(JP=v.body.style.pointerEvents,v.body.style.pointerEvents="none"),d.layersWithOutsidePointerEventsDisabled.add(p)),d.layers.add(p),ej(),()=>{n&&d.layersWithOutsidePointerEventsDisabled.size===1&&(v.body.style.pointerEvents=JP)}},[p,v,n,d]),g.useEffect(()=>()=>{p&&(d.layers.delete(p),d.layersWithOutsidePointerEventsDisabled.delete(p),ej())},[p,d]),g.useEffect(()=>{const P=()=>b({});return document.addEventListener(Y_,P),()=>document.removeEventListener(Y_,P)},[]),E.jsx(Ov.div,{...c,ref:S,style:{pointerEvents:A?T?"auto":"none":void 0,...e.style},onFocusCapture:be(e.onFocusCapture,M.onFocusCapture),onBlurCapture:be(e.onBlurCapture,M.onBlurCapture),onPointerDownCapture:be(e.onPointerDownCapture,R.onPointerDownCapture)})});Iq.displayName=zJ;var HJ="DismissableLayerBranch",GJ=g.forwardRef((e,t)=>{const n=g.useContext($q),r=g.useRef(null),i=Re(t,r);return g.useEffect(()=>{const a=r.current;if(a)return n.branches.add(a),()=>{n.branches.delete(a)}},[n.branches]),E.jsx(Ov.div,{...e,ref:i})});GJ.displayName=HJ;function WJ(e,t=globalThis==null?void 0:globalThis.document){const n=$t(e),r=g.useRef(!1),i=g.useRef(()=>{});return g.useEffect(()=>{const a=u=>{if(u.target&&!r.current){let c=function(){Lq(UJ,n,d,{discrete:!0})};const d={originalEvent:u};u.pointerType==="touch"?(t.removeEventListener("click",i.current),i.current=c,t.addEventListener("click",i.current,{once:!0})):c()}else t.removeEventListener("click",i.current);r.current=!1},s=window.setTimeout(()=>{t.addEventListener("pointerdown",a)},0);return()=>{window.clearTimeout(s),t.removeEventListener("pointerdown",a),t.removeEventListener("click",i.current)}},[t,n]),{onPointerDownCapture:()=>r.current=!0}}function KJ(e,t=globalThis==null?void 0:globalThis.document){const n=$t(e),r=g.useRef(!1);return g.useEffect(()=>{const i=a=>{a.target&&!r.current&&Lq(VJ,n,{originalEvent:a},{discrete:!1})};return t.addEventListener("focusin",i),()=>t.removeEventListener("focusin",i)},[t,n]),{onFocusCapture:()=>r.current=!0,onBlurCapture:()=>r.current=!1}}function ej(){const e=new CustomEvent(Y_);document.dispatchEvent(e)}function Lq(e,t,n,{discrete:r}){const i=n.originalEvent.target,a=new CustomEvent(e,{bubbles:!1,cancelable:!0,detail:n});t&&i.addEventListener(e,t,{once:!0}),r?qJ(i,a):i.dispatchEvent(a)}var Av="Popover",[Bq,mxe]=jn(Av,[fu]),gd=fu(),[YJ,ho]=Bq(Av),Fq=e=>{const{__scopePopover:t,children:n,open:r,defaultOpen:i,onOpenChange:a,modal:s=!1}=e,u=gd(t),c=g.useRef(null),[d,p]=g.useState(!1),[m,v]=Xr({prop:r,defaultProp:i??!1,onChange:a,caller:Av});return E.jsx(rC,{...u,children:E.jsx(YJ,{scope:t,contentId:ln(),triggerRef:c,open:m,onOpenChange:v,onOpenToggle:g.useCallback(()=>v(b=>!b),[v]),hasCustomAnchor:d,onCustomAnchorAdd:g.useCallback(()=>p(!0),[]),onCustomAnchorRemove:g.useCallback(()=>p(!1),[]),modal:s,children:n})})};Fq.displayName=Av;var qq="PopoverAnchor",ZJ=g.forwardRef((e,t)=>{const{__scopePopover:n,...r}=e,i=ho(qq,n),a=gd(n),{onCustomAnchorAdd:s,onCustomAnchorRemove:u}=i;return g.useEffect(()=>(s(),()=>u()),[s,u]),E.jsx(lv,{...a,...r,ref:t})});ZJ.displayName=qq;var zq="PopoverTrigger",Uq=g.forwardRef((e,t)=>{const{__scopePopover:n,...r}=e,i=ho(zq,n),a=gd(n),s=Re(t,i.triggerRef),u=E.jsx(Ov.button,{type:"button","aria-haspopup":"dialog","aria-expanded":i.open,"aria-controls":i.contentId,"data-state":Kq(i.open),...r,ref:s,onClick:be(e.onClick,i.onOpenToggle)});return i.hasCustomAnchor?u:E.jsx(lv,{asChild:!0,...a,children:u})});Uq.displayName=zq;var _C="PopoverPortal",[XJ,QJ]=Bq(_C,{forceMount:void 0}),Vq=e=>{const{__scopePopover:t,forceMount:n,children:r,container:i}=e,a=ho(_C,t);return E.jsx(XJ,{scope:t,forceMount:n,children:E.jsx(zn,{present:n||a.open,children:E.jsx(fd,{asChild:!0,container:i,children:r})})})};Vq.displayName=_C;var kl="PopoverContent",Hq=g.forwardRef((e,t)=>{const n=QJ(kl,e.__scopePopover),{forceMount:r=n.forceMount,...i}=e,a=ho(kl,e.__scopePopover);return E.jsx(zn,{present:r||a.open,children:a.modal?E.jsx(eee,{...i,ref:t}):E.jsx(tee,{...i,ref:t})})});Hq.displayName=kl;var JJ=kq("PopoverContent.RemoveScroll"),eee=g.forwardRef((e,t)=>{const n=ho(kl,e.__scopePopover),r=g.useRef(null),i=Re(t,r),a=g.useRef(!1);return g.useEffect(()=>{const s=r.current;if(s)return uv(s)},[]),E.jsx(dd,{as:JJ,allowPinchZoom:!0,children:E.jsx(Gq,{...e,ref:i,trapFocus:n.open,disableOutsidePointerEvents:!0,onCloseAutoFocus:be(e.onCloseAutoFocus,s=>{var u;s.preventDefault(),a.current||(u=n.triggerRef.current)==null||u.focus()}),onPointerDownOutside:be(e.onPointerDownOutside,s=>{const u=s.detail.originalEvent,c=u.button===0&&u.ctrlKey===!0,d=u.button===2||c;a.current=d},{checkForDefaultPrevented:!1}),onFocusOutside:be(e.onFocusOutside,s=>s.preventDefault(),{checkForDefaultPrevented:!1})})})}),tee=g.forwardRef((e,t)=>{const n=ho(kl,e.__scopePopover),r=g.useRef(!1),i=g.useRef(!1);return E.jsx(Gq,{...e,ref:t,trapFocus:!1,disableOutsidePointerEvents:!1,onCloseAutoFocus:a=>{var s,u;(s=e.onCloseAutoFocus)==null||s.call(e,a),a.defaultPrevented||(r.current||(u=n.triggerRef.current)==null||u.focus(),a.preventDefault()),r.current=!1,i.current=!1},onInteractOutside:a=>{var c,d;(c=e.onInteractOutside)==null||c.call(e,a),a.defaultPrevented||(r.current=!0,a.detail.originalEvent.type==="pointerdown"&&(i.current=!0));const s=a.target;((d=n.triggerRef.current)==null?void 0:d.contains(s))&&a.preventDefault(),a.detail.originalEvent.type==="focusin"&&i.current&&a.preventDefault()}})}),Gq=g.forwardRef((e,t)=>{const{__scopePopover:n,trapFocus:r,onOpenAutoFocus:i,onCloseAutoFocus:a,disableOutsidePointerEvents:s,onEscapeKeyDown:u,onPointerDownOutside:c,onFocusOutside:d,onInteractOutside:p,...m}=e,v=ho(kl,n),b=gd(n);return iv(),E.jsx(ud,{asChild:!0,loop:!0,trapped:r,onMountAutoFocus:i,onUnmountAutoFocus:a,children:E.jsx(Iq,{asChild:!0,disableOutsidePointerEvents:s,onInteractOutside:p,onEscapeKeyDown:u,onPointerDownOutside:c,onFocusOutside:d,onDismiss:()=>v.onOpenChange(!1),children:E.jsx(iC,{"data-state":Kq(v.open),role:"dialog",id:v.contentId,...b,...m,ref:t,style:{...m.style,"--radix-popover-content-transform-origin":"var(--radix-popper-transform-origin)","--radix-popover-content-available-width":"var(--radix-popper-available-width)","--radix-popover-content-available-height":"var(--radix-popper-available-height)","--radix-popover-trigger-width":"var(--radix-popper-anchor-width)","--radix-popover-trigger-height":"var(--radix-popper-anchor-height)"}})})})}),Wq="PopoverClose",nee=g.forwardRef((e,t)=>{const{__scopePopover:n,...r}=e,i=ho(Wq,n);return E.jsx(Ov.button,{type:"button",...r,ref:t,onClick:be(e.onClick,()=>i.onOpenChange(!1))})});nee.displayName=Wq;var ree="PopoverArrow",iee=g.forwardRef((e,t)=>{const{__scopePopover:n,...r}=e,i=gd(n);return E.jsx(aC,{...i,...r,ref:t})});iee.displayName=ree;function Kq(e){return e?"open":"closed"}var aee=Fq,oee=Uq,see=Vq,lee=Hq;function uee({...e}){return E.jsx(aee,{"data-slot":"popover",...e})}function cee({...e}){return E.jsx(oee,{"data-slot":"popover-trigger",...e})}function fee({className:e,align:t="center",sideOffset:n=4,...r}){return E.jsx(see,{children:E.jsx(lee,{"data-slot":"popover-content",align:t,sideOffset:n,className:ke("bg-popover text-popover-foreground data-[state=open]:animate-in data-[state=closed]:animate-out data-[state=closed]:fade-out-0 data-[state=open]:fade-in-0 data-[state=closed]:zoom-out-95 data-[state=open]:zoom-in-95 data-[side=bottom]:slide-in-from-top-2 data-[side=left]:slide-in-from-right-2 data-[side=right]:slide-in-from-left-2 data-[side=top]:slide-in-from-bottom-2 z-50 w-72 origin-(--radix-popover-content-transform-origin) rounded-md border p-4 shadow-md outline-hidden",e),...r})})}var tj=1,dee=.9,pee=.8,hee=.17,Jb=.1,e0=.999,mee=.9999,vee=.99,gee=/[\\\/_+.#"@\[\(\{&]/,yee=/[\\\/_+.#"@\[\(\{&]/g,bee=/[\s-]/,Yq=/[\s-]/g;function Z_(e,t,n,r,i,a,s){if(a===t.length)return i===e.length?tj:vee;var u=`${i},${a}`;if(s[u]!==void 0)return s[u];for(var c=r.charAt(a),d=n.indexOf(c,i),p=0,m,v,b,S;d>=0;)m=Z_(e,t,n,r,d+1,a+1,s),m>p&&(d===i?m*=tj:gee.test(e.charAt(d-1))?(m*=pee,b=e.slice(i,d-1).match(yee),b&&i>0&&(m*=Math.pow(e0,b.length))):bee.test(e.charAt(d-1))?(m*=dee,S=e.slice(i,d-1).match(Yq),S&&i>0&&(m*=Math.pow(e0,S.length))):(m*=hee,i>0&&(m*=Math.pow(e0,d-i))),e.charAt(d)!==t.charAt(a)&&(m*=mee)),(m<Jb&&n.charAt(d-1)===r.charAt(a+1)||r.charAt(a+1)===r.charAt(a)&&n.charAt(d-1)!==r.charAt(a))&&(v=Z_(e,t,n,r,d+1,a+2,s),v*Jb>m&&(m=v*Jb)),m>p&&(p=m),d=n.indexOf(c,d+1);return s[u]=p,p}function nj(e){return e.toLowerCase().replace(Yq," ")}function xee(e,t,n){return e=n&&n.length>0?`${e+" "+n.join(" ")}`:e,Z_(e,t,nj(e),nj(t),0,0,{})}function Zq(e){const t=wee(e),n=g.forwardRef((r,i)=>{const{children:a,...s}=r,u=g.Children.toArray(a),c=u.find(_ee);if(c){const d=c.props.children,p=u.map(m=>m===c?g.Children.count(d)>1?g.Children.only(null):g.isValidElement(d)?d.props.children:null:m);return E.jsx(t,{...s,ref:i,children:g.isValidElement(d)?g.cloneElement(d,void 0,p):null})}return E.jsx(t,{...s,ref:i,children:a})});return n.displayName=`${e}.Slot`,n}function wee(e){const t=g.forwardRef((n,r)=>{const{children:i,...a}=n,s=g.isValidElement(i)?Cee(i):void 0,u=Re(s,r);if(g.isValidElement(i)){const c=Eee(a,i.props);return i.type!==g.Fragment&&(c.ref=u),g.cloneElement(i,c)}return g.Children.count(i)>1?g.Children.only(null):null});return t.displayName=`${e}.SlotClone`,t}var See=Symbol("radix.slottable");function _ee(e){return g.isValidElement(e)&&typeof e.type=="function"&&"__radixId"in e.type&&e.type.__radixId===See}function Eee(e,t){const n={...t};for(const r in t){const i=e[r],a=t[r];/^on[A-Z]/.test(r)?i&&a?n[r]=(...u)=>{const c=a(...u);return i(...u),c}:i&&(n[r]=i):r==="style"?n[r]={...i,...a}:r==="className"&&(n[r]=[i,a].filter(Boolean).join(" "))}return{...e,...n}}function Cee(e){var r,i;let t=(r=Object.getOwnPropertyDescriptor(e.props,"ref"))==null?void 0:r.get,n=t&&"isReactWarning"in t&&t.isReactWarning;return n?e.ref:(t=(i=Object.getOwnPropertyDescriptor(e,"ref"))==null?void 0:i.get,n=t&&"isReactWarning"in t&&t.isReactWarning,n?e.props.ref:e.props.ref||e.ref)}var Oee=["a","button","div","form","h2","h3","img","input","label","li","nav","ol","p","select","span","svg","ul"],vs=Oee.reduce((e,t)=>{const n=Zq(`Primitive.${t}`),r=g.forwardRef((i,a)=>{const{asChild:s,...u}=i,c=s?n:t;return typeof window<"u"&&(window[Symbol.for("radix-ui")]=!0),E.jsx(c,{...u,ref:a})});return r.displayName=`Primitive.${t}`,{...e,[t]:r}},{});function Aee(e,t){e&&ba.flushSync(()=>e.dispatchEvent(t))}var Ree="DismissableLayer",X_="dismissableLayer.update",Tee="dismissableLayer.pointerDownOutside",Mee="dismissableLayer.focusOutside",rj,Xq=g.createContext({layers:new Set,layersWithOutsidePointerEventsDisabled:new Set,branches:new Set}),Qq=g.forwardRef((e,t)=>{const{disableOutsidePointerEvents:n=!1,onEscapeKeyDown:r,onPointerDownOutside:i,onFocusOutside:a,onInteractOutside:s,onDismiss:u,...c}=e,d=g.useContext(Xq),[p,m]=g.useState(null),v=(p==null?void 0:p.ownerDocument)??(globalThis==null?void 0:globalThis.document),[,b]=g.useState({}),S=Re(t,P=>m(P)),w=Array.from(d.layers),[x]=[...d.layersWithOutsidePointerEventsDisabled].slice(-1),C=w.indexOf(x),O=p?w.indexOf(p):-1,A=d.layersWithOutsidePointerEventsDisabled.size>0,T=O>=C,R=Nee(P=>{const N=P.target,k=[...d.branches].some(H=>H.contains(N));!T||k||(i==null||i(P),s==null||s(P),P.defaultPrevented||u==null||u())},v),M=Dee(P=>{const N=P.target;[...d.branches].some(H=>H.contains(N))||(a==null||a(P),s==null||s(P),P.defaultPrevented||u==null||u())},v);return rd(P=>{O===d.layers.size-1&&(r==null||r(P),!P.defaultPrevented&&u&&(P.preventDefault(),u()))},v),g.useEffect(()=>{if(p)return n&&(d.layersWithOutsidePointerEventsDisabled.size===0&&(rj=v.body.style.pointerEvents,v.body.style.pointerEvents="none"),d.layersWithOutsidePointerEventsDisabled.add(p)),d.layers.add(p),ij(),()=>{n&&d.layersWithOutsidePointerEventsDisabled.size===1&&(v.body.style.pointerEvents=rj)}},[p,v,n,d]),g.useEffect(()=>()=>{p&&(d.layers.delete(p),d.layersWithOutsidePointerEventsDisabled.delete(p),ij())},[p,d]),g.useEffect(()=>{const P=()=>b({});return document.addEventListener(X_,P),()=>document.removeEventListener(X_,P)},[]),E.jsx(vs.div,{...c,ref:S,style:{pointerEvents:A?T?"auto":"none":void 0,...e.style},onFocusCapture:be(e.onFocusCapture,M.onFocusCapture),onBlurCapture:be(e.onBlurCapture,M.onBlurCapture),onPointerDownCapture:be(e.onPointerDownCapture,R.onPointerDownCapture)})});Qq.displayName=Ree;var Pee="DismissableLayerBranch",jee=g.forwardRef((e,t)=>{const n=g.useContext(Xq),r=g.useRef(null),i=Re(t,r);return g.useEffect(()=>{const a=r.current;if(a)return n.branches.add(a),()=>{n.branches.delete(a)}},[n.branches]),E.jsx(vs.div,{...e,ref:i})});jee.displayName=Pee;function Nee(e,t=globalThis==null?void 0:globalThis.document){const n=$t(e),r=g.useRef(!1),i=g.useRef(()=>{});return g.useEffect(()=>{const a=u=>{if(u.target&&!r.current){let c=function(){Jq(Tee,n,d,{discrete:!0})};const d={originalEvent:u};u.pointerType==="touch"?(t.removeEventListener("click",i.current),i.current=c,t.addEventListener("click",i.current,{once:!0})):c()}else t.removeEventListener("click",i.current);r.current=!1},s=window.setTimeout(()=>{t.addEventListener("pointerdown",a)},0);return()=>{window.clearTimeout(s),t.removeEventListener("pointerdown",a),t.removeEventListener("click",i.current)}},[t,n]),{onPointerDownCapture:()=>r.current=!0}}function Dee(e,t=globalThis==null?void 0:globalThis.document){const n=$t(e),r=g.useRef(!1);return g.useEffect(()=>{const i=a=>{a.target&&!r.current&&Jq(Mee,n,{originalEvent:a},{discrete:!1})};return t.addEventListener("focusin",i),()=>t.removeEventListener("focusin",i)},[t,n]),{onFocusCapture:()=>r.current=!0,onBlurCapture:()=>r.current=!1}}function ij(){const e=new CustomEvent(X_);document.dispatchEvent(e)}function Jq(e,t,n,{discrete:r}){const i=n.originalEvent.target,a=new CustomEvent(e,{bubbles:!1,cancelable:!0,detail:n});t&&i.addEventListener(e,t,{once:!0}),r?Aee(i,a):i.dispatchEvent(a)}var Rv="Dialog",[ez,vxe]=jn(Rv),[kee,ni]=ez(Rv),tz=e=>{const{__scopeDialog:t,children:n,open:r,defaultOpen:i,onOpenChange:a,modal:s=!0}=e,u=g.useRef(null),c=g.useRef(null),[d,p]=Xr({prop:r,defaultProp:i??!1,onChange:a,caller:Rv});return E.jsx(kee,{scope:t,triggerRef:u,contentRef:c,contentId:ln(),titleId:ln(),descriptionId:ln(),open:d,onOpenChange:p,onOpenToggle:g.useCallback(()=>p(m=>!m),[p]),modal:s,children:n})};tz.displayName=Rv;var nz="DialogTrigger",$ee=g.forwardRef((e,t)=>{const{__scopeDialog:n,...r}=e,i=ni(nz,n),a=Re(t,i.triggerRef);return E.jsx(vs.button,{type:"button","aria-haspopup":"dialog","aria-expanded":i.open,"aria-controls":i.contentId,"data-state":OC(i.open),...r,ref:a,onClick:be(e.onClick,i.onOpenToggle)})});$ee.displayName=nz;var EC="DialogPortal",[Iee,rz]=ez(EC,{forceMount:void 0}),iz=e=>{const{__scopeDialog:t,forceMount:n,children:r,container:i}=e,a=ni(EC,t);return E.jsx(Iee,{scope:t,forceMount:n,children:g.Children.map(r,s=>E.jsx(zn,{present:n||a.open,children:E.jsx(fd,{asChild:!0,container:i,children:s})}))})};iz.displayName=EC;var im="DialogOverlay",az=g.forwardRef((e,t)=>{const n=rz(im,e.__scopeDialog),{forceMount:r=n.forceMount,...i}=e,a=ni(im,e.__scopeDialog);return a.modal?E.jsx(zn,{present:r||a.open,children:E.jsx(Bee,{...i,ref:t})}):null});az.displayName=im;var Lee=Zq("DialogOverlay.RemoveScroll"),Bee=g.forwardRef((e,t)=>{const{__scopeDialog:n,...r}=e,i=ni(im,n);return E.jsx(dd,{as:Lee,allowPinchZoom:!0,shards:[i.contentRef],children:E.jsx(vs.div,{"data-state":OC(i.open),...r,ref:t,style:{pointerEvents:"auto",...r.style}})})}),fs="DialogContent",oz=g.forwardRef((e,t)=>{const n=rz(fs,e.__scopeDialog),{forceMount:r=n.forceMount,...i}=e,a=ni(fs,e.__scopeDialog);return E.jsx(zn,{present:r||a.open,children:a.modal?E.jsx(Fee,{...i,ref:t}):E.jsx(qee,{...i,ref:t})})});oz.displayName=fs;var Fee=g.forwardRef((e,t)=>{const n=ni(fs,e.__scopeDialog),r=g.useRef(null),i=Re(t,n.contentRef,r);return g.useEffect(()=>{const a=r.current;if(a)return uv(a)},[]),E.jsx(sz,{...e,ref:i,trapFocus:n.open,disableOutsidePointerEvents:!0,onCloseAutoFocus:be(e.onCloseAutoFocus,a=>{var s;a.preventDefault(),(s=n.triggerRef.current)==null||s.focus()}),onPointerDownOutside:be(e.onPointerDownOutside,a=>{const s=a.detail.originalEvent,u=s.button===0&&s.ctrlKey===!0;(s.button===2||u)&&a.preventDefault()}),onFocusOutside:be(e.onFocusOutside,a=>a.preventDefault())})}),qee=g.forwardRef((e,t)=>{const n=ni(fs,e.__scopeDialog),r=g.useRef(!1),i=g.useRef(!1);return E.jsx(sz,{...e,ref:t,trapFocus:!1,disableOutsidePointerEvents:!1,onCloseAutoFocus:a=>{var s,u;(s=e.onCloseAutoFocus)==null||s.call(e,a),a.defaultPrevented||(r.current||(u=n.triggerRef.current)==null||u.focus(),a.preventDefault()),r.current=!1,i.current=!1},onInteractOutside:a=>{var c,d;(c=e.onInteractOutside)==null||c.call(e,a),a.defaultPrevented||(r.current=!0,a.detail.originalEvent.type==="pointerdown"&&(i.current=!0));const s=a.target;((d=n.triggerRef.current)==null?void 0:d.contains(s))&&a.preventDefault(),a.detail.originalEvent.type==="focusin"&&i.current&&a.preventDefault()}})}),sz=g.forwardRef((e,t)=>{const{__scopeDialog:n,trapFocus:r,onOpenAutoFocus:i,onCloseAutoFocus:a,...s}=e,u=ni(fs,n),c=g.useRef(null),d=Re(t,c);return iv(),E.jsxs(E.Fragment,{children:[E.jsx(ud,{asChild:!0,loop:!0,trapped:r,onMountAutoFocus:i,onUnmountAutoFocus:a,children:E.jsx(Qq,{role:"dialog",id:u.contentId,"aria-describedby":u.descriptionId,"aria-labelledby":u.titleId,"data-state":OC(u.open),...s,ref:d,onDismiss:()=>u.onOpenChange(!1)})}),E.jsxs(E.Fragment,{children:[E.jsx(Hee,{titleId:u.titleId}),E.jsx(Wee,{contentRef:c,descriptionId:u.descriptionId})]})]})}),CC="DialogTitle",zee=g.forwardRef((e,t)=>{const{__scopeDialog:n,...r}=e,i=ni(CC,n);return E.jsx(vs.h2,{id:i.titleId,...r,ref:t})});zee.displayName=CC;var lz="DialogDescription",Uee=g.forwardRef((e,t)=>{const{__scopeDialog:n,...r}=e,i=ni(lz,n);return E.jsx(vs.p,{id:i.descriptionId,...r,ref:t})});Uee.displayName=lz;var uz="DialogClose",Vee=g.forwardRef((e,t)=>{const{__scopeDialog:n,...r}=e,i=ni(uz,n);return E.jsx(vs.button,{type:"button",...r,ref:t,onClick:be(e.onClick,()=>i.onOpenChange(!1))})});Vee.displayName=uz;function OC(e){return e?"open":"closed"}var cz="DialogTitleWarning",[gxe,fz]=Z5(cz,{contentName:fs,titleName:CC,docsSlug:"dialog"}),Hee=({titleId:e})=>{const t=fz(cz),n=`\`${t.contentName}\` requires a \`${t.titleName}\` for the component to be accessible for screen reader users.

If you want to hide the \`${t.titleName}\`, you can wrap it with our VisuallyHidden component.

For more information, see https://radix-ui.com/primitives/docs/components/${t.docsSlug}`;return g.useEffect(()=>{e&&(document.getElementById(e)||console.error(n))},[n,e]),null},Gee="DialogDescriptionWarning",Wee=({contentRef:e,descriptionId:t})=>{const r=`Warning: Missing \`Description\` or \`aria-describedby={undefined}\` for {${fz(Gee).contentName}}.`;return g.useEffect(()=>{var a;const i=(a=e.current)==null?void 0:a.getAttribute("aria-describedby");t&&i&&(document.getElementById(t)||console.warn(r))},[r,e,t]),null},Kee=tz,Yee=iz,Zee=az,Xee=oz,Ac='[cmdk-group=""]',t0='[cmdk-group-items=""]',Qee='[cmdk-group-heading=""]',dz='[cmdk-item=""]',aj=`${dz}:not([aria-disabled="true"])`,Q_="cmdk-item-select",gl="data-value",Jee=(e,t,n)=>xee(e,t,n),pz=g.createContext(void 0),yd=()=>g.useContext(pz),hz=g.createContext(void 0),AC=()=>g.useContext(hz),mz=g.createContext(void 0),vz=g.forwardRef((e,t)=>{let n=yl(()=>{var q,X;return{search:"",value:(X=(q=e.value)!=null?q:e.defaultValue)!=null?X:"",selectedItemId:void 0,filtered:{count:0,items:new Map,groups:new Set}}}),r=yl(()=>new Set),i=yl(()=>new Map),a=yl(()=>new Map),s=yl(()=>new Set),u=gz(e),{label:c,children:d,value:p,onValueChange:m,filter:v,shouldFilter:b,loop:S,disablePointerSelection:w=!1,vimBindings:x=!0,...C}=e,O=ln(),A=ln(),T=ln(),R=g.useRef(null),M=cte();ds(()=>{if(p!==void 0){let q=p.trim();n.current.value=q,P.emit()}},[p]),ds(()=>{M(6,z)},[]);let P=g.useMemo(()=>({subscribe:q=>(s.current.add(q),()=>s.current.delete(q)),snapshot:()=>n.current,setState:(q,X,oe)=>{var le,ve,de,Se;if(!Object.is(n.current[q],X)){if(n.current[q]=X,q==="search")F(),H(),M(1,L);else if(q==="value"){if(document.activeElement.hasAttribute("cmdk-input")||document.activeElement.hasAttribute("cmdk-root")){let ce=document.getElementById(T);ce?ce.focus():(le=document.getElementById(O))==null||le.focus()}if(M(7,()=>{var ce;n.current.selectedItemId=(ce=Z())==null?void 0:ce.id,P.emit()}),oe||M(5,z),((ve=u.current)==null?void 0:ve.value)!==void 0){let ce=X??"";(Se=(de=u.current).onValueChange)==null||Se.call(de,ce);return}}P.emit()}},emit:()=>{s.current.forEach(q=>q())}}),[]),N=g.useMemo(()=>({value:(q,X,oe)=>{var le;X!==((le=a.current.get(q))==null?void 0:le.value)&&(a.current.set(q,{value:X,keywords:oe}),n.current.filtered.items.set(q,k(X,oe)),M(2,()=>{H(),P.emit()}))},item:(q,X)=>(r.current.add(q),X&&(i.current.has(X)?i.current.get(X).add(q):i.current.set(X,new Set([q]))),M(3,()=>{F(),H(),n.current.value||L(),P.emit()}),()=>{a.current.delete(q),r.current.delete(q),n.current.filtered.items.delete(q);let oe=Z();M(4,()=>{F(),(oe==null?void 0:oe.getAttribute("id"))===q&&L(),P.emit()})}),group:q=>(i.current.has(q)||i.current.set(q,new Set),()=>{a.current.delete(q),i.current.delete(q)}),filter:()=>u.current.shouldFilter,label:c||e["aria-label"],getDisablePointerSelection:()=>u.current.disablePointerSelection,listId:O,inputId:T,labelId:A,listInnerRef:R}),[]);function k(q,X){var oe,le;let ve=(le=(oe=u.current)==null?void 0:oe.filter)!=null?le:Jee;return q?ve(q,n.current.search,X):0}function H(){if(!n.current.search||u.current.shouldFilter===!1)return;let q=n.current.filtered.items,X=[];n.current.filtered.groups.forEach(le=>{let ve=i.current.get(le),de=0;ve.forEach(Se=>{let ce=q.get(Se);de=Math.max(ce,de)}),X.push([le,de])});let oe=R.current;Y().sort((le,ve)=>{var de,Se;let ce=le.getAttribute("id"),we=ve.getAttribute("id");return((de=q.get(we))!=null?de:0)-((Se=q.get(ce))!=null?Se:0)}).forEach(le=>{let ve=le.closest(t0);ve?ve.appendChild(le.parentElement===ve?le:le.closest(`${t0} > *`)):oe.appendChild(le.parentElement===oe?le:le.closest(`${t0} > *`))}),X.sort((le,ve)=>ve[1]-le[1]).forEach(le=>{var ve;let de=(ve=R.current)==null?void 0:ve.querySelector(`${Ac}[${gl}="${encodeURIComponent(le[0])}"]`);de==null||de.parentElement.appendChild(de)})}function L(){let q=Y().find(oe=>oe.getAttribute("aria-disabled")!=="true"),X=q==null?void 0:q.getAttribute(gl);P.setState("value",X||void 0)}function F(){var q,X,oe,le;if(!n.current.search||u.current.shouldFilter===!1){n.current.filtered.count=r.current.size;return}n.current.filtered.groups=new Set;let ve=0;for(let de of r.current){let Se=(X=(q=a.current.get(de))==null?void 0:q.value)!=null?X:"",ce=(le=(oe=a.current.get(de))==null?void 0:oe.keywords)!=null?le:[],we=k(Se,ce);n.current.filtered.items.set(de,we),we>0&&ve++}for(let[de,Se]of i.current)for(let ce of Se)if(n.current.filtered.items.get(ce)>0){n.current.filtered.groups.add(de);break}n.current.filtered.count=ve}function z(){var q,X,oe;let le=Z();le&&(((q=le.parentElement)==null?void 0:q.firstChild)===le&&((oe=(X=le.closest(Ac))==null?void 0:X.querySelector(Qee))==null||oe.scrollIntoView({block:"nearest"})),le.scrollIntoView({block:"nearest"}))}function Z(){var q;return(q=R.current)==null?void 0:q.querySelector(`${dz}[aria-selected="true"]`)}function Y(){var q;return Array.from(((q=R.current)==null?void 0:q.querySelectorAll(aj))||[])}function I(q){let X=Y()[q];X&&P.setState("value",X.getAttribute(gl))}function W(q){var X;let oe=Z(),le=Y(),ve=le.findIndex(Se=>Se===oe),de=le[ve+q];(X=u.current)!=null&&X.loop&&(de=ve+q<0?le[le.length-1]:ve+q===le.length?le[0]:le[ve+q]),de&&P.setState("value",de.getAttribute(gl))}function B(q){let X=Z(),oe=X==null?void 0:X.closest(Ac),le;for(;oe&&!le;)oe=q>0?lte(oe,Ac):ute(oe,Ac),le=oe==null?void 0:oe.querySelector(aj);le?P.setState("value",le.getAttribute(gl)):W(q)}let Q=()=>I(Y().length-1),$=q=>{q.preventDefault(),q.metaKey?Q():q.altKey?B(1):W(1)},J=q=>{q.preventDefault(),q.metaKey?I(0):q.altKey?B(-1):W(-1)};return g.createElement(Ut.div,{ref:t,tabIndex:-1,...C,"cmdk-root":"",onKeyDown:q=>{var X;(X=C.onKeyDown)==null||X.call(C,q);let oe=q.nativeEvent.isComposing||q.keyCode===229;if(!(q.defaultPrevented||oe))switch(q.key){case"n":case"j":{x&&q.ctrlKey&&$(q);break}case"ArrowDown":{$(q);break}case"p":case"k":{x&&q.ctrlKey&&J(q);break}case"ArrowUp":{J(q);break}case"Home":{q.preventDefault(),I(0);break}case"End":{q.preventDefault(),Q();break}case"Enter":{q.preventDefault();let le=Z();if(le){let ve=new Event(Q_);le.dispatchEvent(ve)}}}}},g.createElement("label",{"cmdk-label":"",htmlFor:N.inputId,id:N.labelId,style:dte},c),Tv(e,q=>g.createElement(hz.Provider,{value:P},g.createElement(pz.Provider,{value:N},q))))}),ete=g.forwardRef((e,t)=>{var n,r;let i=ln(),a=g.useRef(null),s=g.useContext(mz),u=yd(),c=gz(e),d=(r=(n=c.current)==null?void 0:n.forceMount)!=null?r:s==null?void 0:s.forceMount;ds(()=>{if(!d)return u.item(i,s==null?void 0:s.id)},[d]);let p=yz(i,a,[e.value,e.children,a],e.keywords),m=AC(),v=uo(M=>M.value&&M.value===p.current),b=uo(M=>d||u.filter()===!1?!0:M.search?M.filtered.items.get(i)>0:!0);g.useEffect(()=>{let M=a.current;if(!(!M||e.disabled))return M.addEventListener(Q_,S),()=>M.removeEventListener(Q_,S)},[b,e.onSelect,e.disabled]);function S(){var M,P;w(),(P=(M=c.current).onSelect)==null||P.call(M,p.current)}function w(){m.setState("value",p.current,!0)}if(!b)return null;let{disabled:x,value:C,onSelect:O,forceMount:A,keywords:T,...R}=e;return g.createElement(Ut.div,{ref:wi(a,t),...R,id:i,"cmdk-item":"",role:"option","aria-disabled":!!x,"aria-selected":!!v,"data-disabled":!!x,"data-selected":!!v,onPointerMove:x||u.getDisablePointerSelection()?void 0:w,onClick:x?void 0:S},e.children)}),tte=g.forwardRef((e,t)=>{let{heading:n,children:r,forceMount:i,...a}=e,s=ln(),u=g.useRef(null),c=g.useRef(null),d=ln(),p=yd(),m=uo(b=>i||p.filter()===!1?!0:b.search?b.filtered.groups.has(s):!0);ds(()=>p.group(s),[]),yz(s,u,[e.value,e.heading,c]);let v=g.useMemo(()=>({id:s,forceMount:i}),[i]);return g.createElement(Ut.div,{ref:wi(u,t),...a,"cmdk-group":"",role:"presentation",hidden:m?void 0:!0},n&&g.createElement("div",{ref:c,"cmdk-group-heading":"","aria-hidden":!0,id:d},n),Tv(e,b=>g.createElement("div",{"cmdk-group-items":"",role:"group","aria-labelledby":n?d:void 0},g.createElement(mz.Provider,{value:v},b))))}),nte=g.forwardRef((e,t)=>{let{alwaysRender:n,...r}=e,i=g.useRef(null),a=uo(s=>!s.search);return!n&&!a?null:g.createElement(Ut.div,{ref:wi(i,t),...r,"cmdk-separator":"",role:"separator"})}),rte=g.forwardRef((e,t)=>{let{onValueChange:n,...r}=e,i=e.value!=null,a=AC(),s=uo(d=>d.search),u=uo(d=>d.selectedItemId),c=yd();return g.useEffect(()=>{e.value!=null&&a.setState("search",e.value)},[e.value]),g.createElement(Ut.input,{ref:t,...r,"cmdk-input":"",autoComplete:"off",autoCorrect:"off",spellCheck:!1,"aria-autocomplete":"list",role:"combobox","aria-expanded":!0,"aria-controls":c.listId,"aria-labelledby":c.labelId,"aria-activedescendant":u,id:c.inputId,type:"text",value:i?e.value:s,onChange:d=>{i||a.setState("search",d.target.value),n==null||n(d.target.value)}})}),ite=g.forwardRef((e,t)=>{let{children:n,label:r="Suggestions",...i}=e,a=g.useRef(null),s=g.useRef(null),u=uo(d=>d.selectedItemId),c=yd();return g.useEffect(()=>{if(s.current&&a.current){let d=s.current,p=a.current,m,v=new ResizeObserver(()=>{m=requestAnimationFrame(()=>{let b=d.offsetHeight;p.style.setProperty("--cmdk-list-height",b.toFixed(1)+"px")})});return v.observe(d),()=>{cancelAnimationFrame(m),v.unobserve(d)}}},[]),g.createElement(Ut.div,{ref:wi(a,t),...i,"cmdk-list":"",role:"listbox",tabIndex:-1,"aria-activedescendant":u,"aria-label":r,id:c.listId},Tv(e,d=>g.createElement("div",{ref:wi(s,c.listInnerRef),"cmdk-list-sizer":""},d)))}),ate=g.forwardRef((e,t)=>{let{open:n,onOpenChange:r,overlayClassName:i,contentClassName:a,container:s,...u}=e;return g.createElement(Kee,{open:n,onOpenChange:r},g.createElement(Yee,{container:s},g.createElement(Zee,{"cmdk-overlay":"",className:i}),g.createElement(Xee,{"aria-label":e.label,"cmdk-dialog":"",className:a},g.createElement(vz,{ref:t,...u}))))}),ote=g.forwardRef((e,t)=>uo(n=>n.filtered.count===0)?g.createElement(Ut.div,{ref:t,...e,"cmdk-empty":"",role:"presentation"}):null),ste=g.forwardRef((e,t)=>{let{progress:n,children:r,label:i="Loading...",...a}=e;return g.createElement(Ut.div,{ref:t,...a,"cmdk-loading":"",role:"progressbar","aria-valuenow":n,"aria-valuemin":0,"aria-valuemax":100,"aria-label":i},Tv(e,s=>g.createElement("div",{"aria-hidden":!0},s)))}),vu=Object.assign(vz,{List:ite,Item:ete,Input:rte,Group:tte,Separator:nte,Dialog:ate,Empty:ote,Loading:ste});function lte(e,t){let n=e.nextElementSibling;for(;n;){if(n.matches(t))return n;n=n.nextElementSibling}}function ute(e,t){let n=e.previousElementSibling;for(;n;){if(n.matches(t))return n;n=n.previousElementSibling}}function gz(e){let t=g.useRef(e);return ds(()=>{t.current=e}),t}var ds=typeof window>"u"?g.useEffect:g.useLayoutEffect;function yl(e){let t=g.useRef();return t.current===void 0&&(t.current=e()),t}function uo(e){let t=AC(),n=()=>e(t.snapshot());return g.useSyncExternalStore(t.subscribe,n,n)}function yz(e,t,n,r=[]){let i=g.useRef(),a=yd();return ds(()=>{var s;let u=(()=>{var d;for(let p of n){if(typeof p=="string")return p.trim();if(typeof p=="object"&&"current"in p)return p.current?(d=p.current.textContent)==null?void 0:d.trim():i.current}})(),c=r.map(d=>d.trim());a.value(e,u,c),(s=t.current)==null||s.setAttribute(gl,u),i.current=u}),i}var cte=()=>{let[e,t]=g.useState(),n=yl(()=>new Map);return ds(()=>{n.current.forEach(r=>r()),n.current=new Map},[e]),(r,i)=>{n.current.set(r,i),t({})}};function fte(e){let t=e.type;return typeof t=="function"?t(e.props):"render"in t?t.render(e.props):e}function Tv({asChild:e,children:t},n){return e&&g.isValidElement(t)?g.cloneElement(fte(t),{ref:t.ref},n(t.props.children)):n(t)}var dte={position:"absolute",width:"1px",height:"1px",padding:"0",margin:"-1px",overflow:"hidden",clip:"rect(0, 0, 0, 0)",whiteSpace:"nowrap",borderWidth:"0"};function pte({className:e,...t}){return E.jsx(vu,{"data-slot":"command",className:ke("bg-popover text-popover-foreground flex h-full w-full flex-col overflow-hidden rounded-md",e),...t})}function hte({className:e,...t}){return E.jsxs("div",{"data-slot":"command-input-wrapper",className:"flex h-9 items-center gap-2 border-b px-3",children:[E.jsx(w8,{className:"size-4 shrink-0 opacity-50"}),E.jsx(vu.Input,{"data-slot":"command-input",className:ke("placeholder:text-muted-foreground flex h-10 w-full rounded-md bg-transparent py-3 text-sm outline-hidden disabled:cursor-not-allowed disabled:opacity-50",e),...t})]})}function mte({className:e,...t}){return E.jsx(vu.List,{"data-slot":"command-list",className:ke("max-h-[300px] scroll-py-1 overflow-x-hidden overflow-y-auto",e),...t})}function vte({...e}){return E.jsx(vu.Empty,{"data-slot":"command-empty",className:"py-6 text-center text-sm",...e})}function gte({className:e,...t}){return E.jsx(vu.Group,{"data-slot":"command-group",className:ke("text-foreground [&_[cmdk-group-heading]]:text-muted-foreground overflow-hidden p-1 [&_[cmdk-group-heading]]:px-2 [&_[cmdk-group-heading]]:py-1.5 [&_[cmdk-group-heading]]:text-xs [&_[cmdk-group-heading]]:font-medium",e),...t})}function yte({className:e,...t}){return E.jsx(vu.Item,{"data-slot":"command-item",className:ke("data-[selected=true]:bg-accent data-[selected=true]:text-accent-foreground [&_svg:not([class*='text-'])]:text-muted-foreground relative flex cursor-default items-center gap-2 rounded-sm px-2 py-1.5 text-sm outline-hidden select-none data-[disabled=true]:pointer-events-none data-[disabled=true]:opacity-50 [&_svg]:pointer-events-none [&_svg]:shrink-0 [&_svg:not([class*='size-'])]:size-4",e),...t})}const bte=Ln.object({input_file:Ln.string().min(1,"Please upload a file"),filter_type:Ln.enum(["Genus","Species"]),model:Ln.string(),filter_species:Ln.string().optional(),filter_mode:Ln.enum(["threshold","maximum"]),threshold:Ln.number().min(0).max(1),sparse_sampling:Ln.boolean(),sparse_sampling_step:Ln.number().min(1).max(500)});function xte(){const[e,t]=g.useState({}),[n,r]=g.useState(null),i=cB({resolver:dB(bte),defaultValues:{filter_type:"Species",model:void 0,filter_mode:"threshold",threshold:.7,sparse_sampling:!1,sparse_sampling_step:1}}),[a,s]=g.useState([]);g.useEffect(()=>{cC().then(v=>{t(v)}).catch(v=>{console.error("Error fetching models:",v)})},[]);const u=i.watch("model")!==void 0&&i.watch("filter_type")!==void 0?`${i.watch("model")}-${i.watch("filter_type")}`.toLowerCase():void 0;g.useEffect(()=>{e&&u&&fC(u).then(v=>{r(v)}).catch(v=>{console.error("Error fetching model metadata:",v)})},[e,u,i]);const c=kE(),d=g.useCallback(async(v,{onSuccess:b,onError:S})=>{const w=v[0];try{const{filename:x}=await fq(w);b(w),console.log("File uploaded successfully:",x),i.setValue("input_file",x)}catch(x){S(w,x),console.error("Error uploading file:",x)}},[i]),p=g.useCallback((v,b)=>{console.log(b,{description:`"${v.name.length>20?`${v.name.slice(0,20)}...`:v.name}" has been rejected`})},[]);function m(v){const b=v.filter_mode==="threshold"?v.threshold:-1;DQ(v.filter_type,v.model,v.input_file,b,v.filter_species,v.sparse_sampling_step).then(S=>{console.log("Filtering response:",S),c(`/filter-result/${S.uuid}`)}).catch(S=>{console.error("Error during filtering:",S)})}return E.jsx(PB,{...i,children:E.jsxs("form",{onSubmit:i.handleSubmit(m),className:"space-y-6",children:[E.jsx(sr,{control:i.control,name:"input_file",render:({field:v})=>E.jsxs(lr,{children:[E.jsx(ur,{children:"Input File"}),E.jsx(cr,{children:E.jsxs(vq,{value:a,onValueChange:s,onUpload:d,onFileReject:p,maxFiles:1,multiple:!0,children:[a.length===0&&E.jsxs(hC,{children:[E.jsxs("div",{className:"flex flex-col items-center gap-1 text-center",children:[E.jsx("div",{className:"flex items-center justify-center rounded-full border p-2.5",children:E.jsx(xL,{className:"size-6 text-muted-foreground"})}),E.jsx("p",{className:"font-medium text-sm",children:"Drag & drop file here"}),E.jsx("p",{className:"text-muted-foreground text-xs",children:"Or click to browse"})]}),E.jsx(mC,{asChild:!0,children:E.jsx(Pr,{variant:"outline",size:"sm",className:"mt-2 w-fit",children:"Browse files"})})]}),E.jsx(vC,{children:a.map((b,S)=>E.jsxs(gC,{value:b,className:"flex-col",children:[E.jsxs("div",{className:"flex w-full items-center gap-2",children:[E.jsx(yC,{}),E.jsx(bC,{}),E.jsx(wC,{asChild:!0,children:E.jsx(Pr,{variant:"ghost",size:"icon",className:"size-7",onClick:()=>{s(w=>w.filter((x,C)=>C!==S)),i.setValue("input_file","")},children:E.jsx(wL,{})})})]}),E.jsx(xC,{})]},S))})]})}),E.jsx(fr,{children:"Upload the file you would like to filter."}),E.jsx(Vr,{})]})}),E.jsx(sr,{control:i.control,name:"filter_type",render:({field:v})=>E.jsxs(lr,{children:[E.jsx(ur,{children:"Filter Type"}),E.jsx(cr,{children:E.jsxs(Kh,{defaultValue:"Species",onValueChange:b=>v.onChange(b),children:[E.jsx(Zh,{className:"w-full",children:E.jsx(Yh,{})}),E.jsxs(Xh,{children:[E.jsx(Go,{value:"Genus",children:"Genus"}),E.jsx(Go,{value:"Species",children:"Species"})]})]})}),E.jsx(fr,{children:"Select the type of filter you would like to perform."}),E.jsx(Vr,{})]})}),E.jsx(sr,{control:i.control,name:"model",render:({field:v})=>{var b;return E.jsxs(lr,{children:[E.jsx(ur,{children:"Model"}),E.jsx(cr,{children:E.jsxs(Kh,{onValueChange:v.onChange,children:[E.jsx(Zh,{className:"w-full",children:E.jsx(Yh,{})}),E.jsx(Xh,{children:(b=e[i.watch("filter_type")])==null?void 0:b.map(S=>E.jsx(Go,{value:S,children:S},S))})]})}),E.jsx(fr,{children:"Select the model you would like to use for filtering."}),E.jsx(Vr,{})]})}}),i.watch("filter_type")==="Species"&&E.jsxs(E.Fragment,{children:[i.watch("model")!=null&&E.jsx(sr,{control:i.control,name:"filter_species",render:({field:v})=>E.jsxs(lr,{className:"flex flex-col",children:[E.jsx(ur,{children:"Filter Species"}),E.jsxs(uee,{children:[E.jsx(cee,{asChild:!0,children:E.jsx(cr,{children:E.jsxs(Pr,{variant:"outline",role:"combobox",className:ke("w-full justify-between",!v.value&&"text-muted-foreground"),children:[v.value?n==null?void 0:n.display_names[v.value]:"Select species...",E.jsx(a8,{className:"ml-2 h-4 w-4 shrink-0 opacity-50"})]})})}),E.jsx(fee,{className:"p-0",children:E.jsxs(pte,{children:[E.jsx(hte,{placeholder:"Search species..."}),E.jsxs(mte,{children:[E.jsx(vte,{children:"No species found."}),E.jsx(gte,{children:(n==null?void 0:n.display_names)&&Object.entries(n.display_names).map(([b,S])=>E.jsxs(yte,{value:S,onSelect:()=>{i.setValue("filter_species",b)},children:[S,E.jsx(jE,{className:ke("ml-auto",b===v.value?"opacity-100":"opacity-0")})]},b))})]})]})})]}),E.jsx(fr,{children:"Select the species you would like to filter for."}),E.jsx(Vr,{})]})}),E.jsx(sr,{control:i.control,name:"filter_mode",render:({field:v})=>E.jsxs(lr,{children:[E.jsx(ur,{children:"Filter Mode"}),E.jsx(cr,{children:E.jsx(NJ,{defaultValue:"threshold",onValueChange:v.onChange,children:E.jsxs(DJ,{className:"w-full",children:[E.jsx(QP,{value:"threshold",children:"Filter by threshold"}),E.jsx(QP,{value:"maximum",children:"Filter by maximum scoring species"})]})})}),E.jsx(fr,{children:"Select the filtering mode. You can either filter by threshold or by maximum scoring species, meaning that contigs will only be included if the selected species is the highest-scoring one."}),E.jsx(Vr,{})]})})]}),i.watch("filter_mode")==="threshold"&&E.jsx(sr,{control:i.control,name:"threshold",render:({field:v})=>E.jsxs(lr,{children:[E.jsx(ur,{children:"Threshold"}),E.jsx(cr,{children:E.jsx(T_,{type:"number",value:v.value,defaultValue:.7,step:.01,min:0,max:1,onChange:b=>v.onChange(Number(b.target.value))})}),E.jsx(fr,{children:"Set the threshold for filtering."}),E.jsx(Vr,{})]})}),E.jsx(sr,{control:i.control,name:"sparse_sampling",render:({field:v})=>E.jsxs(lr,{className:"flex flex-row items-center justify-between pr-2",children:[E.jsxs("div",{className:"space-y-0.5",children:[E.jsx(ur,{children:"Sparse Sampling"}),E.jsx(fr,{children:"Enable sparse sampling for filtering."})]}),E.jsx(cr,{children:E.jsx(TB,{checked:v.value,onCheckedChange:v.onChange})})]})}),i.watch("sparse_sampling")&&E.jsx(sr,{control:i.control,name:"sparse_sampling_step",render:({field:v})=>E.jsxs(lr,{children:[E.jsx(ur,{children:"Sparse Sampling Step"}),E.jsx(cr,{children:E.jsx(T_,{type:"number",value:v.value,onChange:b=>v.onChange(Number(b.target.value))})}),E.jsx(fr,{children:"Set the step size for sparse sampling."}),E.jsx(Vr,{})]})}),E.jsx(Pr,{type:"submit",children:"Filter"})]})})}function wte(){return E.jsx("main",{className:"flex-1 flex flex-col items-center justify-center p-4",children:E.jsxs("div",{className:"w-1/2",children:[E.jsx("h1",{className:"text-xl font-bold",children:"Filter your data"}),E.jsx("p",{children:"Upload your data and select parameters to filter it."}),E.jsx(Mr,{className:"my-4"}),E.jsx(xte,{})]})})}/**
   * table-core
   *
   * Copyright (c) TanStack
   *
   * This source code is licensed under the MIT license found in the
   * LICENSE.md file in the root directory of this source tree.
   *
   * @license MIT
   */function eo(e,t){return typeof e=="function"?e(t):e}function gr(e,t){return n=>{t.setState(r=>({...r,[e]:eo(n,r[e])}))}}function Mv(e){return e instanceof Function}function Ste(e){return Array.isArray(e)&&e.every(t=>typeof t=="number")}function _te(e,t){const n=[],r=i=>{i.forEach(a=>{n.push(a);const s=t(a);s!=null&&s.length&&r(s)})};return r(e),n}function Ge(e,t,n){let r=[],i;return a=>{let s;n.key&&n.debug&&(s=Date.now());const u=e(a);if(!(u.length!==r.length||u.some((p,m)=>r[m]!==p)))return i;r=u;let d;if(n.key&&n.debug&&(d=Date.now()),i=t(...u),n==null||n.onChange==null||n.onChange(i),n.key&&n.debug&&n!=null&&n.debug()){const p=Math.round((Date.now()-s)*100)/100,m=Math.round((Date.now()-d)*100)/100,v=m/16,b=(S,w)=>{for(S=String(S);S.length<w;)S=" "+S;return S};console.info(`%c⏱ ${b(m,5)} /${b(p,5)} ms`,`
            font-size: .6rem;
            font-weight: bold;
            color: hsl(${Math.max(0,Math.min(120-120*v,120))}deg 100% 31%);`,n==null?void 0:n.key)}return i}}function We(e,t,n,r){return{debug:()=>{var i;return(i=e==null?void 0:e.debugAll)!=null?i:e[t]},key:!1,onChange:r}}function Ete(e,t,n,r){const i=()=>{var s;return(s=a.getValue())!=null?s:e.options.renderFallbackValue},a={id:`${t.id}_${n.id}`,row:t,column:n,getValue:()=>t.getValue(r),renderValue:i,getContext:Ge(()=>[e,n,t,a],(s,u,c,d)=>({table:s,column:u,row:c,cell:d,getValue:d.getValue,renderValue:d.renderValue}),We(e.options,"debugCells"))};return e._features.forEach(s=>{s.createCell==null||s.createCell(a,n,t,e)},{}),a}function Cte(e,t,n,r){var i,a;const u={...e._getDefaultColumnDef(),...t},c=u.accessorKey;let d=(i=(a=u.id)!=null?a:c?typeof String.prototype.replaceAll=="function"?c.replaceAll(".","_"):c.replace(/\./g,"_"):void 0)!=null?i:typeof u.header=="string"?u.header:void 0,p;if(u.accessorFn?p=u.accessorFn:c&&(c.includes(".")?p=v=>{let b=v;for(const w of c.split(".")){var S;b=(S=b)==null?void 0:S[w]}return b}:p=v=>v[u.accessorKey]),!d)throw new Error;let m={id:`${String(d)}`,accessorFn:p,parent:r,depth:n,columnDef:u,columns:[],getFlatColumns:Ge(()=>[!0],()=>{var v;return[m,...(v=m.columns)==null?void 0:v.flatMap(b=>b.getFlatColumns())]},We(e.options,"debugColumns")),getLeafColumns:Ge(()=>[e._getOrderColumnsFn()],v=>{var b;if((b=m.columns)!=null&&b.length){let S=m.columns.flatMap(w=>w.getLeafColumns());return v(S)}return[m]},We(e.options,"debugColumns"))};for(const v of e._features)v.createColumn==null||v.createColumn(m,e);return m}const Rn="debugHeaders";function oj(e,t,n){var r;let a={id:(r=n.id)!=null?r:t.id,column:t,index:n.index,isPlaceholder:!!n.isPlaceholder,placeholderId:n.placeholderId,depth:n.depth,subHeaders:[],colSpan:0,rowSpan:0,headerGroup:null,getLeafHeaders:()=>{const s=[],u=c=>{c.subHeaders&&c.subHeaders.length&&c.subHeaders.map(u),s.push(c)};return u(a),s},getContext:()=>({table:e,header:a,column:t})};return e._features.forEach(s=>{s.createHeader==null||s.createHeader(a,e)}),a}const Ote={createTable:e=>{e.getHeaderGroups=Ge(()=>[e.getAllColumns(),e.getVisibleLeafColumns(),e.getState().columnPinning.left,e.getState().columnPinning.right],(t,n,r,i)=>{var a,s;const u=(a=r==null?void 0:r.map(m=>n.find(v=>v.id===m)).filter(Boolean))!=null?a:[],c=(s=i==null?void 0:i.map(m=>n.find(v=>v.id===m)).filter(Boolean))!=null?s:[],d=n.filter(m=>!(r!=null&&r.includes(m.id))&&!(i!=null&&i.includes(m.id)));return Jp(t,[...u,...d,...c],e)},We(e.options,Rn)),e.getCenterHeaderGroups=Ge(()=>[e.getAllColumns(),e.getVisibleLeafColumns(),e.getState().columnPinning.left,e.getState().columnPinning.right],(t,n,r,i)=>(n=n.filter(a=>!(r!=null&&r.includes(a.id))&&!(i!=null&&i.includes(a.id))),Jp(t,n,e,"center")),We(e.options,Rn)),e.getLeftHeaderGroups=Ge(()=>[e.getAllColumns(),e.getVisibleLeafColumns(),e.getState().columnPinning.left],(t,n,r)=>{var i;const a=(i=r==null?void 0:r.map(s=>n.find(u=>u.id===s)).filter(Boolean))!=null?i:[];return Jp(t,a,e,"left")},We(e.options,Rn)),e.getRightHeaderGroups=Ge(()=>[e.getAllColumns(),e.getVisibleLeafColumns(),e.getState().columnPinning.right],(t,n,r)=>{var i;const a=(i=r==null?void 0:r.map(s=>n.find(u=>u.id===s)).filter(Boolean))!=null?i:[];return Jp(t,a,e,"right")},We(e.options,Rn)),e.getFooterGroups=Ge(()=>[e.getHeaderGroups()],t=>[...t].reverse(),We(e.options,Rn)),e.getLeftFooterGroups=Ge(()=>[e.getLeftHeaderGroups()],t=>[...t].reverse(),We(e.options,Rn)),e.getCenterFooterGroups=Ge(()=>[e.getCenterHeaderGroups()],t=>[...t].reverse(),We(e.options,Rn)),e.getRightFooterGroups=Ge(()=>[e.getRightHeaderGroups()],t=>[...t].reverse(),We(e.options,Rn)),e.getFlatHeaders=Ge(()=>[e.getHeaderGroups()],t=>t.map(n=>n.headers).flat(),We(e.options,Rn)),e.getLeftFlatHeaders=Ge(()=>[e.getLeftHeaderGroups()],t=>t.map(n=>n.headers).flat(),We(e.options,Rn)),e.getCenterFlatHeaders=Ge(()=>[e.getCenterHeaderGroups()],t=>t.map(n=>n.headers).flat(),We(e.options,Rn)),e.getRightFlatHeaders=Ge(()=>[e.getRightHeaderGroups()],t=>t.map(n=>n.headers).flat(),We(e.options,Rn)),e.getCenterLeafHeaders=Ge(()=>[e.getCenterFlatHeaders()],t=>t.filter(n=>{var r;return!((r=n.subHeaders)!=null&&r.length)}),We(e.options,Rn)),e.getLeftLeafHeaders=Ge(()=>[e.getLeftFlatHeaders()],t=>t.filter(n=>{var r;return!((r=n.subHeaders)!=null&&r.length)}),We(e.options,Rn)),e.getRightLeafHeaders=Ge(()=>[e.getRightFlatHeaders()],t=>t.filter(n=>{var r;return!((r=n.subHeaders)!=null&&r.length)}),We(e.options,Rn)),e.getLeafHeaders=Ge(()=>[e.getLeftHeaderGroups(),e.getCenterHeaderGroups(),e.getRightHeaderGroups()],(t,n,r)=>{var i,a,s,u,c,d;return[...(i=(a=t[0])==null?void 0:a.headers)!=null?i:[],...(s=(u=n[0])==null?void 0:u.headers)!=null?s:[],...(c=(d=r[0])==null?void 0:d.headers)!=null?c:[]].map(p=>p.getLeafHeaders()).flat()},We(e.options,Rn))}};function Jp(e,t,n,r){var i,a;let s=0;const u=function(v,b){b===void 0&&(b=1),s=Math.max(s,b),v.filter(S=>S.getIsVisible()).forEach(S=>{var w;(w=S.columns)!=null&&w.length&&u(S.columns,b+1)},0)};u(e);let c=[];const d=(v,b)=>{const S={depth:b,id:[r,`${b}`].filter(Boolean).join("_"),headers:[]},w=[];v.forEach(x=>{const C=[...w].reverse()[0],O=x.column.depth===S.depth;let A,T=!1;if(O&&x.column.parent?A=x.column.parent:(A=x.column,T=!0),C&&(C==null?void 0:C.column)===A)C.subHeaders.push(x);else{const R=oj(n,A,{id:[r,b,A.id,x==null?void 0:x.id].filter(Boolean).join("_"),isPlaceholder:T,placeholderId:T?`${w.filter(M=>M.column===A).length}`:void 0,depth:b,index:w.length});R.subHeaders.push(x),w.push(R)}S.headers.push(x),x.headerGroup=S}),c.push(S),b>0&&d(w,b-1)},p=t.map((v,b)=>oj(n,v,{depth:s,index:b}));d(p,s-1),c.reverse();const m=v=>v.filter(S=>S.column.getIsVisible()).map(S=>{let w=0,x=0,C=[0];S.subHeaders&&S.subHeaders.length?(C=[],m(S.subHeaders).forEach(A=>{let{colSpan:T,rowSpan:R}=A;w+=T,C.push(R)})):w=1;const O=Math.min(...C);return x=x+O,S.colSpan=w,S.rowSpan=x,{colSpan:w,rowSpan:x}});return m((i=(a=c[0])==null?void 0:a.headers)!=null?i:[]),c}const Ate=(e,t,n,r,i,a,s)=>{let u={id:t,index:r,original:n,depth:i,parentId:s,_valuesCache:{},_uniqueValuesCache:{},getValue:c=>{if(u._valuesCache.hasOwnProperty(c))return u._valuesCache[c];const d=e.getColumn(c);if(d!=null&&d.accessorFn)return u._valuesCache[c]=d.accessorFn(u.original,r),u._valuesCache[c]},getUniqueValues:c=>{if(u._uniqueValuesCache.hasOwnProperty(c))return u._uniqueValuesCache[c];const d=e.getColumn(c);if(d!=null&&d.accessorFn)return d.columnDef.getUniqueValues?(u._uniqueValuesCache[c]=d.columnDef.getUniqueValues(u.original,r),u._uniqueValuesCache[c]):(u._uniqueValuesCache[c]=[u.getValue(c)],u._uniqueValuesCache[c])},renderValue:c=>{var d;return(d=u.getValue(c))!=null?d:e.options.renderFallbackValue},subRows:[],getLeafRows:()=>_te(u.subRows,c=>c.subRows),getParentRow:()=>u.parentId?e.getRow(u.parentId,!0):void 0,getParentRows:()=>{let c=[],d=u;for(;;){const p=d.getParentRow();if(!p)break;c.push(p),d=p}return c.reverse()},getAllCells:Ge(()=>[e.getAllLeafColumns()],c=>c.map(d=>Ete(e,u,d,d.id)),We(e.options,"debugRows")),_getAllCellsByColumnId:Ge(()=>[u.getAllCells()],c=>c.reduce((d,p)=>(d[p.column.id]=p,d),{}),We(e.options,"debugRows"))};for(let c=0;c<e._features.length;c++){const d=e._features[c];d==null||d.createRow==null||d.createRow(u,e)}return u},Rte={createColumn:(e,t)=>{e._getFacetedRowModel=t.options.getFacetedRowModel&&t.options.getFacetedRowModel(t,e.id),e.getFacetedRowModel=()=>e._getFacetedRowModel?e._getFacetedRowModel():t.getPreFilteredRowModel(),e._getFacetedUniqueValues=t.options.getFacetedUniqueValues&&t.options.getFacetedUniqueValues(t,e.id),e.getFacetedUniqueValues=()=>e._getFacetedUniqueValues?e._getFacetedUniqueValues():new Map,e._getFacetedMinMaxValues=t.options.getFacetedMinMaxValues&&t.options.getFacetedMinMaxValues(t,e.id),e.getFacetedMinMaxValues=()=>{if(e._getFacetedMinMaxValues)return e._getFacetedMinMaxValues()}}},bz=(e,t,n)=>{var r,i;const a=n==null||(r=n.toString())==null?void 0:r.toLowerCase();return!!(!((i=e.getValue(t))==null||(i=i.toString())==null||(i=i.toLowerCase())==null)&&i.includes(a))};bz.autoRemove=e=>Zr(e);const xz=(e,t,n)=>{var r;return!!(!((r=e.getValue(t))==null||(r=r.toString())==null)&&r.includes(n))};xz.autoRemove=e=>Zr(e);const wz=(e,t,n)=>{var r;return((r=e.getValue(t))==null||(r=r.toString())==null?void 0:r.toLowerCase())===(n==null?void 0:n.toLowerCase())};wz.autoRemove=e=>Zr(e);const Sz=(e,t,n)=>{var r;return(r=e.getValue(t))==null?void 0:r.includes(n)};Sz.autoRemove=e=>Zr(e);const _z=(e,t,n)=>!n.some(r=>{var i;return!((i=e.getValue(t))!=null&&i.includes(r))});_z.autoRemove=e=>Zr(e)||!(e!=null&&e.length);const Ez=(e,t,n)=>n.some(r=>{var i;return(i=e.getValue(t))==null?void 0:i.includes(r)});Ez.autoRemove=e=>Zr(e)||!(e!=null&&e.length);const Cz=(e,t,n)=>e.getValue(t)===n;Cz.autoRemove=e=>Zr(e);const Oz=(e,t,n)=>e.getValue(t)==n;Oz.autoRemove=e=>Zr(e);const RC=(e,t,n)=>{let[r,i]=n;const a=e.getValue(t);return a>=r&&a<=i};RC.resolveFilterValue=e=>{let[t,n]=e,r=typeof t!="number"?parseFloat(t):t,i=typeof n!="number"?parseFloat(n):n,a=t===null||Number.isNaN(r)?-1/0:r,s=n===null||Number.isNaN(i)?1/0:i;if(a>s){const u=a;a=s,s=u}return[a,s]};RC.autoRemove=e=>Zr(e)||Zr(e[0])&&Zr(e[1]);const Qi={includesString:bz,includesStringSensitive:xz,equalsString:wz,arrIncludes:Sz,arrIncludesAll:_z,arrIncludesSome:Ez,equals:Cz,weakEquals:Oz,inNumberRange:RC};function Zr(e){return e==null||e===""}const Tte={getDefaultColumnDef:()=>({filterFn:"auto"}),getInitialState:e=>({columnFilters:[],...e}),getDefaultOptions:e=>({onColumnFiltersChange:gr("columnFilters",e),filterFromLeafRows:!1,maxLeafRowFilterDepth:100}),createColumn:(e,t)=>{e.getAutoFilterFn=()=>{const n=t.getCoreRowModel().flatRows[0],r=n==null?void 0:n.getValue(e.id);return typeof r=="string"?Qi.includesString:typeof r=="number"?Qi.inNumberRange:typeof r=="boolean"||r!==null&&typeof r=="object"?Qi.equals:Array.isArray(r)?Qi.arrIncludes:Qi.weakEquals},e.getFilterFn=()=>{var n,r;return Mv(e.columnDef.filterFn)?e.columnDef.filterFn:e.columnDef.filterFn==="auto"?e.getAutoFilterFn():(n=(r=t.options.filterFns)==null?void 0:r[e.columnDef.filterFn])!=null?n:Qi[e.columnDef.filterFn]},e.getCanFilter=()=>{var n,r,i;return((n=e.columnDef.enableColumnFilter)!=null?n:!0)&&((r=t.options.enableColumnFilters)!=null?r:!0)&&((i=t.options.enableFilters)!=null?i:!0)&&!!e.accessorFn},e.getIsFiltered=()=>e.getFilterIndex()>-1,e.getFilterValue=()=>{var n;return(n=t.getState().columnFilters)==null||(n=n.find(r=>r.id===e.id))==null?void 0:n.value},e.getFilterIndex=()=>{var n,r;return(n=(r=t.getState().columnFilters)==null?void 0:r.findIndex(i=>i.id===e.id))!=null?n:-1},e.setFilterValue=n=>{t.setColumnFilters(r=>{const i=e.getFilterFn(),a=r==null?void 0:r.find(p=>p.id===e.id),s=eo(n,a?a.value:void 0);if(sj(i,s,e)){var u;return(u=r==null?void 0:r.filter(p=>p.id!==e.id))!=null?u:[]}const c={id:e.id,value:s};if(a){var d;return(d=r==null?void 0:r.map(p=>p.id===e.id?c:p))!=null?d:[]}return r!=null&&r.length?[...r,c]:[c]})}},createRow:(e,t)=>{e.columnFilters={},e.columnFiltersMeta={}},createTable:e=>{e.setColumnFilters=t=>{const n=e.getAllLeafColumns(),r=i=>{var a;return(a=eo(t,i))==null?void 0:a.filter(s=>{const u=n.find(c=>c.id===s.id);if(u){const c=u.getFilterFn();if(sj(c,s.value,u))return!1}return!0})};e.options.onColumnFiltersChange==null||e.options.onColumnFiltersChange(r)},e.resetColumnFilters=t=>{var n,r;e.setColumnFilters(t?[]:(n=(r=e.initialState)==null?void 0:r.columnFilters)!=null?n:[])},e.getPreFilteredRowModel=()=>e.getCoreRowModel(),e.getFilteredRowModel=()=>(!e._getFilteredRowModel&&e.options.getFilteredRowModel&&(e._getFilteredRowModel=e.options.getFilteredRowModel(e)),e.options.manualFiltering||!e._getFilteredRowModel?e.getPreFilteredRowModel():e._getFilteredRowModel())}};function sj(e,t,n){return(e&&e.autoRemove?e.autoRemove(t,n):!1)||typeof t>"u"||typeof t=="string"&&!t}const Mte=(e,t,n)=>n.reduce((r,i)=>{const a=i.getValue(e);return r+(typeof a=="number"?a:0)},0),Pte=(e,t,n)=>{let r;return n.forEach(i=>{const a=i.getValue(e);a!=null&&(r>a||r===void 0&&a>=a)&&(r=a)}),r},jte=(e,t,n)=>{let r;return n.forEach(i=>{const a=i.getValue(e);a!=null&&(r<a||r===void 0&&a>=a)&&(r=a)}),r},Nte=(e,t,n)=>{let r,i;return n.forEach(a=>{const s=a.getValue(e);s!=null&&(r===void 0?s>=s&&(r=i=s):(r>s&&(r=s),i<s&&(i=s)))}),[r,i]},Dte=(e,t)=>{let n=0,r=0;if(t.forEach(i=>{let a=i.getValue(e);a!=null&&(a=+a)>=a&&(++n,r+=a)}),n)return r/n},kte=(e,t)=>{if(!t.length)return;const n=t.map(a=>a.getValue(e));if(!Ste(n))return;if(n.length===1)return n[0];const r=Math.floor(n.length/2),i=n.sort((a,s)=>a-s);return n.length%2!==0?i[r]:(i[r-1]+i[r])/2},$te=(e,t)=>Array.from(new Set(t.map(n=>n.getValue(e))).values()),Ite=(e,t)=>new Set(t.map(n=>n.getValue(e))).size,Lte=(e,t)=>t.length,n0={sum:Mte,min:Pte,max:jte,extent:Nte,mean:Dte,median:kte,unique:$te,uniqueCount:Ite,count:Lte},Bte={getDefaultColumnDef:()=>({aggregatedCell:e=>{var t,n;return(t=(n=e.getValue())==null||n.toString==null?void 0:n.toString())!=null?t:null},aggregationFn:"auto"}),getInitialState:e=>({grouping:[],...e}),getDefaultOptions:e=>({onGroupingChange:gr("grouping",e),groupedColumnMode:"reorder"}),createColumn:(e,t)=>{e.toggleGrouping=()=>{t.setGrouping(n=>n!=null&&n.includes(e.id)?n.filter(r=>r!==e.id):[...n??[],e.id])},e.getCanGroup=()=>{var n,r;return((n=e.columnDef.enableGrouping)!=null?n:!0)&&((r=t.options.enableGrouping)!=null?r:!0)&&(!!e.accessorFn||!!e.columnDef.getGroupingValue)},e.getIsGrouped=()=>{var n;return(n=t.getState().grouping)==null?void 0:n.includes(e.id)},e.getGroupedIndex=()=>{var n;return(n=t.getState().grouping)==null?void 0:n.indexOf(e.id)},e.getToggleGroupingHandler=()=>{const n=e.getCanGroup();return()=>{n&&e.toggleGrouping()}},e.getAutoAggregationFn=()=>{const n=t.getCoreRowModel().flatRows[0],r=n==null?void 0:n.getValue(e.id);if(typeof r=="number")return n0.sum;if(Object.prototype.toString.call(r)==="[object Date]")return n0.extent},e.getAggregationFn=()=>{var n,r;if(!e)throw new Error;return Mv(e.columnDef.aggregationFn)?e.columnDef.aggregationFn:e.columnDef.aggregationFn==="auto"?e.getAutoAggregationFn():(n=(r=t.options.aggregationFns)==null?void 0:r[e.columnDef.aggregationFn])!=null?n:n0[e.columnDef.aggregationFn]}},createTable:e=>{e.setGrouping=t=>e.options.onGroupingChange==null?void 0:e.options.onGroupingChange(t),e.resetGrouping=t=>{var n,r;e.setGrouping(t?[]:(n=(r=e.initialState)==null?void 0:r.grouping)!=null?n:[])},e.getPreGroupedRowModel=()=>e.getFilteredRowModel(),e.getGroupedRowModel=()=>(!e._getGroupedRowModel&&e.options.getGroupedRowModel&&(e._getGroupedRowModel=e.options.getGroupedRowModel(e)),e.options.manualGrouping||!e._getGroupedRowModel?e.getPreGroupedRowModel():e._getGroupedRowModel())},createRow:(e,t)=>{e.getIsGrouped=()=>!!e.groupingColumnId,e.getGroupingValue=n=>{if(e._groupingValuesCache.hasOwnProperty(n))return e._groupingValuesCache[n];const r=t.getColumn(n);return r!=null&&r.columnDef.getGroupingValue?(e._groupingValuesCache[n]=r.columnDef.getGroupingValue(e.original),e._groupingValuesCache[n]):e.getValue(n)},e._groupingValuesCache={}},createCell:(e,t,n,r)=>{e.getIsGrouped=()=>t.getIsGrouped()&&t.id===n.groupingColumnId,e.getIsPlaceholder=()=>!e.getIsGrouped()&&t.getIsGrouped(),e.getIsAggregated=()=>{var i;return!e.getIsGrouped()&&!e.getIsPlaceholder()&&!!((i=n.subRows)!=null&&i.length)}}};function Fte(e,t,n){if(!(t!=null&&t.length)||!n)return e;const r=e.filter(a=>!t.includes(a.id));return n==="remove"?r:[...t.map(a=>e.find(s=>s.id===a)).filter(Boolean),...r]}const qte={getInitialState:e=>({columnOrder:[],...e}),getDefaultOptions:e=>({onColumnOrderChange:gr("columnOrder",e)}),createColumn:(e,t)=>{e.getIndex=Ge(n=>[Wc(t,n)],n=>n.findIndex(r=>r.id===e.id),We(t.options,"debugColumns")),e.getIsFirstColumn=n=>{var r;return((r=Wc(t,n)[0])==null?void 0:r.id)===e.id},e.getIsLastColumn=n=>{var r;const i=Wc(t,n);return((r=i[i.length-1])==null?void 0:r.id)===e.id}},createTable:e=>{e.setColumnOrder=t=>e.options.onColumnOrderChange==null?void 0:e.options.onColumnOrderChange(t),e.resetColumnOrder=t=>{var n;e.setColumnOrder(t?[]:(n=e.initialState.columnOrder)!=null?n:[])},e._getOrderColumnsFn=Ge(()=>[e.getState().columnOrder,e.getState().grouping,e.options.groupedColumnMode],(t,n,r)=>i=>{let a=[];if(!(t!=null&&t.length))a=i;else{const s=[...t],u=[...i];for(;u.length&&s.length;){const c=s.shift(),d=u.findIndex(p=>p.id===c);d>-1&&a.push(u.splice(d,1)[0])}a=[...a,...u]}return Fte(a,n,r)},We(e.options,"debugTable"))}},r0=()=>({left:[],right:[]}),zte={getInitialState:e=>({columnPinning:r0(),...e}),getDefaultOptions:e=>({onColumnPinningChange:gr("columnPinning",e)}),createColumn:(e,t)=>{e.pin=n=>{const r=e.getLeafColumns().map(i=>i.id).filter(Boolean);t.setColumnPinning(i=>{var a,s;if(n==="right"){var u,c;return{left:((u=i==null?void 0:i.left)!=null?u:[]).filter(m=>!(r!=null&&r.includes(m))),right:[...((c=i==null?void 0:i.right)!=null?c:[]).filter(m=>!(r!=null&&r.includes(m))),...r]}}if(n==="left"){var d,p;return{left:[...((d=i==null?void 0:i.left)!=null?d:[]).filter(m=>!(r!=null&&r.includes(m))),...r],right:((p=i==null?void 0:i.right)!=null?p:[]).filter(m=>!(r!=null&&r.includes(m)))}}return{left:((a=i==null?void 0:i.left)!=null?a:[]).filter(m=>!(r!=null&&r.includes(m))),right:((s=i==null?void 0:i.right)!=null?s:[]).filter(m=>!(r!=null&&r.includes(m)))}})},e.getCanPin=()=>e.getLeafColumns().some(r=>{var i,a,s;return((i=r.columnDef.enablePinning)!=null?i:!0)&&((a=(s=t.options.enableColumnPinning)!=null?s:t.options.enablePinning)!=null?a:!0)}),e.getIsPinned=()=>{const n=e.getLeafColumns().map(u=>u.id),{left:r,right:i}=t.getState().columnPinning,a=n.some(u=>r==null?void 0:r.includes(u)),s=n.some(u=>i==null?void 0:i.includes(u));return a?"left":s?"right":!1},e.getPinnedIndex=()=>{var n,r;const i=e.getIsPinned();return i?(n=(r=t.getState().columnPinning)==null||(r=r[i])==null?void 0:r.indexOf(e.id))!=null?n:-1:0}},createRow:(e,t)=>{e.getCenterVisibleCells=Ge(()=>[e._getAllVisibleCells(),t.getState().columnPinning.left,t.getState().columnPinning.right],(n,r,i)=>{const a=[...r??[],...i??[]];return n.filter(s=>!a.includes(s.column.id))},We(t.options,"debugRows")),e.getLeftVisibleCells=Ge(()=>[e._getAllVisibleCells(),t.getState().columnPinning.left],(n,r)=>(r??[]).map(a=>n.find(s=>s.column.id===a)).filter(Boolean).map(a=>({...a,position:"left"})),We(t.options,"debugRows")),e.getRightVisibleCells=Ge(()=>[e._getAllVisibleCells(),t.getState().columnPinning.right],(n,r)=>(r??[]).map(a=>n.find(s=>s.column.id===a)).filter(Boolean).map(a=>({...a,position:"right"})),We(t.options,"debugRows"))},createTable:e=>{e.setColumnPinning=t=>e.options.onColumnPinningChange==null?void 0:e.options.onColumnPinningChange(t),e.resetColumnPinning=t=>{var n,r;return e.setColumnPinning(t?r0():(n=(r=e.initialState)==null?void 0:r.columnPinning)!=null?n:r0())},e.getIsSomeColumnsPinned=t=>{var n;const r=e.getState().columnPinning;if(!t){var i,a;return!!((i=r.left)!=null&&i.length||(a=r.right)!=null&&a.length)}return!!((n=r[t])!=null&&n.length)},e.getLeftLeafColumns=Ge(()=>[e.getAllLeafColumns(),e.getState().columnPinning.left],(t,n)=>(n??[]).map(r=>t.find(i=>i.id===r)).filter(Boolean),We(e.options,"debugColumns")),e.getRightLeafColumns=Ge(()=>[e.getAllLeafColumns(),e.getState().columnPinning.right],(t,n)=>(n??[]).map(r=>t.find(i=>i.id===r)).filter(Boolean),We(e.options,"debugColumns")),e.getCenterLeafColumns=Ge(()=>[e.getAllLeafColumns(),e.getState().columnPinning.left,e.getState().columnPinning.right],(t,n,r)=>{const i=[...n??[],...r??[]];return t.filter(a=>!i.includes(a.id))},We(e.options,"debugColumns"))}};function Ute(e){return e||(typeof document<"u"?document:null)}const eh={size:150,minSize:20,maxSize:Number.MAX_SAFE_INTEGER},i0=()=>({startOffset:null,startSize:null,deltaOffset:null,deltaPercentage:null,isResizingColumn:!1,columnSizingStart:[]}),Vte={getDefaultColumnDef:()=>eh,getInitialState:e=>({columnSizing:{},columnSizingInfo:i0(),...e}),getDefaultOptions:e=>({columnResizeMode:"onEnd",columnResizeDirection:"ltr",onColumnSizingChange:gr("columnSizing",e),onColumnSizingInfoChange:gr("columnSizingInfo",e)}),createColumn:(e,t)=>{e.getSize=()=>{var n,r,i;const a=t.getState().columnSizing[e.id];return Math.min(Math.max((n=e.columnDef.minSize)!=null?n:eh.minSize,(r=a??e.columnDef.size)!=null?r:eh.size),(i=e.columnDef.maxSize)!=null?i:eh.maxSize)},e.getStart=Ge(n=>[n,Wc(t,n),t.getState().columnSizing],(n,r)=>r.slice(0,e.getIndex(n)).reduce((i,a)=>i+a.getSize(),0),We(t.options,"debugColumns")),e.getAfter=Ge(n=>[n,Wc(t,n),t.getState().columnSizing],(n,r)=>r.slice(e.getIndex(n)+1).reduce((i,a)=>i+a.getSize(),0),We(t.options,"debugColumns")),e.resetSize=()=>{t.setColumnSizing(n=>{let{[e.id]:r,...i}=n;return i})},e.getCanResize=()=>{var n,r;return((n=e.columnDef.enableResizing)!=null?n:!0)&&((r=t.options.enableColumnResizing)!=null?r:!0)},e.getIsResizing=()=>t.getState().columnSizingInfo.isResizingColumn===e.id},createHeader:(e,t)=>{e.getSize=()=>{let n=0;const r=i=>{if(i.subHeaders.length)i.subHeaders.forEach(r);else{var a;n+=(a=i.column.getSize())!=null?a:0}};return r(e),n},e.getStart=()=>{if(e.index>0){const n=e.headerGroup.headers[e.index-1];return n.getStart()+n.getSize()}return 0},e.getResizeHandler=n=>{const r=t.getColumn(e.column.id),i=r==null?void 0:r.getCanResize();return a=>{if(!r||!i||(a.persist==null||a.persist(),a0(a)&&a.touches&&a.touches.length>1))return;const s=e.getSize(),u=e?e.getLeafHeaders().map(C=>[C.column.id,C.column.getSize()]):[[r.id,r.getSize()]],c=a0(a)?Math.round(a.touches[0].clientX):a.clientX,d={},p=(C,O)=>{typeof O=="number"&&(t.setColumnSizingInfo(A=>{var T,R;const M=t.options.columnResizeDirection==="rtl"?-1:1,P=(O-((T=A==null?void 0:A.startOffset)!=null?T:0))*M,N=Math.max(P/((R=A==null?void 0:A.startSize)!=null?R:0),-.999999);return A.columnSizingStart.forEach(k=>{let[H,L]=k;d[H]=Math.round(Math.max(L+L*N,0)*100)/100}),{...A,deltaOffset:P,deltaPercentage:N}}),(t.options.columnResizeMode==="onChange"||C==="end")&&t.setColumnSizing(A=>({...A,...d})))},m=C=>p("move",C),v=C=>{p("end",C),t.setColumnSizingInfo(O=>({...O,isResizingColumn:!1,startOffset:null,startSize:null,deltaOffset:null,deltaPercentage:null,columnSizingStart:[]}))},b=Ute(n),S={moveHandler:C=>m(C.clientX),upHandler:C=>{b==null||b.removeEventListener("mousemove",S.moveHandler),b==null||b.removeEventListener("mouseup",S.upHandler),v(C.clientX)}},w={moveHandler:C=>(C.cancelable&&(C.preventDefault(),C.stopPropagation()),m(C.touches[0].clientX),!1),upHandler:C=>{var O;b==null||b.removeEventListener("touchmove",w.moveHandler),b==null||b.removeEventListener("touchend",w.upHandler),C.cancelable&&(C.preventDefault(),C.stopPropagation()),v((O=C.touches[0])==null?void 0:O.clientX)}},x=Hte()?{passive:!1}:!1;a0(a)?(b==null||b.addEventListener("touchmove",w.moveHandler,x),b==null||b.addEventListener("touchend",w.upHandler,x)):(b==null||b.addEventListener("mousemove",S.moveHandler,x),b==null||b.addEventListener("mouseup",S.upHandler,x)),t.setColumnSizingInfo(C=>({...C,startOffset:c,startSize:s,deltaOffset:0,deltaPercentage:0,columnSizingStart:u,isResizingColumn:r.id}))}}},createTable:e=>{e.setColumnSizing=t=>e.options.onColumnSizingChange==null?void 0:e.options.onColumnSizingChange(t),e.setColumnSizingInfo=t=>e.options.onColumnSizingInfoChange==null?void 0:e.options.onColumnSizingInfoChange(t),e.resetColumnSizing=t=>{var n;e.setColumnSizing(t?{}:(n=e.initialState.columnSizing)!=null?n:{})},e.resetHeaderSizeInfo=t=>{var n;e.setColumnSizingInfo(t?i0():(n=e.initialState.columnSizingInfo)!=null?n:i0())},e.getTotalSize=()=>{var t,n;return(t=(n=e.getHeaderGroups()[0])==null?void 0:n.headers.reduce((r,i)=>r+i.getSize(),0))!=null?t:0},e.getLeftTotalSize=()=>{var t,n;return(t=(n=e.getLeftHeaderGroups()[0])==null?void 0:n.headers.reduce((r,i)=>r+i.getSize(),0))!=null?t:0},e.getCenterTotalSize=()=>{var t,n;return(t=(n=e.getCenterHeaderGroups()[0])==null?void 0:n.headers.reduce((r,i)=>r+i.getSize(),0))!=null?t:0},e.getRightTotalSize=()=>{var t,n;return(t=(n=e.getRightHeaderGroups()[0])==null?void 0:n.headers.reduce((r,i)=>r+i.getSize(),0))!=null?t:0}}};let th=null;function Hte(){if(typeof th=="boolean")return th;let e=!1;try{const t={get passive(){return e=!0,!1}},n=()=>{};window.addEventListener("test",n,t),window.removeEventListener("test",n)}catch{e=!1}return th=e,th}function a0(e){return e.type==="touchstart"}const Gte={getInitialState:e=>({columnVisibility:{},...e}),getDefaultOptions:e=>({onColumnVisibilityChange:gr("columnVisibility",e)}),createColumn:(e,t)=>{e.toggleVisibility=n=>{e.getCanHide()&&t.setColumnVisibility(r=>({...r,[e.id]:n??!e.getIsVisible()}))},e.getIsVisible=()=>{var n,r;const i=e.columns;return(n=i.length?i.some(a=>a.getIsVisible()):(r=t.getState().columnVisibility)==null?void 0:r[e.id])!=null?n:!0},e.getCanHide=()=>{var n,r;return((n=e.columnDef.enableHiding)!=null?n:!0)&&((r=t.options.enableHiding)!=null?r:!0)},e.getToggleVisibilityHandler=()=>n=>{e.toggleVisibility==null||e.toggleVisibility(n.target.checked)}},createRow:(e,t)=>{e._getAllVisibleCells=Ge(()=>[e.getAllCells(),t.getState().columnVisibility],n=>n.filter(r=>r.column.getIsVisible()),We(t.options,"debugRows")),e.getVisibleCells=Ge(()=>[e.getLeftVisibleCells(),e.getCenterVisibleCells(),e.getRightVisibleCells()],(n,r,i)=>[...n,...r,...i],We(t.options,"debugRows"))},createTable:e=>{const t=(n,r)=>Ge(()=>[r(),r().filter(i=>i.getIsVisible()).map(i=>i.id).join("_")],i=>i.filter(a=>a.getIsVisible==null?void 0:a.getIsVisible()),We(e.options,"debugColumns"));e.getVisibleFlatColumns=t("getVisibleFlatColumns",()=>e.getAllFlatColumns()),e.getVisibleLeafColumns=t("getVisibleLeafColumns",()=>e.getAllLeafColumns()),e.getLeftVisibleLeafColumns=t("getLeftVisibleLeafColumns",()=>e.getLeftLeafColumns()),e.getRightVisibleLeafColumns=t("getRightVisibleLeafColumns",()=>e.getRightLeafColumns()),e.getCenterVisibleLeafColumns=t("getCenterVisibleLeafColumns",()=>e.getCenterLeafColumns()),e.setColumnVisibility=n=>e.options.onColumnVisibilityChange==null?void 0:e.options.onColumnVisibilityChange(n),e.resetColumnVisibility=n=>{var r;e.setColumnVisibility(n?{}:(r=e.initialState.columnVisibility)!=null?r:{})},e.toggleAllColumnsVisible=n=>{var r;n=(r=n)!=null?r:!e.getIsAllColumnsVisible(),e.setColumnVisibility(e.getAllLeafColumns().reduce((i,a)=>({...i,[a.id]:n||!(a.getCanHide!=null&&a.getCanHide())}),{}))},e.getIsAllColumnsVisible=()=>!e.getAllLeafColumns().some(n=>!(n.getIsVisible!=null&&n.getIsVisible())),e.getIsSomeColumnsVisible=()=>e.getAllLeafColumns().some(n=>n.getIsVisible==null?void 0:n.getIsVisible()),e.getToggleAllColumnsVisibilityHandler=()=>n=>{var r;e.toggleAllColumnsVisible((r=n.target)==null?void 0:r.checked)}}};function Wc(e,t){return t?t==="center"?e.getCenterVisibleLeafColumns():t==="left"?e.getLeftVisibleLeafColumns():e.getRightVisibleLeafColumns():e.getVisibleLeafColumns()}const Wte={createTable:e=>{e._getGlobalFacetedRowModel=e.options.getFacetedRowModel&&e.options.getFacetedRowModel(e,"__global__"),e.getGlobalFacetedRowModel=()=>e.options.manualFiltering||!e._getGlobalFacetedRowModel?e.getPreFilteredRowModel():e._getGlobalFacetedRowModel(),e._getGlobalFacetedUniqueValues=e.options.getFacetedUniqueValues&&e.options.getFacetedUniqueValues(e,"__global__"),e.getGlobalFacetedUniqueValues=()=>e._getGlobalFacetedUniqueValues?e._getGlobalFacetedUniqueValues():new Map,e._getGlobalFacetedMinMaxValues=e.options.getFacetedMinMaxValues&&e.options.getFacetedMinMaxValues(e,"__global__"),e.getGlobalFacetedMinMaxValues=()=>{if(e._getGlobalFacetedMinMaxValues)return e._getGlobalFacetedMinMaxValues()}}},Kte={getInitialState:e=>({globalFilter:void 0,...e}),getDefaultOptions:e=>({onGlobalFilterChange:gr("globalFilter",e),globalFilterFn:"auto",getColumnCanGlobalFilter:t=>{var n;const r=(n=e.getCoreRowModel().flatRows[0])==null||(n=n._getAllCellsByColumnId()[t.id])==null?void 0:n.getValue();return typeof r=="string"||typeof r=="number"}}),createColumn:(e,t)=>{e.getCanGlobalFilter=()=>{var n,r,i,a;return((n=e.columnDef.enableGlobalFilter)!=null?n:!0)&&((r=t.options.enableGlobalFilter)!=null?r:!0)&&((i=t.options.enableFilters)!=null?i:!0)&&((a=t.options.getColumnCanGlobalFilter==null?void 0:t.options.getColumnCanGlobalFilter(e))!=null?a:!0)&&!!e.accessorFn}},createTable:e=>{e.getGlobalAutoFilterFn=()=>Qi.includesString,e.getGlobalFilterFn=()=>{var t,n;const{globalFilterFn:r}=e.options;return Mv(r)?r:r==="auto"?e.getGlobalAutoFilterFn():(t=(n=e.options.filterFns)==null?void 0:n[r])!=null?t:Qi[r]},e.setGlobalFilter=t=>{e.options.onGlobalFilterChange==null||e.options.onGlobalFilterChange(t)},e.resetGlobalFilter=t=>{e.setGlobalFilter(t?void 0:e.initialState.globalFilter)}}},Yte={getInitialState:e=>({expanded:{},...e}),getDefaultOptions:e=>({onExpandedChange:gr("expanded",e),paginateExpandedRows:!0}),createTable:e=>{let t=!1,n=!1;e._autoResetExpanded=()=>{var r,i;if(!t){e._queue(()=>{t=!0});return}if((r=(i=e.options.autoResetAll)!=null?i:e.options.autoResetExpanded)!=null?r:!e.options.manualExpanding){if(n)return;n=!0,e._queue(()=>{e.resetExpanded(),n=!1})}},e.setExpanded=r=>e.options.onExpandedChange==null?void 0:e.options.onExpandedChange(r),e.toggleAllRowsExpanded=r=>{r??!e.getIsAllRowsExpanded()?e.setExpanded(!0):e.setExpanded({})},e.resetExpanded=r=>{var i,a;e.setExpanded(r?{}:(i=(a=e.initialState)==null?void 0:a.expanded)!=null?i:{})},e.getCanSomeRowsExpand=()=>e.getPrePaginationRowModel().flatRows.some(r=>r.getCanExpand()),e.getToggleAllRowsExpandedHandler=()=>r=>{r.persist==null||r.persist(),e.toggleAllRowsExpanded()},e.getIsSomeRowsExpanded=()=>{const r=e.getState().expanded;return r===!0||Object.values(r).some(Boolean)},e.getIsAllRowsExpanded=()=>{const r=e.getState().expanded;return typeof r=="boolean"?r===!0:!(!Object.keys(r).length||e.getRowModel().flatRows.some(i=>!i.getIsExpanded()))},e.getExpandedDepth=()=>{let r=0;return(e.getState().expanded===!0?Object.keys(e.getRowModel().rowsById):Object.keys(e.getState().expanded)).forEach(a=>{const s=a.split(".");r=Math.max(r,s.length)}),r},e.getPreExpandedRowModel=()=>e.getSortedRowModel(),e.getExpandedRowModel=()=>(!e._getExpandedRowModel&&e.options.getExpandedRowModel&&(e._getExpandedRowModel=e.options.getExpandedRowModel(e)),e.options.manualExpanding||!e._getExpandedRowModel?e.getPreExpandedRowModel():e._getExpandedRowModel())},createRow:(e,t)=>{e.toggleExpanded=n=>{t.setExpanded(r=>{var i;const a=r===!0?!0:!!(r!=null&&r[e.id]);let s={};if(r===!0?Object.keys(t.getRowModel().rowsById).forEach(u=>{s[u]=!0}):s=r,n=(i=n)!=null?i:!a,!a&&n)return{...s,[e.id]:!0};if(a&&!n){const{[e.id]:u,...c}=s;return c}return r})},e.getIsExpanded=()=>{var n;const r=t.getState().expanded;return!!((n=t.options.getIsRowExpanded==null?void 0:t.options.getIsRowExpanded(e))!=null?n:r===!0||r!=null&&r[e.id])},e.getCanExpand=()=>{var n,r,i;return(n=t.options.getRowCanExpand==null?void 0:t.options.getRowCanExpand(e))!=null?n:((r=t.options.enableExpanding)!=null?r:!0)&&!!((i=e.subRows)!=null&&i.length)},e.getIsAllParentsExpanded=()=>{let n=!0,r=e;for(;n&&r.parentId;)r=t.getRow(r.parentId,!0),n=r.getIsExpanded();return n},e.getToggleExpandedHandler=()=>{const n=e.getCanExpand();return()=>{n&&e.toggleExpanded()}}}},J_=0,e1=10,o0=()=>({pageIndex:J_,pageSize:e1}),Zte={getInitialState:e=>({...e,pagination:{...o0(),...e==null?void 0:e.pagination}}),getDefaultOptions:e=>({onPaginationChange:gr("pagination",e)}),createTable:e=>{let t=!1,n=!1;e._autoResetPageIndex=()=>{var r,i;if(!t){e._queue(()=>{t=!0});return}if((r=(i=e.options.autoResetAll)!=null?i:e.options.autoResetPageIndex)!=null?r:!e.options.manualPagination){if(n)return;n=!0,e._queue(()=>{e.resetPageIndex(),n=!1})}},e.setPagination=r=>{const i=a=>eo(r,a);return e.options.onPaginationChange==null?void 0:e.options.onPaginationChange(i)},e.resetPagination=r=>{var i;e.setPagination(r?o0():(i=e.initialState.pagination)!=null?i:o0())},e.setPageIndex=r=>{e.setPagination(i=>{let a=eo(r,i.pageIndex);const s=typeof e.options.pageCount>"u"||e.options.pageCount===-1?Number.MAX_SAFE_INTEGER:e.options.pageCount-1;return a=Math.max(0,Math.min(a,s)),{...i,pageIndex:a}})},e.resetPageIndex=r=>{var i,a;e.setPageIndex(r?J_:(i=(a=e.initialState)==null||(a=a.pagination)==null?void 0:a.pageIndex)!=null?i:J_)},e.resetPageSize=r=>{var i,a;e.setPageSize(r?e1:(i=(a=e.initialState)==null||(a=a.pagination)==null?void 0:a.pageSize)!=null?i:e1)},e.setPageSize=r=>{e.setPagination(i=>{const a=Math.max(1,eo(r,i.pageSize)),s=i.pageSize*i.pageIndex,u=Math.floor(s/a);return{...i,pageIndex:u,pageSize:a}})},e.setPageCount=r=>e.setPagination(i=>{var a;let s=eo(r,(a=e.options.pageCount)!=null?a:-1);return typeof s=="number"&&(s=Math.max(-1,s)),{...i,pageCount:s}}),e.getPageOptions=Ge(()=>[e.getPageCount()],r=>{let i=[];return r&&r>0&&(i=[...new Array(r)].fill(null).map((a,s)=>s)),i},We(e.options,"debugTable")),e.getCanPreviousPage=()=>e.getState().pagination.pageIndex>0,e.getCanNextPage=()=>{const{pageIndex:r}=e.getState().pagination,i=e.getPageCount();return i===-1?!0:i===0?!1:r<i-1},e.previousPage=()=>e.setPageIndex(r=>r-1),e.nextPage=()=>e.setPageIndex(r=>r+1),e.firstPage=()=>e.setPageIndex(0),e.lastPage=()=>e.setPageIndex(e.getPageCount()-1),e.getPrePaginationRowModel=()=>e.getExpandedRowModel(),e.getPaginationRowModel=()=>(!e._getPaginationRowModel&&e.options.getPaginationRowModel&&(e._getPaginationRowModel=e.options.getPaginationRowModel(e)),e.options.manualPagination||!e._getPaginationRowModel?e.getPrePaginationRowModel():e._getPaginationRowModel()),e.getPageCount=()=>{var r;return(r=e.options.pageCount)!=null?r:Math.ceil(e.getRowCount()/e.getState().pagination.pageSize)},e.getRowCount=()=>{var r;return(r=e.options.rowCount)!=null?r:e.getPrePaginationRowModel().rows.length}}},s0=()=>({top:[],bottom:[]}),Xte={getInitialState:e=>({rowPinning:s0(),...e}),getDefaultOptions:e=>({onRowPinningChange:gr("rowPinning",e)}),createRow:(e,t)=>{e.pin=(n,r,i)=>{const a=r?e.getLeafRows().map(c=>{let{id:d}=c;return d}):[],s=i?e.getParentRows().map(c=>{let{id:d}=c;return d}):[],u=new Set([...s,e.id,...a]);t.setRowPinning(c=>{var d,p;if(n==="bottom"){var m,v;return{top:((m=c==null?void 0:c.top)!=null?m:[]).filter(w=>!(u!=null&&u.has(w))),bottom:[...((v=c==null?void 0:c.bottom)!=null?v:[]).filter(w=>!(u!=null&&u.has(w))),...Array.from(u)]}}if(n==="top"){var b,S;return{top:[...((b=c==null?void 0:c.top)!=null?b:[]).filter(w=>!(u!=null&&u.has(w))),...Array.from(u)],bottom:((S=c==null?void 0:c.bottom)!=null?S:[]).filter(w=>!(u!=null&&u.has(w)))}}return{top:((d=c==null?void 0:c.top)!=null?d:[]).filter(w=>!(u!=null&&u.has(w))),bottom:((p=c==null?void 0:c.bottom)!=null?p:[]).filter(w=>!(u!=null&&u.has(w)))}})},e.getCanPin=()=>{var n;const{enableRowPinning:r,enablePinning:i}=t.options;return typeof r=="function"?r(e):(n=r??i)!=null?n:!0},e.getIsPinned=()=>{const n=[e.id],{top:r,bottom:i}=t.getState().rowPinning,a=n.some(u=>r==null?void 0:r.includes(u)),s=n.some(u=>i==null?void 0:i.includes(u));return a?"top":s?"bottom":!1},e.getPinnedIndex=()=>{var n,r;const i=e.getIsPinned();if(!i)return-1;const a=(n=i==="top"?t.getTopRows():t.getBottomRows())==null?void 0:n.map(s=>{let{id:u}=s;return u});return(r=a==null?void 0:a.indexOf(e.id))!=null?r:-1}},createTable:e=>{e.setRowPinning=t=>e.options.onRowPinningChange==null?void 0:e.options.onRowPinningChange(t),e.resetRowPinning=t=>{var n,r;return e.setRowPinning(t?s0():(n=(r=e.initialState)==null?void 0:r.rowPinning)!=null?n:s0())},e.getIsSomeRowsPinned=t=>{var n;const r=e.getState().rowPinning;if(!t){var i,a;return!!((i=r.top)!=null&&i.length||(a=r.bottom)!=null&&a.length)}return!!((n=r[t])!=null&&n.length)},e._getPinnedRows=(t,n,r)=>{var i;return((i=e.options.keepPinnedRows)==null||i?(n??[]).map(s=>{const u=e.getRow(s,!0);return u.getIsAllParentsExpanded()?u:null}):(n??[]).map(s=>t.find(u=>u.id===s))).filter(Boolean).map(s=>({...s,position:r}))},e.getTopRows=Ge(()=>[e.getRowModel().rows,e.getState().rowPinning.top],(t,n)=>e._getPinnedRows(t,n,"top"),We(e.options,"debugRows")),e.getBottomRows=Ge(()=>[e.getRowModel().rows,e.getState().rowPinning.bottom],(t,n)=>e._getPinnedRows(t,n,"bottom"),We(e.options,"debugRows")),e.getCenterRows=Ge(()=>[e.getRowModel().rows,e.getState().rowPinning.top,e.getState().rowPinning.bottom],(t,n,r)=>{const i=new Set([...n??[],...r??[]]);return t.filter(a=>!i.has(a.id))},We(e.options,"debugRows"))}},Qte={getInitialState:e=>({rowSelection:{},...e}),getDefaultOptions:e=>({onRowSelectionChange:gr("rowSelection",e),enableRowSelection:!0,enableMultiRowSelection:!0,enableSubRowSelection:!0}),createTable:e=>{e.setRowSelection=t=>e.options.onRowSelectionChange==null?void 0:e.options.onRowSelectionChange(t),e.resetRowSelection=t=>{var n;return e.setRowSelection(t?{}:(n=e.initialState.rowSelection)!=null?n:{})},e.toggleAllRowsSelected=t=>{e.setRowSelection(n=>{t=typeof t<"u"?t:!e.getIsAllRowsSelected();const r={...n},i=e.getPreGroupedRowModel().flatRows;return t?i.forEach(a=>{a.getCanSelect()&&(r[a.id]=!0)}):i.forEach(a=>{delete r[a.id]}),r})},e.toggleAllPageRowsSelected=t=>e.setRowSelection(n=>{const r=typeof t<"u"?t:!e.getIsAllPageRowsSelected(),i={...n};return e.getRowModel().rows.forEach(a=>{t1(i,a.id,r,!0,e)}),i}),e.getPreSelectedRowModel=()=>e.getCoreRowModel(),e.getSelectedRowModel=Ge(()=>[e.getState().rowSelection,e.getCoreRowModel()],(t,n)=>Object.keys(t).length?l0(e,n):{rows:[],flatRows:[],rowsById:{}},We(e.options,"debugTable")),e.getFilteredSelectedRowModel=Ge(()=>[e.getState().rowSelection,e.getFilteredRowModel()],(t,n)=>Object.keys(t).length?l0(e,n):{rows:[],flatRows:[],rowsById:{}},We(e.options,"debugTable")),e.getGroupedSelectedRowModel=Ge(()=>[e.getState().rowSelection,e.getSortedRowModel()],(t,n)=>Object.keys(t).length?l0(e,n):{rows:[],flatRows:[],rowsById:{}},We(e.options,"debugTable")),e.getIsAllRowsSelected=()=>{const t=e.getFilteredRowModel().flatRows,{rowSelection:n}=e.getState();let r=!!(t.length&&Object.keys(n).length);return r&&t.some(i=>i.getCanSelect()&&!n[i.id])&&(r=!1),r},e.getIsAllPageRowsSelected=()=>{const t=e.getPaginationRowModel().flatRows.filter(i=>i.getCanSelect()),{rowSelection:n}=e.getState();let r=!!t.length;return r&&t.some(i=>!n[i.id])&&(r=!1),r},e.getIsSomeRowsSelected=()=>{var t;const n=Object.keys((t=e.getState().rowSelection)!=null?t:{}).length;return n>0&&n<e.getFilteredRowModel().flatRows.length},e.getIsSomePageRowsSelected=()=>{const t=e.getPaginationRowModel().flatRows;return e.getIsAllPageRowsSelected()?!1:t.filter(n=>n.getCanSelect()).some(n=>n.getIsSelected()||n.getIsSomeSelected())},e.getToggleAllRowsSelectedHandler=()=>t=>{e.toggleAllRowsSelected(t.target.checked)},e.getToggleAllPageRowsSelectedHandler=()=>t=>{e.toggleAllPageRowsSelected(t.target.checked)}},createRow:(e,t)=>{e.toggleSelected=(n,r)=>{const i=e.getIsSelected();t.setRowSelection(a=>{var s;if(n=typeof n<"u"?n:!i,e.getCanSelect()&&i===n)return a;const u={...a};return t1(u,e.id,n,(s=r==null?void 0:r.selectChildren)!=null?s:!0,t),u})},e.getIsSelected=()=>{const{rowSelection:n}=t.getState();return TC(e,n)},e.getIsSomeSelected=()=>{const{rowSelection:n}=t.getState();return n1(e,n)==="some"},e.getIsAllSubRowsSelected=()=>{const{rowSelection:n}=t.getState();return n1(e,n)==="all"},e.getCanSelect=()=>{var n;return typeof t.options.enableRowSelection=="function"?t.options.enableRowSelection(e):(n=t.options.enableRowSelection)!=null?n:!0},e.getCanSelectSubRows=()=>{var n;return typeof t.options.enableSubRowSelection=="function"?t.options.enableSubRowSelection(e):(n=t.options.enableSubRowSelection)!=null?n:!0},e.getCanMultiSelect=()=>{var n;return typeof t.options.enableMultiRowSelection=="function"?t.options.enableMultiRowSelection(e):(n=t.options.enableMultiRowSelection)!=null?n:!0},e.getToggleSelectedHandler=()=>{const n=e.getCanSelect();return r=>{var i;n&&e.toggleSelected((i=r.target)==null?void 0:i.checked)}}}},t1=(e,t,n,r,i)=>{var a;const s=i.getRow(t,!0);n?(s.getCanMultiSelect()||Object.keys(e).forEach(u=>delete e[u]),s.getCanSelect()&&(e[t]=!0)):delete e[t],r&&(a=s.subRows)!=null&&a.length&&s.getCanSelectSubRows()&&s.subRows.forEach(u=>t1(e,u.id,n,r,i))};function l0(e,t){const n=e.getState().rowSelection,r=[],i={},a=function(s,u){return s.map(c=>{var d;const p=TC(c,n);if(p&&(r.push(c),i[c.id]=c),(d=c.subRows)!=null&&d.length&&(c={...c,subRows:a(c.subRows)}),p)return c}).filter(Boolean)};return{rows:a(t.rows),flatRows:r,rowsById:i}}function TC(e,t){var n;return(n=t[e.id])!=null?n:!1}function n1(e,t,n){var r;if(!((r=e.subRows)!=null&&r.length))return!1;let i=!0,a=!1;return e.subRows.forEach(s=>{if(!(a&&!i)&&(s.getCanSelect()&&(TC(s,t)?a=!0:i=!1),s.subRows&&s.subRows.length)){const u=n1(s,t);u==="all"?a=!0:(u==="some"&&(a=!0),i=!1)}}),i?"all":a?"some":!1}const r1=/([0-9]+)/gm,Jte=(e,t,n)=>Az(co(e.getValue(n)).toLowerCase(),co(t.getValue(n)).toLowerCase()),ene=(e,t,n)=>Az(co(e.getValue(n)),co(t.getValue(n))),tne=(e,t,n)=>MC(co(e.getValue(n)).toLowerCase(),co(t.getValue(n)).toLowerCase()),nne=(e,t,n)=>MC(co(e.getValue(n)),co(t.getValue(n))),rne=(e,t,n)=>{const r=e.getValue(n),i=t.getValue(n);return r>i?1:r<i?-1:0},ine=(e,t,n)=>MC(e.getValue(n),t.getValue(n));function MC(e,t){return e===t?0:e>t?1:-1}function co(e){return typeof e=="number"?isNaN(e)||e===1/0||e===-1/0?"":String(e):typeof e=="string"?e:""}function Az(e,t){const n=e.split(r1).filter(Boolean),r=t.split(r1).filter(Boolean);for(;n.length&&r.length;){const i=n.shift(),a=r.shift(),s=parseInt(i,10),u=parseInt(a,10),c=[s,u].sort();if(isNaN(c[0])){if(i>a)return 1;if(a>i)return-1;continue}if(isNaN(c[1]))return isNaN(s)?-1:1;if(s>u)return 1;if(u>s)return-1}return n.length-r.length}const Rc={alphanumeric:Jte,alphanumericCaseSensitive:ene,text:tne,textCaseSensitive:nne,datetime:rne,basic:ine},ane={getInitialState:e=>({sorting:[],...e}),getDefaultColumnDef:()=>({sortingFn:"auto",sortUndefined:1}),getDefaultOptions:e=>({onSortingChange:gr("sorting",e),isMultiSortEvent:t=>t.shiftKey}),createColumn:(e,t)=>{e.getAutoSortingFn=()=>{const n=t.getFilteredRowModel().flatRows.slice(10);let r=!1;for(const i of n){const a=i==null?void 0:i.getValue(e.id);if(Object.prototype.toString.call(a)==="[object Date]")return Rc.datetime;if(typeof a=="string"&&(r=!0,a.split(r1).length>1))return Rc.alphanumeric}return r?Rc.text:Rc.basic},e.getAutoSortDir=()=>{const n=t.getFilteredRowModel().flatRows[0];return typeof(n==null?void 0:n.getValue(e.id))=="string"?"asc":"desc"},e.getSortingFn=()=>{var n,r;if(!e)throw new Error;return Mv(e.columnDef.sortingFn)?e.columnDef.sortingFn:e.columnDef.sortingFn==="auto"?e.getAutoSortingFn():(n=(r=t.options.sortingFns)==null?void 0:r[e.columnDef.sortingFn])!=null?n:Rc[e.columnDef.sortingFn]},e.toggleSorting=(n,r)=>{const i=e.getNextSortingOrder(),a=typeof n<"u"&&n!==null;t.setSorting(s=>{const u=s==null?void 0:s.find(b=>b.id===e.id),c=s==null?void 0:s.findIndex(b=>b.id===e.id);let d=[],p,m=a?n:i==="desc";if(s!=null&&s.length&&e.getCanMultiSort()&&r?u?p="toggle":p="add":s!=null&&s.length&&c!==s.length-1?p="replace":u?p="toggle":p="replace",p==="toggle"&&(a||i||(p="remove")),p==="add"){var v;d=[...s,{id:e.id,desc:m}],d.splice(0,d.length-((v=t.options.maxMultiSortColCount)!=null?v:Number.MAX_SAFE_INTEGER))}else p==="toggle"?d=s.map(b=>b.id===e.id?{...b,desc:m}:b):p==="remove"?d=s.filter(b=>b.id!==e.id):d=[{id:e.id,desc:m}];return d})},e.getFirstSortDir=()=>{var n,r;return((n=(r=e.columnDef.sortDescFirst)!=null?r:t.options.sortDescFirst)!=null?n:e.getAutoSortDir()==="desc")?"desc":"asc"},e.getNextSortingOrder=n=>{var r,i;const a=e.getFirstSortDir(),s=e.getIsSorted();return s?s!==a&&((r=t.options.enableSortingRemoval)==null||r)&&(!(n&&(i=t.options.enableMultiRemove)!=null)||i)?!1:s==="desc"?"asc":"desc":a},e.getCanSort=()=>{var n,r;return((n=e.columnDef.enableSorting)!=null?n:!0)&&((r=t.options.enableSorting)!=null?r:!0)&&!!e.accessorFn},e.getCanMultiSort=()=>{var n,r;return(n=(r=e.columnDef.enableMultiSort)!=null?r:t.options.enableMultiSort)!=null?n:!!e.accessorFn},e.getIsSorted=()=>{var n;const r=(n=t.getState().sorting)==null?void 0:n.find(i=>i.id===e.id);return r?r.desc?"desc":"asc":!1},e.getSortIndex=()=>{var n,r;return(n=(r=t.getState().sorting)==null?void 0:r.findIndex(i=>i.id===e.id))!=null?n:-1},e.clearSorting=()=>{t.setSorting(n=>n!=null&&n.length?n.filter(r=>r.id!==e.id):[])},e.getToggleSortingHandler=()=>{const n=e.getCanSort();return r=>{n&&(r.persist==null||r.persist(),e.toggleSorting==null||e.toggleSorting(void 0,e.getCanMultiSort()?t.options.isMultiSortEvent==null?void 0:t.options.isMultiSortEvent(r):!1))}}},createTable:e=>{e.setSorting=t=>e.options.onSortingChange==null?void 0:e.options.onSortingChange(t),e.resetSorting=t=>{var n,r;e.setSorting(t?[]:(n=(r=e.initialState)==null?void 0:r.sorting)!=null?n:[])},e.getPreSortedRowModel=()=>e.getGroupedRowModel(),e.getSortedRowModel=()=>(!e._getSortedRowModel&&e.options.getSortedRowModel&&(e._getSortedRowModel=e.options.getSortedRowModel(e)),e.options.manualSorting||!e._getSortedRowModel?e.getPreSortedRowModel():e._getSortedRowModel())}},one=[Ote,Gte,qte,zte,Rte,Tte,Wte,Kte,ane,Bte,Yte,Zte,Xte,Qte,Vte];function sne(e){var t,n;const r=[...one,...(t=e._features)!=null?t:[]];let i={_features:r};const a=i._features.reduce((v,b)=>Object.assign(v,b.getDefaultOptions==null?void 0:b.getDefaultOptions(i)),{}),s=v=>i.options.mergeOptions?i.options.mergeOptions(a,v):{...a,...v};let c={...{},...(n=e.initialState)!=null?n:{}};i._features.forEach(v=>{var b;c=(b=v.getInitialState==null?void 0:v.getInitialState(c))!=null?b:c});const d=[];let p=!1;const m={_features:r,options:{...a,...e},initialState:c,_queue:v=>{d.push(v),p||(p=!0,Promise.resolve().then(()=>{for(;d.length;)d.shift()();p=!1}).catch(b=>setTimeout(()=>{throw b})))},reset:()=>{i.setState(i.initialState)},setOptions:v=>{const b=eo(v,i.options);i.options=s(b)},getState:()=>i.options.state,setState:v=>{i.options.onStateChange==null||i.options.onStateChange(v)},_getRowId:(v,b,S)=>{var w;return(w=i.options.getRowId==null?void 0:i.options.getRowId(v,b,S))!=null?w:`${S?[S.id,b].join("."):b}`},getCoreRowModel:()=>(i._getCoreRowModel||(i._getCoreRowModel=i.options.getCoreRowModel(i)),i._getCoreRowModel()),getRowModel:()=>i.getPaginationRowModel(),getRow:(v,b)=>{let S=(b?i.getPrePaginationRowModel():i.getRowModel()).rowsById[v];if(!S&&(S=i.getCoreRowModel().rowsById[v],!S))throw new Error;return S},_getDefaultColumnDef:Ge(()=>[i.options.defaultColumn],v=>{var b;return v=(b=v)!=null?b:{},{header:S=>{const w=S.header.column.columnDef;return w.accessorKey?w.accessorKey:w.accessorFn?w.id:null},cell:S=>{var w,x;return(w=(x=S.renderValue())==null||x.toString==null?void 0:x.toString())!=null?w:null},...i._features.reduce((S,w)=>Object.assign(S,w.getDefaultColumnDef==null?void 0:w.getDefaultColumnDef()),{}),...v}},We(e,"debugColumns")),_getColumnDefs:()=>i.options.columns,getAllColumns:Ge(()=>[i._getColumnDefs()],v=>{const b=function(S,w,x){return x===void 0&&(x=0),S.map(C=>{const O=Cte(i,C,x,w),A=C;return O.columns=A.columns?b(A.columns,O,x+1):[],O})};return b(v)},We(e,"debugColumns")),getAllFlatColumns:Ge(()=>[i.getAllColumns()],v=>v.flatMap(b=>b.getFlatColumns()),We(e,"debugColumns")),_getAllFlatColumnsById:Ge(()=>[i.getAllFlatColumns()],v=>v.reduce((b,S)=>(b[S.id]=S,b),{}),We(e,"debugColumns")),getAllLeafColumns:Ge(()=>[i.getAllColumns(),i._getOrderColumnsFn()],(v,b)=>{let S=v.flatMap(w=>w.getLeafColumns());return b(S)},We(e,"debugColumns")),getColumn:v=>i._getAllFlatColumnsById()[v]};Object.assign(i,m);for(let v=0;v<i._features.length;v++){const b=i._features[v];b==null||b.createTable==null||b.createTable(i)}return i}function lne(){return e=>Ge(()=>[e.options.data],t=>{const n={rows:[],flatRows:[],rowsById:{}},r=function(i,a,s){a===void 0&&(a=0);const u=[];for(let d=0;d<i.length;d++){const p=Ate(e,e._getRowId(i[d],d,s),i[d],d,a,void 0,s==null?void 0:s.id);if(n.flatRows.push(p),n.rowsById[p.id]=p,u.push(p),e.options.getSubRows){var c;p.originalSubRows=e.options.getSubRows(i[d],d),(c=p.originalSubRows)!=null&&c.length&&(p.subRows=r(p.originalSubRows,a+1,p))}}return u};return n.rows=r(t),n},We(e.options,"debugTable","getRowModel",()=>e._autoResetPageIndex()))}/**
   * react-table
   *
   * Copyright (c) TanStack
   *
   * This source code is licensed under the MIT license found in the
   * LICENSE.md file in the root directory of this source tree.
   *
   * @license MIT
   */function lj(e,t){return e?une(e)?g.createElement(e,t):e:null}function une(e){return cne(e)||typeof e=="function"||fne(e)}function cne(e){return typeof e=="function"&&(()=>{const t=Object.getPrototypeOf(e);return t.prototype&&t.prototype.isReactComponent})()}function fne(e){return typeof e=="object"&&typeof e.$$typeof=="symbol"&&["react.memo","react.forward_ref"].includes(e.$$typeof.description)}function dne(e){const t={state:{},onStateChange:()=>{},renderFallbackValue:null,...e},[n]=g.useState(()=>({current:sne(t)})),[r,i]=g.useState(()=>n.current.initialState);return n.current.setOptions(a=>({...a,...e,state:{...r,...e.state},onStateChange:s=>{i(s),e.onStateChange==null||e.onStateChange(s)}})),n.current}function pne({className:e,...t}){return E.jsx("div",{"data-slot":"table-container",className:"relative w-full overflow-x-auto",children:E.jsx("table",{"data-slot":"table",className:ke("w-full caption-bottom text-sm rounded-md border",e),...t})})}function hne({className:e,...t}){return E.jsx("thead",{"data-slot":"table-header",className:ke("[&_tr]:border-b",e),...t})}function mne({className:e,...t}){return E.jsx("tbody",{"data-slot":"table-body",className:ke("[&_tr:last-child]:border-0",e),...t})}function u0({className:e,...t}){return E.jsx("tr",{"data-slot":"table-row",className:ke("hover:bg-muted/50 data-[state=selected]:bg-muted border-b transition-colors",e),...t})}function vne({className:e,...t}){return E.jsx("th",{"data-slot":"table-head",className:ke("bg-gray-100 text-foreground h-10 px-2 text-left align-middle font-medium whitespace-nowrap [&:has([role=checkbox])]:pr-0 [&>[role=checkbox]]:translate-y-[2px]",e),...t})}function uj({className:e,...t}){return E.jsx("td",{"data-slot":"table-cell",className:ke("p-2 align-middle whitespace-nowrap [&:has([role=checkbox])]:pr-0 [&>[role=checkbox]]:translate-y-[2px]",e),...t})}function i1({columns:e,data:t}){var r;const n=dne({data:t,columns:e,getCoreRowModel:lne()});return E.jsx("div",{className:"rounded-md border",children:E.jsxs(pne,{children:[E.jsx(hne,{children:n.getHeaderGroups().map(i=>E.jsx(u0,{children:i.headers.map(a=>E.jsx(vne,{children:a.isPlaceholder?null:lj(a.column.columnDef.header,a.getContext())},a.id))},i.id))}),E.jsx(mne,{children:(r=n.getRowModel().rows)!=null&&r.length?n.getRowModel().rows.map(i=>E.jsx(u0,{"data-state":i.getIsSelected()&&"selected",children:i.getVisibleCells().map(a=>E.jsx(uj,{children:lj(a.column.columnDef.cell,a.getContext())},a.id))},i.id)):E.jsx(u0,{children:E.jsx(uj,{colSpan:e.length,className:"h-24 text-center",children:"No results."})})})]})})}function gne(){const[e,t]=g.useState(null);g.useEffect(()=>{cC().then(r=>{const i=[];Object.entries(r).forEach(([a,s])=>{s.forEach(u=>{i.push({model_type:a,name:u})})}),t(i)}).catch(r=>console.error("Fetch error:",r))},[]);const n=[{header:"Model Name",accessorKey:"name"},{header:"Model Type",accessorKey:"model_type"},{id:"details",cell:({row:r})=>E.jsx(aa,{to:`/models/${r.getValue("name")}-${r.getValue("model_type")}`.toLowerCase(),className:"",children:E.jsx(J6,{})})}];return E.jsx("main",{className:"flex-1 flex flex-col items-center justify-center p-4",children:E.jsxs("div",{className:"w-1/2",children:[E.jsx("h1",{className:"text-xl font-bold",children:"Available models"}),E.jsx("p",{children:"The following models are available for classification and filtering."}),E.jsx(Mr,{className:"my-4"}),E.jsx(i1,{columns:n,data:e||[]})]})})}const Rz=({size:e=24,className:t,...n})=>E.jsx("svg",{xmlns:"http://www.w3.org/2000/svg",width:e,height:e,...n,viewBox:"0 0 24 24",fill:"none",stroke:"currentColor",strokeWidth:"2",strokeLinecap:"round",strokeLinejoin:"round",className:ke("animate-spin",t),children:E.jsx("path",{d:"M21 12a9 9 0 1 1-6.219-8.56"})});var c0,cj;function Qn(){if(cj)return c0;cj=1;var e=Array.isArray;return c0=e,c0}var f0,fj;function Tz(){if(fj)return f0;fj=1;var e=typeof Up=="object"&&Up&&Up.Object===Object&&Up;return f0=e,f0}var d0,dj;function Mi(){if(dj)return d0;dj=1;var e=Tz(),t=typeof self=="object"&&self&&self.Object===Object&&self,n=e||t||Function("return this")();return d0=n,d0}var p0,pj;function bd(){if(pj)return p0;pj=1;var e=Mi(),t=e.Symbol;return p0=t,p0}var h0,hj;function yne(){if(hj)return h0;hj=1;var e=bd(),t=Object.prototype,n=t.hasOwnProperty,r=t.toString,i=e?e.toStringTag:void 0;function a(s){var u=n.call(s,i),c=s[i];try{s[i]=void 0;var d=!0}catch{}var p=r.call(s);return d&&(u?s[i]=c:delete s[i]),p}return h0=a,h0}var m0,mj;function bne(){if(mj)return m0;mj=1;var e=Object.prototype,t=e.toString;function n(r){return t.call(r)}return m0=n,m0}var v0,vj;function xa(){if(vj)return v0;vj=1;var e=bd(),t=yne(),n=bne(),r="[object Null]",i="[object Undefined]",a=e?e.toStringTag:void 0;function s(u){return u==null?u===void 0?i:r:a&&a in Object(u)?t(u):n(u)}return v0=s,v0}var g0,gj;function wa(){if(gj)return g0;gj=1;function e(t){return t!=null&&typeof t=="object"}return g0=e,g0}var y0,yj;function gu(){if(yj)return y0;yj=1;var e=xa(),t=wa(),n="[object Symbol]";function r(i){return typeof i=="symbol"||t(i)&&e(i)==n}return y0=r,y0}var b0,bj;function PC(){if(bj)return b0;bj=1;var e=Qn(),t=gu(),n=/\.|\[(?:[^[\]]*|(["'])(?:(?!\1)[^\\]|\\.)*?\1)\]/,r=/^\w*$/;function i(a,s){if(e(a))return!1;var u=typeof a;return u=="number"||u=="symbol"||u=="boolean"||a==null||t(a)?!0:r.test(a)||!n.test(a)||s!=null&&a in Object(s)}return b0=i,b0}var x0,xj;function mo(){if(xj)return x0;xj=1;function e(t){var n=typeof t;return t!=null&&(n=="object"||n=="function")}return x0=e,x0}var w0,wj;function jC(){if(wj)return w0;wj=1;var e=xa(),t=mo(),n="[object AsyncFunction]",r="[object Function]",i="[object GeneratorFunction]",a="[object Proxy]";function s(u){if(!t(u))return!1;var c=e(u);return c==r||c==i||c==n||c==a}return w0=s,w0}var S0,Sj;function xne(){if(Sj)return S0;Sj=1;var e=Mi(),t=e["__core-js_shared__"];return S0=t,S0}var _0,_j;function wne(){if(_j)return _0;_j=1;var e=xne(),t=function(){var r=/[^.]+$/.exec(e&&e.keys&&e.keys.IE_PROTO||"");return r?"Symbol(src)_1."+r:""}();function n(r){return!!t&&t in r}return _0=n,_0}var E0,Ej;function Mz(){if(Ej)return E0;Ej=1;var e=Function.prototype,t=e.toString;function n(r){if(r!=null){try{return t.call(r)}catch{}try{return r+""}catch{}}return""}return E0=n,E0}var C0,Cj;function Sne(){if(Cj)return C0;Cj=1;var e=jC(),t=wne(),n=mo(),r=Mz(),i=/[\\^$.*+?()[\]{}|]/g,a=/^\[object .+?Constructor\]$/,s=Function.prototype,u=Object.prototype,c=s.toString,d=u.hasOwnProperty,p=RegExp("^"+c.call(d).replace(i,"\\$&").replace(/hasOwnProperty|(function).*?(?=\\\()| for .+?(?=\\\])/g,"$1.*?")+"$");function m(v){if(!n(v)||t(v))return!1;var b=e(v)?p:a;return b.test(r(v))}return C0=m,C0}var O0,Oj;function _ne(){if(Oj)return O0;Oj=1;function e(t,n){return t==null?void 0:t[n]}return O0=e,O0}var A0,Aj;function gs(){if(Aj)return A0;Aj=1;var e=Sne(),t=_ne();function n(r,i){var a=t(r,i);return e(a)?a:void 0}return A0=n,A0}var R0,Rj;function Pv(){if(Rj)return R0;Rj=1;var e=gs(),t=e(Object,"create");return R0=t,R0}var T0,Tj;function Ene(){if(Tj)return T0;Tj=1;var e=Pv();function t(){this.__data__=e?e(null):{},this.size=0}return T0=t,T0}var M0,Mj;function Cne(){if(Mj)return M0;Mj=1;function e(t){var n=this.has(t)&&delete this.__data__[t];return this.size-=n?1:0,n}return M0=e,M0}var P0,Pj;function One(){if(Pj)return P0;Pj=1;var e=Pv(),t="__lodash_hash_undefined__",n=Object.prototype,r=n.hasOwnProperty;function i(a){var s=this.__data__;if(e){var u=s[a];return u===t?void 0:u}return r.call(s,a)?s[a]:void 0}return P0=i,P0}var j0,jj;function Ane(){if(jj)return j0;jj=1;var e=Pv(),t=Object.prototype,n=t.hasOwnProperty;function r(i){var a=this.__data__;return e?a[i]!==void 0:n.call(a,i)}return j0=r,j0}var N0,Nj;function Rne(){if(Nj)return N0;Nj=1;var e=Pv(),t="__lodash_hash_undefined__";function n(r,i){var a=this.__data__;return this.size+=this.has(r)?0:1,a[r]=e&&i===void 0?t:i,this}return N0=n,N0}var D0,Dj;function Tne(){if(Dj)return D0;Dj=1;var e=Ene(),t=Cne(),n=One(),r=Ane(),i=Rne();function a(s){var u=-1,c=s==null?0:s.length;for(this.clear();++u<c;){var d=s[u];this.set(d[0],d[1])}}return a.prototype.clear=e,a.prototype.delete=t,a.prototype.get=n,a.prototype.has=r,a.prototype.set=i,D0=a,D0}var k0,kj;function Mne(){if(kj)return k0;kj=1;function e(){this.__data__=[],this.size=0}return k0=e,k0}var $0,$j;function NC(){if($j)return $0;$j=1;function e(t,n){return t===n||t!==t&&n!==n}return $0=e,$0}var I0,Ij;function jv(){if(Ij)return I0;Ij=1;var e=NC();function t(n,r){for(var i=n.length;i--;)if(e(n[i][0],r))return i;return-1}return I0=t,I0}var L0,Lj;function Pne(){if(Lj)return L0;Lj=1;var e=jv(),t=Array.prototype,n=t.splice;function r(i){var a=this.__data__,s=e(a,i);if(s<0)return!1;var u=a.length-1;return s==u?a.pop():n.call(a,s,1),--this.size,!0}return L0=r,L0}var B0,Bj;function jne(){if(Bj)return B0;Bj=1;var e=jv();function t(n){var r=this.__data__,i=e(r,n);return i<0?void 0:r[i][1]}return B0=t,B0}var F0,Fj;function Nne(){if(Fj)return F0;Fj=1;var e=jv();function t(n){return e(this.__data__,n)>-1}return F0=t,F0}var q0,qj;function Dne(){if(qj)return q0;qj=1;var e=jv();function t(n,r){var i=this.__data__,a=e(i,n);return a<0?(++this.size,i.push([n,r])):i[a][1]=r,this}return q0=t,q0}var z0,zj;function Nv(){if(zj)return z0;zj=1;var e=Mne(),t=Pne(),n=jne(),r=Nne(),i=Dne();function a(s){var u=-1,c=s==null?0:s.length;for(this.clear();++u<c;){var d=s[u];this.set(d[0],d[1])}}return a.prototype.clear=e,a.prototype.delete=t,a.prototype.get=n,a.prototype.has=r,a.prototype.set=i,z0=a,z0}var U0,Uj;function DC(){if(Uj)return U0;Uj=1;var e=gs(),t=Mi(),n=e(t,"Map");return U0=n,U0}var V0,Vj;function kne(){if(Vj)return V0;Vj=1;var e=Tne(),t=Nv(),n=DC();function r(){this.size=0,this.__data__={hash:new e,map:new(n||t),string:new e}}return V0=r,V0}var H0,Hj;function $ne(){if(Hj)return H0;Hj=1;function e(t){var n=typeof t;return n=="string"||n=="number"||n=="symbol"||n=="boolean"?t!=="__proto__":t===null}return H0=e,H0}var G0,Gj;function Dv(){if(Gj)return G0;Gj=1;var e=$ne();function t(n,r){var i=n.__data__;return e(r)?i[typeof r=="string"?"string":"hash"]:i.map}return G0=t,G0}var W0,Wj;function Ine(){if(Wj)return W0;Wj=1;var e=Dv();function t(n){var r=e(this,n).delete(n);return this.size-=r?1:0,r}return W0=t,W0}var K0,Kj;function Lne(){if(Kj)return K0;Kj=1;var e=Dv();function t(n){return e(this,n).get(n)}return K0=t,K0}var Y0,Yj;function Bne(){if(Yj)return Y0;Yj=1;var e=Dv();function t(n){return e(this,n).has(n)}return Y0=t,Y0}var Z0,Zj;function Fne(){if(Zj)return Z0;Zj=1;var e=Dv();function t(n,r){var i=e(this,n),a=i.size;return i.set(n,r),this.size+=i.size==a?0:1,this}return Z0=t,Z0}var X0,Xj;function kC(){if(Xj)return X0;Xj=1;var e=kne(),t=Ine(),n=Lne(),r=Bne(),i=Fne();function a(s){var u=-1,c=s==null?0:s.length;for(this.clear();++u<c;){var d=s[u];this.set(d[0],d[1])}}return a.prototype.clear=e,a.prototype.delete=t,a.prototype.get=n,a.prototype.has=r,a.prototype.set=i,X0=a,X0}var Q0,Qj;function Pz(){if(Qj)return Q0;Qj=1;var e=kC(),t="Expected a function";function n(r,i){if(typeof r!="function"||i!=null&&typeof i!="function")throw new TypeError(t);var a=function(){var s=arguments,u=i?i.apply(this,s):s[0],c=a.cache;if(c.has(u))return c.get(u);var d=r.apply(this,s);return a.cache=c.set(u,d)||c,d};return a.cache=new(n.Cache||e),a}return n.Cache=e,Q0=n,Q0}var J0,Jj;function qne(){if(Jj)return J0;Jj=1;var e=Pz(),t=500;function n(r){var i=e(r,function(s){return a.size===t&&a.clear(),s}),a=i.cache;return i}return J0=n,J0}var ex,eN;function zne(){if(eN)return ex;eN=1;var e=qne(),t=/[^.[\]]+|\[(?:(-?\d+(?:\.\d+)?)|(["'])((?:(?!\2)[^\\]|\\.)*?)\2)\]|(?=(?:\.|\[\])(?:\.|\[\]|$))/g,n=/\\(\\)?/g,r=e(function(i){var a=[];return i.charCodeAt(0)===46&&a.push(""),i.replace(t,function(s,u,c,d){a.push(c?d.replace(n,"$1"):u||s)}),a});return ex=r,ex}var tx,tN;function $C(){if(tN)return tx;tN=1;function e(t,n){for(var r=-1,i=t==null?0:t.length,a=Array(i);++r<i;)a[r]=n(t[r],r,t);return a}return tx=e,tx}var nx,nN;function Une(){if(nN)return nx;nN=1;var e=bd(),t=$C(),n=Qn(),r=gu(),i=e?e.prototype:void 0,a=i?i.toString:void 0;function s(u){if(typeof u=="string")return u;if(n(u))return t(u,s)+"";if(r(u))return a?a.call(u):"";var c=u+"";return c=="0"&&1/u==-1/0?"-0":c}return nx=s,nx}var rx,rN;function jz(){if(rN)return rx;rN=1;var e=Une();function t(n){return n==null?"":e(n)}return rx=t,rx}var ix,iN;function Nz(){if(iN)return ix;iN=1;var e=Qn(),t=PC(),n=zne(),r=jz();function i(a,s){return e(a)?a:t(a,s)?[a]:n(r(a))}return ix=i,ix}var ax,aN;function kv(){if(aN)return ax;aN=1;var e=gu();function t(n){if(typeof n=="string"||e(n))return n;var r=n+"";return r=="0"&&1/n==-1/0?"-0":r}return ax=t,ax}var ox,oN;function IC(){if(oN)return ox;oN=1;var e=Nz(),t=kv();function n(r,i){i=e(i,r);for(var a=0,s=i.length;r!=null&&a<s;)r=r[t(i[a++])];return a&&a==s?r:void 0}return ox=n,ox}var sx,sN;function Dz(){if(sN)return sx;sN=1;var e=IC();function t(n,r,i){var a=n==null?void 0:e(n,r);return a===void 0?i:a}return sx=t,sx}var Vne=Dz();const yi=Rt(Vne);var lx,lN;function Hne(){if(lN)return lx;lN=1;function e(t){return t==null}return lx=e,lx}var Gne=Hne();const vt=Rt(Gne);var ux,uN;function Wne(){if(uN)return ux;uN=1;var e=xa(),t=Qn(),n=wa(),r="[object String]";function i(a){return typeof a=="string"||!t(a)&&n(a)&&e(a)==r}return ux=i,ux}var Kne=Wne();const xd=Rt(Kne);var Yne=jC();const st=Rt(Yne);var Zne=mo();const yu=Rt(Zne);var cx={exports:{}},gt={};/**
 * @license React
 * react-is.production.min.js
 *
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var cN;function Xne(){if(cN)return gt;cN=1;var e=Symbol.for("react.element"),t=Symbol.for("react.portal"),n=Symbol.for("react.fragment"),r=Symbol.for("react.strict_mode"),i=Symbol.for("react.profiler"),a=Symbol.for("react.provider"),s=Symbol.for("react.context"),u=Symbol.for("react.server_context"),c=Symbol.for("react.forward_ref"),d=Symbol.for("react.suspense"),p=Symbol.for("react.suspense_list"),m=Symbol.for("react.memo"),v=Symbol.for("react.lazy"),b=Symbol.for("react.offscreen"),S;S=Symbol.for("react.module.reference");function w(x){if(typeof x=="object"&&x!==null){var C=x.$$typeof;switch(C){case e:switch(x=x.type,x){case n:case i:case r:case d:case p:return x;default:switch(x=x&&x.$$typeof,x){case u:case s:case c:case v:case m:case a:return x;default:return C}}case t:return C}}}return gt.ContextConsumer=s,gt.ContextProvider=a,gt.Element=e,gt.ForwardRef=c,gt.Fragment=n,gt.Lazy=v,gt.Memo=m,gt.Portal=t,gt.Profiler=i,gt.StrictMode=r,gt.Suspense=d,gt.SuspenseList=p,gt.isAsyncMode=function(){return!1},gt.isConcurrentMode=function(){return!1},gt.isContextConsumer=function(x){return w(x)===s},gt.isContextProvider=function(x){return w(x)===a},gt.isElement=function(x){return typeof x=="object"&&x!==null&&x.$$typeof===e},gt.isForwardRef=function(x){return w(x)===c},gt.isFragment=function(x){return w(x)===n},gt.isLazy=function(x){return w(x)===v},gt.isMemo=function(x){return w(x)===m},gt.isPortal=function(x){return w(x)===t},gt.isProfiler=function(x){return w(x)===i},gt.isStrictMode=function(x){return w(x)===r},gt.isSuspense=function(x){return w(x)===d},gt.isSuspenseList=function(x){return w(x)===p},gt.isValidElementType=function(x){return typeof x=="string"||typeof x=="function"||x===n||x===i||x===r||x===d||x===p||x===b||typeof x=="object"&&x!==null&&(x.$$typeof===v||x.$$typeof===m||x.$$typeof===a||x.$$typeof===s||x.$$typeof===c||x.$$typeof===S||x.getModuleId!==void 0)},gt.typeOf=w,gt}var fN;function Qne(){return fN||(fN=1,cx.exports=Xne()),cx.exports}var Jne=Qne(),fx,dN;function kz(){if(dN)return fx;dN=1;var e=xa(),t=wa(),n="[object Number]";function r(i){return typeof i=="number"||t(i)&&e(i)==n}return fx=r,fx}var dx,pN;function ere(){if(pN)return dx;pN=1;var e=kz();function t(n){return e(n)&&n!=+n}return dx=t,dx}var tre=ere();const wd=Rt(tre);var nre=kz();const rre=Rt(nre);var to=function(t){return t===0?0:t>0?1:-1},Ko=function(t){return xd(t)&&t.indexOf("%")===t.length-1},$e=function(t){return rre(t)&&!wd(t)},sn=function(t){return $e(t)||xd(t)},ire=0,LC=function(t){var n=++ire;return"".concat(t||"").concat(n)},hi=function(t,n){var r=arguments.length>2&&arguments[2]!==void 0?arguments[2]:0,i=arguments.length>3&&arguments[3]!==void 0?arguments[3]:!1;if(!$e(t)&&!xd(t))return r;var a;if(Ko(t)){var s=t.indexOf("%");a=n*parseFloat(t.slice(0,s))/100}else a=+t;return wd(a)&&(a=r),i&&a>n&&(a=n),a},bl=function(t){if(!t)return null;var n=Object.keys(t);return n&&n.length?t[n[0]]:null},are=function(t){if(!Array.isArray(t))return!1;for(var n=t.length,r={},i=0;i<n;i++)if(!r[t[i]])r[t[i]]=!0;else return!0;return!1},nh=function(t,n){return $e(t)&&$e(n)?function(r){return t+r*(n-t)}:function(){return n}};function a1(e,t,n){return!e||!e.length?null:e.find(function(r){return r&&(typeof t=="function"?t(r):yi(r,t))===n})}function o1(e,t){for(var n in e)if({}.hasOwnProperty.call(e,n)&&(!{}.hasOwnProperty.call(t,n)||e[n]!==t[n]))return!1;for(var r in t)if({}.hasOwnProperty.call(t,r)&&!{}.hasOwnProperty.call(e,r))return!1;return!0}function s1(e){"@babel/helpers - typeof";return s1=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},s1(e)}var ore=["viewBox","children"],sre=["aria-activedescendant","aria-atomic","aria-autocomplete","aria-busy","aria-checked","aria-colcount","aria-colindex","aria-colspan","aria-controls","aria-current","aria-describedby","aria-details","aria-disabled","aria-errormessage","aria-expanded","aria-flowto","aria-haspopup","aria-hidden","aria-invalid","aria-keyshortcuts","aria-label","aria-labelledby","aria-level","aria-live","aria-modal","aria-multiline","aria-multiselectable","aria-orientation","aria-owns","aria-placeholder","aria-posinset","aria-pressed","aria-readonly","aria-relevant","aria-required","aria-roledescription","aria-rowcount","aria-rowindex","aria-rowspan","aria-selected","aria-setsize","aria-sort","aria-valuemax","aria-valuemin","aria-valuenow","aria-valuetext","className","color","height","id","lang","max","media","method","min","name","style","target","width","role","tabIndex","accentHeight","accumulate","additive","alignmentBaseline","allowReorder","alphabetic","amplitude","arabicForm","ascent","attributeName","attributeType","autoReverse","azimuth","baseFrequency","baselineShift","baseProfile","bbox","begin","bias","by","calcMode","capHeight","clip","clipPath","clipPathUnits","clipRule","colorInterpolation","colorInterpolationFilters","colorProfile","colorRendering","contentScriptType","contentStyleType","cursor","cx","cy","d","decelerate","descent","diffuseConstant","direction","display","divisor","dominantBaseline","dur","dx","dy","edgeMode","elevation","enableBackground","end","exponent","externalResourcesRequired","fill","fillOpacity","fillRule","filter","filterRes","filterUnits","floodColor","floodOpacity","focusable","fontFamily","fontSize","fontSizeAdjust","fontStretch","fontStyle","fontVariant","fontWeight","format","from","fx","fy","g1","g2","glyphName","glyphOrientationHorizontal","glyphOrientationVertical","glyphRef","gradientTransform","gradientUnits","hanging","horizAdvX","horizOriginX","href","ideographic","imageRendering","in2","in","intercept","k1","k2","k3","k4","k","kernelMatrix","kernelUnitLength","kerning","keyPoints","keySplines","keyTimes","lengthAdjust","letterSpacing","lightingColor","limitingConeAngle","local","markerEnd","markerHeight","markerMid","markerStart","markerUnits","markerWidth","mask","maskContentUnits","maskUnits","mathematical","mode","numOctaves","offset","opacity","operator","order","orient","orientation","origin","overflow","overlinePosition","overlineThickness","paintOrder","panose1","pathLength","patternContentUnits","patternTransform","patternUnits","pointerEvents","pointsAtX","pointsAtY","pointsAtZ","preserveAlpha","preserveAspectRatio","primitiveUnits","r","radius","refX","refY","renderingIntent","repeatCount","repeatDur","requiredExtensions","requiredFeatures","restart","result","rotate","rx","ry","seed","shapeRendering","slope","spacing","specularConstant","specularExponent","speed","spreadMethod","startOffset","stdDeviation","stemh","stemv","stitchTiles","stopColor","stopOpacity","strikethroughPosition","strikethroughThickness","string","stroke","strokeDasharray","strokeDashoffset","strokeLinecap","strokeLinejoin","strokeMiterlimit","strokeOpacity","strokeWidth","surfaceScale","systemLanguage","tableValues","targetX","targetY","textAnchor","textDecoration","textLength","textRendering","to","transform","u1","u2","underlinePosition","underlineThickness","unicode","unicodeBidi","unicodeRange","unitsPerEm","vAlphabetic","values","vectorEffect","version","vertAdvY","vertOriginX","vertOriginY","vHanging","vIdeographic","viewTarget","visibility","vMathematical","widths","wordSpacing","writingMode","x1","x2","x","xChannelSelector","xHeight","xlinkActuate","xlinkArcrole","xlinkHref","xlinkRole","xlinkShow","xlinkTitle","xlinkType","xmlBase","xmlLang","xmlns","xmlnsXlink","xmlSpace","y1","y2","y","yChannelSelector","z","zoomAndPan","ref","key","angle"],hN=["points","pathLength"],px={svg:ore,polygon:hN,polyline:hN},BC=["dangerouslySetInnerHTML","onCopy","onCopyCapture","onCut","onCutCapture","onPaste","onPasteCapture","onCompositionEnd","onCompositionEndCapture","onCompositionStart","onCompositionStartCapture","onCompositionUpdate","onCompositionUpdateCapture","onFocus","onFocusCapture","onBlur","onBlurCapture","onChange","onChangeCapture","onBeforeInput","onBeforeInputCapture","onInput","onInputCapture","onReset","onResetCapture","onSubmit","onSubmitCapture","onInvalid","onInvalidCapture","onLoad","onLoadCapture","onError","onErrorCapture","onKeyDown","onKeyDownCapture","onKeyPress","onKeyPressCapture","onKeyUp","onKeyUpCapture","onAbort","onAbortCapture","onCanPlay","onCanPlayCapture","onCanPlayThrough","onCanPlayThroughCapture","onDurationChange","onDurationChangeCapture","onEmptied","onEmptiedCapture","onEncrypted","onEncryptedCapture","onEnded","onEndedCapture","onLoadedData","onLoadedDataCapture","onLoadedMetadata","onLoadedMetadataCapture","onLoadStart","onLoadStartCapture","onPause","onPauseCapture","onPlay","onPlayCapture","onPlaying","onPlayingCapture","onProgress","onProgressCapture","onRateChange","onRateChangeCapture","onSeeked","onSeekedCapture","onSeeking","onSeekingCapture","onStalled","onStalledCapture","onSuspend","onSuspendCapture","onTimeUpdate","onTimeUpdateCapture","onVolumeChange","onVolumeChangeCapture","onWaiting","onWaitingCapture","onAuxClick","onAuxClickCapture","onClick","onClickCapture","onContextMenu","onContextMenuCapture","onDoubleClick","onDoubleClickCapture","onDrag","onDragCapture","onDragEnd","onDragEndCapture","onDragEnter","onDragEnterCapture","onDragExit","onDragExitCapture","onDragLeave","onDragLeaveCapture","onDragOver","onDragOverCapture","onDragStart","onDragStartCapture","onDrop","onDropCapture","onMouseDown","onMouseDownCapture","onMouseEnter","onMouseLeave","onMouseMove","onMouseMoveCapture","onMouseOut","onMouseOutCapture","onMouseOver","onMouseOverCapture","onMouseUp","onMouseUpCapture","onSelect","onSelectCapture","onTouchCancel","onTouchCancelCapture","onTouchEnd","onTouchEndCapture","onTouchMove","onTouchMoveCapture","onTouchStart","onTouchStartCapture","onPointerDown","onPointerDownCapture","onPointerMove","onPointerMoveCapture","onPointerUp","onPointerUpCapture","onPointerCancel","onPointerCancelCapture","onPointerEnter","onPointerEnterCapture","onPointerLeave","onPointerLeaveCapture","onPointerOver","onPointerOverCapture","onPointerOut","onPointerOutCapture","onGotPointerCapture","onGotPointerCaptureCapture","onLostPointerCapture","onLostPointerCaptureCapture","onScroll","onScrollCapture","onWheel","onWheelCapture","onAnimationStart","onAnimationStartCapture","onAnimationEnd","onAnimationEndCapture","onAnimationIteration","onAnimationIterationCapture","onTransitionEnd","onTransitionEndCapture"],am=function(t,n){if(!t||typeof t=="function"||typeof t=="boolean")return null;var r=t;if(g.isValidElement(t)&&(r=t.props),!yu(r))return null;var i={};return Object.keys(r).forEach(function(a){BC.includes(a)&&(i[a]=n||function(s){return r[a](r,s)})}),i},lre=function(t,n,r){return function(i){return t(n,r,i),null}},FC=function(t,n,r){if(!yu(t)||s1(t)!=="object")return null;var i=null;return Object.keys(t).forEach(function(a){var s=t[a];BC.includes(a)&&typeof s=="function"&&(i||(i={}),i[a]=lre(s,n,r))}),i},ure=["children"],cre=["children"];function mN(e,t){if(e==null)return{};var n=fre(e,t),r,i;if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);for(i=0;i<a.length;i++)r=a[i],!(t.indexOf(r)>=0)&&Object.prototype.propertyIsEnumerable.call(e,r)&&(n[r]=e[r])}return n}function fre(e,t){if(e==null)return{};var n={};for(var r in e)if(Object.prototype.hasOwnProperty.call(e,r)){if(t.indexOf(r)>=0)continue;n[r]=e[r]}return n}var vN={click:"onClick",mousedown:"onMouseDown",mouseup:"onMouseUp",mouseover:"onMouseOver",mousemove:"onMouseMove",mouseout:"onMouseOut",mouseenter:"onMouseEnter",mouseleave:"onMouseLeave",touchcancel:"onTouchCancel",touchend:"onTouchEnd",touchmove:"onTouchMove",touchstart:"onTouchStart",contextmenu:"onContextMenu",dblclick:"onDoubleClick"},sa=function(t){return typeof t=="string"?t:t?t.displayName||t.name||"Component":""},gN=null,hx=null,qC=function e(t){if(t===gN&&Array.isArray(hx))return hx;var n=[];return g.Children.forEach(t,function(r){vt(r)||(Jne.isFragment(r)?n=n.concat(e(r.props.children)):n.push(r))}),hx=n,gN=t,n};function la(e,t){var n=[],r=[];return Array.isArray(t)?r=t.map(function(i){return sa(i)}):r=[sa(t)],qC(e).forEach(function(i){var a=yi(i,"type.displayName")||yi(i,"type.name");r.indexOf(a)!==-1&&n.push(i)}),n}function Tr(e,t){var n=la(e,t);return n&&n[0]}var yN=function(t){if(!t||!t.props)return!1;var n=t.props,r=n.width,i=n.height;return!(!$e(r)||r<=0||!$e(i)||i<=0)},dre=["a","altGlyph","altGlyphDef","altGlyphItem","animate","animateColor","animateMotion","animateTransform","circle","clipPath","color-profile","cursor","defs","desc","ellipse","feBlend","feColormatrix","feComponentTransfer","feComposite","feConvolveMatrix","feDiffuseLighting","feDisplacementMap","feDistantLight","feFlood","feFuncA","feFuncB","feFuncG","feFuncR","feGaussianBlur","feImage","feMerge","feMergeNode","feMorphology","feOffset","fePointLight","feSpecularLighting","feSpotLight","feTile","feTurbulence","filter","font","font-face","font-face-format","font-face-name","font-face-url","foreignObject","g","glyph","glyphRef","hkern","image","line","lineGradient","marker","mask","metadata","missing-glyph","mpath","path","pattern","polygon","polyline","radialGradient","rect","script","set","stop","style","svg","switch","symbol","text","textPath","title","tref","tspan","use","view","vkern"],pre=function(t){return t&&t.type&&xd(t.type)&&dre.indexOf(t.type)>=0},hre=function(t,n,r,i){var a,s=(a=px==null?void 0:px[i])!==null&&a!==void 0?a:[];return n.startsWith("data-")||!st(t)&&(i&&s.includes(n)||sre.includes(n))||r&&BC.includes(n)},Je=function(t,n,r){if(!t||typeof t=="function"||typeof t=="boolean")return null;var i=t;if(g.isValidElement(t)&&(i=t.props),!yu(i))return null;var a={};return Object.keys(i).forEach(function(s){var u;hre((u=i)===null||u===void 0?void 0:u[s],s,n,r)&&(a[s]=i[s])}),a},l1=function e(t,n){if(t===n)return!0;var r=g.Children.count(t);if(r!==g.Children.count(n))return!1;if(r===0)return!0;if(r===1)return bN(Array.isArray(t)?t[0]:t,Array.isArray(n)?n[0]:n);for(var i=0;i<r;i++){var a=t[i],s=n[i];if(Array.isArray(a)||Array.isArray(s)){if(!e(a,s))return!1}else if(!bN(a,s))return!1}return!0},bN=function(t,n){if(vt(t)&&vt(n))return!0;if(!vt(t)&&!vt(n)){var r=t.props||{},i=r.children,a=mN(r,ure),s=n.props||{},u=s.children,c=mN(s,cre);return i&&u?o1(a,c)&&l1(i,u):!i&&!u?o1(a,c):!1}return!1},xN=function(t,n){var r=[],i={};return qC(t).forEach(function(a,s){if(pre(a))r.push(a);else if(a){var u=sa(a.type),c=n[u]||{},d=c.handler,p=c.once;if(d&&(!p||!i[u])){var m=d(a,u,s);r.push(m),i[u]=!0}}}),r},mre=function(t){var n=t&&t.type;return n&&vN[n]?vN[n]:null},vre=function(t,n){return qC(n).indexOf(t)},gre=["children","width","height","viewBox","className","style","title","desc"];function u1(){return u1=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)Object.prototype.hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},u1.apply(this,arguments)}function yre(e,t){if(e==null)return{};var n=bre(e,t),r,i;if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);for(i=0;i<a.length;i++)r=a[i],!(t.indexOf(r)>=0)&&Object.prototype.propertyIsEnumerable.call(e,r)&&(n[r]=e[r])}return n}function bre(e,t){if(e==null)return{};var n={};for(var r in e)if(Object.prototype.hasOwnProperty.call(e,r)){if(t.indexOf(r)>=0)continue;n[r]=e[r]}return n}function c1(e){var t=e.children,n=e.width,r=e.height,i=e.viewBox,a=e.className,s=e.style,u=e.title,c=e.desc,d=yre(e,gre),p=i||{width:n,height:r,x:0,y:0},m=ct("recharts-surface",a);return V.createElement("svg",u1({},Je(d,!0,"svg"),{className:m,width:n,height:r,style:s,viewBox:"".concat(p.x," ").concat(p.y," ").concat(p.width," ").concat(p.height)}),V.createElement("title",null,u),V.createElement("desc",null,c),t)}var xre=["children","className"];function f1(){return f1=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)Object.prototype.hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},f1.apply(this,arguments)}function wre(e,t){if(e==null)return{};var n=Sre(e,t),r,i;if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);for(i=0;i<a.length;i++)r=a[i],!(t.indexOf(r)>=0)&&Object.prototype.propertyIsEnumerable.call(e,r)&&(n[r]=e[r])}return n}function Sre(e,t){if(e==null)return{};var n={};for(var r in e)if(Object.prototype.hasOwnProperty.call(e,r)){if(t.indexOf(r)>=0)continue;n[r]=e[r]}return n}var un=V.forwardRef(function(e,t){var n=e.children,r=e.className,i=wre(e,xre),a=ct("recharts-layer",r);return V.createElement("g",f1({className:a},Je(i,!0),{ref:t}),n)}),es=function(t,n){for(var r=arguments.length,i=new Array(r>2?r-2:0),a=2;a<r;a++)i[a-2]=arguments[a]},mx,wN;function _re(){if(wN)return mx;wN=1;function e(t,n,r){var i=-1,a=t.length;n<0&&(n=-n>a?0:a+n),r=r>a?a:r,r<0&&(r+=a),a=n>r?0:r-n>>>0,n>>>=0;for(var s=Array(a);++i<a;)s[i]=t[i+n];return s}return mx=e,mx}var vx,SN;function Ere(){if(SN)return vx;SN=1;var e=_re();function t(n,r,i){var a=n.length;return i=i===void 0?a:i,!r&&i>=a?n:e(n,r,i)}return vx=t,vx}var gx,_N;function $z(){if(_N)return gx;_N=1;var e="\\ud800-\\udfff",t="\\u0300-\\u036f",n="\\ufe20-\\ufe2f",r="\\u20d0-\\u20ff",i=t+n+r,a="\\ufe0e\\ufe0f",s="\\u200d",u=RegExp("["+s+e+i+a+"]");function c(d){return u.test(d)}return gx=c,gx}var yx,EN;function Cre(){if(EN)return yx;EN=1;function e(t){return t.split("")}return yx=e,yx}var bx,CN;function Ore(){if(CN)return bx;CN=1;var e="\\ud800-\\udfff",t="\\u0300-\\u036f",n="\\ufe20-\\ufe2f",r="\\u20d0-\\u20ff",i=t+n+r,a="\\ufe0e\\ufe0f",s="["+e+"]",u="["+i+"]",c="\\ud83c[\\udffb-\\udfff]",d="(?:"+u+"|"+c+")",p="[^"+e+"]",m="(?:\\ud83c[\\udde6-\\uddff]){2}",v="[\\ud800-\\udbff][\\udc00-\\udfff]",b="\\u200d",S=d+"?",w="["+a+"]?",x="(?:"+b+"(?:"+[p,m,v].join("|")+")"+w+S+")*",C=w+S+x,O="(?:"+[p+u+"?",u,m,v,s].join("|")+")",A=RegExp(c+"(?="+c+")|"+O+C,"g");function T(R){return R.match(A)||[]}return bx=T,bx}var xx,ON;function Are(){if(ON)return xx;ON=1;var e=Cre(),t=$z(),n=Ore();function r(i){return t(i)?n(i):e(i)}return xx=r,xx}var wx,AN;function Rre(){if(AN)return wx;AN=1;var e=Ere(),t=$z(),n=Are(),r=jz();function i(a){return function(s){s=r(s);var u=t(s)?n(s):void 0,c=u?u[0]:s.charAt(0),d=u?e(u,1).join(""):s.slice(1);return c[a]()+d}}return wx=i,wx}var Sx,RN;function Tre(){if(RN)return Sx;RN=1;var e=Rre(),t=e("toUpperCase");return Sx=t,Sx}var Mre=Tre();const $v=Rt(Mre);function Dt(e){return function(){return e}}const Iz=Math.cos,om=Math.sin,ri=Math.sqrt,sm=Math.PI,Iv=2*sm,d1=Math.PI,p1=2*d1,zo=1e-6,Pre=p1-zo;function Lz(e){this._+=e[0];for(let t=1,n=e.length;t<n;++t)this._+=arguments[t]+e[t]}function jre(e){let t=Math.floor(e);if(!(t>=0))throw new Error(`invalid digits: ${e}`);if(t>15)return Lz;const n=10**t;return function(r){this._+=r[0];for(let i=1,a=r.length;i<a;++i)this._+=Math.round(arguments[i]*n)/n+r[i]}}class Nre{constructor(t){this._x0=this._y0=this._x1=this._y1=null,this._="",this._append=t==null?Lz:jre(t)}moveTo(t,n){this._append`M${this._x0=this._x1=+t},${this._y0=this._y1=+n}`}closePath(){this._x1!==null&&(this._x1=this._x0,this._y1=this._y0,this._append`Z`)}lineTo(t,n){this._append`L${this._x1=+t},${this._y1=+n}`}quadraticCurveTo(t,n,r,i){this._append`Q${+t},${+n},${this._x1=+r},${this._y1=+i}`}bezierCurveTo(t,n,r,i,a,s){this._append`C${+t},${+n},${+r},${+i},${this._x1=+a},${this._y1=+s}`}arcTo(t,n,r,i,a){if(t=+t,n=+n,r=+r,i=+i,a=+a,a<0)throw new Error(`negative radius: ${a}`);let s=this._x1,u=this._y1,c=r-t,d=i-n,p=s-t,m=u-n,v=p*p+m*m;if(this._x1===null)this._append`M${this._x1=t},${this._y1=n}`;else if(v>zo)if(!(Math.abs(m*c-d*p)>zo)||!a)this._append`L${this._x1=t},${this._y1=n}`;else{let b=r-s,S=i-u,w=c*c+d*d,x=b*b+S*S,C=Math.sqrt(w),O=Math.sqrt(v),A=a*Math.tan((d1-Math.acos((w+v-x)/(2*C*O)))/2),T=A/O,R=A/C;Math.abs(T-1)>zo&&this._append`L${t+T*p},${n+T*m}`,this._append`A${a},${a},0,0,${+(m*b>p*S)},${this._x1=t+R*c},${this._y1=n+R*d}`}}arc(t,n,r,i,a,s){if(t=+t,n=+n,r=+r,s=!!s,r<0)throw new Error(`negative radius: ${r}`);let u=r*Math.cos(i),c=r*Math.sin(i),d=t+u,p=n+c,m=1^s,v=s?i-a:a-i;this._x1===null?this._append`M${d},${p}`:(Math.abs(this._x1-d)>zo||Math.abs(this._y1-p)>zo)&&this._append`L${d},${p}`,r&&(v<0&&(v=v%p1+p1),v>Pre?this._append`A${r},${r},0,1,${m},${t-u},${n-c}A${r},${r},0,1,${m},${this._x1=d},${this._y1=p}`:v>zo&&this._append`A${r},${r},0,${+(v>=d1)},${m},${this._x1=t+r*Math.cos(a)},${this._y1=n+r*Math.sin(a)}`)}rect(t,n,r,i){this._append`M${this._x0=this._x1=+t},${this._y0=this._y1=+n}h${r=+r}v${+i}h${-r}Z`}toString(){return this._}}function zC(e){let t=3;return e.digits=function(n){if(!arguments.length)return t;if(n==null)t=null;else{const r=Math.floor(n);if(!(r>=0))throw new RangeError(`invalid digits: ${n}`);t=r}return e},()=>new Nre(t)}function UC(e){return typeof e=="object"&&"length"in e?e:Array.from(e)}function Bz(e){this._context=e}Bz.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._point=0},lineEnd:function(){(this._line||this._line!==0&&this._point===1)&&this._context.closePath(),this._line=1-this._line},point:function(e,t){switch(e=+e,t=+t,this._point){case 0:this._point=1,this._line?this._context.lineTo(e,t):this._context.moveTo(e,t);break;case 1:this._point=2;default:this._context.lineTo(e,t);break}}};function Lv(e){return new Bz(e)}function Fz(e){return e[0]}function qz(e){return e[1]}function zz(e,t){var n=Dt(!0),r=null,i=Lv,a=null,s=zC(u);e=typeof e=="function"?e:e===void 0?Fz:Dt(e),t=typeof t=="function"?t:t===void 0?qz:Dt(t);function u(c){var d,p=(c=UC(c)).length,m,v=!1,b;for(r==null&&(a=i(b=s())),d=0;d<=p;++d)!(d<p&&n(m=c[d],d,c))===v&&((v=!v)?a.lineStart():a.lineEnd()),v&&a.point(+e(m,d,c),+t(m,d,c));if(b)return a=null,b+""||null}return u.x=function(c){return arguments.length?(e=typeof c=="function"?c:Dt(+c),u):e},u.y=function(c){return arguments.length?(t=typeof c=="function"?c:Dt(+c),u):t},u.defined=function(c){return arguments.length?(n=typeof c=="function"?c:Dt(!!c),u):n},u.curve=function(c){return arguments.length?(i=c,r!=null&&(a=i(r)),u):i},u.context=function(c){return arguments.length?(c==null?r=a=null:a=i(r=c),u):r},u}function rh(e,t,n){var r=null,i=Dt(!0),a=null,s=Lv,u=null,c=zC(d);e=typeof e=="function"?e:e===void 0?Fz:Dt(+e),t=typeof t=="function"?t:Dt(t===void 0?0:+t),n=typeof n=="function"?n:n===void 0?qz:Dt(+n);function d(m){var v,b,S,w=(m=UC(m)).length,x,C=!1,O,A=new Array(w),T=new Array(w);for(a==null&&(u=s(O=c())),v=0;v<=w;++v){if(!(v<w&&i(x=m[v],v,m))===C)if(C=!C)b=v,u.areaStart(),u.lineStart();else{for(u.lineEnd(),u.lineStart(),S=v-1;S>=b;--S)u.point(A[S],T[S]);u.lineEnd(),u.areaEnd()}C&&(A[v]=+e(x,v,m),T[v]=+t(x,v,m),u.point(r?+r(x,v,m):A[v],n?+n(x,v,m):T[v]))}if(O)return u=null,O+""||null}function p(){return zz().defined(i).curve(s).context(a)}return d.x=function(m){return arguments.length?(e=typeof m=="function"?m:Dt(+m),r=null,d):e},d.x0=function(m){return arguments.length?(e=typeof m=="function"?m:Dt(+m),d):e},d.x1=function(m){return arguments.length?(r=m==null?null:typeof m=="function"?m:Dt(+m),d):r},d.y=function(m){return arguments.length?(t=typeof m=="function"?m:Dt(+m),n=null,d):t},d.y0=function(m){return arguments.length?(t=typeof m=="function"?m:Dt(+m),d):t},d.y1=function(m){return arguments.length?(n=m==null?null:typeof m=="function"?m:Dt(+m),d):n},d.lineX0=d.lineY0=function(){return p().x(e).y(t)},d.lineY1=function(){return p().x(e).y(n)},d.lineX1=function(){return p().x(r).y(t)},d.defined=function(m){return arguments.length?(i=typeof m=="function"?m:Dt(!!m),d):i},d.curve=function(m){return arguments.length?(s=m,a!=null&&(u=s(a)),d):s},d.context=function(m){return arguments.length?(m==null?a=u=null:u=s(a=m),d):a},d}class Uz{constructor(t,n){this._context=t,this._x=n}areaStart(){this._line=0}areaEnd(){this._line=NaN}lineStart(){this._point=0}lineEnd(){(this._line||this._line!==0&&this._point===1)&&this._context.closePath(),this._line=1-this._line}point(t,n){switch(t=+t,n=+n,this._point){case 0:{this._point=1,this._line?this._context.lineTo(t,n):this._context.moveTo(t,n);break}case 1:this._point=2;default:{this._x?this._context.bezierCurveTo(this._x0=(this._x0+t)/2,this._y0,this._x0,n,t,n):this._context.bezierCurveTo(this._x0,this._y0=(this._y0+n)/2,t,this._y0,t,n);break}}this._x0=t,this._y0=n}}function Dre(e){return new Uz(e,!0)}function kre(e){return new Uz(e,!1)}const VC={draw(e,t){const n=ri(t/sm);e.moveTo(n,0),e.arc(0,0,n,0,Iv)}},$re={draw(e,t){const n=ri(t/5)/2;e.moveTo(-3*n,-n),e.lineTo(-n,-n),e.lineTo(-n,-3*n),e.lineTo(n,-3*n),e.lineTo(n,-n),e.lineTo(3*n,-n),e.lineTo(3*n,n),e.lineTo(n,n),e.lineTo(n,3*n),e.lineTo(-n,3*n),e.lineTo(-n,n),e.lineTo(-3*n,n),e.closePath()}},Vz=ri(1/3),Ire=Vz*2,Lre={draw(e,t){const n=ri(t/Ire),r=n*Vz;e.moveTo(0,-n),e.lineTo(r,0),e.lineTo(0,n),e.lineTo(-r,0),e.closePath()}},Bre={draw(e,t){const n=ri(t),r=-n/2;e.rect(r,r,n,n)}},Fre=.8908130915292852,Hz=om(sm/10)/om(7*sm/10),qre=om(Iv/10)*Hz,zre=-Iz(Iv/10)*Hz,Ure={draw(e,t){const n=ri(t*Fre),r=qre*n,i=zre*n;e.moveTo(0,-n),e.lineTo(r,i);for(let a=1;a<5;++a){const s=Iv*a/5,u=Iz(s),c=om(s);e.lineTo(c*n,-u*n),e.lineTo(u*r-c*i,c*r+u*i)}e.closePath()}},_x=ri(3),Vre={draw(e,t){const n=-ri(t/(_x*3));e.moveTo(0,n*2),e.lineTo(-_x*n,-n),e.lineTo(_x*n,-n),e.closePath()}},Ar=-.5,Rr=ri(3)/2,h1=1/ri(12),Hre=(h1/2+1)*3,Gre={draw(e,t){const n=ri(t/Hre),r=n/2,i=n*h1,a=r,s=n*h1+n,u=-a,c=s;e.moveTo(r,i),e.lineTo(a,s),e.lineTo(u,c),e.lineTo(Ar*r-Rr*i,Rr*r+Ar*i),e.lineTo(Ar*a-Rr*s,Rr*a+Ar*s),e.lineTo(Ar*u-Rr*c,Rr*u+Ar*c),e.lineTo(Ar*r+Rr*i,Ar*i-Rr*r),e.lineTo(Ar*a+Rr*s,Ar*s-Rr*a),e.lineTo(Ar*u+Rr*c,Ar*c-Rr*u),e.closePath()}};function Wre(e,t){let n=null,r=zC(i);e=typeof e=="function"?e:Dt(e||VC),t=typeof t=="function"?t:Dt(t===void 0?64:+t);function i(){let a;if(n||(n=a=r()),e.apply(this,arguments).draw(n,+t.apply(this,arguments)),a)return n=null,a+""||null}return i.type=function(a){return arguments.length?(e=typeof a=="function"?a:Dt(a),i):e},i.size=function(a){return arguments.length?(t=typeof a=="function"?a:Dt(+a),i):t},i.context=function(a){return arguments.length?(n=a??null,i):n},i}function lm(){}function um(e,t,n){e._context.bezierCurveTo((2*e._x0+e._x1)/3,(2*e._y0+e._y1)/3,(e._x0+2*e._x1)/3,(e._y0+2*e._y1)/3,(e._x0+4*e._x1+t)/6,(e._y0+4*e._y1+n)/6)}function Gz(e){this._context=e}Gz.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._y0=this._y1=NaN,this._point=0},lineEnd:function(){switch(this._point){case 3:um(this,this._x1,this._y1);case 2:this._context.lineTo(this._x1,this._y1);break}(this._line||this._line!==0&&this._point===1)&&this._context.closePath(),this._line=1-this._line},point:function(e,t){switch(e=+e,t=+t,this._point){case 0:this._point=1,this._line?this._context.lineTo(e,t):this._context.moveTo(e,t);break;case 1:this._point=2;break;case 2:this._point=3,this._context.lineTo((5*this._x0+this._x1)/6,(5*this._y0+this._y1)/6);default:um(this,e,t);break}this._x0=this._x1,this._x1=e,this._y0=this._y1,this._y1=t}};function Kre(e){return new Gz(e)}function Wz(e){this._context=e}Wz.prototype={areaStart:lm,areaEnd:lm,lineStart:function(){this._x0=this._x1=this._x2=this._x3=this._x4=this._y0=this._y1=this._y2=this._y3=this._y4=NaN,this._point=0},lineEnd:function(){switch(this._point){case 1:{this._context.moveTo(this._x2,this._y2),this._context.closePath();break}case 2:{this._context.moveTo((this._x2+2*this._x3)/3,(this._y2+2*this._y3)/3),this._context.lineTo((this._x3+2*this._x2)/3,(this._y3+2*this._y2)/3),this._context.closePath();break}case 3:{this.point(this._x2,this._y2),this.point(this._x3,this._y3),this.point(this._x4,this._y4);break}}},point:function(e,t){switch(e=+e,t=+t,this._point){case 0:this._point=1,this._x2=e,this._y2=t;break;case 1:this._point=2,this._x3=e,this._y3=t;break;case 2:this._point=3,this._x4=e,this._y4=t,this._context.moveTo((this._x0+4*this._x1+e)/6,(this._y0+4*this._y1+t)/6);break;default:um(this,e,t);break}this._x0=this._x1,this._x1=e,this._y0=this._y1,this._y1=t}};function Yre(e){return new Wz(e)}function Kz(e){this._context=e}Kz.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._y0=this._y1=NaN,this._point=0},lineEnd:function(){(this._line||this._line!==0&&this._point===3)&&this._context.closePath(),this._line=1-this._line},point:function(e,t){switch(e=+e,t=+t,this._point){case 0:this._point=1;break;case 1:this._point=2;break;case 2:this._point=3;var n=(this._x0+4*this._x1+e)/6,r=(this._y0+4*this._y1+t)/6;this._line?this._context.lineTo(n,r):this._context.moveTo(n,r);break;case 3:this._point=4;default:um(this,e,t);break}this._x0=this._x1,this._x1=e,this._y0=this._y1,this._y1=t}};function Zre(e){return new Kz(e)}function Yz(e){this._context=e}Yz.prototype={areaStart:lm,areaEnd:lm,lineStart:function(){this._point=0},lineEnd:function(){this._point&&this._context.closePath()},point:function(e,t){e=+e,t=+t,this._point?this._context.lineTo(e,t):(this._point=1,this._context.moveTo(e,t))}};function Xre(e){return new Yz(e)}function TN(e){return e<0?-1:1}function MN(e,t,n){var r=e._x1-e._x0,i=t-e._x1,a=(e._y1-e._y0)/(r||i<0&&-0),s=(n-e._y1)/(i||r<0&&-0),u=(a*i+s*r)/(r+i);return(TN(a)+TN(s))*Math.min(Math.abs(a),Math.abs(s),.5*Math.abs(u))||0}function PN(e,t){var n=e._x1-e._x0;return n?(3*(e._y1-e._y0)/n-t)/2:t}function Ex(e,t,n){var r=e._x0,i=e._y0,a=e._x1,s=e._y1,u=(a-r)/3;e._context.bezierCurveTo(r+u,i+u*t,a-u,s-u*n,a,s)}function cm(e){this._context=e}cm.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._y0=this._y1=this._t0=NaN,this._point=0},lineEnd:function(){switch(this._point){case 2:this._context.lineTo(this._x1,this._y1);break;case 3:Ex(this,this._t0,PN(this,this._t0));break}(this._line||this._line!==0&&this._point===1)&&this._context.closePath(),this._line=1-this._line},point:function(e,t){var n=NaN;if(e=+e,t=+t,!(e===this._x1&&t===this._y1)){switch(this._point){case 0:this._point=1,this._line?this._context.lineTo(e,t):this._context.moveTo(e,t);break;case 1:this._point=2;break;case 2:this._point=3,Ex(this,PN(this,n=MN(this,e,t)),n);break;default:Ex(this,this._t0,n=MN(this,e,t));break}this._x0=this._x1,this._x1=e,this._y0=this._y1,this._y1=t,this._t0=n}}};function Zz(e){this._context=new Xz(e)}(Zz.prototype=Object.create(cm.prototype)).point=function(e,t){cm.prototype.point.call(this,t,e)};function Xz(e){this._context=e}Xz.prototype={moveTo:function(e,t){this._context.moveTo(t,e)},closePath:function(){this._context.closePath()},lineTo:function(e,t){this._context.lineTo(t,e)},bezierCurveTo:function(e,t,n,r,i,a){this._context.bezierCurveTo(t,e,r,n,a,i)}};function Qre(e){return new cm(e)}function Jre(e){return new Zz(e)}function Qz(e){this._context=e}Qz.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x=[],this._y=[]},lineEnd:function(){var e=this._x,t=this._y,n=e.length;if(n)if(this._line?this._context.lineTo(e[0],t[0]):this._context.moveTo(e[0],t[0]),n===2)this._context.lineTo(e[1],t[1]);else for(var r=jN(e),i=jN(t),a=0,s=1;s<n;++a,++s)this._context.bezierCurveTo(r[0][a],i[0][a],r[1][a],i[1][a],e[s],t[s]);(this._line||this._line!==0&&n===1)&&this._context.closePath(),this._line=1-this._line,this._x=this._y=null},point:function(e,t){this._x.push(+e),this._y.push(+t)}};function jN(e){var t,n=e.length-1,r,i=new Array(n),a=new Array(n),s=new Array(n);for(i[0]=0,a[0]=2,s[0]=e[0]+2*e[1],t=1;t<n-1;++t)i[t]=1,a[t]=4,s[t]=4*e[t]+2*e[t+1];for(i[n-1]=2,a[n-1]=7,s[n-1]=8*e[n-1]+e[n],t=1;t<n;++t)r=i[t]/a[t-1],a[t]-=r,s[t]-=r*s[t-1];for(i[n-1]=s[n-1]/a[n-1],t=n-2;t>=0;--t)i[t]=(s[t]-i[t+1])/a[t];for(a[n-1]=(e[n]+i[n-1])/2,t=0;t<n-1;++t)a[t]=2*e[t+1]-i[t+1];return[i,a]}function eie(e){return new Qz(e)}function Bv(e,t){this._context=e,this._t=t}Bv.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x=this._y=NaN,this._point=0},lineEnd:function(){0<this._t&&this._t<1&&this._point===2&&this._context.lineTo(this._x,this._y),(this._line||this._line!==0&&this._point===1)&&this._context.closePath(),this._line>=0&&(this._t=1-this._t,this._line=1-this._line)},point:function(e,t){switch(e=+e,t=+t,this._point){case 0:this._point=1,this._line?this._context.lineTo(e,t):this._context.moveTo(e,t);break;case 1:this._point=2;default:{if(this._t<=0)this._context.lineTo(this._x,t),this._context.lineTo(e,t);else{var n=this._x*(1-this._t)+e*this._t;this._context.lineTo(n,this._y),this._context.lineTo(n,t)}break}}this._x=e,this._y=t}};function tie(e){return new Bv(e,.5)}function nie(e){return new Bv(e,0)}function rie(e){return new Bv(e,1)}function $l(e,t){if((s=e.length)>1)for(var n=1,r,i,a=e[t[0]],s,u=a.length;n<s;++n)for(i=a,a=e[t[n]],r=0;r<u;++r)a[r][1]+=a[r][0]=isNaN(i[r][1])?i[r][0]:i[r][1]}function m1(e){for(var t=e.length,n=new Array(t);--t>=0;)n[t]=t;return n}function iie(e,t){return e[t]}function aie(e){const t=[];return t.key=e,t}function oie(){var e=Dt([]),t=m1,n=$l,r=iie;function i(a){var s=Array.from(e.apply(this,arguments),aie),u,c=s.length,d=-1,p;for(const m of a)for(u=0,++d;u<c;++u)(s[u][d]=[0,+r(m,s[u].key,d,a)]).data=m;for(u=0,p=UC(t(s));u<c;++u)s[p[u]].index=u;return n(s,p),s}return i.keys=function(a){return arguments.length?(e=typeof a=="function"?a:Dt(Array.from(a)),i):e},i.value=function(a){return arguments.length?(r=typeof a=="function"?a:Dt(+a),i):r},i.order=function(a){return arguments.length?(t=a==null?m1:typeof a=="function"?a:Dt(Array.from(a)),i):t},i.offset=function(a){return arguments.length?(n=a??$l,i):n},i}function sie(e,t){if((r=e.length)>0){for(var n,r,i=0,a=e[0].length,s;i<a;++i){for(s=n=0;n<r;++n)s+=e[n][i][1]||0;if(s)for(n=0;n<r;++n)e[n][i][1]/=s}$l(e,t)}}function lie(e,t){if((i=e.length)>0){for(var n=0,r=e[t[0]],i,a=r.length;n<a;++n){for(var s=0,u=0;s<i;++s)u+=e[s][n][1]||0;r[n][1]+=r[n][0]=-u/2}$l(e,t)}}function uie(e,t){if(!(!((s=e.length)>0)||!((a=(i=e[t[0]]).length)>0))){for(var n=0,r=1,i,a,s;r<a;++r){for(var u=0,c=0,d=0;u<s;++u){for(var p=e[t[u]],m=p[r][1]||0,v=p[r-1][1]||0,b=(m-v)/2,S=0;S<u;++S){var w=e[t[S]],x=w[r][1]||0,C=w[r-1][1]||0;b+=x-C}c+=m,d+=b*m}i[r-1][1]+=i[r-1][0]=n,c&&(n-=d/c)}i[r-1][1]+=i[r-1][0]=n,$l(e,t)}}function wf(e){"@babel/helpers - typeof";return wf=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},wf(e)}var cie=["type","size","sizeType"];function v1(){return v1=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)Object.prototype.hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},v1.apply(this,arguments)}function NN(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function DN(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?NN(Object(n),!0).forEach(function(r){fie(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):NN(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function fie(e,t,n){return t=die(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function die(e){var t=pie(e,"string");return wf(t)=="symbol"?t:t+""}function pie(e,t){if(wf(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(wf(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function hie(e,t){if(e==null)return{};var n=mie(e,t),r,i;if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);for(i=0;i<a.length;i++)r=a[i],!(t.indexOf(r)>=0)&&Object.prototype.propertyIsEnumerable.call(e,r)&&(n[r]=e[r])}return n}function mie(e,t){if(e==null)return{};var n={};for(var r in e)if(Object.prototype.hasOwnProperty.call(e,r)){if(t.indexOf(r)>=0)continue;n[r]=e[r]}return n}var Jz={symbolCircle:VC,symbolCross:$re,symbolDiamond:Lre,symbolSquare:Bre,symbolStar:Ure,symbolTriangle:Vre,symbolWye:Gre},vie=Math.PI/180,gie=function(t){var n="symbol".concat($v(t));return Jz[n]||VC},yie=function(t,n,r){if(n==="area")return t;switch(r){case"cross":return 5*t*t/9;case"diamond":return .5*t*t/Math.sqrt(3);case"square":return t*t;case"star":{var i=18*vie;return 1.25*t*t*(Math.tan(i)-Math.tan(i*2)*Math.pow(Math.tan(i),2))}case"triangle":return Math.sqrt(3)*t*t/4;case"wye":return(21-10*Math.sqrt(3))*t*t/8;default:return Math.PI*t*t/4}},bie=function(t,n){Jz["symbol".concat($v(t))]=n},eU=function(t){var n=t.type,r=n===void 0?"circle":n,i=t.size,a=i===void 0?64:i,s=t.sizeType,u=s===void 0?"area":s,c=hie(t,cie),d=DN(DN({},c),{},{type:r,size:a,sizeType:u}),p=function(){var x=gie(r),C=Wre().type(x).size(yie(a,u,r));return C()},m=d.className,v=d.cx,b=d.cy,S=Je(d,!0);return v===+v&&b===+b&&a===+a?V.createElement("path",v1({},S,{className:ct("recharts-symbols",m),transform:"translate(".concat(v,", ").concat(b,")"),d:p()})):null};eU.registerSymbol=bie;function Il(e){"@babel/helpers - typeof";return Il=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},Il(e)}function g1(){return g1=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)Object.prototype.hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},g1.apply(this,arguments)}function kN(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function xie(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?kN(Object(n),!0).forEach(function(r){Sf(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):kN(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function wie(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}function Sie(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,nU(r.key),r)}}function _ie(e,t,n){return t&&Sie(e.prototype,t),Object.defineProperty(e,"prototype",{writable:!1}),e}function Eie(e,t,n){return t=fm(t),Cie(e,tU()?Reflect.construct(t,n||[],fm(e).constructor):t.apply(e,n))}function Cie(e,t){if(t&&(Il(t)==="object"||typeof t=="function"))return t;if(t!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return Oie(e)}function Oie(e){if(e===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return e}function tU(){try{var e=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(tU=function(){return!!e})()}function fm(e){return fm=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(n){return n.__proto__||Object.getPrototypeOf(n)},fm(e)}function Aie(e,t){if(typeof t!="function"&&t!==null)throw new TypeError("Super expression must either be null or a function");e.prototype=Object.create(t&&t.prototype,{constructor:{value:e,writable:!0,configurable:!0}}),Object.defineProperty(e,"prototype",{writable:!1}),t&&y1(e,t)}function y1(e,t){return y1=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(r,i){return r.__proto__=i,r},y1(e,t)}function Sf(e,t,n){return t=nU(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function nU(e){var t=Rie(e,"string");return Il(t)=="symbol"?t:t+""}function Rie(e,t){if(Il(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(Il(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(e)}var qr=32,HC=function(e){function t(){return wie(this,t),Eie(this,t,arguments)}return Aie(t,e),_ie(t,[{key:"renderIcon",value:function(r){var i=this.props.inactiveColor,a=qr/2,s=qr/6,u=qr/3,c=r.inactive?i:r.color;if(r.type==="plainline")return V.createElement("line",{strokeWidth:4,fill:"none",stroke:c,strokeDasharray:r.payload.strokeDasharray,x1:0,y1:a,x2:qr,y2:a,className:"recharts-legend-icon"});if(r.type==="line")return V.createElement("path",{strokeWidth:4,fill:"none",stroke:c,d:"M0,".concat(a,"h").concat(u,`
            A`).concat(s,",").concat(s,",0,1,1,").concat(2*u,",").concat(a,`
            H`).concat(qr,"M").concat(2*u,",").concat(a,`
            A`).concat(s,",").concat(s,",0,1,1,").concat(u,",").concat(a),className:"recharts-legend-icon"});if(r.type==="rect")return V.createElement("path",{stroke:"none",fill:c,d:"M0,".concat(qr/8,"h").concat(qr,"v").concat(qr*3/4,"h").concat(-32,"z"),className:"recharts-legend-icon"});if(V.isValidElement(r.legendIcon)){var d=xie({},r);return delete d.legendIcon,V.cloneElement(r.legendIcon,d)}return V.createElement(eU,{fill:c,cx:a,cy:a,size:qr,sizeType:"diameter",type:r.type})}},{key:"renderItems",value:function(){var r=this,i=this.props,a=i.payload,s=i.iconSize,u=i.layout,c=i.formatter,d=i.inactiveColor,p={x:0,y:0,width:qr,height:qr},m={display:u==="horizontal"?"inline-block":"block",marginRight:10},v={display:"inline-block",verticalAlign:"middle",marginRight:4};return a.map(function(b,S){var w=b.formatter||c,x=ct(Sf(Sf({"recharts-legend-item":!0},"legend-item-".concat(S),!0),"inactive",b.inactive));if(b.type==="none")return null;var C=st(b.value)?null:b.value;es(!st(b.value),`The name property is also required when using a function for the dataKey of a chart's cartesian components. Ex: <Bar name="Name of my Data"/>`);var O=b.inactive?d:b.color;return V.createElement("li",g1({className:x,style:m,key:"legend-item-".concat(S)},FC(r.props,b,S)),V.createElement(c1,{width:s,height:s,viewBox:p,style:v},r.renderIcon(b)),V.createElement("span",{className:"recharts-legend-item-text",style:{color:O}},w?w(C,b,S):C))})}},{key:"render",value:function(){var r=this.props,i=r.payload,a=r.layout,s=r.align;if(!i||!i.length)return null;var u={padding:0,margin:0,textAlign:a==="horizontal"?s:"left"};return V.createElement("ul",{className:"recharts-default-legend",style:u},this.renderItems())}}])}(g.PureComponent);Sf(HC,"displayName","Legend");Sf(HC,"defaultProps",{iconSize:14,layout:"horizontal",align:"center",verticalAlign:"middle",inactiveColor:"#ccc"});var Cx,$N;function Tie(){if($N)return Cx;$N=1;var e=Nv();function t(){this.__data__=new e,this.size=0}return Cx=t,Cx}var Ox,IN;function Mie(){if(IN)return Ox;IN=1;function e(t){var n=this.__data__,r=n.delete(t);return this.size=n.size,r}return Ox=e,Ox}var Ax,LN;function Pie(){if(LN)return Ax;LN=1;function e(t){return this.__data__.get(t)}return Ax=e,Ax}var Rx,BN;function jie(){if(BN)return Rx;BN=1;function e(t){return this.__data__.has(t)}return Rx=e,Rx}var Tx,FN;function Nie(){if(FN)return Tx;FN=1;var e=Nv(),t=DC(),n=kC(),r=200;function i(a,s){var u=this.__data__;if(u instanceof e){var c=u.__data__;if(!t||c.length<r-1)return c.push([a,s]),this.size=++u.size,this;u=this.__data__=new n(c)}return u.set(a,s),this.size=u.size,this}return Tx=i,Tx}var Mx,qN;function rU(){if(qN)return Mx;qN=1;var e=Nv(),t=Tie(),n=Mie(),r=Pie(),i=jie(),a=Nie();function s(u){var c=this.__data__=new e(u);this.size=c.size}return s.prototype.clear=t,s.prototype.delete=n,s.prototype.get=r,s.prototype.has=i,s.prototype.set=a,Mx=s,Mx}var Px,zN;function Die(){if(zN)return Px;zN=1;var e="__lodash_hash_undefined__";function t(n){return this.__data__.set(n,e),this}return Px=t,Px}var jx,UN;function kie(){if(UN)return jx;UN=1;function e(t){return this.__data__.has(t)}return jx=e,jx}var Nx,VN;function iU(){if(VN)return Nx;VN=1;var e=kC(),t=Die(),n=kie();function r(i){var a=-1,s=i==null?0:i.length;for(this.__data__=new e;++a<s;)this.add(i[a])}return r.prototype.add=r.prototype.push=t,r.prototype.has=n,Nx=r,Nx}var Dx,HN;function aU(){if(HN)return Dx;HN=1;function e(t,n){for(var r=-1,i=t==null?0:t.length;++r<i;)if(n(t[r],r,t))return!0;return!1}return Dx=e,Dx}var kx,GN;function oU(){if(GN)return kx;GN=1;function e(t,n){return t.has(n)}return kx=e,kx}var $x,WN;function sU(){if(WN)return $x;WN=1;var e=iU(),t=aU(),n=oU(),r=1,i=2;function a(s,u,c,d,p,m){var v=c&r,b=s.length,S=u.length;if(b!=S&&!(v&&S>b))return!1;var w=m.get(s),x=m.get(u);if(w&&x)return w==u&&x==s;var C=-1,O=!0,A=c&i?new e:void 0;for(m.set(s,u),m.set(u,s);++C<b;){var T=s[C],R=u[C];if(d)var M=v?d(R,T,C,u,s,m):d(T,R,C,s,u,m);if(M!==void 0){if(M)continue;O=!1;break}if(A){if(!t(u,function(P,N){if(!n(A,N)&&(T===P||p(T,P,c,d,m)))return A.push(N)})){O=!1;break}}else if(!(T===R||p(T,R,c,d,m))){O=!1;break}}return m.delete(s),m.delete(u),O}return $x=a,$x}var Ix,KN;function $ie(){if(KN)return Ix;KN=1;var e=Mi(),t=e.Uint8Array;return Ix=t,Ix}var Lx,YN;function Iie(){if(YN)return Lx;YN=1;function e(t){var n=-1,r=Array(t.size);return t.forEach(function(i,a){r[++n]=[a,i]}),r}return Lx=e,Lx}var Bx,ZN;function GC(){if(ZN)return Bx;ZN=1;function e(t){var n=-1,r=Array(t.size);return t.forEach(function(i){r[++n]=i}),r}return Bx=e,Bx}var Fx,XN;function Lie(){if(XN)return Fx;XN=1;var e=bd(),t=$ie(),n=NC(),r=sU(),i=Iie(),a=GC(),s=1,u=2,c="[object Boolean]",d="[object Date]",p="[object Error]",m="[object Map]",v="[object Number]",b="[object RegExp]",S="[object Set]",w="[object String]",x="[object Symbol]",C="[object ArrayBuffer]",O="[object DataView]",A=e?e.prototype:void 0,T=A?A.valueOf:void 0;function R(M,P,N,k,H,L,F){switch(N){case O:if(M.byteLength!=P.byteLength||M.byteOffset!=P.byteOffset)return!1;M=M.buffer,P=P.buffer;case C:return!(M.byteLength!=P.byteLength||!L(new t(M),new t(P)));case c:case d:case v:return n(+M,+P);case p:return M.name==P.name&&M.message==P.message;case b:case w:return M==P+"";case m:var z=i;case S:var Z=k&s;if(z||(z=a),M.size!=P.size&&!Z)return!1;var Y=F.get(M);if(Y)return Y==P;k|=u,F.set(M,P);var I=r(z(M),z(P),k,H,L,F);return F.delete(M),I;case x:if(T)return T.call(M)==T.call(P)}return!1}return Fx=R,Fx}var qx,QN;function lU(){if(QN)return qx;QN=1;function e(t,n){for(var r=-1,i=n.length,a=t.length;++r<i;)t[a+r]=n[r];return t}return qx=e,qx}var zx,JN;function Bie(){if(JN)return zx;JN=1;var e=lU(),t=Qn();function n(r,i,a){var s=i(r);return t(r)?s:e(s,a(r))}return zx=n,zx}var Ux,eD;function Fie(){if(eD)return Ux;eD=1;function e(t,n){for(var r=-1,i=t==null?0:t.length,a=0,s=[];++r<i;){var u=t[r];n(u,r,t)&&(s[a++]=u)}return s}return Ux=e,Ux}var Vx,tD;function qie(){if(tD)return Vx;tD=1;function e(){return[]}return Vx=e,Vx}var Hx,nD;function zie(){if(nD)return Hx;nD=1;var e=Fie(),t=qie(),n=Object.prototype,r=n.propertyIsEnumerable,i=Object.getOwnPropertySymbols,a=i?function(s){return s==null?[]:(s=Object(s),e(i(s),function(u){return r.call(s,u)}))}:t;return Hx=a,Hx}var Gx,rD;function Uie(){if(rD)return Gx;rD=1;function e(t,n){for(var r=-1,i=Array(t);++r<t;)i[r]=n(r);return i}return Gx=e,Gx}var Wx,iD;function Vie(){if(iD)return Wx;iD=1;var e=xa(),t=wa(),n="[object Arguments]";function r(i){return t(i)&&e(i)==n}return Wx=r,Wx}var Kx,aD;function WC(){if(aD)return Kx;aD=1;var e=Vie(),t=wa(),n=Object.prototype,r=n.hasOwnProperty,i=n.propertyIsEnumerable,a=e(function(){return arguments}())?e:function(s){return t(s)&&r.call(s,"callee")&&!i.call(s,"callee")};return Kx=a,Kx}var Fc={exports:{}},Yx,oD;function Hie(){if(oD)return Yx;oD=1;function e(){return!1}return Yx=e,Yx}Fc.exports;var sD;function uU(){return sD||(sD=1,function(e,t){var n=Mi(),r=Hie(),i=t&&!t.nodeType&&t,a=i&&!0&&e&&!e.nodeType&&e,s=a&&a.exports===i,u=s?n.Buffer:void 0,c=u?u.isBuffer:void 0,d=c||r;e.exports=d}(Fc,Fc.exports)),Fc.exports}var Zx,lD;function KC(){if(lD)return Zx;lD=1;var e=9007199254740991,t=/^(?:0|[1-9]\d*)$/;function n(r,i){var a=typeof r;return i=i??e,!!i&&(a=="number"||a!="symbol"&&t.test(r))&&r>-1&&r%1==0&&r<i}return Zx=n,Zx}var Xx,uD;function YC(){if(uD)return Xx;uD=1;var e=9007199254740991;function t(n){return typeof n=="number"&&n>-1&&n%1==0&&n<=e}return Xx=t,Xx}var Qx,cD;function Gie(){if(cD)return Qx;cD=1;var e=xa(),t=YC(),n=wa(),r="[object Arguments]",i="[object Array]",a="[object Boolean]",s="[object Date]",u="[object Error]",c="[object Function]",d="[object Map]",p="[object Number]",m="[object Object]",v="[object RegExp]",b="[object Set]",S="[object String]",w="[object WeakMap]",x="[object ArrayBuffer]",C="[object DataView]",O="[object Float32Array]",A="[object Float64Array]",T="[object Int8Array]",R="[object Int16Array]",M="[object Int32Array]",P="[object Uint8Array]",N="[object Uint8ClampedArray]",k="[object Uint16Array]",H="[object Uint32Array]",L={};L[O]=L[A]=L[T]=L[R]=L[M]=L[P]=L[N]=L[k]=L[H]=!0,L[r]=L[i]=L[x]=L[a]=L[C]=L[s]=L[u]=L[c]=L[d]=L[p]=L[m]=L[v]=L[b]=L[S]=L[w]=!1;function F(z){return n(z)&&t(z.length)&&!!L[e(z)]}return Qx=F,Qx}var Jx,fD;function cU(){if(fD)return Jx;fD=1;function e(t){return function(n){return t(n)}}return Jx=e,Jx}var qc={exports:{}};qc.exports;var dD;function Wie(){return dD||(dD=1,function(e,t){var n=Tz(),r=t&&!t.nodeType&&t,i=r&&!0&&e&&!e.nodeType&&e,a=i&&i.exports===r,s=a&&n.process,u=function(){try{var c=i&&i.require&&i.require("util").types;return c||s&&s.binding&&s.binding("util")}catch{}}();e.exports=u}(qc,qc.exports)),qc.exports}var ew,pD;function fU(){if(pD)return ew;pD=1;var e=Gie(),t=cU(),n=Wie(),r=n&&n.isTypedArray,i=r?t(r):e;return ew=i,ew}var tw,hD;function Kie(){if(hD)return tw;hD=1;var e=Uie(),t=WC(),n=Qn(),r=uU(),i=KC(),a=fU(),s=Object.prototype,u=s.hasOwnProperty;function c(d,p){var m=n(d),v=!m&&t(d),b=!m&&!v&&r(d),S=!m&&!v&&!b&&a(d),w=m||v||b||S,x=w?e(d.length,String):[],C=x.length;for(var O in d)(p||u.call(d,O))&&!(w&&(O=="length"||b&&(O=="offset"||O=="parent")||S&&(O=="buffer"||O=="byteLength"||O=="byteOffset")||i(O,C)))&&x.push(O);return x}return tw=c,tw}var nw,mD;function Yie(){if(mD)return nw;mD=1;var e=Object.prototype;function t(n){var r=n&&n.constructor,i=typeof r=="function"&&r.prototype||e;return n===i}return nw=t,nw}var rw,vD;function dU(){if(vD)return rw;vD=1;function e(t,n){return function(r){return t(n(r))}}return rw=e,rw}var iw,gD;function Zie(){if(gD)return iw;gD=1;var e=dU(),t=e(Object.keys,Object);return iw=t,iw}var aw,yD;function Xie(){if(yD)return aw;yD=1;var e=Yie(),t=Zie(),n=Object.prototype,r=n.hasOwnProperty;function i(a){if(!e(a))return t(a);var s=[];for(var u in Object(a))r.call(a,u)&&u!="constructor"&&s.push(u);return s}return aw=i,aw}var ow,bD;function Sd(){if(bD)return ow;bD=1;var e=jC(),t=YC();function n(r){return r!=null&&t(r.length)&&!e(r)}return ow=n,ow}var sw,xD;function Fv(){if(xD)return sw;xD=1;var e=Kie(),t=Xie(),n=Sd();function r(i){return n(i)?e(i):t(i)}return sw=r,sw}var lw,wD;function Qie(){if(wD)return lw;wD=1;var e=Bie(),t=zie(),n=Fv();function r(i){return e(i,n,t)}return lw=r,lw}var uw,SD;function Jie(){if(SD)return uw;SD=1;var e=Qie(),t=1,n=Object.prototype,r=n.hasOwnProperty;function i(a,s,u,c,d,p){var m=u&t,v=e(a),b=v.length,S=e(s),w=S.length;if(b!=w&&!m)return!1;for(var x=b;x--;){var C=v[x];if(!(m?C in s:r.call(s,C)))return!1}var O=p.get(a),A=p.get(s);if(O&&A)return O==s&&A==a;var T=!0;p.set(a,s),p.set(s,a);for(var R=m;++x<b;){C=v[x];var M=a[C],P=s[C];if(c)var N=m?c(P,M,C,s,a,p):c(M,P,C,a,s,p);if(!(N===void 0?M===P||d(M,P,u,c,p):N)){T=!1;break}R||(R=C=="constructor")}if(T&&!R){var k=a.constructor,H=s.constructor;k!=H&&"constructor"in a&&"constructor"in s&&!(typeof k=="function"&&k instanceof k&&typeof H=="function"&&H instanceof H)&&(T=!1)}return p.delete(a),p.delete(s),T}return uw=i,uw}var cw,_D;function eae(){if(_D)return cw;_D=1;var e=gs(),t=Mi(),n=e(t,"DataView");return cw=n,cw}var fw,ED;function tae(){if(ED)return fw;ED=1;var e=gs(),t=Mi(),n=e(t,"Promise");return fw=n,fw}var dw,CD;function pU(){if(CD)return dw;CD=1;var e=gs(),t=Mi(),n=e(t,"Set");return dw=n,dw}var pw,OD;function nae(){if(OD)return pw;OD=1;var e=gs(),t=Mi(),n=e(t,"WeakMap");return pw=n,pw}var hw,AD;function rae(){if(AD)return hw;AD=1;var e=eae(),t=DC(),n=tae(),r=pU(),i=nae(),a=xa(),s=Mz(),u="[object Map]",c="[object Object]",d="[object Promise]",p="[object Set]",m="[object WeakMap]",v="[object DataView]",b=s(e),S=s(t),w=s(n),x=s(r),C=s(i),O=a;return(e&&O(new e(new ArrayBuffer(1)))!=v||t&&O(new t)!=u||n&&O(n.resolve())!=d||r&&O(new r)!=p||i&&O(new i)!=m)&&(O=function(A){var T=a(A),R=T==c?A.constructor:void 0,M=R?s(R):"";if(M)switch(M){case b:return v;case S:return u;case w:return d;case x:return p;case C:return m}return T}),hw=O,hw}var mw,RD;function iae(){if(RD)return mw;RD=1;var e=rU(),t=sU(),n=Lie(),r=Jie(),i=rae(),a=Qn(),s=uU(),u=fU(),c=1,d="[object Arguments]",p="[object Array]",m="[object Object]",v=Object.prototype,b=v.hasOwnProperty;function S(w,x,C,O,A,T){var R=a(w),M=a(x),P=R?p:i(w),N=M?p:i(x);P=P==d?m:P,N=N==d?m:N;var k=P==m,H=N==m,L=P==N;if(L&&s(w)){if(!s(x))return!1;R=!0,k=!1}if(L&&!k)return T||(T=new e),R||u(w)?t(w,x,C,O,A,T):n(w,x,P,C,O,A,T);if(!(C&c)){var F=k&&b.call(w,"__wrapped__"),z=H&&b.call(x,"__wrapped__");if(F||z){var Z=F?w.value():w,Y=z?x.value():x;return T||(T=new e),A(Z,Y,C,O,T)}}return L?(T||(T=new e),r(w,x,C,O,A,T)):!1}return mw=S,mw}var vw,TD;function ZC(){if(TD)return vw;TD=1;var e=iae(),t=wa();function n(r,i,a,s,u){return r===i?!0:r==null||i==null||!t(r)&&!t(i)?r!==r&&i!==i:e(r,i,a,s,n,u)}return vw=n,vw}var gw,MD;function aae(){if(MD)return gw;MD=1;var e=rU(),t=ZC(),n=1,r=2;function i(a,s,u,c){var d=u.length,p=d,m=!c;if(a==null)return!p;for(a=Object(a);d--;){var v=u[d];if(m&&v[2]?v[1]!==a[v[0]]:!(v[0]in a))return!1}for(;++d<p;){v=u[d];var b=v[0],S=a[b],w=v[1];if(m&&v[2]){if(S===void 0&&!(b in a))return!1}else{var x=new e;if(c)var C=c(S,w,b,a,s,x);if(!(C===void 0?t(w,S,n|r,c,x):C))return!1}}return!0}return gw=i,gw}var yw,PD;function hU(){if(PD)return yw;PD=1;var e=mo();function t(n){return n===n&&!e(n)}return yw=t,yw}var bw,jD;function oae(){if(jD)return bw;jD=1;var e=hU(),t=Fv();function n(r){for(var i=t(r),a=i.length;a--;){var s=i[a],u=r[s];i[a]=[s,u,e(u)]}return i}return bw=n,bw}var xw,ND;function mU(){if(ND)return xw;ND=1;function e(t,n){return function(r){return r==null?!1:r[t]===n&&(n!==void 0||t in Object(r))}}return xw=e,xw}var ww,DD;function sae(){if(DD)return ww;DD=1;var e=aae(),t=oae(),n=mU();function r(i){var a=t(i);return a.length==1&&a[0][2]?n(a[0][0],a[0][1]):function(s){return s===i||e(s,i,a)}}return ww=r,ww}var Sw,kD;function lae(){if(kD)return Sw;kD=1;function e(t,n){return t!=null&&n in Object(t)}return Sw=e,Sw}var _w,$D;function uae(){if($D)return _w;$D=1;var e=Nz(),t=WC(),n=Qn(),r=KC(),i=YC(),a=kv();function s(u,c,d){c=e(c,u);for(var p=-1,m=c.length,v=!1;++p<m;){var b=a(c[p]);if(!(v=u!=null&&d(u,b)))break;u=u[b]}return v||++p!=m?v:(m=u==null?0:u.length,!!m&&i(m)&&r(b,m)&&(n(u)||t(u)))}return _w=s,_w}var Ew,ID;function cae(){if(ID)return Ew;ID=1;var e=lae(),t=uae();function n(r,i){return r!=null&&t(r,i,e)}return Ew=n,Ew}var Cw,LD;function fae(){if(LD)return Cw;LD=1;var e=ZC(),t=Dz(),n=cae(),r=PC(),i=hU(),a=mU(),s=kv(),u=1,c=2;function d(p,m){return r(p)&&i(m)?a(s(p),m):function(v){var b=t(v,p);return b===void 0&&b===m?n(v,p):e(m,b,u|c)}}return Cw=d,Cw}var Ow,BD;function bu(){if(BD)return Ow;BD=1;function e(t){return t}return Ow=e,Ow}var Aw,FD;function dae(){if(FD)return Aw;FD=1;function e(t){return function(n){return n==null?void 0:n[t]}}return Aw=e,Aw}var Rw,qD;function pae(){if(qD)return Rw;qD=1;var e=IC();function t(n){return function(r){return e(r,n)}}return Rw=t,Rw}var Tw,zD;function hae(){if(zD)return Tw;zD=1;var e=dae(),t=pae(),n=PC(),r=kv();function i(a){return n(a)?e(r(a)):t(a)}return Tw=i,Tw}var Mw,UD;function Pi(){if(UD)return Mw;UD=1;var e=sae(),t=fae(),n=bu(),r=Qn(),i=hae();function a(s){return typeof s=="function"?s:s==null?n:typeof s=="object"?r(s)?t(s[0],s[1]):e(s):i(s)}return Mw=a,Mw}var Pw,VD;function vU(){if(VD)return Pw;VD=1;function e(t,n,r,i){for(var a=t.length,s=r+(i?1:-1);i?s--:++s<a;)if(n(t[s],s,t))return s;return-1}return Pw=e,Pw}var jw,HD;function mae(){if(HD)return jw;HD=1;function e(t){return t!==t}return jw=e,jw}var Nw,GD;function vae(){if(GD)return Nw;GD=1;function e(t,n,r){for(var i=r-1,a=t.length;++i<a;)if(t[i]===n)return i;return-1}return Nw=e,Nw}var Dw,WD;function gae(){if(WD)return Dw;WD=1;var e=vU(),t=mae(),n=vae();function r(i,a,s){return a===a?n(i,a,s):e(i,t,s)}return Dw=r,Dw}var kw,KD;function yae(){if(KD)return kw;KD=1;var e=gae();function t(n,r){var i=n==null?0:n.length;return!!i&&e(n,r,0)>-1}return kw=t,kw}var $w,YD;function bae(){if(YD)return $w;YD=1;function e(t,n,r){for(var i=-1,a=t==null?0:t.length;++i<a;)if(r(n,t[i]))return!0;return!1}return $w=e,$w}var Iw,ZD;function xae(){if(ZD)return Iw;ZD=1;function e(){}return Iw=e,Iw}var Lw,XD;function wae(){if(XD)return Lw;XD=1;var e=pU(),t=xae(),n=GC(),r=1/0,i=e&&1/n(new e([,-0]))[1]==r?function(a){return new e(a)}:t;return Lw=i,Lw}var Bw,QD;function Sae(){if(QD)return Bw;QD=1;var e=iU(),t=yae(),n=bae(),r=oU(),i=wae(),a=GC(),s=200;function u(c,d,p){var m=-1,v=t,b=c.length,S=!0,w=[],x=w;if(p)S=!1,v=n;else if(b>=s){var C=d?null:i(c);if(C)return a(C);S=!1,v=r,x=new e}else x=d?[]:w;e:for(;++m<b;){var O=c[m],A=d?d(O):O;if(O=p||O!==0?O:0,S&&A===A){for(var T=x.length;T--;)if(x[T]===A)continue e;d&&x.push(A),w.push(O)}else v(x,A,p)||(x!==w&&x.push(A),w.push(O))}return w}return Bw=u,Bw}var Fw,JD;function _ae(){if(JD)return Fw;JD=1;var e=Pi(),t=Sae();function n(r,i){return r&&r.length?t(r,e(i,2)):[]}return Fw=n,Fw}var Eae=_ae();const e2=Rt(Eae);function gU(e,t,n){return t===!0?e2(e,n):st(t)?e2(e,t):e}function Ll(e){"@babel/helpers - typeof";return Ll=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},Ll(e)}var Cae=["ref"];function t2(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function Zi(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?t2(Object(n),!0).forEach(function(r){qv(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):t2(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function Oae(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}function n2(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,bU(r.key),r)}}function Aae(e,t,n){return t&&n2(e.prototype,t),n&&n2(e,n),Object.defineProperty(e,"prototype",{writable:!1}),e}function Rae(e,t,n){return t=dm(t),Tae(e,yU()?Reflect.construct(t,n||[],dm(e).constructor):t.apply(e,n))}function Tae(e,t){if(t&&(Ll(t)==="object"||typeof t=="function"))return t;if(t!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return Mae(e)}function Mae(e){if(e===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return e}function yU(){try{var e=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(yU=function(){return!!e})()}function dm(e){return dm=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(n){return n.__proto__||Object.getPrototypeOf(n)},dm(e)}function Pae(e,t){if(typeof t!="function"&&t!==null)throw new TypeError("Super expression must either be null or a function");e.prototype=Object.create(t&&t.prototype,{constructor:{value:e,writable:!0,configurable:!0}}),Object.defineProperty(e,"prototype",{writable:!1}),t&&b1(e,t)}function b1(e,t){return b1=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(r,i){return r.__proto__=i,r},b1(e,t)}function qv(e,t,n){return t=bU(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function bU(e){var t=jae(e,"string");return Ll(t)=="symbol"?t:t+""}function jae(e,t){if(Ll(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(Ll(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(e)}function Nae(e,t){if(e==null)return{};var n=Dae(e,t),r,i;if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);for(i=0;i<a.length;i++)r=a[i],!(t.indexOf(r)>=0)&&Object.prototype.propertyIsEnumerable.call(e,r)&&(n[r]=e[r])}return n}function Dae(e,t){if(e==null)return{};var n={};for(var r in e)if(Object.prototype.hasOwnProperty.call(e,r)){if(t.indexOf(r)>=0)continue;n[r]=e[r]}return n}function kae(e){return e.value}function $ae(e,t){if(V.isValidElement(e))return V.cloneElement(e,t);if(typeof e=="function")return V.createElement(e,t);t.ref;var n=Nae(t,Cae);return V.createElement(HC,n)}var r2=1,Al=function(e){function t(){var n;Oae(this,t);for(var r=arguments.length,i=new Array(r),a=0;a<r;a++)i[a]=arguments[a];return n=Rae(this,t,[].concat(i)),qv(n,"lastBoundingBox",{width:-1,height:-1}),n}return Pae(t,e),Aae(t,[{key:"componentDidMount",value:function(){this.updateBBox()}},{key:"componentDidUpdate",value:function(){this.updateBBox()}},{key:"getBBox",value:function(){if(this.wrapperNode&&this.wrapperNode.getBoundingClientRect){var r=this.wrapperNode.getBoundingClientRect();return r.height=this.wrapperNode.offsetHeight,r.width=this.wrapperNode.offsetWidth,r}return null}},{key:"updateBBox",value:function(){var r=this.props.onBBoxUpdate,i=this.getBBox();i?(Math.abs(i.width-this.lastBoundingBox.width)>r2||Math.abs(i.height-this.lastBoundingBox.height)>r2)&&(this.lastBoundingBox.width=i.width,this.lastBoundingBox.height=i.height,r&&r(i)):(this.lastBoundingBox.width!==-1||this.lastBoundingBox.height!==-1)&&(this.lastBoundingBox.width=-1,this.lastBoundingBox.height=-1,r&&r(null))}},{key:"getBBoxSnapshot",value:function(){return this.lastBoundingBox.width>=0&&this.lastBoundingBox.height>=0?Zi({},this.lastBoundingBox):{width:0,height:0}}},{key:"getDefaultPosition",value:function(r){var i=this.props,a=i.layout,s=i.align,u=i.verticalAlign,c=i.margin,d=i.chartWidth,p=i.chartHeight,m,v;if(!r||(r.left===void 0||r.left===null)&&(r.right===void 0||r.right===null))if(s==="center"&&a==="vertical"){var b=this.getBBoxSnapshot();m={left:((d||0)-b.width)/2}}else m=s==="right"?{right:c&&c.right||0}:{left:c&&c.left||0};if(!r||(r.top===void 0||r.top===null)&&(r.bottom===void 0||r.bottom===null))if(u==="middle"){var S=this.getBBoxSnapshot();v={top:((p||0)-S.height)/2}}else v=u==="bottom"?{bottom:c&&c.bottom||0}:{top:c&&c.top||0};return Zi(Zi({},m),v)}},{key:"render",value:function(){var r=this,i=this.props,a=i.content,s=i.width,u=i.height,c=i.wrapperStyle,d=i.payloadUniqBy,p=i.payload,m=Zi(Zi({position:"absolute",width:s||"auto",height:u||"auto"},this.getDefaultPosition(c)),c);return V.createElement("div",{className:"recharts-legend-wrapper",style:m,ref:function(b){r.wrapperNode=b}},$ae(a,Zi(Zi({},this.props),{},{payload:gU(p,d,kae)})))}}],[{key:"getWithHeight",value:function(r,i){var a=Zi(Zi({},this.defaultProps),r.props),s=a.layout;return s==="vertical"&&$e(r.props.height)?{height:r.props.height}:s==="horizontal"?{width:r.props.width||i}:null}}])}(g.PureComponent);qv(Al,"displayName","Legend");qv(Al,"defaultProps",{iconSize:14,layout:"horizontal",align:"center",verticalAlign:"bottom"});var qw,i2;function Iae(){if(i2)return qw;i2=1;var e=bd(),t=WC(),n=Qn(),r=e?e.isConcatSpreadable:void 0;function i(a){return n(a)||t(a)||!!(r&&a&&a[r])}return qw=i,qw}var zw,a2;function xU(){if(a2)return zw;a2=1;var e=lU(),t=Iae();function n(r,i,a,s,u){var c=-1,d=r.length;for(a||(a=t),u||(u=[]);++c<d;){var p=r[c];i>0&&a(p)?i>1?n(p,i-1,a,s,u):e(u,p):s||(u[u.length]=p)}return u}return zw=n,zw}var Uw,o2;function Lae(){if(o2)return Uw;o2=1;function e(t){return function(n,r,i){for(var a=-1,s=Object(n),u=i(n),c=u.length;c--;){var d=u[t?c:++a];if(r(s[d],d,s)===!1)break}return n}}return Uw=e,Uw}var Vw,s2;function Bae(){if(s2)return Vw;s2=1;var e=Lae(),t=e();return Vw=t,Vw}var Hw,l2;function wU(){if(l2)return Hw;l2=1;var e=Bae(),t=Fv();function n(r,i){return r&&e(r,i,t)}return Hw=n,Hw}var Gw,u2;function Fae(){if(u2)return Gw;u2=1;var e=Sd();function t(n,r){return function(i,a){if(i==null)return i;if(!e(i))return n(i,a);for(var s=i.length,u=r?s:-1,c=Object(i);(r?u--:++u<s)&&a(c[u],u,c)!==!1;);return i}}return Gw=t,Gw}var Ww,c2;function XC(){if(c2)return Ww;c2=1;var e=wU(),t=Fae(),n=t(e);return Ww=n,Ww}var Kw,f2;function SU(){if(f2)return Kw;f2=1;var e=XC(),t=Sd();function n(r,i){var a=-1,s=t(r)?Array(r.length):[];return e(r,function(u,c,d){s[++a]=i(u,c,d)}),s}return Kw=n,Kw}var Yw,d2;function qae(){if(d2)return Yw;d2=1;function e(t,n){var r=t.length;for(t.sort(n);r--;)t[r]=t[r].value;return t}return Yw=e,Yw}var Zw,p2;function zae(){if(p2)return Zw;p2=1;var e=gu();function t(n,r){if(n!==r){var i=n!==void 0,a=n===null,s=n===n,u=e(n),c=r!==void 0,d=r===null,p=r===r,m=e(r);if(!d&&!m&&!u&&n>r||u&&c&&p&&!d&&!m||a&&c&&p||!i&&p||!s)return 1;if(!a&&!u&&!m&&n<r||m&&i&&s&&!a&&!u||d&&i&&s||!c&&s||!p)return-1}return 0}return Zw=t,Zw}var Xw,h2;function Uae(){if(h2)return Xw;h2=1;var e=zae();function t(n,r,i){for(var a=-1,s=n.criteria,u=r.criteria,c=s.length,d=i.length;++a<c;){var p=e(s[a],u[a]);if(p){if(a>=d)return p;var m=i[a];return p*(m=="desc"?-1:1)}}return n.index-r.index}return Xw=t,Xw}var Qw,m2;function Vae(){if(m2)return Qw;m2=1;var e=$C(),t=IC(),n=Pi(),r=SU(),i=qae(),a=cU(),s=Uae(),u=bu(),c=Qn();function d(p,m,v){m.length?m=e(m,function(w){return c(w)?function(x){return t(x,w.length===1?w[0]:w)}:w}):m=[u];var b=-1;m=e(m,a(n));var S=r(p,function(w,x,C){var O=e(m,function(A){return A(w)});return{criteria:O,index:++b,value:w}});return i(S,function(w,x){return s(w,x,v)})}return Qw=d,Qw}var Jw,v2;function Hae(){if(v2)return Jw;v2=1;function e(t,n,r){switch(r.length){case 0:return t.call(n);case 1:return t.call(n,r[0]);case 2:return t.call(n,r[0],r[1]);case 3:return t.call(n,r[0],r[1],r[2])}return t.apply(n,r)}return Jw=e,Jw}var eS,g2;function Gae(){if(g2)return eS;g2=1;var e=Hae(),t=Math.max;function n(r,i,a){return i=t(i===void 0?r.length-1:i,0),function(){for(var s=arguments,u=-1,c=t(s.length-i,0),d=Array(c);++u<c;)d[u]=s[i+u];u=-1;for(var p=Array(i+1);++u<i;)p[u]=s[u];return p[i]=a(d),e(r,this,p)}}return eS=n,eS}var tS,y2;function Wae(){if(y2)return tS;y2=1;function e(t){return function(){return t}}return tS=e,tS}var nS,b2;function _U(){if(b2)return nS;b2=1;var e=gs(),t=function(){try{var n=e(Object,"defineProperty");return n({},"",{}),n}catch{}}();return nS=t,nS}var rS,x2;function Kae(){if(x2)return rS;x2=1;var e=Wae(),t=_U(),n=bu(),r=t?function(i,a){return t(i,"toString",{configurable:!0,enumerable:!1,value:e(a),writable:!0})}:n;return rS=r,rS}var iS,w2;function Yae(){if(w2)return iS;w2=1;var e=800,t=16,n=Date.now;function r(i){var a=0,s=0;return function(){var u=n(),c=t-(u-s);if(s=u,c>0){if(++a>=e)return arguments[0]}else a=0;return i.apply(void 0,arguments)}}return iS=r,iS}var aS,S2;function Zae(){if(S2)return aS;S2=1;var e=Kae(),t=Yae(),n=t(e);return aS=n,aS}var oS,_2;function Xae(){if(_2)return oS;_2=1;var e=bu(),t=Gae(),n=Zae();function r(i,a){return n(t(i,a,e),i+"")}return oS=r,oS}var sS,E2;function zv(){if(E2)return sS;E2=1;var e=NC(),t=Sd(),n=KC(),r=mo();function i(a,s,u){if(!r(u))return!1;var c=typeof s;return(c=="number"?t(u)&&n(s,u.length):c=="string"&&s in u)?e(u[s],a):!1}return sS=i,sS}var lS,C2;function Qae(){if(C2)return lS;C2=1;var e=xU(),t=Vae(),n=Xae(),r=zv(),i=n(function(a,s){if(a==null)return[];var u=s.length;return u>1&&r(a,s[0],s[1])?s=[]:u>2&&r(s[0],s[1],s[2])&&(s=[s[0]]),t(a,e(s,1),[])});return lS=i,lS}var Jae=Qae();const QC=Rt(Jae);function _f(e){"@babel/helpers - typeof";return _f=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},_f(e)}function x1(){return x1=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)Object.prototype.hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},x1.apply(this,arguments)}function eoe(e,t){return ioe(e)||roe(e,t)||noe(e,t)||toe()}function toe(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function noe(e,t){if(e){if(typeof e=="string")return O2(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);if(n==="Object"&&e.constructor&&(n=e.constructor.name),n==="Map"||n==="Set")return Array.from(e);if(n==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return O2(e,t)}}function O2(e,t){(t==null||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}function roe(e,t){var n=e==null?null:typeof Symbol<"u"&&e[Symbol.iterator]||e["@@iterator"];if(n!=null){var r,i,a,s,u=[],c=!0,d=!1;try{if(a=(n=n.call(e)).next,t!==0)for(;!(c=(r=a.call(n)).done)&&(u.push(r.value),u.length!==t);c=!0);}catch(p){d=!0,i=p}finally{try{if(!c&&n.return!=null&&(s=n.return(),Object(s)!==s))return}finally{if(d)throw i}}return u}}function ioe(e){if(Array.isArray(e))return e}function A2(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function uS(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?A2(Object(n),!0).forEach(function(r){aoe(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):A2(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function aoe(e,t,n){return t=ooe(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function ooe(e){var t=soe(e,"string");return _f(t)=="symbol"?t:t+""}function soe(e,t){if(_f(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(_f(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function loe(e){return Array.isArray(e)&&sn(e[0])&&sn(e[1])?e.join(" ~ "):e}var uoe=function(t){var n=t.separator,r=n===void 0?" : ":n,i=t.contentStyle,a=i===void 0?{}:i,s=t.itemStyle,u=s===void 0?{}:s,c=t.labelStyle,d=c===void 0?{}:c,p=t.payload,m=t.formatter,v=t.itemSorter,b=t.wrapperClassName,S=t.labelClassName,w=t.label,x=t.labelFormatter,C=t.accessibilityLayer,O=C===void 0?!1:C,A=function(){if(p&&p.length){var F={padding:0,margin:0},z=(v?QC(p,v):p).map(function(Z,Y){if(Z.type==="none")return null;var I=uS({display:"block",paddingTop:4,paddingBottom:4,color:Z.color||"#000"},u),W=Z.formatter||m||loe,B=Z.value,Q=Z.name,$=B,J=Q;if(W&&$!=null&&J!=null){var q=W(B,Q,Z,Y,p);if(Array.isArray(q)){var X=eoe(q,2);$=X[0],J=X[1]}else $=q}return V.createElement("li",{className:"recharts-tooltip-item",key:"tooltip-item-".concat(Y),style:I},sn(J)?V.createElement("span",{className:"recharts-tooltip-item-name"},J):null,sn(J)?V.createElement("span",{className:"recharts-tooltip-item-separator"},r):null,V.createElement("span",{className:"recharts-tooltip-item-value"},$),V.createElement("span",{className:"recharts-tooltip-item-unit"},Z.unit||""))});return V.createElement("ul",{className:"recharts-tooltip-item-list",style:F},z)}return null},T=uS({margin:0,padding:10,backgroundColor:"#fff",border:"1px solid #ccc",whiteSpace:"nowrap"},a),R=uS({margin:0},d),M=!vt(w),P=M?w:"",N=ct("recharts-default-tooltip",b),k=ct("recharts-tooltip-label",S);M&&x&&p!==void 0&&p!==null&&(P=x(w,p));var H=O?{role:"status","aria-live":"assertive"}:{};return V.createElement("div",x1({className:N,style:T},H),V.createElement("p",{className:k,style:R},V.isValidElement(P)?P:"".concat(P)),A())};function Ef(e){"@babel/helpers - typeof";return Ef=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},Ef(e)}function ih(e,t,n){return t=coe(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function coe(e){var t=foe(e,"string");return Ef(t)=="symbol"?t:t+""}function foe(e,t){if(Ef(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(Ef(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}var Tc="recharts-tooltip-wrapper",doe={visibility:"hidden"};function poe(e){var t=e.coordinate,n=e.translateX,r=e.translateY;return ct(Tc,ih(ih(ih(ih({},"".concat(Tc,"-right"),$e(n)&&t&&$e(t.x)&&n>=t.x),"".concat(Tc,"-left"),$e(n)&&t&&$e(t.x)&&n<t.x),"".concat(Tc,"-bottom"),$e(r)&&t&&$e(t.y)&&r>=t.y),"".concat(Tc,"-top"),$e(r)&&t&&$e(t.y)&&r<t.y))}function R2(e){var t=e.allowEscapeViewBox,n=e.coordinate,r=e.key,i=e.offsetTopLeft,a=e.position,s=e.reverseDirection,u=e.tooltipDimension,c=e.viewBox,d=e.viewBoxDimension;if(a&&$e(a[r]))return a[r];var p=n[r]-u-i,m=n[r]+i;if(t[r])return s[r]?p:m;if(s[r]){var v=p,b=c[r];return v<b?Math.max(m,c[r]):Math.max(p,c[r])}var S=m+u,w=c[r]+d;return S>w?Math.max(p,c[r]):Math.max(m,c[r])}function hoe(e){var t=e.translateX,n=e.translateY,r=e.useTranslate3d;return{transform:r?"translate3d(".concat(t,"px, ").concat(n,"px, 0)"):"translate(".concat(t,"px, ").concat(n,"px)")}}function moe(e){var t=e.allowEscapeViewBox,n=e.coordinate,r=e.offsetTopLeft,i=e.position,a=e.reverseDirection,s=e.tooltipBox,u=e.useTranslate3d,c=e.viewBox,d,p,m;return s.height>0&&s.width>0&&n?(p=R2({allowEscapeViewBox:t,coordinate:n,key:"x",offsetTopLeft:r,position:i,reverseDirection:a,tooltipDimension:s.width,viewBox:c,viewBoxDimension:c.width}),m=R2({allowEscapeViewBox:t,coordinate:n,key:"y",offsetTopLeft:r,position:i,reverseDirection:a,tooltipDimension:s.height,viewBox:c,viewBoxDimension:c.height}),d=hoe({translateX:p,translateY:m,useTranslate3d:u})):d=doe,{cssProperties:d,cssClasses:poe({translateX:p,translateY:m,coordinate:n})}}function Bl(e){"@babel/helpers - typeof";return Bl=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},Bl(e)}function T2(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function M2(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?T2(Object(n),!0).forEach(function(r){S1(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):T2(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function voe(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}function goe(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,CU(r.key),r)}}function yoe(e,t,n){return t&&goe(e.prototype,t),Object.defineProperty(e,"prototype",{writable:!1}),e}function boe(e,t,n){return t=pm(t),xoe(e,EU()?Reflect.construct(t,n||[],pm(e).constructor):t.apply(e,n))}function xoe(e,t){if(t&&(Bl(t)==="object"||typeof t=="function"))return t;if(t!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return woe(e)}function woe(e){if(e===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return e}function EU(){try{var e=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(EU=function(){return!!e})()}function pm(e){return pm=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(n){return n.__proto__||Object.getPrototypeOf(n)},pm(e)}function Soe(e,t){if(typeof t!="function"&&t!==null)throw new TypeError("Super expression must either be null or a function");e.prototype=Object.create(t&&t.prototype,{constructor:{value:e,writable:!0,configurable:!0}}),Object.defineProperty(e,"prototype",{writable:!1}),t&&w1(e,t)}function w1(e,t){return w1=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(r,i){return r.__proto__=i,r},w1(e,t)}function S1(e,t,n){return t=CU(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function CU(e){var t=_oe(e,"string");return Bl(t)=="symbol"?t:t+""}function _oe(e,t){if(Bl(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(Bl(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(e)}var P2=1,Eoe=function(e){function t(){var n;voe(this,t);for(var r=arguments.length,i=new Array(r),a=0;a<r;a++)i[a]=arguments[a];return n=boe(this,t,[].concat(i)),S1(n,"state",{dismissed:!1,dismissedAtCoordinate:{x:0,y:0},lastBoundingBox:{width:-1,height:-1}}),S1(n,"handleKeyDown",function(s){if(s.key==="Escape"){var u,c,d,p;n.setState({dismissed:!0,dismissedAtCoordinate:{x:(u=(c=n.props.coordinate)===null||c===void 0?void 0:c.x)!==null&&u!==void 0?u:0,y:(d=(p=n.props.coordinate)===null||p===void 0?void 0:p.y)!==null&&d!==void 0?d:0}})}}),n}return Soe(t,e),yoe(t,[{key:"updateBBox",value:function(){if(this.wrapperNode&&this.wrapperNode.getBoundingClientRect){var r=this.wrapperNode.getBoundingClientRect();(Math.abs(r.width-this.state.lastBoundingBox.width)>P2||Math.abs(r.height-this.state.lastBoundingBox.height)>P2)&&this.setState({lastBoundingBox:{width:r.width,height:r.height}})}else(this.state.lastBoundingBox.width!==-1||this.state.lastBoundingBox.height!==-1)&&this.setState({lastBoundingBox:{width:-1,height:-1}})}},{key:"componentDidMount",value:function(){document.addEventListener("keydown",this.handleKeyDown),this.updateBBox()}},{key:"componentWillUnmount",value:function(){document.removeEventListener("keydown",this.handleKeyDown)}},{key:"componentDidUpdate",value:function(){var r,i;this.props.active&&this.updateBBox(),this.state.dismissed&&(((r=this.props.coordinate)===null||r===void 0?void 0:r.x)!==this.state.dismissedAtCoordinate.x||((i=this.props.coordinate)===null||i===void 0?void 0:i.y)!==this.state.dismissedAtCoordinate.y)&&(this.state.dismissed=!1)}},{key:"render",value:function(){var r=this,i=this.props,a=i.active,s=i.allowEscapeViewBox,u=i.animationDuration,c=i.animationEasing,d=i.children,p=i.coordinate,m=i.hasPayload,v=i.isAnimationActive,b=i.offset,S=i.position,w=i.reverseDirection,x=i.useTranslate3d,C=i.viewBox,O=i.wrapperStyle,A=moe({allowEscapeViewBox:s,coordinate:p,offsetTopLeft:b,position:S,reverseDirection:w,tooltipBox:this.state.lastBoundingBox,useTranslate3d:x,viewBox:C}),T=A.cssClasses,R=A.cssProperties,M=M2(M2({transition:v&&a?"transform ".concat(u,"ms ").concat(c):void 0},R),{},{pointerEvents:"none",visibility:!this.state.dismissed&&a&&m?"visible":"hidden",position:"absolute",top:0,left:0},O);return V.createElement("div",{tabIndex:-1,className:T,style:M,ref:function(N){r.wrapperNode=N}},d)}}])}(g.PureComponent),Coe=function(){return!(typeof window<"u"&&window.document&&window.document.createElement&&window.setTimeout)},Uv={isSsr:Coe()};function Fl(e){"@babel/helpers - typeof";return Fl=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},Fl(e)}function j2(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function N2(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?j2(Object(n),!0).forEach(function(r){JC(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):j2(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function Ooe(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}function Aoe(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,AU(r.key),r)}}function Roe(e,t,n){return t&&Aoe(e.prototype,t),Object.defineProperty(e,"prototype",{writable:!1}),e}function Toe(e,t,n){return t=hm(t),Moe(e,OU()?Reflect.construct(t,n||[],hm(e).constructor):t.apply(e,n))}function Moe(e,t){if(t&&(Fl(t)==="object"||typeof t=="function"))return t;if(t!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return Poe(e)}function Poe(e){if(e===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return e}function OU(){try{var e=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(OU=function(){return!!e})()}function hm(e){return hm=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(n){return n.__proto__||Object.getPrototypeOf(n)},hm(e)}function joe(e,t){if(typeof t!="function"&&t!==null)throw new TypeError("Super expression must either be null or a function");e.prototype=Object.create(t&&t.prototype,{constructor:{value:e,writable:!0,configurable:!0}}),Object.defineProperty(e,"prototype",{writable:!1}),t&&_1(e,t)}function _1(e,t){return _1=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(r,i){return r.__proto__=i,r},_1(e,t)}function JC(e,t,n){return t=AU(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function AU(e){var t=Noe(e,"string");return Fl(t)=="symbol"?t:t+""}function Noe(e,t){if(Fl(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(Fl(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(e)}function Doe(e){return e.dataKey}function koe(e,t){return V.isValidElement(e)?V.cloneElement(e,t):typeof e=="function"?V.createElement(e,t):V.createElement(uoe,t)}var fi=function(e){function t(){return Ooe(this,t),Toe(this,t,arguments)}return joe(t,e),Roe(t,[{key:"render",value:function(){var r=this,i=this.props,a=i.active,s=i.allowEscapeViewBox,u=i.animationDuration,c=i.animationEasing,d=i.content,p=i.coordinate,m=i.filterNull,v=i.isAnimationActive,b=i.offset,S=i.payload,w=i.payloadUniqBy,x=i.position,C=i.reverseDirection,O=i.useTranslate3d,A=i.viewBox,T=i.wrapperStyle,R=S??[];m&&R.length&&(R=gU(S.filter(function(P){return P.value!=null&&(P.hide!==!0||r.props.includeHidden)}),w,Doe));var M=R.length>0;return V.createElement(Eoe,{allowEscapeViewBox:s,animationDuration:u,animationEasing:c,isAnimationActive:v,active:a,coordinate:p,hasPayload:M,offset:b,position:x,reverseDirection:C,useTranslate3d:O,viewBox:A,wrapperStyle:T},koe(d,N2(N2({},this.props),{},{payload:R})))}}])}(g.PureComponent);JC(fi,"displayName","Tooltip");JC(fi,"defaultProps",{accessibilityLayer:!1,allowEscapeViewBox:{x:!1,y:!1},animationDuration:400,animationEasing:"ease",contentStyle:{},coordinate:{x:0,y:0},cursor:!0,cursorStyle:{},filterNull:!0,isAnimationActive:!Uv.isSsr,itemStyle:{},labelStyle:{},offset:10,reverseDirection:{x:!1,y:!1},separator:" : ",trigger:"hover",useTranslate3d:!1,viewBox:{x:0,y:0,height:0,width:0},wrapperStyle:{}});var cS,D2;function $oe(){if(D2)return cS;D2=1;var e=Mi(),t=function(){return e.Date.now()};return cS=t,cS}var fS,k2;function Ioe(){if(k2)return fS;k2=1;var e=/\s/;function t(n){for(var r=n.length;r--&&e.test(n.charAt(r)););return r}return fS=t,fS}var dS,$2;function Loe(){if($2)return dS;$2=1;var e=Ioe(),t=/^\s+/;function n(r){return r&&r.slice(0,e(r)+1).replace(t,"")}return dS=n,dS}var pS,I2;function RU(){if(I2)return pS;I2=1;var e=Loe(),t=mo(),n=gu(),r=NaN,i=/^[-+]0x[0-9a-f]+$/i,a=/^0b[01]+$/i,s=/^0o[0-7]+$/i,u=parseInt;function c(d){if(typeof d=="number")return d;if(n(d))return r;if(t(d)){var p=typeof d.valueOf=="function"?d.valueOf():d;d=t(p)?p+"":p}if(typeof d!="string")return d===0?d:+d;d=e(d);var m=a.test(d);return m||s.test(d)?u(d.slice(2),m?2:8):i.test(d)?r:+d}return pS=c,pS}var hS,L2;function Boe(){if(L2)return hS;L2=1;var e=mo(),t=$oe(),n=RU(),r="Expected a function",i=Math.max,a=Math.min;function s(u,c,d){var p,m,v,b,S,w,x=0,C=!1,O=!1,A=!0;if(typeof u!="function")throw new TypeError(r);c=n(c)||0,e(d)&&(C=!!d.leading,O="maxWait"in d,v=O?i(n(d.maxWait)||0,c):v,A="trailing"in d?!!d.trailing:A);function T(z){var Z=p,Y=m;return p=m=void 0,x=z,b=u.apply(Y,Z),b}function R(z){return x=z,S=setTimeout(N,c),C?T(z):b}function M(z){var Z=z-w,Y=z-x,I=c-Z;return O?a(I,v-Y):I}function P(z){var Z=z-w,Y=z-x;return w===void 0||Z>=c||Z<0||O&&Y>=v}function N(){var z=t();if(P(z))return k(z);S=setTimeout(N,M(z))}function k(z){return S=void 0,A&&p?T(z):(p=m=void 0,b)}function H(){S!==void 0&&clearTimeout(S),x=0,p=w=m=S=void 0}function L(){return S===void 0?b:k(t())}function F(){var z=t(),Z=P(z);if(p=arguments,m=this,w=z,Z){if(S===void 0)return R(w);if(O)return clearTimeout(S),S=setTimeout(N,c),T(w)}return S===void 0&&(S=setTimeout(N,c)),b}return F.cancel=H,F.flush=L,F}return hS=s,hS}var mS,B2;function Foe(){if(B2)return mS;B2=1;var e=Boe(),t=mo(),n="Expected a function";function r(i,a,s){var u=!0,c=!0;if(typeof i!="function")throw new TypeError(n);return t(s)&&(u="leading"in s?!!s.leading:u,c="trailing"in s?!!s.trailing:c),e(i,a,{leading:u,maxWait:a,trailing:c})}return mS=r,mS}var qoe=Foe();const TU=Rt(qoe);function Cf(e){"@babel/helpers - typeof";return Cf=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},Cf(e)}function F2(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function ah(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?F2(Object(n),!0).forEach(function(r){zoe(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):F2(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function zoe(e,t,n){return t=Uoe(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function Uoe(e){var t=Voe(e,"string");return Cf(t)=="symbol"?t:t+""}function Voe(e,t){if(Cf(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(Cf(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function Hoe(e,t){return Yoe(e)||Koe(e,t)||Woe(e,t)||Goe()}function Goe(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function Woe(e,t){if(e){if(typeof e=="string")return q2(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);if(n==="Object"&&e.constructor&&(n=e.constructor.name),n==="Map"||n==="Set")return Array.from(e);if(n==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return q2(e,t)}}function q2(e,t){(t==null||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}function Koe(e,t){var n=e==null?null:typeof Symbol<"u"&&e[Symbol.iterator]||e["@@iterator"];if(n!=null){var r,i,a,s,u=[],c=!0,d=!1;try{if(a=(n=n.call(e)).next,t!==0)for(;!(c=(r=a.call(n)).done)&&(u.push(r.value),u.length!==t);c=!0);}catch(p){d=!0,i=p}finally{try{if(!c&&n.return!=null&&(s=n.return(),Object(s)!==s))return}finally{if(d)throw i}}return u}}function Yoe(e){if(Array.isArray(e))return e}var Zoe=g.forwardRef(function(e,t){var n=e.aspect,r=e.initialDimension,i=r===void 0?{width:-1,height:-1}:r,a=e.width,s=a===void 0?"100%":a,u=e.height,c=u===void 0?"100%":u,d=e.minWidth,p=d===void 0?0:d,m=e.minHeight,v=e.maxHeight,b=e.children,S=e.debounce,w=S===void 0?0:S,x=e.id,C=e.className,O=e.onResize,A=e.style,T=A===void 0?{}:A,R=g.useRef(null),M=g.useRef();M.current=O,g.useImperativeHandle(t,function(){return Object.defineProperty(R.current,"current",{get:function(){return console.warn("The usage of ref.current.current is deprecated and will no longer be supported."),R.current},configurable:!0})});var P=g.useState({containerWidth:i.width,containerHeight:i.height}),N=Hoe(P,2),k=N[0],H=N[1],L=g.useCallback(function(z,Z){H(function(Y){var I=Math.round(z),W=Math.round(Z);return Y.containerWidth===I&&Y.containerHeight===W?Y:{containerWidth:I,containerHeight:W}})},[]);g.useEffect(function(){var z=function(Q){var $,J=Q[0].contentRect,q=J.width,X=J.height;L(q,X),($=M.current)===null||$===void 0||$.call(M,q,X)};w>0&&(z=TU(z,w,{trailing:!0,leading:!1}));var Z=new ResizeObserver(z),Y=R.current.getBoundingClientRect(),I=Y.width,W=Y.height;return L(I,W),Z.observe(R.current),function(){Z.disconnect()}},[L,w]);var F=g.useMemo(function(){var z=k.containerWidth,Z=k.containerHeight;if(z<0||Z<0)return null;es(Ko(s)||Ko(c),`The width(%s) and height(%s) are both fixed numbers,
       maybe you don't need to use a ResponsiveContainer.`,s,c),es(!n||n>0,"The aspect(%s) must be greater than zero.",n);var Y=Ko(s)?z:s,I=Ko(c)?Z:c;n&&n>0&&(Y?I=Y/n:I&&(Y=I*n),v&&I>v&&(I=v)),es(Y>0||I>0,`The width(%s) and height(%s) of chart should be greater than 0,
       please check the style of container, or the props width(%s) and height(%s),
       or add a minWidth(%s) or minHeight(%s) or use aspect(%s) to control the
       height and width.`,Y,I,s,c,p,m,n);var W=!Array.isArray(b)&&sa(b.type).endsWith("Chart");return V.Children.map(b,function(B){return V.isValidElement(B)?g.cloneElement(B,ah({width:Y,height:I},W?{style:ah({height:"100%",width:"100%",maxHeight:I,maxWidth:Y},B.props.style)}:{})):B})},[n,b,c,v,m,p,k,s]);return V.createElement("div",{id:x?"".concat(x):void 0,className:ct("recharts-responsive-container",C),style:ah(ah({},T),{},{width:s,height:c,minWidth:p,minHeight:m,maxHeight:v}),ref:R},F)});function Of(e){"@babel/helpers - typeof";return Of=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},Of(e)}function z2(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function E1(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?z2(Object(n),!0).forEach(function(r){Xoe(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):z2(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function Xoe(e,t,n){return t=Qoe(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function Qoe(e){var t=Joe(e,"string");return Of(t)=="symbol"?t:t+""}function Joe(e,t){if(Of(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(Of(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}var hl={widthCache:{},cacheCount:0},ese=2e3,tse={position:"absolute",top:"-20000px",left:0,padding:0,margin:0,border:"none",whiteSpace:"pre"},U2="recharts_measurement_span";function nse(e){var t=E1({},e);return Object.keys(t).forEach(function(n){t[n]||delete t[n]}),t}var V2=function(t){var n=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{};if(t==null||Uv.isSsr)return{width:0,height:0};var r=nse(n),i=JSON.stringify({text:t,copyStyle:r});if(hl.widthCache[i])return hl.widthCache[i];try{var a=document.getElementById(U2);a||(a=document.createElement("span"),a.setAttribute("id",U2),a.setAttribute("aria-hidden","true"),document.body.appendChild(a));var s=E1(E1({},tse),r);Object.assign(a.style,s),a.textContent="".concat(t);var u=a.getBoundingClientRect(),c={width:u.width,height:u.height};return hl.widthCache[i]=c,++hl.cacheCount>ese&&(hl.cacheCount=0,hl.widthCache={}),c}catch{return{width:0,height:0}}},rse=function(t){return{top:t.top+window.scrollY-document.documentElement.clientTop,left:t.left+window.scrollX-document.documentElement.clientLeft}};function Af(e){"@babel/helpers - typeof";return Af=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},Af(e)}function mm(e,t){return sse(e)||ose(e,t)||ase(e,t)||ise()}function ise(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function ase(e,t){if(e){if(typeof e=="string")return H2(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);if(n==="Object"&&e.constructor&&(n=e.constructor.name),n==="Map"||n==="Set")return Array.from(e);if(n==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return H2(e,t)}}function H2(e,t){(t==null||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}function ose(e,t){var n=e==null?null:typeof Symbol<"u"&&e[Symbol.iterator]||e["@@iterator"];if(n!=null){var r,i,a,s,u=[],c=!0,d=!1;try{if(a=(n=n.call(e)).next,t===0){if(Object(n)!==n)return;c=!1}else for(;!(c=(r=a.call(n)).done)&&(u.push(r.value),u.length!==t);c=!0);}catch(p){d=!0,i=p}finally{try{if(!c&&n.return!=null&&(s=n.return(),Object(s)!==s))return}finally{if(d)throw i}}return u}}function sse(e){if(Array.isArray(e))return e}function lse(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}function G2(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,cse(r.key),r)}}function use(e,t,n){return t&&G2(e.prototype,t),n&&G2(e,n),Object.defineProperty(e,"prototype",{writable:!1}),e}function cse(e){var t=fse(e,"string");return Af(t)=="symbol"?t:t+""}function fse(e,t){if(Af(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(Af(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(e)}var W2=/(-?\d+(?:\.\d+)?[a-zA-Z%]*)([*/])(-?\d+(?:\.\d+)?[a-zA-Z%]*)/,K2=/(-?\d+(?:\.\d+)?[a-zA-Z%]*)([+-])(-?\d+(?:\.\d+)?[a-zA-Z%]*)/,dse=/^px|cm|vh|vw|em|rem|%|mm|in|pt|pc|ex|ch|vmin|vmax|Q$/,pse=/(-?\d+(?:\.\d+)?)([a-zA-Z%]+)?/,MU={cm:96/2.54,mm:96/25.4,pt:96/72,pc:96/6,in:96,Q:96/(2.54*40),px:1},hse=Object.keys(MU),wl="NaN";function mse(e,t){return e*MU[t]}var oh=function(){function e(t,n){lse(this,e),this.num=t,this.unit=n,this.num=t,this.unit=n,Number.isNaN(t)&&(this.unit=""),n!==""&&!dse.test(n)&&(this.num=NaN,this.unit=""),hse.includes(n)&&(this.num=mse(t,n),this.unit="px")}return use(e,[{key:"add",value:function(n){return this.unit!==n.unit?new e(NaN,""):new e(this.num+n.num,this.unit)}},{key:"subtract",value:function(n){return this.unit!==n.unit?new e(NaN,""):new e(this.num-n.num,this.unit)}},{key:"multiply",value:function(n){return this.unit!==""&&n.unit!==""&&this.unit!==n.unit?new e(NaN,""):new e(this.num*n.num,this.unit||n.unit)}},{key:"divide",value:function(n){return this.unit!==""&&n.unit!==""&&this.unit!==n.unit?new e(NaN,""):new e(this.num/n.num,this.unit||n.unit)}},{key:"toString",value:function(){return"".concat(this.num).concat(this.unit)}},{key:"isNaN",value:function(){return Number.isNaN(this.num)}}],[{key:"parse",value:function(n){var r,i=(r=pse.exec(n))!==null&&r!==void 0?r:[],a=mm(i,3),s=a[1],u=a[2];return new e(parseFloat(s),u??"")}}])}();function PU(e){if(e.includes(wl))return wl;for(var t=e;t.includes("*")||t.includes("/");){var n,r=(n=W2.exec(t))!==null&&n!==void 0?n:[],i=mm(r,4),a=i[1],s=i[2],u=i[3],c=oh.parse(a??""),d=oh.parse(u??""),p=s==="*"?c.multiply(d):c.divide(d);if(p.isNaN())return wl;t=t.replace(W2,p.toString())}for(;t.includes("+")||/.-\d+(?:\.\d+)?/.test(t);){var m,v=(m=K2.exec(t))!==null&&m!==void 0?m:[],b=mm(v,4),S=b[1],w=b[2],x=b[3],C=oh.parse(S??""),O=oh.parse(x??""),A=w==="+"?C.add(O):C.subtract(O);if(A.isNaN())return wl;t=t.replace(K2,A.toString())}return t}var Y2=/\(([^()]*)\)/;function vse(e){for(var t=e;t.includes("(");){var n=Y2.exec(t),r=mm(n,2),i=r[1];t=t.replace(Y2,PU(i))}return t}function gse(e){var t=e.replace(/\s+/g,"");return t=vse(t),t=PU(t),t}function yse(e){try{return gse(e)}catch{return wl}}function vS(e){var t=yse(e.slice(5,-1));return t===wl?"":t}var bse=["x","y","lineHeight","capHeight","scaleToFit","textAnchor","verticalAnchor","fill"],xse=["dx","dy","angle","className","breakAll"];function C1(){return C1=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)Object.prototype.hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},C1.apply(this,arguments)}function Z2(e,t){if(e==null)return{};var n=wse(e,t),r,i;if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);for(i=0;i<a.length;i++)r=a[i],!(t.indexOf(r)>=0)&&Object.prototype.propertyIsEnumerable.call(e,r)&&(n[r]=e[r])}return n}function wse(e,t){if(e==null)return{};var n={};for(var r in e)if(Object.prototype.hasOwnProperty.call(e,r)){if(t.indexOf(r)>=0)continue;n[r]=e[r]}return n}function X2(e,t){return Cse(e)||Ese(e,t)||_se(e,t)||Sse()}function Sse(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function _se(e,t){if(e){if(typeof e=="string")return Q2(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);if(n==="Object"&&e.constructor&&(n=e.constructor.name),n==="Map"||n==="Set")return Array.from(e);if(n==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return Q2(e,t)}}function Q2(e,t){(t==null||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}function Ese(e,t){var n=e==null?null:typeof Symbol<"u"&&e[Symbol.iterator]||e["@@iterator"];if(n!=null){var r,i,a,s,u=[],c=!0,d=!1;try{if(a=(n=n.call(e)).next,t===0){if(Object(n)!==n)return;c=!1}else for(;!(c=(r=a.call(n)).done)&&(u.push(r.value),u.length!==t);c=!0);}catch(p){d=!0,i=p}finally{try{if(!c&&n.return!=null&&(s=n.return(),Object(s)!==s))return}finally{if(d)throw i}}return u}}function Cse(e){if(Array.isArray(e))return e}var jU=/[ \f\n\r\t\v\u2028\u2029]+/,NU=function(t){var n=t.children,r=t.breakAll,i=t.style;try{var a=[];vt(n)||(r?a=n.toString().split(""):a=n.toString().split(jU));var s=a.map(function(c){return{word:c,width:V2(c,i).width}}),u=r?0:V2(" ",i).width;return{wordsWithComputedWidth:s,spaceWidth:u}}catch{return null}},Ose=function(t,n,r,i,a){var s=t.maxLines,u=t.children,c=t.style,d=t.breakAll,p=$e(s),m=u,v=function(){var Y=arguments.length>0&&arguments[0]!==void 0?arguments[0]:[];return Y.reduce(function(I,W){var B=W.word,Q=W.width,$=I[I.length-1];if($&&(i==null||a||$.width+Q+r<Number(i)))$.words.push(B),$.width+=Q+r;else{var J={words:[B],width:Q};I.push(J)}return I},[])},b=v(n),S=function(Y){return Y.reduce(function(I,W){return I.width>W.width?I:W})};if(!p)return b;for(var w="…",x=function(Y){var I=m.slice(0,Y),W=NU({breakAll:d,style:c,children:I+w}).wordsWithComputedWidth,B=v(W),Q=B.length>s||S(B).width>Number(i);return[Q,B]},C=0,O=m.length-1,A=0,T;C<=O&&A<=m.length-1;){var R=Math.floor((C+O)/2),M=R-1,P=x(M),N=X2(P,2),k=N[0],H=N[1],L=x(R),F=X2(L,1),z=F[0];if(!k&&!z&&(C=R+1),k&&z&&(O=R-1),!k&&z){T=H;break}A++}return T||b},J2=function(t){var n=vt(t)?[]:t.toString().split(jU);return[{words:n}]},Ase=function(t){var n=t.width,r=t.scaleToFit,i=t.children,a=t.style,s=t.breakAll,u=t.maxLines;if((n||r)&&!Uv.isSsr){var c,d,p=NU({breakAll:s,children:i,style:a});if(p){var m=p.wordsWithComputedWidth,v=p.spaceWidth;c=m,d=v}else return J2(i);return Ose({breakAll:s,children:i,maxLines:u,style:a},c,d,n,r)}return J2(i)},ek="#808080",Rf=function(t){var n=t.x,r=n===void 0?0:n,i=t.y,a=i===void 0?0:i,s=t.lineHeight,u=s===void 0?"1em":s,c=t.capHeight,d=c===void 0?"0.71em":c,p=t.scaleToFit,m=p===void 0?!1:p,v=t.textAnchor,b=v===void 0?"start":v,S=t.verticalAnchor,w=S===void 0?"end":S,x=t.fill,C=x===void 0?ek:x,O=Z2(t,bse),A=g.useMemo(function(){return Ase({breakAll:O.breakAll,children:O.children,maxLines:O.maxLines,scaleToFit:m,style:O.style,width:O.width})},[O.breakAll,O.children,O.maxLines,m,O.style,O.width]),T=O.dx,R=O.dy,M=O.angle,P=O.className,N=O.breakAll,k=Z2(O,xse);if(!sn(r)||!sn(a))return null;var H=r+($e(T)?T:0),L=a+($e(R)?R:0),F;switch(w){case"start":F=vS("calc(".concat(d,")"));break;case"middle":F=vS("calc(".concat((A.length-1)/2," * -").concat(u," + (").concat(d," / 2))"));break;default:F=vS("calc(".concat(A.length-1," * -").concat(u,")"));break}var z=[];if(m){var Z=A[0].width,Y=O.width;z.push("scale(".concat(($e(Y)?Y/Z:1)/Z,")"))}return M&&z.push("rotate(".concat(M,", ").concat(H,", ").concat(L,")")),z.length&&(k.transform=z.join(" ")),V.createElement("text",C1({},Je(k,!0),{x:H,y:L,className:ct("recharts-text",P),textAnchor:b,fill:C.includes("url")?ek:C}),A.map(function(I,W){var B=I.words.join(N?"":" ");return V.createElement("tspan",{x:H,dy:W===0?F:u,key:"".concat(B,"-").concat(W)},B)}))};function no(e,t){return e==null||t==null?NaN:e<t?-1:e>t?1:e>=t?0:NaN}function Rse(e,t){return e==null||t==null?NaN:t<e?-1:t>e?1:t>=e?0:NaN}function eO(e){let t,n,r;e.length!==2?(t=no,n=(u,c)=>no(e(u),c),r=(u,c)=>e(u)-c):(t=e===no||e===Rse?e:Tse,n=e,r=e);function i(u,c,d=0,p=u.length){if(d<p){if(t(c,c)!==0)return p;do{const m=d+p>>>1;n(u[m],c)<0?d=m+1:p=m}while(d<p)}return d}function a(u,c,d=0,p=u.length){if(d<p){if(t(c,c)!==0)return p;do{const m=d+p>>>1;n(u[m],c)<=0?d=m+1:p=m}while(d<p)}return d}function s(u,c,d=0,p=u.length){const m=i(u,c,d,p-1);return m>d&&r(u[m-1],c)>-r(u[m],c)?m-1:m}return{left:i,center:s,right:a}}function Tse(){return 0}function DU(e){return e===null?NaN:+e}function*Mse(e,t){for(let n of e)n!=null&&(n=+n)>=n&&(yield n)}const Pse=eO(no),_d=Pse.right;eO(DU).center;class tk extends Map{constructor(t,n=Dse){if(super(),Object.defineProperties(this,{_intern:{value:new Map},_key:{value:n}}),t!=null)for(const[r,i]of t)this.set(r,i)}get(t){return super.get(nk(this,t))}has(t){return super.has(nk(this,t))}set(t,n){return super.set(jse(this,t),n)}delete(t){return super.delete(Nse(this,t))}}function nk({_intern:e,_key:t},n){const r=t(n);return e.has(r)?e.get(r):n}function jse({_intern:e,_key:t},n){const r=t(n);return e.has(r)?e.get(r):(e.set(r,n),n)}function Nse({_intern:e,_key:t},n){const r=t(n);return e.has(r)&&(n=e.get(r),e.delete(r)),n}function Dse(e){return e!==null&&typeof e=="object"?e.valueOf():e}function kse(e=no){if(e===no)return kU;if(typeof e!="function")throw new TypeError("compare is not a function");return(t,n)=>{const r=e(t,n);return r||r===0?r:(e(n,n)===0)-(e(t,t)===0)}}function kU(e,t){return(e==null||!(e>=e))-(t==null||!(t>=t))||(e<t?-1:e>t?1:0)}const $se=Math.sqrt(50),Ise=Math.sqrt(10),Lse=Math.sqrt(2);function vm(e,t,n){const r=(t-e)/Math.max(0,n),i=Math.floor(Math.log10(r)),a=r/Math.pow(10,i),s=a>=$se?10:a>=Ise?5:a>=Lse?2:1;let u,c,d;return i<0?(d=Math.pow(10,-i)/s,u=Math.round(e*d),c=Math.round(t*d),u/d<e&&++u,c/d>t&&--c,d=-d):(d=Math.pow(10,i)*s,u=Math.round(e/d),c=Math.round(t/d),u*d<e&&++u,c*d>t&&--c),c<u&&.5<=n&&n<2?vm(e,t,n*2):[u,c,d]}function O1(e,t,n){if(t=+t,e=+e,n=+n,!(n>0))return[];if(e===t)return[e];const r=t<e,[i,a,s]=r?vm(t,e,n):vm(e,t,n);if(!(a>=i))return[];const u=a-i+1,c=new Array(u);if(r)if(s<0)for(let d=0;d<u;++d)c[d]=(a-d)/-s;else for(let d=0;d<u;++d)c[d]=(a-d)*s;else if(s<0)for(let d=0;d<u;++d)c[d]=(i+d)/-s;else for(let d=0;d<u;++d)c[d]=(i+d)*s;return c}function A1(e,t,n){return t=+t,e=+e,n=+n,vm(e,t,n)[2]}function R1(e,t,n){t=+t,e=+e,n=+n;const r=t<e,i=r?A1(t,e,n):A1(e,t,n);return(r?-1:1)*(i<0?1/-i:i)}function rk(e,t){let n;for(const r of e)r!=null&&(n<r||n===void 0&&r>=r)&&(n=r);return n}function ik(e,t){let n;for(const r of e)r!=null&&(n>r||n===void 0&&r>=r)&&(n=r);return n}function $U(e,t,n=0,r=1/0,i){if(t=Math.floor(t),n=Math.floor(Math.max(0,n)),r=Math.floor(Math.min(e.length-1,r)),!(n<=t&&t<=r))return e;for(i=i===void 0?kU:kse(i);r>n;){if(r-n>600){const c=r-n+1,d=t-n+1,p=Math.log(c),m=.5*Math.exp(2*p/3),v=.5*Math.sqrt(p*m*(c-m)/c)*(d-c/2<0?-1:1),b=Math.max(n,Math.floor(t-d*m/c+v)),S=Math.min(r,Math.floor(t+(c-d)*m/c+v));$U(e,t,b,S,i)}const a=e[t];let s=n,u=r;for(Mc(e,n,t),i(e[r],a)>0&&Mc(e,n,r);s<u;){for(Mc(e,s,u),++s,--u;i(e[s],a)<0;)++s;for(;i(e[u],a)>0;)--u}i(e[n],a)===0?Mc(e,n,u):(++u,Mc(e,u,r)),u<=t&&(n=u+1),t<=u&&(r=u-1)}return e}function Mc(e,t,n){const r=e[t];e[t]=e[n],e[n]=r}function Bse(e,t,n){if(e=Float64Array.from(Mse(e)),!(!(r=e.length)||isNaN(t=+t))){if(t<=0||r<2)return ik(e);if(t>=1)return rk(e);var r,i=(r-1)*t,a=Math.floor(i),s=rk($U(e,a).subarray(0,a+1)),u=ik(e.subarray(a+1));return s+(u-s)*(i-a)}}function Fse(e,t,n=DU){if(!(!(r=e.length)||isNaN(t=+t))){if(t<=0||r<2)return+n(e[0],0,e);if(t>=1)return+n(e[r-1],r-1,e);var r,i=(r-1)*t,a=Math.floor(i),s=+n(e[a],a,e),u=+n(e[a+1],a+1,e);return s+(u-s)*(i-a)}}function qse(e,t,n){e=+e,t=+t,n=(i=arguments.length)<2?(t=e,e=0,1):i<3?1:+n;for(var r=-1,i=Math.max(0,Math.ceil((t-e)/n))|0,a=new Array(i);++r<i;)a[r]=e+r*n;return a}function Ir(e,t){switch(arguments.length){case 0:break;case 1:this.range(e);break;default:this.range(t).domain(e);break}return this}function Sa(e,t){switch(arguments.length){case 0:break;case 1:{typeof e=="function"?this.interpolator(e):this.range(e);break}default:{this.domain(e),typeof t=="function"?this.interpolator(t):this.range(t);break}}return this}const T1=Symbol("implicit");function tO(){var e=new tk,t=[],n=[],r=T1;function i(a){let s=e.get(a);if(s===void 0){if(r!==T1)return r;e.set(a,s=t.push(a)-1)}return n[s%n.length]}return i.domain=function(a){if(!arguments.length)return t.slice();t=[],e=new tk;for(const s of a)e.has(s)||e.set(s,t.push(s)-1);return i},i.range=function(a){return arguments.length?(n=Array.from(a),i):n.slice()},i.unknown=function(a){return arguments.length?(r=a,i):r},i.copy=function(){return tO(t,n).unknown(r)},Ir.apply(i,arguments),i}function Tf(){var e=tO().unknown(void 0),t=e.domain,n=e.range,r=0,i=1,a,s,u=!1,c=0,d=0,p=.5;delete e.unknown;function m(){var v=t().length,b=i<r,S=b?i:r,w=b?r:i;a=(w-S)/Math.max(1,v-c+d*2),u&&(a=Math.floor(a)),S+=(w-S-a*(v-c))*p,s=a*(1-c),u&&(S=Math.round(S),s=Math.round(s));var x=qse(v).map(function(C){return S+a*C});return n(b?x.reverse():x)}return e.domain=function(v){return arguments.length?(t(v),m()):t()},e.range=function(v){return arguments.length?([r,i]=v,r=+r,i=+i,m()):[r,i]},e.rangeRound=function(v){return[r,i]=v,r=+r,i=+i,u=!0,m()},e.bandwidth=function(){return s},e.step=function(){return a},e.round=function(v){return arguments.length?(u=!!v,m()):u},e.padding=function(v){return arguments.length?(c=Math.min(1,d=+v),m()):c},e.paddingInner=function(v){return arguments.length?(c=Math.min(1,v),m()):c},e.paddingOuter=function(v){return arguments.length?(d=+v,m()):d},e.align=function(v){return arguments.length?(p=Math.max(0,Math.min(1,v)),m()):p},e.copy=function(){return Tf(t(),[r,i]).round(u).paddingInner(c).paddingOuter(d).align(p)},Ir.apply(m(),arguments)}function IU(e){var t=e.copy;return e.padding=e.paddingOuter,delete e.paddingInner,delete e.paddingOuter,e.copy=function(){return IU(t())},e}function Kc(){return IU(Tf.apply(null,arguments).paddingInner(1))}function nO(e,t,n){e.prototype=t.prototype=n,n.constructor=e}function LU(e,t){var n=Object.create(e.prototype);for(var r in t)n[r]=t[r];return n}function Ed(){}var Mf=.7,gm=1/Mf,Rl="\\s*([+-]?\\d+)\\s*",Pf="\\s*([+-]?(?:\\d*\\.)?\\d+(?:[eE][+-]?\\d+)?)\\s*",bi="\\s*([+-]?(?:\\d*\\.)?\\d+(?:[eE][+-]?\\d+)?)%\\s*",zse=/^#([0-9a-f]{3,8})$/,Use=new RegExp(`^rgb\\(${Rl},${Rl},${Rl}\\)$`),Vse=new RegExp(`^rgb\\(${bi},${bi},${bi}\\)$`),Hse=new RegExp(`^rgba\\(${Rl},${Rl},${Rl},${Pf}\\)$`),Gse=new RegExp(`^rgba\\(${bi},${bi},${bi},${Pf}\\)$`),Wse=new RegExp(`^hsl\\(${Pf},${bi},${bi}\\)$`),Kse=new RegExp(`^hsla\\(${Pf},${bi},${bi},${Pf}\\)$`),ak={aliceblue:15792383,antiquewhite:16444375,aqua:65535,aquamarine:8388564,azure:15794175,beige:16119260,bisque:16770244,black:0,blanchedalmond:16772045,blue:255,blueviolet:9055202,brown:10824234,burlywood:14596231,cadetblue:6266528,chartreuse:8388352,chocolate:13789470,coral:16744272,cornflowerblue:6591981,cornsilk:16775388,crimson:14423100,cyan:65535,darkblue:139,darkcyan:35723,darkgoldenrod:12092939,darkgray:11119017,darkgreen:25600,darkgrey:11119017,darkkhaki:12433259,darkmagenta:9109643,darkolivegreen:5597999,darkorange:16747520,darkorchid:10040012,darkred:9109504,darksalmon:15308410,darkseagreen:9419919,darkslateblue:4734347,darkslategray:3100495,darkslategrey:3100495,darkturquoise:52945,darkviolet:9699539,deeppink:16716947,deepskyblue:49151,dimgray:6908265,dimgrey:6908265,dodgerblue:2003199,firebrick:11674146,floralwhite:16775920,forestgreen:2263842,fuchsia:16711935,gainsboro:14474460,ghostwhite:16316671,gold:16766720,goldenrod:14329120,gray:8421504,green:32768,greenyellow:11403055,grey:8421504,honeydew:15794160,hotpink:16738740,indianred:13458524,indigo:4915330,ivory:16777200,khaki:15787660,lavender:15132410,lavenderblush:16773365,lawngreen:8190976,lemonchiffon:16775885,lightblue:11393254,lightcoral:15761536,lightcyan:14745599,lightgoldenrodyellow:16448210,lightgray:13882323,lightgreen:9498256,lightgrey:13882323,lightpink:16758465,lightsalmon:16752762,lightseagreen:2142890,lightskyblue:8900346,lightslategray:7833753,lightslategrey:7833753,lightsteelblue:11584734,lightyellow:16777184,lime:65280,limegreen:3329330,linen:16445670,magenta:16711935,maroon:8388608,mediumaquamarine:6737322,mediumblue:205,mediumorchid:12211667,mediumpurple:9662683,mediumseagreen:3978097,mediumslateblue:8087790,mediumspringgreen:64154,mediumturquoise:4772300,mediumvioletred:13047173,midnightblue:1644912,mintcream:16121850,mistyrose:16770273,moccasin:16770229,navajowhite:16768685,navy:128,oldlace:16643558,olive:8421376,olivedrab:7048739,orange:16753920,orangered:16729344,orchid:14315734,palegoldenrod:15657130,palegreen:10025880,paleturquoise:11529966,palevioletred:14381203,papayawhip:16773077,peachpuff:16767673,peru:13468991,pink:16761035,plum:14524637,powderblue:11591910,purple:8388736,rebeccapurple:6697881,red:16711680,rosybrown:12357519,royalblue:4286945,saddlebrown:9127187,salmon:16416882,sandybrown:16032864,seagreen:3050327,seashell:16774638,sienna:10506797,silver:12632256,skyblue:8900331,slateblue:6970061,slategray:7372944,slategrey:7372944,snow:16775930,springgreen:65407,steelblue:4620980,tan:13808780,teal:32896,thistle:14204888,tomato:16737095,turquoise:4251856,violet:15631086,wheat:16113331,white:16777215,whitesmoke:16119285,yellow:16776960,yellowgreen:10145074};nO(Ed,jf,{copy(e){return Object.assign(new this.constructor,this,e)},displayable(){return this.rgb().displayable()},hex:ok,formatHex:ok,formatHex8:Yse,formatHsl:Zse,formatRgb:sk,toString:sk});function ok(){return this.rgb().formatHex()}function Yse(){return this.rgb().formatHex8()}function Zse(){return BU(this).formatHsl()}function sk(){return this.rgb().formatRgb()}function jf(e){var t,n;return e=(e+"").trim().toLowerCase(),(t=zse.exec(e))?(n=t[1].length,t=parseInt(t[1],16),n===6?lk(t):n===3?new Yn(t>>8&15|t>>4&240,t>>4&15|t&240,(t&15)<<4|t&15,1):n===8?sh(t>>24&255,t>>16&255,t>>8&255,(t&255)/255):n===4?sh(t>>12&15|t>>8&240,t>>8&15|t>>4&240,t>>4&15|t&240,((t&15)<<4|t&15)/255):null):(t=Use.exec(e))?new Yn(t[1],t[2],t[3],1):(t=Vse.exec(e))?new Yn(t[1]*255/100,t[2]*255/100,t[3]*255/100,1):(t=Hse.exec(e))?sh(t[1],t[2],t[3],t[4]):(t=Gse.exec(e))?sh(t[1]*255/100,t[2]*255/100,t[3]*255/100,t[4]):(t=Wse.exec(e))?fk(t[1],t[2]/100,t[3]/100,1):(t=Kse.exec(e))?fk(t[1],t[2]/100,t[3]/100,t[4]):ak.hasOwnProperty(e)?lk(ak[e]):e==="transparent"?new Yn(NaN,NaN,NaN,0):null}function lk(e){return new Yn(e>>16&255,e>>8&255,e&255,1)}function sh(e,t,n,r){return r<=0&&(e=t=n=NaN),new Yn(e,t,n,r)}function Xse(e){return e instanceof Ed||(e=jf(e)),e?(e=e.rgb(),new Yn(e.r,e.g,e.b,e.opacity)):new Yn}function M1(e,t,n,r){return arguments.length===1?Xse(e):new Yn(e,t,n,r??1)}function Yn(e,t,n,r){this.r=+e,this.g=+t,this.b=+n,this.opacity=+r}nO(Yn,M1,LU(Ed,{brighter(e){return e=e==null?gm:Math.pow(gm,e),new Yn(this.r*e,this.g*e,this.b*e,this.opacity)},darker(e){return e=e==null?Mf:Math.pow(Mf,e),new Yn(this.r*e,this.g*e,this.b*e,this.opacity)},rgb(){return this},clamp(){return new Yn(ts(this.r),ts(this.g),ts(this.b),ym(this.opacity))},displayable(){return-.5<=this.r&&this.r<255.5&&-.5<=this.g&&this.g<255.5&&-.5<=this.b&&this.b<255.5&&0<=this.opacity&&this.opacity<=1},hex:uk,formatHex:uk,formatHex8:Qse,formatRgb:ck,toString:ck}));function uk(){return`#${Yo(this.r)}${Yo(this.g)}${Yo(this.b)}`}function Qse(){return`#${Yo(this.r)}${Yo(this.g)}${Yo(this.b)}${Yo((isNaN(this.opacity)?1:this.opacity)*255)}`}function ck(){const e=ym(this.opacity);return`${e===1?"rgb(":"rgba("}${ts(this.r)}, ${ts(this.g)}, ${ts(this.b)}${e===1?")":`, ${e})`}`}function ym(e){return isNaN(e)?1:Math.max(0,Math.min(1,e))}function ts(e){return Math.max(0,Math.min(255,Math.round(e)||0))}function Yo(e){return e=ts(e),(e<16?"0":"")+e.toString(16)}function fk(e,t,n,r){return r<=0?e=t=n=NaN:n<=0||n>=1?e=t=NaN:t<=0&&(e=NaN),new Wr(e,t,n,r)}function BU(e){if(e instanceof Wr)return new Wr(e.h,e.s,e.l,e.opacity);if(e instanceof Ed||(e=jf(e)),!e)return new Wr;if(e instanceof Wr)return e;e=e.rgb();var t=e.r/255,n=e.g/255,r=e.b/255,i=Math.min(t,n,r),a=Math.max(t,n,r),s=NaN,u=a-i,c=(a+i)/2;return u?(t===a?s=(n-r)/u+(n<r)*6:n===a?s=(r-t)/u+2:s=(t-n)/u+4,u/=c<.5?a+i:2-a-i,s*=60):u=c>0&&c<1?0:s,new Wr(s,u,c,e.opacity)}function Jse(e,t,n,r){return arguments.length===1?BU(e):new Wr(e,t,n,r??1)}function Wr(e,t,n,r){this.h=+e,this.s=+t,this.l=+n,this.opacity=+r}nO(Wr,Jse,LU(Ed,{brighter(e){return e=e==null?gm:Math.pow(gm,e),new Wr(this.h,this.s,this.l*e,this.opacity)},darker(e){return e=e==null?Mf:Math.pow(Mf,e),new Wr(this.h,this.s,this.l*e,this.opacity)},rgb(){var e=this.h%360+(this.h<0)*360,t=isNaN(e)||isNaN(this.s)?0:this.s,n=this.l,r=n+(n<.5?n:1-n)*t,i=2*n-r;return new Yn(gS(e>=240?e-240:e+120,i,r),gS(e,i,r),gS(e<120?e+240:e-120,i,r),this.opacity)},clamp(){return new Wr(dk(this.h),lh(this.s),lh(this.l),ym(this.opacity))},displayable(){return(0<=this.s&&this.s<=1||isNaN(this.s))&&0<=this.l&&this.l<=1&&0<=this.opacity&&this.opacity<=1},formatHsl(){const e=ym(this.opacity);return`${e===1?"hsl(":"hsla("}${dk(this.h)}, ${lh(this.s)*100}%, ${lh(this.l)*100}%${e===1?")":`, ${e})`}`}}));function dk(e){return e=(e||0)%360,e<0?e+360:e}function lh(e){return Math.max(0,Math.min(1,e||0))}function gS(e,t,n){return(e<60?t+(n-t)*e/60:e<180?n:e<240?t+(n-t)*(240-e)/60:t)*255}const rO=e=>()=>e;function ele(e,t){return function(n){return e+n*t}}function tle(e,t,n){return e=Math.pow(e,n),t=Math.pow(t,n)-e,n=1/n,function(r){return Math.pow(e+r*t,n)}}function nle(e){return(e=+e)==1?FU:function(t,n){return n-t?tle(t,n,e):rO(isNaN(t)?n:t)}}function FU(e,t){var n=t-e;return n?ele(e,n):rO(isNaN(e)?t:e)}const pk=function e(t){var n=nle(t);function r(i,a){var s=n((i=M1(i)).r,(a=M1(a)).r),u=n(i.g,a.g),c=n(i.b,a.b),d=FU(i.opacity,a.opacity);return function(p){return i.r=s(p),i.g=u(p),i.b=c(p),i.opacity=d(p),i+""}}return r.gamma=e,r}(1);function rle(e,t){t||(t=[]);var n=e?Math.min(t.length,e.length):0,r=t.slice(),i;return function(a){for(i=0;i<n;++i)r[i]=e[i]*(1-a)+t[i]*a;return r}}function ile(e){return ArrayBuffer.isView(e)&&!(e instanceof DataView)}function ale(e,t){var n=t?t.length:0,r=e?Math.min(n,e.length):0,i=new Array(r),a=new Array(n),s;for(s=0;s<r;++s)i[s]=xu(e[s],t[s]);for(;s<n;++s)a[s]=t[s];return function(u){for(s=0;s<r;++s)a[s]=i[s](u);return a}}function ole(e,t){var n=new Date;return e=+e,t=+t,function(r){return n.setTime(e*(1-r)+t*r),n}}function bm(e,t){return e=+e,t=+t,function(n){return e*(1-n)+t*n}}function sle(e,t){var n={},r={},i;(e===null||typeof e!="object")&&(e={}),(t===null||typeof t!="object")&&(t={});for(i in t)i in e?n[i]=xu(e[i],t[i]):r[i]=t[i];return function(a){for(i in n)r[i]=n[i](a);return r}}var P1=/[-+]?(?:\d+\.?\d*|\.?\d+)(?:[eE][-+]?\d+)?/g,yS=new RegExp(P1.source,"g");function lle(e){return function(){return e}}function ule(e){return function(t){return e(t)+""}}function cle(e,t){var n=P1.lastIndex=yS.lastIndex=0,r,i,a,s=-1,u=[],c=[];for(e=e+"",t=t+"";(r=P1.exec(e))&&(i=yS.exec(t));)(a=i.index)>n&&(a=t.slice(n,a),u[s]?u[s]+=a:u[++s]=a),(r=r[0])===(i=i[0])?u[s]?u[s]+=i:u[++s]=i:(u[++s]=null,c.push({i:s,x:bm(r,i)})),n=yS.lastIndex;return n<t.length&&(a=t.slice(n),u[s]?u[s]+=a:u[++s]=a),u.length<2?c[0]?ule(c[0].x):lle(t):(t=c.length,function(d){for(var p=0,m;p<t;++p)u[(m=c[p]).i]=m.x(d);return u.join("")})}function xu(e,t){var n=typeof t,r;return t==null||n==="boolean"?rO(t):(n==="number"?bm:n==="string"?(r=jf(t))?(t=r,pk):cle:t instanceof jf?pk:t instanceof Date?ole:ile(t)?rle:Array.isArray(t)?ale:typeof t.valueOf!="function"&&typeof t.toString!="function"||isNaN(t)?sle:bm)(e,t)}function iO(e,t){return e=+e,t=+t,function(n){return Math.round(e*(1-n)+t*n)}}function fle(e,t){t===void 0&&(t=e,e=xu);for(var n=0,r=t.length-1,i=t[0],a=new Array(r<0?0:r);n<r;)a[n]=e(i,i=t[++n]);return function(s){var u=Math.max(0,Math.min(r-1,Math.floor(s*=r)));return a[u](s-u)}}function dle(e){return function(){return e}}function xm(e){return+e}var hk=[0,1];function Fn(e){return e}function j1(e,t){return(t-=e=+e)?function(n){return(n-e)/t}:dle(isNaN(t)?NaN:.5)}function ple(e,t){var n;return e>t&&(n=e,e=t,t=n),function(r){return Math.max(e,Math.min(t,r))}}function hle(e,t,n){var r=e[0],i=e[1],a=t[0],s=t[1];return i<r?(r=j1(i,r),a=n(s,a)):(r=j1(r,i),a=n(a,s)),function(u){return a(r(u))}}function mle(e,t,n){var r=Math.min(e.length,t.length)-1,i=new Array(r),a=new Array(r),s=-1;for(e[r]<e[0]&&(e=e.slice().reverse(),t=t.slice().reverse());++s<r;)i[s]=j1(e[s],e[s+1]),a[s]=n(t[s],t[s+1]);return function(u){var c=_d(e,u,1,r)-1;return a[c](i[c](u))}}function Cd(e,t){return t.domain(e.domain()).range(e.range()).interpolate(e.interpolate()).clamp(e.clamp()).unknown(e.unknown())}function Vv(){var e=hk,t=hk,n=xu,r,i,a,s=Fn,u,c,d;function p(){var v=Math.min(e.length,t.length);return s!==Fn&&(s=ple(e[0],e[v-1])),u=v>2?mle:hle,c=d=null,m}function m(v){return v==null||isNaN(v=+v)?a:(c||(c=u(e.map(r),t,n)))(r(s(v)))}return m.invert=function(v){return s(i((d||(d=u(t,e.map(r),bm)))(v)))},m.domain=function(v){return arguments.length?(e=Array.from(v,xm),p()):e.slice()},m.range=function(v){return arguments.length?(t=Array.from(v),p()):t.slice()},m.rangeRound=function(v){return t=Array.from(v),n=iO,p()},m.clamp=function(v){return arguments.length?(s=v?!0:Fn,p()):s!==Fn},m.interpolate=function(v){return arguments.length?(n=v,p()):n},m.unknown=function(v){return arguments.length?(a=v,m):a},function(v,b){return r=v,i=b,p()}}function aO(){return Vv()(Fn,Fn)}function vle(e){return Math.abs(e=Math.round(e))>=1e21?e.toLocaleString("en").replace(/,/g,""):e.toString(10)}function wm(e,t){if((n=(e=t?e.toExponential(t-1):e.toExponential()).indexOf("e"))<0)return null;var n,r=e.slice(0,n);return[r.length>1?r[0]+r.slice(2):r,+e.slice(n+1)]}function ql(e){return e=wm(Math.abs(e)),e?e[1]:NaN}function gle(e,t){return function(n,r){for(var i=n.length,a=[],s=0,u=e[0],c=0;i>0&&u>0&&(c+u+1>r&&(u=Math.max(1,r-c)),a.push(n.substring(i-=u,i+u)),!((c+=u+1)>r));)u=e[s=(s+1)%e.length];return a.reverse().join(t)}}function yle(e){return function(t){return t.replace(/[0-9]/g,function(n){return e[+n]})}}var ble=/^(?:(.)?([<>=^]))?([+\-( ])?([$#])?(0)?(\d+)?(,)?(\.\d+)?(~)?([a-z%])?$/i;function Nf(e){if(!(t=ble.exec(e)))throw new Error("invalid format: "+e);var t;return new oO({fill:t[1],align:t[2],sign:t[3],symbol:t[4],zero:t[5],width:t[6],comma:t[7],precision:t[8]&&t[8].slice(1),trim:t[9],type:t[10]})}Nf.prototype=oO.prototype;function oO(e){this.fill=e.fill===void 0?" ":e.fill+"",this.align=e.align===void 0?">":e.align+"",this.sign=e.sign===void 0?"-":e.sign+"",this.symbol=e.symbol===void 0?"":e.symbol+"",this.zero=!!e.zero,this.width=e.width===void 0?void 0:+e.width,this.comma=!!e.comma,this.precision=e.precision===void 0?void 0:+e.precision,this.trim=!!e.trim,this.type=e.type===void 0?"":e.type+""}oO.prototype.toString=function(){return this.fill+this.align+this.sign+this.symbol+(this.zero?"0":"")+(this.width===void 0?"":Math.max(1,this.width|0))+(this.comma?",":"")+(this.precision===void 0?"":"."+Math.max(0,this.precision|0))+(this.trim?"~":"")+this.type};function xle(e){e:for(var t=e.length,n=1,r=-1,i;n<t;++n)switch(e[n]){case".":r=i=n;break;case"0":r===0&&(r=n),i=n;break;default:if(!+e[n])break e;r>0&&(r=0);break}return r>0?e.slice(0,r)+e.slice(i+1):e}var qU;function wle(e,t){var n=wm(e,t);if(!n)return e+"";var r=n[0],i=n[1],a=i-(qU=Math.max(-8,Math.min(8,Math.floor(i/3)))*3)+1,s=r.length;return a===s?r:a>s?r+new Array(a-s+1).join("0"):a>0?r.slice(0,a)+"."+r.slice(a):"0."+new Array(1-a).join("0")+wm(e,Math.max(0,t+a-1))[0]}function mk(e,t){var n=wm(e,t);if(!n)return e+"";var r=n[0],i=n[1];return i<0?"0."+new Array(-i).join("0")+r:r.length>i+1?r.slice(0,i+1)+"."+r.slice(i+1):r+new Array(i-r.length+2).join("0")}const vk={"%":(e,t)=>(e*100).toFixed(t),b:e=>Math.round(e).toString(2),c:e=>e+"",d:vle,e:(e,t)=>e.toExponential(t),f:(e,t)=>e.toFixed(t),g:(e,t)=>e.toPrecision(t),o:e=>Math.round(e).toString(8),p:(e,t)=>mk(e*100,t),r:mk,s:wle,X:e=>Math.round(e).toString(16).toUpperCase(),x:e=>Math.round(e).toString(16)};function gk(e){return e}var yk=Array.prototype.map,bk=["y","z","a","f","p","n","µ","m","","k","M","G","T","P","E","Z","Y"];function Sle(e){var t=e.grouping===void 0||e.thousands===void 0?gk:gle(yk.call(e.grouping,Number),e.thousands+""),n=e.currency===void 0?"":e.currency[0]+"",r=e.currency===void 0?"":e.currency[1]+"",i=e.decimal===void 0?".":e.decimal+"",a=e.numerals===void 0?gk:yle(yk.call(e.numerals,String)),s=e.percent===void 0?"%":e.percent+"",u=e.minus===void 0?"−":e.minus+"",c=e.nan===void 0?"NaN":e.nan+"";function d(m){m=Nf(m);var v=m.fill,b=m.align,S=m.sign,w=m.symbol,x=m.zero,C=m.width,O=m.comma,A=m.precision,T=m.trim,R=m.type;R==="n"?(O=!0,R="g"):vk[R]||(A===void 0&&(A=12),T=!0,R="g"),(x||v==="0"&&b==="=")&&(x=!0,v="0",b="=");var M=w==="$"?n:w==="#"&&/[boxX]/.test(R)?"0"+R.toLowerCase():"",P=w==="$"?r:/[%p]/.test(R)?s:"",N=vk[R],k=/[defgprs%]/.test(R);A=A===void 0?6:/[gprs]/.test(R)?Math.max(1,Math.min(21,A)):Math.max(0,Math.min(20,A));function H(L){var F=M,z=P,Z,Y,I;if(R==="c")z=N(L)+z,L="";else{L=+L;var W=L<0||1/L<0;if(L=isNaN(L)?c:N(Math.abs(L),A),T&&(L=xle(L)),W&&+L==0&&S!=="+"&&(W=!1),F=(W?S==="("?S:u:S==="-"||S==="("?"":S)+F,z=(R==="s"?bk[8+qU/3]:"")+z+(W&&S==="("?")":""),k){for(Z=-1,Y=L.length;++Z<Y;)if(I=L.charCodeAt(Z),48>I||I>57){z=(I===46?i+L.slice(Z+1):L.slice(Z))+z,L=L.slice(0,Z);break}}}O&&!x&&(L=t(L,1/0));var B=F.length+L.length+z.length,Q=B<C?new Array(C-B+1).join(v):"";switch(O&&x&&(L=t(Q+L,Q.length?C-z.length:1/0),Q=""),b){case"<":L=F+L+z+Q;break;case"=":L=F+Q+L+z;break;case"^":L=Q.slice(0,B=Q.length>>1)+F+L+z+Q.slice(B);break;default:L=Q+F+L+z;break}return a(L)}return H.toString=function(){return m+""},H}function p(m,v){var b=d((m=Nf(m),m.type="f",m)),S=Math.max(-8,Math.min(8,Math.floor(ql(v)/3)))*3,w=Math.pow(10,-S),x=bk[8+S/3];return function(C){return b(w*C)+x}}return{format:d,formatPrefix:p}}var uh,sO,zU;_le({thousands:",",grouping:[3],currency:["$",""]});function _le(e){return uh=Sle(e),sO=uh.format,zU=uh.formatPrefix,uh}function Ele(e){return Math.max(0,-ql(Math.abs(e)))}function Cle(e,t){return Math.max(0,Math.max(-8,Math.min(8,Math.floor(ql(t)/3)))*3-ql(Math.abs(e)))}function Ole(e,t){return e=Math.abs(e),t=Math.abs(t)-e,Math.max(0,ql(t)-ql(e))+1}function UU(e,t,n,r){var i=R1(e,t,n),a;switch(r=Nf(r??",f"),r.type){case"s":{var s=Math.max(Math.abs(e),Math.abs(t));return r.precision==null&&!isNaN(a=Cle(i,s))&&(r.precision=a),zU(r,s)}case"":case"e":case"g":case"p":case"r":{r.precision==null&&!isNaN(a=Ole(i,Math.max(Math.abs(e),Math.abs(t))))&&(r.precision=a-(r.type==="e"));break}case"f":case"%":{r.precision==null&&!isNaN(a=Ele(i))&&(r.precision=a-(r.type==="%")*2);break}}return sO(r)}function vo(e){var t=e.domain;return e.ticks=function(n){var r=t();return O1(r[0],r[r.length-1],n??10)},e.tickFormat=function(n,r){var i=t();return UU(i[0],i[i.length-1],n??10,r)},e.nice=function(n){n==null&&(n=10);var r=t(),i=0,a=r.length-1,s=r[i],u=r[a],c,d,p=10;for(u<s&&(d=s,s=u,u=d,d=i,i=a,a=d);p-- >0;){if(d=A1(s,u,n),d===c)return r[i]=s,r[a]=u,t(r);if(d>0)s=Math.floor(s/d)*d,u=Math.ceil(u/d)*d;else if(d<0)s=Math.ceil(s*d)/d,u=Math.floor(u*d)/d;else break;c=d}return e},e}function Sm(){var e=aO();return e.copy=function(){return Cd(e,Sm())},Ir.apply(e,arguments),vo(e)}function VU(e){var t;function n(r){return r==null||isNaN(r=+r)?t:r}return n.invert=n,n.domain=n.range=function(r){return arguments.length?(e=Array.from(r,xm),n):e.slice()},n.unknown=function(r){return arguments.length?(t=r,n):t},n.copy=function(){return VU(e).unknown(t)},e=arguments.length?Array.from(e,xm):[0,1],vo(n)}function HU(e,t){e=e.slice();var n=0,r=e.length-1,i=e[n],a=e[r],s;return a<i&&(s=n,n=r,r=s,s=i,i=a,a=s),e[n]=t.floor(i),e[r]=t.ceil(a),e}function xk(e){return Math.log(e)}function wk(e){return Math.exp(e)}function Ale(e){return-Math.log(-e)}function Rle(e){return-Math.exp(-e)}function Tle(e){return isFinite(e)?+("1e"+e):e<0?0:e}function Mle(e){return e===10?Tle:e===Math.E?Math.exp:t=>Math.pow(e,t)}function Ple(e){return e===Math.E?Math.log:e===10&&Math.log10||e===2&&Math.log2||(e=Math.log(e),t=>Math.log(t)/e)}function Sk(e){return(t,n)=>-e(-t,n)}function lO(e){const t=e(xk,wk),n=t.domain;let r=10,i,a;function s(){return i=Ple(r),a=Mle(r),n()[0]<0?(i=Sk(i),a=Sk(a),e(Ale,Rle)):e(xk,wk),t}return t.base=function(u){return arguments.length?(r=+u,s()):r},t.domain=function(u){return arguments.length?(n(u),s()):n()},t.ticks=u=>{const c=n();let d=c[0],p=c[c.length-1];const m=p<d;m&&([d,p]=[p,d]);let v=i(d),b=i(p),S,w;const x=u==null?10:+u;let C=[];if(!(r%1)&&b-v<x){if(v=Math.floor(v),b=Math.ceil(b),d>0){for(;v<=b;++v)for(S=1;S<r;++S)if(w=v<0?S/a(-v):S*a(v),!(w<d)){if(w>p)break;C.push(w)}}else for(;v<=b;++v)for(S=r-1;S>=1;--S)if(w=v>0?S/a(-v):S*a(v),!(w<d)){if(w>p)break;C.push(w)}C.length*2<x&&(C=O1(d,p,x))}else C=O1(v,b,Math.min(b-v,x)).map(a);return m?C.reverse():C},t.tickFormat=(u,c)=>{if(u==null&&(u=10),c==null&&(c=r===10?"s":","),typeof c!="function"&&(!(r%1)&&(c=Nf(c)).precision==null&&(c.trim=!0),c=sO(c)),u===1/0)return c;const d=Math.max(1,r*u/t.ticks().length);return p=>{let m=p/a(Math.round(i(p)));return m*r<r-.5&&(m*=r),m<=d?c(p):""}},t.nice=()=>n(HU(n(),{floor:u=>a(Math.floor(i(u))),ceil:u=>a(Math.ceil(i(u)))})),t}function GU(){const e=lO(Vv()).domain([1,10]);return e.copy=()=>Cd(e,GU()).base(e.base()),Ir.apply(e,arguments),e}function _k(e){return function(t){return Math.sign(t)*Math.log1p(Math.abs(t/e))}}function Ek(e){return function(t){return Math.sign(t)*Math.expm1(Math.abs(t))*e}}function uO(e){var t=1,n=e(_k(t),Ek(t));return n.constant=function(r){return arguments.length?e(_k(t=+r),Ek(t)):t},vo(n)}function WU(){var e=uO(Vv());return e.copy=function(){return Cd(e,WU()).constant(e.constant())},Ir.apply(e,arguments)}function Ck(e){return function(t){return t<0?-Math.pow(-t,e):Math.pow(t,e)}}function jle(e){return e<0?-Math.sqrt(-e):Math.sqrt(e)}function Nle(e){return e<0?-e*e:e*e}function cO(e){var t=e(Fn,Fn),n=1;function r(){return n===1?e(Fn,Fn):n===.5?e(jle,Nle):e(Ck(n),Ck(1/n))}return t.exponent=function(i){return arguments.length?(n=+i,r()):n},vo(t)}function fO(){var e=cO(Vv());return e.copy=function(){return Cd(e,fO()).exponent(e.exponent())},Ir.apply(e,arguments),e}function Dle(){return fO.apply(null,arguments).exponent(.5)}function Ok(e){return Math.sign(e)*e*e}function kle(e){return Math.sign(e)*Math.sqrt(Math.abs(e))}function KU(){var e=aO(),t=[0,1],n=!1,r;function i(a){var s=kle(e(a));return isNaN(s)?r:n?Math.round(s):s}return i.invert=function(a){return e.invert(Ok(a))},i.domain=function(a){return arguments.length?(e.domain(a),i):e.domain()},i.range=function(a){return arguments.length?(e.range((t=Array.from(a,xm)).map(Ok)),i):t.slice()},i.rangeRound=function(a){return i.range(a).round(!0)},i.round=function(a){return arguments.length?(n=!!a,i):n},i.clamp=function(a){return arguments.length?(e.clamp(a),i):e.clamp()},i.unknown=function(a){return arguments.length?(r=a,i):r},i.copy=function(){return KU(e.domain(),t).round(n).clamp(e.clamp()).unknown(r)},Ir.apply(i,arguments),vo(i)}function YU(){var e=[],t=[],n=[],r;function i(){var s=0,u=Math.max(1,t.length);for(n=new Array(u-1);++s<u;)n[s-1]=Fse(e,s/u);return a}function a(s){return s==null||isNaN(s=+s)?r:t[_d(n,s)]}return a.invertExtent=function(s){var u=t.indexOf(s);return u<0?[NaN,NaN]:[u>0?n[u-1]:e[0],u<n.length?n[u]:e[e.length-1]]},a.domain=function(s){if(!arguments.length)return e.slice();e=[];for(let u of s)u!=null&&!isNaN(u=+u)&&e.push(u);return e.sort(no),i()},a.range=function(s){return arguments.length?(t=Array.from(s),i()):t.slice()},a.unknown=function(s){return arguments.length?(r=s,a):r},a.quantiles=function(){return n.slice()},a.copy=function(){return YU().domain(e).range(t).unknown(r)},Ir.apply(a,arguments)}function ZU(){var e=0,t=1,n=1,r=[.5],i=[0,1],a;function s(c){return c!=null&&c<=c?i[_d(r,c,0,n)]:a}function u(){var c=-1;for(r=new Array(n);++c<n;)r[c]=((c+1)*t-(c-n)*e)/(n+1);return s}return s.domain=function(c){return arguments.length?([e,t]=c,e=+e,t=+t,u()):[e,t]},s.range=function(c){return arguments.length?(n=(i=Array.from(c)).length-1,u()):i.slice()},s.invertExtent=function(c){var d=i.indexOf(c);return d<0?[NaN,NaN]:d<1?[e,r[0]]:d>=n?[r[n-1],t]:[r[d-1],r[d]]},s.unknown=function(c){return arguments.length&&(a=c),s},s.thresholds=function(){return r.slice()},s.copy=function(){return ZU().domain([e,t]).range(i).unknown(a)},Ir.apply(vo(s),arguments)}function XU(){var e=[.5],t=[0,1],n,r=1;function i(a){return a!=null&&a<=a?t[_d(e,a,0,r)]:n}return i.domain=function(a){return arguments.length?(e=Array.from(a),r=Math.min(e.length,t.length-1),i):e.slice()},i.range=function(a){return arguments.length?(t=Array.from(a),r=Math.min(e.length,t.length-1),i):t.slice()},i.invertExtent=function(a){var s=t.indexOf(a);return[e[s-1],e[s]]},i.unknown=function(a){return arguments.length?(n=a,i):n},i.copy=function(){return XU().domain(e).range(t).unknown(n)},Ir.apply(i,arguments)}const bS=new Date,xS=new Date;function fn(e,t,n,r){function i(a){return e(a=arguments.length===0?new Date:new Date(+a)),a}return i.floor=a=>(e(a=new Date(+a)),a),i.ceil=a=>(e(a=new Date(a-1)),t(a,1),e(a),a),i.round=a=>{const s=i(a),u=i.ceil(a);return a-s<u-a?s:u},i.offset=(a,s)=>(t(a=new Date(+a),s==null?1:Math.floor(s)),a),i.range=(a,s,u)=>{const c=[];if(a=i.ceil(a),u=u==null?1:Math.floor(u),!(a<s)||!(u>0))return c;let d;do c.push(d=new Date(+a)),t(a,u),e(a);while(d<a&&a<s);return c},i.filter=a=>fn(s=>{if(s>=s)for(;e(s),!a(s);)s.setTime(s-1)},(s,u)=>{if(s>=s)if(u<0)for(;++u<=0;)for(;t(s,-1),!a(s););else for(;--u>=0;)for(;t(s,1),!a(s););}),n&&(i.count=(a,s)=>(bS.setTime(+a),xS.setTime(+s),e(bS),e(xS),Math.floor(n(bS,xS))),i.every=a=>(a=Math.floor(a),!isFinite(a)||!(a>0)?null:a>1?i.filter(r?s=>r(s)%a===0:s=>i.count(0,s)%a===0):i)),i}const _m=fn(()=>{},(e,t)=>{e.setTime(+e+t)},(e,t)=>t-e);_m.every=e=>(e=Math.floor(e),!isFinite(e)||!(e>0)?null:e>1?fn(t=>{t.setTime(Math.floor(t/e)*e)},(t,n)=>{t.setTime(+t+n*e)},(t,n)=>(n-t)/e):_m);_m.range;const na=1e3,jr=na*60,ra=jr*60,ha=ra*24,dO=ha*7,Ak=ha*30,wS=ha*365,Zo=fn(e=>{e.setTime(e-e.getMilliseconds())},(e,t)=>{e.setTime(+e+t*na)},(e,t)=>(t-e)/na,e=>e.getUTCSeconds());Zo.range;const pO=fn(e=>{e.setTime(e-e.getMilliseconds()-e.getSeconds()*na)},(e,t)=>{e.setTime(+e+t*jr)},(e,t)=>(t-e)/jr,e=>e.getMinutes());pO.range;const hO=fn(e=>{e.setUTCSeconds(0,0)},(e,t)=>{e.setTime(+e+t*jr)},(e,t)=>(t-e)/jr,e=>e.getUTCMinutes());hO.range;const mO=fn(e=>{e.setTime(e-e.getMilliseconds()-e.getSeconds()*na-e.getMinutes()*jr)},(e,t)=>{e.setTime(+e+t*ra)},(e,t)=>(t-e)/ra,e=>e.getHours());mO.range;const vO=fn(e=>{e.setUTCMinutes(0,0,0)},(e,t)=>{e.setTime(+e+t*ra)},(e,t)=>(t-e)/ra,e=>e.getUTCHours());vO.range;const Od=fn(e=>e.setHours(0,0,0,0),(e,t)=>e.setDate(e.getDate()+t),(e,t)=>(t-e-(t.getTimezoneOffset()-e.getTimezoneOffset())*jr)/ha,e=>e.getDate()-1);Od.range;const Hv=fn(e=>{e.setUTCHours(0,0,0,0)},(e,t)=>{e.setUTCDate(e.getUTCDate()+t)},(e,t)=>(t-e)/ha,e=>e.getUTCDate()-1);Hv.range;const QU=fn(e=>{e.setUTCHours(0,0,0,0)},(e,t)=>{e.setUTCDate(e.getUTCDate()+t)},(e,t)=>(t-e)/ha,e=>Math.floor(e/ha));QU.range;function ys(e){return fn(t=>{t.setDate(t.getDate()-(t.getDay()+7-e)%7),t.setHours(0,0,0,0)},(t,n)=>{t.setDate(t.getDate()+n*7)},(t,n)=>(n-t-(n.getTimezoneOffset()-t.getTimezoneOffset())*jr)/dO)}const Gv=ys(0),Em=ys(1),$le=ys(2),Ile=ys(3),zl=ys(4),Lle=ys(5),Ble=ys(6);Gv.range;Em.range;$le.range;Ile.range;zl.range;Lle.range;Ble.range;function bs(e){return fn(t=>{t.setUTCDate(t.getUTCDate()-(t.getUTCDay()+7-e)%7),t.setUTCHours(0,0,0,0)},(t,n)=>{t.setUTCDate(t.getUTCDate()+n*7)},(t,n)=>(n-t)/dO)}const Wv=bs(0),Cm=bs(1),Fle=bs(2),qle=bs(3),Ul=bs(4),zle=bs(5),Ule=bs(6);Wv.range;Cm.range;Fle.range;qle.range;Ul.range;zle.range;Ule.range;const gO=fn(e=>{e.setDate(1),e.setHours(0,0,0,0)},(e,t)=>{e.setMonth(e.getMonth()+t)},(e,t)=>t.getMonth()-e.getMonth()+(t.getFullYear()-e.getFullYear())*12,e=>e.getMonth());gO.range;const yO=fn(e=>{e.setUTCDate(1),e.setUTCHours(0,0,0,0)},(e,t)=>{e.setUTCMonth(e.getUTCMonth()+t)},(e,t)=>t.getUTCMonth()-e.getUTCMonth()+(t.getUTCFullYear()-e.getUTCFullYear())*12,e=>e.getUTCMonth());yO.range;const ma=fn(e=>{e.setMonth(0,1),e.setHours(0,0,0,0)},(e,t)=>{e.setFullYear(e.getFullYear()+t)},(e,t)=>t.getFullYear()-e.getFullYear(),e=>e.getFullYear());ma.every=e=>!isFinite(e=Math.floor(e))||!(e>0)?null:fn(t=>{t.setFullYear(Math.floor(t.getFullYear()/e)*e),t.setMonth(0,1),t.setHours(0,0,0,0)},(t,n)=>{t.setFullYear(t.getFullYear()+n*e)});ma.range;const va=fn(e=>{e.setUTCMonth(0,1),e.setUTCHours(0,0,0,0)},(e,t)=>{e.setUTCFullYear(e.getUTCFullYear()+t)},(e,t)=>t.getUTCFullYear()-e.getUTCFullYear(),e=>e.getUTCFullYear());va.every=e=>!isFinite(e=Math.floor(e))||!(e>0)?null:fn(t=>{t.setUTCFullYear(Math.floor(t.getUTCFullYear()/e)*e),t.setUTCMonth(0,1),t.setUTCHours(0,0,0,0)},(t,n)=>{t.setUTCFullYear(t.getUTCFullYear()+n*e)});va.range;function JU(e,t,n,r,i,a){const s=[[Zo,1,na],[Zo,5,5*na],[Zo,15,15*na],[Zo,30,30*na],[a,1,jr],[a,5,5*jr],[a,15,15*jr],[a,30,30*jr],[i,1,ra],[i,3,3*ra],[i,6,6*ra],[i,12,12*ra],[r,1,ha],[r,2,2*ha],[n,1,dO],[t,1,Ak],[t,3,3*Ak],[e,1,wS]];function u(d,p,m){const v=p<d;v&&([d,p]=[p,d]);const b=m&&typeof m.range=="function"?m:c(d,p,m),S=b?b.range(d,+p+1):[];return v?S.reverse():S}function c(d,p,m){const v=Math.abs(p-d)/m,b=eO(([,,x])=>x).right(s,v);if(b===s.length)return e.every(R1(d/wS,p/wS,m));if(b===0)return _m.every(Math.max(R1(d,p,m),1));const[S,w]=s[v/s[b-1][2]<s[b][2]/v?b-1:b];return S.every(w)}return[u,c]}const[Vle,Hle]=JU(va,yO,Wv,QU,vO,hO),[Gle,Wle]=JU(ma,gO,Gv,Od,mO,pO);function SS(e){if(0<=e.y&&e.y<100){var t=new Date(-1,e.m,e.d,e.H,e.M,e.S,e.L);return t.setFullYear(e.y),t}return new Date(e.y,e.m,e.d,e.H,e.M,e.S,e.L)}function _S(e){if(0<=e.y&&e.y<100){var t=new Date(Date.UTC(-1,e.m,e.d,e.H,e.M,e.S,e.L));return t.setUTCFullYear(e.y),t}return new Date(Date.UTC(e.y,e.m,e.d,e.H,e.M,e.S,e.L))}function Pc(e,t,n){return{y:e,m:t,d:n,H:0,M:0,S:0,L:0}}function Kle(e){var t=e.dateTime,n=e.date,r=e.time,i=e.periods,a=e.days,s=e.shortDays,u=e.months,c=e.shortMonths,d=jc(i),p=Nc(i),m=jc(a),v=Nc(a),b=jc(s),S=Nc(s),w=jc(u),x=Nc(u),C=jc(c),O=Nc(c),A={a:W,A:B,b:Q,B:$,c:null,d:Nk,e:Nk,f:gue,g:Aue,G:Tue,H:hue,I:mue,j:vue,L:eV,m:yue,M:bue,p:J,q,Q:$k,s:Ik,S:xue,u:wue,U:Sue,V:_ue,w:Eue,W:Cue,x:null,X:null,y:Oue,Y:Rue,Z:Mue,"%":kk},T={a:X,A:oe,b:le,B:ve,c:null,d:Dk,e:Dk,f:Due,g:Vue,G:Gue,H:Pue,I:jue,j:Nue,L:nV,m:kue,M:$ue,p:de,q:Se,Q:$k,s:Ik,S:Iue,u:Lue,U:Bue,V:Fue,w:que,W:zue,x:null,X:null,y:Uue,Y:Hue,Z:Wue,"%":kk},R={a:H,A:L,b:F,B:z,c:Z,d:Pk,e:Pk,f:cue,g:Mk,G:Tk,H:jk,I:jk,j:oue,L:uue,m:aue,M:sue,p:k,q:iue,Q:due,s:pue,S:lue,u:Jle,U:eue,V:tue,w:Qle,W:nue,x:Y,X:I,y:Mk,Y:Tk,Z:rue,"%":fue};A.x=M(n,A),A.X=M(r,A),A.c=M(t,A),T.x=M(n,T),T.X=M(r,T),T.c=M(t,T);function M(ce,we){return function(_e){var se=[],Ye=-1,Be=0,dt=ce.length,ft,it,xt;for(_e instanceof Date||(_e=new Date(+_e));++Ye<dt;)ce.charCodeAt(Ye)===37&&(se.push(ce.slice(Be,Ye)),(it=Rk[ft=ce.charAt(++Ye)])!=null?ft=ce.charAt(++Ye):it=ft==="e"?" ":"0",(xt=we[ft])&&(ft=xt(_e,it)),se.push(ft),Be=Ye+1);return se.push(ce.slice(Be,Ye)),se.join("")}}function P(ce,we){return function(_e){var se=Pc(1900,void 0,1),Ye=N(se,ce,_e+="",0),Be,dt;if(Ye!=_e.length)return null;if("Q"in se)return new Date(se.Q);if("s"in se)return new Date(se.s*1e3+("L"in se?se.L:0));if(we&&!("Z"in se)&&(se.Z=0),"p"in se&&(se.H=se.H%12+se.p*12),se.m===void 0&&(se.m="q"in se?se.q:0),"V"in se){if(se.V<1||se.V>53)return null;"w"in se||(se.w=1),"Z"in se?(Be=_S(Pc(se.y,0,1)),dt=Be.getUTCDay(),Be=dt>4||dt===0?Cm.ceil(Be):Cm(Be),Be=Hv.offset(Be,(se.V-1)*7),se.y=Be.getUTCFullYear(),se.m=Be.getUTCMonth(),se.d=Be.getUTCDate()+(se.w+6)%7):(Be=SS(Pc(se.y,0,1)),dt=Be.getDay(),Be=dt>4||dt===0?Em.ceil(Be):Em(Be),Be=Od.offset(Be,(se.V-1)*7),se.y=Be.getFullYear(),se.m=Be.getMonth(),se.d=Be.getDate()+(se.w+6)%7)}else("W"in se||"U"in se)&&("w"in se||(se.w="u"in se?se.u%7:"W"in se?1:0),dt="Z"in se?_S(Pc(se.y,0,1)).getUTCDay():SS(Pc(se.y,0,1)).getDay(),se.m=0,se.d="W"in se?(se.w+6)%7+se.W*7-(dt+5)%7:se.w+se.U*7-(dt+6)%7);return"Z"in se?(se.H+=se.Z/100|0,se.M+=se.Z%100,_S(se)):SS(se)}}function N(ce,we,_e,se){for(var Ye=0,Be=we.length,dt=_e.length,ft,it;Ye<Be;){if(se>=dt)return-1;if(ft=we.charCodeAt(Ye++),ft===37){if(ft=we.charAt(Ye++),it=R[ft in Rk?we.charAt(Ye++):ft],!it||(se=it(ce,_e,se))<0)return-1}else if(ft!=_e.charCodeAt(se++))return-1}return se}function k(ce,we,_e){var se=d.exec(we.slice(_e));return se?(ce.p=p.get(se[0].toLowerCase()),_e+se[0].length):-1}function H(ce,we,_e){var se=b.exec(we.slice(_e));return se?(ce.w=S.get(se[0].toLowerCase()),_e+se[0].length):-1}function L(ce,we,_e){var se=m.exec(we.slice(_e));return se?(ce.w=v.get(se[0].toLowerCase()),_e+se[0].length):-1}function F(ce,we,_e){var se=C.exec(we.slice(_e));return se?(ce.m=O.get(se[0].toLowerCase()),_e+se[0].length):-1}function z(ce,we,_e){var se=w.exec(we.slice(_e));return se?(ce.m=x.get(se[0].toLowerCase()),_e+se[0].length):-1}function Z(ce,we,_e){return N(ce,t,we,_e)}function Y(ce,we,_e){return N(ce,n,we,_e)}function I(ce,we,_e){return N(ce,r,we,_e)}function W(ce){return s[ce.getDay()]}function B(ce){return a[ce.getDay()]}function Q(ce){return c[ce.getMonth()]}function $(ce){return u[ce.getMonth()]}function J(ce){return i[+(ce.getHours()>=12)]}function q(ce){return 1+~~(ce.getMonth()/3)}function X(ce){return s[ce.getUTCDay()]}function oe(ce){return a[ce.getUTCDay()]}function le(ce){return c[ce.getUTCMonth()]}function ve(ce){return u[ce.getUTCMonth()]}function de(ce){return i[+(ce.getUTCHours()>=12)]}function Se(ce){return 1+~~(ce.getUTCMonth()/3)}return{format:function(ce){var we=M(ce+="",A);return we.toString=function(){return ce},we},parse:function(ce){var we=P(ce+="",!1);return we.toString=function(){return ce},we},utcFormat:function(ce){var we=M(ce+="",T);return we.toString=function(){return ce},we},utcParse:function(ce){var we=P(ce+="",!0);return we.toString=function(){return ce},we}}}var Rk={"-":"",_:" ",0:"0"},yn=/^\s*\d+/,Yle=/^%/,Zle=/[\\^$*+?|[\]().{}]/g;function ht(e,t,n){var r=e<0?"-":"",i=(r?-e:e)+"",a=i.length;return r+(a<n?new Array(n-a+1).join(t)+i:i)}function Xle(e){return e.replace(Zle,"\\$&")}function jc(e){return new RegExp("^(?:"+e.map(Xle).join("|")+")","i")}function Nc(e){return new Map(e.map((t,n)=>[t.toLowerCase(),n]))}function Qle(e,t,n){var r=yn.exec(t.slice(n,n+1));return r?(e.w=+r[0],n+r[0].length):-1}function Jle(e,t,n){var r=yn.exec(t.slice(n,n+1));return r?(e.u=+r[0],n+r[0].length):-1}function eue(e,t,n){var r=yn.exec(t.slice(n,n+2));return r?(e.U=+r[0],n+r[0].length):-1}function tue(e,t,n){var r=yn.exec(t.slice(n,n+2));return r?(e.V=+r[0],n+r[0].length):-1}function nue(e,t,n){var r=yn.exec(t.slice(n,n+2));return r?(e.W=+r[0],n+r[0].length):-1}function Tk(e,t,n){var r=yn.exec(t.slice(n,n+4));return r?(e.y=+r[0],n+r[0].length):-1}function Mk(e,t,n){var r=yn.exec(t.slice(n,n+2));return r?(e.y=+r[0]+(+r[0]>68?1900:2e3),n+r[0].length):-1}function rue(e,t,n){var r=/^(Z)|([+-]\d\d)(?::?(\d\d))?/.exec(t.slice(n,n+6));return r?(e.Z=r[1]?0:-(r[2]+(r[3]||"00")),n+r[0].length):-1}function iue(e,t,n){var r=yn.exec(t.slice(n,n+1));return r?(e.q=r[0]*3-3,n+r[0].length):-1}function aue(e,t,n){var r=yn.exec(t.slice(n,n+2));return r?(e.m=r[0]-1,n+r[0].length):-1}function Pk(e,t,n){var r=yn.exec(t.slice(n,n+2));return r?(e.d=+r[0],n+r[0].length):-1}function oue(e,t,n){var r=yn.exec(t.slice(n,n+3));return r?(e.m=0,e.d=+r[0],n+r[0].length):-1}function jk(e,t,n){var r=yn.exec(t.slice(n,n+2));return r?(e.H=+r[0],n+r[0].length):-1}function sue(e,t,n){var r=yn.exec(t.slice(n,n+2));return r?(e.M=+r[0],n+r[0].length):-1}function lue(e,t,n){var r=yn.exec(t.slice(n,n+2));return r?(e.S=+r[0],n+r[0].length):-1}function uue(e,t,n){var r=yn.exec(t.slice(n,n+3));return r?(e.L=+r[0],n+r[0].length):-1}function cue(e,t,n){var r=yn.exec(t.slice(n,n+6));return r?(e.L=Math.floor(r[0]/1e3),n+r[0].length):-1}function fue(e,t,n){var r=Yle.exec(t.slice(n,n+1));return r?n+r[0].length:-1}function due(e,t,n){var r=yn.exec(t.slice(n));return r?(e.Q=+r[0],n+r[0].length):-1}function pue(e,t,n){var r=yn.exec(t.slice(n));return r?(e.s=+r[0],n+r[0].length):-1}function Nk(e,t){return ht(e.getDate(),t,2)}function hue(e,t){return ht(e.getHours(),t,2)}function mue(e,t){return ht(e.getHours()%12||12,t,2)}function vue(e,t){return ht(1+Od.count(ma(e),e),t,3)}function eV(e,t){return ht(e.getMilliseconds(),t,3)}function gue(e,t){return eV(e,t)+"000"}function yue(e,t){return ht(e.getMonth()+1,t,2)}function bue(e,t){return ht(e.getMinutes(),t,2)}function xue(e,t){return ht(e.getSeconds(),t,2)}function wue(e){var t=e.getDay();return t===0?7:t}function Sue(e,t){return ht(Gv.count(ma(e)-1,e),t,2)}function tV(e){var t=e.getDay();return t>=4||t===0?zl(e):zl.ceil(e)}function _ue(e,t){return e=tV(e),ht(zl.count(ma(e),e)+(ma(e).getDay()===4),t,2)}function Eue(e){return e.getDay()}function Cue(e,t){return ht(Em.count(ma(e)-1,e),t,2)}function Oue(e,t){return ht(e.getFullYear()%100,t,2)}function Aue(e,t){return e=tV(e),ht(e.getFullYear()%100,t,2)}function Rue(e,t){return ht(e.getFullYear()%1e4,t,4)}function Tue(e,t){var n=e.getDay();return e=n>=4||n===0?zl(e):zl.ceil(e),ht(e.getFullYear()%1e4,t,4)}function Mue(e){var t=e.getTimezoneOffset();return(t>0?"-":(t*=-1,"+"))+ht(t/60|0,"0",2)+ht(t%60,"0",2)}function Dk(e,t){return ht(e.getUTCDate(),t,2)}function Pue(e,t){return ht(e.getUTCHours(),t,2)}function jue(e,t){return ht(e.getUTCHours()%12||12,t,2)}function Nue(e,t){return ht(1+Hv.count(va(e),e),t,3)}function nV(e,t){return ht(e.getUTCMilliseconds(),t,3)}function Due(e,t){return nV(e,t)+"000"}function kue(e,t){return ht(e.getUTCMonth()+1,t,2)}function $ue(e,t){return ht(e.getUTCMinutes(),t,2)}function Iue(e,t){return ht(e.getUTCSeconds(),t,2)}function Lue(e){var t=e.getUTCDay();return t===0?7:t}function Bue(e,t){return ht(Wv.count(va(e)-1,e),t,2)}function rV(e){var t=e.getUTCDay();return t>=4||t===0?Ul(e):Ul.ceil(e)}function Fue(e,t){return e=rV(e),ht(Ul.count(va(e),e)+(va(e).getUTCDay()===4),t,2)}function que(e){return e.getUTCDay()}function zue(e,t){return ht(Cm.count(va(e)-1,e),t,2)}function Uue(e,t){return ht(e.getUTCFullYear()%100,t,2)}function Vue(e,t){return e=rV(e),ht(e.getUTCFullYear()%100,t,2)}function Hue(e,t){return ht(e.getUTCFullYear()%1e4,t,4)}function Gue(e,t){var n=e.getUTCDay();return e=n>=4||n===0?Ul(e):Ul.ceil(e),ht(e.getUTCFullYear()%1e4,t,4)}function Wue(){return"+0000"}function kk(){return"%"}function $k(e){return+e}function Ik(e){return Math.floor(+e/1e3)}var ml,iV,aV;Kue({dateTime:"%x, %X",date:"%-m/%-d/%Y",time:"%-I:%M:%S %p",periods:["AM","PM"],days:["Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday"],shortDays:["Sun","Mon","Tue","Wed","Thu","Fri","Sat"],months:["January","February","March","April","May","June","July","August","September","October","November","December"],shortMonths:["Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"]});function Kue(e){return ml=Kle(e),iV=ml.format,ml.parse,aV=ml.utcFormat,ml.utcParse,ml}function Yue(e){return new Date(e)}function Zue(e){return e instanceof Date?+e:+new Date(+e)}function bO(e,t,n,r,i,a,s,u,c,d){var p=aO(),m=p.invert,v=p.domain,b=d(".%L"),S=d(":%S"),w=d("%I:%M"),x=d("%I %p"),C=d("%a %d"),O=d("%b %d"),A=d("%B"),T=d("%Y");function R(M){return(c(M)<M?b:u(M)<M?S:s(M)<M?w:a(M)<M?x:r(M)<M?i(M)<M?C:O:n(M)<M?A:T)(M)}return p.invert=function(M){return new Date(m(M))},p.domain=function(M){return arguments.length?v(Array.from(M,Zue)):v().map(Yue)},p.ticks=function(M){var P=v();return e(P[0],P[P.length-1],M??10)},p.tickFormat=function(M,P){return P==null?R:d(P)},p.nice=function(M){var P=v();return(!M||typeof M.range!="function")&&(M=t(P[0],P[P.length-1],M??10)),M?v(HU(P,M)):p},p.copy=function(){return Cd(p,bO(e,t,n,r,i,a,s,u,c,d))},p}function Xue(){return Ir.apply(bO(Gle,Wle,ma,gO,Gv,Od,mO,pO,Zo,iV).domain([new Date(2e3,0,1),new Date(2e3,0,2)]),arguments)}function Que(){return Ir.apply(bO(Vle,Hle,va,yO,Wv,Hv,vO,hO,Zo,aV).domain([Date.UTC(2e3,0,1),Date.UTC(2e3,0,2)]),arguments)}function Kv(){var e=0,t=1,n,r,i,a,s=Fn,u=!1,c;function d(m){return m==null||isNaN(m=+m)?c:s(i===0?.5:(m=(a(m)-n)*i,u?Math.max(0,Math.min(1,m)):m))}d.domain=function(m){return arguments.length?([e,t]=m,n=a(e=+e),r=a(t=+t),i=n===r?0:1/(r-n),d):[e,t]},d.clamp=function(m){return arguments.length?(u=!!m,d):u},d.interpolator=function(m){return arguments.length?(s=m,d):s};function p(m){return function(v){var b,S;return arguments.length?([b,S]=v,s=m(b,S),d):[s(0),s(1)]}}return d.range=p(xu),d.rangeRound=p(iO),d.unknown=function(m){return arguments.length?(c=m,d):c},function(m){return a=m,n=m(e),r=m(t),i=n===r?0:1/(r-n),d}}function go(e,t){return t.domain(e.domain()).interpolator(e.interpolator()).clamp(e.clamp()).unknown(e.unknown())}function oV(){var e=vo(Kv()(Fn));return e.copy=function(){return go(e,oV())},Sa.apply(e,arguments)}function sV(){var e=lO(Kv()).domain([1,10]);return e.copy=function(){return go(e,sV()).base(e.base())},Sa.apply(e,arguments)}function lV(){var e=uO(Kv());return e.copy=function(){return go(e,lV()).constant(e.constant())},Sa.apply(e,arguments)}function xO(){var e=cO(Kv());return e.copy=function(){return go(e,xO()).exponent(e.exponent())},Sa.apply(e,arguments)}function Jue(){return xO.apply(null,arguments).exponent(.5)}function uV(){var e=[],t=Fn;function n(r){if(r!=null&&!isNaN(r=+r))return t((_d(e,r,1)-1)/(e.length-1))}return n.domain=function(r){if(!arguments.length)return e.slice();e=[];for(let i of r)i!=null&&!isNaN(i=+i)&&e.push(i);return e.sort(no),n},n.interpolator=function(r){return arguments.length?(t=r,n):t},n.range=function(){return e.map((r,i)=>t(i/(e.length-1)))},n.quantiles=function(r){return Array.from({length:r+1},(i,a)=>Bse(e,a/r))},n.copy=function(){return uV(t).domain(e)},Sa.apply(n,arguments)}function Yv(){var e=0,t=.5,n=1,r=1,i,a,s,u,c,d=Fn,p,m=!1,v;function b(w){return isNaN(w=+w)?v:(w=.5+((w=+p(w))-a)*(r*w<r*a?u:c),d(m?Math.max(0,Math.min(1,w)):w))}b.domain=function(w){return arguments.length?([e,t,n]=w,i=p(e=+e),a=p(t=+t),s=p(n=+n),u=i===a?0:.5/(a-i),c=a===s?0:.5/(s-a),r=a<i?-1:1,b):[e,t,n]},b.clamp=function(w){return arguments.length?(m=!!w,b):m},b.interpolator=function(w){return arguments.length?(d=w,b):d};function S(w){return function(x){var C,O,A;return arguments.length?([C,O,A]=x,d=fle(w,[C,O,A]),b):[d(0),d(.5),d(1)]}}return b.range=S(xu),b.rangeRound=S(iO),b.unknown=function(w){return arguments.length?(v=w,b):v},function(w){return p=w,i=w(e),a=w(t),s=w(n),u=i===a?0:.5/(a-i),c=a===s?0:.5/(s-a),r=a<i?-1:1,b}}function cV(){var e=vo(Yv()(Fn));return e.copy=function(){return go(e,cV())},Sa.apply(e,arguments)}function fV(){var e=lO(Yv()).domain([.1,1,10]);return e.copy=function(){return go(e,fV()).base(e.base())},Sa.apply(e,arguments)}function dV(){var e=uO(Yv());return e.copy=function(){return go(e,dV()).constant(e.constant())},Sa.apply(e,arguments)}function wO(){var e=cO(Yv());return e.copy=function(){return go(e,wO()).exponent(e.exponent())},Sa.apply(e,arguments)}function ece(){return wO.apply(null,arguments).exponent(.5)}const Lk=Object.freeze(Object.defineProperty({__proto__:null,scaleBand:Tf,scaleDiverging:cV,scaleDivergingLog:fV,scaleDivergingPow:wO,scaleDivergingSqrt:ece,scaleDivergingSymlog:dV,scaleIdentity:VU,scaleImplicit:T1,scaleLinear:Sm,scaleLog:GU,scaleOrdinal:tO,scalePoint:Kc,scalePow:fO,scaleQuantile:YU,scaleQuantize:ZU,scaleRadial:KU,scaleSequential:oV,scaleSequentialLog:sV,scaleSequentialPow:xO,scaleSequentialQuantile:uV,scaleSequentialSqrt:Jue,scaleSequentialSymlog:lV,scaleSqrt:Dle,scaleSymlog:WU,scaleThreshold:XU,scaleTime:Xue,scaleUtc:Que,tickFormat:UU},Symbol.toStringTag,{value:"Module"}));var ES,Bk;function Zv(){if(Bk)return ES;Bk=1;var e=gu();function t(n,r,i){for(var a=-1,s=n.length;++a<s;){var u=n[a],c=r(u);if(c!=null&&(d===void 0?c===c&&!e(c):i(c,d)))var d=c,p=u}return p}return ES=t,ES}var CS,Fk;function pV(){if(Fk)return CS;Fk=1;function e(t,n){return t>n}return CS=e,CS}var OS,qk;function tce(){if(qk)return OS;qk=1;var e=Zv(),t=pV(),n=bu();function r(i){return i&&i.length?e(i,n,t):void 0}return OS=r,OS}var nce=tce();const Xv=Rt(nce);var AS,zk;function hV(){if(zk)return AS;zk=1;function e(t,n){return t<n}return AS=e,AS}var RS,Uk;function rce(){if(Uk)return RS;Uk=1;var e=Zv(),t=hV(),n=bu();function r(i){return i&&i.length?e(i,n,t):void 0}return RS=r,RS}var ice=rce();const Qv=Rt(ice);var TS,Vk;function ace(){if(Vk)return TS;Vk=1;var e=$C(),t=Pi(),n=SU(),r=Qn();function i(a,s){var u=r(a)?e:n;return u(a,t(s,3))}return TS=i,TS}var MS,Hk;function oce(){if(Hk)return MS;Hk=1;var e=xU(),t=ace();function n(r,i){return e(t(r,i),1)}return MS=n,MS}var sce=oce();const lce=Rt(sce);var PS,Gk;function uce(){if(Gk)return PS;Gk=1;var e=ZC();function t(n,r){return e(n,r)}return PS=t,PS}var cce=uce();const SO=Rt(cce);var wu=1e9,fce={precision:20,rounding:4,toExpNeg:-7,toExpPos:21,LN10:"2.302585092994045684017991454684364207601101488628772976033327900967572609677352480235997205089598298341967784042286"},EO,Lt=!0,Dr="[DecimalError] ",ns=Dr+"Invalid argument: ",_O=Dr+"Exponent out of range: ",Su=Math.floor,Uo=Math.pow,dce=/^(\d+(\.\d*)?|\.\d+)(e[+-]?\d+)?$/i,hr,gn=1e7,kt=7,mV=9007199254740991,Om=Su(mV/kt),Pe={};Pe.absoluteValue=Pe.abs=function(){var e=new this.constructor(this);return e.s&&(e.s=1),e};Pe.comparedTo=Pe.cmp=function(e){var t,n,r,i,a=this;if(e=new a.constructor(e),a.s!==e.s)return a.s||-e.s;if(a.e!==e.e)return a.e>e.e^a.s<0?1:-1;for(r=a.d.length,i=e.d.length,t=0,n=r<i?r:i;t<n;++t)if(a.d[t]!==e.d[t])return a.d[t]>e.d[t]^a.s<0?1:-1;return r===i?0:r>i^a.s<0?1:-1};Pe.decimalPlaces=Pe.dp=function(){var e=this,t=e.d.length-1,n=(t-e.e)*kt;if(t=e.d[t],t)for(;t%10==0;t/=10)n--;return n<0?0:n};Pe.dividedBy=Pe.div=function(e){return ua(this,new this.constructor(e))};Pe.dividedToIntegerBy=Pe.idiv=function(e){var t=this,n=t.constructor;return Pt(ua(t,new n(e),0,1),n.precision)};Pe.equals=Pe.eq=function(e){return!this.cmp(e)};Pe.exponent=function(){return en(this)};Pe.greaterThan=Pe.gt=function(e){return this.cmp(e)>0};Pe.greaterThanOrEqualTo=Pe.gte=function(e){return this.cmp(e)>=0};Pe.isInteger=Pe.isint=function(){return this.e>this.d.length-2};Pe.isNegative=Pe.isneg=function(){return this.s<0};Pe.isPositive=Pe.ispos=function(){return this.s>0};Pe.isZero=function(){return this.s===0};Pe.lessThan=Pe.lt=function(e){return this.cmp(e)<0};Pe.lessThanOrEqualTo=Pe.lte=function(e){return this.cmp(e)<1};Pe.logarithm=Pe.log=function(e){var t,n=this,r=n.constructor,i=r.precision,a=i+5;if(e===void 0)e=new r(10);else if(e=new r(e),e.s<1||e.eq(hr))throw Error(Dr+"NaN");if(n.s<1)throw Error(Dr+(n.s?"NaN":"-Infinity"));return n.eq(hr)?new r(0):(Lt=!1,t=ua(Df(n,a),Df(e,a),a),Lt=!0,Pt(t,i))};Pe.minus=Pe.sub=function(e){var t=this;return e=new t.constructor(e),t.s==e.s?yV(t,e):vV(t,(e.s=-e.s,e))};Pe.modulo=Pe.mod=function(e){var t,n=this,r=n.constructor,i=r.precision;if(e=new r(e),!e.s)throw Error(Dr+"NaN");return n.s?(Lt=!1,t=ua(n,e,0,1).times(e),Lt=!0,n.minus(t)):Pt(new r(n),i)};Pe.naturalExponential=Pe.exp=function(){return gV(this)};Pe.naturalLogarithm=Pe.ln=function(){return Df(this)};Pe.negated=Pe.neg=function(){var e=new this.constructor(this);return e.s=-e.s||0,e};Pe.plus=Pe.add=function(e){var t=this;return e=new t.constructor(e),t.s==e.s?vV(t,e):yV(t,(e.s=-e.s,e))};Pe.precision=Pe.sd=function(e){var t,n,r,i=this;if(e!==void 0&&e!==!!e&&e!==1&&e!==0)throw Error(ns+e);if(t=en(i)+1,r=i.d.length-1,n=r*kt+1,r=i.d[r],r){for(;r%10==0;r/=10)n--;for(r=i.d[0];r>=10;r/=10)n++}return e&&t>n?t:n};Pe.squareRoot=Pe.sqrt=function(){var e,t,n,r,i,a,s,u=this,c=u.constructor;if(u.s<1){if(!u.s)return new c(0);throw Error(Dr+"NaN")}for(e=en(u),Lt=!1,i=Math.sqrt(+u),i==0||i==1/0?(t=mi(u.d),(t.length+e)%2==0&&(t+="0"),i=Math.sqrt(t),e=Su((e+1)/2)-(e<0||e%2),i==1/0?t="5e"+e:(t=i.toExponential(),t=t.slice(0,t.indexOf("e")+1)+e),r=new c(t)):r=new c(i.toString()),n=c.precision,i=s=n+3;;)if(a=r,r=a.plus(ua(u,a,s+2)).times(.5),mi(a.d).slice(0,s)===(t=mi(r.d)).slice(0,s)){if(t=t.slice(s-3,s+1),i==s&&t=="4999"){if(Pt(a,n+1,0),a.times(a).eq(u)){r=a;break}}else if(t!="9999")break;s+=4}return Lt=!0,Pt(r,n)};Pe.times=Pe.mul=function(e){var t,n,r,i,a,s,u,c,d,p=this,m=p.constructor,v=p.d,b=(e=new m(e)).d;if(!p.s||!e.s)return new m(0);for(e.s*=p.s,n=p.e+e.e,c=v.length,d=b.length,c<d&&(a=v,v=b,b=a,s=c,c=d,d=s),a=[],s=c+d,r=s;r--;)a.push(0);for(r=d;--r>=0;){for(t=0,i=c+r;i>r;)u=a[i]+b[r]*v[i-r-1]+t,a[i--]=u%gn|0,t=u/gn|0;a[i]=(a[i]+t)%gn|0}for(;!a[--s];)a.pop();return t?++n:a.shift(),e.d=a,e.e=n,Lt?Pt(e,m.precision):e};Pe.toDecimalPlaces=Pe.todp=function(e,t){var n=this,r=n.constructor;return n=new r(n),e===void 0?n:(Oi(e,0,wu),t===void 0?t=r.rounding:Oi(t,0,8),Pt(n,e+en(n)+1,t))};Pe.toExponential=function(e,t){var n,r=this,i=r.constructor;return e===void 0?n=ps(r,!0):(Oi(e,0,wu),t===void 0?t=i.rounding:Oi(t,0,8),r=Pt(new i(r),e+1,t),n=ps(r,!0,e+1)),n};Pe.toFixed=function(e,t){var n,r,i=this,a=i.constructor;return e===void 0?ps(i):(Oi(e,0,wu),t===void 0?t=a.rounding:Oi(t,0,8),r=Pt(new a(i),e+en(i)+1,t),n=ps(r.abs(),!1,e+en(r)+1),i.isneg()&&!i.isZero()?"-"+n:n)};Pe.toInteger=Pe.toint=function(){var e=this,t=e.constructor;return Pt(new t(e),en(e)+1,t.rounding)};Pe.toNumber=function(){return+this};Pe.toPower=Pe.pow=function(e){var t,n,r,i,a,s,u=this,c=u.constructor,d=12,p=+(e=new c(e));if(!e.s)return new c(hr);if(u=new c(u),!u.s){if(e.s<1)throw Error(Dr+"Infinity");return u}if(u.eq(hr))return u;if(r=c.precision,e.eq(hr))return Pt(u,r);if(t=e.e,n=e.d.length-1,s=t>=n,a=u.s,s){if((n=p<0?-p:p)<=mV){for(i=new c(hr),t=Math.ceil(r/kt+4),Lt=!1;n%2&&(i=i.times(u),Kk(i.d,t)),n=Su(n/2),n!==0;)u=u.times(u),Kk(u.d,t);return Lt=!0,e.s<0?new c(hr).div(i):Pt(i,r)}}else if(a<0)throw Error(Dr+"NaN");return a=a<0&&e.d[Math.max(t,n)]&1?-1:1,u.s=1,Lt=!1,i=e.times(Df(u,r+d)),Lt=!0,i=gV(i),i.s=a,i};Pe.toPrecision=function(e,t){var n,r,i=this,a=i.constructor;return e===void 0?(n=en(i),r=ps(i,n<=a.toExpNeg||n>=a.toExpPos)):(Oi(e,1,wu),t===void 0?t=a.rounding:Oi(t,0,8),i=Pt(new a(i),e,t),n=en(i),r=ps(i,e<=n||n<=a.toExpNeg,e)),r};Pe.toSignificantDigits=Pe.tosd=function(e,t){var n=this,r=n.constructor;return e===void 0?(e=r.precision,t=r.rounding):(Oi(e,1,wu),t===void 0?t=r.rounding:Oi(t,0,8)),Pt(new r(n),e,t)};Pe.toString=Pe.valueOf=Pe.val=Pe.toJSON=Pe[Symbol.for("nodejs.util.inspect.custom")]=function(){var e=this,t=en(e),n=e.constructor;return ps(e,t<=n.toExpNeg||t>=n.toExpPos)};function vV(e,t){var n,r,i,a,s,u,c,d,p=e.constructor,m=p.precision;if(!e.s||!t.s)return t.s||(t=new p(e)),Lt?Pt(t,m):t;if(c=e.d,d=t.d,s=e.e,i=t.e,c=c.slice(),a=s-i,a){for(a<0?(r=c,a=-a,u=d.length):(r=d,i=s,u=c.length),s=Math.ceil(m/kt),u=s>u?s+1:u+1,a>u&&(a=u,r.length=1),r.reverse();a--;)r.push(0);r.reverse()}for(u=c.length,a=d.length,u-a<0&&(a=u,r=d,d=c,c=r),n=0;a;)n=(c[--a]=c[a]+d[a]+n)/gn|0,c[a]%=gn;for(n&&(c.unshift(n),++i),u=c.length;c[--u]==0;)c.pop();return t.d=c,t.e=i,Lt?Pt(t,m):t}function Oi(e,t,n){if(e!==~~e||e<t||e>n)throw Error(ns+e)}function mi(e){var t,n,r,i=e.length-1,a="",s=e[0];if(i>0){for(a+=s,t=1;t<i;t++)r=e[t]+"",n=kt-r.length,n&&(a+=Qa(n)),a+=r;s=e[t],r=s+"",n=kt-r.length,n&&(a+=Qa(n))}else if(s===0)return"0";for(;s%10===0;)s/=10;return a+s}var ua=function(){function e(r,i){var a,s=0,u=r.length;for(r=r.slice();u--;)a=r[u]*i+s,r[u]=a%gn|0,s=a/gn|0;return s&&r.unshift(s),r}function t(r,i,a,s){var u,c;if(a!=s)c=a>s?1:-1;else for(u=c=0;u<a;u++)if(r[u]!=i[u]){c=r[u]>i[u]?1:-1;break}return c}function n(r,i,a){for(var s=0;a--;)r[a]-=s,s=r[a]<i[a]?1:0,r[a]=s*gn+r[a]-i[a];for(;!r[0]&&r.length>1;)r.shift()}return function(r,i,a,s){var u,c,d,p,m,v,b,S,w,x,C,O,A,T,R,M,P,N,k=r.constructor,H=r.s==i.s?1:-1,L=r.d,F=i.d;if(!r.s)return new k(r);if(!i.s)throw Error(Dr+"Division by zero");for(c=r.e-i.e,P=F.length,R=L.length,b=new k(H),S=b.d=[],d=0;F[d]==(L[d]||0);)++d;if(F[d]>(L[d]||0)&&--c,a==null?O=a=k.precision:s?O=a+(en(r)-en(i))+1:O=a,O<0)return new k(0);if(O=O/kt+2|0,d=0,P==1)for(p=0,F=F[0],O++;(d<R||p)&&O--;d++)A=p*gn+(L[d]||0),S[d]=A/F|0,p=A%F|0;else{for(p=gn/(F[0]+1)|0,p>1&&(F=e(F,p),L=e(L,p),P=F.length,R=L.length),T=P,w=L.slice(0,P),x=w.length;x<P;)w[x++]=0;N=F.slice(),N.unshift(0),M=F[0],F[1]>=gn/2&&++M;do p=0,u=t(F,w,P,x),u<0?(C=w[0],P!=x&&(C=C*gn+(w[1]||0)),p=C/M|0,p>1?(p>=gn&&(p=gn-1),m=e(F,p),v=m.length,x=w.length,u=t(m,w,v,x),u==1&&(p--,n(m,P<v?N:F,v))):(p==0&&(u=p=1),m=F.slice()),v=m.length,v<x&&m.unshift(0),n(w,m,x),u==-1&&(x=w.length,u=t(F,w,P,x),u<1&&(p++,n(w,P<x?N:F,x))),x=w.length):u===0&&(p++,w=[0]),S[d++]=p,u&&w[0]?w[x++]=L[T]||0:(w=[L[T]],x=1);while((T++<R||w[0]!==void 0)&&O--)}return S[0]||S.shift(),b.e=c,Pt(b,s?a+en(b)+1:a)}}();function gV(e,t){var n,r,i,a,s,u,c=0,d=0,p=e.constructor,m=p.precision;if(en(e)>16)throw Error(_O+en(e));if(!e.s)return new p(hr);for(Lt=!1,u=m,s=new p(.03125);e.abs().gte(.1);)e=e.times(s),d+=5;for(r=Math.log(Uo(2,d))/Math.LN10*2+5|0,u+=r,n=i=a=new p(hr),p.precision=u;;){if(i=Pt(i.times(e),u),n=n.times(++c),s=a.plus(ua(i,n,u)),mi(s.d).slice(0,u)===mi(a.d).slice(0,u)){for(;d--;)a=Pt(a.times(a),u);return p.precision=m,t==null?(Lt=!0,Pt(a,m)):a}a=s}}function en(e){for(var t=e.e*kt,n=e.d[0];n>=10;n/=10)t++;return t}function jS(e,t,n){if(t>e.LN10.sd())throw Lt=!0,n&&(e.precision=n),Error(Dr+"LN10 precision limit exceeded");return Pt(new e(e.LN10),t)}function Qa(e){for(var t="";e--;)t+="0";return t}function Df(e,t){var n,r,i,a,s,u,c,d,p,m=1,v=10,b=e,S=b.d,w=b.constructor,x=w.precision;if(b.s<1)throw Error(Dr+(b.s?"NaN":"-Infinity"));if(b.eq(hr))return new w(0);if(t==null?(Lt=!1,d=x):d=t,b.eq(10))return t==null&&(Lt=!0),jS(w,d);if(d+=v,w.precision=d,n=mi(S),r=n.charAt(0),a=en(b),Math.abs(a)<15e14){for(;r<7&&r!=1||r==1&&n.charAt(1)>3;)b=b.times(e),n=mi(b.d),r=n.charAt(0),m++;a=en(b),r>1?(b=new w("0."+n),a++):b=new w(r+"."+n.slice(1))}else return c=jS(w,d+2,x).times(a+""),b=Df(new w(r+"."+n.slice(1)),d-v).plus(c),w.precision=x,t==null?(Lt=!0,Pt(b,x)):b;for(u=s=b=ua(b.minus(hr),b.plus(hr),d),p=Pt(b.times(b),d),i=3;;){if(s=Pt(s.times(p),d),c=u.plus(ua(s,new w(i),d)),mi(c.d).slice(0,d)===mi(u.d).slice(0,d))return u=u.times(2),a!==0&&(u=u.plus(jS(w,d+2,x).times(a+""))),u=ua(u,new w(m),d),w.precision=x,t==null?(Lt=!0,Pt(u,x)):u;u=c,i+=2}}function Wk(e,t){var n,r,i;for((n=t.indexOf("."))>-1&&(t=t.replace(".","")),(r=t.search(/e/i))>0?(n<0&&(n=r),n+=+t.slice(r+1),t=t.substring(0,r)):n<0&&(n=t.length),r=0;t.charCodeAt(r)===48;)++r;for(i=t.length;t.charCodeAt(i-1)===48;)--i;if(t=t.slice(r,i),t){if(i-=r,n=n-r-1,e.e=Su(n/kt),e.d=[],r=(n+1)%kt,n<0&&(r+=kt),r<i){for(r&&e.d.push(+t.slice(0,r)),i-=kt;r<i;)e.d.push(+t.slice(r,r+=kt));t=t.slice(r),r=kt-t.length}else r-=i;for(;r--;)t+="0";if(e.d.push(+t),Lt&&(e.e>Om||e.e<-Om))throw Error(_O+n)}else e.s=0,e.e=0,e.d=[0];return e}function Pt(e,t,n){var r,i,a,s,u,c,d,p,m=e.d;for(s=1,a=m[0];a>=10;a/=10)s++;if(r=t-s,r<0)r+=kt,i=t,d=m[p=0];else{if(p=Math.ceil((r+1)/kt),a=m.length,p>=a)return e;for(d=a=m[p],s=1;a>=10;a/=10)s++;r%=kt,i=r-kt+s}if(n!==void 0&&(a=Uo(10,s-i-1),u=d/a%10|0,c=t<0||m[p+1]!==void 0||d%a,c=n<4?(u||c)&&(n==0||n==(e.s<0?3:2)):u>5||u==5&&(n==4||c||n==6&&(r>0?i>0?d/Uo(10,s-i):0:m[p-1])%10&1||n==(e.s<0?8:7))),t<1||!m[0])return c?(a=en(e),m.length=1,t=t-a-1,m[0]=Uo(10,(kt-t%kt)%kt),e.e=Su(-t/kt)||0):(m.length=1,m[0]=e.e=e.s=0),e;if(r==0?(m.length=p,a=1,p--):(m.length=p+1,a=Uo(10,kt-r),m[p]=i>0?(d/Uo(10,s-i)%Uo(10,i)|0)*a:0),c)for(;;)if(p==0){(m[0]+=a)==gn&&(m[0]=1,++e.e);break}else{if(m[p]+=a,m[p]!=gn)break;m[p--]=0,a=1}for(r=m.length;m[--r]===0;)m.pop();if(Lt&&(e.e>Om||e.e<-Om))throw Error(_O+en(e));return e}function yV(e,t){var n,r,i,a,s,u,c,d,p,m,v=e.constructor,b=v.precision;if(!e.s||!t.s)return t.s?t.s=-t.s:t=new v(e),Lt?Pt(t,b):t;if(c=e.d,m=t.d,r=t.e,d=e.e,c=c.slice(),s=d-r,s){for(p=s<0,p?(n=c,s=-s,u=m.length):(n=m,r=d,u=c.length),i=Math.max(Math.ceil(b/kt),u)+2,s>i&&(s=i,n.length=1),n.reverse(),i=s;i--;)n.push(0);n.reverse()}else{for(i=c.length,u=m.length,p=i<u,p&&(u=i),i=0;i<u;i++)if(c[i]!=m[i]){p=c[i]<m[i];break}s=0}for(p&&(n=c,c=m,m=n,t.s=-t.s),u=c.length,i=m.length-u;i>0;--i)c[u++]=0;for(i=m.length;i>s;){if(c[--i]<m[i]){for(a=i;a&&c[--a]===0;)c[a]=gn-1;--c[a],c[i]+=gn}c[i]-=m[i]}for(;c[--u]===0;)c.pop();for(;c[0]===0;c.shift())--r;return c[0]?(t.d=c,t.e=r,Lt?Pt(t,b):t):new v(0)}function ps(e,t,n){var r,i=en(e),a=mi(e.d),s=a.length;return t?(n&&(r=n-s)>0?a=a.charAt(0)+"."+a.slice(1)+Qa(r):s>1&&(a=a.charAt(0)+"."+a.slice(1)),a=a+(i<0?"e":"e+")+i):i<0?(a="0."+Qa(-i-1)+a,n&&(r=n-s)>0&&(a+=Qa(r))):i>=s?(a+=Qa(i+1-s),n&&(r=n-i-1)>0&&(a=a+"."+Qa(r))):((r=i+1)<s&&(a=a.slice(0,r)+"."+a.slice(r)),n&&(r=n-s)>0&&(i+1===s&&(a+="."),a+=Qa(r))),e.s<0?"-"+a:a}function Kk(e,t){if(e.length>t)return e.length=t,!0}function bV(e){var t,n,r;function i(a){var s=this;if(!(s instanceof i))return new i(a);if(s.constructor=i,a instanceof i){s.s=a.s,s.e=a.e,s.d=(a=a.d)?a.slice():a;return}if(typeof a=="number"){if(a*0!==0)throw Error(ns+a);if(a>0)s.s=1;else if(a<0)a=-a,s.s=-1;else{s.s=0,s.e=0,s.d=[0];return}if(a===~~a&&a<1e7){s.e=0,s.d=[a];return}return Wk(s,a.toString())}else if(typeof a!="string")throw Error(ns+a);if(a.charCodeAt(0)===45?(a=a.slice(1),s.s=-1):s.s=1,dce.test(a))Wk(s,a);else throw Error(ns+a)}if(i.prototype=Pe,i.ROUND_UP=0,i.ROUND_DOWN=1,i.ROUND_CEIL=2,i.ROUND_FLOOR=3,i.ROUND_HALF_UP=4,i.ROUND_HALF_DOWN=5,i.ROUND_HALF_EVEN=6,i.ROUND_HALF_CEIL=7,i.ROUND_HALF_FLOOR=8,i.clone=bV,i.config=i.set=pce,e===void 0&&(e={}),e)for(r=["precision","rounding","toExpNeg","toExpPos","LN10"],t=0;t<r.length;)e.hasOwnProperty(n=r[t++])||(e[n]=this[n]);return i.config(e),i}function pce(e){if(!e||typeof e!="object")throw Error(Dr+"Object expected");var t,n,r,i=["precision",1,wu,"rounding",0,8,"toExpNeg",-1/0,0,"toExpPos",0,1/0];for(t=0;t<i.length;t+=3)if((r=e[n=i[t]])!==void 0)if(Su(r)===r&&r>=i[t+1]&&r<=i[t+2])this[n]=r;else throw Error(ns+n+": "+r);if((r=e[n="LN10"])!==void 0)if(r==Math.LN10)this[n]=new this(r);else throw Error(ns+n+": "+r);return this}var EO=bV(fce);hr=new EO(1);const Ot=EO;function hce(e){return yce(e)||gce(e)||vce(e)||mce()}function mce(){throw new TypeError(`Invalid attempt to spread non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function vce(e,t){if(e){if(typeof e=="string")return N1(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);if(n==="Object"&&e.constructor&&(n=e.constructor.name),n==="Map"||n==="Set")return Array.from(e);if(n==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return N1(e,t)}}function gce(e){if(typeof Symbol<"u"&&Symbol.iterator in Object(e))return Array.from(e)}function yce(e){if(Array.isArray(e))return N1(e)}function N1(e,t){(t==null||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}var bce=function(t){return t},xV={},wV=function(t){return t===xV},Yk=function(t){return function n(){return arguments.length===0||arguments.length===1&&wV(arguments.length<=0?void 0:arguments[0])?n:t.apply(void 0,arguments)}},xce=function e(t,n){return t===1?n:Yk(function(){for(var r=arguments.length,i=new Array(r),a=0;a<r;a++)i[a]=arguments[a];var s=i.filter(function(u){return u!==xV}).length;return s>=t?n.apply(void 0,i):e(t-s,Yk(function(){for(var u=arguments.length,c=new Array(u),d=0;d<u;d++)c[d]=arguments[d];var p=i.map(function(m){return wV(m)?c.shift():m});return n.apply(void 0,hce(p).concat(c))}))})},Jv=function(t){return xce(t.length,t)},D1=function(t,n){for(var r=[],i=t;i<n;++i)r[i-t]=i;return r},wce=Jv(function(e,t){return Array.isArray(t)?t.map(e):Object.keys(t).map(function(n){return t[n]}).map(e)}),Sce=function(){for(var t=arguments.length,n=new Array(t),r=0;r<t;r++)n[r]=arguments[r];if(!n.length)return bce;var i=n.reverse(),a=i[0],s=i.slice(1);return function(){return s.reduce(function(u,c){return c(u)},a.apply(void 0,arguments))}},k1=function(t){return Array.isArray(t)?t.reverse():t.split("").reverse.join("")},SV=function(t){var n=null,r=null;return function(){for(var i=arguments.length,a=new Array(i),s=0;s<i;s++)a[s]=arguments[s];return n&&a.every(function(u,c){return u===n[c]})||(n=a,r=t.apply(void 0,a)),r}};function _ce(e){var t;return e===0?t=1:t=Math.floor(new Ot(e).abs().log(10).toNumber())+1,t}function Ece(e,t,n){for(var r=new Ot(e),i=0,a=[];r.lt(t)&&i<1e5;)a.push(r.toNumber()),r=r.add(n),i++;return a}var Cce=Jv(function(e,t,n){var r=+e,i=+t;return r+n*(i-r)}),Oce=Jv(function(e,t,n){var r=t-+e;return r=r||1/0,(n-e)/r}),Ace=Jv(function(e,t,n){var r=t-+e;return r=r||1/0,Math.max(0,Math.min(1,(n-e)/r))});const eg={rangeStep:Ece,getDigitCount:_ce,interpolateNumber:Cce,uninterpolateNumber:Oce,uninterpolateTruncation:Ace};function $1(e){return Mce(e)||Tce(e)||_V(e)||Rce()}function Rce(){throw new TypeError(`Invalid attempt to spread non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function Tce(e){if(typeof Symbol<"u"&&Symbol.iterator in Object(e))return Array.from(e)}function Mce(e){if(Array.isArray(e))return I1(e)}function kf(e,t){return Nce(e)||jce(e,t)||_V(e,t)||Pce()}function Pce(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function _V(e,t){if(e){if(typeof e=="string")return I1(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);if(n==="Object"&&e.constructor&&(n=e.constructor.name),n==="Map"||n==="Set")return Array.from(e);if(n==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return I1(e,t)}}function I1(e,t){(t==null||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}function jce(e,t){if(!(typeof Symbol>"u"||!(Symbol.iterator in Object(e)))){var n=[],r=!0,i=!1,a=void 0;try{for(var s=e[Symbol.iterator](),u;!(r=(u=s.next()).done)&&(n.push(u.value),!(t&&n.length===t));r=!0);}catch(c){i=!0,a=c}finally{try{!r&&s.return!=null&&s.return()}finally{if(i)throw a}}return n}}function Nce(e){if(Array.isArray(e))return e}function EV(e){var t=kf(e,2),n=t[0],r=t[1],i=n,a=r;return n>r&&(i=r,a=n),[i,a]}function CV(e,t,n){if(e.lte(0))return new Ot(0);var r=eg.getDigitCount(e.toNumber()),i=new Ot(10).pow(r),a=e.div(i),s=r!==1?.05:.1,u=new Ot(Math.ceil(a.div(s).toNumber())).add(n).mul(s),c=u.mul(i);return t?c:new Ot(Math.ceil(c))}function Dce(e,t,n){var r=1,i=new Ot(e);if(!i.isint()&&n){var a=Math.abs(e);a<1?(r=new Ot(10).pow(eg.getDigitCount(e)-1),i=new Ot(Math.floor(i.div(r).toNumber())).mul(r)):a>1&&(i=new Ot(Math.floor(e)))}else e===0?i=new Ot(Math.floor((t-1)/2)):n||(i=new Ot(Math.floor(e)));var s=Math.floor((t-1)/2),u=Sce(wce(function(c){return i.add(new Ot(c-s).mul(r)).toNumber()}),D1);return u(0,t)}function OV(e,t,n,r){var i=arguments.length>4&&arguments[4]!==void 0?arguments[4]:0;if(!Number.isFinite((t-e)/(n-1)))return{step:new Ot(0),tickMin:new Ot(0),tickMax:new Ot(0)};var a=CV(new Ot(t).sub(e).div(n-1),r,i),s;e<=0&&t>=0?s=new Ot(0):(s=new Ot(e).add(t).div(2),s=s.sub(new Ot(s).mod(a)));var u=Math.ceil(s.sub(e).div(a).toNumber()),c=Math.ceil(new Ot(t).sub(s).div(a).toNumber()),d=u+c+1;return d>n?OV(e,t,n,r,i+1):(d<n&&(c=t>0?c+(n-d):c,u=t>0?u:u+(n-d)),{step:a,tickMin:s.sub(new Ot(u).mul(a)),tickMax:s.add(new Ot(c).mul(a))})}function kce(e){var t=kf(e,2),n=t[0],r=t[1],i=arguments.length>1&&arguments[1]!==void 0?arguments[1]:6,a=arguments.length>2&&arguments[2]!==void 0?arguments[2]:!0,s=Math.max(i,2),u=EV([n,r]),c=kf(u,2),d=c[0],p=c[1];if(d===-1/0||p===1/0){var m=p===1/0?[d].concat($1(D1(0,i-1).map(function(){return 1/0}))):[].concat($1(D1(0,i-1).map(function(){return-1/0})),[p]);return n>r?k1(m):m}if(d===p)return Dce(d,i,a);var v=OV(d,p,s,a),b=v.step,S=v.tickMin,w=v.tickMax,x=eg.rangeStep(S,w.add(new Ot(.1).mul(b)),b);return n>r?k1(x):x}function $ce(e,t){var n=kf(e,2),r=n[0],i=n[1],a=arguments.length>2&&arguments[2]!==void 0?arguments[2]:!0,s=EV([r,i]),u=kf(s,2),c=u[0],d=u[1];if(c===-1/0||d===1/0)return[r,i];if(c===d)return[c];var p=Math.max(t,2),m=CV(new Ot(d).sub(c).div(p-1),a,0),v=[].concat($1(eg.rangeStep(new Ot(c),new Ot(d).sub(new Ot(.99).mul(m)),m)),[d]);return r>i?k1(v):v}var Ice=SV(kce),Lce=SV($ce),Bce="Invariant failed";function Vl(e,t){throw new Error(Bce)}var Fce=["offset","layout","width","dataKey","data","dataPointFormatter","xAxis","yAxis"];function Hl(e){"@babel/helpers - typeof";return Hl=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},Hl(e)}function Am(){return Am=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)Object.prototype.hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},Am.apply(this,arguments)}function qce(e,t){return Hce(e)||Vce(e,t)||Uce(e,t)||zce()}function zce(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function Uce(e,t){if(e){if(typeof e=="string")return Zk(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);if(n==="Object"&&e.constructor&&(n=e.constructor.name),n==="Map"||n==="Set")return Array.from(e);if(n==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return Zk(e,t)}}function Zk(e,t){(t==null||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}function Vce(e,t){var n=e==null?null:typeof Symbol<"u"&&e[Symbol.iterator]||e["@@iterator"];if(n!=null){var r,i,a,s,u=[],c=!0,d=!1;try{if(a=(n=n.call(e)).next,t!==0)for(;!(c=(r=a.call(n)).done)&&(u.push(r.value),u.length!==t);c=!0);}catch(p){d=!0,i=p}finally{try{if(!c&&n.return!=null&&(s=n.return(),Object(s)!==s))return}finally{if(d)throw i}}return u}}function Hce(e){if(Array.isArray(e))return e}function Gce(e,t){if(e==null)return{};var n=Wce(e,t),r,i;if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);for(i=0;i<a.length;i++)r=a[i],!(t.indexOf(r)>=0)&&Object.prototype.propertyIsEnumerable.call(e,r)&&(n[r]=e[r])}return n}function Wce(e,t){if(e==null)return{};var n={};for(var r in e)if(Object.prototype.hasOwnProperty.call(e,r)){if(t.indexOf(r)>=0)continue;n[r]=e[r]}return n}function Kce(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}function Yce(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,TV(r.key),r)}}function Zce(e,t,n){return t&&Yce(e.prototype,t),Object.defineProperty(e,"prototype",{writable:!1}),e}function Xce(e,t,n){return t=Rm(t),Qce(e,AV()?Reflect.construct(t,n||[],Rm(e).constructor):t.apply(e,n))}function Qce(e,t){if(t&&(Hl(t)==="object"||typeof t=="function"))return t;if(t!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return Jce(e)}function Jce(e){if(e===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return e}function AV(){try{var e=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(AV=function(){return!!e})()}function Rm(e){return Rm=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(n){return n.__proto__||Object.getPrototypeOf(n)},Rm(e)}function efe(e,t){if(typeof t!="function"&&t!==null)throw new TypeError("Super expression must either be null or a function");e.prototype=Object.create(t&&t.prototype,{constructor:{value:e,writable:!0,configurable:!0}}),Object.defineProperty(e,"prototype",{writable:!1}),t&&L1(e,t)}function L1(e,t){return L1=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(r,i){return r.__proto__=i,r},L1(e,t)}function RV(e,t,n){return t=TV(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function TV(e){var t=tfe(e,"string");return Hl(t)=="symbol"?t:t+""}function tfe(e,t){if(Hl(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(Hl(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(e)}var CO=function(e){function t(){return Kce(this,t),Xce(this,t,arguments)}return efe(t,e),Zce(t,[{key:"render",value:function(){var r=this.props,i=r.offset,a=r.layout,s=r.width,u=r.dataKey,c=r.data,d=r.dataPointFormatter,p=r.xAxis,m=r.yAxis,v=Gce(r,Fce),b=Je(v,!1);this.props.direction==="x"&&p.type!=="number"&&Vl();var S=c.map(function(w){var x=d(w,u),C=x.x,O=x.y,A=x.value,T=x.errorVal;if(!T)return null;var R=[],M,P;if(Array.isArray(T)){var N=qce(T,2);M=N[0],P=N[1]}else M=P=T;if(a==="vertical"){var k=p.scale,H=O+i,L=H+s,F=H-s,z=k(A-M),Z=k(A+P);R.push({x1:Z,y1:L,x2:Z,y2:F}),R.push({x1:z,y1:H,x2:Z,y2:H}),R.push({x1:z,y1:L,x2:z,y2:F})}else if(a==="horizontal"){var Y=m.scale,I=C+i,W=I-s,B=I+s,Q=Y(A-M),$=Y(A+P);R.push({x1:W,y1:$,x2:B,y2:$}),R.push({x1:I,y1:Q,x2:I,y2:$}),R.push({x1:W,y1:Q,x2:B,y2:Q})}return V.createElement(un,Am({className:"recharts-errorBar",key:"bar-".concat(R.map(function(J){return"".concat(J.x1,"-").concat(J.x2,"-").concat(J.y1,"-").concat(J.y2)}))},b),R.map(function(J){return V.createElement("line",Am({},J,{key:"line-".concat(J.x1,"-").concat(J.x2,"-").concat(J.y1,"-").concat(J.y2)}))}))});return V.createElement(un,{className:"recharts-errorBars"},S)}}])}(V.Component);RV(CO,"defaultProps",{stroke:"black",strokeWidth:1.5,width:5,offset:0,layout:"horizontal"});RV(CO,"displayName","ErrorBar");function $f(e){"@babel/helpers - typeof";return $f=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},$f(e)}function Xk(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function Bo(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?Xk(Object(n),!0).forEach(function(r){nfe(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):Xk(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function nfe(e,t,n){return t=rfe(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function rfe(e){var t=ife(e,"string");return $f(t)=="symbol"?t:t+""}function ife(e,t){if($f(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if($f(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}var MV=function(t){var n=t.children,r=t.formattedGraphicalItems,i=t.legendWidth,a=t.legendContent,s=Tr(n,Al);if(!s)return null;var u=Al.defaultProps,c=u!==void 0?Bo(Bo({},u),s.props):{},d;return s.props&&s.props.payload?d=s.props&&s.props.payload:a==="children"?d=(r||[]).reduce(function(p,m){var v=m.item,b=m.props,S=b.sectors||b.data||[];return p.concat(S.map(function(w){return{type:s.props.iconType||v.props.legendType,value:w.name,color:w.fill,payload:w}}))},[]):d=(r||[]).map(function(p){var m=p.item,v=m.type.defaultProps,b=v!==void 0?Bo(Bo({},v),m.props):{},S=b.dataKey,w=b.name,x=b.legendType,C=b.hide;return{inactive:C,dataKey:S,type:c.iconType||x||"square",color:OO(m),value:w||S,payload:b}}),Bo(Bo(Bo({},c),Al.getWithHeight(s,i)),{},{payload:d,item:s})};function If(e){"@babel/helpers - typeof";return If=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},If(e)}function Qk(e){return lfe(e)||sfe(e)||ofe(e)||afe()}function afe(){throw new TypeError(`Invalid attempt to spread non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function ofe(e,t){if(e){if(typeof e=="string")return B1(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);if(n==="Object"&&e.constructor&&(n=e.constructor.name),n==="Map"||n==="Set")return Array.from(e);if(n==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return B1(e,t)}}function sfe(e){if(typeof Symbol<"u"&&e[Symbol.iterator]!=null||e["@@iterator"]!=null)return Array.from(e)}function lfe(e){if(Array.isArray(e))return B1(e)}function B1(e,t){(t==null||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}function Jk(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function Wt(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?Jk(Object(n),!0).forEach(function(r){Tl(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):Jk(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function Tl(e,t,n){return t=ufe(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function ufe(e){var t=cfe(e,"string");return If(t)=="symbol"?t:t+""}function cfe(e,t){if(If(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(If(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function ga(e,t,n){return vt(e)||vt(t)?n:sn(t)?yi(e,t,n):st(t)?t(e):n}function Yc(e,t,n,r){var i=lce(e,function(u){return ga(u,t)});if(n==="number"){var a=i.filter(function(u){return $e(u)||parseFloat(u)});return a.length?[Qv(a),Xv(a)]:[1/0,-1/0]}var s=r?i.filter(function(u){return!vt(u)}):i;return s.map(function(u){return sn(u)||u instanceof Date?u:""})}var ffe=function(t){var n,r=arguments.length>1&&arguments[1]!==void 0?arguments[1]:[],i=arguments.length>2?arguments[2]:void 0,a=arguments.length>3?arguments[3]:void 0,s=-1,u=(n=r==null?void 0:r.length)!==null&&n!==void 0?n:0;if(u<=1)return 0;if(a&&a.axisType==="angleAxis"&&Math.abs(Math.abs(a.range[1]-a.range[0])-360)<=1e-6)for(var c=a.range,d=0;d<u;d++){var p=d>0?i[d-1].coordinate:i[u-1].coordinate,m=i[d].coordinate,v=d>=u-1?i[0].coordinate:i[d+1].coordinate,b=void 0;if(to(m-p)!==to(v-m)){var S=[];if(to(v-m)===to(c[1]-c[0])){b=v;var w=m+c[1]-c[0];S[0]=Math.min(w,(w+p)/2),S[1]=Math.max(w,(w+p)/2)}else{b=p;var x=v+c[1]-c[0];S[0]=Math.min(m,(x+m)/2),S[1]=Math.max(m,(x+m)/2)}var C=[Math.min(m,(b+m)/2),Math.max(m,(b+m)/2)];if(t>C[0]&&t<=C[1]||t>=S[0]&&t<=S[1]){s=i[d].index;break}}else{var O=Math.min(p,v),A=Math.max(p,v);if(t>(O+m)/2&&t<=(A+m)/2){s=i[d].index;break}}}else for(var T=0;T<u;T++)if(T===0&&t<=(r[T].coordinate+r[T+1].coordinate)/2||T>0&&T<u-1&&t>(r[T].coordinate+r[T-1].coordinate)/2&&t<=(r[T].coordinate+r[T+1].coordinate)/2||T===u-1&&t>(r[T].coordinate+r[T-1].coordinate)/2){s=r[T].index;break}return s},OO=function(t){var n,r=t,i=r.type.displayName,a=(n=t.type)!==null&&n!==void 0&&n.defaultProps?Wt(Wt({},t.type.defaultProps),t.props):t.props,s=a.stroke,u=a.fill,c;switch(i){case"Line":c=s;break;case"Area":case"Radar":c=s&&s!=="none"?s:u;break;default:c=u;break}return c},dfe=function(t){var n=t.barSize,r=t.totalSize,i=t.stackGroups,a=i===void 0?{}:i;if(!a)return{};for(var s={},u=Object.keys(a),c=0,d=u.length;c<d;c++)for(var p=a[u[c]].stackGroups,m=Object.keys(p),v=0,b=m.length;v<b;v++){var S=p[m[v]],w=S.items,x=S.cateAxisId,C=w.filter(function(P){return sa(P.type).indexOf("Bar")>=0});if(C&&C.length){var O=C[0].type.defaultProps,A=O!==void 0?Wt(Wt({},O),C[0].props):C[0].props,T=A.barSize,R=A[x];s[R]||(s[R]=[]);var M=vt(T)?n:T;s[R].push({item:C[0],stackList:C.slice(1),barSize:vt(M)?void 0:hi(M,r,0)})}}return s},pfe=function(t){var n=t.barGap,r=t.barCategoryGap,i=t.bandSize,a=t.sizeList,s=a===void 0?[]:a,u=t.maxBarSize,c=s.length;if(c<1)return null;var d=hi(n,i,0,!0),p,m=[];if(s[0].barSize===+s[0].barSize){var v=!1,b=i/c,S=s.reduce(function(T,R){return T+R.barSize||0},0);S+=(c-1)*d,S>=i&&(S-=(c-1)*d,d=0),S>=i&&b>0&&(v=!0,b*=.9,S=c*b);var w=(i-S)/2>>0,x={offset:w-d,size:0};p=s.reduce(function(T,R){var M={item:R.item,position:{offset:x.offset+x.size+d,size:v?b:R.barSize}},P=[].concat(Qk(T),[M]);return x=P[P.length-1].position,R.stackList&&R.stackList.length&&R.stackList.forEach(function(N){P.push({item:N,position:x})}),P},m)}else{var C=hi(r,i,0,!0);i-2*C-(c-1)*d<=0&&(d=0);var O=(i-2*C-(c-1)*d)/c;O>1&&(O>>=0);var A=u===+u?Math.min(O,u):O;p=s.reduce(function(T,R,M){var P=[].concat(Qk(T),[{item:R.item,position:{offset:C+(O+d)*M+(O-A)/2,size:A}}]);return R.stackList&&R.stackList.length&&R.stackList.forEach(function(N){P.push({item:N,position:P[P.length-1].position})}),P},m)}return p},hfe=function(t,n,r,i){var a=r.children,s=r.width,u=r.margin,c=s-(u.left||0)-(u.right||0),d=MV({children:a,legendWidth:c});if(d){var p=i||{},m=p.width,v=p.height,b=d.align,S=d.verticalAlign,w=d.layout;if((w==="vertical"||w==="horizontal"&&S==="middle")&&b!=="center"&&$e(t[b]))return Wt(Wt({},t),{},Tl({},b,t[b]+(m||0)));if((w==="horizontal"||w==="vertical"&&b==="center")&&S!=="middle"&&$e(t[S]))return Wt(Wt({},t),{},Tl({},S,t[S]+(v||0)))}return t},mfe=function(t,n,r){return vt(n)?!0:t==="horizontal"?n==="yAxis":t==="vertical"||r==="x"?n==="xAxis":r==="y"?n==="yAxis":!0},PV=function(t,n,r,i,a){var s=n.props.children,u=la(s,CO).filter(function(d){return mfe(i,a,d.props.direction)});if(u&&u.length){var c=u.map(function(d){return d.props.dataKey});return t.reduce(function(d,p){var m=ga(p,r);if(vt(m))return d;var v=Array.isArray(m)?[Qv(m),Xv(m)]:[m,m],b=c.reduce(function(S,w){var x=ga(p,w,0),C=v[0]-Math.abs(Array.isArray(x)?x[0]:x),O=v[1]+Math.abs(Array.isArray(x)?x[1]:x);return[Math.min(C,S[0]),Math.max(O,S[1])]},[1/0,-1/0]);return[Math.min(b[0],d[0]),Math.max(b[1],d[1])]},[1/0,-1/0])}return null},vfe=function(t,n,r,i,a){var s=n.map(function(u){return PV(t,u,r,a,i)}).filter(function(u){return!vt(u)});return s&&s.length?s.reduce(function(u,c){return[Math.min(u[0],c[0]),Math.max(u[1],c[1])]},[1/0,-1/0]):null},jV=function(t,n,r,i,a){var s=n.map(function(c){var d=c.props.dataKey;return r==="number"&&d&&PV(t,c,d,i)||Yc(t,d,r,a)});if(r==="number")return s.reduce(function(c,d){return[Math.min(c[0],d[0]),Math.max(c[1],d[1])]},[1/0,-1/0]);var u={};return s.reduce(function(c,d){for(var p=0,m=d.length;p<m;p++)u[d[p]]||(u[d[p]]=!0,c.push(d[p]));return c},[])},NV=function(t,n){return t==="horizontal"&&n==="xAxis"||t==="vertical"&&n==="yAxis"||t==="centric"&&n==="angleAxis"||t==="radial"&&n==="radiusAxis"},zc=function(t,n,r){if(!t)return null;var i=t.scale,a=t.duplicateDomain,s=t.type,u=t.range,c=t.realScaleType==="scaleBand"?i.bandwidth()/2:2,d=(n||r)&&s==="category"&&i.bandwidth?i.bandwidth()/c:0;if(d=t.axisType==="angleAxis"&&(u==null?void 0:u.length)>=2?to(u[0]-u[1])*2*d:d,n&&(t.ticks||t.niceTicks)){var p=(t.ticks||t.niceTicks).map(function(m){var v=a?a.indexOf(m):m;return{coordinate:i(v)+d,value:m,offset:d}});return p.filter(function(m){return!wd(m.coordinate)})}return t.isCategorical&&t.categoricalDomain?t.categoricalDomain.map(function(m,v){return{coordinate:i(m)+d,value:m,index:v,offset:d}}):i.ticks&&!r?i.ticks(t.tickCount).map(function(m){return{coordinate:i(m)+d,value:m,offset:d}}):i.domain().map(function(m,v){return{coordinate:i(m)+d,value:a?a[m]:m,index:v,offset:d}})},NS=new WeakMap,ch=function(t,n){if(typeof n!="function")return t;NS.has(t)||NS.set(t,new WeakMap);var r=NS.get(t);if(r.has(n))return r.get(n);var i=function(){t.apply(void 0,arguments),n.apply(void 0,arguments)};return r.set(n,i),i},gfe=function(t,n,r){var i=t.scale,a=t.type,s=t.layout,u=t.axisType;if(i==="auto")return s==="radial"&&u==="radiusAxis"?{scale:Tf(),realScaleType:"band"}:s==="radial"&&u==="angleAxis"?{scale:Sm(),realScaleType:"linear"}:a==="category"&&n&&(n.indexOf("LineChart")>=0||n.indexOf("AreaChart")>=0||n.indexOf("ComposedChart")>=0)?{scale:Kc(),realScaleType:"point"}:a==="category"?{scale:Tf(),realScaleType:"band"}:{scale:Sm(),realScaleType:"linear"};if(xd(i)){var c="scale".concat($v(i));return{scale:(Lk[c]||Kc)(),realScaleType:Lk[c]?c:"point"}}return st(i)?{scale:i}:{scale:Kc(),realScaleType:"point"}},e$=1e-4,yfe=function(t){var n=t.domain();if(!(!n||n.length<=2)){var r=n.length,i=t.range(),a=Math.min(i[0],i[1])-e$,s=Math.max(i[0],i[1])+e$,u=t(n[0]),c=t(n[r-1]);(u<a||u>s||c<a||c>s)&&t.domain([n[0],n[r-1]])}},bfe=function(t){var n=t.length;if(!(n<=0))for(var r=0,i=t[0].length;r<i;++r)for(var a=0,s=0,u=0;u<n;++u){var c=wd(t[u][r][1])?t[u][r][0]:t[u][r][1];c>=0?(t[u][r][0]=a,t[u][r][1]=a+c,a=t[u][r][1]):(t[u][r][0]=s,t[u][r][1]=s+c,s=t[u][r][1])}},xfe=function(t){var n=t.length;if(!(n<=0))for(var r=0,i=t[0].length;r<i;++r)for(var a=0,s=0;s<n;++s){var u=wd(t[s][r][1])?t[s][r][0]:t[s][r][1];u>=0?(t[s][r][0]=a,t[s][r][1]=a+u,a=t[s][r][1]):(t[s][r][0]=0,t[s][r][1]=0)}},wfe={sign:bfe,expand:sie,none:$l,silhouette:lie,wiggle:uie,positive:xfe},Sfe=function(t,n,r){var i=n.map(function(u){return u.props.dataKey}),a=wfe[r],s=oie().keys(i).value(function(u,c){return+ga(u,c,0)}).order(m1).offset(a);return s(t)},_fe=function(t,n,r,i,a,s){if(!t)return null;var u=s?n.reverse():n,c={},d=u.reduce(function(m,v){var b,S=(b=v.type)!==null&&b!==void 0&&b.defaultProps?Wt(Wt({},v.type.defaultProps),v.props):v.props,w=S.stackId,x=S.hide;if(x)return m;var C=S[r],O=m[C]||{hasStack:!1,stackGroups:{}};if(sn(w)){var A=O.stackGroups[w]||{numericAxisId:r,cateAxisId:i,items:[]};A.items.push(v),O.hasStack=!0,O.stackGroups[w]=A}else O.stackGroups[LC("_stackId_")]={numericAxisId:r,cateAxisId:i,items:[v]};return Wt(Wt({},m),{},Tl({},C,O))},c),p={};return Object.keys(d).reduce(function(m,v){var b=d[v];if(b.hasStack){var S={};b.stackGroups=Object.keys(b.stackGroups).reduce(function(w,x){var C=b.stackGroups[x];return Wt(Wt({},w),{},Tl({},x,{numericAxisId:r,cateAxisId:i,items:C.items,stackedData:Sfe(t,C.items,a)}))},S)}return Wt(Wt({},m),{},Tl({},v,b))},p)},Efe=function(t,n){var r=n.realScaleType,i=n.type,a=n.tickCount,s=n.originalDomain,u=n.allowDecimals,c=r||n.scale;if(c!=="auto"&&c!=="linear")return null;if(a&&i==="number"&&s&&(s[0]==="auto"||s[1]==="auto")){var d=t.domain();if(!d.length)return null;var p=Ice(d,a,u);return t.domain([Qv(p),Xv(p)]),{niceTicks:p}}if(a&&i==="number"){var m=t.domain(),v=Lce(m,a,u);return{niceTicks:v}}return null},Cfe=function(t,n){var r,i=(r=t.type)!==null&&r!==void 0&&r.defaultProps?Wt(Wt({},t.type.defaultProps),t.props):t.props,a=i.stackId;if(sn(a)){var s=n[a];if(s){var u=s.items.indexOf(t);return u>=0?s.stackedData[u]:null}}return null},Ofe=function(t){return t.reduce(function(n,r){return[Qv(r.concat([n[0]]).filter($e)),Xv(r.concat([n[1]]).filter($e))]},[1/0,-1/0])},DV=function(t,n,r){return Object.keys(t).reduce(function(i,a){var s=t[a],u=s.stackedData,c=u.reduce(function(d,p){var m=Ofe(p.slice(n,r+1));return[Math.min(d[0],m[0]),Math.max(d[1],m[1])]},[1/0,-1/0]);return[Math.min(c[0],i[0]),Math.max(c[1],i[1])]},[1/0,-1/0]).map(function(i){return i===1/0||i===-1/0?0:i})},t$=/^dataMin[\s]*-[\s]*([0-9]+([.]{1}[0-9]+){0,1})$/,n$=/^dataMax[\s]*\+[\s]*([0-9]+([.]{1}[0-9]+){0,1})$/,F1=function(t,n,r){if(st(t))return t(n,r);if(!Array.isArray(t))return n;var i=[];if($e(t[0]))i[0]=r?t[0]:Math.min(t[0],n[0]);else if(t$.test(t[0])){var a=+t$.exec(t[0])[1];i[0]=n[0]-a}else st(t[0])?i[0]=t[0](n[0]):i[0]=n[0];if($e(t[1]))i[1]=r?t[1]:Math.max(t[1],n[1]);else if(n$.test(t[1])){var s=+n$.exec(t[1])[1];i[1]=n[1]+s}else st(t[1])?i[1]=t[1](n[1]):i[1]=n[1];return i},q1=function(t,n,r){if(t&&t.scale&&t.scale.bandwidth){var i=t.scale.bandwidth();if(!r||i>0)return i}if(t&&n&&n.length>=2){for(var a=QC(n,function(m){return m.coordinate}),s=1/0,u=1,c=a.length;u<c;u++){var d=a[u],p=a[u-1];s=Math.min((d.coordinate||0)-(p.coordinate||0),s)}return s===1/0?0:s}return r?void 0:0},r$=function(t,n,r){return!t||!t.length||SO(t,yi(r,"type.defaultProps.domain"))?n:t},Afe=function(t,n){var r=t.type.defaultProps?Wt(Wt({},t.type.defaultProps),t.props):t.props,i=r.dataKey,a=r.name,s=r.unit,u=r.formatter,c=r.tooltipType,d=r.chartType,p=r.hide;return Wt(Wt({},Je(t,!1)),{},{dataKey:i,unit:s,formatter:u,name:a||i,color:OO(t),value:ga(n,i),type:c,payload:n,chartType:d,hide:p})};function Lf(e){"@babel/helpers - typeof";return Lf=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},Lf(e)}function i$(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function Ji(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?i$(Object(n),!0).forEach(function(r){kV(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):i$(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function kV(e,t,n){return t=Rfe(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function Rfe(e){var t=Tfe(e,"string");return Lf(t)=="symbol"?t:t+""}function Tfe(e,t){if(Lf(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(Lf(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function Mfe(e,t){return Dfe(e)||Nfe(e,t)||jfe(e,t)||Pfe()}function Pfe(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function jfe(e,t){if(e){if(typeof e=="string")return a$(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);if(n==="Object"&&e.constructor&&(n=e.constructor.name),n==="Map"||n==="Set")return Array.from(e);if(n==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return a$(e,t)}}function a$(e,t){(t==null||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}function Nfe(e,t){var n=e==null?null:typeof Symbol<"u"&&e[Symbol.iterator]||e["@@iterator"];if(n!=null){var r,i,a,s,u=[],c=!0,d=!1;try{if(a=(n=n.call(e)).next,t!==0)for(;!(c=(r=a.call(n)).done)&&(u.push(r.value),u.length!==t);c=!0);}catch(p){d=!0,i=p}finally{try{if(!c&&n.return!=null&&(s=n.return(),Object(s)!==s))return}finally{if(d)throw i}}return u}}function Dfe(e){if(Array.isArray(e))return e}var Tm=Math.PI/180,kfe=function(t){return t*180/Math.PI},At=function(t,n,r,i){return{x:t+Math.cos(-Tm*i)*r,y:n+Math.sin(-Tm*i)*r}},$fe=function(t,n){var r=arguments.length>2&&arguments[2]!==void 0?arguments[2]:{top:0,right:0,bottom:0,left:0};return Math.min(Math.abs(t-(r.left||0)-(r.right||0)),Math.abs(n-(r.top||0)-(r.bottom||0)))/2},Ife=function(t,n,r,i,a){var s=t.width,u=t.height,c=t.startAngle,d=t.endAngle,p=hi(t.cx,s,s/2),m=hi(t.cy,u,u/2),v=$fe(s,u,r),b=hi(t.innerRadius,v,0),S=hi(t.outerRadius,v,v*.8),w=Object.keys(n);return w.reduce(function(x,C){var O=n[C],A=O.domain,T=O.reversed,R;if(vt(O.range))i==="angleAxis"?R=[c,d]:i==="radiusAxis"&&(R=[b,S]),T&&(R=[R[1],R[0]]);else{R=O.range;var M=R,P=Mfe(M,2);c=P[0],d=P[1]}var N=gfe(O,a),k=N.realScaleType,H=N.scale;H.domain(A).range(R),yfe(H);var L=Efe(H,Ji(Ji({},O),{},{realScaleType:k})),F=Ji(Ji(Ji({},O),L),{},{range:R,radius:S,realScaleType:k,scale:H,cx:p,cy:m,innerRadius:b,outerRadius:S,startAngle:c,endAngle:d});return Ji(Ji({},x),{},kV({},C,F))},{})},Lfe=function(t,n){var r=t.x,i=t.y,a=n.x,s=n.y;return Math.sqrt(Math.pow(r-a,2)+Math.pow(i-s,2))},Bfe=function(t,n){var r=t.x,i=t.y,a=n.cx,s=n.cy,u=Lfe({x:r,y:i},{x:a,y:s});if(u<=0)return{radius:u};var c=(r-a)/u,d=Math.acos(c);return i>s&&(d=2*Math.PI-d),{radius:u,angle:kfe(d),angleInRadian:d}},Ffe=function(t){var n=t.startAngle,r=t.endAngle,i=Math.floor(n/360),a=Math.floor(r/360),s=Math.min(i,a);return{startAngle:n-s*360,endAngle:r-s*360}},qfe=function(t,n){var r=n.startAngle,i=n.endAngle,a=Math.floor(r/360),s=Math.floor(i/360),u=Math.min(a,s);return t+u*360},o$=function(t,n){var r=t.x,i=t.y,a=Bfe({x:r,y:i},n),s=a.radius,u=a.angle,c=n.innerRadius,d=n.outerRadius;if(s<c||s>d)return!1;if(s===0)return!0;var p=Ffe(n),m=p.startAngle,v=p.endAngle,b=u,S;if(m<=v){for(;b>v;)b-=360;for(;b<m;)b+=360;S=b>=m&&b<=v}else{for(;b>m;)b-=360;for(;b<v;)b+=360;S=b>=v&&b<=m}return S?Ji(Ji({},n),{},{radius:s,angle:qfe(b,n)}):null},$V=function(t){return!g.isValidElement(t)&&!st(t)&&typeof t!="boolean"?t.className:""};function Bf(e){"@babel/helpers - typeof";return Bf=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},Bf(e)}var zfe=["offset"];function Ufe(e){return Wfe(e)||Gfe(e)||Hfe(e)||Vfe()}function Vfe(){throw new TypeError(`Invalid attempt to spread non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function Hfe(e,t){if(e){if(typeof e=="string")return z1(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);if(n==="Object"&&e.constructor&&(n=e.constructor.name),n==="Map"||n==="Set")return Array.from(e);if(n==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return z1(e,t)}}function Gfe(e){if(typeof Symbol<"u"&&e[Symbol.iterator]!=null||e["@@iterator"]!=null)return Array.from(e)}function Wfe(e){if(Array.isArray(e))return z1(e)}function z1(e,t){(t==null||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}function Kfe(e,t){if(e==null)return{};var n=Yfe(e,t),r,i;if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);for(i=0;i<a.length;i++)r=a[i],!(t.indexOf(r)>=0)&&Object.prototype.propertyIsEnumerable.call(e,r)&&(n[r]=e[r])}return n}function Yfe(e,t){if(e==null)return{};var n={};for(var r in e)if(Object.prototype.hasOwnProperty.call(e,r)){if(t.indexOf(r)>=0)continue;n[r]=e[r]}return n}function s$(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function an(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?s$(Object(n),!0).forEach(function(r){Zfe(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):s$(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function Zfe(e,t,n){return t=Xfe(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function Xfe(e){var t=Qfe(e,"string");return Bf(t)=="symbol"?t:t+""}function Qfe(e,t){if(Bf(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(Bf(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function Ff(){return Ff=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)Object.prototype.hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},Ff.apply(this,arguments)}var Jfe=function(t){var n=t.value,r=t.formatter,i=vt(t.children)?n:t.children;return st(r)?r(i):i},ede=function(t,n){var r=to(n-t),i=Math.min(Math.abs(n-t),360);return r*i},tde=function(t,n,r){var i=t.position,a=t.viewBox,s=t.offset,u=t.className,c=a,d=c.cx,p=c.cy,m=c.innerRadius,v=c.outerRadius,b=c.startAngle,S=c.endAngle,w=c.clockWise,x=(m+v)/2,C=ede(b,S),O=C>=0?1:-1,A,T;i==="insideStart"?(A=b+O*s,T=w):i==="insideEnd"?(A=S-O*s,T=!w):i==="end"&&(A=S+O*s,T=w),T=C<=0?T:!T;var R=At(d,p,x,A),M=At(d,p,x,A+(T?1:-1)*359),P="M".concat(R.x,",").concat(R.y,`
    A`).concat(x,",").concat(x,",0,1,").concat(T?0:1,`,
    `).concat(M.x,",").concat(M.y),N=vt(t.id)?LC("recharts-radial-line-"):t.id;return V.createElement("text",Ff({},r,{dominantBaseline:"central",className:ct("recharts-radial-bar-label",u)}),V.createElement("defs",null,V.createElement("path",{id:N,d:P})),V.createElement("textPath",{xlinkHref:"#".concat(N)},n))},nde=function(t){var n=t.viewBox,r=t.offset,i=t.position,a=n,s=a.cx,u=a.cy,c=a.innerRadius,d=a.outerRadius,p=a.startAngle,m=a.endAngle,v=(p+m)/2;if(i==="outside"){var b=At(s,u,d+r,v),S=b.x,w=b.y;return{x:S,y:w,textAnchor:S>=s?"start":"end",verticalAnchor:"middle"}}if(i==="center")return{x:s,y:u,textAnchor:"middle",verticalAnchor:"middle"};if(i==="centerTop")return{x:s,y:u,textAnchor:"middle",verticalAnchor:"start"};if(i==="centerBottom")return{x:s,y:u,textAnchor:"middle",verticalAnchor:"end"};var x=(c+d)/2,C=At(s,u,x,v),O=C.x,A=C.y;return{x:O,y:A,textAnchor:"middle",verticalAnchor:"middle"}},rde=function(t){var n=t.viewBox,r=t.parentViewBox,i=t.offset,a=t.position,s=n,u=s.x,c=s.y,d=s.width,p=s.height,m=p>=0?1:-1,v=m*i,b=m>0?"end":"start",S=m>0?"start":"end",w=d>=0?1:-1,x=w*i,C=w>0?"end":"start",O=w>0?"start":"end";if(a==="top"){var A={x:u+d/2,y:c-m*i,textAnchor:"middle",verticalAnchor:b};return an(an({},A),r?{height:Math.max(c-r.y,0),width:d}:{})}if(a==="bottom"){var T={x:u+d/2,y:c+p+v,textAnchor:"middle",verticalAnchor:S};return an(an({},T),r?{height:Math.max(r.y+r.height-(c+p),0),width:d}:{})}if(a==="left"){var R={x:u-x,y:c+p/2,textAnchor:C,verticalAnchor:"middle"};return an(an({},R),r?{width:Math.max(R.x-r.x,0),height:p}:{})}if(a==="right"){var M={x:u+d+x,y:c+p/2,textAnchor:O,verticalAnchor:"middle"};return an(an({},M),r?{width:Math.max(r.x+r.width-M.x,0),height:p}:{})}var P=r?{width:d,height:p}:{};return a==="insideLeft"?an({x:u+x,y:c+p/2,textAnchor:O,verticalAnchor:"middle"},P):a==="insideRight"?an({x:u+d-x,y:c+p/2,textAnchor:C,verticalAnchor:"middle"},P):a==="insideTop"?an({x:u+d/2,y:c+v,textAnchor:"middle",verticalAnchor:S},P):a==="insideBottom"?an({x:u+d/2,y:c+p-v,textAnchor:"middle",verticalAnchor:b},P):a==="insideTopLeft"?an({x:u+x,y:c+v,textAnchor:O,verticalAnchor:S},P):a==="insideTopRight"?an({x:u+d-x,y:c+v,textAnchor:C,verticalAnchor:S},P):a==="insideBottomLeft"?an({x:u+x,y:c+p-v,textAnchor:O,verticalAnchor:b},P):a==="insideBottomRight"?an({x:u+d-x,y:c+p-v,textAnchor:C,verticalAnchor:b},P):yu(a)&&($e(a.x)||Ko(a.x))&&($e(a.y)||Ko(a.y))?an({x:u+hi(a.x,d),y:c+hi(a.y,p),textAnchor:"end",verticalAnchor:"end"},P):an({x:u+d/2,y:c+p/2,textAnchor:"middle",verticalAnchor:"middle"},P)},ide=function(t){return"cx"in t&&$e(t.cx)};function Tn(e){var t=e.offset,n=t===void 0?5:t,r=Kfe(e,zfe),i=an({offset:n},r),a=i.viewBox,s=i.position,u=i.value,c=i.children,d=i.content,p=i.className,m=p===void 0?"":p,v=i.textBreakAll;if(!a||vt(u)&&vt(c)&&!g.isValidElement(d)&&!st(d))return null;if(g.isValidElement(d))return g.cloneElement(d,i);var b;if(st(d)){if(b=g.createElement(d,i),g.isValidElement(b))return b}else b=Jfe(i);var S=ide(a),w=Je(i,!0);if(S&&(s==="insideStart"||s==="insideEnd"||s==="end"))return tde(i,b,w);var x=S?nde(i):rde(i);return V.createElement(Rf,Ff({className:ct("recharts-label",m)},w,x,{breakAll:v}),b)}Tn.displayName="Label";var IV=function(t){var n=t.cx,r=t.cy,i=t.angle,a=t.startAngle,s=t.endAngle,u=t.r,c=t.radius,d=t.innerRadius,p=t.outerRadius,m=t.x,v=t.y,b=t.top,S=t.left,w=t.width,x=t.height,C=t.clockWise,O=t.labelViewBox;if(O)return O;if($e(w)&&$e(x)){if($e(m)&&$e(v))return{x:m,y:v,width:w,height:x};if($e(b)&&$e(S))return{x:b,y:S,width:w,height:x}}return $e(m)&&$e(v)?{x:m,y:v,width:0,height:0}:$e(n)&&$e(r)?{cx:n,cy:r,startAngle:a||i||0,endAngle:s||i||0,innerRadius:d||0,outerRadius:p||c||u||0,clockWise:C}:t.viewBox?t.viewBox:{}},ade=function(t,n){return t?t===!0?V.createElement(Tn,{key:"label-implicit",viewBox:n}):sn(t)?V.createElement(Tn,{key:"label-implicit",viewBox:n,value:t}):g.isValidElement(t)?t.type===Tn?g.cloneElement(t,{key:"label-implicit",viewBox:n}):V.createElement(Tn,{key:"label-implicit",content:t,viewBox:n}):st(t)?V.createElement(Tn,{key:"label-implicit",content:t,viewBox:n}):yu(t)?V.createElement(Tn,Ff({viewBox:n},t,{key:"label-implicit"})):null:null},ode=function(t,n){var r=arguments.length>2&&arguments[2]!==void 0?arguments[2]:!0;if(!t||!t.children&&r&&!t.label)return null;var i=t.children,a=IV(t),s=la(i,Tn).map(function(c,d){return g.cloneElement(c,{viewBox:n||a,key:"label-".concat(d)})});if(!r)return s;var u=ade(t.label,n||a);return[u].concat(Ufe(s))};Tn.parseViewBox=IV;Tn.renderCallByParent=ode;var DS,l$;function sde(){if(l$)return DS;l$=1;function e(t){var n=t==null?0:t.length;return n?t[n-1]:void 0}return DS=e,DS}var lde=sde();const LV=Rt(lde);function qf(e){"@babel/helpers - typeof";return qf=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},qf(e)}var ude=["valueAccessor"],cde=["data","dataKey","clockWise","id","textBreakAll"];function fde(e){return mde(e)||hde(e)||pde(e)||dde()}function dde(){throw new TypeError(`Invalid attempt to spread non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function pde(e,t){if(e){if(typeof e=="string")return U1(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);if(n==="Object"&&e.constructor&&(n=e.constructor.name),n==="Map"||n==="Set")return Array.from(e);if(n==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return U1(e,t)}}function hde(e){if(typeof Symbol<"u"&&e[Symbol.iterator]!=null||e["@@iterator"]!=null)return Array.from(e)}function mde(e){if(Array.isArray(e))return U1(e)}function U1(e,t){(t==null||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}function Mm(){return Mm=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)Object.prototype.hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},Mm.apply(this,arguments)}function u$(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function c$(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?u$(Object(n),!0).forEach(function(r){vde(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):u$(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function vde(e,t,n){return t=gde(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function gde(e){var t=yde(e,"string");return qf(t)=="symbol"?t:t+""}function yde(e,t){if(qf(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(qf(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function f$(e,t){if(e==null)return{};var n=bde(e,t),r,i;if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);for(i=0;i<a.length;i++)r=a[i],!(t.indexOf(r)>=0)&&Object.prototype.propertyIsEnumerable.call(e,r)&&(n[r]=e[r])}return n}function bde(e,t){if(e==null)return{};var n={};for(var r in e)if(Object.prototype.hasOwnProperty.call(e,r)){if(t.indexOf(r)>=0)continue;n[r]=e[r]}return n}var xde=function(t){return Array.isArray(t.value)?LV(t.value):t.value};function rs(e){var t=e.valueAccessor,n=t===void 0?xde:t,r=f$(e,ude),i=r.data,a=r.dataKey,s=r.clockWise,u=r.id,c=r.textBreakAll,d=f$(r,cde);return!i||!i.length?null:V.createElement(un,{className:"recharts-label-list"},i.map(function(p,m){var v=vt(a)?n(p,m):ga(p&&p.payload,a),b=vt(u)?{}:{id:"".concat(u,"-").concat(m)};return V.createElement(Tn,Mm({},Je(p,!0),d,b,{parentViewBox:p.parentViewBox,value:v,textBreakAll:c,viewBox:Tn.parseViewBox(vt(s)?p:c$(c$({},p),{},{clockWise:s})),key:"label-".concat(m),index:m}))}))}rs.displayName="LabelList";function wde(e,t){return e?e===!0?V.createElement(rs,{key:"labelList-implicit",data:t}):V.isValidElement(e)||st(e)?V.createElement(rs,{key:"labelList-implicit",data:t,content:e}):yu(e)?V.createElement(rs,Mm({data:t},e,{key:"labelList-implicit"})):null:null}function Sde(e,t){var n=arguments.length>2&&arguments[2]!==void 0?arguments[2]:!0;if(!e||!e.children&&n&&!e.label)return null;var r=e.children,i=la(r,rs).map(function(s,u){return g.cloneElement(s,{data:t,key:"labelList-".concat(u)})});if(!n)return i;var a=wde(e.label,t);return[a].concat(fde(i))}rs.renderCallByParent=Sde;function zf(e){"@babel/helpers - typeof";return zf=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},zf(e)}function V1(){return V1=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)Object.prototype.hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},V1.apply(this,arguments)}function d$(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function p$(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?d$(Object(n),!0).forEach(function(r){_de(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):d$(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function _de(e,t,n){return t=Ede(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function Ede(e){var t=Cde(e,"string");return zf(t)=="symbol"?t:t+""}function Cde(e,t){if(zf(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(zf(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}var Ode=function(t,n){var r=to(n-t),i=Math.min(Math.abs(n-t),359.999);return r*i},fh=function(t){var n=t.cx,r=t.cy,i=t.radius,a=t.angle,s=t.sign,u=t.isExternal,c=t.cornerRadius,d=t.cornerIsExternal,p=c*(u?1:-1)+i,m=Math.asin(c/p)/Tm,v=d?a:a+s*m,b=At(n,r,p,v),S=At(n,r,i,v),w=d?a-s*m:a,x=At(n,r,p*Math.cos(m*Tm),w);return{center:b,circleTangency:S,lineTangency:x,theta:m}},BV=function(t){var n=t.cx,r=t.cy,i=t.innerRadius,a=t.outerRadius,s=t.startAngle,u=t.endAngle,c=Ode(s,u),d=s+c,p=At(n,r,a,s),m=At(n,r,a,d),v="M ".concat(p.x,",").concat(p.y,`
    A `).concat(a,",").concat(a,`,0,
    `).concat(+(Math.abs(c)>180),",").concat(+(s>d),`,
    `).concat(m.x,",").concat(m.y,`
  `);if(i>0){var b=At(n,r,i,s),S=At(n,r,i,d);v+="L ".concat(S.x,",").concat(S.y,`
            A `).concat(i,",").concat(i,`,0,
            `).concat(+(Math.abs(c)>180),",").concat(+(s<=d),`,
            `).concat(b.x,",").concat(b.y," Z")}else v+="L ".concat(n,",").concat(r," Z");return v},Ade=function(t){var n=t.cx,r=t.cy,i=t.innerRadius,a=t.outerRadius,s=t.cornerRadius,u=t.forceCornerRadius,c=t.cornerIsExternal,d=t.startAngle,p=t.endAngle,m=to(p-d),v=fh({cx:n,cy:r,radius:a,angle:d,sign:m,cornerRadius:s,cornerIsExternal:c}),b=v.circleTangency,S=v.lineTangency,w=v.theta,x=fh({cx:n,cy:r,radius:a,angle:p,sign:-m,cornerRadius:s,cornerIsExternal:c}),C=x.circleTangency,O=x.lineTangency,A=x.theta,T=c?Math.abs(d-p):Math.abs(d-p)-w-A;if(T<0)return u?"M ".concat(S.x,",").concat(S.y,`
        a`).concat(s,",").concat(s,",0,0,1,").concat(s*2,`,0
        a`).concat(s,",").concat(s,",0,0,1,").concat(-s*2,`,0
      `):BV({cx:n,cy:r,innerRadius:i,outerRadius:a,startAngle:d,endAngle:p});var R="M ".concat(S.x,",").concat(S.y,`
    A`).concat(s,",").concat(s,",0,0,").concat(+(m<0),",").concat(b.x,",").concat(b.y,`
    A`).concat(a,",").concat(a,",0,").concat(+(T>180),",").concat(+(m<0),",").concat(C.x,",").concat(C.y,`
    A`).concat(s,",").concat(s,",0,0,").concat(+(m<0),",").concat(O.x,",").concat(O.y,`
  `);if(i>0){var M=fh({cx:n,cy:r,radius:i,angle:d,sign:m,isExternal:!0,cornerRadius:s,cornerIsExternal:c}),P=M.circleTangency,N=M.lineTangency,k=M.theta,H=fh({cx:n,cy:r,radius:i,angle:p,sign:-m,isExternal:!0,cornerRadius:s,cornerIsExternal:c}),L=H.circleTangency,F=H.lineTangency,z=H.theta,Z=c?Math.abs(d-p):Math.abs(d-p)-k-z;if(Z<0&&s===0)return"".concat(R,"L").concat(n,",").concat(r,"Z");R+="L".concat(F.x,",").concat(F.y,`
      A`).concat(s,",").concat(s,",0,0,").concat(+(m<0),",").concat(L.x,",").concat(L.y,`
      A`).concat(i,",").concat(i,",0,").concat(+(Z>180),",").concat(+(m>0),",").concat(P.x,",").concat(P.y,`
      A`).concat(s,",").concat(s,",0,0,").concat(+(m<0),",").concat(N.x,",").concat(N.y,"Z")}else R+="L".concat(n,",").concat(r,"Z");return R},Rde={cx:0,cy:0,innerRadius:0,outerRadius:0,startAngle:0,endAngle:0,cornerRadius:0,forceCornerRadius:!1,cornerIsExternal:!1},Tde=function(t){var n=p$(p$({},Rde),t),r=n.cx,i=n.cy,a=n.innerRadius,s=n.outerRadius,u=n.cornerRadius,c=n.forceCornerRadius,d=n.cornerIsExternal,p=n.startAngle,m=n.endAngle,v=n.className;if(s<a||p===m)return null;var b=ct("recharts-sector",v),S=s-a,w=hi(u,S,0,!0),x;return w>0&&Math.abs(p-m)<360?x=Ade({cx:r,cy:i,innerRadius:a,outerRadius:s,cornerRadius:Math.min(w,S/2),forceCornerRadius:c,cornerIsExternal:d,startAngle:p,endAngle:m}):x=BV({cx:r,cy:i,innerRadius:a,outerRadius:s,startAngle:p,endAngle:m}),V.createElement("path",V1({},Je(n,!0),{className:b,d:x,role:"img"}))};function Uf(e){"@babel/helpers - typeof";return Uf=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},Uf(e)}function H1(){return H1=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)Object.prototype.hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},H1.apply(this,arguments)}function h$(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function m$(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?h$(Object(n),!0).forEach(function(r){Mde(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):h$(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function Mde(e,t,n){return t=Pde(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function Pde(e){var t=jde(e,"string");return Uf(t)=="symbol"?t:t+""}function jde(e,t){if(Uf(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(Uf(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}var v$={curveBasisClosed:Yre,curveBasisOpen:Zre,curveBasis:Kre,curveBumpX:Dre,curveBumpY:kre,curveLinearClosed:Xre,curveLinear:Lv,curveMonotoneX:Qre,curveMonotoneY:Jre,curveNatural:eie,curveStep:tie,curveStepAfter:rie,curveStepBefore:nie},dh=function(t){return t.x===+t.x&&t.y===+t.y},Dc=function(t){return t.x},kc=function(t){return t.y},Nde=function(t,n){if(st(t))return t;var r="curve".concat($v(t));return(r==="curveMonotone"||r==="curveBump")&&n?v$["".concat(r).concat(n==="vertical"?"Y":"X")]:v$[r]||Lv},Dde=function(t){var n=t.type,r=n===void 0?"linear":n,i=t.points,a=i===void 0?[]:i,s=t.baseLine,u=t.layout,c=t.connectNulls,d=c===void 0?!1:c,p=Nde(r,u),m=d?a.filter(function(w){return dh(w)}):a,v;if(Array.isArray(s)){var b=d?s.filter(function(w){return dh(w)}):s,S=m.map(function(w,x){return m$(m$({},w),{},{base:b[x]})});return u==="vertical"?v=rh().y(kc).x1(Dc).x0(function(w){return w.base.x}):v=rh().x(Dc).y1(kc).y0(function(w){return w.base.y}),v.defined(dh).curve(p),v(S)}return u==="vertical"&&$e(s)?v=rh().y(kc).x1(Dc).x0(s):$e(s)?v=rh().x(Dc).y1(kc).y0(s):v=zz().x(Dc).y(kc),v.defined(dh).curve(p),v(m)},g$=function(t){var n=t.className,r=t.points,i=t.path,a=t.pathRef;if((!r||!r.length)&&!i)return null;var s=r&&r.length?Dde(t):i;return V.createElement("path",H1({},Je(t,!1),am(t),{className:ct("recharts-curve",n),d:s,ref:a}))},kS={exports:{}},$S,y$;function kde(){if(y$)return $S;y$=1;var e="SECRET_DO_NOT_PASS_THIS_OR_YOU_WILL_BE_FIRED";return $S=e,$S}var IS,b$;function $de(){if(b$)return IS;b$=1;var e=kde();function t(){}function n(){}return n.resetWarningCache=t,IS=function(){function r(s,u,c,d,p,m){if(m!==e){var v=new Error("Calling PropTypes validators directly is not supported by the `prop-types` package. Use PropTypes.checkPropTypes() to call them. Read more at http://fb.me/use-check-prop-types");throw v.name="Invariant Violation",v}}r.isRequired=r;function i(){return r}var a={array:r,bigint:r,bool:r,func:r,number:r,object:r,string:r,symbol:r,any:r,arrayOf:i,element:r,elementType:r,instanceOf:i,node:r,objectOf:i,oneOf:i,oneOfType:i,shape:i,exact:i,checkPropTypes:n,resetWarningCache:t};return a.PropTypes=a,a},IS}var x$;function Ide(){return x$||(x$=1,kS.exports=$de()()),kS.exports}var Lde=Ide();const yt=Rt(Lde);var Bde=Object.getOwnPropertyNames,Fde=Object.getOwnPropertySymbols,qde=Object.prototype.hasOwnProperty;function w$(e,t){return function(r,i,a){return e(r,i,a)&&t(r,i,a)}}function ph(e){return function(n,r,i){if(!n||!r||typeof n!="object"||typeof r!="object")return e(n,r,i);var a=i.cache,s=a.get(n),u=a.get(r);if(s&&u)return s===r&&u===n;a.set(n,r),a.set(r,n);var c=e(n,r,i);return a.delete(n),a.delete(r),c}}function S$(e){return Bde(e).concat(Fde(e))}var zde=Object.hasOwn||function(e,t){return qde.call(e,t)};function xs(e,t){return e===t||!e&&!t&&e!==e&&t!==t}var Ude="__v",Vde="__o",Hde="_owner",_$=Object.getOwnPropertyDescriptor,E$=Object.keys;function Gde(e,t,n){var r=e.length;if(t.length!==r)return!1;for(;r-- >0;)if(!n.equals(e[r],t[r],r,r,e,t,n))return!1;return!0}function Wde(e,t){return xs(e.getTime(),t.getTime())}function Kde(e,t){return e.name===t.name&&e.message===t.message&&e.cause===t.cause&&e.stack===t.stack}function Yde(e,t){return e===t}function C$(e,t,n){var r=e.size;if(r!==t.size)return!1;if(!r)return!0;for(var i=new Array(r),a=e.entries(),s,u,c=0;(s=a.next())&&!s.done;){for(var d=t.entries(),p=!1,m=0;(u=d.next())&&!u.done;){if(i[m]){m++;continue}var v=s.value,b=u.value;if(n.equals(v[0],b[0],c,m,e,t,n)&&n.equals(v[1],b[1],v[0],b[0],e,t,n)){p=i[m]=!0;break}m++}if(!p)return!1;c++}return!0}var Zde=xs;function Xde(e,t,n){var r=E$(e),i=r.length;if(E$(t).length!==i)return!1;for(;i-- >0;)if(!FV(e,t,n,r[i]))return!1;return!0}function $c(e,t,n){var r=S$(e),i=r.length;if(S$(t).length!==i)return!1;for(var a,s,u;i-- >0;)if(a=r[i],!FV(e,t,n,a)||(s=_$(e,a),u=_$(t,a),(s||u)&&(!s||!u||s.configurable!==u.configurable||s.enumerable!==u.enumerable||s.writable!==u.writable)))return!1;return!0}function Qde(e,t){return xs(e.valueOf(),t.valueOf())}function Jde(e,t){return e.source===t.source&&e.flags===t.flags}function O$(e,t,n){var r=e.size;if(r!==t.size)return!1;if(!r)return!0;for(var i=new Array(r),a=e.values(),s,u;(s=a.next())&&!s.done;){for(var c=t.values(),d=!1,p=0;(u=c.next())&&!u.done;){if(!i[p]&&n.equals(s.value,u.value,s.value,u.value,e,t,n)){d=i[p]=!0;break}p++}if(!d)return!1}return!0}function epe(e,t){var n=e.length;if(t.length!==n)return!1;for(;n-- >0;)if(e[n]!==t[n])return!1;return!0}function tpe(e,t){return e.hostname===t.hostname&&e.pathname===t.pathname&&e.protocol===t.protocol&&e.port===t.port&&e.hash===t.hash&&e.username===t.username&&e.password===t.password}function FV(e,t,n,r){return(r===Hde||r===Vde||r===Ude)&&(e.$$typeof||t.$$typeof)?!0:zde(t,r)&&n.equals(e[r],t[r],r,r,e,t,n)}var npe="[object Arguments]",rpe="[object Boolean]",ipe="[object Date]",ape="[object Error]",ope="[object Map]",spe="[object Number]",lpe="[object Object]",upe="[object RegExp]",cpe="[object Set]",fpe="[object String]",dpe="[object URL]",ppe=Array.isArray,A$=typeof ArrayBuffer=="function"&&ArrayBuffer.isView?ArrayBuffer.isView:null,R$=Object.assign,hpe=Object.prototype.toString.call.bind(Object.prototype.toString);function mpe(e){var t=e.areArraysEqual,n=e.areDatesEqual,r=e.areErrorsEqual,i=e.areFunctionsEqual,a=e.areMapsEqual,s=e.areNumbersEqual,u=e.areObjectsEqual,c=e.arePrimitiveWrappersEqual,d=e.areRegExpsEqual,p=e.areSetsEqual,m=e.areTypedArraysEqual,v=e.areUrlsEqual;return function(S,w,x){if(S===w)return!0;if(S==null||w==null)return!1;var C=typeof S;if(C!==typeof w)return!1;if(C!=="object")return C==="number"?s(S,w,x):C==="function"?i(S,w,x):!1;var O=S.constructor;if(O!==w.constructor)return!1;if(O===Object)return u(S,w,x);if(ppe(S))return t(S,w,x);if(A$!=null&&A$(S))return m(S,w,x);if(O===Date)return n(S,w,x);if(O===RegExp)return d(S,w,x);if(O===Map)return a(S,w,x);if(O===Set)return p(S,w,x);var A=hpe(S);return A===ipe?n(S,w,x):A===upe?d(S,w,x):A===ope?a(S,w,x):A===cpe?p(S,w,x):A===lpe?typeof S.then!="function"&&typeof w.then!="function"&&u(S,w,x):A===dpe?v(S,w,x):A===ape?r(S,w,x):A===npe?u(S,w,x):A===rpe||A===spe||A===fpe?c(S,w,x):!1}}function vpe(e){var t=e.circular,n=e.createCustomConfig,r=e.strict,i={areArraysEqual:r?$c:Gde,areDatesEqual:Wde,areErrorsEqual:Kde,areFunctionsEqual:Yde,areMapsEqual:r?w$(C$,$c):C$,areNumbersEqual:Zde,areObjectsEqual:r?$c:Xde,arePrimitiveWrappersEqual:Qde,areRegExpsEqual:Jde,areSetsEqual:r?w$(O$,$c):O$,areTypedArraysEqual:r?$c:epe,areUrlsEqual:tpe};if(n&&(i=R$({},i,n(i))),t){var a=ph(i.areArraysEqual),s=ph(i.areMapsEqual),u=ph(i.areObjectsEqual),c=ph(i.areSetsEqual);i=R$({},i,{areArraysEqual:a,areMapsEqual:s,areObjectsEqual:u,areSetsEqual:c})}return i}function gpe(e){return function(t,n,r,i,a,s,u){return e(t,n,u)}}function ype(e){var t=e.circular,n=e.comparator,r=e.createState,i=e.equals,a=e.strict;if(r)return function(c,d){var p=r(),m=p.cache,v=m===void 0?t?new WeakMap:void 0:m,b=p.meta;return n(c,d,{cache:v,equals:i,meta:b,strict:a})};if(t)return function(c,d){return n(c,d,{cache:new WeakMap,equals:i,meta:void 0,strict:a})};var s={cache:void 0,equals:i,meta:void 0,strict:a};return function(c,d){return n(c,d,s)}}var bpe=yo();yo({strict:!0});yo({circular:!0});yo({circular:!0,strict:!0});yo({createInternalComparator:function(){return xs}});yo({strict:!0,createInternalComparator:function(){return xs}});yo({circular:!0,createInternalComparator:function(){return xs}});yo({circular:!0,createInternalComparator:function(){return xs},strict:!0});function yo(e){e===void 0&&(e={});var t=e.circular,n=t===void 0?!1:t,r=e.createInternalComparator,i=e.createState,a=e.strict,s=a===void 0?!1:a,u=vpe(e),c=mpe(u),d=r?r(c):gpe(c);return ype({circular:n,comparator:c,createState:i,equals:d,strict:s})}function xpe(e){typeof requestAnimationFrame<"u"&&requestAnimationFrame(e)}function T$(e){var t=arguments.length>1&&arguments[1]!==void 0?arguments[1]:0,n=-1,r=function i(a){n<0&&(n=a),a-n>t?(e(a),n=-1):xpe(i)};requestAnimationFrame(r)}function G1(e){"@babel/helpers - typeof";return G1=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},G1(e)}function wpe(e){return Cpe(e)||Epe(e)||_pe(e)||Spe()}function Spe(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function _pe(e,t){if(e){if(typeof e=="string")return M$(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);if(n==="Object"&&e.constructor&&(n=e.constructor.name),n==="Map"||n==="Set")return Array.from(e);if(n==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return M$(e,t)}}function M$(e,t){(t==null||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}function Epe(e){if(typeof Symbol<"u"&&e[Symbol.iterator]!=null||e["@@iterator"]!=null)return Array.from(e)}function Cpe(e){if(Array.isArray(e))return e}function Ope(){var e={},t=function(){return null},n=!1,r=function i(a){if(!n){if(Array.isArray(a)){if(!a.length)return;var s=a,u=wpe(s),c=u[0],d=u.slice(1);if(typeof c=="number"){T$(i.bind(null,d),c);return}i(c),T$(i.bind(null,d));return}G1(a)==="object"&&(e=a,t(e)),typeof a=="function"&&a()}};return{stop:function(){n=!0},start:function(a){n=!1,r(a)},subscribe:function(a){return t=a,function(){t=function(){return null}}}}}function Vf(e){"@babel/helpers - typeof";return Vf=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},Vf(e)}function P$(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function j$(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?P$(Object(n),!0).forEach(function(r){qV(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):P$(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function qV(e,t,n){return t=Ape(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function Ape(e){var t=Rpe(e,"string");return Vf(t)==="symbol"?t:String(t)}function Rpe(e,t){if(Vf(e)!=="object"||e===null)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(Vf(r)!=="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}var Tpe=function(t,n){return[Object.keys(t),Object.keys(n)].reduce(function(r,i){return r.filter(function(a){return i.includes(a)})})},Mpe=function(t){return t},Ppe=function(t){return t.replace(/([A-Z])/g,function(n){return"-".concat(n.toLowerCase())})},Zc=function(t,n){return Object.keys(n).reduce(function(r,i){return j$(j$({},r),{},qV({},i,t(i,n[i])))},{})},N$=function(t,n,r){return t.map(function(i){return"".concat(Ppe(i)," ").concat(n,"ms ").concat(r)}).join(",")};function jpe(e,t){return kpe(e)||Dpe(e,t)||zV(e,t)||Npe()}function Npe(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function Dpe(e,t){var n=e==null?null:typeof Symbol<"u"&&e[Symbol.iterator]||e["@@iterator"];if(n!=null){var r,i,a,s,u=[],c=!0,d=!1;try{if(a=(n=n.call(e)).next,t!==0)for(;!(c=(r=a.call(n)).done)&&(u.push(r.value),u.length!==t);c=!0);}catch(p){d=!0,i=p}finally{try{if(!c&&n.return!=null&&(s=n.return(),Object(s)!==s))return}finally{if(d)throw i}}return u}}function kpe(e){if(Array.isArray(e))return e}function $pe(e){return Bpe(e)||Lpe(e)||zV(e)||Ipe()}function Ipe(){throw new TypeError(`Invalid attempt to spread non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function zV(e,t){if(e){if(typeof e=="string")return W1(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);if(n==="Object"&&e.constructor&&(n=e.constructor.name),n==="Map"||n==="Set")return Array.from(e);if(n==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return W1(e,t)}}function Lpe(e){if(typeof Symbol<"u"&&e[Symbol.iterator]!=null||e["@@iterator"]!=null)return Array.from(e)}function Bpe(e){if(Array.isArray(e))return W1(e)}function W1(e,t){(t==null||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}var Pm=1e-4,UV=function(t,n){return[0,3*t,3*n-6*t,3*t-3*n+1]},VV=function(t,n){return t.map(function(r,i){return r*Math.pow(n,i)}).reduce(function(r,i){return r+i})},D$=function(t,n){return function(r){var i=UV(t,n);return VV(i,r)}},Fpe=function(t,n){return function(r){var i=UV(t,n),a=[].concat($pe(i.map(function(s,u){return s*u}).slice(1)),[0]);return VV(a,r)}},k$=function(){for(var t=arguments.length,n=new Array(t),r=0;r<t;r++)n[r]=arguments[r];var i=n[0],a=n[1],s=n[2],u=n[3];if(n.length===1)switch(n[0]){case"linear":i=0,a=0,s=1,u=1;break;case"ease":i=.25,a=.1,s=.25,u=1;break;case"ease-in":i=.42,a=0,s=1,u=1;break;case"ease-out":i=.42,a=0,s=.58,u=1;break;case"ease-in-out":i=0,a=0,s=.58,u=1;break;default:{var c=n[0].split("(");if(c[0]==="cubic-bezier"&&c[1].split(")")[0].split(",").length===4){var d=c[1].split(")")[0].split(",").map(function(x){return parseFloat(x)}),p=jpe(d,4);i=p[0],a=p[1],s=p[2],u=p[3]}}}var m=D$(i,s),v=D$(a,u),b=Fpe(i,s),S=function(C){return C>1?1:C<0?0:C},w=function(C){for(var O=C>1?1:C,A=O,T=0;T<8;++T){var R=m(A)-O,M=b(A);if(Math.abs(R-O)<Pm||M<Pm)return v(A);A=S(A-R/M)}return v(A)};return w.isStepper=!1,w},qpe=function(){var t=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{},n=t.stiff,r=n===void 0?100:n,i=t.damping,a=i===void 0?8:i,s=t.dt,u=s===void 0?17:s,c=function(p,m,v){var b=-(p-m)*r,S=v*a,w=v+(b-S)*u/1e3,x=v*u/1e3+p;return Math.abs(x-m)<Pm&&Math.abs(w)<Pm?[m,0]:[x,w]};return c.isStepper=!0,c.dt=u,c},zpe=function(){for(var t=arguments.length,n=new Array(t),r=0;r<t;r++)n[r]=arguments[r];var i=n[0];if(typeof i=="string")switch(i){case"ease":case"ease-in-out":case"ease-out":case"ease-in":case"linear":return k$(i);case"spring":return qpe();default:if(i.split("(")[0]==="cubic-bezier")return k$(i)}return typeof i=="function"?i:null};function Hf(e){"@babel/helpers - typeof";return Hf=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},Hf(e)}function $$(e){return Hpe(e)||Vpe(e)||HV(e)||Upe()}function Upe(){throw new TypeError(`Invalid attempt to spread non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function Vpe(e){if(typeof Symbol<"u"&&e[Symbol.iterator]!=null||e["@@iterator"]!=null)return Array.from(e)}function Hpe(e){if(Array.isArray(e))return Y1(e)}function I$(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function En(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?I$(Object(n),!0).forEach(function(r){K1(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):I$(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function K1(e,t,n){return t=Gpe(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function Gpe(e){var t=Wpe(e,"string");return Hf(t)==="symbol"?t:String(t)}function Wpe(e,t){if(Hf(e)!=="object"||e===null)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(Hf(r)!=="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function Kpe(e,t){return Xpe(e)||Zpe(e,t)||HV(e,t)||Ype()}function Ype(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function HV(e,t){if(e){if(typeof e=="string")return Y1(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);if(n==="Object"&&e.constructor&&(n=e.constructor.name),n==="Map"||n==="Set")return Array.from(e);if(n==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return Y1(e,t)}}function Y1(e,t){(t==null||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}function Zpe(e,t){var n=e==null?null:typeof Symbol<"u"&&e[Symbol.iterator]||e["@@iterator"];if(n!=null){var r,i,a,s,u=[],c=!0,d=!1;try{if(a=(n=n.call(e)).next,t!==0)for(;!(c=(r=a.call(n)).done)&&(u.push(r.value),u.length!==t);c=!0);}catch(p){d=!0,i=p}finally{try{if(!c&&n.return!=null&&(s=n.return(),Object(s)!==s))return}finally{if(d)throw i}}return u}}function Xpe(e){if(Array.isArray(e))return e}var jm=function(t,n,r){return t+(n-t)*r},Z1=function(t){var n=t.from,r=t.to;return n!==r},Qpe=function e(t,n,r){var i=Zc(function(a,s){if(Z1(s)){var u=t(s.from,s.to,s.velocity),c=Kpe(u,2),d=c[0],p=c[1];return En(En({},s),{},{from:d,velocity:p})}return s},n);return r<1?Zc(function(a,s){return Z1(s)?En(En({},s),{},{velocity:jm(s.velocity,i[a].velocity,r),from:jm(s.from,i[a].from,r)}):s},n):e(t,i,r-1)};const Jpe=function(e,t,n,r,i){var a=Tpe(e,t),s=a.reduce(function(x,C){return En(En({},x),{},K1({},C,[e[C],t[C]]))},{}),u=a.reduce(function(x,C){return En(En({},x),{},K1({},C,{from:e[C],velocity:0,to:t[C]}))},{}),c=-1,d,p,m=function(){return null},v=function(){return Zc(function(C,O){return O.from},u)},b=function(){return!Object.values(u).filter(Z1).length},S=function(C){d||(d=C);var O=C-d,A=O/n.dt;u=Qpe(n,u,A),i(En(En(En({},e),t),v())),d=C,b()||(c=requestAnimationFrame(m))},w=function(C){p||(p=C);var O=(C-p)/r,A=Zc(function(R,M){return jm.apply(void 0,$$(M).concat([n(O)]))},s);if(i(En(En(En({},e),t),A)),O<1)c=requestAnimationFrame(m);else{var T=Zc(function(R,M){return jm.apply(void 0,$$(M).concat([n(1)]))},s);i(En(En(En({},e),t),T))}};return m=n.isStepper?S:w,function(){return requestAnimationFrame(m),function(){cancelAnimationFrame(c)}}};function Gl(e){"@babel/helpers - typeof";return Gl=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},Gl(e)}var ehe=["children","begin","duration","attributeName","easing","isActive","steps","from","to","canBegin","onAnimationEnd","shouldReAnimate","onAnimationReStart"];function the(e,t){if(e==null)return{};var n=nhe(e,t),r,i;if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);for(i=0;i<a.length;i++)r=a[i],!(t.indexOf(r)>=0)&&Object.prototype.propertyIsEnumerable.call(e,r)&&(n[r]=e[r])}return n}function nhe(e,t){if(e==null)return{};var n={},r=Object.keys(e),i,a;for(a=0;a<r.length;a++)i=r[a],!(t.indexOf(i)>=0)&&(n[i]=e[i]);return n}function LS(e){return ohe(e)||ahe(e)||ihe(e)||rhe()}function rhe(){throw new TypeError(`Invalid attempt to spread non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function ihe(e,t){if(e){if(typeof e=="string")return X1(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);if(n==="Object"&&e.constructor&&(n=e.constructor.name),n==="Map"||n==="Set")return Array.from(e);if(n==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return X1(e,t)}}function ahe(e){if(typeof Symbol<"u"&&e[Symbol.iterator]!=null||e["@@iterator"]!=null)return Array.from(e)}function ohe(e){if(Array.isArray(e))return X1(e)}function X1(e,t){(t==null||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}function L$(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function zr(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?L$(Object(n),!0).forEach(function(r){Uc(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):L$(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function Uc(e,t,n){return t=GV(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function she(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}function lhe(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,GV(r.key),r)}}function uhe(e,t,n){return t&&lhe(e.prototype,t),Object.defineProperty(e,"prototype",{writable:!1}),e}function GV(e){var t=che(e,"string");return Gl(t)==="symbol"?t:String(t)}function che(e,t){if(Gl(e)!=="object"||e===null)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(Gl(r)!=="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function fhe(e,t){if(typeof t!="function"&&t!==null)throw new TypeError("Super expression must either be null or a function");e.prototype=Object.create(t&&t.prototype,{constructor:{value:e,writable:!0,configurable:!0}}),Object.defineProperty(e,"prototype",{writable:!1}),t&&Q1(e,t)}function Q1(e,t){return Q1=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(r,i){return r.__proto__=i,r},Q1(e,t)}function dhe(e){var t=phe();return function(){var r=Nm(e),i;if(t){var a=Nm(this).constructor;i=Reflect.construct(r,arguments,a)}else i=r.apply(this,arguments);return J1(this,i)}}function J1(e,t){if(t&&(Gl(t)==="object"||typeof t=="function"))return t;if(t!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return eE(e)}function eE(e){if(e===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return e}function phe(){if(typeof Reflect>"u"||!Reflect.construct||Reflect.construct.sham)return!1;if(typeof Proxy=="function")return!0;try{return Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){})),!0}catch{return!1}}function Nm(e){return Nm=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(n){return n.__proto__||Object.getPrototypeOf(n)},Nm(e)}var Wl=function(e){fhe(n,e);var t=dhe(n);function n(r,i){var a;she(this,n),a=t.call(this,r,i);var s=a.props,u=s.isActive,c=s.attributeName,d=s.from,p=s.to,m=s.steps,v=s.children,b=s.duration;if(a.handleStyleChange=a.handleStyleChange.bind(eE(a)),a.changeStyle=a.changeStyle.bind(eE(a)),!u||b<=0)return a.state={style:{}},typeof v=="function"&&(a.state={style:p}),J1(a);if(m&&m.length)a.state={style:m[0].style};else if(d){if(typeof v=="function")return a.state={style:d},J1(a);a.state={style:c?Uc({},c,d):d}}else a.state={style:{}};return a}return uhe(n,[{key:"componentDidMount",value:function(){var i=this.props,a=i.isActive,s=i.canBegin;this.mounted=!0,!(!a||!s)&&this.runAnimation(this.props)}},{key:"componentDidUpdate",value:function(i){var a=this.props,s=a.isActive,u=a.canBegin,c=a.attributeName,d=a.shouldReAnimate,p=a.to,m=a.from,v=this.state.style;if(u){if(!s){var b={style:c?Uc({},c,p):p};this.state&&v&&(c&&v[c]!==p||!c&&v!==p)&&this.setState(b);return}if(!(bpe(i.to,p)&&i.canBegin&&i.isActive)){var S=!i.canBegin||!i.isActive;this.manager&&this.manager.stop(),this.stopJSAnimation&&this.stopJSAnimation();var w=S||d?m:i.to;if(this.state&&v){var x={style:c?Uc({},c,w):w};(c&&v[c]!==w||!c&&v!==w)&&this.setState(x)}this.runAnimation(zr(zr({},this.props),{},{from:w,begin:0}))}}}},{key:"componentWillUnmount",value:function(){this.mounted=!1;var i=this.props.onAnimationEnd;this.unSubscribe&&this.unSubscribe(),this.manager&&(this.manager.stop(),this.manager=null),this.stopJSAnimation&&this.stopJSAnimation(),i&&i()}},{key:"handleStyleChange",value:function(i){this.changeStyle(i)}},{key:"changeStyle",value:function(i){this.mounted&&this.setState({style:i})}},{key:"runJSAnimation",value:function(i){var a=this,s=i.from,u=i.to,c=i.duration,d=i.easing,p=i.begin,m=i.onAnimationEnd,v=i.onAnimationStart,b=Jpe(s,u,zpe(d),c,this.changeStyle),S=function(){a.stopJSAnimation=b()};this.manager.start([v,p,S,c,m])}},{key:"runStepAnimation",value:function(i){var a=this,s=i.steps,u=i.begin,c=i.onAnimationStart,d=s[0],p=d.style,m=d.duration,v=m===void 0?0:m,b=function(w,x,C){if(C===0)return w;var O=x.duration,A=x.easing,T=A===void 0?"ease":A,R=x.style,M=x.properties,P=x.onAnimationEnd,N=C>0?s[C-1]:x,k=M||Object.keys(R);if(typeof T=="function"||T==="spring")return[].concat(LS(w),[a.runJSAnimation.bind(a,{from:N.style,to:R,duration:O,easing:T}),O]);var H=N$(k,O,T),L=zr(zr(zr({},N.style),R),{},{transition:H});return[].concat(LS(w),[L,O,P]).filter(Mpe)};return this.manager.start([c].concat(LS(s.reduce(b,[p,Math.max(v,u)])),[i.onAnimationEnd]))}},{key:"runAnimation",value:function(i){this.manager||(this.manager=Ope());var a=i.begin,s=i.duration,u=i.attributeName,c=i.to,d=i.easing,p=i.onAnimationStart,m=i.onAnimationEnd,v=i.steps,b=i.children,S=this.manager;if(this.unSubscribe=S.subscribe(this.handleStyleChange),typeof d=="function"||typeof b=="function"||d==="spring"){this.runJSAnimation(i);return}if(v.length>1){this.runStepAnimation(i);return}var w=u?Uc({},u,c):c,x=N$(Object.keys(w),s,d);S.start([p,a,zr(zr({},w),{},{transition:x}),s,m])}},{key:"render",value:function(){var i=this.props,a=i.children;i.begin;var s=i.duration;i.attributeName,i.easing;var u=i.isActive;i.steps,i.from,i.to,i.canBegin,i.onAnimationEnd,i.shouldReAnimate,i.onAnimationReStart;var c=the(i,ehe),d=g.Children.count(a),p=this.state.style;if(typeof a=="function")return a(p);if(!u||d===0||s<=0)return a;var m=function(b){var S=b.props,w=S.style,x=w===void 0?{}:w,C=S.className,O=g.cloneElement(b,zr(zr({},c),{},{style:zr(zr({},x),p),className:C}));return O};return d===1?m(g.Children.only(a)):V.createElement("div",null,g.Children.map(a,function(v){return m(v)}))}}]),n}(g.PureComponent);Wl.displayName="Animate";Wl.defaultProps={begin:0,duration:1e3,from:"",to:"",attributeName:"",easing:"ease",isActive:!0,canBegin:!0,steps:[],onAnimationEnd:function(){},onAnimationStart:function(){}};Wl.propTypes={from:yt.oneOfType([yt.object,yt.string]),to:yt.oneOfType([yt.object,yt.string]),attributeName:yt.string,duration:yt.number,begin:yt.number,easing:yt.oneOfType([yt.string,yt.func]),steps:yt.arrayOf(yt.shape({duration:yt.number.isRequired,style:yt.object.isRequired,easing:yt.oneOfType([yt.oneOf(["ease","ease-in","ease-out","ease-in-out","linear"]),yt.func]),properties:yt.arrayOf("string"),onAnimationEnd:yt.func})),children:yt.oneOfType([yt.node,yt.func]),isActive:yt.bool,canBegin:yt.bool,onAnimationEnd:yt.func,shouldReAnimate:yt.bool,onAnimationStart:yt.func,onAnimationReStart:yt.func};function Gf(e){"@babel/helpers - typeof";return Gf=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},Gf(e)}function Dm(){return Dm=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)Object.prototype.hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},Dm.apply(this,arguments)}function hhe(e,t){return yhe(e)||ghe(e,t)||vhe(e,t)||mhe()}function mhe(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function vhe(e,t){if(e){if(typeof e=="string")return B$(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);if(n==="Object"&&e.constructor&&(n=e.constructor.name),n==="Map"||n==="Set")return Array.from(e);if(n==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return B$(e,t)}}function B$(e,t){(t==null||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}function ghe(e,t){var n=e==null?null:typeof Symbol<"u"&&e[Symbol.iterator]||e["@@iterator"];if(n!=null){var r,i,a,s,u=[],c=!0,d=!1;try{if(a=(n=n.call(e)).next,t!==0)for(;!(c=(r=a.call(n)).done)&&(u.push(r.value),u.length!==t);c=!0);}catch(p){d=!0,i=p}finally{try{if(!c&&n.return!=null&&(s=n.return(),Object(s)!==s))return}finally{if(d)throw i}}return u}}function yhe(e){if(Array.isArray(e))return e}function F$(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function q$(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?F$(Object(n),!0).forEach(function(r){bhe(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):F$(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function bhe(e,t,n){return t=xhe(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function xhe(e){var t=whe(e,"string");return Gf(t)=="symbol"?t:t+""}function whe(e,t){if(Gf(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(Gf(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}var z$=function(t,n,r,i,a){var s=Math.min(Math.abs(r)/2,Math.abs(i)/2),u=i>=0?1:-1,c=r>=0?1:-1,d=i>=0&&r>=0||i<0&&r<0?1:0,p;if(s>0&&a instanceof Array){for(var m=[0,0,0,0],v=0,b=4;v<b;v++)m[v]=a[v]>s?s:a[v];p="M".concat(t,",").concat(n+u*m[0]),m[0]>0&&(p+="A ".concat(m[0],",").concat(m[0],",0,0,").concat(d,",").concat(t+c*m[0],",").concat(n)),p+="L ".concat(t+r-c*m[1],",").concat(n),m[1]>0&&(p+="A ".concat(m[1],",").concat(m[1],",0,0,").concat(d,`,
        `).concat(t+r,",").concat(n+u*m[1])),p+="L ".concat(t+r,",").concat(n+i-u*m[2]),m[2]>0&&(p+="A ".concat(m[2],",").concat(m[2],",0,0,").concat(d,`,
        `).concat(t+r-c*m[2],",").concat(n+i)),p+="L ".concat(t+c*m[3],",").concat(n+i),m[3]>0&&(p+="A ".concat(m[3],",").concat(m[3],",0,0,").concat(d,`,
        `).concat(t,",").concat(n+i-u*m[3])),p+="Z"}else if(s>0&&a===+a&&a>0){var S=Math.min(s,a);p="M ".concat(t,",").concat(n+u*S,`
            A `).concat(S,",").concat(S,",0,0,").concat(d,",").concat(t+c*S,",").concat(n,`
            L `).concat(t+r-c*S,",").concat(n,`
            A `).concat(S,",").concat(S,",0,0,").concat(d,",").concat(t+r,",").concat(n+u*S,`
            L `).concat(t+r,",").concat(n+i-u*S,`
            A `).concat(S,",").concat(S,",0,0,").concat(d,",").concat(t+r-c*S,",").concat(n+i,`
            L `).concat(t+c*S,",").concat(n+i,`
            A `).concat(S,",").concat(S,",0,0,").concat(d,",").concat(t,",").concat(n+i-u*S," Z")}else p="M ".concat(t,",").concat(n," h ").concat(r," v ").concat(i," h ").concat(-r," Z");return p},She=function(t,n){if(!t||!n)return!1;var r=t.x,i=t.y,a=n.x,s=n.y,u=n.width,c=n.height;if(Math.abs(u)>0&&Math.abs(c)>0){var d=Math.min(a,a+u),p=Math.max(a,a+u),m=Math.min(s,s+c),v=Math.max(s,s+c);return r>=d&&r<=p&&i>=m&&i<=v}return!1},_he={x:0,y:0,width:0,height:0,radius:0,isAnimationActive:!1,isUpdateAnimationActive:!1,animationBegin:0,animationDuration:1500,animationEasing:"ease"},WV=function(t){var n=q$(q$({},_he),t),r=g.useRef(),i=g.useState(-1),a=hhe(i,2),s=a[0],u=a[1];g.useEffect(function(){if(r.current&&r.current.getTotalLength)try{var T=r.current.getTotalLength();T&&u(T)}catch{}},[]);var c=n.x,d=n.y,p=n.width,m=n.height,v=n.radius,b=n.className,S=n.animationEasing,w=n.animationDuration,x=n.animationBegin,C=n.isAnimationActive,O=n.isUpdateAnimationActive;if(c!==+c||d!==+d||p!==+p||m!==+m||p===0||m===0)return null;var A=ct("recharts-rectangle",b);return O?V.createElement(Wl,{canBegin:s>0,from:{width:p,height:m,x:c,y:d},to:{width:p,height:m,x:c,y:d},duration:w,animationEasing:S,isActive:O},function(T){var R=T.width,M=T.height,P=T.x,N=T.y;return V.createElement(Wl,{canBegin:s>0,from:"0px ".concat(s===-1?1:s,"px"),to:"".concat(s,"px 0px"),attributeName:"strokeDasharray",begin:x,duration:w,isActive:C,easing:S},V.createElement("path",Dm({},Je(n,!0),{className:A,d:z$(P,N,R,M,v),ref:r})))}):V.createElement("path",Dm({},Je(n,!0),{className:A,d:z$(c,d,p,m,v)}))},Ehe=["points","className","baseLinePoints","connectNulls"];function Sl(){return Sl=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)Object.prototype.hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},Sl.apply(this,arguments)}function Che(e,t){if(e==null)return{};var n=Ohe(e,t),r,i;if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);for(i=0;i<a.length;i++)r=a[i],!(t.indexOf(r)>=0)&&Object.prototype.propertyIsEnumerable.call(e,r)&&(n[r]=e[r])}return n}function Ohe(e,t){if(e==null)return{};var n={};for(var r in e)if(Object.prototype.hasOwnProperty.call(e,r)){if(t.indexOf(r)>=0)continue;n[r]=e[r]}return n}function U$(e){return Mhe(e)||The(e)||Rhe(e)||Ahe()}function Ahe(){throw new TypeError(`Invalid attempt to spread non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function Rhe(e,t){if(e){if(typeof e=="string")return tE(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);if(n==="Object"&&e.constructor&&(n=e.constructor.name),n==="Map"||n==="Set")return Array.from(e);if(n==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return tE(e,t)}}function The(e){if(typeof Symbol<"u"&&e[Symbol.iterator]!=null||e["@@iterator"]!=null)return Array.from(e)}function Mhe(e){if(Array.isArray(e))return tE(e)}function tE(e,t){(t==null||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}var V$=function(t){return t&&t.x===+t.x&&t.y===+t.y},Phe=function(){var t=arguments.length>0&&arguments[0]!==void 0?arguments[0]:[],n=[[]];return t.forEach(function(r){V$(r)?n[n.length-1].push(r):n[n.length-1].length>0&&n.push([])}),V$(t[0])&&n[n.length-1].push(t[0]),n[n.length-1].length<=0&&(n=n.slice(0,-1)),n},Xc=function(t,n){var r=Phe(t);n&&(r=[r.reduce(function(a,s){return[].concat(U$(a),U$(s))},[])]);var i=r.map(function(a){return a.reduce(function(s,u,c){return"".concat(s).concat(c===0?"M":"L").concat(u.x,",").concat(u.y)},"")}).join("");return r.length===1?"".concat(i,"Z"):i},jhe=function(t,n,r){var i=Xc(t,r);return"".concat(i.slice(-1)==="Z"?i.slice(0,-1):i,"L").concat(Xc(n.reverse(),r).slice(1))},KV=function(t){var n=t.points,r=t.className,i=t.baseLinePoints,a=t.connectNulls,s=Che(t,Ehe);if(!n||!n.length)return null;var u=ct("recharts-polygon",r);if(i&&i.length){var c=s.stroke&&s.stroke!=="none",d=jhe(n,i,a);return V.createElement("g",{className:u},V.createElement("path",Sl({},Je(s,!0),{fill:d.slice(-1)==="Z"?s.fill:"none",stroke:"none",d})),c?V.createElement("path",Sl({},Je(s,!0),{fill:"none",d:Xc(n,a)})):null,c?V.createElement("path",Sl({},Je(s,!0),{fill:"none",d:Xc(i,a)})):null)}var p=Xc(n,a);return V.createElement("path",Sl({},Je(s,!0),{fill:p.slice(-1)==="Z"?s.fill:"none",className:u,d:p}))};function nE(){return nE=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)Object.prototype.hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},nE.apply(this,arguments)}var tg=function(t){var n=t.cx,r=t.cy,i=t.r,a=t.className,s=ct("recharts-dot",a);return n===+n&&r===+r&&i===+i?V.createElement("circle",nE({},Je(t,!1),am(t),{className:s,cx:n,cy:r,r:i})):null};function Wf(e){"@babel/helpers - typeof";return Wf=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},Wf(e)}var Nhe=["x","y","top","left","width","height","className"];function rE(){return rE=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)Object.prototype.hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},rE.apply(this,arguments)}function H$(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function Dhe(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?H$(Object(n),!0).forEach(function(r){khe(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):H$(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function khe(e,t,n){return t=$he(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function $he(e){var t=Ihe(e,"string");return Wf(t)=="symbol"?t:t+""}function Ihe(e,t){if(Wf(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(Wf(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function Lhe(e,t){if(e==null)return{};var n=Bhe(e,t),r,i;if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);for(i=0;i<a.length;i++)r=a[i],!(t.indexOf(r)>=0)&&Object.prototype.propertyIsEnumerable.call(e,r)&&(n[r]=e[r])}return n}function Bhe(e,t){if(e==null)return{};var n={};for(var r in e)if(Object.prototype.hasOwnProperty.call(e,r)){if(t.indexOf(r)>=0)continue;n[r]=e[r]}return n}var Fhe=function(t,n,r,i,a,s){return"M".concat(t,",").concat(a,"v").concat(i,"M").concat(s,",").concat(n,"h").concat(r)},qhe=function(t){var n=t.x,r=n===void 0?0:n,i=t.y,a=i===void 0?0:i,s=t.top,u=s===void 0?0:s,c=t.left,d=c===void 0?0:c,p=t.width,m=p===void 0?0:p,v=t.height,b=v===void 0?0:v,S=t.className,w=Lhe(t,Nhe),x=Dhe({x:r,y:a,top:u,left:d,width:m,height:b},w);return!$e(r)||!$e(a)||!$e(m)||!$e(b)||!$e(u)||!$e(d)?null:V.createElement("path",rE({},Je(x,!0),{className:ct("recharts-cross",S),d:Fhe(r,a,m,b,u,d)}))},zhe=["cx","cy","innerRadius","outerRadius","gridType","radialLines"];function Kf(e){"@babel/helpers - typeof";return Kf=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},Kf(e)}function Uhe(e,t){if(e==null)return{};var n=Vhe(e,t),r,i;if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);for(i=0;i<a.length;i++)r=a[i],!(t.indexOf(r)>=0)&&Object.prototype.propertyIsEnumerable.call(e,r)&&(n[r]=e[r])}return n}function Vhe(e,t){if(e==null)return{};var n={};for(var r in e)if(Object.prototype.hasOwnProperty.call(e,r)){if(t.indexOf(r)>=0)continue;n[r]=e[r]}return n}function ya(){return ya=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)Object.prototype.hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},ya.apply(this,arguments)}function G$(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function Yf(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?G$(Object(n),!0).forEach(function(r){Hhe(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):G$(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function Hhe(e,t,n){return t=Ghe(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function Ghe(e){var t=Whe(e,"string");return Kf(t)=="symbol"?t:t+""}function Whe(e,t){if(Kf(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(Kf(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}var Khe=function(t,n,r,i){var a="";return i.forEach(function(s,u){var c=At(n,r,t,s);u?a+="L ".concat(c.x,",").concat(c.y):a+="M ".concat(c.x,",").concat(c.y)}),a+="Z",a},Yhe=function(t){var n=t.cx,r=t.cy,i=t.innerRadius,a=t.outerRadius,s=t.polarAngles,u=t.radialLines;if(!s||!s.length||!u)return null;var c=Yf({stroke:"#ccc"},Je(t,!1));return V.createElement("g",{className:"recharts-polar-grid-angle"},s.map(function(d){var p=At(n,r,i,d),m=At(n,r,a,d);return V.createElement("line",ya({},c,{key:"line-".concat(d),x1:p.x,y1:p.y,x2:m.x,y2:m.y}))}))},Zhe=function(t){var n=t.cx,r=t.cy,i=t.radius,a=t.index,s=Yf(Yf({stroke:"#ccc"},Je(t,!1)),{},{fill:"none"});return V.createElement("circle",ya({},s,{className:ct("recharts-polar-grid-concentric-circle",t.className),key:"circle-".concat(a),cx:n,cy:r,r:i}))},Xhe=function(t){var n=t.radius,r=t.index,i=Yf(Yf({stroke:"#ccc"},Je(t,!1)),{},{fill:"none"});return V.createElement("path",ya({},i,{className:ct("recharts-polar-grid-concentric-polygon",t.className),key:"path-".concat(r),d:Khe(n,t.cx,t.cy,t.polarAngles)}))},Qhe=function(t){var n=t.polarRadius,r=t.gridType;return!n||!n.length?null:V.createElement("g",{className:"recharts-polar-grid-concentric"},n.map(function(i,a){var s=a;return r==="circle"?V.createElement(Zhe,ya({key:s},t,{radius:i,index:a})):V.createElement(Xhe,ya({key:s},t,{radius:i,index:a}))}))},YV=function(t){var n=t.cx,r=n===void 0?0:n,i=t.cy,a=i===void 0?0:i,s=t.innerRadius,u=s===void 0?0:s,c=t.outerRadius,d=c===void 0?0:c,p=t.gridType,m=p===void 0?"polygon":p,v=t.radialLines,b=v===void 0?!0:v,S=Uhe(t,zhe);return d<=0?null:V.createElement("g",{className:"recharts-polar-grid"},V.createElement(Yhe,ya({cx:r,cy:a,innerRadius:u,outerRadius:d,gridType:m,radialLines:b},S)),V.createElement(Qhe,ya({cx:r,cy:a,innerRadius:u,outerRadius:d,gridType:m,radialLines:b},S)))};YV.displayName="PolarGrid";var BS,W$;function Jhe(){if(W$)return BS;W$=1;var e=Zv(),t=pV(),n=Pi();function r(i,a){return i&&i.length?e(i,n(a,2),t):void 0}return BS=r,BS}var eme=Jhe();const tme=Rt(eme);var FS,K$;function nme(){if(K$)return FS;K$=1;var e=Zv(),t=Pi(),n=hV();function r(i,a){return i&&i.length?e(i,t(a,2),n):void 0}return FS=r,FS}var rme=nme();const ime=Rt(rme);var ame=["cx","cy","angle","ticks","axisLine"],ome=["ticks","tick","angle","tickFormatter","stroke"];function Kl(e){"@babel/helpers - typeof";return Kl=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},Kl(e)}function Qc(){return Qc=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)Object.prototype.hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},Qc.apply(this,arguments)}function Y$(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function Fo(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?Y$(Object(n),!0).forEach(function(r){ng(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):Y$(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function Z$(e,t){if(e==null)return{};var n=sme(e,t),r,i;if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);for(i=0;i<a.length;i++)r=a[i],!(t.indexOf(r)>=0)&&Object.prototype.propertyIsEnumerable.call(e,r)&&(n[r]=e[r])}return n}function sme(e,t){if(e==null)return{};var n={};for(var r in e)if(Object.prototype.hasOwnProperty.call(e,r)){if(t.indexOf(r)>=0)continue;n[r]=e[r]}return n}function lme(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}function X$(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,XV(r.key),r)}}function ume(e,t,n){return t&&X$(e.prototype,t),n&&X$(e,n),Object.defineProperty(e,"prototype",{writable:!1}),e}function cme(e,t,n){return t=km(t),fme(e,ZV()?Reflect.construct(t,n||[],km(e).constructor):t.apply(e,n))}function fme(e,t){if(t&&(Kl(t)==="object"||typeof t=="function"))return t;if(t!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return dme(e)}function dme(e){if(e===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return e}function ZV(){try{var e=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(ZV=function(){return!!e})()}function km(e){return km=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(n){return n.__proto__||Object.getPrototypeOf(n)},km(e)}function pme(e,t){if(typeof t!="function"&&t!==null)throw new TypeError("Super expression must either be null or a function");e.prototype=Object.create(t&&t.prototype,{constructor:{value:e,writable:!0,configurable:!0}}),Object.defineProperty(e,"prototype",{writable:!1}),t&&iE(e,t)}function iE(e,t){return iE=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(r,i){return r.__proto__=i,r},iE(e,t)}function ng(e,t,n){return t=XV(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function XV(e){var t=hme(e,"string");return Kl(t)=="symbol"?t:t+""}function hme(e,t){if(Kl(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(Kl(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}var rg=function(e){function t(){return lme(this,t),cme(this,t,arguments)}return pme(t,e),ume(t,[{key:"getTickValueCoord",value:function(r){var i=r.coordinate,a=this.props,s=a.angle,u=a.cx,c=a.cy;return At(u,c,i,s)}},{key:"getTickTextAnchor",value:function(){var r=this.props.orientation,i;switch(r){case"left":i="end";break;case"right":i="start";break;default:i="middle";break}return i}},{key:"getViewBox",value:function(){var r=this.props,i=r.cx,a=r.cy,s=r.angle,u=r.ticks,c=tme(u,function(p){return p.coordinate||0}),d=ime(u,function(p){return p.coordinate||0});return{cx:i,cy:a,startAngle:s,endAngle:s,innerRadius:d.coordinate||0,outerRadius:c.coordinate||0}}},{key:"renderAxisLine",value:function(){var r=this.props,i=r.cx,a=r.cy,s=r.angle,u=r.ticks,c=r.axisLine,d=Z$(r,ame),p=u.reduce(function(S,w){return[Math.min(S[0],w.coordinate),Math.max(S[1],w.coordinate)]},[1/0,-1/0]),m=At(i,a,p[0],s),v=At(i,a,p[1],s),b=Fo(Fo(Fo({},Je(d,!1)),{},{fill:"none"},Je(c,!1)),{},{x1:m.x,y1:m.y,x2:v.x,y2:v.y});return V.createElement("line",Qc({className:"recharts-polar-radius-axis-line"},b))}},{key:"renderTicks",value:function(){var r=this,i=this.props,a=i.ticks,s=i.tick,u=i.angle,c=i.tickFormatter,d=i.stroke,p=Z$(i,ome),m=this.getTickTextAnchor(),v=Je(p,!1),b=Je(s,!1),S=a.map(function(w,x){var C=r.getTickValueCoord(w),O=Fo(Fo(Fo(Fo({textAnchor:m,transform:"rotate(".concat(90-u,", ").concat(C.x,", ").concat(C.y,")")},v),{},{stroke:"none",fill:d},b),{},{index:x},C),{},{payload:w});return V.createElement(un,Qc({className:ct("recharts-polar-radius-axis-tick",$V(s)),key:"tick-".concat(w.coordinate)},FC(r.props,w,x)),t.renderTickItem(s,O,c?c(w.value,x):w.value))});return V.createElement(un,{className:"recharts-polar-radius-axis-ticks"},S)}},{key:"render",value:function(){var r=this.props,i=r.ticks,a=r.axisLine,s=r.tick;return!i||!i.length?null:V.createElement(un,{className:ct("recharts-polar-radius-axis",this.props.className)},a&&this.renderAxisLine(),s&&this.renderTicks(),Tn.renderCallByParent(this.props,this.getViewBox()))}}],[{key:"renderTickItem",value:function(r,i,a){var s;return V.isValidElement(r)?s=V.cloneElement(r,i):st(r)?s=r(i):s=V.createElement(Rf,Qc({},i,{className:"recharts-polar-radius-axis-tick-value"}),a),s}}])}(g.PureComponent);ng(rg,"displayName","PolarRadiusAxis");ng(rg,"axisType","radiusAxis");ng(rg,"defaultProps",{type:"number",radiusAxisId:0,cx:0,cy:0,angle:0,orientation:"right",stroke:"#ccc",axisLine:!0,tick:!0,tickCount:5,allowDataOverflow:!1,scale:"auto",allowDuplicatedCategory:!0});function Yl(e){"@babel/helpers - typeof";return Yl=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},Yl(e)}function Vo(){return Vo=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)Object.prototype.hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},Vo.apply(this,arguments)}function Q$(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function qo(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?Q$(Object(n),!0).forEach(function(r){ig(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):Q$(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function mme(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}function J$(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,JV(r.key),r)}}function vme(e,t,n){return t&&J$(e.prototype,t),n&&J$(e,n),Object.defineProperty(e,"prototype",{writable:!1}),e}function gme(e,t,n){return t=$m(t),yme(e,QV()?Reflect.construct(t,n||[],$m(e).constructor):t.apply(e,n))}function yme(e,t){if(t&&(Yl(t)==="object"||typeof t=="function"))return t;if(t!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return bme(e)}function bme(e){if(e===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return e}function QV(){try{var e=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(QV=function(){return!!e})()}function $m(e){return $m=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(n){return n.__proto__||Object.getPrototypeOf(n)},$m(e)}function xme(e,t){if(typeof t!="function"&&t!==null)throw new TypeError("Super expression must either be null or a function");e.prototype=Object.create(t&&t.prototype,{constructor:{value:e,writable:!0,configurable:!0}}),Object.defineProperty(e,"prototype",{writable:!1}),t&&aE(e,t)}function aE(e,t){return aE=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(r,i){return r.__proto__=i,r},aE(e,t)}function ig(e,t,n){return t=JV(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function JV(e){var t=wme(e,"string");return Yl(t)=="symbol"?t:t+""}function wme(e,t){if(Yl(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(Yl(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(e)}var Sme=Math.PI/180,_me=1e-5,Ad=function(e){function t(){return mme(this,t),gme(this,t,arguments)}return xme(t,e),vme(t,[{key:"getTickLineCoord",value:function(r){var i=this.props,a=i.cx,s=i.cy,u=i.radius,c=i.orientation,d=i.tickSize,p=d||8,m=At(a,s,u,r.coordinate),v=At(a,s,u+(c==="inner"?-1:1)*p,r.coordinate);return{x1:m.x,y1:m.y,x2:v.x,y2:v.y}}},{key:"getTickTextAnchor",value:function(r){var i=this.props.orientation,a=Math.cos(-r.coordinate*Sme),s;return a>_me?s=i==="outer"?"start":"end":a<-1e-5?s=i==="outer"?"end":"start":s="middle",s}},{key:"renderAxisLine",value:function(){var r=this.props,i=r.cx,a=r.cy,s=r.radius,u=r.axisLine,c=r.axisLineType,d=qo(qo({},Je(this.props,!1)),{},{fill:"none"},Je(u,!1));if(c==="circle")return V.createElement(tg,Vo({className:"recharts-polar-angle-axis-line"},d,{cx:i,cy:a,r:s}));var p=this.props.ticks,m=p.map(function(v){return At(i,a,s,v.coordinate)});return V.createElement(KV,Vo({className:"recharts-polar-angle-axis-line"},d,{points:m}))}},{key:"renderTicks",value:function(){var r=this,i=this.props,a=i.ticks,s=i.tick,u=i.tickLine,c=i.tickFormatter,d=i.stroke,p=Je(this.props,!1),m=Je(s,!1),v=qo(qo({},p),{},{fill:"none"},Je(u,!1)),b=a.map(function(S,w){var x=r.getTickLineCoord(S),C=r.getTickTextAnchor(S),O=qo(qo(qo({textAnchor:C},p),{},{stroke:"none",fill:d},m),{},{index:w,payload:S,x:x.x2,y:x.y2});return V.createElement(un,Vo({className:ct("recharts-polar-angle-axis-tick",$V(s)),key:"tick-".concat(S.coordinate)},FC(r.props,S,w)),u&&V.createElement("line",Vo({className:"recharts-polar-angle-axis-tick-line"},v,x)),s&&t.renderTickItem(s,O,c?c(S.value,w):S.value))});return V.createElement(un,{className:"recharts-polar-angle-axis-ticks"},b)}},{key:"render",value:function(){var r=this.props,i=r.ticks,a=r.radius,s=r.axisLine;return a<=0||!i||!i.length?null:V.createElement(un,{className:ct("recharts-polar-angle-axis",this.props.className)},s&&this.renderAxisLine(),this.renderTicks())}}],[{key:"renderTickItem",value:function(r,i,a){var s;return V.isValidElement(r)?s=V.cloneElement(r,i):st(r)?s=r(i):s=V.createElement(Rf,Vo({},i,{className:"recharts-polar-angle-axis-tick-value"}),a),s}}])}(g.PureComponent);ig(Ad,"displayName","PolarAngleAxis");ig(Ad,"axisType","angleAxis");ig(Ad,"defaultProps",{type:"category",angleAxisId:0,scale:"auto",cx:0,cy:0,orientation:"outer",axisLine:!0,tickLine:!0,tickSize:8,tick:!0,hide:!1,allowDuplicatedCategory:!0});var qS,eI;function Eme(){if(eI)return qS;eI=1;var e=dU(),t=e(Object.getPrototypeOf,Object);return qS=t,qS}var zS,tI;function Cme(){if(tI)return zS;tI=1;var e=xa(),t=Eme(),n=wa(),r="[object Object]",i=Function.prototype,a=Object.prototype,s=i.toString,u=a.hasOwnProperty,c=s.call(Object);function d(p){if(!n(p)||e(p)!=r)return!1;var m=t(p);if(m===null)return!0;var v=u.call(m,"constructor")&&m.constructor;return typeof v=="function"&&v instanceof v&&s.call(v)==c}return zS=d,zS}Cme();var US,nI;function Ome(){if(nI)return US;nI=1;var e=xa(),t=wa(),n="[object Boolean]";function r(i){return i===!0||i===!1||t(i)&&e(i)==n}return US=r,US}Ome();function ag(e,t){return t!=null&&"trapezoids"in e.props}function og(e,t){return t!=null&&"sectors"in e.props}function Zf(e,t){return t!=null&&"points"in e.props}function Ame(e,t){var n,r,i=e.x===(t==null||(n=t.labelViewBox)===null||n===void 0?void 0:n.x)||e.x===t.x,a=e.y===(t==null||(r=t.labelViewBox)===null||r===void 0?void 0:r.y)||e.y===t.y;return i&&a}function Rme(e,t){var n=e.endAngle===t.endAngle,r=e.startAngle===t.startAngle;return n&&r}function Tme(e,t){var n=e.x===t.x,r=e.y===t.y,i=e.z===t.z;return n&&r&&i}function Mme(e,t){var n;return ag(e,t)?n=Ame:og(e,t)?n=Rme:Zf(e,t)&&(n=Tme),n}function Pme(e,t){var n;return ag(e,t)?n="trapezoids":og(e,t)?n="sectors":Zf(e,t)&&(n="points"),n}function jme(e,t){if(ag(e,t)){var n;return(n=t.tooltipPayload)===null||n===void 0||(n=n[0])===null||n===void 0||(n=n.payload)===null||n===void 0?void 0:n.payload}if(og(e,t)){var r;return(r=t.tooltipPayload)===null||r===void 0||(r=r[0])===null||r===void 0||(r=r.payload)===null||r===void 0?void 0:r.payload}return Zf(e,t)?t.payload:{}}function Nme(e){var t=e.activeTooltipItem,n=e.graphicalItem,r=e.itemData,i=Pme(n,t),a=jme(n,t),s=r.filter(function(c,d){var p=SO(a,c),m=n.props[i].filter(function(S){var w=Mme(n,t);return w(S,t)}),v=n.props[i].indexOf(m[m.length-1]),b=d===v;return p&&b}),u=r.indexOf(s[s.length-1]);return u}var VS,rI;function Dme(){if(rI)return VS;rI=1;function e(t){return t&&t.length?t[0]:void 0}return VS=e,VS}var HS,iI;function kme(){return iI||(iI=1,HS=Dme()),HS}var $me=kme();const Ime=Rt($me);var Lme=["key"];function Zl(e){"@babel/helpers - typeof";return Zl=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},Zl(e)}function Bme(e,t){if(e==null)return{};var n=Fme(e,t),r,i;if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);for(i=0;i<a.length;i++)r=a[i],!(t.indexOf(r)>=0)&&Object.prototype.propertyIsEnumerable.call(e,r)&&(n[r]=e[r])}return n}function Fme(e,t){if(e==null)return{};var n={};for(var r in e)if(Object.prototype.hasOwnProperty.call(e,r)){if(t.indexOf(r)>=0)continue;n[r]=e[r]}return n}function Im(){return Im=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)Object.prototype.hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},Im.apply(this,arguments)}function aI(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function $n(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?aI(Object(n),!0).forEach(function(r){ta(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):aI(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function qme(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}function oI(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,t3(r.key),r)}}function zme(e,t,n){return t&&oI(e.prototype,t),n&&oI(e,n),Object.defineProperty(e,"prototype",{writable:!1}),e}function Ume(e,t,n){return t=Lm(t),Vme(e,e3()?Reflect.construct(t,n||[],Lm(e).constructor):t.apply(e,n))}function Vme(e,t){if(t&&(Zl(t)==="object"||typeof t=="function"))return t;if(t!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return Hme(e)}function Hme(e){if(e===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return e}function e3(){try{var e=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(e3=function(){return!!e})()}function Lm(e){return Lm=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(n){return n.__proto__||Object.getPrototypeOf(n)},Lm(e)}function Gme(e,t){if(typeof t!="function"&&t!==null)throw new TypeError("Super expression must either be null or a function");e.prototype=Object.create(t&&t.prototype,{constructor:{value:e,writable:!0,configurable:!0}}),Object.defineProperty(e,"prototype",{writable:!1}),t&&oE(e,t)}function oE(e,t){return oE=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(r,i){return r.__proto__=i,r},oE(e,t)}function ta(e,t,n){return t=t3(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function t3(e){var t=Wme(e,"string");return Zl(t)=="symbol"?t:t+""}function Wme(e,t){if(Zl(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(Zl(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(e)}var Rd=function(e){function t(){var n;qme(this,t);for(var r=arguments.length,i=new Array(r),a=0;a<r;a++)i[a]=arguments[a];return n=Ume(this,t,[].concat(i)),ta(n,"state",{isAnimationFinished:!1}),ta(n,"handleAnimationEnd",function(){var s=n.props.onAnimationEnd;n.setState({isAnimationFinished:!0}),st(s)&&s()}),ta(n,"handleAnimationStart",function(){var s=n.props.onAnimationStart;n.setState({isAnimationFinished:!1}),st(s)&&s()}),ta(n,"handleMouseEnter",function(s){var u=n.props.onMouseEnter;u&&u(n.props,s)}),ta(n,"handleMouseLeave",function(s){var u=n.props.onMouseLeave;u&&u(n.props,s)}),n}return Gme(t,e),zme(t,[{key:"renderDots",value:function(r){var i=this.props,a=i.dot,s=i.dataKey,u=Je(this.props,!1),c=Je(a,!0),d=r.map(function(p,m){var v=$n($n($n({key:"dot-".concat(m),r:3},u),c),{},{dataKey:s,cx:p.x,cy:p.y,index:m,payload:p});return t.renderDotItem(a,v)});return V.createElement(un,{className:"recharts-radar-dots"},d)}},{key:"renderPolygonStatically",value:function(r){var i=this.props,a=i.shape,s=i.dot,u=i.isRange,c=i.baseLinePoints,d=i.connectNulls,p;return V.isValidElement(a)?p=V.cloneElement(a,$n($n({},this.props),{},{points:r})):st(a)?p=a($n($n({},this.props),{},{points:r})):p=V.createElement(KV,Im({},Je(this.props,!0),{onMouseEnter:this.handleMouseEnter,onMouseLeave:this.handleMouseLeave,points:r,baseLinePoints:u?c:null,connectNulls:d})),V.createElement(un,{className:"recharts-radar-polygon"},p,s?this.renderDots(r):null)}},{key:"renderPolygonWithAnimation",value:function(){var r=this,i=this.props,a=i.points,s=i.isAnimationActive,u=i.animationBegin,c=i.animationDuration,d=i.animationEasing,p=i.animationId,m=this.state.prevPoints;return V.createElement(Wl,{begin:u,duration:c,isActive:s,easing:d,from:{t:0},to:{t:1},key:"radar-".concat(p),onAnimationEnd:this.handleAnimationEnd,onAnimationStart:this.handleAnimationStart},function(v){var b=v.t,S=m&&m.length/a.length,w=a.map(function(x,C){var O=m&&m[Math.floor(C*S)];if(O){var A=nh(O.x,x.x),T=nh(O.y,x.y);return $n($n({},x),{},{x:A(b),y:T(b)})}var R=nh(x.cx,x.x),M=nh(x.cy,x.y);return $n($n({},x),{},{x:R(b),y:M(b)})});return r.renderPolygonStatically(w)})}},{key:"renderPolygon",value:function(){var r=this.props,i=r.points,a=r.isAnimationActive,s=r.isRange,u=this.state.prevPoints;return a&&i&&i.length&&!s&&(!u||!SO(u,i))?this.renderPolygonWithAnimation():this.renderPolygonStatically(i)}},{key:"render",value:function(){var r=this.props,i=r.hide,a=r.className,s=r.points,u=r.isAnimationActive;if(i||!s||!s.length)return null;var c=this.state.isAnimationFinished,d=ct("recharts-radar",a);return V.createElement(un,{className:d},this.renderPolygon(),(!u||c)&&rs.renderCallByParent(this.props,s))}}],[{key:"getDerivedStateFromProps",value:function(r,i){return r.animationId!==i.prevAnimationId?{prevAnimationId:r.animationId,curPoints:r.points,prevPoints:i.curPoints}:r.points!==i.curPoints?{curPoints:r.points}:null}},{key:"renderDotItem",value:function(r,i){var a;if(V.isValidElement(r))a=V.cloneElement(r,i);else if(st(r))a=r(i);else{var s=i.key,u=Bme(i,Lme);a=V.createElement(tg,Im({},u,{key:s,className:ct("recharts-radar-dot",typeof r!="boolean"?r.className:"")}))}return a}}])}(g.PureComponent);ta(Rd,"displayName","Radar");ta(Rd,"defaultProps",{angleAxisId:0,radiusAxisId:0,hide:!1,activeDot:!0,dot:!1,legendType:"rect",isAnimationActive:!Uv.isSsr,animationBegin:0,animationDuration:1500,animationEasing:"ease"});ta(Rd,"getComposedData",function(e){var t=e.radiusAxis,n=e.angleAxis,r=e.displayedData,i=e.dataKey,a=e.bandSize,s=n.cx,u=n.cy,c=!1,d=[],p=n.type!=="number"?a??0:0;r.forEach(function(v,b){var S=ga(v,n.dataKey,b),w=ga(v,i),x=n.scale(S)+p,C=Array.isArray(w)?LV(w):w,O=vt(C)?void 0:t.scale(C);Array.isArray(w)&&w.length>=2&&(c=!0),d.push($n($n({},At(s,u,O,x)),{},{name:S,value:w,cx:s,cy:u,radius:O,angle:x,payload:v}))});var m=[];return c&&d.forEach(function(v){if(Array.isArray(v.value)){var b=Ime(v.value),S=vt(b)?void 0:t.scale(b);m.push($n($n({},v),{},{radius:S},At(s,u,S,v.angle)))}else m.push(v)}),{points:d,isRange:c,baseLinePoints:m}});var GS,sI;function Kme(){if(sI)return GS;sI=1;var e=Math.ceil,t=Math.max;function n(r,i,a,s){for(var u=-1,c=t(e((i-r)/(a||1)),0),d=Array(c);c--;)d[s?c:++u]=r,r+=a;return d}return GS=n,GS}var WS,lI;function n3(){if(lI)return WS;lI=1;var e=RU(),t=1/0,n=17976931348623157e292;function r(i){if(!i)return i===0?i:0;if(i=e(i),i===t||i===-1/0){var a=i<0?-1:1;return a*n}return i===i?i:0}return WS=r,WS}var KS,uI;function Yme(){if(uI)return KS;uI=1;var e=Kme(),t=zv(),n=n3();function r(i){return function(a,s,u){return u&&typeof u!="number"&&t(a,s,u)&&(s=u=void 0),a=n(a),s===void 0?(s=a,a=0):s=n(s),u=u===void 0?a<s?1:-1:n(u),e(a,s,u,i)}}return KS=r,KS}var YS,cI;function Zme(){if(cI)return YS;cI=1;var e=Yme(),t=e();return YS=t,YS}var Xme=Zme();const Bm=Rt(Xme);function Xf(e){"@babel/helpers - typeof";return Xf=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},Xf(e)}function fI(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function dI(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?fI(Object(n),!0).forEach(function(r){r3(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):fI(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function r3(e,t,n){return t=Qme(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function Qme(e){var t=Jme(e,"string");return Xf(t)=="symbol"?t:t+""}function Jme(e,t){if(Xf(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(Xf(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}var eve=["Webkit","Moz","O","ms"],tve=function(t,n){var r=t.replace(/(\w)/,function(a){return a.toUpperCase()}),i=eve.reduce(function(a,s){return dI(dI({},a),{},r3({},s+r,n))},{});return i[t]=n,i};function Xl(e){"@babel/helpers - typeof";return Xl=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},Xl(e)}function Fm(){return Fm=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)Object.prototype.hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},Fm.apply(this,arguments)}function pI(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function ZS(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?pI(Object(n),!0).forEach(function(r){or(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):pI(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function nve(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}function hI(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,a3(r.key),r)}}function rve(e,t,n){return t&&hI(e.prototype,t),n&&hI(e,n),Object.defineProperty(e,"prototype",{writable:!1}),e}function ive(e,t,n){return t=qm(t),ave(e,i3()?Reflect.construct(t,n||[],qm(e).constructor):t.apply(e,n))}function ave(e,t){if(t&&(Xl(t)==="object"||typeof t=="function"))return t;if(t!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return ove(e)}function ove(e){if(e===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return e}function i3(){try{var e=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(i3=function(){return!!e})()}function qm(e){return qm=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(n){return n.__proto__||Object.getPrototypeOf(n)},qm(e)}function sve(e,t){if(typeof t!="function"&&t!==null)throw new TypeError("Super expression must either be null or a function");e.prototype=Object.create(t&&t.prototype,{constructor:{value:e,writable:!0,configurable:!0}}),Object.defineProperty(e,"prototype",{writable:!1}),t&&sE(e,t)}function sE(e,t){return sE=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(r,i){return r.__proto__=i,r},sE(e,t)}function or(e,t,n){return t=a3(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function a3(e){var t=lve(e,"string");return Xl(t)=="symbol"?t:t+""}function lve(e,t){if(Xl(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(Xl(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(e)}var uve=function(t){var n=t.data,r=t.startIndex,i=t.endIndex,a=t.x,s=t.width,u=t.travellerWidth;if(!n||!n.length)return{};var c=n.length,d=Kc().domain(Bm(0,c)).range([a,a+s-u]),p=d.domain().map(function(m){return d(m)});return{isTextActive:!1,isSlideMoving:!1,isTravellerMoving:!1,isTravellerFocused:!1,startX:d(r),endX:d(i),scale:d,scaleValues:p}},mI=function(t){return t.changedTouches&&!!t.changedTouches.length},Ql=function(e){function t(n){var r;return nve(this,t),r=ive(this,t,[n]),or(r,"handleDrag",function(i){r.leaveTimer&&(clearTimeout(r.leaveTimer),r.leaveTimer=null),r.state.isTravellerMoving?r.handleTravellerMove(i):r.state.isSlideMoving&&r.handleSlideDrag(i)}),or(r,"handleTouchMove",function(i){i.changedTouches!=null&&i.changedTouches.length>0&&r.handleDrag(i.changedTouches[0])}),or(r,"handleDragEnd",function(){r.setState({isTravellerMoving:!1,isSlideMoving:!1},function(){var i=r.props,a=i.endIndex,s=i.onDragEnd,u=i.startIndex;s==null||s({endIndex:a,startIndex:u})}),r.detachDragEndListener()}),or(r,"handleLeaveWrapper",function(){(r.state.isTravellerMoving||r.state.isSlideMoving)&&(r.leaveTimer=window.setTimeout(r.handleDragEnd,r.props.leaveTimeOut))}),or(r,"handleEnterSlideOrTraveller",function(){r.setState({isTextActive:!0})}),or(r,"handleLeaveSlideOrTraveller",function(){r.setState({isTextActive:!1})}),or(r,"handleSlideDragStart",function(i){var a=mI(i)?i.changedTouches[0]:i;r.setState({isTravellerMoving:!1,isSlideMoving:!0,slideMoveStartX:a.pageX}),r.attachDragEndListener()}),r.travellerDragStartHandlers={startX:r.handleTravellerDragStart.bind(r,"startX"),endX:r.handleTravellerDragStart.bind(r,"endX")},r.state={},r}return sve(t,e),rve(t,[{key:"componentWillUnmount",value:function(){this.leaveTimer&&(clearTimeout(this.leaveTimer),this.leaveTimer=null),this.detachDragEndListener()}},{key:"getIndex",value:function(r){var i=r.startX,a=r.endX,s=this.state.scaleValues,u=this.props,c=u.gap,d=u.data,p=d.length-1,m=Math.min(i,a),v=Math.max(i,a),b=t.getIndexInRange(s,m),S=t.getIndexInRange(s,v);return{startIndex:b-b%c,endIndex:S===p?p:S-S%c}}},{key:"getTextOfTick",value:function(r){var i=this.props,a=i.data,s=i.tickFormatter,u=i.dataKey,c=ga(a[r],u,r);return st(s)?s(c,r):c}},{key:"attachDragEndListener",value:function(){window.addEventListener("mouseup",this.handleDragEnd,!0),window.addEventListener("touchend",this.handleDragEnd,!0),window.addEventListener("mousemove",this.handleDrag,!0)}},{key:"detachDragEndListener",value:function(){window.removeEventListener("mouseup",this.handleDragEnd,!0),window.removeEventListener("touchend",this.handleDragEnd,!0),window.removeEventListener("mousemove",this.handleDrag,!0)}},{key:"handleSlideDrag",value:function(r){var i=this.state,a=i.slideMoveStartX,s=i.startX,u=i.endX,c=this.props,d=c.x,p=c.width,m=c.travellerWidth,v=c.startIndex,b=c.endIndex,S=c.onChange,w=r.pageX-a;w>0?w=Math.min(w,d+p-m-u,d+p-m-s):w<0&&(w=Math.max(w,d-s,d-u));var x=this.getIndex({startX:s+w,endX:u+w});(x.startIndex!==v||x.endIndex!==b)&&S&&S(x),this.setState({startX:s+w,endX:u+w,slideMoveStartX:r.pageX})}},{key:"handleTravellerDragStart",value:function(r,i){var a=mI(i)?i.changedTouches[0]:i;this.setState({isSlideMoving:!1,isTravellerMoving:!0,movingTravellerId:r,brushMoveStartX:a.pageX}),this.attachDragEndListener()}},{key:"handleTravellerMove",value:function(r){var i=this.state,a=i.brushMoveStartX,s=i.movingTravellerId,u=i.endX,c=i.startX,d=this.state[s],p=this.props,m=p.x,v=p.width,b=p.travellerWidth,S=p.onChange,w=p.gap,x=p.data,C={startX:this.state.startX,endX:this.state.endX},O=r.pageX-a;O>0?O=Math.min(O,m+v-b-d):O<0&&(O=Math.max(O,m-d)),C[s]=d+O;var A=this.getIndex(C),T=A.startIndex,R=A.endIndex,M=function(){var N=x.length-1;return s==="startX"&&(u>c?T%w===0:R%w===0)||u<c&&R===N||s==="endX"&&(u>c?R%w===0:T%w===0)||u>c&&R===N};this.setState(or(or({},s,d+O),"brushMoveStartX",r.pageX),function(){S&&M()&&S(A)})}},{key:"handleTravellerMoveKeyboard",value:function(r,i){var a=this,s=this.state,u=s.scaleValues,c=s.startX,d=s.endX,p=this.state[i],m=u.indexOf(p);if(m!==-1){var v=m+r;if(!(v===-1||v>=u.length)){var b=u[v];i==="startX"&&b>=d||i==="endX"&&b<=c||this.setState(or({},i,b),function(){a.props.onChange(a.getIndex({startX:a.state.startX,endX:a.state.endX}))})}}}},{key:"renderBackground",value:function(){var r=this.props,i=r.x,a=r.y,s=r.width,u=r.height,c=r.fill,d=r.stroke;return V.createElement("rect",{stroke:d,fill:c,x:i,y:a,width:s,height:u})}},{key:"renderPanorama",value:function(){var r=this.props,i=r.x,a=r.y,s=r.width,u=r.height,c=r.data,d=r.children,p=r.padding,m=g.Children.only(d);return m?V.cloneElement(m,{x:i,y:a,width:s,height:u,margin:p,compact:!0,data:c}):null}},{key:"renderTravellerLayer",value:function(r,i){var a,s,u=this,c=this.props,d=c.y,p=c.travellerWidth,m=c.height,v=c.traveller,b=c.ariaLabel,S=c.data,w=c.startIndex,x=c.endIndex,C=Math.max(r,this.props.x),O=ZS(ZS({},Je(this.props,!1)),{},{x:C,y:d,width:p,height:m}),A=b||"Min value: ".concat((a=S[w])===null||a===void 0?void 0:a.name,", Max value: ").concat((s=S[x])===null||s===void 0?void 0:s.name);return V.createElement(un,{tabIndex:0,role:"slider","aria-label":A,"aria-valuenow":r,className:"recharts-brush-traveller",onMouseEnter:this.handleEnterSlideOrTraveller,onMouseLeave:this.handleLeaveSlideOrTraveller,onMouseDown:this.travellerDragStartHandlers[i],onTouchStart:this.travellerDragStartHandlers[i],onKeyDown:function(R){["ArrowLeft","ArrowRight"].includes(R.key)&&(R.preventDefault(),R.stopPropagation(),u.handleTravellerMoveKeyboard(R.key==="ArrowRight"?1:-1,i))},onFocus:function(){u.setState({isTravellerFocused:!0})},onBlur:function(){u.setState({isTravellerFocused:!1})},style:{cursor:"col-resize"}},t.renderTraveller(v,O))}},{key:"renderSlide",value:function(r,i){var a=this.props,s=a.y,u=a.height,c=a.stroke,d=a.travellerWidth,p=Math.min(r,i)+d,m=Math.max(Math.abs(i-r)-d,0);return V.createElement("rect",{className:"recharts-brush-slide",onMouseEnter:this.handleEnterSlideOrTraveller,onMouseLeave:this.handleLeaveSlideOrTraveller,onMouseDown:this.handleSlideDragStart,onTouchStart:this.handleSlideDragStart,style:{cursor:"move"},stroke:"none",fill:c,fillOpacity:.2,x:p,y:s,width:m,height:u})}},{key:"renderText",value:function(){var r=this.props,i=r.startIndex,a=r.endIndex,s=r.y,u=r.height,c=r.travellerWidth,d=r.stroke,p=this.state,m=p.startX,v=p.endX,b=5,S={pointerEvents:"none",fill:d};return V.createElement(un,{className:"recharts-brush-texts"},V.createElement(Rf,Fm({textAnchor:"end",verticalAnchor:"middle",x:Math.min(m,v)-b,y:s+u/2},S),this.getTextOfTick(i)),V.createElement(Rf,Fm({textAnchor:"start",verticalAnchor:"middle",x:Math.max(m,v)+c+b,y:s+u/2},S),this.getTextOfTick(a)))}},{key:"render",value:function(){var r=this.props,i=r.data,a=r.className,s=r.children,u=r.x,c=r.y,d=r.width,p=r.height,m=r.alwaysShowText,v=this.state,b=v.startX,S=v.endX,w=v.isTextActive,x=v.isSlideMoving,C=v.isTravellerMoving,O=v.isTravellerFocused;if(!i||!i.length||!$e(u)||!$e(c)||!$e(d)||!$e(p)||d<=0||p<=0)return null;var A=ct("recharts-brush",a),T=V.Children.count(s)===1,R=tve("userSelect","none");return V.createElement(un,{className:A,onMouseLeave:this.handleLeaveWrapper,onTouchMove:this.handleTouchMove,style:R},this.renderBackground(),T&&this.renderPanorama(),this.renderSlide(b,S),this.renderTravellerLayer(b,"startX"),this.renderTravellerLayer(S,"endX"),(w||x||C||O||m)&&this.renderText())}}],[{key:"renderDefaultTraveller",value:function(r){var i=r.x,a=r.y,s=r.width,u=r.height,c=r.stroke,d=Math.floor(a+u/2)-1;return V.createElement(V.Fragment,null,V.createElement("rect",{x:i,y:a,width:s,height:u,fill:c,stroke:"none"}),V.createElement("line",{x1:i+1,y1:d,x2:i+s-1,y2:d,fill:"none",stroke:"#fff"}),V.createElement("line",{x1:i+1,y1:d+2,x2:i+s-1,y2:d+2,fill:"none",stroke:"#fff"}))}},{key:"renderTraveller",value:function(r,i){var a;return V.isValidElement(r)?a=V.cloneElement(r,i):st(r)?a=r(i):a=t.renderDefaultTraveller(i),a}},{key:"getDerivedStateFromProps",value:function(r,i){var a=r.data,s=r.width,u=r.x,c=r.travellerWidth,d=r.updateId,p=r.startIndex,m=r.endIndex;if(a!==i.prevData||d!==i.prevUpdateId)return ZS({prevData:a,prevTravellerWidth:c,prevUpdateId:d,prevX:u,prevWidth:s},a&&a.length?uve({data:a,width:s,x:u,travellerWidth:c,startIndex:p,endIndex:m}):{scale:null,scaleValues:null});if(i.scale&&(s!==i.prevWidth||u!==i.prevX||c!==i.prevTravellerWidth)){i.scale.range([u,u+s-c]);var v=i.scale.domain().map(function(b){return i.scale(b)});return{prevData:a,prevTravellerWidth:c,prevUpdateId:d,prevX:u,prevWidth:s,startX:i.scale(r.startIndex),endX:i.scale(r.endIndex),scaleValues:v}}return null}},{key:"getIndexInRange",value:function(r,i){for(var a=r.length,s=0,u=a-1;u-s>1;){var c=Math.floor((s+u)/2);r[c]>i?u=c:s=c}return i>=r[u]?u:s}}])}(g.PureComponent);or(Ql,"displayName","Brush");or(Ql,"defaultProps",{height:40,travellerWidth:5,gap:1,fill:"#fff",stroke:"#666",padding:{top:1,right:1,bottom:1,left:1},leaveTimeOut:1e3,alwaysShowText:!1});var XS,vI;function cve(){if(vI)return XS;vI=1;var e=XC();function t(n,r){var i;return e(n,function(a,s,u){return i=r(a,s,u),!i}),!!i}return XS=t,XS}var QS,gI;function fve(){if(gI)return QS;gI=1;var e=aU(),t=Pi(),n=cve(),r=Qn(),i=zv();function a(s,u,c){var d=r(s)?e:n;return c&&i(s,u,c)&&(u=void 0),d(s,t(u,3))}return QS=a,QS}var dve=fve();const pve=Rt(dve);var xi=function(t,n){var r=t.alwaysShow,i=t.ifOverflow;return r&&(i="extendDomain"),i===n},JS,yI;function hve(){if(yI)return JS;yI=1;var e=_U();function t(n,r,i){r=="__proto__"&&e?e(n,r,{configurable:!0,enumerable:!0,value:i,writable:!0}):n[r]=i}return JS=t,JS}var e_,bI;function mve(){if(bI)return e_;bI=1;var e=hve(),t=wU(),n=Pi();function r(i,a){var s={};return a=n(a,3),t(i,function(u,c,d){e(s,c,a(u,c,d))}),s}return e_=r,e_}var vve=mve();const gve=Rt(vve);var t_,xI;function yve(){if(xI)return t_;xI=1;function e(t,n){for(var r=-1,i=t==null?0:t.length;++r<i;)if(!n(t[r],r,t))return!1;return!0}return t_=e,t_}var n_,wI;function bve(){if(wI)return n_;wI=1;var e=XC();function t(n,r){var i=!0;return e(n,function(a,s,u){return i=!!r(a,s,u),i}),i}return n_=t,n_}var r_,SI;function xve(){if(SI)return r_;SI=1;var e=yve(),t=bve(),n=Pi(),r=Qn(),i=zv();function a(s,u,c){var d=r(s)?e:t;return c&&i(s,u,c)&&(u=void 0),d(s,n(u,3))}return r_=a,r_}var wve=xve();const Sve=Rt(wve);function Qf(e){"@babel/helpers - typeof";return Qf=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},Qf(e)}function _ve(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}function _I(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,o3(r.key),r)}}function Eve(e,t,n){return t&&_I(e.prototype,t),n&&_I(e,n),Object.defineProperty(e,"prototype",{writable:!1}),e}function EI(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function hh(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?EI(Object(n),!0).forEach(function(r){AO(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):EI(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function AO(e,t,n){return t=o3(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function o3(e){var t=Cve(e,"string");return Qf(t)=="symbol"?t:t+""}function Cve(e,t){if(Qf(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(Qf(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}var s3=function(t,n){var r=t.x,i=t.y,a=n.x,s=n.y;return{x:Math.min(r,a),y:Math.min(i,s),width:Math.abs(a-r),height:Math.abs(s-i)}},Ove=function(t){var n=t.x1,r=t.y1,i=t.x2,a=t.y2;return s3({x:n,y:r},{x:i,y:a})},l3=function(){function e(t){_ve(this,e),this.scale=t}return Eve(e,[{key:"domain",get:function(){return this.scale.domain}},{key:"range",get:function(){return this.scale.range}},{key:"rangeMin",get:function(){return this.range()[0]}},{key:"rangeMax",get:function(){return this.range()[1]}},{key:"bandwidth",get:function(){return this.scale.bandwidth}},{key:"apply",value:function(n){var r=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{},i=r.bandAware,a=r.position;if(n!==void 0){if(a)switch(a){case"start":return this.scale(n);case"middle":{var s=this.bandwidth?this.bandwidth()/2:0;return this.scale(n)+s}case"end":{var u=this.bandwidth?this.bandwidth():0;return this.scale(n)+u}default:return this.scale(n)}if(i){var c=this.bandwidth?this.bandwidth()/2:0;return this.scale(n)+c}return this.scale(n)}}},{key:"isInRange",value:function(n){var r=this.range(),i=r[0],a=r[r.length-1];return i<=a?n>=i&&n<=a:n>=a&&n<=i}}],[{key:"create",value:function(n){return new e(n)}}])}();AO(l3,"EPS",1e-4);var RO=function(t){var n=Object.keys(t).reduce(function(r,i){return hh(hh({},r),{},AO({},i,l3.create(t[i])))},{});return hh(hh({},n),{},{apply:function(i){var a=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{},s=a.bandAware,u=a.position;return gve(i,function(c,d){return n[d].apply(c,{bandAware:s,position:u})})},isInRange:function(i){return Sve(i,function(a,s){return n[s].isInRange(a)})}})},i_,CI;function Ave(){if(CI)return i_;CI=1;var e=Pi(),t=Sd(),n=Fv();function r(i){return function(a,s,u){var c=Object(a);if(!t(a)){var d=e(s,3);a=n(a),s=function(m){return d(c[m],m,c)}}var p=i(a,s,u);return p>-1?c[d?a[p]:p]:void 0}}return i_=r,i_}var a_,OI;function Rve(){if(OI)return a_;OI=1;var e=n3();function t(n){var r=e(n),i=r%1;return r===r?i?r-i:r:0}return a_=t,a_}var o_,AI;function Tve(){if(AI)return o_;AI=1;var e=vU(),t=Pi(),n=Rve(),r=Math.max;function i(a,s,u){var c=a==null?0:a.length;if(!c)return-1;var d=u==null?0:n(u);return d<0&&(d=r(c+d,0)),e(a,t(s,3),d)}return o_=i,o_}var s_,RI;function Mve(){if(RI)return s_;RI=1;var e=Ave(),t=Tve(),n=e(t);return s_=n,s_}Mve();var Pve=Pz();const jve=Rt(Pve);var Nve=jve(function(e){return{x:e.left,y:e.top,width:e.width,height:e.height}},function(e){return["l",e.left,"t",e.top,"w",e.width,"h",e.height].join("")}),u3=g.createContext(void 0),c3=g.createContext(void 0),f3=g.createContext(void 0),Dve=g.createContext({}),d3=g.createContext(void 0),kve=g.createContext(0),$ve=g.createContext(0),TI=function(t){var n=t.state,r=n.xAxisMap,i=n.yAxisMap,a=n.offset,s=t.clipPathId,u=t.children,c=t.width,d=t.height,p=Nve(a);return V.createElement(u3.Provider,{value:r},V.createElement(c3.Provider,{value:i},V.createElement(Dve.Provider,{value:a},V.createElement(f3.Provider,{value:p},V.createElement(d3.Provider,{value:s},V.createElement(kve.Provider,{value:d},V.createElement($ve.Provider,{value:c},u)))))))},Ive=function(){return g.useContext(d3)},Lve=function(t){var n=g.useContext(u3);n==null&&Vl();var r=n[t];return r==null&&Vl(),r},Bve=function(t){var n=g.useContext(c3);n==null&&Vl();var r=n[t];return r==null&&Vl(),r},Fve=function(){var t=g.useContext(f3);return t};function Jl(e){"@babel/helpers - typeof";return Jl=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},Jl(e)}function qve(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}function zve(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,h3(r.key),r)}}function Uve(e,t,n){return t&&zve(e.prototype,t),Object.defineProperty(e,"prototype",{writable:!1}),e}function Vve(e,t,n){return t=zm(t),Hve(e,p3()?Reflect.construct(t,n||[],zm(e).constructor):t.apply(e,n))}function Hve(e,t){if(t&&(Jl(t)==="object"||typeof t=="function"))return t;if(t!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return Gve(e)}function Gve(e){if(e===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return e}function p3(){try{var e=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(p3=function(){return!!e})()}function zm(e){return zm=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(n){return n.__proto__||Object.getPrototypeOf(n)},zm(e)}function Wve(e,t){if(typeof t!="function"&&t!==null)throw new TypeError("Super expression must either be null or a function");e.prototype=Object.create(t&&t.prototype,{constructor:{value:e,writable:!0,configurable:!0}}),Object.defineProperty(e,"prototype",{writable:!1}),t&&lE(e,t)}function lE(e,t){return lE=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(r,i){return r.__proto__=i,r},lE(e,t)}function MI(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function PI(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?MI(Object(n),!0).forEach(function(r){TO(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):MI(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function TO(e,t,n){return t=h3(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function h3(e){var t=Kve(e,"string");return Jl(t)=="symbol"?t:t+""}function Kve(e,t){if(Jl(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(Jl(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(e)}function Yve(e,t){return Jve(e)||Qve(e,t)||Xve(e,t)||Zve()}function Zve(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function Xve(e,t){if(e){if(typeof e=="string")return jI(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);if(n==="Object"&&e.constructor&&(n=e.constructor.name),n==="Map"||n==="Set")return Array.from(e);if(n==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return jI(e,t)}}function jI(e,t){(t==null||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}function Qve(e,t){var n=e==null?null:typeof Symbol<"u"&&e[Symbol.iterator]||e["@@iterator"];if(n!=null){var r,i,a,s,u=[],c=!0,d=!1;try{if(a=(n=n.call(e)).next,t!==0)for(;!(c=(r=a.call(n)).done)&&(u.push(r.value),u.length!==t);c=!0);}catch(p){d=!0,i=p}finally{try{if(!c&&n.return!=null&&(s=n.return(),Object(s)!==s))return}finally{if(d)throw i}}return u}}function Jve(e){if(Array.isArray(e))return e}function uE(){return uE=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)Object.prototype.hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},uE.apply(this,arguments)}var ege=function(t,n){var r;return V.isValidElement(t)?r=V.cloneElement(t,n):st(t)?r=t(n):r=V.createElement("line",uE({},n,{className:"recharts-reference-line-line"})),r},tge=function(t,n,r,i,a,s,u,c,d){var p=a.x,m=a.y,v=a.width,b=a.height;if(r){var S=d.y,w=t.y.apply(S,{position:s});if(xi(d,"discard")&&!t.y.isInRange(w))return null;var x=[{x:p+v,y:w},{x:p,y:w}];return c==="left"?x.reverse():x}if(n){var C=d.x,O=t.x.apply(C,{position:s});if(xi(d,"discard")&&!t.x.isInRange(O))return null;var A=[{x:O,y:m+b},{x:O,y:m}];return u==="top"?A.reverse():A}if(i){var T=d.segment,R=T.map(function(M){return t.apply(M,{position:s})});return xi(d,"discard")&&pve(R,function(M){return!t.isInRange(M)})?null:R}return null};function nge(e){var t=e.x,n=e.y,r=e.segment,i=e.xAxisId,a=e.yAxisId,s=e.shape,u=e.className,c=e.alwaysShow,d=Ive(),p=Lve(i),m=Bve(a),v=Fve();if(!d||!v)return null;es(c===void 0,'The alwaysShow prop is deprecated. Please use ifOverflow="extendDomain" instead.');var b=RO({x:p.scale,y:m.scale}),S=sn(t),w=sn(n),x=r&&r.length===2,C=tge(b,S,w,x,v,e.position,p.orientation,m.orientation,e);if(!C)return null;var O=Yve(C,2),A=O[0],T=A.x,R=A.y,M=O[1],P=M.x,N=M.y,k=xi(e,"hidden")?"url(#".concat(d,")"):void 0,H=PI(PI({clipPath:k},Je(e,!0)),{},{x1:T,y1:R,x2:P,y2:N});return V.createElement(un,{className:ct("recharts-reference-line",u)},ege(s,H),Tn.renderCallByParent(e,Ove({x1:T,y1:R,x2:P,y2:N})))}var MO=function(e){function t(){return qve(this,t),Vve(this,t,arguments)}return Wve(t,e),Uve(t,[{key:"render",value:function(){return V.createElement(nge,this.props)}}])}(V.Component);TO(MO,"displayName","ReferenceLine");TO(MO,"defaultProps",{isFront:!1,ifOverflow:"discard",xAxisId:0,yAxisId:0,fill:"none",stroke:"#ccc",fillOpacity:1,strokeWidth:1,position:"middle"});function cE(){return cE=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)Object.prototype.hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},cE.apply(this,arguments)}function eu(e){"@babel/helpers - typeof";return eu=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},eu(e)}function NI(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function DI(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?NI(Object(n),!0).forEach(function(r){sg(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):NI(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function rge(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}function ige(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,v3(r.key),r)}}function age(e,t,n){return t&&ige(e.prototype,t),Object.defineProperty(e,"prototype",{writable:!1}),e}function oge(e,t,n){return t=Um(t),sge(e,m3()?Reflect.construct(t,n||[],Um(e).constructor):t.apply(e,n))}function sge(e,t){if(t&&(eu(t)==="object"||typeof t=="function"))return t;if(t!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return lge(e)}function lge(e){if(e===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return e}function m3(){try{var e=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(m3=function(){return!!e})()}function Um(e){return Um=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(n){return n.__proto__||Object.getPrototypeOf(n)},Um(e)}function uge(e,t){if(typeof t!="function"&&t!==null)throw new TypeError("Super expression must either be null or a function");e.prototype=Object.create(t&&t.prototype,{constructor:{value:e,writable:!0,configurable:!0}}),Object.defineProperty(e,"prototype",{writable:!1}),t&&fE(e,t)}function fE(e,t){return fE=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(r,i){return r.__proto__=i,r},fE(e,t)}function sg(e,t,n){return t=v3(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function v3(e){var t=cge(e,"string");return eu(t)=="symbol"?t:t+""}function cge(e,t){if(eu(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(eu(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(e)}var fge=function(t){var n=t.x,r=t.y,i=t.xAxis,a=t.yAxis,s=RO({x:i.scale,y:a.scale}),u=s.apply({x:n,y:r},{bandAware:!0});return xi(t,"discard")&&!s.isInRange(u)?null:u},lg=function(e){function t(){return rge(this,t),oge(this,t,arguments)}return uge(t,e),age(t,[{key:"render",value:function(){var r=this.props,i=r.x,a=r.y,s=r.r,u=r.alwaysShow,c=r.clipPathId,d=sn(i),p=sn(a);if(es(u===void 0,'The alwaysShow prop is deprecated. Please use ifOverflow="extendDomain" instead.'),!d||!p)return null;var m=fge(this.props);if(!m)return null;var v=m.x,b=m.y,S=this.props,w=S.shape,x=S.className,C=xi(this.props,"hidden")?"url(#".concat(c,")"):void 0,O=DI(DI({clipPath:C},Je(this.props,!0)),{},{cx:v,cy:b});return V.createElement(un,{className:ct("recharts-reference-dot",x)},t.renderDot(w,O),Tn.renderCallByParent(this.props,{x:v-s,y:b-s,width:2*s,height:2*s}))}}])}(V.Component);sg(lg,"displayName","ReferenceDot");sg(lg,"defaultProps",{isFront:!1,ifOverflow:"discard",xAxisId:0,yAxisId:0,r:10,fill:"#fff",stroke:"#ccc",fillOpacity:1,strokeWidth:1});sg(lg,"renderDot",function(e,t){var n;return V.isValidElement(e)?n=V.cloneElement(e,t):st(e)?n=e(t):n=V.createElement(tg,cE({},t,{cx:t.cx,cy:t.cy,className:"recharts-reference-dot-dot"})),n});function dE(){return dE=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)Object.prototype.hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},dE.apply(this,arguments)}function tu(e){"@babel/helpers - typeof";return tu=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},tu(e)}function kI(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function $I(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?kI(Object(n),!0).forEach(function(r){ug(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):kI(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function dge(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}function pge(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,y3(r.key),r)}}function hge(e,t,n){return t&&pge(e.prototype,t),Object.defineProperty(e,"prototype",{writable:!1}),e}function mge(e,t,n){return t=Vm(t),vge(e,g3()?Reflect.construct(t,n||[],Vm(e).constructor):t.apply(e,n))}function vge(e,t){if(t&&(tu(t)==="object"||typeof t=="function"))return t;if(t!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return gge(e)}function gge(e){if(e===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return e}function g3(){try{var e=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(g3=function(){return!!e})()}function Vm(e){return Vm=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(n){return n.__proto__||Object.getPrototypeOf(n)},Vm(e)}function yge(e,t){if(typeof t!="function"&&t!==null)throw new TypeError("Super expression must either be null or a function");e.prototype=Object.create(t&&t.prototype,{constructor:{value:e,writable:!0,configurable:!0}}),Object.defineProperty(e,"prototype",{writable:!1}),t&&pE(e,t)}function pE(e,t){return pE=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(r,i){return r.__proto__=i,r},pE(e,t)}function ug(e,t,n){return t=y3(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function y3(e){var t=bge(e,"string");return tu(t)=="symbol"?t:t+""}function bge(e,t){if(tu(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(tu(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(e)}var xge=function(t,n,r,i,a){var s=a.x1,u=a.x2,c=a.y1,d=a.y2,p=a.xAxis,m=a.yAxis;if(!p||!m)return null;var v=RO({x:p.scale,y:m.scale}),b={x:t?v.x.apply(s,{position:"start"}):v.x.rangeMin,y:r?v.y.apply(c,{position:"start"}):v.y.rangeMin},S={x:n?v.x.apply(u,{position:"end"}):v.x.rangeMax,y:i?v.y.apply(d,{position:"end"}):v.y.rangeMax};return xi(a,"discard")&&(!v.isInRange(b)||!v.isInRange(S))?null:s3(b,S)},cg=function(e){function t(){return dge(this,t),mge(this,t,arguments)}return yge(t,e),hge(t,[{key:"render",value:function(){var r=this.props,i=r.x1,a=r.x2,s=r.y1,u=r.y2,c=r.className,d=r.alwaysShow,p=r.clipPathId;es(d===void 0,'The alwaysShow prop is deprecated. Please use ifOverflow="extendDomain" instead.');var m=sn(i),v=sn(a),b=sn(s),S=sn(u),w=this.props.shape;if(!m&&!v&&!b&&!S&&!w)return null;var x=xge(m,v,b,S,this.props);if(!x&&!w)return null;var C=xi(this.props,"hidden")?"url(#".concat(p,")"):void 0;return V.createElement(un,{className:ct("recharts-reference-area",c)},t.renderRect(w,$I($I({clipPath:C},Je(this.props,!0)),x)),Tn.renderCallByParent(this.props,x))}}])}(V.Component);ug(cg,"displayName","ReferenceArea");ug(cg,"defaultProps",{isFront:!1,ifOverflow:"discard",xAxisId:0,yAxisId:0,r:10,fill:"#ccc",fillOpacity:.5,stroke:"none",strokeWidth:1});ug(cg,"renderRect",function(e,t){var n;return V.isValidElement(e)?n=V.cloneElement(e,t):st(e)?n=e(t):n=V.createElement(WV,dE({},t,{className:"recharts-reference-area-rect"})),n});function II(e){return Ege(e)||_ge(e)||Sge(e)||wge()}function wge(){throw new TypeError(`Invalid attempt to spread non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function Sge(e,t){if(e){if(typeof e=="string")return hE(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);if(n==="Object"&&e.constructor&&(n=e.constructor.name),n==="Map"||n==="Set")return Array.from(e);if(n==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return hE(e,t)}}function _ge(e){if(typeof Symbol<"u"&&e[Symbol.iterator]!=null||e["@@iterator"]!=null)return Array.from(e)}function Ege(e){if(Array.isArray(e))return hE(e)}function hE(e,t){(t==null||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}var mE=function(t,n,r,i,a){var s=la(t,MO),u=la(t,lg),c=[].concat(II(s),II(u)),d=la(t,cg),p="".concat(i,"Id"),m=i[0],v=n;if(c.length&&(v=c.reduce(function(w,x){if(x.props[p]===r&&xi(x.props,"extendDomain")&&$e(x.props[m])){var C=x.props[m];return[Math.min(w[0],C),Math.max(w[1],C)]}return w},v)),d.length){var b="".concat(m,"1"),S="".concat(m,"2");v=d.reduce(function(w,x){if(x.props[p]===r&&xi(x.props,"extendDomain")&&$e(x.props[b])&&$e(x.props[S])){var C=x.props[b],O=x.props[S];return[Math.min(w[0],C,O),Math.max(w[1],C,O)]}return w},v)}return a&&a.length&&(v=a.reduce(function(w,x){return $e(x)?[Math.min(w[0],x),Math.max(w[1],x)]:w},v)),v},l_={exports:{}},LI;function Cge(){return LI||(LI=1,function(e){var t=Object.prototype.hasOwnProperty,n="~";function r(){}Object.create&&(r.prototype=Object.create(null),new r().__proto__||(n=!1));function i(c,d,p){this.fn=c,this.context=d,this.once=p||!1}function a(c,d,p,m,v){if(typeof p!="function")throw new TypeError("The listener must be a function");var b=new i(p,m||c,v),S=n?n+d:d;return c._events[S]?c._events[S].fn?c._events[S]=[c._events[S],b]:c._events[S].push(b):(c._events[S]=b,c._eventsCount++),c}function s(c,d){--c._eventsCount===0?c._events=new r:delete c._events[d]}function u(){this._events=new r,this._eventsCount=0}u.prototype.eventNames=function(){var d=[],p,m;if(this._eventsCount===0)return d;for(m in p=this._events)t.call(p,m)&&d.push(n?m.slice(1):m);return Object.getOwnPropertySymbols?d.concat(Object.getOwnPropertySymbols(p)):d},u.prototype.listeners=function(d){var p=n?n+d:d,m=this._events[p];if(!m)return[];if(m.fn)return[m.fn];for(var v=0,b=m.length,S=new Array(b);v<b;v++)S[v]=m[v].fn;return S},u.prototype.listenerCount=function(d){var p=n?n+d:d,m=this._events[p];return m?m.fn?1:m.length:0},u.prototype.emit=function(d,p,m,v,b,S){var w=n?n+d:d;if(!this._events[w])return!1;var x=this._events[w],C=arguments.length,O,A;if(x.fn){switch(x.once&&this.removeListener(d,x.fn,void 0,!0),C){case 1:return x.fn.call(x.context),!0;case 2:return x.fn.call(x.context,p),!0;case 3:return x.fn.call(x.context,p,m),!0;case 4:return x.fn.call(x.context,p,m,v),!0;case 5:return x.fn.call(x.context,p,m,v,b),!0;case 6:return x.fn.call(x.context,p,m,v,b,S),!0}for(A=1,O=new Array(C-1);A<C;A++)O[A-1]=arguments[A];x.fn.apply(x.context,O)}else{var T=x.length,R;for(A=0;A<T;A++)switch(x[A].once&&this.removeListener(d,x[A].fn,void 0,!0),C){case 1:x[A].fn.call(x[A].context);break;case 2:x[A].fn.call(x[A].context,p);break;case 3:x[A].fn.call(x[A].context,p,m);break;case 4:x[A].fn.call(x[A].context,p,m,v);break;default:if(!O)for(R=1,O=new Array(C-1);R<C;R++)O[R-1]=arguments[R];x[A].fn.apply(x[A].context,O)}}return!0},u.prototype.on=function(d,p,m){return a(this,d,p,m,!1)},u.prototype.once=function(d,p,m){return a(this,d,p,m,!0)},u.prototype.removeListener=function(d,p,m,v){var b=n?n+d:d;if(!this._events[b])return this;if(!p)return s(this,b),this;var S=this._events[b];if(S.fn)S.fn===p&&(!v||S.once)&&(!m||S.context===m)&&s(this,b);else{for(var w=0,x=[],C=S.length;w<C;w++)(S[w].fn!==p||v&&!S[w].once||m&&S[w].context!==m)&&x.push(S[w]);x.length?this._events[b]=x.length===1?x[0]:x:s(this,b)}return this},u.prototype.removeAllListeners=function(d){var p;return d?(p=n?n+d:d,this._events[p]&&s(this,p)):(this._events=new r,this._eventsCount=0),this},u.prototype.off=u.prototype.removeListener,u.prototype.addListener=u.prototype.on,u.prefixed=n,u.EventEmitter=u,e.exports=u}(l_)),l_.exports}var Oge=Cge();const Age=Rt(Oge);var u_=new Age,c_="recharts.syncMouseEvents";function Jf(e){"@babel/helpers - typeof";return Jf=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},Jf(e)}function Rge(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}function Tge(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,b3(r.key),r)}}function Mge(e,t,n){return t&&Tge(e.prototype,t),Object.defineProperty(e,"prototype",{writable:!1}),e}function f_(e,t,n){return t=b3(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function b3(e){var t=Pge(e,"string");return Jf(t)=="symbol"?t:t+""}function Pge(e,t){if(Jf(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(Jf(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(e)}var jge=function(){function e(){Rge(this,e),f_(this,"activeIndex",0),f_(this,"coordinateList",[]),f_(this,"layout","horizontal")}return Mge(e,[{key:"setDetails",value:function(n){var r,i=n.coordinateList,a=i===void 0?null:i,s=n.container,u=s===void 0?null:s,c=n.layout,d=c===void 0?null:c,p=n.offset,m=p===void 0?null:p,v=n.mouseHandlerCallback,b=v===void 0?null:v;this.coordinateList=(r=a??this.coordinateList)!==null&&r!==void 0?r:[],this.container=u??this.container,this.layout=d??this.layout,this.offset=m??this.offset,this.mouseHandlerCallback=b??this.mouseHandlerCallback,this.activeIndex=Math.min(Math.max(this.activeIndex,0),this.coordinateList.length-1)}},{key:"focus",value:function(){this.spoofMouse()}},{key:"keyboardEvent",value:function(n){if(this.coordinateList.length!==0)switch(n.key){case"ArrowRight":{if(this.layout!=="horizontal")return;this.activeIndex=Math.min(this.activeIndex+1,this.coordinateList.length-1),this.spoofMouse();break}case"ArrowLeft":{if(this.layout!=="horizontal")return;this.activeIndex=Math.max(this.activeIndex-1,0),this.spoofMouse();break}}}},{key:"setIndex",value:function(n){this.activeIndex=n}},{key:"spoofMouse",value:function(){var n,r;if(this.layout==="horizontal"&&this.coordinateList.length!==0){var i=this.container.getBoundingClientRect(),a=i.x,s=i.y,u=i.height,c=this.coordinateList[this.activeIndex].coordinate,d=((n=window)===null||n===void 0?void 0:n.scrollX)||0,p=((r=window)===null||r===void 0?void 0:r.scrollY)||0,m=a+c+d,v=s+this.offset.top+u/2+p;this.mouseHandlerCallback({pageX:m,pageY:v})}}}])}();function Nge(e,t,n){if(n==="number"&&t===!0&&Array.isArray(e)){var r=e==null?void 0:e[0],i=e==null?void 0:e[1];if(r&&i&&$e(r)&&$e(i))return!0}return!1}function Dge(e,t,n,r){var i=r/2;return{stroke:"none",fill:"#ccc",x:e==="horizontal"?t.x-i:n.left+.5,y:e==="horizontal"?n.top+.5:t.y-i,width:e==="horizontal"?r:n.width-1,height:e==="horizontal"?n.height-1:r}}function x3(e){var t=e.cx,n=e.cy,r=e.radius,i=e.startAngle,a=e.endAngle,s=At(t,n,r,i),u=At(t,n,r,a);return{points:[s,u],cx:t,cy:n,radius:r,startAngle:i,endAngle:a}}function kge(e,t,n){var r,i,a,s;if(e==="horizontal")r=t.x,a=r,i=n.top,s=n.top+n.height;else if(e==="vertical")i=t.y,s=i,r=n.left,a=n.left+n.width;else if(t.cx!=null&&t.cy!=null)if(e==="centric"){var u=t.cx,c=t.cy,d=t.innerRadius,p=t.outerRadius,m=t.angle,v=At(u,c,d,m),b=At(u,c,p,m);r=v.x,i=v.y,a=b.x,s=b.y}else return x3(t);return[{x:r,y:i},{x:a,y:s}]}function ed(e){"@babel/helpers - typeof";return ed=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},ed(e)}function BI(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function mh(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?BI(Object(n),!0).forEach(function(r){$ge(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):BI(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function $ge(e,t,n){return t=Ige(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function Ige(e){var t=Lge(e,"string");return ed(t)=="symbol"?t:t+""}function Lge(e,t){if(ed(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(ed(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function Bge(e){var t,n,r=e.element,i=e.tooltipEventType,a=e.isActive,s=e.activeCoordinate,u=e.activePayload,c=e.offset,d=e.activeTooltipIndex,p=e.tooltipAxisBandSize,m=e.layout,v=e.chartName,b=(t=r.props.cursor)!==null&&t!==void 0?t:(n=r.type.defaultProps)===null||n===void 0?void 0:n.cursor;if(!r||!b||!a||!s||v!=="ScatterChart"&&i!=="axis")return null;var S,w=g$;if(v==="ScatterChart")S=s,w=qhe;else if(v==="BarChart")S=Dge(m,s,c,p),w=WV;else if(m==="radial"){var x=x3(s),C=x.cx,O=x.cy,A=x.radius,T=x.startAngle,R=x.endAngle;S={cx:C,cy:O,startAngle:T,endAngle:R,innerRadius:A,outerRadius:A},w=Tde}else S={points:kge(m,s,c)},w=g$;var M=mh(mh(mh(mh({stroke:"#ccc",pointerEvents:"none"},c),S),Je(b,!1)),{},{payload:u,payloadIndex:d,className:ct("recharts-tooltip-cursor",b.className)});return g.isValidElement(b)?g.cloneElement(b,M):g.createElement(w,M)}var Fge=["item"],qge=["children","className","width","height","style","compact","title","desc"];function nu(e){"@babel/helpers - typeof";return nu=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},nu(e)}function _l(){return _l=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)Object.prototype.hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},_l.apply(this,arguments)}function FI(e,t){return Vge(e)||Uge(e,t)||S3(e,t)||zge()}function zge(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function Uge(e,t){var n=e==null?null:typeof Symbol<"u"&&e[Symbol.iterator]||e["@@iterator"];if(n!=null){var r,i,a,s,u=[],c=!0,d=!1;try{if(a=(n=n.call(e)).next,t!==0)for(;!(c=(r=a.call(n)).done)&&(u.push(r.value),u.length!==t);c=!0);}catch(p){d=!0,i=p}finally{try{if(!c&&n.return!=null&&(s=n.return(),Object(s)!==s))return}finally{if(d)throw i}}return u}}function Vge(e){if(Array.isArray(e))return e}function qI(e,t){if(e==null)return{};var n=Hge(e,t),r,i;if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);for(i=0;i<a.length;i++)r=a[i],!(t.indexOf(r)>=0)&&Object.prototype.propertyIsEnumerable.call(e,r)&&(n[r]=e[r])}return n}function Hge(e,t){if(e==null)return{};var n={};for(var r in e)if(Object.prototype.hasOwnProperty.call(e,r)){if(t.indexOf(r)>=0)continue;n[r]=e[r]}return n}function Gge(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}function Wge(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,_3(r.key),r)}}function Kge(e,t,n){return t&&Wge(e.prototype,t),Object.defineProperty(e,"prototype",{writable:!1}),e}function Yge(e,t,n){return t=Hm(t),Zge(e,w3()?Reflect.construct(t,n||[],Hm(e).constructor):t.apply(e,n))}function Zge(e,t){if(t&&(nu(t)==="object"||typeof t=="function"))return t;if(t!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return Xge(e)}function Xge(e){if(e===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return e}function w3(){try{var e=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(w3=function(){return!!e})()}function Hm(e){return Hm=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(n){return n.__proto__||Object.getPrototypeOf(n)},Hm(e)}function Qge(e,t){if(typeof t!="function"&&t!==null)throw new TypeError("Super expression must either be null or a function");e.prototype=Object.create(t&&t.prototype,{constructor:{value:e,writable:!0,configurable:!0}}),Object.defineProperty(e,"prototype",{writable:!1}),t&&vE(e,t)}function vE(e,t){return vE=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(r,i){return r.__proto__=i,r},vE(e,t)}function ru(e){return tye(e)||eye(e)||S3(e)||Jge()}function Jge(){throw new TypeError(`Invalid attempt to spread non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function S3(e,t){if(e){if(typeof e=="string")return gE(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);if(n==="Object"&&e.constructor&&(n=e.constructor.name),n==="Map"||n==="Set")return Array.from(e);if(n==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return gE(e,t)}}function eye(e){if(typeof Symbol<"u"&&e[Symbol.iterator]!=null||e["@@iterator"]!=null)return Array.from(e)}function tye(e){if(Array.isArray(e))return gE(e)}function gE(e,t){(t==null||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}function zI(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(i){return Object.getOwnPropertyDescriptor(e,i).enumerable})),n.push.apply(n,r)}return n}function pe(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?zI(Object(n),!0).forEach(function(r){ze(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):zI(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function ze(e,t,n){return t=_3(t),t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function _3(e){var t=nye(e,"string");return nu(t)=="symbol"?t:t+""}function nye(e,t){if(nu(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(nu(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}var rye={xAxis:["bottom","top"],yAxis:["left","right"]},iye={width:"100%",height:"100%"},E3={x:0,y:0};function vh(e){return e}var aye=function(t,n){return n==="horizontal"?t.x:n==="vertical"?t.y:n==="centric"?t.angle:t.radius},oye=function(t,n,r,i){var a=n.find(function(p){return p&&p.index===r});if(a){if(t==="horizontal")return{x:a.coordinate,y:i.y};if(t==="vertical")return{x:i.x,y:a.coordinate};if(t==="centric"){var s=a.coordinate,u=i.radius;return pe(pe(pe({},i),At(i.cx,i.cy,u,s)),{},{angle:s,radius:u})}var c=a.coordinate,d=i.angle;return pe(pe(pe({},i),At(i.cx,i.cy,c,d)),{},{angle:d,radius:c})}return E3},fg=function(t,n){var r=n.graphicalItems,i=n.dataStartIndex,a=n.dataEndIndex,s=(r??[]).reduce(function(u,c){var d=c.props.data;return d&&d.length?[].concat(ru(u),ru(d)):u},[]);return s.length>0?s:t&&t.length&&$e(i)&&$e(a)?t.slice(i,a+1):[]};function C3(e){return e==="number"?[0,"auto"]:void 0}var yE=function(t,n,r,i){var a=t.graphicalItems,s=t.tooltipAxis,u=fg(n,t);return r<0||!a||!a.length||r>=u.length?null:a.reduce(function(c,d){var p,m=(p=d.props.data)!==null&&p!==void 0?p:n;m&&t.dataStartIndex+t.dataEndIndex!==0&&t.dataEndIndex-t.dataStartIndex>=r&&(m=m.slice(t.dataStartIndex,t.dataEndIndex+1));var v;if(s.dataKey&&!s.allowDuplicatedCategory){var b=m===void 0?u:m;v=a1(b,s.dataKey,i)}else v=m&&m[r]||u[r];return v?[].concat(ru(c),[Afe(d,v)]):c},[])},UI=function(t,n,r,i){var a=i||{x:t.chartX,y:t.chartY},s=aye(a,r),u=t.orderedTooltipTicks,c=t.tooltipAxis,d=t.tooltipTicks,p=ffe(s,u,d,c);if(p>=0&&d){var m=d[p]&&d[p].value,v=yE(t,n,p,m),b=oye(r,u,p,a);return{activeTooltipIndex:p,activeLabel:m,activePayload:v,activeCoordinate:b}}return null},sye=function(t,n){var r=n.axes,i=n.graphicalItems,a=n.axisType,s=n.axisIdKey,u=n.stackGroups,c=n.dataStartIndex,d=n.dataEndIndex,p=t.layout,m=t.children,v=t.stackOffset,b=NV(p,a);return r.reduce(function(S,w){var x,C=w.type.defaultProps!==void 0?pe(pe({},w.type.defaultProps),w.props):w.props,O=C.type,A=C.dataKey,T=C.allowDataOverflow,R=C.allowDuplicatedCategory,M=C.scale,P=C.ticks,N=C.includeHidden,k=C[s];if(S[k])return S;var H=fg(t.data,{graphicalItems:i.filter(function(q){var X,oe=s in q.props?q.props[s]:(X=q.type.defaultProps)===null||X===void 0?void 0:X[s];return oe===k}),dataStartIndex:c,dataEndIndex:d}),L=H.length,F,z,Z;Nge(C.domain,T,O)&&(F=F1(C.domain,null,T),b&&(O==="number"||M!=="auto")&&(Z=Yc(H,A,"category")));var Y=C3(O);if(!F||F.length===0){var I,W=(I=C.domain)!==null&&I!==void 0?I:Y;if(A){if(F=Yc(H,A,O),O==="category"&&b){var B=are(F);R&&B?(z=F,F=Bm(0,L)):R||(F=r$(W,F,w).reduce(function(q,X){return q.indexOf(X)>=0?q:[].concat(ru(q),[X])},[]))}else if(O==="category")R?F=F.filter(function(q){return q!==""&&!vt(q)}):F=r$(W,F,w).reduce(function(q,X){return q.indexOf(X)>=0||X===""||vt(X)?q:[].concat(ru(q),[X])},[]);else if(O==="number"){var Q=vfe(H,i.filter(function(q){var X,oe,le=s in q.props?q.props[s]:(X=q.type.defaultProps)===null||X===void 0?void 0:X[s],ve="hide"in q.props?q.props.hide:(oe=q.type.defaultProps)===null||oe===void 0?void 0:oe.hide;return le===k&&(N||!ve)}),A,a,p);Q&&(F=Q)}b&&(O==="number"||M!=="auto")&&(Z=Yc(H,A,"category"))}else b?F=Bm(0,L):u&&u[k]&&u[k].hasStack&&O==="number"?F=v==="expand"?[0,1]:DV(u[k].stackGroups,c,d):F=jV(H,i.filter(function(q){var X=s in q.props?q.props[s]:q.type.defaultProps[s],oe="hide"in q.props?q.props.hide:q.type.defaultProps.hide;return X===k&&(N||!oe)}),O,p,!0);if(O==="number")F=mE(m,F,k,a,P),W&&(F=F1(W,F,T));else if(O==="category"&&W){var $=W,J=F.every(function(q){return $.indexOf(q)>=0});J&&(F=$)}}return pe(pe({},S),{},ze({},k,pe(pe({},C),{},{axisType:a,domain:F,categoricalDomain:Z,duplicateDomain:z,originalDomain:(x=C.domain)!==null&&x!==void 0?x:Y,isCategorical:b,layout:p})))},{})},lye=function(t,n){var r=n.graphicalItems,i=n.Axis,a=n.axisType,s=n.axisIdKey,u=n.stackGroups,c=n.dataStartIndex,d=n.dataEndIndex,p=t.layout,m=t.children,v=fg(t.data,{graphicalItems:r,dataStartIndex:c,dataEndIndex:d}),b=v.length,S=NV(p,a),w=-1;return r.reduce(function(x,C){var O=C.type.defaultProps!==void 0?pe(pe({},C.type.defaultProps),C.props):C.props,A=O[s],T=C3("number");if(!x[A]){w++;var R;return S?R=Bm(0,b):u&&u[A]&&u[A].hasStack?(R=DV(u[A].stackGroups,c,d),R=mE(m,R,A,a)):(R=F1(T,jV(v,r.filter(function(M){var P,N,k=s in M.props?M.props[s]:(P=M.type.defaultProps)===null||P===void 0?void 0:P[s],H="hide"in M.props?M.props.hide:(N=M.type.defaultProps)===null||N===void 0?void 0:N.hide;return k===A&&!H}),"number",p),i.defaultProps.allowDataOverflow),R=mE(m,R,A,a)),pe(pe({},x),{},ze({},A,pe(pe({axisType:a},i.defaultProps),{},{hide:!0,orientation:yi(rye,"".concat(a,".").concat(w%2),null),domain:R,originalDomain:T,isCategorical:S,layout:p})))}return x},{})},uye=function(t,n){var r=n.axisType,i=r===void 0?"xAxis":r,a=n.AxisComp,s=n.graphicalItems,u=n.stackGroups,c=n.dataStartIndex,d=n.dataEndIndex,p=t.children,m="".concat(i,"Id"),v=la(p,a),b={};return v&&v.length?b=sye(t,{axes:v,graphicalItems:s,axisType:i,axisIdKey:m,stackGroups:u,dataStartIndex:c,dataEndIndex:d}):s&&s.length&&(b=lye(t,{Axis:a,graphicalItems:s,axisType:i,axisIdKey:m,stackGroups:u,dataStartIndex:c,dataEndIndex:d})),b},cye=function(t){var n=bl(t),r=zc(n,!1,!0);return{tooltipTicks:r,orderedTooltipTicks:QC(r,function(i){return i.coordinate}),tooltipAxis:n,tooltipAxisBandSize:q1(n,r)}},VI=function(t){var n=t.children,r=t.defaultShowTooltip,i=Tr(n,Ql),a=0,s=0;return t.data&&t.data.length!==0&&(s=t.data.length-1),i&&i.props&&(i.props.startIndex>=0&&(a=i.props.startIndex),i.props.endIndex>=0&&(s=i.props.endIndex)),{chartX:0,chartY:0,dataStartIndex:a,dataEndIndex:s,activeTooltipIndex:-1,isTooltipActive:!!r}},fye=function(t){return!t||!t.length?!1:t.some(function(n){var r=sa(n&&n.type);return r&&r.indexOf("Bar")>=0})},HI=function(t){return t==="horizontal"?{numericAxisName:"yAxis",cateAxisName:"xAxis"}:t==="vertical"?{numericAxisName:"xAxis",cateAxisName:"yAxis"}:t==="centric"?{numericAxisName:"radiusAxis",cateAxisName:"angleAxis"}:{numericAxisName:"angleAxis",cateAxisName:"radiusAxis"}},dye=function(t,n){var r=t.props,i=t.graphicalItems,a=t.xAxisMap,s=a===void 0?{}:a,u=t.yAxisMap,c=u===void 0?{}:u,d=r.width,p=r.height,m=r.children,v=r.margin||{},b=Tr(m,Ql),S=Tr(m,Al),w=Object.keys(c).reduce(function(R,M){var P=c[M],N=P.orientation;return!P.mirror&&!P.hide?pe(pe({},R),{},ze({},N,R[N]+P.width)):R},{left:v.left||0,right:v.right||0}),x=Object.keys(s).reduce(function(R,M){var P=s[M],N=P.orientation;return!P.mirror&&!P.hide?pe(pe({},R),{},ze({},N,yi(R,"".concat(N))+P.height)):R},{top:v.top||0,bottom:v.bottom||0}),C=pe(pe({},x),w),O=C.bottom;b&&(C.bottom+=b.props.height||Ql.defaultProps.height),S&&n&&(C=hfe(C,i,r,n));var A=d-C.left-C.right,T=p-C.top-C.bottom;return pe(pe({brushBottom:O},C),{},{width:Math.max(A,0),height:Math.max(T,0)})},pye=function(t,n){if(n==="xAxis")return t[n].width;if(n==="yAxis")return t[n].height},hye=function(t){var n=t.chartName,r=t.GraphicalChild,i=t.defaultTooltipEventType,a=i===void 0?"axis":i,s=t.validateTooltipEventTypes,u=s===void 0?["axis"]:s,c=t.axisComponents,d=t.legendContent,p=t.formatAxisMap,m=t.defaultProps,v=function(C,O){var A=O.graphicalItems,T=O.stackGroups,R=O.offset,M=O.updateId,P=O.dataStartIndex,N=O.dataEndIndex,k=C.barSize,H=C.layout,L=C.barGap,F=C.barCategoryGap,z=C.maxBarSize,Z=HI(H),Y=Z.numericAxisName,I=Z.cateAxisName,W=fye(A),B=[];return A.forEach(function(Q,$){var J=fg(C.data,{graphicalItems:[Q],dataStartIndex:P,dataEndIndex:N}),q=Q.type.defaultProps!==void 0?pe(pe({},Q.type.defaultProps),Q.props):Q.props,X=q.dataKey,oe=q.maxBarSize,le=q["".concat(Y,"Id")],ve=q["".concat(I,"Id")],de={},Se=c.reduce(function(Bt,G){var re=O["".concat(G.axisType,"Map")],fe=q["".concat(G.axisType,"Id")];re&&re[fe]||G.axisType==="zAxis"||Vl();var Ae=re[fe];return pe(pe({},Bt),{},ze(ze({},G.axisType,Ae),"".concat(G.axisType,"Ticks"),zc(Ae)))},de),ce=Se[I],we=Se["".concat(I,"Ticks")],_e=T&&T[le]&&T[le].hasStack&&Cfe(Q,T[le].stackGroups),se=sa(Q.type).indexOf("Bar")>=0,Ye=q1(ce,we),Be=[],dt=W&&dfe({barSize:k,stackGroups:T,totalSize:pye(Se,I)});if(se){var ft,it,xt=vt(oe)?z:oe,Tt=(ft=(it=q1(ce,we,!0))!==null&&it!==void 0?it:xt)!==null&&ft!==void 0?ft:0;Be=pfe({barGap:L,barCategoryGap:F,bandSize:Tt!==Ye?Tt:Ye,sizeList:dt[ve],maxBarSize:xt}),Tt!==Ye&&(Be=Be.map(function(Bt){return pe(pe({},Bt),{},{position:pe(pe({},Bt.position),{},{offset:Bt.position.offset-Tt/2})})}))}var bn=Q&&Q.type&&Q.type.getComposedData;bn&&B.push({props:pe(pe({},bn(pe(pe({},Se),{},{displayedData:J,props:C,dataKey:X,item:Q,bandSize:Ye,barPosition:Be,offset:R,stackedData:_e,layout:H,dataStartIndex:P,dataEndIndex:N}))),{},ze(ze(ze({key:Q.key||"item-".concat($)},Y,Se[Y]),I,Se[I]),"animationId",M)),childIndex:vre(Q,C.children),item:Q})}),B},b=function(C,O){var A=C.props,T=C.dataStartIndex,R=C.dataEndIndex,M=C.updateId;if(!yN({props:A}))return null;var P=A.children,N=A.layout,k=A.stackOffset,H=A.data,L=A.reverseStackOrder,F=HI(N),z=F.numericAxisName,Z=F.cateAxisName,Y=la(P,r),I=_fe(H,Y,"".concat(z,"Id"),"".concat(Z,"Id"),k,L),W=c.reduce(function(q,X){var oe="".concat(X.axisType,"Map");return pe(pe({},q),{},ze({},oe,uye(A,pe(pe({},X),{},{graphicalItems:Y,stackGroups:X.axisType===z&&I,dataStartIndex:T,dataEndIndex:R}))))},{}),B=dye(pe(pe({},W),{},{props:A,graphicalItems:Y}),O==null?void 0:O.legendBBox);Object.keys(W).forEach(function(q){W[q]=p(A,W[q],B,q.replace("Map",""),n)});var Q=W["".concat(Z,"Map")],$=cye(Q),J=v(A,pe(pe({},W),{},{dataStartIndex:T,dataEndIndex:R,updateId:M,graphicalItems:Y,stackGroups:I,offset:B}));return pe(pe({formattedGraphicalItems:J,graphicalItems:Y,offset:B,stackGroups:I},$),W)},S=function(x){function C(O){var A,T,R;return Gge(this,C),R=Yge(this,C,[O]),ze(R,"eventEmitterSymbol",Symbol("rechartsEventEmitter")),ze(R,"accessibilityManager",new jge),ze(R,"handleLegendBBoxUpdate",function(M){if(M){var P=R.state,N=P.dataStartIndex,k=P.dataEndIndex,H=P.updateId;R.setState(pe({legendBBox:M},b({props:R.props,dataStartIndex:N,dataEndIndex:k,updateId:H},pe(pe({},R.state),{},{legendBBox:M}))))}}),ze(R,"handleReceiveSyncEvent",function(M,P,N){if(R.props.syncId===M){if(N===R.eventEmitterSymbol&&typeof R.props.syncMethod!="function")return;R.applySyncEvent(P)}}),ze(R,"handleBrushChange",function(M){var P=M.startIndex,N=M.endIndex;if(P!==R.state.dataStartIndex||N!==R.state.dataEndIndex){var k=R.state.updateId;R.setState(function(){return pe({dataStartIndex:P,dataEndIndex:N},b({props:R.props,dataStartIndex:P,dataEndIndex:N,updateId:k},R.state))}),R.triggerSyncEvent({dataStartIndex:P,dataEndIndex:N})}}),ze(R,"handleMouseEnter",function(M){var P=R.getMouseInfo(M);if(P){var N=pe(pe({},P),{},{isTooltipActive:!0});R.setState(N),R.triggerSyncEvent(N);var k=R.props.onMouseEnter;st(k)&&k(N,M)}}),ze(R,"triggeredAfterMouseMove",function(M){var P=R.getMouseInfo(M),N=P?pe(pe({},P),{},{isTooltipActive:!0}):{isTooltipActive:!1};R.setState(N),R.triggerSyncEvent(N);var k=R.props.onMouseMove;st(k)&&k(N,M)}),ze(R,"handleItemMouseEnter",function(M){R.setState(function(){return{isTooltipActive:!0,activeItem:M,activePayload:M.tooltipPayload,activeCoordinate:M.tooltipPosition||{x:M.cx,y:M.cy}}})}),ze(R,"handleItemMouseLeave",function(){R.setState(function(){return{isTooltipActive:!1}})}),ze(R,"handleMouseMove",function(M){M.persist(),R.throttleTriggeredAfterMouseMove(M)}),ze(R,"handleMouseLeave",function(M){R.throttleTriggeredAfterMouseMove.cancel();var P={isTooltipActive:!1};R.setState(P),R.triggerSyncEvent(P);var N=R.props.onMouseLeave;st(N)&&N(P,M)}),ze(R,"handleOuterEvent",function(M){var P=mre(M),N=yi(R.props,"".concat(P));if(P&&st(N)){var k,H;/.*touch.*/i.test(P)?H=R.getMouseInfo(M.changedTouches[0]):H=R.getMouseInfo(M),N((k=H)!==null&&k!==void 0?k:{},M)}}),ze(R,"handleClick",function(M){var P=R.getMouseInfo(M);if(P){var N=pe(pe({},P),{},{isTooltipActive:!0});R.setState(N),R.triggerSyncEvent(N);var k=R.props.onClick;st(k)&&k(N,M)}}),ze(R,"handleMouseDown",function(M){var P=R.props.onMouseDown;if(st(P)){var N=R.getMouseInfo(M);P(N,M)}}),ze(R,"handleMouseUp",function(M){var P=R.props.onMouseUp;if(st(P)){var N=R.getMouseInfo(M);P(N,M)}}),ze(R,"handleTouchMove",function(M){M.changedTouches!=null&&M.changedTouches.length>0&&R.throttleTriggeredAfterMouseMove(M.changedTouches[0])}),ze(R,"handleTouchStart",function(M){M.changedTouches!=null&&M.changedTouches.length>0&&R.handleMouseDown(M.changedTouches[0])}),ze(R,"handleTouchEnd",function(M){M.changedTouches!=null&&M.changedTouches.length>0&&R.handleMouseUp(M.changedTouches[0])}),ze(R,"handleDoubleClick",function(M){var P=R.props.onDoubleClick;if(st(P)){var N=R.getMouseInfo(M);P(N,M)}}),ze(R,"handleContextMenu",function(M){var P=R.props.onContextMenu;if(st(P)){var N=R.getMouseInfo(M);P(N,M)}}),ze(R,"triggerSyncEvent",function(M){R.props.syncId!==void 0&&u_.emit(c_,R.props.syncId,M,R.eventEmitterSymbol)}),ze(R,"applySyncEvent",function(M){var P=R.props,N=P.layout,k=P.syncMethod,H=R.state.updateId,L=M.dataStartIndex,F=M.dataEndIndex;if(M.dataStartIndex!==void 0||M.dataEndIndex!==void 0)R.setState(pe({dataStartIndex:L,dataEndIndex:F},b({props:R.props,dataStartIndex:L,dataEndIndex:F,updateId:H},R.state)));else if(M.activeTooltipIndex!==void 0){var z=M.chartX,Z=M.chartY,Y=M.activeTooltipIndex,I=R.state,W=I.offset,B=I.tooltipTicks;if(!W)return;if(typeof k=="function")Y=k(B,M);else if(k==="value"){Y=-1;for(var Q=0;Q<B.length;Q++)if(B[Q].value===M.activeLabel){Y=Q;break}}var $=pe(pe({},W),{},{x:W.left,y:W.top}),J=Math.min(z,$.x+$.width),q=Math.min(Z,$.y+$.height),X=B[Y]&&B[Y].value,oe=yE(R.state,R.props.data,Y),le=B[Y]?{x:N==="horizontal"?B[Y].coordinate:J,y:N==="horizontal"?q:B[Y].coordinate}:E3;R.setState(pe(pe({},M),{},{activeLabel:X,activeCoordinate:le,activePayload:oe,activeTooltipIndex:Y}))}else R.setState(M)}),ze(R,"renderCursor",function(M){var P,N=R.state,k=N.isTooltipActive,H=N.activeCoordinate,L=N.activePayload,F=N.offset,z=N.activeTooltipIndex,Z=N.tooltipAxisBandSize,Y=R.getTooltipEventType(),I=(P=M.props.active)!==null&&P!==void 0?P:k,W=R.props.layout,B=M.key||"_recharts-cursor";return V.createElement(Bge,{key:B,activeCoordinate:H,activePayload:L,activeTooltipIndex:z,chartName:n,element:M,isActive:I,layout:W,offset:F,tooltipAxisBandSize:Z,tooltipEventType:Y})}),ze(R,"renderPolarAxis",function(M,P,N){var k=yi(M,"type.axisType"),H=yi(R.state,"".concat(k,"Map")),L=M.type.defaultProps,F=L!==void 0?pe(pe({},L),M.props):M.props,z=H&&H[F["".concat(k,"Id")]];return g.cloneElement(M,pe(pe({},z),{},{className:ct(k,z.className),key:M.key||"".concat(P,"-").concat(N),ticks:zc(z,!0)}))}),ze(R,"renderPolarGrid",function(M){var P=M.props,N=P.radialLines,k=P.polarAngles,H=P.polarRadius,L=R.state,F=L.radiusAxisMap,z=L.angleAxisMap,Z=bl(F),Y=bl(z),I=Y.cx,W=Y.cy,B=Y.innerRadius,Q=Y.outerRadius;return g.cloneElement(M,{polarAngles:Array.isArray(k)?k:zc(Y,!0).map(function($){return $.coordinate}),polarRadius:Array.isArray(H)?H:zc(Z,!0).map(function($){return $.coordinate}),cx:I,cy:W,innerRadius:B,outerRadius:Q,key:M.key||"polar-grid",radialLines:N})}),ze(R,"renderLegend",function(){var M=R.state.formattedGraphicalItems,P=R.props,N=P.children,k=P.width,H=P.height,L=R.props.margin||{},F=k-(L.left||0)-(L.right||0),z=MV({children:N,formattedGraphicalItems:M,legendWidth:F,legendContent:d});if(!z)return null;var Z=z.item,Y=qI(z,Fge);return g.cloneElement(Z,pe(pe({},Y),{},{chartWidth:k,chartHeight:H,margin:L,onBBoxUpdate:R.handleLegendBBoxUpdate}))}),ze(R,"renderTooltip",function(){var M,P=R.props,N=P.children,k=P.accessibilityLayer,H=Tr(N,fi);if(!H)return null;var L=R.state,F=L.isTooltipActive,z=L.activeCoordinate,Z=L.activePayload,Y=L.activeLabel,I=L.offset,W=(M=H.props.active)!==null&&M!==void 0?M:F;return g.cloneElement(H,{viewBox:pe(pe({},I),{},{x:I.left,y:I.top}),active:W,label:Y,payload:W?Z:[],coordinate:z,accessibilityLayer:k})}),ze(R,"renderBrush",function(M){var P=R.props,N=P.margin,k=P.data,H=R.state,L=H.offset,F=H.dataStartIndex,z=H.dataEndIndex,Z=H.updateId;return g.cloneElement(M,{key:M.key||"_recharts-brush",onChange:ch(R.handleBrushChange,M.props.onChange),data:k,x:$e(M.props.x)?M.props.x:L.left,y:$e(M.props.y)?M.props.y:L.top+L.height+L.brushBottom-(N.bottom||0),width:$e(M.props.width)?M.props.width:L.width,startIndex:F,endIndex:z,updateId:"brush-".concat(Z)})}),ze(R,"renderReferenceElement",function(M,P,N){if(!M)return null;var k=R,H=k.clipPathId,L=R.state,F=L.xAxisMap,z=L.yAxisMap,Z=L.offset,Y=M.type.defaultProps||{},I=M.props,W=I.xAxisId,B=W===void 0?Y.xAxisId:W,Q=I.yAxisId,$=Q===void 0?Y.yAxisId:Q;return g.cloneElement(M,{key:M.key||"".concat(P,"-").concat(N),xAxis:F[B],yAxis:z[$],viewBox:{x:Z.left,y:Z.top,width:Z.width,height:Z.height},clipPathId:H})}),ze(R,"renderActivePoints",function(M){var P=M.item,N=M.activePoint,k=M.basePoint,H=M.childIndex,L=M.isRange,F=[],z=P.props.key,Z=P.item.type.defaultProps!==void 0?pe(pe({},P.item.type.defaultProps),P.item.props):P.item.props,Y=Z.activeDot,I=Z.dataKey,W=pe(pe({index:H,dataKey:I,cx:N.x,cy:N.y,r:4,fill:OO(P.item),strokeWidth:2,stroke:"#fff",payload:N.payload,value:N.value},Je(Y,!1)),am(Y));return F.push(C.renderActiveDot(Y,W,"".concat(z,"-activePoint-").concat(H))),k?F.push(C.renderActiveDot(Y,pe(pe({},W),{},{cx:k.x,cy:k.y}),"".concat(z,"-basePoint-").concat(H))):L&&F.push(null),F}),ze(R,"renderGraphicChild",function(M,P,N){var k=R.filterFormatItem(M,P,N);if(!k)return null;var H=R.getTooltipEventType(),L=R.state,F=L.isTooltipActive,z=L.tooltipAxis,Z=L.activeTooltipIndex,Y=L.activeLabel,I=R.props.children,W=Tr(I,fi),B=k.props,Q=B.points,$=B.isRange,J=B.baseLine,q=k.item.type.defaultProps!==void 0?pe(pe({},k.item.type.defaultProps),k.item.props):k.item.props,X=q.activeDot,oe=q.hide,le=q.activeBar,ve=q.activeShape,de=!!(!oe&&F&&W&&(X||le||ve)),Se={};H!=="axis"&&W&&W.props.trigger==="click"?Se={onClick:ch(R.handleItemMouseEnter,M.props.onClick)}:H!=="axis"&&(Se={onMouseLeave:ch(R.handleItemMouseLeave,M.props.onMouseLeave),onMouseEnter:ch(R.handleItemMouseEnter,M.props.onMouseEnter)});var ce=g.cloneElement(M,pe(pe({},k.props),Se));function we(G){return typeof z.dataKey=="function"?z.dataKey(G.payload):null}if(de)if(Z>=0){var _e,se;if(z.dataKey&&!z.allowDuplicatedCategory){var Ye=typeof z.dataKey=="function"?we:"payload.".concat(z.dataKey.toString());_e=a1(Q,Ye,Y),se=$&&J&&a1(J,Ye,Y)}else _e=Q==null?void 0:Q[Z],se=$&&J&&J[Z];if(ve||le){var Be=M.props.activeIndex!==void 0?M.props.activeIndex:Z;return[g.cloneElement(M,pe(pe(pe({},k.props),Se),{},{activeIndex:Be})),null,null]}if(!vt(_e))return[ce].concat(ru(R.renderActivePoints({item:k,activePoint:_e,basePoint:se,childIndex:Z,isRange:$})))}else{var dt,ft=(dt=R.getItemByXY(R.state.activeCoordinate))!==null&&dt!==void 0?dt:{graphicalItem:ce},it=ft.graphicalItem,xt=it.item,Tt=xt===void 0?M:xt,bn=it.childIndex,Bt=pe(pe(pe({},k.props),Se),{},{activeIndex:bn});return[g.cloneElement(Tt,Bt),null,null]}return $?[ce,null,null]:[ce,null]}),ze(R,"renderCustomized",function(M,P,N){return g.cloneElement(M,pe(pe({key:"recharts-customized-".concat(N)},R.props),R.state))}),ze(R,"renderMap",{CartesianGrid:{handler:vh,once:!0},ReferenceArea:{handler:R.renderReferenceElement},ReferenceLine:{handler:vh},ReferenceDot:{handler:R.renderReferenceElement},XAxis:{handler:vh},YAxis:{handler:vh},Brush:{handler:R.renderBrush,once:!0},Bar:{handler:R.renderGraphicChild},Line:{handler:R.renderGraphicChild},Area:{handler:R.renderGraphicChild},Radar:{handler:R.renderGraphicChild},RadialBar:{handler:R.renderGraphicChild},Scatter:{handler:R.renderGraphicChild},Pie:{handler:R.renderGraphicChild},Funnel:{handler:R.renderGraphicChild},Tooltip:{handler:R.renderCursor,once:!0},PolarGrid:{handler:R.renderPolarGrid,once:!0},PolarAngleAxis:{handler:R.renderPolarAxis},PolarRadiusAxis:{handler:R.renderPolarAxis},Customized:{handler:R.renderCustomized}}),R.clipPathId="".concat((A=O.id)!==null&&A!==void 0?A:LC("recharts"),"-clip"),R.throttleTriggeredAfterMouseMove=TU(R.triggeredAfterMouseMove,(T=O.throttleDelay)!==null&&T!==void 0?T:1e3/60),R.state={},R}return Qge(C,x),Kge(C,[{key:"componentDidMount",value:function(){var A,T;this.addListener(),this.accessibilityManager.setDetails({container:this.container,offset:{left:(A=this.props.margin.left)!==null&&A!==void 0?A:0,top:(T=this.props.margin.top)!==null&&T!==void 0?T:0},coordinateList:this.state.tooltipTicks,mouseHandlerCallback:this.triggeredAfterMouseMove,layout:this.props.layout}),this.displayDefaultTooltip()}},{key:"displayDefaultTooltip",value:function(){var A=this.props,T=A.children,R=A.data,M=A.height,P=A.layout,N=Tr(T,fi);if(N){var k=N.props.defaultIndex;if(!(typeof k!="number"||k<0||k>this.state.tooltipTicks.length-1)){var H=this.state.tooltipTicks[k]&&this.state.tooltipTicks[k].value,L=yE(this.state,R,k,H),F=this.state.tooltipTicks[k].coordinate,z=(this.state.offset.top+M)/2,Z=P==="horizontal",Y=Z?{x:F,y:z}:{y:F,x:z},I=this.state.formattedGraphicalItems.find(function(B){var Q=B.item;return Q.type.name==="Scatter"});I&&(Y=pe(pe({},Y),I.props.points[k].tooltipPosition),L=I.props.points[k].tooltipPayload);var W={activeTooltipIndex:k,isTooltipActive:!0,activeLabel:H,activePayload:L,activeCoordinate:Y};this.setState(W),this.renderCursor(N),this.accessibilityManager.setIndex(k)}}}},{key:"getSnapshotBeforeUpdate",value:function(A,T){if(!this.props.accessibilityLayer)return null;if(this.state.tooltipTicks!==T.tooltipTicks&&this.accessibilityManager.setDetails({coordinateList:this.state.tooltipTicks}),this.props.layout!==A.layout&&this.accessibilityManager.setDetails({layout:this.props.layout}),this.props.margin!==A.margin){var R,M;this.accessibilityManager.setDetails({offset:{left:(R=this.props.margin.left)!==null&&R!==void 0?R:0,top:(M=this.props.margin.top)!==null&&M!==void 0?M:0}})}return null}},{key:"componentDidUpdate",value:function(A){l1([Tr(A.children,fi)],[Tr(this.props.children,fi)])||this.displayDefaultTooltip()}},{key:"componentWillUnmount",value:function(){this.removeListener(),this.throttleTriggeredAfterMouseMove.cancel()}},{key:"getTooltipEventType",value:function(){var A=Tr(this.props.children,fi);if(A&&typeof A.props.shared=="boolean"){var T=A.props.shared?"axis":"item";return u.indexOf(T)>=0?T:a}return a}},{key:"getMouseInfo",value:function(A){if(!this.container)return null;var T=this.container,R=T.getBoundingClientRect(),M=rse(R),P={chartX:Math.round(A.pageX-M.left),chartY:Math.round(A.pageY-M.top)},N=R.width/T.offsetWidth||1,k=this.inRange(P.chartX,P.chartY,N);if(!k)return null;var H=this.state,L=H.xAxisMap,F=H.yAxisMap,z=this.getTooltipEventType(),Z=UI(this.state,this.props.data,this.props.layout,k);if(z!=="axis"&&L&&F){var Y=bl(L).scale,I=bl(F).scale,W=Y&&Y.invert?Y.invert(P.chartX):null,B=I&&I.invert?I.invert(P.chartY):null;return pe(pe({},P),{},{xValue:W,yValue:B},Z)}return Z?pe(pe({},P),Z):null}},{key:"inRange",value:function(A,T){var R=arguments.length>2&&arguments[2]!==void 0?arguments[2]:1,M=this.props.layout,P=A/R,N=T/R;if(M==="horizontal"||M==="vertical"){var k=this.state.offset,H=P>=k.left&&P<=k.left+k.width&&N>=k.top&&N<=k.top+k.height;return H?{x:P,y:N}:null}var L=this.state,F=L.angleAxisMap,z=L.radiusAxisMap;if(F&&z){var Z=bl(F);return o$({x:P,y:N},Z)}return null}},{key:"parseEventsOfWrapper",value:function(){var A=this.props.children,T=this.getTooltipEventType(),R=Tr(A,fi),M={};R&&T==="axis"&&(R.props.trigger==="click"?M={onClick:this.handleClick}:M={onMouseEnter:this.handleMouseEnter,onDoubleClick:this.handleDoubleClick,onMouseMove:this.handleMouseMove,onMouseLeave:this.handleMouseLeave,onTouchMove:this.handleTouchMove,onTouchStart:this.handleTouchStart,onTouchEnd:this.handleTouchEnd,onContextMenu:this.handleContextMenu});var P=am(this.props,this.handleOuterEvent);return pe(pe({},P),M)}},{key:"addListener",value:function(){u_.on(c_,this.handleReceiveSyncEvent)}},{key:"removeListener",value:function(){u_.removeListener(c_,this.handleReceiveSyncEvent)}},{key:"filterFormatItem",value:function(A,T,R){for(var M=this.state.formattedGraphicalItems,P=0,N=M.length;P<N;P++){var k=M[P];if(k.item===A||k.props.key===A.key||T===sa(k.item.type)&&R===k.childIndex)return k}return null}},{key:"renderClipPath",value:function(){var A=this.clipPathId,T=this.state.offset,R=T.left,M=T.top,P=T.height,N=T.width;return V.createElement("defs",null,V.createElement("clipPath",{id:A},V.createElement("rect",{x:R,y:M,height:P,width:N})))}},{key:"getXScales",value:function(){var A=this.state.xAxisMap;return A?Object.entries(A).reduce(function(T,R){var M=FI(R,2),P=M[0],N=M[1];return pe(pe({},T),{},ze({},P,N.scale))},{}):null}},{key:"getYScales",value:function(){var A=this.state.yAxisMap;return A?Object.entries(A).reduce(function(T,R){var M=FI(R,2),P=M[0],N=M[1];return pe(pe({},T),{},ze({},P,N.scale))},{}):null}},{key:"getXScaleByAxisId",value:function(A){var T;return(T=this.state.xAxisMap)===null||T===void 0||(T=T[A])===null||T===void 0?void 0:T.scale}},{key:"getYScaleByAxisId",value:function(A){var T;return(T=this.state.yAxisMap)===null||T===void 0||(T=T[A])===null||T===void 0?void 0:T.scale}},{key:"getItemByXY",value:function(A){var T=this.state,R=T.formattedGraphicalItems,M=T.activeItem;if(R&&R.length)for(var P=0,N=R.length;P<N;P++){var k=R[P],H=k.props,L=k.item,F=L.type.defaultProps!==void 0?pe(pe({},L.type.defaultProps),L.props):L.props,z=sa(L.type);if(z==="Bar"){var Z=(H.data||[]).find(function(B){return She(A,B)});if(Z)return{graphicalItem:k,payload:Z}}else if(z==="RadialBar"){var Y=(H.data||[]).find(function(B){return o$(A,B)});if(Y)return{graphicalItem:k,payload:Y}}else if(ag(k,M)||og(k,M)||Zf(k,M)){var I=Nme({graphicalItem:k,activeTooltipItem:M,itemData:F.data}),W=F.activeIndex===void 0?I:F.activeIndex;return{graphicalItem:pe(pe({},k),{},{childIndex:W}),payload:Zf(k,M)?F.data[I]:k.props.data[I]}}}return null}},{key:"render",value:function(){var A=this;if(!yN(this))return null;var T=this.props,R=T.children,M=T.className,P=T.width,N=T.height,k=T.style,H=T.compact,L=T.title,F=T.desc,z=qI(T,qge),Z=Je(z,!1);if(H)return V.createElement(TI,{state:this.state,width:this.props.width,height:this.props.height,clipPathId:this.clipPathId},V.createElement(c1,_l({},Z,{width:P,height:N,title:L,desc:F}),this.renderClipPath(),xN(R,this.renderMap)));if(this.props.accessibilityLayer){var Y,I;Z.tabIndex=(Y=this.props.tabIndex)!==null&&Y!==void 0?Y:0,Z.role=(I=this.props.role)!==null&&I!==void 0?I:"application",Z.onKeyDown=function(B){A.accessibilityManager.keyboardEvent(B)},Z.onFocus=function(){A.accessibilityManager.focus()}}var W=this.parseEventsOfWrapper();return V.createElement(TI,{state:this.state,width:this.props.width,height:this.props.height,clipPathId:this.clipPathId},V.createElement("div",_l({className:ct("recharts-wrapper",M),style:pe({position:"relative",cursor:"default",width:P,height:N},k)},W,{ref:function(Q){A.container=Q}}),V.createElement(c1,_l({},Z,{width:P,height:N,title:L,desc:F,style:iye}),this.renderClipPath(),xN(R,this.renderMap)),this.renderLegend(),this.renderTooltip()))}}])}(g.Component);ze(S,"displayName",n),ze(S,"defaultProps",pe({layout:"horizontal",stackOffset:"none",barCategoryGap:"10%",barGap:4,margin:{top:5,right:5,bottom:5,left:5},reverseStackOrder:!1,syncMethod:"index"},m)),ze(S,"getDerivedStateFromProps",function(x,C){var O=x.dataKey,A=x.data,T=x.children,R=x.width,M=x.height,P=x.layout,N=x.stackOffset,k=x.margin,H=C.dataStartIndex,L=C.dataEndIndex;if(C.updateId===void 0){var F=VI(x);return pe(pe(pe({},F),{},{updateId:0},b(pe(pe({props:x},F),{},{updateId:0}),C)),{},{prevDataKey:O,prevData:A,prevWidth:R,prevHeight:M,prevLayout:P,prevStackOffset:N,prevMargin:k,prevChildren:T})}if(O!==C.prevDataKey||A!==C.prevData||R!==C.prevWidth||M!==C.prevHeight||P!==C.prevLayout||N!==C.prevStackOffset||!o1(k,C.prevMargin)){var z=VI(x),Z={chartX:C.chartX,chartY:C.chartY,isTooltipActive:C.isTooltipActive},Y=pe(pe({},UI(C,A,P)),{},{updateId:C.updateId+1}),I=pe(pe(pe({},z),Z),Y);return pe(pe(pe({},I),b(pe({props:x},I),C)),{},{prevDataKey:O,prevData:A,prevWidth:R,prevHeight:M,prevLayout:P,prevStackOffset:N,prevMargin:k,prevChildren:T})}if(!l1(T,C.prevChildren)){var W,B,Q,$,J=Tr(T,Ql),q=J&&(W=(B=J.props)===null||B===void 0?void 0:B.startIndex)!==null&&W!==void 0?W:H,X=J&&(Q=($=J.props)===null||$===void 0?void 0:$.endIndex)!==null&&Q!==void 0?Q:L,oe=q!==H||X!==L,le=!vt(A),ve=le&&!oe?C.updateId:C.updateId+1;return pe(pe({updateId:ve},b(pe(pe({props:x},C),{},{updateId:ve,dataStartIndex:q,dataEndIndex:X}),C)),{},{prevChildren:T,dataStartIndex:q,dataEndIndex:X})}return null}),ze(S,"renderActiveDot",function(x,C,O){var A;return g.isValidElement(x)?A=g.cloneElement(x,C):st(x)?A=x(C):A=V.createElement(tg,C),V.createElement(un,{className:"recharts-active-dot",key:O},A)});var w=g.forwardRef(function(C,O){return V.createElement(S,_l({},C,{ref:O}))});return w.displayName=S.displayName,w},mye=hye({chartName:"RadarChart",GraphicalChild:Rd,axisComponents:[{axisType:"angleAxis",AxisComp:Ad},{axisType:"radiusAxis",AxisComp:rg}],formatAxisMap:Ife,defaultProps:{layout:"centric",startAngle:90,endAngle:-270,cx:"50%",cy:"50%",innerRadius:0,outerRadius:"80%"}});const vye={light:"",dark:".dark"},O3=g.createContext(null);function gye(){const e=g.useContext(O3);if(!e)throw new Error("useChart must be used within a <ChartContainer />");return e}function yye({id:e,className:t,children:n,config:r,...i}){const a=g.useId(),s=`chart-${e||a.replace(/:/g,"")}`;return E.jsx(O3.Provider,{value:{config:r},children:E.jsxs("div",{"data-slot":"chart","data-chart":s,className:ke("[&_.recharts-cartesian-axis-tick_text]:fill-muted-foreground [&_.recharts-cartesian-grid_line[stroke='#ccc']]:stroke-border/50 [&_.recharts-curve.recharts-tooltip-cursor]:stroke-border [&_.recharts-polar-grid_[stroke='#ccc']]:stroke-border [&_.recharts-radial-bar-background-sector]:fill-muted [&_.recharts-rectangle.recharts-tooltip-cursor]:fill-muted [&_.recharts-reference-line_[stroke='#ccc']]:stroke-border flex aspect-video justify-center text-xs [&_.recharts-dot[stroke='#fff']]:stroke-transparent [&_.recharts-layer]:outline-hidden [&_.recharts-sector]:outline-hidden [&_.recharts-sector[stroke='#fff']]:stroke-transparent [&_.recharts-surface]:outline-hidden",t),...i,children:[E.jsx(bye,{id:s,config:r}),E.jsx(Zoe,{children:n})]})})}const bye=({id:e,config:t})=>{const n=Object.entries(t).filter(([,r])=>r.theme||r.color);return n.length?E.jsx("style",{dangerouslySetInnerHTML:{__html:Object.entries(vye).map(([r,i])=>`
${i} [data-chart=${e}] {
${n.map(([a,s])=>{var c;const u=((c=s.theme)==null?void 0:c[r])||s.color;return u?`  --color-${a}: ${u};`:null}).join(`
`)}
}
`).join(`
`)}}):null},xye=fi;function wye({active:e,payload:t,className:n,indicator:r="dot",hideLabel:i=!1,hideIndicator:a=!1,label:s,labelFormatter:u,labelClassName:c,formatter:d,color:p,nameKey:m,labelKey:v}){const{config:b}=gye(),S=g.useMemo(()=>{var T;if(i||!(t!=null&&t.length))return null;const[x]=t,C=`${v||(x==null?void 0:x.dataKey)||(x==null?void 0:x.name)||"value"}`,O=GI(b,x,C),A=!v&&typeof s=="string"?((T=b[s])==null?void 0:T.label)||s:O==null?void 0:O.label;return u?E.jsx("div",{className:ke("font-medium",c),children:u(A,t)}):A?E.jsx("div",{className:ke("font-medium",c),children:A}):null},[s,u,t,i,c,b,v]);if(!e||!(t!=null&&t.length))return null;const w=t.length===1&&r!=="dot";return E.jsxs("div",{className:ke("border-border/50 bg-background grid min-w-[8rem] items-start gap-1.5 rounded-lg border px-2.5 py-1.5 text-xs shadow-xl",n),children:[w?null:S,E.jsx("div",{className:"grid gap-1.5",children:t.map((x,C)=>{const O=`${m||x.name||x.dataKey||"value"}`,A=GI(b,x,O),T=p||x.payload.fill||x.color;return E.jsx("div",{className:ke("[&>svg]:text-muted-foreground flex w-full flex-wrap items-stretch gap-2 [&>svg]:h-2.5 [&>svg]:w-2.5",r==="dot"&&"items-center"),children:d&&(x==null?void 0:x.value)!==void 0&&x.name?d(x.value,x.name,x,C,x.payload):E.jsxs(E.Fragment,{children:[A!=null&&A.icon?E.jsx(A.icon,{}):!a&&E.jsx("div",{className:ke("shrink-0 rounded-[2px] border-(--color-border) bg-(--color-bg)",{"h-2.5 w-2.5":r==="dot","w-1":r==="line","w-0 border-[1.5px] border-dashed bg-transparent":r==="dashed","my-0.5":w&&r==="dashed"}),style:{"--color-bg":T,"--color-border":T}}),E.jsxs("div",{className:ke("flex flex-1 justify-between leading-none",w?"items-end":"items-center"),children:[E.jsxs("div",{className:"grid gap-1.5",children:[w?S:null,E.jsx("span",{className:"text-muted-foreground",children:(A==null?void 0:A.label)||x.name})]}),x.value&&E.jsx("span",{className:"text-foreground font-mono font-medium tabular-nums",children:x.value.toLocaleString()})]})]})},x.dataKey)})})]})}function GI(e,t,n){if(typeof t!="object"||t===null)return;const r="payload"in t&&typeof t.payload=="object"&&t.payload!==null?t.payload:void 0;let i=n;return n in t&&typeof t[n]=="string"?i=t[n]:r&&n in r&&typeof r[n]=="string"&&(i=r[n]),i in e?e[i]:e[n]}const Sye={score:{label:"Score",color:"hsl(var(--chart-1))"}};function _ye({data:e}){return E.jsx(yye,{config:Sye,className:"mx-auto aspect-square max-h-[1000px]",children:E.jsxs(mye,{data:e,children:[E.jsx(xye,{cursor:!1,content:E.jsx(wye,{})}),E.jsx(Ad,{dataKey:"taxon"}),E.jsx(YV,{gridType:"circle"}),E.jsx(Rd,{dataKey:"score",fill:"var(--primary)",fillOpacity:.6,animationDuration:500,max:1})]})})}function Eye(e){const t=Cye(e),n=g.forwardRef((r,i)=>{const{children:a,...s}=r,u=g.Children.toArray(a),c=u.find(Aye);if(c){const d=c.props.children,p=u.map(m=>m===c?g.Children.count(d)>1?g.Children.only(null):g.isValidElement(d)?d.props.children:null:m);return E.jsx(t,{...s,ref:i,children:g.isValidElement(d)?g.cloneElement(d,void 0,p):null})}return E.jsx(t,{...s,ref:i,children:a})});return n.displayName=`${e}.Slot`,n}function Cye(e){const t=g.forwardRef((n,r)=>{const{children:i,...a}=n,s=g.isValidElement(i)?Tye(i):void 0,u=Re(s,r);if(g.isValidElement(i)){const c=Rye(a,i.props);return i.type!==g.Fragment&&(c.ref=u),g.cloneElement(i,c)}return g.Children.count(i)>1?g.Children.only(null):null});return t.displayName=`${e}.SlotClone`,t}var Oye=Symbol("radix.slottable");function Aye(e){return g.isValidElement(e)&&typeof e.type=="function"&&"__radixId"in e.type&&e.type.__radixId===Oye}function Rye(e,t){const n={...t};for(const r in t){const i=e[r],a=t[r];/^on[A-Z]/.test(r)?i&&a?n[r]=(...u)=>{const c=a(...u);return i(...u),c}:i&&(n[r]=i):r==="style"?n[r]={...i,...a}:r==="className"&&(n[r]=[i,a].filter(Boolean).join(" "))}return{...e,...n}}function Tye(e){var r,i;let t=(r=Object.getOwnPropertyDescriptor(e.props,"ref"))==null?void 0:r.get,n=t&&"isReactWarning"in t&&t.isReactWarning;return n?e.ref:(t=(i=Object.getOwnPropertyDescriptor(e,"ref"))==null?void 0:i.get,n=t&&"isReactWarning"in t&&t.isReactWarning,n?e.props.ref:e.props.ref||e.ref)}var Mye=["a","button","div","form","h2","h3","img","input","label","li","nav","ol","p","select","span","svg","ul"],Pye=Mye.reduce((e,t)=>{const n=Eye(`Primitive.${t}`),r=g.forwardRef((i,a)=>{const{asChild:s,...u}=i,c=s?n:t;return typeof window<"u"&&(window[Symbol.for("radix-ui")]=!0),E.jsx(c,{...u,ref:a})});return r.displayName=`Primitive.${t}`,{...e,[t]:r}},{});function Gm(e){const t=jye(e),n=g.forwardRef((r,i)=>{const{children:a,...s}=r,u=g.Children.toArray(a),c=u.find(Dye);if(c){const d=c.props.children,p=u.map(m=>m===c?g.Children.count(d)>1?g.Children.only(null):g.isValidElement(d)?d.props.children:null:m);return E.jsx(t,{...s,ref:i,children:g.isValidElement(d)?g.cloneElement(d,void 0,p):null})}return E.jsx(t,{...s,ref:i,children:a})});return n.displayName=`${e}.Slot`,n}function jye(e){const t=g.forwardRef((n,r)=>{const{children:i,...a}=n,s=g.isValidElement(i)?$ye(i):void 0,u=Re(s,r);if(g.isValidElement(i)){const c=kye(a,i.props);return i.type!==g.Fragment&&(c.ref=u),g.cloneElement(i,c)}return g.Children.count(i)>1?g.Children.only(null):null});return t.displayName=`${e}.SlotClone`,t}var Nye=Symbol("radix.slottable");function Dye(e){return g.isValidElement(e)&&typeof e.type=="function"&&"__radixId"in e.type&&e.type.__radixId===Nye}function kye(e,t){const n={...t};for(const r in t){const i=e[r],a=t[r];/^on[A-Z]/.test(r)?i&&a?n[r]=(...u)=>{const c=a(...u);return i(...u),c}:i&&(n[r]=i):r==="style"?n[r]={...i,...a}:r==="className"&&(n[r]=[i,a].filter(Boolean).join(" "))}return{...e,...n}}function $ye(e){var r,i;let t=(r=Object.getOwnPropertyDescriptor(e.props,"ref"))==null?void 0:r.get,n=t&&"isReactWarning"in t&&t.isReactWarning;return n?e.ref:(t=(i=Object.getOwnPropertyDescriptor(e,"ref"))==null?void 0:i.get,n=t&&"isReactWarning"in t&&t.isReactWarning,n?e.props.ref:e.props.ref||e.ref)}function Iye(e){const t=e+"CollectionProvider",[n,r]=jn(t),[i,a]=n(t,{collectionRef:{current:null},itemMap:new Map}),s=w=>{const{scope:x,children:C}=w,O=V.useRef(null),A=V.useRef(new Map).current;return E.jsx(i,{scope:x,itemMap:A,collectionRef:O,children:C})};s.displayName=t;const u=e+"CollectionSlot",c=Gm(u),d=V.forwardRef((w,x)=>{const{scope:C,children:O}=w,A=a(u,C),T=Re(x,A.collectionRef);return E.jsx(c,{ref:T,children:O})});d.displayName=u;const p=e+"CollectionItemSlot",m="data-radix-collection-item",v=Gm(p),b=V.forwardRef((w,x)=>{const{scope:C,children:O,...A}=w,T=V.useRef(null),R=Re(x,T),M=a(p,C);return V.useEffect(()=>(M.itemMap.set(T,{ref:T,...A}),()=>void M.itemMap.delete(T))),E.jsx(v,{[m]:"",ref:R,children:O})});b.displayName=p;function S(w){const x=a(e+"CollectionConsumer",w);return V.useCallback(()=>{const O=x.collectionRef.current;if(!O)return[];const A=Array.from(O.querySelectorAll(`[${m}]`));return Array.from(x.itemMap.values()).sort((M,P)=>A.indexOf(M.ref.current)-A.indexOf(P.ref.current))},[x.collectionRef,x.itemMap])}return[{Provider:s,Slot:d,ItemSlot:b},S,r]}var Lye=["a","button","div","form","h2","h3","img","input","label","li","nav","ol","p","select","span","svg","ul"],ws=Lye.reduce((e,t)=>{const n=Gm(`Primitive.${t}`),r=g.forwardRef((i,a)=>{const{asChild:s,...u}=i,c=s?n:t;return typeof window<"u"&&(window[Symbol.for("radix-ui")]=!0),E.jsx(c,{...u,ref:a})});return r.displayName=`Primitive.${t}`,{...e,[t]:r}},{});function A3(e,t){e&&ba.flushSync(()=>e.dispatchEvent(t))}var Bye="DismissableLayer",bE="dismissableLayer.update",Fye="dismissableLayer.pointerDownOutside",qye="dismissableLayer.focusOutside",WI,R3=g.createContext({layers:new Set,layersWithOutsidePointerEventsDisabled:new Set,branches:new Set}),T3=g.forwardRef((e,t)=>{const{disableOutsidePointerEvents:n=!1,onEscapeKeyDown:r,onPointerDownOutside:i,onFocusOutside:a,onInteractOutside:s,onDismiss:u,...c}=e,d=g.useContext(R3),[p,m]=g.useState(null),v=(p==null?void 0:p.ownerDocument)??(globalThis==null?void 0:globalThis.document),[,b]=g.useState({}),S=Re(t,P=>m(P)),w=Array.from(d.layers),[x]=[...d.layersWithOutsidePointerEventsDisabled].slice(-1),C=w.indexOf(x),O=p?w.indexOf(p):-1,A=d.layersWithOutsidePointerEventsDisabled.size>0,T=O>=C,R=Vye(P=>{const N=P.target,k=[...d.branches].some(H=>H.contains(N));!T||k||(i==null||i(P),s==null||s(P),P.defaultPrevented||u==null||u())},v),M=Hye(P=>{const N=P.target;[...d.branches].some(H=>H.contains(N))||(a==null||a(P),s==null||s(P),P.defaultPrevented||u==null||u())},v);return rd(P=>{O===d.layers.size-1&&(r==null||r(P),!P.defaultPrevented&&u&&(P.preventDefault(),u()))},v),g.useEffect(()=>{if(p)return n&&(d.layersWithOutsidePointerEventsDisabled.size===0&&(WI=v.body.style.pointerEvents,v.body.style.pointerEvents="none"),d.layersWithOutsidePointerEventsDisabled.add(p)),d.layers.add(p),KI(),()=>{n&&d.layersWithOutsidePointerEventsDisabled.size===1&&(v.body.style.pointerEvents=WI)}},[p,v,n,d]),g.useEffect(()=>()=>{p&&(d.layers.delete(p),d.layersWithOutsidePointerEventsDisabled.delete(p),KI())},[p,d]),g.useEffect(()=>{const P=()=>b({});return document.addEventListener(bE,P),()=>document.removeEventListener(bE,P)},[]),E.jsx(ws.div,{...c,ref:S,style:{pointerEvents:A?T?"auto":"none":void 0,...e.style},onFocusCapture:be(e.onFocusCapture,M.onFocusCapture),onBlurCapture:be(e.onBlurCapture,M.onBlurCapture),onPointerDownCapture:be(e.onPointerDownCapture,R.onPointerDownCapture)})});T3.displayName=Bye;var zye="DismissableLayerBranch",Uye=g.forwardRef((e,t)=>{const n=g.useContext(R3),r=g.useRef(null),i=Re(t,r);return g.useEffect(()=>{const a=r.current;if(a)return n.branches.add(a),()=>{n.branches.delete(a)}},[n.branches]),E.jsx(ws.div,{...e,ref:i})});Uye.displayName=zye;function Vye(e,t=globalThis==null?void 0:globalThis.document){const n=$t(e),r=g.useRef(!1),i=g.useRef(()=>{});return g.useEffect(()=>{const a=u=>{if(u.target&&!r.current){let c=function(){M3(Fye,n,d,{discrete:!0})};const d={originalEvent:u};u.pointerType==="touch"?(t.removeEventListener("click",i.current),i.current=c,t.addEventListener("click",i.current,{once:!0})):c()}else t.removeEventListener("click",i.current);r.current=!1},s=window.setTimeout(()=>{t.addEventListener("pointerdown",a)},0);return()=>{window.clearTimeout(s),t.removeEventListener("pointerdown",a),t.removeEventListener("click",i.current)}},[t,n]),{onPointerDownCapture:()=>r.current=!0}}function Hye(e,t=globalThis==null?void 0:globalThis.document){const n=$t(e),r=g.useRef(!1);return g.useEffect(()=>{const i=a=>{a.target&&!r.current&&M3(qye,n,{originalEvent:a},{discrete:!1})};return t.addEventListener("focusin",i),()=>t.removeEventListener("focusin",i)},[t,n]),{onFocusCapture:()=>r.current=!0,onBlurCapture:()=>r.current=!1}}function KI(){const e=new CustomEvent(bE);document.dispatchEvent(e)}function M3(e,t,n,{discrete:r}){const i=n.originalEvent.target,a=new CustomEvent(e,{bubbles:!1,cancelable:!0,detail:n});t&&i.addEventListener(e,t,{once:!0}),r?A3(i,a):i.dispatchEvent(a)}var xE=["Enter"," "],Gye=["ArrowDown","PageUp","Home"],P3=["ArrowUp","PageDown","End"],Wye=[...Gye,...P3],Kye={ltr:[...xE,"ArrowRight"],rtl:[...xE,"ArrowLeft"]},Yye={ltr:["ArrowLeft"],rtl:["ArrowRight"]},Td="Menu",[td,Zye,Xye]=Iye(Td),[Ss,j3]=jn(Td,[Xye,fu,_v]),dg=fu(),N3=_v(),[Qye,_s]=Ss(Td),[Jye,Md]=Ss(Td),D3=e=>{const{__scopeMenu:t,open:n=!1,children:r,dir:i,onOpenChange:a,modal:s=!0}=e,u=dg(t),[c,d]=g.useState(null),p=g.useRef(!1),m=$t(a),v=iu(i);return g.useEffect(()=>{const b=()=>{p.current=!0,document.addEventListener("pointerdown",S,{capture:!0,once:!0}),document.addEventListener("pointermove",S,{capture:!0,once:!0})},S=()=>p.current=!1;return document.addEventListener("keydown",b,{capture:!0}),()=>{document.removeEventListener("keydown",b,{capture:!0}),document.removeEventListener("pointerdown",S,{capture:!0}),document.removeEventListener("pointermove",S,{capture:!0})}},[]),E.jsx(rC,{...u,children:E.jsx(Qye,{scope:t,open:n,onOpenChange:m,content:c,onContentChange:d,children:E.jsx(Jye,{scope:t,onClose:g.useCallback(()=>m(!1),[m]),isUsingKeyboardRef:p,dir:v,modal:s,children:r})})})};D3.displayName=Td;var ebe="MenuAnchor",PO=g.forwardRef((e,t)=>{const{__scopeMenu:n,...r}=e,i=dg(n);return E.jsx(lv,{...i,...r,ref:t})});PO.displayName=ebe;var jO="MenuPortal",[tbe,k3]=Ss(jO,{forceMount:void 0}),$3=e=>{const{__scopeMenu:t,forceMount:n,children:r,container:i}=e,a=_s(jO,t);return E.jsx(tbe,{scope:t,forceMount:n,children:E.jsx(zn,{present:n||a.open,children:E.jsx(fd,{asChild:!0,container:i,children:r})})})};$3.displayName=jO;var Nr="MenuContent",[nbe,NO]=Ss(Nr),I3=g.forwardRef((e,t)=>{const n=k3(Nr,e.__scopeMenu),{forceMount:r=n.forceMount,...i}=e,a=_s(Nr,e.__scopeMenu),s=Md(Nr,e.__scopeMenu);return E.jsx(td.Provider,{scope:e.__scopeMenu,children:E.jsx(zn,{present:r||a.open,children:E.jsx(td.Slot,{scope:e.__scopeMenu,children:s.modal?E.jsx(rbe,{...i,ref:t}):E.jsx(ibe,{...i,ref:t})})})})}),rbe=g.forwardRef((e,t)=>{const n=_s(Nr,e.__scopeMenu),r=g.useRef(null),i=Re(t,r);return g.useEffect(()=>{const a=r.current;if(a)return uv(a)},[]),E.jsx(DO,{...e,ref:i,trapFocus:n.open,disableOutsidePointerEvents:n.open,disableOutsideScroll:!0,onFocusOutside:be(e.onFocusOutside,a=>a.preventDefault(),{checkForDefaultPrevented:!1}),onDismiss:()=>n.onOpenChange(!1)})}),ibe=g.forwardRef((e,t)=>{const n=_s(Nr,e.__scopeMenu);return E.jsx(DO,{...e,ref:t,trapFocus:!1,disableOutsidePointerEvents:!1,disableOutsideScroll:!1,onDismiss:()=>n.onOpenChange(!1)})}),abe=Gm("MenuContent.ScrollLock"),DO=g.forwardRef((e,t)=>{const{__scopeMenu:n,loop:r=!1,trapFocus:i,onOpenAutoFocus:a,onCloseAutoFocus:s,disableOutsidePointerEvents:u,onEntryFocus:c,onEscapeKeyDown:d,onPointerDownOutside:p,onFocusOutside:m,onInteractOutside:v,onDismiss:b,disableOutsideScroll:S,...w}=e,x=_s(Nr,n),C=Md(Nr,n),O=dg(n),A=N3(n),T=Zye(n),[R,M]=g.useState(null),P=g.useRef(null),N=Re(t,P,x.onContentChange),k=g.useRef(0),H=g.useRef(""),L=g.useRef(0),F=g.useRef(null),z=g.useRef("right"),Z=g.useRef(0),Y=S?dd:g.Fragment,I=S?{as:abe,allowPinchZoom:!0}:void 0,W=Q=>{var de,Se;const $=H.current+Q,J=T().filter(ce=>!ce.disabled),q=document.activeElement,X=(de=J.find(ce=>ce.ref.current===q))==null?void 0:de.textValue,oe=J.map(ce=>ce.textValue),le=gbe(oe,$,X),ve=(Se=J.find(ce=>ce.textValue===le))==null?void 0:Se.ref.current;(function ce(we){H.current=we,window.clearTimeout(k.current),we!==""&&(k.current=window.setTimeout(()=>ce(""),1e3))})($),ve&&setTimeout(()=>ve.focus())};g.useEffect(()=>()=>window.clearTimeout(k.current),[]),iv();const B=g.useCallback(Q=>{var J,q;return z.current===((J=F.current)==null?void 0:J.side)&&bbe(Q,(q=F.current)==null?void 0:q.area)},[]);return E.jsx(nbe,{scope:n,searchRef:H,onItemEnter:g.useCallback(Q=>{B(Q)&&Q.preventDefault()},[B]),onItemLeave:g.useCallback(Q=>{var $;B(Q)||(($=P.current)==null||$.focus(),M(null))},[B]),onTriggerLeave:g.useCallback(Q=>{B(Q)&&Q.preventDefault()},[B]),pointerGraceTimerRef:L,onPointerGraceIntentChange:g.useCallback(Q=>{F.current=Q},[]),children:E.jsx(Y,{...I,children:E.jsx(ud,{asChild:!0,trapped:i,onMountAutoFocus:be(a,Q=>{var $;Q.preventDefault(),($=P.current)==null||$.focus({preventScroll:!0})}),onUnmountAutoFocus:s,children:E.jsx(T3,{asChild:!0,disableOutsidePointerEvents:u,onEscapeKeyDown:d,onPointerDownOutside:p,onFocusOutside:m,onInteractOutside:v,onDismiss:b,children:E.jsx(Eq,{asChild:!0,...A,dir:C.dir,orientation:"vertical",loop:r,currentTabStopId:R,onCurrentTabStopIdChange:M,onEntryFocus:be(c,Q=>{C.isUsingKeyboardRef.current||Q.preventDefault()}),preventScrollOnEntryFocus:!0,children:E.jsx(iC,{role:"menu","aria-orientation":"vertical","data-state":J3(x.open),"data-radix-menu-content":"",dir:C.dir,...O,...w,ref:N,style:{outline:"none",...w.style},onKeyDown:be(w.onKeyDown,Q=>{const J=Q.target.closest("[data-radix-menu-content]")===Q.currentTarget,q=Q.ctrlKey||Q.altKey||Q.metaKey,X=Q.key.length===1;J&&(Q.key==="Tab"&&Q.preventDefault(),!q&&X&&W(Q.key));const oe=P.current;if(Q.target!==oe||!Wye.includes(Q.key))return;Q.preventDefault();const ve=T().filter(de=>!de.disabled).map(de=>de.ref.current);P3.includes(Q.key)&&ve.reverse(),mbe(ve)}),onBlur:be(e.onBlur,Q=>{Q.currentTarget.contains(Q.target)||(window.clearTimeout(k.current),H.current="")}),onPointerMove:be(e.onPointerMove,nd(Q=>{const $=Q.target,J=Z.current!==Q.clientX;if(Q.currentTarget.contains($)&&J){const q=Q.clientX>Z.current?"right":"left";z.current=q,Z.current=Q.clientX}}))})})})})})})});I3.displayName=Nr;var obe="MenuGroup",kO=g.forwardRef((e,t)=>{const{__scopeMenu:n,...r}=e;return E.jsx(ws.div,{role:"group",...r,ref:t})});kO.displayName=obe;var sbe="MenuLabel",L3=g.forwardRef((e,t)=>{const{__scopeMenu:n,...r}=e;return E.jsx(ws.div,{...r,ref:t})});L3.displayName=sbe;var Wm="MenuItem",YI="menu.itemSelect",pg=g.forwardRef((e,t)=>{const{disabled:n=!1,onSelect:r,...i}=e,a=g.useRef(null),s=Md(Wm,e.__scopeMenu),u=NO(Wm,e.__scopeMenu),c=Re(t,a),d=g.useRef(!1),p=()=>{const m=a.current;if(!n&&m){const v=new CustomEvent(YI,{bubbles:!0,cancelable:!0});m.addEventListener(YI,b=>r==null?void 0:r(b),{once:!0}),A3(m,v),v.defaultPrevented?d.current=!1:s.onClose()}};return E.jsx(B3,{...i,ref:c,disabled:n,onClick:be(e.onClick,p),onPointerDown:m=>{var v;(v=e.onPointerDown)==null||v.call(e,m),d.current=!0},onPointerUp:be(e.onPointerUp,m=>{var v;d.current||(v=m.currentTarget)==null||v.click()}),onKeyDown:be(e.onKeyDown,m=>{const v=u.searchRef.current!=="";n||v&&m.key===" "||xE.includes(m.key)&&(m.currentTarget.click(),m.preventDefault())})})});pg.displayName=Wm;var B3=g.forwardRef((e,t)=>{const{__scopeMenu:n,disabled:r=!1,textValue:i,...a}=e,s=NO(Wm,n),u=N3(n),c=g.useRef(null),d=Re(t,c),[p,m]=g.useState(!1),[v,b]=g.useState("");return g.useEffect(()=>{const S=c.current;S&&b((S.textContent??"").trim())},[a.children]),E.jsx(td.ItemSlot,{scope:n,disabled:r,textValue:i??v,children:E.jsx(Cq,{asChild:!0,...u,focusable:!r,children:E.jsx(ws.div,{role:"menuitem","data-highlighted":p?"":void 0,"aria-disabled":r||void 0,"data-disabled":r?"":void 0,...a,ref:d,onPointerMove:be(e.onPointerMove,nd(S=>{r?s.onItemLeave(S):(s.onItemEnter(S),S.defaultPrevented||S.currentTarget.focus({preventScroll:!0}))})),onPointerLeave:be(e.onPointerLeave,nd(S=>s.onItemLeave(S))),onFocus:be(e.onFocus,()=>m(!0)),onBlur:be(e.onBlur,()=>m(!1))})})})}),lbe="MenuCheckboxItem",F3=g.forwardRef((e,t)=>{const{checked:n=!1,onCheckedChange:r,...i}=e;return E.jsx(H3,{scope:e.__scopeMenu,checked:n,children:E.jsx(pg,{role:"menuitemcheckbox","aria-checked":Km(n)?"mixed":n,...i,ref:t,"data-state":IO(n),onSelect:be(i.onSelect,()=>r==null?void 0:r(Km(n)?!0:!n),{checkForDefaultPrevented:!1})})})});F3.displayName=lbe;var q3="MenuRadioGroup",[ube,cbe]=Ss(q3,{value:void 0,onValueChange:()=>{}}),z3=g.forwardRef((e,t)=>{const{value:n,onValueChange:r,...i}=e,a=$t(r);return E.jsx(ube,{scope:e.__scopeMenu,value:n,onValueChange:a,children:E.jsx(kO,{...i,ref:t})})});z3.displayName=q3;var U3="MenuRadioItem",V3=g.forwardRef((e,t)=>{const{value:n,...r}=e,i=cbe(U3,e.__scopeMenu),a=n===i.value;return E.jsx(H3,{scope:e.__scopeMenu,checked:a,children:E.jsx(pg,{role:"menuitemradio","aria-checked":a,...r,ref:t,"data-state":IO(a),onSelect:be(r.onSelect,()=>{var s;return(s=i.onValueChange)==null?void 0:s.call(i,n)},{checkForDefaultPrevented:!1})})})});V3.displayName=U3;var $O="MenuItemIndicator",[H3,fbe]=Ss($O,{checked:!1}),G3=g.forwardRef((e,t)=>{const{__scopeMenu:n,forceMount:r,...i}=e,a=fbe($O,n);return E.jsx(zn,{present:r||Km(a.checked)||a.checked===!0,children:E.jsx(ws.span,{...i,ref:t,"data-state":IO(a.checked)})})});G3.displayName=$O;var dbe="MenuSeparator",W3=g.forwardRef((e,t)=>{const{__scopeMenu:n,...r}=e;return E.jsx(ws.div,{role:"separator","aria-orientation":"horizontal",...r,ref:t})});W3.displayName=dbe;var pbe="MenuArrow",K3=g.forwardRef((e,t)=>{const{__scopeMenu:n,...r}=e,i=dg(n);return E.jsx(aC,{...i,...r,ref:t})});K3.displayName=pbe;var hbe="MenuSub",[yxe,Y3]=Ss(hbe),Vc="MenuSubTrigger",Z3=g.forwardRef((e,t)=>{const n=_s(Vc,e.__scopeMenu),r=Md(Vc,e.__scopeMenu),i=Y3(Vc,e.__scopeMenu),a=NO(Vc,e.__scopeMenu),s=g.useRef(null),{pointerGraceTimerRef:u,onPointerGraceIntentChange:c}=a,d={__scopeMenu:e.__scopeMenu},p=g.useCallback(()=>{s.current&&window.clearTimeout(s.current),s.current=null},[]);return g.useEffect(()=>p,[p]),g.useEffect(()=>{const m=u.current;return()=>{window.clearTimeout(m),c(null)}},[u,c]),E.jsx(PO,{asChild:!0,...d,children:E.jsx(B3,{id:i.triggerId,"aria-haspopup":"menu","aria-expanded":n.open,"aria-controls":i.contentId,"data-state":J3(n.open),...e,ref:wi(t,i.onTriggerChange),onClick:m=>{var v;(v=e.onClick)==null||v.call(e,m),!(e.disabled||m.defaultPrevented)&&(m.currentTarget.focus(),n.open||n.onOpenChange(!0))},onPointerMove:be(e.onPointerMove,nd(m=>{a.onItemEnter(m),!m.defaultPrevented&&!e.disabled&&!n.open&&!s.current&&(a.onPointerGraceIntentChange(null),s.current=window.setTimeout(()=>{n.onOpenChange(!0),p()},100))})),onPointerLeave:be(e.onPointerLeave,nd(m=>{var b,S;p();const v=(b=n.content)==null?void 0:b.getBoundingClientRect();if(v){const w=(S=n.content)==null?void 0:S.dataset.side,x=w==="right",C=x?-5:5,O=v[x?"left":"right"],A=v[x?"right":"left"];a.onPointerGraceIntentChange({area:[{x:m.clientX+C,y:m.clientY},{x:O,y:v.top},{x:A,y:v.top},{x:A,y:v.bottom},{x:O,y:v.bottom}],side:w}),window.clearTimeout(u.current),u.current=window.setTimeout(()=>a.onPointerGraceIntentChange(null),300)}else{if(a.onTriggerLeave(m),m.defaultPrevented)return;a.onPointerGraceIntentChange(null)}})),onKeyDown:be(e.onKeyDown,m=>{var b;const v=a.searchRef.current!=="";e.disabled||v&&m.key===" "||Kye[r.dir].includes(m.key)&&(n.onOpenChange(!0),(b=n.content)==null||b.focus(),m.preventDefault())})})})});Z3.displayName=Vc;var X3="MenuSubContent",Q3=g.forwardRef((e,t)=>{const n=k3(Nr,e.__scopeMenu),{forceMount:r=n.forceMount,...i}=e,a=_s(Nr,e.__scopeMenu),s=Md(Nr,e.__scopeMenu),u=Y3(X3,e.__scopeMenu),c=g.useRef(null),d=Re(t,c);return E.jsx(td.Provider,{scope:e.__scopeMenu,children:E.jsx(zn,{present:r||a.open,children:E.jsx(td.Slot,{scope:e.__scopeMenu,children:E.jsx(DO,{id:u.contentId,"aria-labelledby":u.triggerId,...i,ref:d,align:"start",side:s.dir==="rtl"?"left":"right",disableOutsidePointerEvents:!1,disableOutsideScroll:!1,trapFocus:!1,onOpenAutoFocus:p=>{var m;s.isUsingKeyboardRef.current&&((m=c.current)==null||m.focus()),p.preventDefault()},onCloseAutoFocus:p=>p.preventDefault(),onFocusOutside:be(e.onFocusOutside,p=>{p.target!==u.trigger&&a.onOpenChange(!1)}),onEscapeKeyDown:be(e.onEscapeKeyDown,p=>{s.onClose(),p.preventDefault()}),onKeyDown:be(e.onKeyDown,p=>{var b;const m=p.currentTarget.contains(p.target),v=Yye[s.dir].includes(p.key);m&&v&&(a.onOpenChange(!1),(b=u.trigger)==null||b.focus(),p.preventDefault())})})})})})});Q3.displayName=X3;function J3(e){return e?"open":"closed"}function Km(e){return e==="indeterminate"}function IO(e){return Km(e)?"indeterminate":e?"checked":"unchecked"}function mbe(e){const t=document.activeElement;for(const n of e)if(n===t||(n.focus(),document.activeElement!==t))return}function vbe(e,t){return e.map((n,r)=>e[(t+r)%e.length])}function gbe(e,t,n){const i=t.length>1&&Array.from(t).every(d=>d===t[0])?t[0]:t,a=n?e.indexOf(n):-1;let s=vbe(e,Math.max(a,0));i.length===1&&(s=s.filter(d=>d!==n));const c=s.find(d=>d.toLowerCase().startsWith(i.toLowerCase()));return c!==n?c:void 0}function ybe(e,t){const{x:n,y:r}=e;let i=!1;for(let a=0,s=t.length-1;a<t.length;s=a++){const u=t[a],c=t[s],d=u.x,p=u.y,m=c.x,v=c.y;p>r!=v>r&&n<(m-d)*(r-p)/(v-p)+d&&(i=!i)}return i}function bbe(e,t){if(!t)return!1;const n={x:e.clientX,y:e.clientY};return ybe(n,t)}function nd(e){return t=>t.pointerType==="mouse"?e(t):void 0}var xbe=D3,wbe=PO,Sbe=$3,_be=I3,Ebe=kO,Cbe=L3,Obe=pg,Abe=F3,Rbe=z3,Tbe=V3,Mbe=G3,Pbe=W3,jbe=K3,Nbe=Z3,Dbe=Q3,hg="DropdownMenu",[kbe,bxe]=jn(hg,[j3]),Un=j3(),[$be,e4]=kbe(hg),t4=e=>{const{__scopeDropdownMenu:t,children:n,dir:r,open:i,defaultOpen:a,onOpenChange:s,modal:u=!0}=e,c=Un(t),d=g.useRef(null),[p,m]=Xr({prop:i,defaultProp:a??!1,onChange:s,caller:hg});return E.jsx($be,{scope:t,triggerId:ln(),triggerRef:d,contentId:ln(),open:p,onOpenChange:m,onOpenToggle:g.useCallback(()=>m(v=>!v),[m]),modal:u,children:E.jsx(xbe,{...c,open:p,onOpenChange:m,dir:r,modal:u,children:n})})};t4.displayName=hg;var n4="DropdownMenuTrigger",r4=g.forwardRef((e,t)=>{const{__scopeDropdownMenu:n,disabled:r=!1,...i}=e,a=e4(n4,n),s=Un(n);return E.jsx(wbe,{asChild:!0,...s,children:E.jsx(Pye.button,{type:"button",id:a.triggerId,"aria-haspopup":"menu","aria-expanded":a.open,"aria-controls":a.open?a.contentId:void 0,"data-state":a.open?"open":"closed","data-disabled":r?"":void 0,disabled:r,...i,ref:wi(t,a.triggerRef),onPointerDown:be(e.onPointerDown,u=>{!r&&u.button===0&&u.ctrlKey===!1&&(a.onOpenToggle(),a.open||u.preventDefault())}),onKeyDown:be(e.onKeyDown,u=>{r||(["Enter"," "].includes(u.key)&&a.onOpenToggle(),u.key==="ArrowDown"&&a.onOpenChange(!0),["Enter"," ","ArrowDown"].includes(u.key)&&u.preventDefault())})})})});r4.displayName=n4;var Ibe="DropdownMenuPortal",i4=e=>{const{__scopeDropdownMenu:t,...n}=e,r=Un(t);return E.jsx(Sbe,{...r,...n})};i4.displayName=Ibe;var a4="DropdownMenuContent",o4=g.forwardRef((e,t)=>{const{__scopeDropdownMenu:n,...r}=e,i=e4(a4,n),a=Un(n),s=g.useRef(!1);return E.jsx(_be,{id:i.contentId,"aria-labelledby":i.triggerId,...a,...r,ref:t,onCloseAutoFocus:be(e.onCloseAutoFocus,u=>{var c;s.current||(c=i.triggerRef.current)==null||c.focus(),s.current=!1,u.preventDefault()}),onInteractOutside:be(e.onInteractOutside,u=>{const c=u.detail.originalEvent,d=c.button===0&&c.ctrlKey===!0,p=c.button===2||d;(!i.modal||p)&&(s.current=!0)}),style:{...e.style,"--radix-dropdown-menu-content-transform-origin":"var(--radix-popper-transform-origin)","--radix-dropdown-menu-content-available-width":"var(--radix-popper-available-width)","--radix-dropdown-menu-content-available-height":"var(--radix-popper-available-height)","--radix-dropdown-menu-trigger-width":"var(--radix-popper-anchor-width)","--radix-dropdown-menu-trigger-height":"var(--radix-popper-anchor-height)"}})});o4.displayName=a4;var Lbe="DropdownMenuGroup",Bbe=g.forwardRef((e,t)=>{const{__scopeDropdownMenu:n,...r}=e,i=Un(n);return E.jsx(Ebe,{...i,...r,ref:t})});Bbe.displayName=Lbe;var Fbe="DropdownMenuLabel",s4=g.forwardRef((e,t)=>{const{__scopeDropdownMenu:n,...r}=e,i=Un(n);return E.jsx(Cbe,{...i,...r,ref:t})});s4.displayName=Fbe;var qbe="DropdownMenuItem",zbe=g.forwardRef((e,t)=>{const{__scopeDropdownMenu:n,...r}=e,i=Un(n);return E.jsx(Obe,{...i,...r,ref:t})});zbe.displayName=qbe;var Ube="DropdownMenuCheckboxItem",l4=g.forwardRef((e,t)=>{const{__scopeDropdownMenu:n,...r}=e,i=Un(n);return E.jsx(Abe,{...i,...r,ref:t})});l4.displayName=Ube;var Vbe="DropdownMenuRadioGroup",Hbe=g.forwardRef((e,t)=>{const{__scopeDropdownMenu:n,...r}=e,i=Un(n);return E.jsx(Rbe,{...i,...r,ref:t})});Hbe.displayName=Vbe;var Gbe="DropdownMenuRadioItem",Wbe=g.forwardRef((e,t)=>{const{__scopeDropdownMenu:n,...r}=e,i=Un(n);return E.jsx(Tbe,{...i,...r,ref:t})});Wbe.displayName=Gbe;var Kbe="DropdownMenuItemIndicator",u4=g.forwardRef((e,t)=>{const{__scopeDropdownMenu:n,...r}=e,i=Un(n);return E.jsx(Mbe,{...i,...r,ref:t})});u4.displayName=Kbe;var Ybe="DropdownMenuSeparator",c4=g.forwardRef((e,t)=>{const{__scopeDropdownMenu:n,...r}=e,i=Un(n);return E.jsx(Pbe,{...i,...r,ref:t})});c4.displayName=Ybe;var Zbe="DropdownMenuArrow",Xbe=g.forwardRef((e,t)=>{const{__scopeDropdownMenu:n,...r}=e,i=Un(n);return E.jsx(jbe,{...i,...r,ref:t})});Xbe.displayName=Zbe;var Qbe="DropdownMenuSubTrigger",Jbe=g.forwardRef((e,t)=>{const{__scopeDropdownMenu:n,...r}=e,i=Un(n);return E.jsx(Nbe,{...i,...r,ref:t})});Jbe.displayName=Qbe;var e0e="DropdownMenuSubContent",t0e=g.forwardRef((e,t)=>{const{__scopeDropdownMenu:n,...r}=e,i=Un(n);return E.jsx(Dbe,{...i,...r,ref:t,style:{...e.style,"--radix-dropdown-menu-content-transform-origin":"var(--radix-popper-transform-origin)","--radix-dropdown-menu-content-available-width":"var(--radix-popper-available-width)","--radix-dropdown-menu-content-available-height":"var(--radix-popper-available-height)","--radix-dropdown-menu-trigger-width":"var(--radix-popper-anchor-width)","--radix-dropdown-menu-trigger-height":"var(--radix-popper-anchor-height)"}})});t0e.displayName=e0e;var n0e=t4,r0e=r4,i0e=i4,a0e=o4,o0e=s4,s0e=l4,l0e=u4,u0e=c4;function f4({...e}){return E.jsx(n0e,{"data-slot":"dropdown-menu",...e})}function d4({...e}){return E.jsx(r0e,{"data-slot":"dropdown-menu-trigger",...e})}function p4({className:e,sideOffset:t=4,...n}){return E.jsx(i0e,{children:E.jsx(a0e,{"data-slot":"dropdown-menu-content",sideOffset:t,className:ke("bg-popover text-popover-foreground data-[state=open]:animate-in data-[state=closed]:animate-out data-[state=closed]:fade-out-0 data-[state=open]:fade-in-0 data-[state=closed]:zoom-out-95 data-[state=open]:zoom-in-95 data-[side=bottom]:slide-in-from-top-2 data-[side=left]:slide-in-from-right-2 data-[side=right]:slide-in-from-left-2 data-[side=top]:slide-in-from-bottom-2 z-50 max-h-(--radix-dropdown-menu-content-available-height) min-w-[8rem] origin-(--radix-dropdown-menu-content-transform-origin) overflow-x-hidden overflow-y-auto rounded-md border p-1 shadow-md",e),...n})})}function c0e({className:e,children:t,checked:n,...r}){return E.jsxs(s0e,{"data-slot":"dropdown-menu-checkbox-item",className:ke("focus:bg-accent focus:text-accent-foreground relative flex cursor-default items-center gap-2 rounded-sm py-1.5 pr-2 pl-8 text-sm outline-hidden select-none data-[disabled]:pointer-events-none data-[disabled]:opacity-50 [&_svg]:pointer-events-none [&_svg]:shrink-0 [&_svg:not([class*='size-'])]:size-4",e),checked:n,...r,children:[E.jsx("span",{className:"pointer-events-none absolute left-2 flex size-3.5 items-center justify-center",children:E.jsx(l0e,{children:E.jsx(jE,{className:"size-4"})})}),t]})}function f0e({className:e,inset:t,...n}){return E.jsx(o0e,{"data-slot":"dropdown-menu-label","data-inset":t,className:ke("px-2 py-1.5 text-sm font-medium data-[inset]:pl-8",e),...n})}function d0e({className:e,...t}){return E.jsx(u0e,{"data-slot":"dropdown-menu-separator",className:ke("bg-border -mx-1 my-1 h-px",e),...t})}function p0e({triggerButtonText:e,labelText:t,items:n,disabled:r=!1}){return E.jsxs(f4,{children:[E.jsx(d4,{asChild:!0,children:E.jsxs(Pr,{variant:"ghost",disabled:r,children:[e,E.jsx(Zm,{})]})}),E.jsxs(p4,{className:"w-56",children:[t&&E.jsxs(E.Fragment,{children:[E.jsx(f0e,{children:t}),E.jsx(d0e,{})]}),n.map(i=>E.jsx(c0e,{checked:i.checked,onCheckedChange:i.onCheckedChange,disabled:i.disabled,children:i.label},i.id))]})]})}function wE(e){const t=h0e(e),n=g.forwardRef((r,i)=>{const{children:a,...s}=r,u=g.Children.toArray(a),c=u.find(v0e);if(c){const d=c.props.children,p=u.map(m=>m===c?g.Children.count(d)>1?g.Children.only(null):g.isValidElement(d)?d.props.children:null:m);return E.jsx(t,{...s,ref:i,children:g.isValidElement(d)?g.cloneElement(d,void 0,p):null})}return E.jsx(t,{...s,ref:i,children:a})});return n.displayName=`${e}.Slot`,n}function h0e(e){const t=g.forwardRef((n,r)=>{const{children:i,...a}=n,s=g.isValidElement(i)?y0e(i):void 0,u=Re(s,r);if(g.isValidElement(i)){const c=g0e(a,i.props);return i.type!==g.Fragment&&(c.ref=u),g.cloneElement(i,c)}return g.Children.count(i)>1?g.Children.only(null):null});return t.displayName=`${e}.SlotClone`,t}var m0e=Symbol("radix.slottable");function v0e(e){return g.isValidElement(e)&&typeof e.type=="function"&&"__radixId"in e.type&&e.type.__radixId===m0e}function g0e(e,t){const n={...t};for(const r in t){const i=e[r],a=t[r];/^on[A-Z]/.test(r)?i&&a?n[r]=(...u)=>{const c=a(...u);return i(...u),c}:i&&(n[r]=i):r==="style"?n[r]={...i,...a}:r==="className"&&(n[r]=[i,a].filter(Boolean).join(" "))}return{...e,...n}}function y0e(e){var r,i;let t=(r=Object.getOwnPropertyDescriptor(e.props,"ref"))==null?void 0:r.get,n=t&&"isReactWarning"in t&&t.isReactWarning;return n?e.ref:(t=(i=Object.getOwnPropertyDescriptor(e,"ref"))==null?void 0:i.get,n=t&&"isReactWarning"in t&&t.isReactWarning,n?e.props.ref:e.props.ref||e.ref)}var b0e=["a","button","div","form","h2","h3","img","input","label","li","nav","ol","p","select","span","svg","ul"],Pd=b0e.reduce((e,t)=>{const n=wE(`Primitive.${t}`),r=g.forwardRef((i,a)=>{const{asChild:s,...u}=i,c=s?n:t;return typeof window<"u"&&(window[Symbol.for("radix-ui")]=!0),E.jsx(c,{...u,ref:a})});return r.displayName=`Primitive.${t}`,{...e,[t]:r}},{});function x0e(e){const t=e+"CollectionProvider",[n,r]=jn(t),[i,a]=n(t,{collectionRef:{current:null},itemMap:new Map}),s=w=>{const{scope:x,children:C}=w,O=V.useRef(null),A=V.useRef(new Map).current;return E.jsx(i,{scope:x,itemMap:A,collectionRef:O,children:C})};s.displayName=t;const u=e+"CollectionSlot",c=wE(u),d=V.forwardRef((w,x)=>{const{scope:C,children:O}=w,A=a(u,C),T=Re(x,A.collectionRef);return E.jsx(c,{ref:T,children:O})});d.displayName=u;const p=e+"CollectionItemSlot",m="data-radix-collection-item",v=wE(p),b=V.forwardRef((w,x)=>{const{scope:C,children:O,...A}=w,T=V.useRef(null),R=Re(x,T),M=a(p,C);return V.useEffect(()=>(M.itemMap.set(T,{ref:T,...A}),()=>void M.itemMap.delete(T))),E.jsx(v,{[m]:"",ref:R,children:O})});b.displayName=p;function S(w){const x=a(e+"CollectionConsumer",w);return V.useCallback(()=>{const O=x.collectionRef.current;if(!O)return[];const A=Array.from(O.querySelectorAll(`[${m}]`));return Array.from(x.itemMap.values()).sort((M,P)=>A.indexOf(M.ref.current)-A.indexOf(P.ref.current))},[x.collectionRef,x.itemMap])}return[{Provider:s,Slot:d,ItemSlot:b},S,r]}var h4=["PageUp","PageDown"],m4=["ArrowUp","ArrowDown","ArrowLeft","ArrowRight"],v4={"from-left":["Home","PageDown","ArrowDown","ArrowLeft"],"from-right":["Home","PageDown","ArrowDown","ArrowRight"],"from-bottom":["Home","PageDown","ArrowDown","ArrowLeft"],"from-top":["Home","PageDown","ArrowUp","ArrowLeft"]},_u="Slider",[SE,w0e,S0e]=x0e(_u),[g4,xxe]=jn(_u,[S0e]),[_0e,mg]=g4(_u),y4=g.forwardRef((e,t)=>{const{name:n,min:r=0,max:i=100,step:a=1,orientation:s="horizontal",disabled:u=!1,minStepsBetweenThumbs:c=0,defaultValue:d=[r],value:p,onValueChange:m=()=>{},onValueCommit:v=()=>{},inverted:b=!1,form:S,...w}=e,x=g.useRef(new Set),C=g.useRef(0),A=s==="horizontal"?E0e:C0e,[T=[],R]=Xr({prop:p,defaultProp:d,onChange:L=>{var z;(z=[...x.current][C.current])==null||z.focus(),m(L)}}),M=g.useRef(T);function P(L){const F=M0e(T,L);H(L,F)}function N(L){H(L,C.current)}function k(){const L=M.current[C.current];T[C.current]!==L&&v(T)}function H(L,F,{commit:z}={commit:!1}){const Z=D0e(a),Y=k0e(Math.round((L-r)/a)*a+r,Z),I=qh(Y,[r,i]);R((W=[])=>{const B=R0e(W,I,F);if(N0e(B,c*a)){C.current=B.indexOf(I);const Q=String(B)!==String(W);return Q&&z&&v(B),Q?B:W}else return W})}return E.jsx(_0e,{scope:e.__scopeSlider,name:n,disabled:u,min:r,max:i,valueIndexToChangeRef:C,thumbs:x.current,values:T,orientation:s,form:S,children:E.jsx(SE.Provider,{scope:e.__scopeSlider,children:E.jsx(SE.Slot,{scope:e.__scopeSlider,children:E.jsx(A,{"aria-disabled":u,"data-disabled":u?"":void 0,...w,ref:t,onPointerDown:be(w.onPointerDown,()=>{u||(M.current=T)}),min:r,max:i,inverted:b,onSlideStart:u?void 0:P,onSlideMove:u?void 0:N,onSlideEnd:u?void 0:k,onHomeKeyDown:()=>!u&&H(r,0,{commit:!0}),onEndKeyDown:()=>!u&&H(i,T.length-1,{commit:!0}),onStepKeyDown:({event:L,direction:F})=>{if(!u){const Y=h4.includes(L.key)||L.shiftKey&&m4.includes(L.key)?10:1,I=C.current,W=T[I],B=a*Y*F;H(W+B,I,{commit:!0})}}})})})})});y4.displayName=_u;var[b4,x4]=g4(_u,{startEdge:"left",endEdge:"right",size:"width",direction:1}),E0e=g.forwardRef((e,t)=>{const{min:n,max:r,dir:i,inverted:a,onSlideStart:s,onSlideMove:u,onSlideEnd:c,onStepKeyDown:d,...p}=e,[m,v]=g.useState(null),b=Re(t,A=>v(A)),S=g.useRef(void 0),w=iu(i),x=w==="ltr",C=x&&!a||!x&&a;function O(A){const T=S.current||m.getBoundingClientRect(),R=[0,T.width],P=LO(R,C?[n,r]:[r,n]);return S.current=T,P(A-T.left)}return E.jsx(b4,{scope:e.__scopeSlider,startEdge:C?"left":"right",endEdge:C?"right":"left",direction:C?1:-1,size:"width",children:E.jsx(w4,{dir:w,"data-orientation":"horizontal",...p,ref:b,style:{...p.style,"--radix-slider-thumb-transform":"translateX(-50%)"},onSlideStart:A=>{const T=O(A.clientX);s==null||s(T)},onSlideMove:A=>{const T=O(A.clientX);u==null||u(T)},onSlideEnd:()=>{S.current=void 0,c==null||c()},onStepKeyDown:A=>{const R=v4[C?"from-left":"from-right"].includes(A.key);d==null||d({event:A,direction:R?-1:1})}})})}),C0e=g.forwardRef((e,t)=>{const{min:n,max:r,inverted:i,onSlideStart:a,onSlideMove:s,onSlideEnd:u,onStepKeyDown:c,...d}=e,p=g.useRef(null),m=Re(t,p),v=g.useRef(void 0),b=!i;function S(w){const x=v.current||p.current.getBoundingClientRect(),C=[0,x.height],A=LO(C,b?[r,n]:[n,r]);return v.current=x,A(w-x.top)}return E.jsx(b4,{scope:e.__scopeSlider,startEdge:b?"bottom":"top",endEdge:b?"top":"bottom",size:"height",direction:b?1:-1,children:E.jsx(w4,{"data-orientation":"vertical",...d,ref:m,style:{...d.style,"--radix-slider-thumb-transform":"translateY(50%)"},onSlideStart:w=>{const x=S(w.clientY);a==null||a(x)},onSlideMove:w=>{const x=S(w.clientY);s==null||s(x)},onSlideEnd:()=>{v.current=void 0,u==null||u()},onStepKeyDown:w=>{const C=v4[b?"from-bottom":"from-top"].includes(w.key);c==null||c({event:w,direction:C?-1:1})}})})}),w4=g.forwardRef((e,t)=>{const{__scopeSlider:n,onSlideStart:r,onSlideMove:i,onSlideEnd:a,onHomeKeyDown:s,onEndKeyDown:u,onStepKeyDown:c,...d}=e,p=mg(_u,n);return E.jsx(Pd.span,{...d,ref:t,onKeyDown:be(e.onKeyDown,m=>{m.key==="Home"?(s(m),m.preventDefault()):m.key==="End"?(u(m),m.preventDefault()):h4.concat(m4).includes(m.key)&&(c(m),m.preventDefault())}),onPointerDown:be(e.onPointerDown,m=>{const v=m.target;v.setPointerCapture(m.pointerId),m.preventDefault(),p.thumbs.has(v)?v.focus():r(m)}),onPointerMove:be(e.onPointerMove,m=>{m.target.hasPointerCapture(m.pointerId)&&i(m)}),onPointerUp:be(e.onPointerUp,m=>{const v=m.target;v.hasPointerCapture(m.pointerId)&&(v.releasePointerCapture(m.pointerId),a(m))})})}),S4="SliderTrack",_4=g.forwardRef((e,t)=>{const{__scopeSlider:n,...r}=e,i=mg(S4,n);return E.jsx(Pd.span,{"data-disabled":i.disabled?"":void 0,"data-orientation":i.orientation,...r,ref:t})});_4.displayName=S4;var _E="SliderRange",E4=g.forwardRef((e,t)=>{const{__scopeSlider:n,...r}=e,i=mg(_E,n),a=x4(_E,n),s=g.useRef(null),u=Re(t,s),c=i.values.length,d=i.values.map(v=>A4(v,i.min,i.max)),p=c>1?Math.min(...d):0,m=100-Math.max(...d);return E.jsx(Pd.span,{"data-orientation":i.orientation,"data-disabled":i.disabled?"":void 0,...r,ref:u,style:{...e.style,[a.startEdge]:p+"%",[a.endEdge]:m+"%"}})});E4.displayName=_E;var EE="SliderThumb",C4=g.forwardRef((e,t)=>{const n=w0e(e.__scopeSlider),[r,i]=g.useState(null),a=Re(t,u=>i(u)),s=g.useMemo(()=>r?n().findIndex(u=>u.ref.current===r):-1,[n,r]);return E.jsx(O0e,{...e,ref:a,index:s})}),O0e=g.forwardRef((e,t)=>{const{__scopeSlider:n,index:r,name:i,...a}=e,s=mg(EE,n),u=x4(EE,n),[c,d]=g.useState(null),p=Re(t,O=>d(O)),m=c?s.form||!!c.closest("form"):!0,v=WE(c),b=s.values[r],S=b===void 0?0:A4(b,s.min,s.max),w=T0e(r,s.values.length),x=v==null?void 0:v[u.size],C=x?P0e(x,S,u.direction):0;return g.useEffect(()=>{if(c)return s.thumbs.add(c),()=>{s.thumbs.delete(c)}},[c,s.thumbs]),E.jsxs("span",{style:{transform:"var(--radix-slider-thumb-transform)",position:"absolute",[u.startEdge]:`calc(${S}% + ${C}px)`},children:[E.jsx(SE.ItemSlot,{scope:e.__scopeSlider,children:E.jsx(Pd.span,{role:"slider","aria-label":e["aria-label"]||w,"aria-valuemin":s.min,"aria-valuenow":b,"aria-valuemax":s.max,"aria-orientation":s.orientation,"data-orientation":s.orientation,"data-disabled":s.disabled?"":void 0,tabIndex:s.disabled?void 0:0,...a,ref:p,style:b===void 0?{display:"none"}:e.style,onFocus:be(e.onFocus,()=>{s.valueIndexToChangeRef.current=r})})}),m&&E.jsx(O4,{name:i??(s.name?s.name+(s.values.length>1?"[]":""):void 0),form:s.form,value:b},r)]})});C4.displayName=EE;var A0e="RadioBubbleInput",O4=g.forwardRef(({__scopeSlider:e,value:t,...n},r)=>{const i=g.useRef(null),a=Re(i,r),s=Ym(t);return g.useEffect(()=>{const u=i.current;if(!u)return;const c=window.HTMLInputElement.prototype,p=Object.getOwnPropertyDescriptor(c,"value").set;if(s!==t&&p){const m=new Event("input",{bubbles:!0});p.call(u,t),u.dispatchEvent(m)}},[s,t]),E.jsx(Pd.input,{style:{display:"none"},...n,ref:a,defaultValue:t})});O4.displayName=A0e;function R0e(e=[],t,n){const r=[...e];return r[n]=t,r.sort((i,a)=>i-a)}function A4(e,t,n){const a=100/(n-t)*(e-t);return qh(a,[0,100])}function T0e(e,t){return t>2?`Value ${e+1} of ${t}`:t===2?["Minimum","Maximum"][e]:void 0}function M0e(e,t){if(e.length===1)return 0;const n=e.map(i=>Math.abs(i-t)),r=Math.min(...n);return n.indexOf(r)}function P0e(e,t,n){const r=e/2,a=LO([0,50],[0,r]);return(r-a(t)*n)*n}function j0e(e){return e.slice(0,-1).map((t,n)=>e[n+1]-t)}function N0e(e,t){if(t>0){const n=j0e(e);return Math.min(...n)>=t}return!0}function LO(e,t){return n=>{if(e[0]===e[1]||t[0]===t[1])return t[0];const r=(t[1]-t[0])/(e[1]-e[0]);return t[0]+r*(n-e[0])}}function D0e(e){return(String(e).split(".")[1]||"").length}function k0e(e,t){const n=Math.pow(10,t);return Math.round(e*n)/n}var $0e=y4,I0e=_4,L0e=E4,B0e=C4;function F0e({className:e,defaultValue:t,value:n,min:r=0,max:i=100,...a}){const s=g.useMemo(()=>Array.isArray(n)?n:Array.isArray(t)?t:[r,i],[n,t,r,i]);return E.jsxs($0e,{"data-slot":"slider",defaultValue:t,value:n,min:r,max:i,className:ke("relative flex w-full touch-none items-center select-none data-[disabled]:opacity-50 data-[orientation=vertical]:h-full data-[orientation=vertical]:min-h-44 data-[orientation=vertical]:w-auto data-[orientation=vertical]:flex-col",e),...a,children:[E.jsx(I0e,{"data-slot":"slider-track",className:ke("bg-muted relative grow overflow-hidden rounded-full data-[orientation=horizontal]:h-1.5 data-[orientation=horizontal]:w-full data-[orientation=vertical]:h-full data-[orientation=vertical]:w-1.5"),children:E.jsx(L0e,{"data-slot":"slider-range",className:ke("bg-primary absolute data-[orientation=horizontal]:h-full data-[orientation=vertical]:w-full")})}),Array.from({length:s.length},(u,c)=>E.jsx(B0e,{"data-slot":"slider-thumb",className:"border-primary bg-background ring-ring/50 block size-4 shrink-0 rounded-full border shadow-sm transition-[color,box-shadow] hover:ring-4 focus-visible:ring-4 focus-visible:outline-hidden disabled:pointer-events-none disabled:opacity-50"},c))]})}function q0e({triggerButtonText:e,value:t,onValueChange:n,max:r,step:i,disabled:a=!1}){return E.jsxs(f4,{children:[E.jsx(d4,{asChild:!0,children:E.jsxs(Pr,{variant:"ghost",disabled:a,children:[e,E.jsx(Zm,{})]})}),E.jsx(p4,{className:"w-56",children:E.jsx(F0e,{value:[t],onValueChange:n,max:r,step:i,className:"my-3"})})]})}const z0e=()=>{const[e,t]=g.useState([]),n=(i,a)=>{t(e.map(s=>s.id===i?{...s,checked:a}:s))};return{checkboxItems:e.map(i=>({...i,onCheckedChange:a=>n(i.id,a)})),setItems:t}};function U0e(){const{checkboxItems:e,setItems:t}=z0e(),{classification_uuid:n}=$E(),[r,i]=g.useState(null),[a,s]=g.useState(null),[u,c]=g.useState(null),[d,p]=g.useState(15),m=JSON.stringify(e.map(v=>({id:v.id,checked:v.checked})));return g.useEffect(()=>{const v=()=>{n&&NQ(n).then(b=>{i(b),fC(b.model_slug).then(S=>{c(S)}).catch(S=>{console.error("Error fetching model metadata:",S),setTimeout(v,500)})}).catch(b=>{console.error("Error fetching classification result:",b),setTimeout(v,500)})};v()},[n,i,c]),g.useEffect(()=>{if(r){const v=Object.keys(r.scores);if(v.length<=20){const b=v.filter(S=>S!=="total").map(S=>({id:S,label:S,checked:!0}));t(b)}}},[r,t]),g.useEffect(()=>{if(r&&u){const v=e.filter(O=>O.checked).map(O=>O.id);console.log("selectedContigs",v);const b=Object.entries(r.hits).filter(([O])=>v.includes(O)),S={};b.forEach(([,O])=>{Object.entries(O).forEach(([A,T])=>{S[A]||(S[A]=0),S[A]+=T})});const w=Object.entries(r.num_kmers).filter(([O])=>v.includes(O)).reduce((O,[A,T])=>O+T,0),x=Object.entries(S).map(([O,A])=>({taxon:u.display_names[O].replace(u.model_display_name,u.model_display_name.charAt(0)+".")||O,score:A/w}));x.sort((O,A)=>A.score-O.score);const C=x.slice(0,d);s(C)}},[r,u,m,d]),E.jsx("main",{className:"flex-1 flex flex-col items-center justify-center p-4",children:E.jsxs("div",{className:"w-1/2",children:[!a&&E.jsxs("div",{className:"flex items-center justify-center",children:[E.jsx(Rz,{className:"size-10"}),E.jsx("span",{className:"m-2 text-xl",children:"Classifying..."})]}),a&&E.jsxs(E.Fragment,{children:[E.jsx("h1",{className:"text-xl font-bold",children:"Classification Results"}),E.jsxs("p",{children:[r==null?void 0:r.input_source," was classified using the ",E.jsxs(aa,{className:"font-medium underline underline-offset-4",to:`/models/${u==null?void 0:u.model_slug}`,children:[u==null?void 0:u.model_display_name," ",u==null?void 0:u.model_type.toLowerCase()]})," model. ",(r==null?void 0:r.sparse_sampling_step)===1?"No sparse sampling was used.":"Sparse sampling was used with step "+(r==null?void 0:r.sparse_sampling_step)+"."]}),E.jsxs("p",{className:"mt-4",children:[E.jsx("b",{children:"Prediction:"})," ",(u==null?void 0:u.display_names[r==null?void 0:r.prediction])||"No prediction available."]}),E.jsx(Pr,{className:"w-full mt-4",onClick:()=>{const v=new Blob([JSON.stringify(r,null,2)],{type:"application/json"}),b=URL.createObjectURL(v),S=document.createElement("a");S.href=b,S.download="result-"+n+".json",document.body.appendChild(S),S.click(),document.body.removeChild(S),URL.revokeObjectURL(b)},children:"Download Full Result"}),E.jsx(Mr,{className:"my-4 h-px"}),E.jsxs("div",{className:"flex justify-end space-x-2",children:[E.jsx(q0e,{triggerButtonText:"#Results",value:d,onValueChange:p,max:u==null?void 0:u.display_names.length,step:1,disabled:!1}),E.jsx(p0e,{triggerButtonText:"Contigs",labelText:"Select Contigs",items:e,disabled:e.length===0})]}),E.jsx(_ye,{data:a})]})]})})}function V0e(){const{filter_uuid:e}=$E(),[t,n]=g.useState(null);return g.useEffect(()=>{const r=()=>{e&&kQ(e).then(i=>{n(i)}).catch(i=>{console.error("Error fetching filtering result:",i),setTimeout(r,500)})};r()},[e,n]),E.jsx("main",{className:"flex-1 flex flex-col items-center justify-center p-4",children:E.jsxs("div",{className:"w-1/2",children:[!t&&E.jsxs("div",{className:"flex items-center justify-center",children:[E.jsx(Rz,{className:"size-10"}),E.jsx("span",{className:"m-2 text-xl",children:"Filtering..."})]}),t&&E.jsxs(E.Fragment,{children:[E.jsx("h1",{className:"text-xl font-bold",children:"Filtering Results"}),E.jsx("p",{children:"Your file was filtered successfully."}),t.message!="Filtering completed successfully."&&E.jsxs("div",{className:"flex items-center gap-2 text-red-500",children:[E.jsx(_8,{className:"w-5 h-5"}),E.jsx("p",{children:"All sequences were filtered out."})]}),E.jsx(Mr,{className:"my-4 h-px"}),E.jsxs("div",{className:"flex flex-col gap-3 mt-4",children:[t.message=="Filtering completed successfully."&&E.jsx(Pr,{asChild:!0,children:E.jsx("a",{href:`${window.location.origin}/api/download-filtered?uuid=${e}`,download:!0,children:"Download the filtered file"})}),E.jsx(Pr,{variant:"outline",asChild:!0,children:E.jsx(aa,{to:`/result/${e}`,children:"View underlying classification results"})})]})]})]})})}function H0e(){const{model_slug:e}=$E(),[t,n]=g.useState(null);return g.useEffect(()=>{e&&fC(e).then(r=>{n(r)}).catch(r=>console.error("Fetch error:",r))},[e]),E.jsx("main",{className:"flex-1 flex flex-col items-center justify-center p-4",children:E.jsxs("div",{className:"w-1/2",children:[t&&E.jsxs(E.Fragment,{children:[E.jsxs("h1",{className:"text-xl font-bold",children:[t.model_display_name," ",t.model_type," Model"]}),E.jsxs("div",{className:"flex h-5 items-center space-x-4",children:[E.jsxs("div",{children:["k = ",t.k]}),E.jsx(Mr,{orientation:"vertical"}),E.jsxs("div",{children:["num_hashes = ",t.num_hashes]}),E.jsx(Mr,{orientation:"vertical"}),E.jsxs("div",{children:["fpr = ",t.fpr]}),t.model_type==="Species"&&E.jsxs(E.Fragment,{children:[E.jsx(Mr,{orientation:"vertical"}),E.jsxs("div",{children:["kernel = ",t.kernel]}),E.jsx(Mr,{orientation:"vertical"}),E.jsxs("div",{children:["C = ",t.C]})]})]}),E.jsx(Mr,{className:"my-4"}),E.jsx("h2",{className:"text-l font-bold",children:"Author"}),E.jsxs("div",{className:"flex flex-col gap-2",children:[E.jsxs("p",{children:["Model author: ",t.author==null?"N/A":t.author]}),E.jsxs("p",{children:["Model author email: ",t.author_email==null?"N/A":E.jsx(aa,{to:`mailto:${t.author_email}`,className:"font-medium underline underline-offset-4",children:t.author_email})]})]}),E.jsx(Mr,{className:"my-4"}),E.jsx("h2",{className:"text-l font-bold mb-2",children:"Model display names"}),E.jsx(i1,{columns:[{accessorKey:"key",header:"ID"},{accessorKey:"value",header:"Display name"}],data:Object.entries(t.display_names).map(([r,i])=>({key:r,value:i}))}),E.jsx(Mr,{className:"my-4"}),E.jsx("h2",{className:"text-l font-bold mb-2",children:"K-mer Index Training Data"}),typeof t.training_accessions=="object"&&Array.isArray(t.training_accessions)&&E.jsxs("p",{children:["The model was trained on the following accessions:",E.jsx("ul",{className:"list-disc list-inside",children:t.training_accessions.map((r,i)=>E.jsx("li",{children:E.jsx(aa,{to:`https://www.ncbi.nlm.nih.gov/datasets/genome/${r}/`,target:"_blank",rel:"noopener noreferrer",className:"font-medium underline underline-offset-4",children:r})},i))})]}),typeof t.training_accessions=="object"&&!Array.isArray(t.training_accessions)&&E.jsx(i1,{columns:[{accessorKey:"key",header:"Name"},{accessorKey:"value",header:"Accessions",cell:({getValue:r})=>{const i=r();return Array.isArray(i)?i.map((a,s)=>E.jsxs(E.Fragment,{children:[s>0&&", ",E.jsx(aa,{to:`https://www.ncbi.nlm.nih.gov/datasets/genome/${a}/`,target:"_blank",rel:"noopener noreferrer",className:"font-medium underline underline-offset-4",children:a})]})):i}}],data:Object.entries(t.training_accessions).map(([r,i])=>({key:t.display_names[r],value:i}))})]}),!t&&E.jsxs(E.Fragment,{children:[E.jsx("h1",{className:"text-xl font-bold",children:"Error"}),E.jsxs("p",{children:["No model could be found with slug ",E.jsx("span",{className:"font-mono",children:e}),"."]})]})]})})}function G0e(){return E.jsx(_G,{basename:"/xspect-web",children:E.jsxs("div",{className:"min-h-screen flex flex-col",children:[E.jsx(jG,{}),E.jsxs(J9,{children:[E.jsx(Ya,{path:"/classify",element:E.jsx(nJ,{})}),E.jsx(Ya,{path:"/filter",element:E.jsx(wte,{})}),E.jsx(Ya,{path:"/models",element:E.jsx(gne,{})}),E.jsx(Ya,{path:"/models/:model_slug",element:E.jsx(H0e,{})}),E.jsx(Ya,{path:"/result/:classification_uuid",element:E.jsx(U0e,{})}),E.jsx(Ya,{path:"/filter-result/:filter_uuid",element:E.jsx(V0e,{})}),E.jsx(Ya,{path:"*",element:E.jsx(NG,{})})]})]})})}Y5.createRoot(document.getElementById("root")).render(E.jsx(g.StrictMode,{children:E.jsx(G0e,{})}));
